define(["base/js/namespace","base/js/dialog"],(function(e,t){return function(e){var t={};function n(i){if(t[i])return t[i].exports;var r=t[i]={i:i,l:!1,exports:{}};return e[i].call(r.exports,r,r.exports,n),r.l=!0,r.exports}return n.m=e,n.c=t,n.d=function(e,t,i){n.o(e,t)||Object.defineProperty(e,t,{enumerable:!0,get:i})},n.r=function(e){"undefined"!=typeof Symbol&&Symbol.toStringTag&&Object.defineProperty(e,Symbol.toStringTag,{value:"Module"}),Object.defineProperty(e,"__esModule",{value:!0})},n.t=function(e,t){if(1&t&&(e=n(e)),8&t)return e;if(4&t&&"object"==typeof e&&e&&e.__esModule)return e;var i=Object.create(null);if(n.r(i),Object.defineProperty(i,"default",{enumerable:!0,value:e}),2&t&&"string"!=typeof e)for(var r in e)n.d(i,r,function(t){return e[t]}.bind(null,r));return i},n.n=function(e){var t=e&&e.__esModule?function(){return e.default}:function(){return e};return n.d(t,"a",t),t},n.o=function(e,t){return Object.prototype.hasOwnProperty.call(e,t)},n.p="",n(n.s=150)}([function(e,t,n){"use strict";e.exports=n(64)},function(e,t,n){e.exports=n(61)},function(e,t,n){"use strict";(function(e,i){n.d(t,"a",(function(){return x})),n.d(t,"b",(function(){return Pe})),n.d(t,"c",(function(){return be})),n.d(t,"d",(function(){return Qe})),n.d(t,"e",(function(){return te})),n.d(t,"f",(function(){return tt})),n.d(t,"g",(function(){return _})),n.d(t,"h",(function(){return ot})),n.d(t,"i",(function(){return _t})),n.d(t,"j",(function(){return Lt})),n.d(t,"k",(function(){return jt})),n.d(t,"l",(function(){return Y})),n.d(t,"m",(function(){return qe})),n.d(t,"n",(function(){return He})),n.d(t,"o",(function(){return ht})),n.d(t,"p",(function(){return le}));
/*! *****************************************************************************
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License"); you may not use
this file except in compliance with the License. You may obtain a copy of the
License at http://www.apache.org/licenses/LICENSE-2.0

THIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED
WARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,
MERCHANTABLITY OR NON-INFRINGEMENT.

See the Apache Version 2.0 License for specific language governing permissions
and limitations under the License.
***************************************************************************** */
var r=function(e,t){return(r=Object.setPrototypeOf||{__proto__:[]}instanceof Array&&function(e,t){e.__proto__=t}||function(e,t){for(var n in t)t.hasOwnProperty(n)&&(e[n]=t[n])})(e,t)};var o=function(){return(o=Object.assign||function(e){for(var t,n=1,i=arguments.length;n<i;n++)for(var r in t=arguments[n])Object.prototype.hasOwnProperty.call(t,r)&&(e[r]=t[r]);return e}).apply(this,arguments)};function s(e){var t="function"==typeof Symbol&&e[Symbol.iterator],n=0;return t?t.call(e):{next:function(){return e&&n>=e.length&&(e=void 0),{value:e&&e[n++],done:!e}}}}function a(e,t){var n="function"==typeof Symbol&&e[Symbol.iterator];if(!n)return e;var i,r,o=n.call(e),s=[];try{for(;(void 0===t||t-- >0)&&!(i=o.next()).done;)s.push(i.value)}catch(e){r={error:e}}finally{try{i&&!i.done&&(n=o.return)&&n.call(o)}finally{if(r)throw r.error}}return s}function l(){for(var e=[],t=0;t<arguments.length;t++)e=e.concat(a(arguments[t]));return e}var u=[];Object.freeze(u);var c={};function h(){return++De.mobxGuid}function f(e){throw d(!1,e),"X"}function d(e,t){if(!e)throw new Error("[mobx] "+(t||"An invariant failed, however the error is obfuscated because this is a production build."))}Object.freeze(c);function p(e){var t=!1;return function(){if(!t)return t=!0,e.apply(this,arguments)}}var g=function(){};function m(e){return null!==e&&"object"==typeof e}function v(e){if(null===e||"object"!=typeof e)return!1;var t=Object.getPrototypeOf(e);return t===Object.prototype||null===t}function b(e,t,n){Object.defineProperty(e,t,{enumerable:!1,writable:!0,configurable:!0,value:n})}function y(e,t){var n="isMobX"+e;return t.prototype[n]=!0,function(e){return m(e)&&!0===e[n]}}function w(e){return e instanceof Map}function A(e){return e instanceof Set}function C(e){var t=new Set;for(var n in e)t.add(n);return Object.getOwnPropertySymbols(e).forEach((function(n){Object.getOwnPropertyDescriptor(e,n).enumerable&&t.add(n)})),Array.from(t)}function E(e){return e&&e.toString?e.toString():new String(e).toString()}function F(e){return null===e?null:"object"==typeof e?""+e:e}var x=Symbol("mobx administration"),k=function(){function e(e){void 0===e&&(e="Atom@"+h()),this.name=e,this.isPendingUnobservation=!1,this.isBeingObserved=!1,this.observers=new Set,this.diffValue=0,this.lastAccessedBy=0,this.lowestObserverState=X.NOT_TRACKING}return e.prototype.onBecomeObserved=function(){this.onBecomeObservedListeners&&this.onBecomeObservedListeners.forEach((function(e){return e()}))},e.prototype.onBecomeUnobserved=function(){this.onBecomeUnobservedListeners&&this.onBecomeUnobservedListeners.forEach((function(e){return e()}))},e.prototype.reportObserved=function(){return Oe(this)},e.prototype.reportChanged=function(){Te(),function(e){if(e.lowestObserverState===X.STALE)return;e.lowestObserverState=X.STALE,e.observers.forEach((function(t){t.dependenciesState===X.UP_TO_DATE&&(t.isTracing!==J.NONE&&Me(t,e),t.onBecomeStale()),t.dependenciesState=X.STALE}))}(this),Re()},e.prototype.toString=function(){return this.name},e}(),S=y("Atom",k);function _(e,t,n){void 0===t&&(t=g),void 0===n&&(n=g);var i,r=new k(e);return t!==g&&et("onBecomeObserved",r,t,i),n!==g&&Ze(r,n),r}var D={identity:function(e,t){return e===t},structural:function(e,t){return Kt(e,t)},default:function(e,t){return Object.is(e,t)},shallow:function(e,t){return Kt(e,t,1)}},$=Symbol("mobx did run lazy initializers"),B=Symbol("mobx pending decorators"),L={},T={};function R(e,t){var n=t?L:T;return n[e]||(n[e]={configurable:!0,enumerable:t,get:function(){return O(this),this[e]},set:function(t){O(this),this[e]=t}})}function O(e){var t,n;if(!0!==e[$]){var i=e[B];if(i){b(e,$,!0);var r=l(Object.getOwnPropertySymbols(i),Object.keys(i));try{for(var o=s(r),a=o.next();!a.done;a=o.next()){var u=i[a.value];u.propertyCreator(e,u.prop,u.descriptor,u.decoratorTarget,u.decoratorArguments)}}catch(e){t={error:e}}finally{try{a&&!a.done&&(n=o.return)&&n.call(o)}finally{if(t)throw t.error}}}}}function M(e,t){return function(){var n,i=function(i,r,s,a){if(!0===a)return t(i,r,s,i,n),null;if(!Object.prototype.hasOwnProperty.call(i,B)){var l=i[B];b(i,B,o({},l))}return i[B][r]={prop:r,propertyCreator:t,descriptor:s,decoratorTarget:i,decoratorArguments:n},R(r,e)};return P(arguments)?(n=u,i.apply(null,arguments)):(n=Array.prototype.slice.call(arguments),i)}}function P(e){return(2===e.length||3===e.length)&&("string"==typeof e[1]||"symbol"==typeof e[1])||4===e.length&&!0===e[3]}function I(e,t,n){return ut(e)?e:Array.isArray(e)?Y.array(e,{name:n}):v(e)?Y.object(e,void 0,{name:n}):w(e)?Y.map(e,{name:n}):A(e)?Y.set(e,{name:n}):e}function N(e){return e}function W(t){d(t);var n=M(!0,(function(e,n,i,r,o){var s=i?i.initializer?i.initializer.call(e):i.value:void 0;Pt(e).addObservableProp(n,s,t)})),i=(void 0!==e&&e.env,n);return i.enhancer=t,i}var z={deep:!0,name:void 0,defaultDecorator:void 0,proxy:!0};function j(e){return null==e?z:"string"==typeof e?{name:e,deep:!0,proxy:!0}:e}Object.freeze(z);var H=W(I),V=W((function(e,t,n){return null==e?e:jt(e)||_t(e)||Lt(e)||Ot(e)?e:Array.isArray(e)?Y.array(e,{name:n,deep:!1}):v(e)?Y.object(e,void 0,{name:n,deep:!1}):w(e)?Y.map(e,{name:n,deep:!1}):A(e)?Y.set(e,{name:n,deep:!1}):f(!1)})),U=W(N),K=W((function(e,t,n){return Kt(e,t)?t:e}));function G(e){return e.defaultDecorator?e.defaultDecorator.enhancer:!1===e.deep?N:I}var q={box:function(e,t){arguments.length>2&&Q("box");var n=j(t);return new Ae(e,G(n),n.name,!0,n.equals)},array:function(e,t){arguments.length>2&&Q("array");var n=j(t);return Et(e,G(n),n.name)},map:function(e,t){arguments.length>2&&Q("map");var n=j(t);return new Bt(e,G(n),n.name)},set:function(e,t){arguments.length>2&&Q("set");var n=j(t);return new Rt(e,G(n),n.name)},object:function(e,t,n){"string"==typeof arguments[1]&&Q("object");var i=j(n);if(!1===i.proxy)return nt({},e,t,i);var r=it(i),o=nt({},void 0,void 0,i),s=gt(o);return rt(s,e,t,r),s},ref:U,shallow:V,deep:H,struct:K},Y=function(e,t,n){if("string"==typeof arguments[1]||"symbol"==typeof arguments[1])return H.apply(null,arguments);if(ut(e))return e;var i=v(e)?Y.object(e,t,n):Array.isArray(e)?Y.array(e,t):w(e)?Y.map(e,t):A(e)?Y.set(e,t):e;if(i!==e)return i;f(!1)};function Q(e){f("Expected one or two arguments to observable."+e+". Did you accidentally try to use observable."+e+" as decorator?")}Object.keys(q).forEach((function(e){return Y[e]=q[e]}));var X,J,Z=M(!1,(function(e,t,n,i,r){var s=n.get,a=n.set,l=r[0]||{};Pt(e).addComputedProp(e,t,o({get:s,set:a,context:e},l))})),ee=Z({equals:D.structural}),te=function(e,t,n){if("string"==typeof t)return Z.apply(null,arguments);if(null!==e&&"object"==typeof e&&1===arguments.length)return Z.apply(null,arguments);var i="object"==typeof t?t:{};return i.get=e,i.set="function"==typeof t?t:i.set,i.name=i.name||e.name||"",new Ce(i)};te.struct=ee,function(e){e[e.NOT_TRACKING=-1]="NOT_TRACKING",e[e.UP_TO_DATE=0]="UP_TO_DATE",e[e.POSSIBLY_STALE=1]="POSSIBLY_STALE",e[e.STALE=2]="STALE"}(X||(X={})),function(e){e[e.NONE=0]="NONE",e[e.LOG=1]="LOG",e[e.BREAK=2]="BREAK"}(J||(J={}));var ne=function(e){this.cause=e};function ie(e){return e instanceof ne}function re(e){switch(e.dependenciesState){case X.UP_TO_DATE:return!1;case X.NOT_TRACKING:case X.STALE:return!0;case X.POSSIBLY_STALE:for(var t=he(!0),n=ue(),i=e.observing,r=i.length,o=0;o<r;o++){var s=i[o];if(Ee(s)){if(De.disableErrorBoundaries)s.get();else try{s.get()}catch(e){return ce(n),fe(t),!0}if(e.dependenciesState===X.STALE)return ce(n),fe(t),!0}}return de(e),ce(n),fe(t),!1}}function oe(e){var t=e.observers.size>0;De.computationDepth>0&&t&&f(!1),De.allowStateChanges||!t&&"strict"!==De.enforceActions||f(!1)}function se(e,t,n){var i=he(!0);de(e),e.newObserving=new Array(e.observing.length+100),e.unboundDepsCount=0,e.runId=++De.runId;var r,o=De.trackingDerivation;if(De.trackingDerivation=e,!0===De.disableErrorBoundaries)r=t.call(n);else try{r=t.call(n)}catch(e){r=new ne(e)}return De.trackingDerivation=o,function(e){for(var t=e.observing,n=e.observing=e.newObserving,i=X.UP_TO_DATE,r=0,o=e.unboundDepsCount,s=0;s<o;s++){0===(a=n[s]).diffValue&&(a.diffValue=1,r!==s&&(n[r]=a),r++),a.dependenciesState>i&&(i=a.dependenciesState)}n.length=r,e.newObserving=null,o=t.length;for(;o--;){0===(a=t[o]).diffValue&&Be(a,e),a.diffValue=0}for(;r--;){var a;1===(a=n[r]).diffValue&&(a.diffValue=0,$e(a,e))}i!==X.UP_TO_DATE&&(e.dependenciesState=i,e.onBecomeStale())}(e),fe(i),r}function ae(e){var t=e.observing;e.observing=[];for(var n=t.length;n--;)Be(t[n],e);e.dependenciesState=X.NOT_TRACKING}function le(e){var t=ue();try{return e()}finally{ce(t)}}function ue(){var e=De.trackingDerivation;return De.trackingDerivation=null,e}function ce(e){De.trackingDerivation=e}function he(e){var t=De.allowStateReads;return De.allowStateReads=e,t}function fe(e){De.allowStateReads=e}function de(e){if(e.dependenciesState!==X.UP_TO_DATE){e.dependenciesState=X.UP_TO_DATE;for(var t=e.observing,n=t.length;n--;)t[n].lowestObserverState=X.UP_TO_DATE}}var pe=0,ge=1;function me(e,t,n){var i=function(){return ve(e,t,n||this,arguments)};return i.isMobxAction=!0,i}function ve(e,t,n,i){var r=function(e,t,n){var i=0;var r=ue();Te();var o=ye(!0),s=he(!0),a={prevDerivation:r,prevAllowStateChanges:o,prevAllowStateReads:s,notifySpy:!1,startTime:i,actionId:ge++,parentActionId:pe};return pe=a.actionId,a}();try{return t.apply(n,i)}catch(e){throw r.error=e,e}finally{!function(e){pe!==e.actionId&&f("invalid action stack. did you forget to finish an action?");pe=e.parentActionId,void 0!==e.error&&(De.suppressReactionErrors=!0);we(e.prevAllowStateChanges),fe(e.prevAllowStateReads),Re(),ce(e.prevDerivation),e.notifySpy&&!1;De.suppressReactionErrors=!1}(r)}}function be(e,t){var n,i=ye(e);try{n=t()}finally{we(i)}return n}function ye(e){var t=De.allowStateChanges;return De.allowStateChanges=e,t}function we(e){De.allowStateChanges=e}var Ae=function(e){function t(t,n,i,r,o){void 0===i&&(i="ObservableValue@"+h()),void 0===r&&(r=!0),void 0===o&&(o=D.default);var s=e.call(this,i)||this;return s.enhancer=n,s.name=i,s.equals=o,s.hasUnreportedChange=!1,s.value=n(t,void 0,i),s}return function(e,t){function n(){this.constructor=e}r(e,t),e.prototype=null===t?Object.create(t):(n.prototype=t.prototype,new n)}(t,e),t.prototype.dehanceValue=function(e){return void 0!==this.dehancer?this.dehancer(e):e},t.prototype.set=function(e){this.value;if((e=this.prepareNewValue(e))!==De.UNCHANGED){0,this.setNewValue(e)}},t.prototype.prepareNewValue=function(e){if(oe(this),mt(this)){var t=bt(this,{object:this,type:"update",newValue:e});if(!t)return De.UNCHANGED;e=t.newValue}return e=this.enhancer(e,this.value,this.name),this.equals(this.value,e)?De.UNCHANGED:e},t.prototype.setNewValue=function(e){var t=this.value;this.value=e,this.reportChanged(),yt(this)&&At(this,{type:"update",object:this,newValue:e,oldValue:t})},t.prototype.get=function(){return this.reportObserved(),this.dehanceValue(this.value)},t.prototype.intercept=function(e){return vt(this,e)},t.prototype.observe=function(e,t){return t&&e({object:this,type:"update",newValue:this.value,oldValue:void 0}),wt(this,e)},t.prototype.toJSON=function(){return this.get()},t.prototype.toString=function(){return this.name+"["+this.value+"]"},t.prototype.valueOf=function(){return F(this.get())},t.prototype[Symbol.toPrimitive]=function(){return this.valueOf()},t}(k),Ce=(y("ObservableValue",Ae),function(){function e(e){this.dependenciesState=X.NOT_TRACKING,this.observing=[],this.newObserving=null,this.isBeingObserved=!1,this.isPendingUnobservation=!1,this.observers=new Set,this.diffValue=0,this.runId=0,this.lastAccessedBy=0,this.lowestObserverState=X.UP_TO_DATE,this.unboundDepsCount=0,this.__mapid="#"+h(),this.value=new ne(null),this.isComputing=!1,this.isRunningSetter=!1,this.isTracing=J.NONE,this.derivation=e.get,this.name=e.name||"ComputedValue@"+h(),e.set&&(this.setter=me(this.name+"-setter",e.set)),this.equals=e.equals||(e.compareStructural||e.struct?D.structural:D.default),this.scope=e.context,this.requiresReaction=!!e.requiresReaction,this.keepAlive=!!e.keepAlive}return e.prototype.onBecomeStale=function(){!function(e){if(e.lowestObserverState!==X.UP_TO_DATE)return;e.lowestObserverState=X.POSSIBLY_STALE,e.observers.forEach((function(t){t.dependenciesState===X.UP_TO_DATE&&(t.dependenciesState=X.POSSIBLY_STALE,t.isTracing!==J.NONE&&Me(t,e),t.onBecomeStale())}))}(this)},e.prototype.onBecomeObserved=function(){this.onBecomeObservedListeners&&this.onBecomeObservedListeners.forEach((function(e){return e()}))},e.prototype.onBecomeUnobserved=function(){this.onBecomeUnobservedListeners&&this.onBecomeUnobservedListeners.forEach((function(e){return e()}))},e.prototype.get=function(){this.isComputing&&f("Cycle detected in computation "+this.name+": "+this.derivation),0!==De.inBatch||0!==this.observers.size||this.keepAlive?(Oe(this),re(this)&&this.trackAndCompute()&&function(e){if(e.lowestObserverState===X.STALE)return;e.lowestObserverState=X.STALE,e.observers.forEach((function(t){t.dependenciesState===X.POSSIBLY_STALE?t.dependenciesState=X.STALE:t.dependenciesState===X.UP_TO_DATE&&(e.lowestObserverState=X.UP_TO_DATE)}))}(this)):re(this)&&(this.warnAboutUntrackedRead(),Te(),this.value=this.computeValue(!1),Re());var e=this.value;if(ie(e))throw e.cause;return e},e.prototype.peek=function(){var e=this.computeValue(!1);if(ie(e))throw e.cause;return e},e.prototype.set=function(e){if(this.setter){d(!this.isRunningSetter,"The setter of computed value '"+this.name+"' is trying to update itself. Did you intend to update an _observable_ value, instead of the computed property?"),this.isRunningSetter=!0;try{this.setter.call(this.scope,e)}finally{this.isRunningSetter=!1}}else d(!1,!1)},e.prototype.trackAndCompute=function(){var e=this.value,t=this.dependenciesState===X.NOT_TRACKING,n=this.computeValue(!0),i=t||ie(e)||ie(n)||!this.equals(e,n);return i&&(this.value=n),i},e.prototype.computeValue=function(e){var t;if(this.isComputing=!0,De.computationDepth++,e)t=se(this,this.derivation,this.scope);else if(!0===De.disableErrorBoundaries)t=this.derivation.call(this.scope);else try{t=this.derivation.call(this.scope)}catch(e){t=new ne(e)}return De.computationDepth--,this.isComputing=!1,t},e.prototype.suspend=function(){this.keepAlive||(ae(this),this.value=void 0)},e.prototype.observe=function(e,t){var n=this,i=!0,r=void 0;return Qe((function(){var o=n.get();if(!i||t){var s=ue();e({type:"update",object:n,newValue:o,oldValue:r}),ce(s)}i=!1,r=o}))},e.prototype.warnAboutUntrackedRead=function(){},e.prototype.toJSON=function(){return this.get()},e.prototype.toString=function(){return this.name+"["+this.derivation.toString()+"]"},e.prototype.valueOf=function(){return F(this.get())},e.prototype[Symbol.toPrimitive]=function(){return this.valueOf()},e}()),Ee=y("ComputedValue",Ce),Fe=function(){this.version=5,this.UNCHANGED={},this.trackingDerivation=null,this.computationDepth=0,this.runId=0,this.mobxGuid=0,this.inBatch=0,this.pendingUnobservations=[],this.pendingReactions=[],this.isRunningReactions=!1,this.allowStateChanges=!0,this.allowStateReads=!0,this.enforceActions=!1,this.spyListeners=[],this.globalReactionErrorHandlers=[],this.computedRequiresReaction=!1,this.reactionRequiresObservable=!1,this.observableRequiresReaction=!1,this.computedConfigurable=!1,this.disableErrorBoundaries=!1,this.suppressReactionErrors=!1},xe={};function ke(){return"undefined"!=typeof window?window:void 0!==i?i:"undefined"!=typeof self?self:xe}var Se=!0,_e=!1,De=function(){var e=ke();return e.__mobxInstanceCount>0&&!e.__mobxGlobals&&(Se=!1),e.__mobxGlobals&&e.__mobxGlobals.version!==(new Fe).version&&(Se=!1),Se?e.__mobxGlobals?(e.__mobxInstanceCount+=1,e.__mobxGlobals.UNCHANGED||(e.__mobxGlobals.UNCHANGED={}),e.__mobxGlobals):(e.__mobxInstanceCount=1,e.__mobxGlobals=new Fe):(setTimeout((function(){_e||f("There are multiple, different versions of MobX active. Make sure MobX is loaded only once or use `configure({ isolateGlobalState: true })`")}),1),new Fe)}();function $e(e,t){e.observers.add(t),e.lowestObserverState>t.dependenciesState&&(e.lowestObserverState=t.dependenciesState)}function Be(e,t){e.observers.delete(t),0===e.observers.size&&Le(e)}function Le(e){!1===e.isPendingUnobservation&&(e.isPendingUnobservation=!0,De.pendingUnobservations.push(e))}function Te(){De.inBatch++}function Re(){if(0==--De.inBatch){Ne();for(var e=De.pendingUnobservations,t=0;t<e.length;t++){var n=e[t];n.isPendingUnobservation=!1,0===n.observers.size&&(n.isBeingObserved&&(n.isBeingObserved=!1,n.onBecomeUnobserved()),n instanceof Ce&&n.suspend())}De.pendingUnobservations=[]}}function Oe(e){var t=De.trackingDerivation;return null!==t?(t.runId!==e.lastAccessedBy&&(e.lastAccessedBy=t.runId,t.newObserving[t.unboundDepsCount++]=e,e.isBeingObserved||(e.isBeingObserved=!0,e.onBecomeObserved())),!0):(0===e.observers.size&&De.inBatch>0&&Le(e),!1)}function Me(e,t){if(console.log("[mobx.trace] '"+e.name+"' is invalidated due to a change in: '"+t.name+"'"),e.isTracing===J.BREAK){var n=[];!function e(t,n,i){if(n.length>=1e3)return void n.push("(and many more)");n.push(""+new Array(i).join("\t")+t.name),t.dependencies&&t.dependencies.forEach((function(t){return e(t,n,i+1)}))}(ot(e),n,1),new Function("debugger;\n/*\nTracing '"+e.name+"'\n\nYou are entering this break point because derivation '"+e.name+"' is being traced and '"+t.name+"' is now forcing it to update.\nJust follow the stacktrace you should now see in the devtools to see precisely what piece of your code is causing this update\nThe stackframe you are looking for is at least ~6-8 stack-frames up.\n\n"+(e instanceof Ce?e.derivation.toString().replace(/[*]\//g,"/"):"")+"\n\nThe dependencies for this derivation are:\n\n"+n.join("\n")+"\n*/\n    ")()}}var Pe=function(){function e(e,t,n,i){void 0===e&&(e="Reaction@"+h()),void 0===i&&(i=!1),this.name=e,this.onInvalidate=t,this.errorHandler=n,this.requiresObservable=i,this.observing=[],this.newObserving=[],this.dependenciesState=X.NOT_TRACKING,this.diffValue=0,this.runId=0,this.unboundDepsCount=0,this.__mapid="#"+h(),this.isDisposed=!1,this._isScheduled=!1,this._isTrackPending=!1,this._isRunning=!1,this.isTracing=J.NONE}return e.prototype.onBecomeStale=function(){this.schedule()},e.prototype.schedule=function(){this._isScheduled||(this._isScheduled=!0,De.pendingReactions.push(this),Ne())},e.prototype.isScheduled=function(){return this._isScheduled},e.prototype.runReaction=function(){if(!this.isDisposed){if(Te(),this._isScheduled=!1,re(this)){this._isTrackPending=!0;try{this.onInvalidate(),this._isTrackPending}catch(e){this.reportExceptionInDerivation(e)}}Re()}},e.prototype.track=function(e){if(!this.isDisposed){Te();0,this._isRunning=!0;var t=se(this,e,void 0);this._isRunning=!1,this._isTrackPending=!1,this.isDisposed&&ae(this),ie(t)&&this.reportExceptionInDerivation(t.cause),Re()}},e.prototype.reportExceptionInDerivation=function(e){var t=this;if(this.errorHandler)this.errorHandler(e,this);else{if(De.disableErrorBoundaries)throw e;var n="[mobx] Encountered an uncaught exception that was thrown by a reaction or observer component, in: '"+this+"'";De.suppressReactionErrors?console.warn("[mobx] (error in reaction '"+this.name+"' suppressed, fix error of causing action below)"):console.error(n,e),De.globalReactionErrorHandlers.forEach((function(n){return n(e,t)}))}},e.prototype.dispose=function(){this.isDisposed||(this.isDisposed=!0,this._isRunning||(Te(),ae(this),Re()))},e.prototype.getDisposer=function(){var e=this.dispose.bind(this);return e[x]=this,e},e.prototype.toString=function(){return"Reaction["+this.name+"]"},e.prototype.trace=function(e){void 0===e&&(e=!1),function(){for(var e=[],t=0;t<arguments.length;t++)e[t]=arguments[t];var n=!1;"boolean"==typeof e[e.length-1]&&(n=e.pop());var i=ct(e);if(!i)return f(!1);i.isTracing===J.NONE&&console.log("[mobx.trace] '"+i.name+"' tracing enabled");i.isTracing=n?J.BREAK:J.LOG}(this,e)},e}();var Ie=function(e){return e()};function Ne(){De.inBatch>0||De.isRunningReactions||Ie(We)}function We(){De.isRunningReactions=!0;for(var e=De.pendingReactions,t=0;e.length>0;){100==++t&&(console.error("Reaction doesn't converge to a stable state after 100 iterations. Probably there is a cycle in the reactive function: "+e[0]),e.splice(0));for(var n=e.splice(0),i=0,r=n.length;i<r;i++)n[i].runReaction()}De.isRunningReactions=!1}var ze=y("Reaction",Pe);function je(e){var t=Ie;Ie=function(n){return e((function(){return t(n)}))}}function He(e){return console.warn("[mobx.spy] Is a no-op in production builds"),function(){}}function Ve(){f(!1)}function Ue(e){return function(t,n,i){if(i){if(i.value)return{value:me(e,i.value),enumerable:!1,configurable:!0,writable:!0};var r=i.initializer;return{enumerable:!1,configurable:!0,writable:!0,initializer:function(){return me(e,r.call(this))}}}return Ke(e).apply(this,arguments)}}function Ke(e){return function(t,n,i){Object.defineProperty(t,n,{configurable:!0,enumerable:!1,get:function(){},set:function(t){b(this,n,Ge(e,t))}})}}var Ge=function(e,t,n,i){return 1===arguments.length&&"function"==typeof e?me(e.name||"<unnamed action>",e):2===arguments.length&&"function"==typeof t?me(e,t):1===arguments.length&&"string"==typeof e?Ue(e):!0!==i?Ue(t).apply(null,arguments):void b(e,t,me(e.name||t,n.value,this))};function qe(e,t){"string"==typeof e||e.name;return ve(0,"function"==typeof e?e:t,this,void 0)}function Ye(e,t,n){b(e,t,me(t,n.bind(e)))}function Qe(e,t){void 0===t&&(t=c);var n,i=t&&t.name||e.name||"Autorun@"+h();if(!t.scheduler&&!t.delay)n=new Pe(i,(function(){this.track(s)}),t.onError,t.requiresObservable);else{var r=Je(t),o=!1;n=new Pe(i,(function(){o||(o=!0,r((function(){o=!1,n.isDisposed||n.track(s)})))}),t.onError,t.requiresObservable)}function s(){e(n)}return n.schedule(),n.getDisposer()}Ge.bound=function(e,t,n,i){return!0===i?(Ye(e,t,n.value),null):n?{configurable:!0,enumerable:!1,get:function(){return Ye(this,t,n.value||n.initializer.call(this)),this[t]},set:Ve}:{enumerable:!1,configurable:!0,set:function(e){Ye(this,t,e)},get:function(){}}};var Xe=function(e){return e()};function Je(e){return e.scheduler?e.scheduler:e.delay?function(t){return setTimeout(t,e.delay)}:Xe}function Ze(e,t,n){return et("onBecomeUnobserved",e,t,n)}function et(e,t,n,i){var r="function"==typeof i?Ht(t,n):Ht(t),o="function"==typeof i?i:n,s=e+"Listeners";return r[s]?r[s].add(o):r[s]=new Set([o]),"function"!=typeof r[e]?f(!1):function(){var e=r[s];e&&(e.delete(o),0===e.size&&delete r[s])}}function tt(e){var t=e.enforceActions,n=e.computedRequiresReaction,i=e.computedConfigurable,r=e.disableErrorBoundaries,o=e.reactionScheduler,s=e.reactionRequiresObservable,a=e.observableRequiresReaction;if(!0===e.isolateGlobalState&&((De.pendingReactions.length||De.inBatch||De.isRunningReactions)&&f("isolateGlobalState should be called before MobX is running any reactions"),_e=!0,Se&&(0==--ke().__mobxInstanceCount&&(ke().__mobxGlobals=void 0),De=new Fe)),void 0!==t){var l=void 0;switch(t){case!0:case"observed":l=!0;break;case!1:case"never":l=!1;break;case"strict":case"always":l="strict";break;default:f("Invalid value for 'enforceActions': '"+t+"', expected 'never', 'always' or 'observed'")}De.enforceActions=l,De.allowStateChanges=!0!==l&&"strict"!==l}void 0!==n&&(De.computedRequiresReaction=!!n),void 0!==s&&(De.reactionRequiresObservable=!!s),void 0!==a&&(De.observableRequiresReaction=!!a,De.allowStateReads=!De.observableRequiresReaction),void 0!==i&&(De.computedConfigurable=!!i),void 0!==r&&(!0===r&&console.warn("WARNING: Debug feature only. MobX will NOT recover from errors when `disableErrorBoundaries` is enabled."),De.disableErrorBoundaries=!!r),o&&je(o)}function nt(e,t,n,i){var r=it(i=j(i));return O(e),Pt(e,i.name,r.enhancer),t&&rt(e,t,n,r),e}function it(e){return e.defaultDecorator||(!1===e.deep?U:H)}function rt(e,t,n,i){var r,o;Te();try{var a=C(t);try{for(var l=s(a),u=l.next();!u.done;u=l.next()){var c=u.value,h=Object.getOwnPropertyDescriptor(t,c);0;var f=(n&&c in n?n[c]:h.get?Z:i)(e,c,h,!0);f&&Object.defineProperty(e,c,f)}}catch(e){r={error:e}}finally{try{u&&!u.done&&(o=l.return)&&o.call(l)}finally{if(r)throw r.error}}}finally{Re()}}function ot(e,t){return st(Ht(e,t))}function st(e){var t,n,i={name:e.name};return e.observing&&e.observing.length>0&&(i.dependencies=(t=e.observing,n=[],t.forEach((function(e){-1===n.indexOf(e)&&n.push(e)})),n).map(st)),i}function at(){this.message="FLOW_CANCELLED"}function lt(e,t){return null!=e&&(void 0!==t?!!jt(e)&&e[x].values.has(t):jt(e)||!!e[x]||S(e)||ze(e)||Ee(e))}function ut(e){return 1!==arguments.length&&f(!1),lt(e)}at.prototype=Object.create(Error.prototype);function ct(e){switch(e.length){case 0:return De.trackingDerivation;case 1:return Ht(e[0]);case 2:return Ht(e[0],e[1])}}function ht(e,t){void 0===t&&(t=void 0),Te();try{return e.apply(t)}finally{Re()}}function ft(e){return e[x]}function dt(e){return"string"==typeof e||"number"==typeof e||"symbol"==typeof e}var pt={has:function(e,t){if(t===x||"constructor"===t||t===$)return!0;var n=ft(e);return dt(t)?n.has(t):t in e},get:function(e,t){if(t===x||"constructor"===t||t===$)return e[t];var n=ft(e),i=n.values.get(t);if(i instanceof k){var r=i.get();return void 0===r&&n.has(t),r}return dt(t)&&n.has(t),e[t]},set:function(e,t,n){return!!dt(t)&&(function e(t,n,i){if(2!==arguments.length||Ot(t))if(jt(t)){var r=t[x],o=r.values.get(n);o?r.write(n,i):r.addObservableProp(n,i,r.defaultEnhancer)}else if(Lt(t))t.set(n,i);else if(Ot(t))t.add(n);else{if(!_t(t))return f(!1);"number"!=typeof n&&(n=parseInt(n,10)),d(n>=0,"Not a valid index: '"+n+"'"),Te(),n>=t.length&&(t.length=n+1),t[n]=i,Re()}else{Te();var s=n;try{for(var a in s)e(t,a,s[a])}finally{Re()}}}(e,t,n),!0)},deleteProperty:function(e,t){return!!dt(t)&&(ft(e).remove(t),!0)},ownKeys:function(e){return ft(e).keysAtom.reportObserved(),Reflect.ownKeys(e)},preventExtensions:function(e){return f("Dynamic observable objects cannot be frozen"),!1}};function gt(e){var t=new Proxy(e,pt);return e[x].proxy=t,t}function mt(e){return void 0!==e.interceptors&&e.interceptors.length>0}function vt(e,t){var n=e.interceptors||(e.interceptors=[]);return n.push(t),p((function(){var e=n.indexOf(t);-1!==e&&n.splice(e,1)}))}function bt(e,t){var n=ue();try{for(var i=l(e.interceptors||[]),r=0,o=i.length;r<o&&(d(!(t=i[r](t))||t.type,"Intercept handlers should return nothing or a change object"),t);r++);return t}finally{ce(n)}}function yt(e){return void 0!==e.changeListeners&&e.changeListeners.length>0}function wt(e,t){var n=e.changeListeners||(e.changeListeners=[]);return n.push(t),p((function(){var e=n.indexOf(t);-1!==e&&n.splice(e,1)}))}function At(e,t){var n=ue(),i=e.changeListeners;if(i){for(var r=0,o=(i=i.slice()).length;r<o;r++)i[r](t);ce(n)}}var Ct={get:function(e,t){return t===x?e[x]:"length"===t?e[x].getArrayLength():"number"==typeof t?xt.get.call(e,t):"string"!=typeof t||isNaN(t)?xt.hasOwnProperty(t)?xt[t]:e[t]:xt.get.call(e,parseInt(t))},set:function(e,t,n){return"length"===t&&e[x].setArrayLength(n),"number"==typeof t&&xt.set.call(e,t,n),"symbol"==typeof t||isNaN(t)?e[t]=n:xt.set.call(e,parseInt(t),n),!0},preventExtensions:function(e){return f("Observable arrays cannot be frozen"),!1}};function Et(e,t,n,i){void 0===n&&(n="ObservableArray@"+h()),void 0===i&&(i=!1);var r,o,s,a=new Ft(n,t,i);r=a.values,o=x,s=a,Object.defineProperty(r,o,{enumerable:!1,writable:!1,configurable:!0,value:s});var l=new Proxy(a.values,Ct);if(a.proxy=l,e&&e.length){var u=ye(!0);a.spliceWithArray(0,0,e),we(u)}return l}var Ft=function(){function e(e,t,n){this.owned=n,this.values=[],this.proxy=void 0,this.lastKnownLength=0,this.atom=new k(e||"ObservableArray@"+h()),this.enhancer=function(n,i){return t(n,i,e+"[..]")}}return e.prototype.dehanceValue=function(e){return void 0!==this.dehancer?this.dehancer(e):e},e.prototype.dehanceValues=function(e){return void 0!==this.dehancer&&e.length>0?e.map(this.dehancer):e},e.prototype.intercept=function(e){return vt(this,e)},e.prototype.observe=function(e,t){return void 0===t&&(t=!1),t&&e({object:this.proxy,type:"splice",index:0,added:this.values.slice(),addedCount:this.values.length,removed:[],removedCount:0}),wt(this,e)},e.prototype.getArrayLength=function(){return this.atom.reportObserved(),this.values.length},e.prototype.setArrayLength=function(e){if("number"!=typeof e||e<0)throw new Error("[mobx.array] Out of range: "+e);var t=this.values.length;if(e!==t)if(e>t){for(var n=new Array(e-t),i=0;i<e-t;i++)n[i]=void 0;this.spliceWithArray(t,0,n)}else this.spliceWithArray(e,t-e)},e.prototype.updateArrayLength=function(e,t){if(e!==this.lastKnownLength)throw new Error("[mobx] Modification exception: the internal structure of an observable array was changed.");this.lastKnownLength+=t},e.prototype.spliceWithArray=function(e,t,n){var i=this;oe(this.atom);var r=this.values.length;if(void 0===e?e=0:e>r?e=r:e<0&&(e=Math.max(0,r+e)),t=1===arguments.length?r-e:null==t?0:Math.max(0,Math.min(t,r-e)),void 0===n&&(n=u),mt(this)){var o=bt(this,{object:this.proxy,type:"splice",index:e,removedCount:t,added:n});if(!o)return u;t=o.removedCount,n=o.added}n=0===n.length?n:n.map((function(e){return i.enhancer(e,void 0)}));var s=this.spliceItemsIntoValues(e,t,n);return 0===t&&0===n.length||this.notifyArraySplice(e,n,s),this.dehanceValues(s)},e.prototype.spliceItemsIntoValues=function(e,t,n){var i;if(n.length<1e4)return(i=this.values).splice.apply(i,l([e,t],n));var r=this.values.slice(e,e+t);return this.values=this.values.slice(0,e).concat(n,this.values.slice(e+t)),r},e.prototype.notifyArrayChildUpdate=function(e,t,n){var i=!this.owned&&!1,r=yt(this),o=r||i?{object:this.proxy,type:"update",index:e,newValue:t,oldValue:n}:null;this.atom.reportChanged(),r&&At(this,o)},e.prototype.notifyArraySplice=function(e,t,n){var i=!this.owned&&!1,r=yt(this),o=r||i?{object:this.proxy,type:"splice",index:e,removed:n,added:t,removedCount:n.length,addedCount:t.length}:null;this.atom.reportChanged(),r&&At(this,o)},e}(),xt={intercept:function(e){return this[x].intercept(e)},observe:function(e,t){return void 0===t&&(t=!1),this[x].observe(e,t)},clear:function(){return this.splice(0)},replace:function(e){var t=this[x];return t.spliceWithArray(0,t.values.length,e)},toJS:function(){return this.slice()},toJSON:function(){return this.toJS()},splice:function(e,t){for(var n=[],i=2;i<arguments.length;i++)n[i-2]=arguments[i];var r=this[x];switch(arguments.length){case 0:return[];case 1:return r.spliceWithArray(e);case 2:return r.spliceWithArray(e,t)}return r.spliceWithArray(e,t,n)},spliceWithArray:function(e,t,n){return this[x].spliceWithArray(e,t,n)},push:function(){for(var e=[],t=0;t<arguments.length;t++)e[t]=arguments[t];var n=this[x];return n.spliceWithArray(n.values.length,0,e),n.values.length},pop:function(){return this.splice(Math.max(this[x].values.length-1,0),1)[0]},shift:function(){return this.splice(0,1)[0]},unshift:function(){for(var e=[],t=0;t<arguments.length;t++)e[t]=arguments[t];var n=this[x];return n.spliceWithArray(0,0,e),n.values.length},reverse:function(){var e=this.slice();return e.reverse.apply(e,arguments)},sort:function(e){var t=this.slice();return t.sort.apply(t,arguments)},remove:function(e){var t=this[x],n=t.dehanceValues(t.values).indexOf(e);return n>-1&&(this.splice(n,1),!0)},get:function(e){var t=this[x];if(t){if(e<t.values.length)return t.atom.reportObserved(),t.dehanceValue(t.values[e]);console.warn("[mobx.array] Attempt to read an array index ("+e+") that is out of bounds ("+t.values.length+"). Please check length first. Out of bound indices will not be tracked by MobX")}},set:function(e,t){var n=this[x],i=n.values;if(e<i.length){oe(n.atom);var r=i[e];if(mt(n)){var o=bt(n,{type:"update",object:n.proxy,index:e,newValue:t});if(!o)return;t=o.newValue}(t=n.enhancer(t,r))!==r&&(i[e]=t,n.notifyArrayChildUpdate(e,t,r))}else{if(e!==i.length)throw new Error("[mobx.array] Index out of bounds, "+e+" is larger than "+i.length);n.spliceWithArray(e,0,[t])}}};["concat","every","filter","forEach","indexOf","join","lastIndexOf","map","reduce","reduceRight","slice","some","toString","toLocaleString"].forEach((function(e){xt[e]=function(){var t=this[x];t.atom.reportObserved();var n=t.dehanceValues(t.values);return n[e].apply(n,arguments)}}));var kt,St=y("ObservableArrayAdministration",Ft);function _t(e){return m(e)&&St(e[x])}var Dt,$t={},Bt=function(){function e(e,t,n){if(void 0===t&&(t=I),void 0===n&&(n="ObservableMap@"+h()),this.enhancer=t,this.name=n,this[kt]=$t,this._keysAtom=_(this.name+".keys()"),this[Symbol.toStringTag]="Map","function"!=typeof Map)throw new Error("mobx.map requires Map polyfill for the current browser. Check babel-polyfill or core-js/es6/map.js");this._data=new Map,this._hasMap=new Map,this.merge(e)}return e.prototype._has=function(e){return this._data.has(e)},e.prototype.has=function(e){var t=this;if(!De.trackingDerivation)return this._has(e);var n=this._hasMap.get(e);if(!n){var i=n=new Ae(this._has(e),N,this.name+"."+E(e)+"?",!1);this._hasMap.set(e,i),Ze(i,(function(){return t._hasMap.delete(e)}))}return n.get()},e.prototype.set=function(e,t){var n=this._has(e);if(mt(this)){var i=bt(this,{type:n?"update":"add",object:this,newValue:t,name:e});if(!i)return this;t=i.newValue}return n?this._updateValue(e,t):this._addValue(e,t),this},e.prototype.delete=function(e){var t=this;if(mt(this)&&!(i=bt(this,{type:"delete",object:this,name:e})))return!1;if(this._has(e)){var n=yt(this),i=n?{type:"delete",object:this,oldValue:this._data.get(e).value,name:e}:null;return ht((function(){t._keysAtom.reportChanged(),t._updateHasMapEntry(e,!1),t._data.get(e).setNewValue(void 0),t._data.delete(e)})),n&&At(this,i),!0}return!1},e.prototype._updateHasMapEntry=function(e,t){var n=this._hasMap.get(e);n&&n.setNewValue(t)},e.prototype._updateValue=function(e,t){var n=this._data.get(e);if((t=n.prepareNewValue(t))!==De.UNCHANGED){var i=yt(this),r=i?{type:"update",object:this,oldValue:n.value,name:e,newValue:t}:null;0,n.setNewValue(t),i&&At(this,r)}},e.prototype._addValue=function(e,t){var n=this;oe(this._keysAtom),ht((function(){var i=new Ae(t,n.enhancer,n.name+"."+E(e),!1);n._data.set(e,i),t=i.value,n._updateHasMapEntry(e,!0),n._keysAtom.reportChanged()}));var i=yt(this),r=i?{type:"add",object:this,name:e,newValue:t}:null;i&&At(this,r)},e.prototype.get=function(e){return this.has(e)?this.dehanceValue(this._data.get(e).get()):this.dehanceValue(void 0)},e.prototype.dehanceValue=function(e){return void 0!==this.dehancer?this.dehancer(e):e},e.prototype.keys=function(){return this._keysAtom.reportObserved(),this._data.keys()},e.prototype.values=function(){var e=this,t=0,n=Array.from(this.keys());return Yt({next:function(){return t<n.length?{value:e.get(n[t++]),done:!1}:{done:!0}}})},e.prototype.entries=function(){var e=this,t=0,n=Array.from(this.keys());return Yt({next:function(){if(t<n.length){var i=n[t++];return{value:[i,e.get(i)],done:!1}}return{done:!0}}})},e.prototype[(kt=x,Symbol.iterator)]=function(){return this.entries()},e.prototype.forEach=function(e,t){var n,i;try{for(var r=s(this),o=r.next();!o.done;o=r.next()){var l=a(o.value,2),u=l[0],c=l[1];e.call(t,c,u,this)}}catch(e){n={error:e}}finally{try{o&&!o.done&&(i=r.return)&&i.call(r)}finally{if(n)throw n.error}}},e.prototype.merge=function(e){var t=this;return Lt(e)&&(e=e.toJS()),ht((function(){v(e)?C(e).forEach((function(n){return t.set(n,e[n])})):Array.isArray(e)?e.forEach((function(e){var n=a(e,2),i=n[0],r=n[1];return t.set(i,r)})):w(e)?(e.constructor!==Map&&f("Cannot initialize from classes that inherit from Map: "+e.constructor.name),e.forEach((function(e,n){return t.set(n,e)}))):null!=e&&f("Cannot initialize map from "+e)})),this},e.prototype.clear=function(){var e=this;ht((function(){le((function(){var t,n;try{for(var i=s(e.keys()),r=i.next();!r.done;r=i.next()){var o=r.value;e.delete(o)}}catch(e){t={error:e}}finally{try{r&&!r.done&&(n=i.return)&&n.call(i)}finally{if(t)throw t.error}}}))}))},e.prototype.replace=function(e){var t=this;return ht((function(){var n,i=v(n=e)?Object.keys(n):Array.isArray(n)?n.map((function(e){return a(e,1)[0]})):w(n)||Lt(n)?Array.from(n.keys()):f("Cannot get keys from '"+n+"'");Array.from(t.keys()).filter((function(e){return-1===i.indexOf(e)})).forEach((function(e){return t.delete(e)})),t.merge(e)})),this},Object.defineProperty(e.prototype,"size",{get:function(){return this._keysAtom.reportObserved(),this._data.size},enumerable:!0,configurable:!0}),e.prototype.toPOJO=function(){var e,t,n={};try{for(var i=s(this),r=i.next();!r.done;r=i.next()){var o=a(r.value,2),l=o[0],u=o[1];n["symbol"==typeof l?l:E(l)]=u}}catch(t){e={error:t}}finally{try{r&&!r.done&&(t=i.return)&&t.call(i)}finally{if(e)throw e.error}}return n},e.prototype.toJS=function(){return new Map(this)},e.prototype.toJSON=function(){return this.toPOJO()},e.prototype.toString=function(){var e=this;return this.name+"[{ "+Array.from(this.keys()).map((function(t){return E(t)+": "+e.get(t)})).join(", ")+" }]"},e.prototype.observe=function(e,t){return wt(this,e)},e.prototype.intercept=function(e){return vt(this,e)},e}(),Lt=y("ObservableMap",Bt),Tt={},Rt=function(){function e(e,t,n){if(void 0===t&&(t=I),void 0===n&&(n="ObservableSet@"+h()),this.name=n,this[Dt]=Tt,this._data=new Set,this._atom=_(this.name),this[Symbol.toStringTag]="Set","function"!=typeof Set)throw new Error("mobx.set requires Set polyfill for the current browser. Check babel-polyfill or core-js/es6/set.js");this.enhancer=function(e,i){return t(e,i,n)},e&&this.replace(e)}return e.prototype.dehanceValue=function(e){return void 0!==this.dehancer?this.dehancer(e):e},e.prototype.clear=function(){var e=this;ht((function(){le((function(){var t,n;try{for(var i=s(e._data.values()),r=i.next();!r.done;r=i.next()){var o=r.value;e.delete(o)}}catch(e){t={error:e}}finally{try{r&&!r.done&&(n=i.return)&&n.call(i)}finally{if(t)throw t.error}}}))}))},e.prototype.forEach=function(e,t){var n,i;try{for(var r=s(this),o=r.next();!o.done;o=r.next()){var a=o.value;e.call(t,a,a,this)}}catch(e){n={error:e}}finally{try{o&&!o.done&&(i=r.return)&&i.call(r)}finally{if(n)throw n.error}}},Object.defineProperty(e.prototype,"size",{get:function(){return this._atom.reportObserved(),this._data.size},enumerable:!0,configurable:!0}),e.prototype.add=function(e){var t=this;if((oe(this._atom),mt(this))&&!(i=bt(this,{type:"add",object:this,newValue:e})))return this;if(!this.has(e)){ht((function(){t._data.add(t.enhancer(e,void 0)),t._atom.reportChanged()}));var n=yt(this),i=n?{type:"add",object:this,newValue:e}:null;0,n&&At(this,i)}return this},e.prototype.delete=function(e){var t=this;if(mt(this)&&!(i=bt(this,{type:"delete",object:this,oldValue:e})))return!1;if(this.has(e)){var n=yt(this),i=n?{type:"delete",object:this,oldValue:e}:null;return ht((function(){t._atom.reportChanged(),t._data.delete(e)})),n&&At(this,i),!0}return!1},e.prototype.has=function(e){return this._atom.reportObserved(),this._data.has(this.dehanceValue(e))},e.prototype.entries=function(){var e=0,t=Array.from(this.keys()),n=Array.from(this.values());return Yt({next:function(){var i=e;return e+=1,i<n.length?{value:[t[i],n[i]],done:!1}:{done:!0}}})},e.prototype.keys=function(){return this.values()},e.prototype.values=function(){this._atom.reportObserved();var e=this,t=0,n=Array.from(this._data.values());return Yt({next:function(){return t<n.length?{value:e.dehanceValue(n[t++]),done:!1}:{done:!0}}})},e.prototype.replace=function(e){var t=this;return Ot(e)&&(e=e.toJS()),ht((function(){Array.isArray(e)?(t.clear(),e.forEach((function(e){return t.add(e)}))):A(e)?(t.clear(),e.forEach((function(e){return t.add(e)}))):null!=e&&f("Cannot initialize set from "+e)})),this},e.prototype.observe=function(e,t){return wt(this,e)},e.prototype.intercept=function(e){return vt(this,e)},e.prototype.toJS=function(){return new Set(this)},e.prototype.toString=function(){return this.name+"[ "+Array.from(this).join(", ")+" ]"},e.prototype[(Dt=x,Symbol.iterator)]=function(){return this.values()},e}(),Ot=y("ObservableSet",Rt),Mt=function(){function e(e,t,n,i){void 0===t&&(t=new Map),this.target=e,this.values=t,this.name=n,this.defaultEnhancer=i,this.keysAtom=new k(n+".keys")}return e.prototype.read=function(e){return this.values.get(e).get()},e.prototype.write=function(e,t){var n=this.target,i=this.values.get(e);if(i instanceof Ce)i.set(t);else{if(mt(this)){if(!(o=bt(this,{type:"update",object:this.proxy||n,name:e,newValue:t})))return;t=o.newValue}if((t=i.prepareNewValue(t))!==De.UNCHANGED){var r=yt(this),o=r?{type:"update",object:this.proxy||n,oldValue:i.value,name:e,newValue:t}:null;0,i.setNewValue(t),r&&At(this,o)}}},e.prototype.has=function(e){var t=this.pendingKeys||(this.pendingKeys=new Map),n=t.get(e);if(n)return n.get();var i=!!this.values.get(e);return n=new Ae(i,N,this.name+"."+E(e)+"?",!1),t.set(e,n),n.get()},e.prototype.addObservableProp=function(e,t,n){void 0===n&&(n=this.defaultEnhancer);var i=this.target;if(mt(this)){var r=bt(this,{object:this.proxy||i,name:e,type:"add",newValue:t});if(!r)return;t=r.newValue}var o=new Ae(t,n,this.name+"."+E(e),!1);this.values.set(e,o),t=o.value,Object.defineProperty(i,e,function(e){return It[e]||(It[e]={configurable:!0,enumerable:!0,get:function(){return this[x].read(e)},set:function(t){this[x].write(e,t)}})}(e)),this.notifyPropertyAddition(e,t)},e.prototype.addComputedProp=function(e,t,n){var i,r,o,s=this.target;n.name=n.name||this.name+"."+E(t),this.values.set(t,new Ce(n)),(e===s||(i=e,r=t,!(o=Object.getOwnPropertyDescriptor(i,r))||!1!==o.configurable&&!1!==o.writable))&&Object.defineProperty(e,t,function(e){return Nt[e]||(Nt[e]={configurable:De.computedConfigurable,enumerable:!1,get:function(){return Wt(this).read(e)},set:function(t){Wt(this).write(e,t)}})}(t))},e.prototype.remove=function(e){if(this.values.has(e)){var t=this.target;if(mt(this))if(!(s=bt(this,{object:this.proxy||t,name:e,type:"remove"})))return;try{Te();var n=yt(this),i=this.values.get(e),r=i&&i.get();if(i&&i.set(void 0),this.keysAtom.reportChanged(),this.values.delete(e),this.pendingKeys){var o=this.pendingKeys.get(e);o&&o.set(!1)}delete this.target[e];var s=n?{type:"remove",object:this.proxy||t,oldValue:r,name:e}:null;0,n&&At(this,s)}finally{Re()}}},e.prototype.illegalAccess=function(e,t){console.warn("Property '"+t+"' of '"+e+"' was accessed through the prototype chain. Use 'decorate' instead to declare the prop or access it statically through it's owner")},e.prototype.observe=function(e,t){return wt(this,e)},e.prototype.intercept=function(e){return vt(this,e)},e.prototype.notifyPropertyAddition=function(e,t){var n=yt(this),i=n?{type:"add",object:this.proxy||this.target,name:e,newValue:t}:null;if(n&&At(this,i),this.pendingKeys){var r=this.pendingKeys.get(e);r&&r.set(!0)}this.keysAtom.reportChanged()},e.prototype.getKeys=function(){var e,t;this.keysAtom.reportObserved();var n=[];try{for(var i=s(this.values),r=i.next();!r.done;r=i.next()){var o=a(r.value,2),l=o[0];o[1]instanceof Ae&&n.push(l)}}catch(t){e={error:t}}finally{try{r&&!r.done&&(t=i.return)&&t.call(i)}finally{if(e)throw e.error}}return n},e}();function Pt(e,t,n){if(void 0===t&&(t=""),void 0===n&&(n=I),Object.prototype.hasOwnProperty.call(e,x))return e[x];v(e)||(t=(e.constructor.name||"ObservableObject")+"@"+h()),t||(t="ObservableObject@"+h());var i=new Mt(e,new Map,E(t),n);return b(e,x,i),i}var It=Object.create(null),Nt=Object.create(null);function Wt(e){var t=e[x];return t||(O(e),e[x])}var zt=y("ObservableObjectAdministration",Mt);function jt(e){return!!m(e)&&(O(e),zt(e[x]))}function Ht(e,t){if("object"==typeof e&&null!==e){if(_t(e))return void 0!==t&&f(!1),e[x].atom;if(Ot(e))return e[x];if(Lt(e)){var n=e;return void 0===t?n._keysAtom:((i=n._data.get(t)||n._hasMap.get(t))||f(!1),i)}var i;if(O(e),t&&!e[x]&&e[t],jt(e))return t?((i=e[x].values.get(t))||f(!1),i):f(!1);if(S(e)||Ee(e)||ze(e))return e}else if("function"==typeof e&&ze(e[x]))return e[x];return f(!1)}function Vt(e,t){return e||f("Expecting some object"),void 0!==t?Vt(Ht(e,t)):S(e)||Ee(e)||ze(e)?e:Lt(e)||Ot(e)?e:(O(e),e[x]?e[x]:void f(!1))}var Ut=Object.prototype.toString;function Kt(e,t,n){return void 0===n&&(n=-1),function e(t,n,i,r,o){if(t===n)return 0!==t||1/t==1/n;if(null==t||null==n)return!1;if(t!=t)return n!=n;var s=typeof t;if("function"!==s&&"object"!==s&&"object"!=typeof n)return!1;t=Gt(t),n=Gt(n);var a=Ut.call(t);if(a!==Ut.call(n))return!1;switch(a){case"[object RegExp]":case"[object String]":return""+t==""+n;case"[object Number]":return+t!=+t?+n!=+n:0==+t?1/+t==1/n:+t==+n;case"[object Date]":case"[object Boolean]":return+t==+n;case"[object Symbol]":return"undefined"!=typeof Symbol&&Symbol.valueOf.call(t)===Symbol.valueOf.call(n)}var l="[object Array]"===a;if(!l){if("object"!=typeof t||"object"!=typeof n)return!1;var u=t.constructor,c=n.constructor;if(u!==c&&!("function"==typeof u&&u instanceof u&&"function"==typeof c&&c instanceof c)&&"constructor"in t&&"constructor"in n)return!1}if(0===i)return!1;i<0&&(i=-1);o=o||[];var h=(r=r||[]).length;for(;h--;)if(r[h]===t)return o[h]===n;if(r.push(t),o.push(n),l){if((h=t.length)!==n.length)return!1;for(;h--;)if(!e(t[h],n[h],i-1,r,o))return!1}else{var f=Object.keys(t),d=void 0;if(h=f.length,Object.keys(n).length!==h)return!1;for(;h--;)if(d=f[h],!qt(n,d)||!e(t[d],n[d],i-1,r,o))return!1}return r.pop(),o.pop(),!0}(e,t,n)}function Gt(e){return _t(e)?e.slice():w(e)||Lt(e)?Array.from(e.entries()):A(e)||Ot(e)?Array.from(e.entries()):e}function qt(e,t){return Object.prototype.hasOwnProperty.call(e,t)}function Yt(e){return e[Symbol.iterator]=Qt,e}function Qt(){return this}if("undefined"==typeof Proxy||"undefined"==typeof Symbol)throw new Error("[mobx] MobX 5+ requires Proxy and Symbol objects. If your environment doesn't support Symbol or Proxy objects, please downgrade to MobX 4. For React Native Android, consider upgrading JSCore.");"object"==typeof __MOBX_DEVTOOLS_GLOBAL_HOOK__&&__MOBX_DEVTOOLS_GLOBAL_HOOK__.injectMobx({spy:He,extras:{getDebugName:function(e,t){return(void 0!==t?Ht(e,t):jt(e)||Lt(e)||Ot(e)?Vt(e):Ht(e)).name}},$mobx:x})}).call(this,n(71),n(26))},function(e,t){function n(e,t,n,i,r,o,s){try{var a=e[o](s),l=a.value}catch(e){return void n(e)}a.done?t(l):Promise.resolve(l).then(i,r)}e.exports=function(e){return function(){var t=this,i=arguments;return new Promise((function(r,o){var s=e.apply(t,i);function a(e){n(s,r,o,a,l,"next",e)}function l(e){n(s,r,o,a,l,"throw",e)}a(void 0)}))}}},function(e,t){e.exports=function(e,t,n,i,r){var o={};return Object.keys(i).forEach((function(e){o[e]=i[e]})),o.enumerable=!!o.enumerable,o.configurable=!!o.configurable,("value"in o||o.initializer)&&(o.writable=!0),o=n.slice().reverse().reduce((function(n,i){return i(e,t,n)||n}),o),r&&void 0!==o.initializer&&(o.value=o.initializer?o.initializer.call(r):void 0,o.initializer=void 0),void 0===o.initializer&&(Object.defineProperty(e,t,o),o=null),o}},function(e,t){e.exports=function(e,t){if(!(e instanceof t))throw new TypeError("Cannot call a class as a function")}},function(e,t){function n(e,t){for(var n=0;n<t.length;n++){var i=t[n];i.enumerable=i.enumerable||!1,i.configurable=!0,"value"in i&&(i.writable=!0),Object.defineProperty(e,i.key,i)}}e.exports=function(e,t,i){return t&&n(e.prototype,t),i&&n(e,i),e}},function(e,t){e.exports=function(e){var t=typeof e;return null!=e&&("object"==t||"function"==t)}},function(e,t,n){"use strict";!function e(){if("undefined"!=typeof __REACT_DEVTOOLS_GLOBAL_HOOK__&&"function"==typeof __REACT_DEVTOOLS_GLOBAL_HOOK__.checkDCE){0;try{__REACT_DEVTOOLS_GLOBAL_HOOK__.checkDCE(e)}catch(e){console.error(e)}}}(),e.exports=n(65)},function(t,n){t.exports=e},function(e,t){e.exports=function(e,t,n,i){n&&Object.defineProperty(e,t,{enumerable:n.enumerable,configurable:n.configurable,writable:n.writable,value:n.initializer?n.initializer.call(i):void 0})}},function(e,t,n){var i=n(41),r="object"==typeof self&&self&&self.Object===Object&&self,o=i||r||Function("return this")();e.exports=o},function(e,t){e.exports=function(e){return null!=e&&"object"==typeof e}},function(e,t,n){!function(){var e=function(){return this}();e||"undefined"==typeof window||(e=window);var t=function(e,n,i){"string"==typeof e?(2==arguments.length&&(i=n),t.modules[e]||(t.payloads[e]=i,t.modules[e]=null)):t.original?t.original.apply(this,arguments):(console.error("dropping module because define wasn't a string."),console.trace())};t.modules={},t.payloads={};var n,i,r=function(e,t,n){if("string"==typeof t){var i=a(e,t);if(null!=i)return n&&n(),i}else if("[object Array]"===Object.prototype.toString.call(t)){for(var r=[],s=0,l=t.length;s<l;++s){var u=a(e,t[s]);if(null==u&&o.original)return;r.push(u)}return n&&n.apply(null,r)||!0}},o=function(e,t){var n=r("",e,t);return null==n&&o.original?o.original.apply(this,arguments):n},s=function(e,t){if(-1!==t.indexOf("!")){var n=t.split("!");return s(e,n[0])+"!"+s(e,n[1])}if("."==t.charAt(0))for(t=e.split("/").slice(0,-1).join("/")+"/"+t;-1!==t.indexOf(".")&&i!=t;){var i=t;t=t.replace(/\/\.\//,"/").replace(/[^\/]+\/\.\.\//,"")}return t},a=function(e,n){n=s(e,n);var i=t.modules[n];if(!i){if("function"==typeof(i=t.payloads[n])){var o={},a={id:n,uri:"",exports:o,packaged:!0};o=i((function(e,t){return r(n,e,t)}),o,a)||a.exports,t.modules[n]=o,delete t.payloads[n]}i=t.modules[n]=o||i}return i};i=e,(n="ace")&&(e[n]||(e[n]={}),i=e[n]),i.define&&i.define.packaged||(t.original=i.define,i.define=t,i.define.packaged=!0),i.acequire&&i.acequire.packaged||(o.original=i.acequire,i.acequire=o,i.acequire.packaged=!0)}(),ace.define("ace/lib/regexp",["require","exports","module"],(function(e,t,n){"use strict";var i,r={exec:RegExp.prototype.exec,test:RegExp.prototype.test,match:String.prototype.match,replace:String.prototype.replace,split:String.prototype.split},o=void 0===r.exec.call(/()??/,"")[1],s=(i=/^/g,r.test.call(i,""),!i.lastIndex);function a(e){return(e.global?"g":"")+(e.ignoreCase?"i":"")+(e.multiline?"m":"")+(e.extended?"x":"")+(e.sticky?"y":"")}function l(e,t,n){if(Array.prototype.indexOf)return e.indexOf(t,n);for(var i=n||0;i<e.length;i++)if(e[i]===t)return i;return-1}s&&o||(RegExp.prototype.exec=function(e){var t,n,i=r.exec.apply(this,arguments);if("string"==typeof e&&i){if(!o&&i.length>1&&l(i,"")>-1&&(n=RegExp(this.source,r.replace.call(a(this),"g","")),r.replace.call(e.slice(i.index),n,(function(){for(var e=1;e<arguments.length-2;e++)void 0===arguments[e]&&(i[e]=void 0)}))),this._xregexp&&this._xregexp.captureNames)for(var u=1;u<i.length;u++)(t=this._xregexp.captureNames[u-1])&&(i[t]=i[u]);!s&&this.global&&!i[0].length&&this.lastIndex>i.index&&this.lastIndex--}return i},s||(RegExp.prototype.test=function(e){var t=r.exec.call(this,e);return t&&this.global&&!t[0].length&&this.lastIndex>t.index&&this.lastIndex--,!!t}))})),ace.define("ace/lib/es5-shim",["require","exports","module"],(function(e,t,n){function i(){}Function.prototype.bind||(Function.prototype.bind=function(e){var t=this;if("function"!=typeof t)throw new TypeError("Function.prototype.bind called on incompatible "+t);var n=f.call(arguments,1),r=function(){if(this instanceof r){var i=t.apply(this,n.concat(f.call(arguments)));return Object(i)===i?i:this}return t.apply(e,n.concat(f.call(arguments)))};return t.prototype&&(i.prototype=t.prototype,r.prototype=new i,i.prototype=null),r});var r,o,s,a,l,u=Function.prototype.call,c=Array.prototype,h=Object.prototype,f=c.slice,d=u.bind(h.toString),p=u.bind(h.hasOwnProperty);if((l=p(h,"__defineGetter__"))&&(r=u.bind(h.__defineGetter__),o=u.bind(h.__defineSetter__),s=u.bind(h.__lookupGetter__),a=u.bind(h.__lookupSetter__)),2!=[1,2].splice(0).length)if(function(){function e(e){var t=new Array(e+2);return t[0]=t[1]=0,t}var t,n=[];if(n.splice.apply(n,e(20)),n.splice.apply(n,e(26)),t=n.length,n.splice(5,0,"XXX"),n.length,t+1==n.length)return!0}()){var g=Array.prototype.splice;Array.prototype.splice=function(e,t){return arguments.length?g.apply(this,[void 0===e?0:e,void 0===t?this.length-e:t].concat(f.call(arguments,2))):[]}}else Array.prototype.splice=function(e,t){var n=this.length;e>0?e>n&&(e=n):null==e?e=0:e<0&&(e=Math.max(n+e,0)),e+t<n||(t=n-e);var i=this.slice(e,e+t),r=f.call(arguments,2),o=r.length;if(e===n)o&&this.push.apply(this,r);else{var s=Math.min(t,n-e),a=e+s,l=a+o-s,u=n-a,c=n-s;if(l<a)for(var h=0;h<u;++h)this[l+h]=this[a+h];else if(l>a)for(h=u;h--;)this[l+h]=this[a+h];if(o&&e===c)this.length=c,this.push.apply(this,r);else for(this.length=c+o,h=0;h<o;++h)this[e+h]=r[h]}return i};Array.isArray||(Array.isArray=function(e){return"[object Array]"==d(e)});var m,v,b=Object("a"),y="a"!=b[0]||!(0 in b);if(Array.prototype.forEach||(Array.prototype.forEach=function(e){var t=L(this),n=y&&"[object String]"==d(this)?this.split(""):t,i=arguments[1],r=-1,o=n.length>>>0;if("[object Function]"!=d(e))throw new TypeError;for(;++r<o;)r in n&&e.call(i,n[r],r,t)}),Array.prototype.map||(Array.prototype.map=function(e){var t=L(this),n=y&&"[object String]"==d(this)?this.split(""):t,i=n.length>>>0,r=Array(i),o=arguments[1];if("[object Function]"!=d(e))throw new TypeError(e+" is not a function");for(var s=0;s<i;s++)s in n&&(r[s]=e.call(o,n[s],s,t));return r}),Array.prototype.filter||(Array.prototype.filter=function(e){var t,n=L(this),i=y&&"[object String]"==d(this)?this.split(""):n,r=i.length>>>0,o=[],s=arguments[1];if("[object Function]"!=d(e))throw new TypeError(e+" is not a function");for(var a=0;a<r;a++)a in i&&(t=i[a],e.call(s,t,a,n)&&o.push(t));return o}),Array.prototype.every||(Array.prototype.every=function(e){var t=L(this),n=y&&"[object String]"==d(this)?this.split(""):t,i=n.length>>>0,r=arguments[1];if("[object Function]"!=d(e))throw new TypeError(e+" is not a function");for(var o=0;o<i;o++)if(o in n&&!e.call(r,n[o],o,t))return!1;return!0}),Array.prototype.some||(Array.prototype.some=function(e){var t=L(this),n=y&&"[object String]"==d(this)?this.split(""):t,i=n.length>>>0,r=arguments[1];if("[object Function]"!=d(e))throw new TypeError(e+" is not a function");for(var o=0;o<i;o++)if(o in n&&e.call(r,n[o],o,t))return!0;return!1}),Array.prototype.reduce||(Array.prototype.reduce=function(e){var t=L(this),n=y&&"[object String]"==d(this)?this.split(""):t,i=n.length>>>0;if("[object Function]"!=d(e))throw new TypeError(e+" is not a function");if(!i&&1==arguments.length)throw new TypeError("reduce of empty array with no initial value");var r,o=0;if(arguments.length>=2)r=arguments[1];else for(;;){if(o in n){r=n[o++];break}if(++o>=i)throw new TypeError("reduce of empty array with no initial value")}for(;o<i;o++)o in n&&(r=e.call(void 0,r,n[o],o,t));return r}),Array.prototype.reduceRight||(Array.prototype.reduceRight=function(e){var t=L(this),n=y&&"[object String]"==d(this)?this.split(""):t,i=n.length>>>0;if("[object Function]"!=d(e))throw new TypeError(e+" is not a function");if(!i&&1==arguments.length)throw new TypeError("reduceRight of empty array with no initial value");var r,o=i-1;if(arguments.length>=2)r=arguments[1];else for(;;){if(o in n){r=n[o--];break}if(--o<0)throw new TypeError("reduceRight of empty array with no initial value")}do{o in this&&(r=e.call(void 0,r,n[o],o,t))}while(o--);return r}),Array.prototype.indexOf&&-1==[0,1].indexOf(1,2)||(Array.prototype.indexOf=function(e){var t=y&&"[object String]"==d(this)?this.split(""):L(this),n=t.length>>>0;if(!n)return-1;var i=0;for(arguments.length>1&&(i=B(arguments[1])),i=i>=0?i:Math.max(0,n+i);i<n;i++)if(i in t&&t[i]===e)return i;return-1}),Array.prototype.lastIndexOf&&-1==[0,1].lastIndexOf(0,-3)||(Array.prototype.lastIndexOf=function(e){var t=y&&"[object String]"==d(this)?this.split(""):L(this),n=t.length>>>0;if(!n)return-1;var i=n-1;for(arguments.length>1&&(i=Math.min(i,B(arguments[1]))),i=i>=0?i:n-Math.abs(i);i>=0;i--)if(i in t&&e===t[i])return i;return-1}),Object.getPrototypeOf||(Object.getPrototypeOf=function(e){return e.__proto__||(e.constructor?e.constructor.prototype:h)}),!Object.getOwnPropertyDescriptor){Object.getOwnPropertyDescriptor=function(e,t){if("object"!=typeof e&&"function"!=typeof e||null===e)throw new TypeError("Object.getOwnPropertyDescriptor called on a non-object: "+e);if(p(e,t)){var n;if(n={enumerable:!0,configurable:!0},l){var i=e.__proto__;e.__proto__=h;var r=s(e,t),o=a(e,t);if(e.__proto__=i,r||o)return r&&(n.get=r),o&&(n.set=o),n}return n.value=e[t],n}}}(Object.getOwnPropertyNames||(Object.getOwnPropertyNames=function(e){return Object.keys(e)}),Object.create)||(m=null===Object.prototype.__proto__?function(){return{__proto__:null}}:function(){var e={};for(var t in e)e[t]=null;return e.constructor=e.hasOwnProperty=e.propertyIsEnumerable=e.isPrototypeOf=e.toLocaleString=e.toString=e.valueOf=e.__proto__=null,e},Object.create=function(e,t){var n;if(null===e)n=m();else{if("object"!=typeof e)throw new TypeError("typeof prototype["+typeof e+"] != 'object'");var i=function(){};i.prototype=e,(n=new i).__proto__=e}return void 0!==t&&Object.defineProperties(n,t),n});function w(e){try{return Object.defineProperty(e,"sentinel",{}),"sentinel"in e}catch(e){}}if(Object.defineProperty){var A=w({}),C="undefined"==typeof document||w(document.createElement("div"));if(!A||!C)var E=Object.defineProperty}if(!Object.defineProperty||E){Object.defineProperty=function(e,t,n){if("object"!=typeof e&&"function"!=typeof e||null===e)throw new TypeError("Object.defineProperty called on non-object: "+e);if("object"!=typeof n&&"function"!=typeof n||null===n)throw new TypeError("Property description must be an object: "+n);if(E)try{return E.call(Object,e,t,n)}catch(e){}if(p(n,"value"))if(l&&(s(e,t)||a(e,t))){var i=e.__proto__;e.__proto__=h,delete e[t],e[t]=n.value,e.__proto__=i}else e[t]=n.value;else{if(!l)throw new TypeError("getters & setters can not be defined on this javascript engine");p(n,"get")&&r(e,t,n.get),p(n,"set")&&o(e,t,n.set)}return e}}Object.defineProperties||(Object.defineProperties=function(e,t){for(var n in t)p(t,n)&&Object.defineProperty(e,n,t[n]);return e}),Object.seal||(Object.seal=function(e){return e}),Object.freeze||(Object.freeze=function(e){return e});try{Object.freeze((function(){}))}catch(e){Object.freeze=(v=Object.freeze,function(e){return"function"==typeof e?e:v(e)})}if(Object.preventExtensions||(Object.preventExtensions=function(e){return e}),Object.isSealed||(Object.isSealed=function(e){return!1}),Object.isFrozen||(Object.isFrozen=function(e){return!1}),Object.isExtensible||(Object.isExtensible=function(e){if(Object(e)===e)throw new TypeError;for(var t="";p(e,t);)t+="?";e[t]=!0;var n=p(e,t);return delete e[t],n}),!Object.keys){var F=!0,x=["toString","toLocaleString","valueOf","hasOwnProperty","isPrototypeOf","propertyIsEnumerable","constructor"],k=x.length;for(var S in{toString:null})F=!1;Object.keys=function(e){if("object"!=typeof e&&"function"!=typeof e||null===e)throw new TypeError("Object.keys called on a non-object");var t=[];for(var n in e)p(e,n)&&t.push(n);if(F)for(var i=0,r=k;i<r;i++){var o=x[i];p(e,o)&&t.push(o)}return t}}Date.now||(Date.now=function(){return(new Date).getTime()});var _="\t\n\v\f\r   ᠎             　\u2028\u2029\ufeff";if(!String.prototype.trim||_.trim()){_="["+_+"]";var D=new RegExp("^"+_+_+"*"),$=new RegExp(_+_+"*$");String.prototype.trim=function(){return String(this).replace(D,"").replace($,"")}}function B(e){return(e=+e)!=e?e=0:0!==e&&e!==1/0&&e!==-1/0&&(e=(e>0||-1)*Math.floor(Math.abs(e))),e}var L=function(e){if(null==e)throw new TypeError("can't convert "+e+" to object");return Object(e)}})),ace.define("ace/lib/fixoldbrowsers",["require","exports","module","ace/lib/regexp","ace/lib/es5-shim"],(function(e,t,n){"use strict";e("./regexp"),e("./es5-shim")})),ace.define("ace/lib/dom",["require","exports","module"],(function(e,t,n){"use strict";t.getDocumentHead=function(e){return e||(e=document),e.head||e.getElementsByTagName("head")[0]||e.documentElement},t.createElement=function(e,t){return document.createElementNS?document.createElementNS(t||"http://www.w3.org/1999/xhtml",e):document.createElement(e)},t.hasCssClass=function(e,t){return-1!==(e.className+"").split(/\s+/g).indexOf(t)},t.addCssClass=function(e,n){t.hasCssClass(e,n)||(e.className+=" "+n)},t.removeCssClass=function(e,t){for(var n=e.className.split(/\s+/g);;){var i=n.indexOf(t);if(-1==i)break;n.splice(i,1)}e.className=n.join(" ")},t.toggleCssClass=function(e,t){for(var n=e.className.split(/\s+/g),i=!0;;){var r=n.indexOf(t);if(-1==r)break;i=!1,n.splice(r,1)}return i&&n.push(t),e.className=n.join(" "),i},t.setCssClass=function(e,n,i){i?t.addCssClass(e,n):t.removeCssClass(e,n)},t.hasCssString=function(e,t){var n,i=0;if((t=t||document).createStyleSheet&&(n=t.styleSheets)){for(;i<n.length;)if(n[i++].owningElement.id===e)return!0}else if(n=t.getElementsByTagName("style"))for(;i<n.length;)if(n[i++].id===e)return!0;return!1},t.importCssString=function(e,n,i){if(i=i||document,n&&t.hasCssString(n,i))return null;var r;n&&(e+="\n/*# sourceURL=ace/css/"+n+" */"),i.createStyleSheet?((r=i.createStyleSheet()).cssText=e,n&&(r.owningElement.id=n)):((r=t.createElement("style")).appendChild(i.createTextNode(e)),n&&(r.id=n),t.getDocumentHead(i).appendChild(r))},t.importCssStylsheet=function(e,n){if(n.createStyleSheet)n.createStyleSheet(e);else{var i=t.createElement("link");i.rel="stylesheet",i.href=e,t.getDocumentHead(n).appendChild(i)}},t.getInnerWidth=function(e){return parseInt(t.computedStyle(e,"paddingLeft"),10)+parseInt(t.computedStyle(e,"paddingRight"),10)+e.clientWidth},t.getInnerHeight=function(e){return parseInt(t.computedStyle(e,"paddingTop"),10)+parseInt(t.computedStyle(e,"paddingBottom"),10)+e.clientHeight},t.scrollbarWidth=function(e){var n=t.createElement("ace_inner");n.style.width="100%",n.style.minWidth="0px",n.style.height="200px",n.style.display="block";var i=t.createElement("ace_outer"),r=i.style;r.position="absolute",r.left="-10000px",r.overflow="hidden",r.width="200px",r.minWidth="0px",r.height="150px",r.display="block",i.appendChild(n);var o=e.documentElement;o.appendChild(i);var s=n.offsetWidth;r.overflow="scroll";var a=n.offsetWidth;return s==a&&(a=i.clientWidth),o.removeChild(i),s-a},"undefined"!=typeof document?(void 0!==window.pageYOffset?(t.getPageScrollTop=function(){return window.pageYOffset},t.getPageScrollLeft=function(){return window.pageXOffset}):(t.getPageScrollTop=function(){return document.body.scrollTop},t.getPageScrollLeft=function(){return document.body.scrollLeft}),window.getComputedStyle?t.computedStyle=function(e,t){return t?(window.getComputedStyle(e,"")||{})[t]||"":window.getComputedStyle(e,"")||{}}:t.computedStyle=function(e,t){return t?e.currentStyle[t]:e.currentStyle},t.setInnerHtml=function(e,t){var n=e.cloneNode(!1);return n.innerHTML=t,e.parentNode.replaceChild(n,e),n},"textContent"in document.documentElement?(t.setInnerText=function(e,t){e.textContent=t},t.getInnerText=function(e){return e.textContent}):(t.setInnerText=function(e,t){e.innerText=t},t.getInnerText=function(e){return e.innerText}),t.getParentWindow=function(e){return e.defaultView||e.parentWindow}):t.importCssString=function(){}})),ace.define("ace/lib/oop",["require","exports","module"],(function(e,t,n){"use strict";t.inherits=function(e,t){e.super_=t,e.prototype=Object.create(t.prototype,{constructor:{value:e,enumerable:!1,writable:!0,configurable:!0}})},t.mixin=function(e,t){for(var n in t)e[n]=t[n];return e},t.implement=function(e,n){t.mixin(e,n)}})),ace.define("ace/lib/keys",["require","exports","module","ace/lib/fixoldbrowsers","ace/lib/oop"],(function(e,t,n){"use strict";e("./fixoldbrowsers");var i=e("./oop"),r=function(){var e,t,n={MODIFIER_KEYS:{16:"Shift",17:"Ctrl",18:"Alt",224:"Meta"},KEY_MODS:{ctrl:1,alt:2,option:2,shift:4,super:8,meta:8,command:8,cmd:8},FUNCTION_KEYS:{8:"Backspace",9:"Tab",13:"Return",19:"Pause",27:"Esc",32:"Space",33:"PageUp",34:"PageDown",35:"End",36:"Home",37:"Left",38:"Up",39:"Right",40:"Down",44:"Print",45:"Insert",46:"Delete",96:"Numpad0",97:"Numpad1",98:"Numpad2",99:"Numpad3",100:"Numpad4",101:"Numpad5",102:"Numpad6",103:"Numpad7",104:"Numpad8",105:"Numpad9","-13":"NumpadEnter",112:"F1",113:"F2",114:"F3",115:"F4",116:"F5",117:"F6",118:"F7",119:"F8",120:"F9",121:"F10",122:"F11",123:"F12",144:"Numlock",145:"Scrolllock"},PRINTABLE_KEYS:{32:" ",48:"0",49:"1",50:"2",51:"3",52:"4",53:"5",54:"6",55:"7",56:"8",57:"9",59:";",61:"=",65:"a",66:"b",67:"c",68:"d",69:"e",70:"f",71:"g",72:"h",73:"i",74:"j",75:"k",76:"l",77:"m",78:"n",79:"o",80:"p",81:"q",82:"r",83:"s",84:"t",85:"u",86:"v",87:"w",88:"x",89:"y",90:"z",107:"+",109:"-",110:".",186:";",187:"=",188:",",189:"-",190:".",191:"/",192:"`",219:"[",220:"\\",221:"]",222:"'",111:"/",106:"*"}};for(t in n.FUNCTION_KEYS)e=n.FUNCTION_KEYS[t].toLowerCase(),n[e]=parseInt(t,10);for(t in n.PRINTABLE_KEYS)e=n.PRINTABLE_KEYS[t].toLowerCase(),n[e]=parseInt(t,10);return i.mixin(n,n.MODIFIER_KEYS),i.mixin(n,n.PRINTABLE_KEYS),i.mixin(n,n.FUNCTION_KEYS),n.enter=n.return,n.escape=n.esc,n.del=n.delete,n[173]="-",function(){for(var e=["cmd","ctrl","alt","shift"],t=Math.pow(2,e.length);t--;)n.KEY_MODS[t]=e.filter((function(e){return t&n.KEY_MODS[e]})).join("-")+"-"}(),n.KEY_MODS[0]="",n.KEY_MODS[-1]="input-",n}();i.mixin(t,r),t.keyCodeToString=function(e){var t=r[e];return"string"!=typeof t&&(t=String.fromCharCode(e)),t.toLowerCase()}})),ace.define("ace/lib/useragent",["require","exports","module"],(function(e,t,n){"use strict";if(t.OS={LINUX:"LINUX",MAC:"MAC",WINDOWS:"WINDOWS"},t.getOS=function(){return t.isMac?t.OS.MAC:t.isLinux?t.OS.LINUX:t.OS.WINDOWS},"object"==typeof navigator){var i=(navigator.platform.match(/mac|win|linux/i)||["other"])[0].toLowerCase(),r=navigator.userAgent;t.isWin="win"==i,t.isMac="mac"==i,t.isLinux="linux"==i,t.isIE="Microsoft Internet Explorer"==navigator.appName||navigator.appName.indexOf("MSAppHost")>=0?parseFloat((r.match(/(?:MSIE |Trident\/[0-9]+[\.0-9]+;.*rv:)([0-9]+[\.0-9]+)/)||[])[1]):parseFloat((r.match(/(?:Trident\/[0-9]+[\.0-9]+;.*rv:)([0-9]+[\.0-9]+)/)||[])[1]),t.isOldIE=t.isIE&&t.isIE<9,t.isGecko=t.isMozilla=(window.Controllers||window.controllers)&&"Gecko"===window.navigator.product,t.isOldGecko=t.isGecko&&parseInt((r.match(/rv:(\d+)/)||[])[1],10)<4,t.isOpera=window.opera&&"[object Opera]"==Object.prototype.toString.call(window.opera),t.isWebKit=parseFloat(r.split("WebKit/")[1])||void 0,t.isChrome=parseFloat(r.split(" Chrome/")[1])||void 0,t.isAIR=r.indexOf("AdobeAIR")>=0,t.isIPad=r.indexOf("iPad")>=0,t.isChromeOS=r.indexOf(" CrOS ")>=0,t.isIOS=/iPad|iPhone|iPod/.test(r)&&!window.MSStream,t.isIOS&&(t.isMac=!0)}})),ace.define("ace/lib/event",["require","exports","module","ace/lib/keys","ace/lib/useragent"],(function(e,t,n){"use strict";var i=e("./keys"),r=e("./useragent"),o=null,s=0;t.addListener=function(e,t,n){if(e.addEventListener)return e.addEventListener(t,n,!1);if(e.attachEvent){var i=function(){n.call(e,window.event)};n._wrapper=i,e.attachEvent("on"+t,i)}},t.removeListener=function(e,t,n){if(e.removeEventListener)return e.removeEventListener(t,n,!1);e.detachEvent&&e.detachEvent("on"+t,n._wrapper||n)},t.stopEvent=function(e){return t.stopPropagation(e),t.preventDefault(e),!1},t.stopPropagation=function(e){e.stopPropagation?e.stopPropagation():e.cancelBubble=!0},t.preventDefault=function(e){e.preventDefault?e.preventDefault():e.returnValue=!1},t.getButton=function(e){return"dblclick"==e.type?0:"contextmenu"==e.type||r.isMac&&e.ctrlKey&&!e.altKey&&!e.shiftKey?2:e.preventDefault?e.button:{1:0,2:2,4:1}[e.button]},t.capture=function(e,n,i){function r(e){n&&n(e),i&&i(e),t.removeListener(document,"mousemove",n,!0),t.removeListener(document,"mouseup",r,!0),t.removeListener(document,"dragstart",r,!0)}return t.addListener(document,"mousemove",n,!0),t.addListener(document,"mouseup",r,!0),t.addListener(document,"dragstart",r,!0),r},t.addTouchMoveListener=function(e,n){var i,r;t.addListener(e,"touchstart",(function(e){var t=e.touches[0];i=t.clientX,r=t.clientY})),t.addListener(e,"touchmove",(function(e){var t=e.touches;if(!(t.length>1)){var o=t[0];e.wheelX=i-o.clientX,e.wheelY=r-o.clientY,i=o.clientX,r=o.clientY,n(e)}}))},t.addMouseWheelListener=function(e,n){"onmousewheel"in e?t.addListener(e,"mousewheel",(function(e){void 0!==e.wheelDeltaX?(e.wheelX=-e.wheelDeltaX/8,e.wheelY=-e.wheelDeltaY/8):(e.wheelX=0,e.wheelY=-e.wheelDelta/8),n(e)})):"onwheel"in e?t.addListener(e,"wheel",(function(e){switch(e.deltaMode){case e.DOM_DELTA_PIXEL:e.wheelX=.35*e.deltaX||0,e.wheelY=.35*e.deltaY||0;break;case e.DOM_DELTA_LINE:case e.DOM_DELTA_PAGE:e.wheelX=5*(e.deltaX||0),e.wheelY=5*(e.deltaY||0)}n(e)})):t.addListener(e,"DOMMouseScroll",(function(e){e.axis&&e.axis==e.HORIZONTAL_AXIS?(e.wheelX=5*(e.detail||0),e.wheelY=0):(e.wheelX=0,e.wheelY=5*(e.detail||0)),n(e)}))},t.addMultiMouseDownListener=function(e,n,i,o){var s,a,l,u=0,c={2:"dblclick",3:"tripleclick",4:"quadclick"};function h(e){if(0!==t.getButton(e)?u=0:e.detail>1?++u>4&&(u=1):u=1,r.isIE){var h=Math.abs(e.clientX-s)>5||Math.abs(e.clientY-a)>5;l&&!h||(u=1),l&&clearTimeout(l),l=setTimeout((function(){l=null}),n[u-1]||600),1==u&&(s=e.clientX,a=e.clientY)}if(e._clicks=u,i[o]("mousedown",e),u>4)u=0;else if(u>1)return i[o](c[u],e)}function f(e){u=2,l&&clearTimeout(l),l=setTimeout((function(){l=null}),n[u-1]||600),i[o]("mousedown",e),i[o](c[u],e)}Array.isArray(e)||(e=[e]),e.forEach((function(e){t.addListener(e,"mousedown",h),r.isOldIE&&t.addListener(e,"dblclick",f)}))};var a=!r.isMac||!r.isOpera||"KeyboardEvent"in window?function(e){return 0|(e.ctrlKey?1:0)|(e.altKey?2:0)|(e.shiftKey?4:0)|(e.metaKey?8:0)}:function(e){return 0|(e.metaKey?1:0)|(e.altKey?2:0)|(e.shiftKey?4:0)|(e.ctrlKey?8:0)};function l(e,t,n){var l=a(t);if(!r.isMac&&o){if(t.getModifierState&&(t.getModifierState("OS")||t.getModifierState("Win"))&&(l|=8),o.altGr){if(3==(3&l))return;o.altGr=0}if(18===n||17===n){var u="location"in t?t.location:t.keyLocation;if(17===n&&1===u)1==o[n]&&(s=t.timeStamp);else if(18===n&&3===l&&2===u){t.timeStamp-s<50&&(o.altGr=!0)}}}if((n in i.MODIFIER_KEYS&&(n=-1),8&l&&n>=91&&n<=93&&(n=-1),!l&&13===n)&&(3===(u="location"in t?t.location:t.keyLocation)&&(e(t,l,-n),t.defaultPrevented)))return;if(r.isChromeOS&&8&l){if(e(t,l,n),t.defaultPrevented)return;l&=-9}return!!(l||n in i.FUNCTION_KEYS||n in i.PRINTABLE_KEYS)&&e(t,l,n)}function u(){o=Object.create(null)}if(t.getModifierString=function(e){return i.KEY_MODS[a(e)]},t.addCommandKeyListener=function(e,n){var i=t.addListener;if(r.isOldGecko||r.isOpera&&!("KeyboardEvent"in window)){var s=null;i(e,"keydown",(function(e){s=e.keyCode})),i(e,"keypress",(function(e){return l(n,e,s)}))}else{var a=null;i(e,"keydown",(function(e){o[e.keyCode]=(o[e.keyCode]||0)+1;var t=l(n,e,e.keyCode);return a=e.defaultPrevented,t})),i(e,"keypress",(function(e){a&&(e.ctrlKey||e.altKey||e.shiftKey||e.metaKey)&&(t.stopEvent(e),a=null)})),i(e,"keyup",(function(e){o[e.keyCode]=null})),o||(u(),i(window,"focus",u))}},"object"==typeof window&&window.postMessage&&!r.isOldIE){t.nextTick=function(e,n){n=n||window;t.addListener(n,"message",(function i(r){"zero-timeout-message-1"==r.data&&(t.stopPropagation(r),t.removeListener(n,"message",i),e())})),n.postMessage("zero-timeout-message-1","*")}}t.nextFrame="object"==typeof window&&(window.requestAnimationFrame||window.mozRequestAnimationFrame||window.webkitRequestAnimationFrame||window.msRequestAnimationFrame||window.oRequestAnimationFrame),t.nextFrame?t.nextFrame=t.nextFrame.bind(window):t.nextFrame=function(e){setTimeout(e,17)}})),ace.define("ace/lib/lang",["require","exports","module"],(function(e,t,n){"use strict";t.last=function(e){return e[e.length-1]},t.stringReverse=function(e){return e.split("").reverse().join("")},t.stringRepeat=function(e,t){for(var n="";t>0;)1&t&&(n+=e),(t>>=1)&&(e+=e);return n};var i=/^\s\s*/,r=/\s\s*$/;t.stringTrimLeft=function(e){return e.replace(i,"")},t.stringTrimRight=function(e){return e.replace(r,"")},t.copyObject=function(e){var t={};for(var n in e)t[n]=e[n];return t},t.copyArray=function(e){for(var t=[],n=0,i=e.length;n<i;n++)e[n]&&"object"==typeof e[n]?t[n]=this.copyObject(e[n]):t[n]=e[n];return t},t.deepCopy=function e(t){if("object"!=typeof t||!t)return t;var n;if(Array.isArray(t)){n=[];for(var i=0;i<t.length;i++)n[i]=e(t[i]);return n}if("[object Object]"!==Object.prototype.toString.call(t))return t;for(var i in n={},t)n[i]=e(t[i]);return n},t.arrayToMap=function(e){for(var t={},n=0;n<e.length;n++)t[e[n]]=1;return t},t.createMap=function(e){var t=Object.create(null);for(var n in e)t[n]=e[n];return t},t.arrayRemove=function(e,t){for(var n=0;n<=e.length;n++)t===e[n]&&e.splice(n,1)},t.escapeRegExp=function(e){return e.replace(/([.*+?^${}()|[\]\/\\])/g,"\\$1")},t.escapeHTML=function(e){return e.replace(/&/g,"&#38;").replace(/"/g,"&#34;").replace(/'/g,"&#39;").replace(/</g,"&#60;")},t.getMatchOffsets=function(e,t){var n=[];return e.replace(t,(function(e){n.push({offset:arguments[arguments.length-2],length:e.length})})),n},t.deferredCall=function(e){var t=null,n=function(){t=null,e()},i=function(e){return i.cancel(),t=setTimeout(n,e||0),i};return i.schedule=i,i.call=function(){return this.cancel(),e(),i},i.cancel=function(){return clearTimeout(t),t=null,i},i.isPending=function(){return t},i},t.delayedCall=function(e,t){var n=null,i=function(){n=null,e()},r=function(e){null==n&&(n=setTimeout(i,e||t))};return r.delay=function(e){n&&clearTimeout(n),n=setTimeout(i,e||t)},r.schedule=r,r.call=function(){this.cancel(),e()},r.cancel=function(){n&&clearTimeout(n),n=null},r.isPending=function(){return n},r}})),ace.define("ace/keyboard/textinput_ios",["require","exports","module","ace/lib/event","ace/lib/useragent","ace/lib/dom","ace/lib/lang","ace/lib/keys"],(function(e,t,n){"use strict";var i=e("../lib/event"),r=e("../lib/useragent"),o=e("../lib/dom"),s=e("../lib/lang"),a=e("../lib/keys"),l=a.KEY_MODS,u=r.isChrome<18,c=r.isIE;t.TextInput=function(e,t){var n=o.createElement("textarea");n.className=r.isIOS?"ace_text-input ace_text-input-ios":"ace_text-input",r.isTouchPad&&n.setAttribute("x-palm-disable-auto-cap",!0),n.setAttribute("wrap","off"),n.setAttribute("autocorrect","off"),n.setAttribute("autocapitalize","off"),n.setAttribute("spellcheck",!1),n.style.opacity="0",e.insertBefore(n,e.firstChild);var h=!1,f=!1,d=!1,p=!1,g="",m=!0;try{var v=document.activeElement===n}catch(e){}i.addListener(n,"blur",(function(e){t.onBlur(e),v=!1})),i.addListener(n,"focus",(function(e){v=!0,t.onFocus(e),w()})),this.focus=function(){if(g)return n.focus();n.style.position="fixed",n.focus()},this.blur=function(){n.blur()},this.isFocused=function(){return v};var b=s.delayedCall((function(){v&&w(m)})),y=s.delayedCall((function(){p||(n.value="\n aaaa a\n",v&&w())}));function w(e){if(!p){if(p=!0,C)t=0,i=e?0:n.value.length-1;else var t=4,i=5;try{n.setSelectionRange(t,i)}catch(e){}p=!1}}function A(){p||(n.value="\n aaaa a\n",r.isWebKit&&y.schedule())}r.isWebKit||t.addEventListener("changeSelection",(function(){t.selection.isEmpty()!=m&&(m=!m,b.schedule())})),A(),v&&t.onFocus();var C=null;this.setInputHandler=function(e){C=e},this.getInputHandler=function(){return C};var E=!1,F=function(e){4===n.selectionStart&&5===n.selectionEnd||(C&&(e=C(e),C=null),d?(w(),e&&t.onPaste(e),d=!1):e=="\n aaaa a\n".substr(0)&&4===n.selectionStart?E?t.execCommand("del",{source:"ace"}):t.execCommand("backspace",{source:"ace"}):h||("\n aaaa a\n"==e.substring(0,9)&&e.length>"\n aaaa a\n".length?e=e.substr(9):e.substr(0,4)=="\n aaaa a\n".substr(0,4)?e=e.substr(4,e.length-"\n aaaa a\n".length+1):e.charAt(e.length-1)=="\n aaaa a\n".charAt(0)&&(e=e.slice(0,-1)),e=="\n aaaa a\n".charAt(0)||e.charAt(e.length-1)=="\n aaaa a\n".charAt(0)&&(e=e.slice(0,-1)),e&&t.onTextInput(e)),h&&(h=!1),E&&(E=!1))},x=function(e){if(!p){var t=n.value;F(t),A()}},k=function(e,t,n){var i=e.clipboardData||window.clipboardData;if(i&&!u){var r=c||n?"Text":"text/plain";try{return t?!1!==i.setData(r,t):i.getData(r)}catch(e){if(!n)return k(e,t,!0)}}},S=function(e,o){var s=t.getCopyText();if(!s)return i.preventDefault(e);k(e,s)?(r.isIOS&&(f=o,n.value="\n aa"+s+"a a\n",n.setSelectionRange(4,4+s.length),h={value:s}),o?t.onCut():t.onCopy(),r.isIOS||i.preventDefault(e)):(h=!0,n.value=s,n.select(),setTimeout((function(){h=!1,A(),w(),o?t.onCut():t.onCopy()})))};i.addCommandKeyListener(n,t.onCommandKey.bind(t)),i.addListener(n,"select",(function(e){!function(e){return 0===e.selectionStart&&e.selectionEnd===e.value.length}(n)?C&&w(t.selection.isEmpty()):(t.selectAll(),w())})),i.addListener(n,"input",x),i.addListener(n,"cut",(function(e){S(e,!0)})),i.addListener(n,"copy",(function(e){S(e,!1)})),i.addListener(n,"paste",(function(e){var o=k(e);"string"==typeof o?(o&&t.onPaste(o,e),r.isIE&&setTimeout(w),i.preventDefault(e)):(n.value="",d=!0)}));var _,D=function(){if(p&&t.onCompositionUpdate&&!t.$readOnly){var e=n.value.replace(/\x01/g,"");if(p.lastValue!==e&&(t.onCompositionUpdate(e),p.lastValue&&t.undo(),p.canUndo&&(p.lastValue=e),p.lastValue)){var i=t.selection.getRange();t.insert(p.lastValue),t.session.markUndoGroup(),p.range=t.selection.getRange(),t.selection.setRange(i),t.selection.clearSelection()}}},$=function(e){if(t.onCompositionEnd&&!t.$readOnly){var i=p;p=!1;var o=setTimeout((function(){o=null;var e=n.value.replace(/\x01/g,"");p||(e==i.lastValue?A():!i.lastValue&&e&&(A(),F(e)))}));C=function(e){return o&&clearTimeout(o),(e=e.replace(/\x01/g,""))==i.lastValue?"":(i.lastValue&&o&&t.undo(),e)},t.onCompositionEnd(),t.removeListener("mousedown",$),"compositionend"==e.type&&i.range&&t.selection.setRange(i.range),(!!r.isChrome&&r.isChrome>=53||!!r.isWebKit&&r.isWebKit>=603)&&x()}},B=s.delayedCall(D,50);function L(){clearTimeout(_),_=setTimeout((function(){g&&(n.style.cssText=g,g=""),null==t.renderer.$keepTextAreaAtCursor&&(t.renderer.$keepTextAreaAtCursor=!0,t.renderer.$moveTextAreaToCursor())}),0)}i.addListener(n,"compositionstart",(function(e){p||!t.onCompositionStart||t.$readOnly||((p={}).canUndo=t.session.$undoManager,t.onCompositionStart(),setTimeout(D,0),t.on("mousedown",$),p.canUndo&&!t.selection.isEmpty()&&(t.insert(""),t.session.markUndoGroup(),t.selection.clearSelection()),t.session.markUndoGroup())})),r.isGecko?i.addListener(n,"text",(function(){B.schedule()})):(i.addListener(n,"keyup",(function(){B.schedule()})),i.addListener(n,"keydown",(function(){B.schedule()}))),i.addListener(n,"compositionend",$),this.getElement=function(){return n},this.setReadOnly=function(e){n.readOnly=e},this.onContextMenu=function(e){E=!0,w(t.selection.isEmpty()),t._emit("nativecontextmenu",{target:t,domEvent:e}),this.moveToMouse(e,!0)},this.moveToMouse=function(e,s){g||(g=n.style.cssText),n.style.cssText=(s?"z-index:100000;":"")+"height:"+n.style.height+";"+(r.isIE?"opacity:0.1;":"");var a=t.container.getBoundingClientRect(),l=o.computedStyle(t.container),u=a.top+(parseInt(l.borderTopWidth)||0),c=a.left+(parseInt(a.borderLeftWidth)||0),h=a.bottom-u-n.clientHeight-2,f=function(e){n.style.left=e.clientX-c-2+"px",n.style.top=Math.min(e.clientY-u-2,h)+"px"};f(e),"mousedown"==e.type&&(t.renderer.$keepTextAreaAtCursor&&(t.renderer.$keepTextAreaAtCursor=null),clearTimeout(_),r.isWin&&i.capture(t.container,f,L))},this.onContextMenuClose=L;var T=function(e){t.textInput.onContextMenu(e),L()};if(i.addListener(n,"mouseup",T),i.addListener(n,"mousedown",(function(e){e.preventDefault(),L()})),i.addListener(t.renderer.scroller,"contextmenu",T),i.addListener(n,"contextmenu",T),r.isIOS){var R=null,O=!1;e.addEventListener("keydown",(function(e){R&&clearTimeout(R),O=!0})),e.addEventListener("keyup",(function(e){R=setTimeout((function(){O=!1}),100)}));var M=function(e){if(document.activeElement===n&&!O){if(f)return setTimeout((function(){f=!1}),100);var i=n.selectionStart,r=n.selectionEnd;if(n.setSelectionRange(4,5),i==r)switch(i){case 0:t.onCommandKey(null,0,a.up);break;case 1:t.onCommandKey(null,0,a.home);break;case 2:t.onCommandKey(null,l.option,a.left);break;case 4:t.onCommandKey(null,0,a.left);break;case 5:t.onCommandKey(null,0,a.right);break;case 7:t.onCommandKey(null,l.option,a.right);break;case 8:t.onCommandKey(null,0,a.end);break;case 9:t.onCommandKey(null,0,a.down)}else{switch(r){case 6:t.onCommandKey(null,l.shift,a.right);break;case 7:t.onCommandKey(null,l.shift|l.option,a.right);break;case 8:t.onCommandKey(null,l.shift,a.end);break;case 9:t.onCommandKey(null,l.shift,a.down)}switch(i){case 0:t.onCommandKey(null,l.shift,a.up);break;case 1:t.onCommandKey(null,l.shift,a.home);break;case 2:t.onCommandKey(null,l.shift|l.option,a.left);break;case 3:t.onCommandKey(null,l.shift,a.left)}}}};document.addEventListener("selectionchange",M),t.on("destroy",(function(){document.removeEventListener("selectionchange",M)}))}}})),ace.define("ace/keyboard/textinput",["require","exports","module","ace/lib/event","ace/lib/useragent","ace/lib/dom","ace/lib/lang","ace/keyboard/textinput_ios"],(function(e,t,n){"use strict";var i=e("../lib/event"),r=e("../lib/useragent"),o=e("../lib/dom"),s=e("../lib/lang"),a=r.isChrome<18,l=r.isIE,u=e("./textinput_ios").TextInput;t.TextInput=function(e,t){if(r.isIOS)return u.call(this,e,t);var n=o.createElement("textarea");n.className="ace_text-input",n.setAttribute("wrap","off"),n.setAttribute("autocorrect","off"),n.setAttribute("autocapitalize","off"),n.setAttribute("spellcheck",!1),n.style.opacity="0",e.insertBefore(n,e.firstChild);var c=!1,h=!1,f=!1,d="",p=!0;try{var g=document.activeElement===n}catch(e){}i.addListener(n,"blur",(function(e){t.onBlur(e),g=!1})),i.addListener(n,"focus",(function(e){g=!0,t.onFocus(e),b()})),this.focus=function(){if(d)return n.focus();var e=n.style.top;n.style.position="fixed",n.style.top="0px",n.focus(),setTimeout((function(){n.style.position="","0px"==n.style.top&&(n.style.top=e)}),0)},this.blur=function(){n.blur()},this.isFocused=function(){return g};var m=s.delayedCall((function(){g&&b(p)})),v=s.delayedCall((function(){f||(n.value="\u2028\u2028",g&&b())}));function b(e){if(!f){if(f=!0,w)var t=0,i=e?0:n.value.length-1;else t=e?2:1,i=2;try{n.setSelectionRange(t,i)}catch(e){}f=!1}}function y(){f||(n.value="\u2028\u2028",r.isWebKit&&v.schedule())}r.isWebKit||t.addEventListener("changeSelection",(function(){t.selection.isEmpty()!=p&&(p=!p,m.schedule())})),y(),g&&t.onFocus();var w=null;this.setInputHandler=function(e){w=e},this.getInputHandler=function(){return w};var A=!1,C=function(e){w&&(e=w(e),w=null),h?(b(),e&&t.onPaste(e),h=!1):e=="\u2028\u2028".charAt(0)?A?t.execCommand("del",{source:"ace"}):t.execCommand("backspace",{source:"ace"}):("\u2028\u2028"==e.substring(0,2)?e=e.substr(2):e.charAt(0)=="\u2028\u2028".charAt(0)?e=e.substr(1):e.charAt(e.length-1)=="\u2028\u2028".charAt(0)&&(e=e.slice(0,-1)),e.charAt(e.length-1)=="\u2028\u2028".charAt(0)&&(e=e.slice(0,-1)),e&&t.onTextInput(e)),A&&(A=!1)},E=function(e){if(!f){var t=n.value;C(t),y()}},F=function(e,t,n){var i=e.clipboardData||window.clipboardData;if(i&&!a){var r=l||n?"Text":"text/plain";try{return t?!1!==i.setData(r,t):i.getData(r)}catch(e){if(!n)return F(e,t,!0)}}},x=function(e,r){var o=t.getCopyText();if(!o)return i.preventDefault(e);F(e,o)?(r?t.onCut():t.onCopy(),i.preventDefault(e)):(c=!0,n.value=o,n.select(),setTimeout((function(){c=!1,y(),b(),r?t.onCut():t.onCopy()})))},k=function(e){x(e,!0)},S=function(e){x(e,!1)},_=function(e){var o=F(e);"string"==typeof o?(o&&t.onPaste(o,e),r.isIE&&setTimeout(b),i.preventDefault(e)):(n.value="",h=!0)};i.addCommandKeyListener(n,t.onCommandKey.bind(t)),i.addListener(n,"select",(function(e){c?c=!1:!function(e){return 0===e.selectionStart&&e.selectionEnd===e.value.length}(n)?w&&b(t.selection.isEmpty()):(t.selectAll(),b())})),i.addListener(n,"input",E),i.addListener(n,"cut",k),i.addListener(n,"copy",S),i.addListener(n,"paste",_),"oncut"in n&&"oncopy"in n&&"onpaste"in n||i.addListener(e,"keydown",(function(e){if((!r.isMac||e.metaKey)&&e.ctrlKey)switch(e.keyCode){case 67:S(e);break;case 86:_(e);break;case 88:k(e)}}));var D,$=function(){if(f&&t.onCompositionUpdate&&!t.$readOnly){var e=n.value.replace(/\u2028/g,"");if(f.lastValue!==e&&(t.onCompositionUpdate(e),f.lastValue&&t.undo(),f.canUndo&&(f.lastValue=e),f.lastValue)){var i=t.selection.getRange();t.insert(f.lastValue),t.session.markUndoGroup(),f.range=t.selection.getRange(),t.selection.setRange(i),t.selection.clearSelection()}}},B=function(e){if(t.onCompositionEnd&&!t.$readOnly){var i=f;f=!1;var o=setTimeout((function(){o=null;var e=n.value.replace(/\u2028/g,"");f||(e==i.lastValue?y():!i.lastValue&&e&&(y(),C(e)))}));w=function(e){return o&&clearTimeout(o),(e=e.replace(/\u2028/g,""))==i.lastValue?"":(i.lastValue&&o&&t.undo(),e)},t.onCompositionEnd(),t.removeListener("mousedown",B),"compositionend"==e.type&&i.range&&t.selection.setRange(i.range),(!!r.isChrome&&r.isChrome>=53||!!r.isWebKit&&r.isWebKit>=603)&&E()}},L=s.delayedCall($,50);function T(){clearTimeout(D),D=setTimeout((function(){d&&(n.style.cssText=d,d=""),null==t.renderer.$keepTextAreaAtCursor&&(t.renderer.$keepTextAreaAtCursor=!0,t.renderer.$moveTextAreaToCursor())}),0)}i.addListener(n,"compositionstart",(function(e){f||!t.onCompositionStart||t.$readOnly||((f={}).canUndo=t.session.$undoManager,t.onCompositionStart(),setTimeout($,0),t.on("mousedown",B),f.canUndo&&!t.selection.isEmpty()&&(t.insert(""),t.session.markUndoGroup(),t.selection.clearSelection()),t.session.markUndoGroup())})),r.isGecko?i.addListener(n,"text",(function(){L.schedule()})):(i.addListener(n,"keyup",(function(){L.schedule()})),i.addListener(n,"keydown",(function(){L.schedule()}))),i.addListener(n,"compositionend",B),this.getElement=function(){return n},this.setReadOnly=function(e){n.readOnly=e},this.onContextMenu=function(e){A=!0,b(t.selection.isEmpty()),t._emit("nativecontextmenu",{target:t,domEvent:e}),this.moveToMouse(e,!0)},this.moveToMouse=function(e,s){d||(d=n.style.cssText),n.style.cssText=(s?"z-index:100000;":"")+"height:"+n.style.height+";"+(r.isIE?"opacity:0.1;":"");var a=t.container.getBoundingClientRect(),l=o.computedStyle(t.container),u=a.top+(parseInt(l.borderTopWidth)||0),c=a.left+(parseInt(a.borderLeftWidth)||0),h=a.bottom-u-n.clientHeight-2,f=function(e){n.style.left=e.clientX-c-2+"px",n.style.top=Math.min(e.clientY-u-2,h)+"px"};f(e),"mousedown"==e.type&&(t.renderer.$keepTextAreaAtCursor&&(t.renderer.$keepTextAreaAtCursor=null),clearTimeout(D),r.isWin&&i.capture(t.container,f,T))},this.onContextMenuClose=T;var R=function(e){t.textInput.onContextMenu(e),T()};i.addListener(n,"mouseup",R),i.addListener(n,"mousedown",(function(e){e.preventDefault(),T()})),i.addListener(t.renderer.scroller,"contextmenu",R),i.addListener(n,"contextmenu",R)}})),ace.define("ace/mouse/default_handlers",["require","exports","module","ace/lib/dom","ace/lib/event","ace/lib/useragent"],(function(e,t,n){"use strict";e("../lib/dom"),e("../lib/event");var i=e("../lib/useragent");function r(e){e.$clickSelection=null;var t=e.editor;t.setDefaultHandler("mousedown",this.onMouseDown.bind(e)),t.setDefaultHandler("dblclick",this.onDoubleClick.bind(e)),t.setDefaultHandler("tripleclick",this.onTripleClick.bind(e)),t.setDefaultHandler("quadclick",this.onQuadClick.bind(e)),t.setDefaultHandler("mousewheel",this.onMouseWheel.bind(e)),t.setDefaultHandler("touchmove",this.onTouchMove.bind(e));["select","startSelect","selectEnd","selectAllEnd","selectByWordsEnd","selectByLinesEnd","dragWait","dragWaitEnd","focusWait"].forEach((function(t){e[t]=this[t]}),this),e.selectByLines=this.extendSelectionBy.bind(e,"getLineRange"),e.selectByWords=this.extendSelectionBy.bind(e,"getWordRange")}function o(e,t){if(e.start.row==e.end.row)var n=2*t.column-e.start.column-e.end.column;else if(e.start.row!=e.end.row-1||e.start.column||e.end.column)n=2*t.row-e.start.row-e.end.row;else var n=t.column-4;return n<0?{cursor:e.start,anchor:e.end}:{cursor:e.end,anchor:e.start}}(function(){this.onMouseDown=function(e){var t=e.inSelection(),n=e.getDocumentPosition();this.mousedownEvent=e;var r=this.editor,o=e.getButton();if(0!==o){var s=r.getSelectionRange().isEmpty();return r.$blockScrolling++,(s||1==o)&&r.selection.moveToPosition(n),r.$blockScrolling--,void(2==o&&(r.textInput.onContextMenu(e.domEvent),i.isMozilla||e.preventDefault()))}return this.mousedownEvent.time=Date.now(),!t||r.isFocused()||(r.focus(),!this.$focusTimout||this.$clickSelection||r.inMultiSelectMode)?(this.captureMouse(e),this.startSelect(n,e.domEvent._clicks>1),e.preventDefault()):(this.setState("focusWait"),void this.captureMouse(e))},this.startSelect=function(e,t){e=e||this.editor.renderer.screenToTextCoordinates(this.x,this.y);var n=this.editor;n.$blockScrolling++,this.mousedownEvent.getShiftKey()?n.selection.selectToPosition(e):t||n.selection.moveToPosition(e),t||this.select(),n.renderer.scroller.setCapture&&n.renderer.scroller.setCapture(),n.setStyle("ace_selecting"),this.setState("select"),n.$blockScrolling--},this.select=function(){var e,t=this.editor,n=t.renderer.screenToTextCoordinates(this.x,this.y);if(t.$blockScrolling++,this.$clickSelection){var i=this.$clickSelection.comparePoint(n);if(-1==i)e=this.$clickSelection.end;else if(1==i)e=this.$clickSelection.start;else{var r=o(this.$clickSelection,n);n=r.cursor,e=r.anchor}t.selection.setSelectionAnchor(e.row,e.column)}t.selection.selectToPosition(n),t.$blockScrolling--,t.renderer.scrollCursorIntoView()},this.extendSelectionBy=function(e){var t,n=this.editor,i=n.renderer.screenToTextCoordinates(this.x,this.y),r=n.selection[e](i.row,i.column);if(n.$blockScrolling++,this.$clickSelection){var s=this.$clickSelection.comparePoint(r.start),a=this.$clickSelection.comparePoint(r.end);if(-1==s&&a<=0)t=this.$clickSelection.end,r.end.row==i.row&&r.end.column==i.column||(i=r.start);else if(1==a&&s>=0)t=this.$clickSelection.start,r.start.row==i.row&&r.start.column==i.column||(i=r.end);else if(-1==s&&1==a)i=r.end,t=r.start;else{var l=o(this.$clickSelection,i);i=l.cursor,t=l.anchor}n.selection.setSelectionAnchor(t.row,t.column)}n.selection.selectToPosition(i),n.$blockScrolling--,n.renderer.scrollCursorIntoView()},this.selectEnd=this.selectAllEnd=this.selectByWordsEnd=this.selectByLinesEnd=function(){this.$clickSelection=null,this.editor.unsetStyle("ace_selecting"),this.editor.renderer.scroller.releaseCapture&&this.editor.renderer.scroller.releaseCapture()},this.focusWait=function(){var e,t,n,i,r=(e=this.mousedownEvent.x,t=this.mousedownEvent.y,n=this.x,i=this.y,Math.sqrt(Math.pow(n-e,2)+Math.pow(i-t,2))),o=Date.now();(r>0||o-this.mousedownEvent.time>this.$focusTimout)&&this.startSelect(this.mousedownEvent.getDocumentPosition())},this.onDoubleClick=function(e){var t=e.getDocumentPosition(),n=this.editor,i=n.session.getBracketRange(t);i?(i.isEmpty()&&(i.start.column--,i.end.column++),this.setState("select")):(i=n.selection.getWordRange(t.row,t.column),this.setState("selectByWords")),this.$clickSelection=i,this.select()},this.onTripleClick=function(e){var t=e.getDocumentPosition(),n=this.editor;this.setState("selectByLines");var i=n.getSelectionRange();i.isMultiLine()&&i.contains(t.row,t.column)?(this.$clickSelection=n.selection.getLineRange(i.start.row),this.$clickSelection.end=n.selection.getLineRange(i.end.row).end):this.$clickSelection=n.selection.getLineRange(t.row),this.select()},this.onQuadClick=function(e){var t=this.editor;t.selectAll(),this.$clickSelection=t.getSelectionRange(),this.setState("selectAll")},this.onMouseWheel=function(e){if(!e.getAccelKey()){e.getShiftKey()&&e.wheelY&&!e.wheelX&&(e.wheelX=e.wheelY,e.wheelY=0);var t=this.editor;this.$lastScroll||(this.$lastScroll={t:0,vx:0,vy:0,allowed:0});var n=this.$lastScroll,i=e.domEvent.timeStamp,r=i-n.t,o=e.wheelX/r,s=e.wheelY/r;r<250&&(o=(o+n.vx)/2,s=(s+n.vy)/2);var a=Math.abs(o/s),l=!1;if(a>=1&&t.renderer.isScrollableBy(e.wheelX*e.speed,0)&&(l=!0),a<=1&&t.renderer.isScrollableBy(0,e.wheelY*e.speed)&&(l=!0),l)n.allowed=i;else if(i-n.allowed<250){Math.abs(o)<=1.1*Math.abs(n.vx)&&Math.abs(s)<=1.1*Math.abs(n.vy)?(l=!0,n.allowed=i):n.allowed=0}return n.t=i,n.vx=o,n.vy=s,l?(t.renderer.scrollBy(e.wheelX*e.speed,e.wheelY*e.speed),e.stop()):void 0}},this.onTouchMove=function(e){this.editor._emit("mousewheel",e)}}).call(r.prototype),t.DefaultHandlers=r})),ace.define("ace/tooltip",["require","exports","module","ace/lib/oop","ace/lib/dom"],(function(e,t,n){"use strict";e("./lib/oop");var i=e("./lib/dom");function r(e){this.isOpen=!1,this.$element=null,this.$parentNode=e}(function(){this.$init=function(){return this.$element=i.createElement("div"),this.$element.className="ace_tooltip",this.$element.style.display="none",this.$parentNode.appendChild(this.$element),this.$element},this.getElement=function(){return this.$element||this.$init()},this.setText=function(e){i.setInnerText(this.getElement(),e)},this.setHtml=function(e){this.getElement().innerHTML=e},this.setPosition=function(e,t){this.getElement().style.left=e+"px",this.getElement().style.top=t+"px"},this.setClassName=function(e){i.addCssClass(this.getElement(),e)},this.show=function(e,t,n){null!=e&&this.setText(e),null!=t&&null!=n&&this.setPosition(t,n),this.isOpen||(this.getElement().style.display="block",this.isOpen=!0)},this.hide=function(){this.isOpen&&(this.getElement().style.display="none",this.isOpen=!1)},this.getHeight=function(){return this.getElement().offsetHeight},this.getWidth=function(){return this.getElement().offsetWidth},this.destroy=function(){this.isOpen=!1,this.$element&&this.$element.parentNode&&this.$element.parentNode.removeChild(this.$element)}}).call(r.prototype),t.Tooltip=r})),ace.define("ace/mouse/default_gutter_handler",["require","exports","module","ace/lib/dom","ace/lib/oop","ace/lib/event","ace/tooltip"],(function(e,t,n){"use strict";var i=e("../lib/dom"),r=e("../lib/oop"),o=e("../lib/event"),s=e("../tooltip").Tooltip;function a(e){s.call(this,e)}r.inherits(a,s),function(){this.setPosition=function(e,t){var n=window.innerWidth||document.documentElement.clientWidth,i=window.innerHeight||document.documentElement.clientHeight,r=this.getWidth(),o=this.getHeight();(e+=15)+r>n&&(e-=e+r-n),(t+=15)+o>i&&(t-=20+o),s.prototype.setPosition.call(this,e,t)}}.call(a.prototype),t.GutterHandler=function(e){var t,n,r,s=e.editor,l=s.renderer.$gutterLayer,u=new a(s.container);function c(){t&&(t=clearTimeout(t)),r&&(u.hide(),r=null,s._signal("hideGutterTooltip",u),s.removeEventListener("mousewheel",c))}function h(e){u.setPosition(e.x,e.y)}e.editor.setDefaultHandler("guttermousedown",(function(t){if(s.isFocused()&&0==t.getButton()&&"foldWidgets"!=l.getRegion(t)){var n=t.getDocumentPosition().row,i=s.session.selection;if(t.getShiftKey())i.selectTo(n,0);else{if(2==t.domEvent.detail)return s.selectAll(),t.preventDefault();e.$clickSelection=s.selection.getLineRange(n)}return e.setState("selectByLines"),e.captureMouse(t),t.preventDefault()}})),e.editor.setDefaultHandler("guttermousemove",(function(o){var a=o.domEvent.target||o.domEvent.srcElement;if(i.hasCssClass(a,"ace_fold-widget"))return c();r&&e.$tooltipFollowsMouse&&h(o),n=o,t||(t=setTimeout((function(){t=null,n&&!e.isMousePressed?function(){var t=n.getDocumentPosition().row,i=l.$annotations[t];if(!i)return c();if(t==s.session.getLength()){var o=s.renderer.pixelToScreenCoordinates(0,n.y).row,a=n.$pos;if(o>s.session.documentToScreenRow(a.row,a.column))return c()}if(r!=i)if(r=i.text.join("<br/>"),u.setHtml(r),u.show(),s._signal("showGutterTooltip",u),s.on("mousewheel",c),e.$tooltipFollowsMouse)h(n);else{var f=n.domEvent.target.getBoundingClientRect(),d=u.getElement().style;d.left=f.right+"px",d.top=f.bottom+"px"}}():c()}),50))})),o.addListener(s.renderer.$gutter,"mouseout",(function(e){n=null,r&&!t&&(t=setTimeout((function(){t=null,c()}),50))})),s.on("changeSession",c)}})),ace.define("ace/mouse/mouse_event",["require","exports","module","ace/lib/event","ace/lib/useragent"],(function(e,t,n){"use strict";var i=e("../lib/event"),r=e("../lib/useragent"),o=t.MouseEvent=function(e,t){this.domEvent=e,this.editor=t,this.x=this.clientX=e.clientX,this.y=this.clientY=e.clientY,this.$pos=null,this.$inSelection=null,this.propagationStopped=!1,this.defaultPrevented=!1};(function(){this.stopPropagation=function(){i.stopPropagation(this.domEvent),this.propagationStopped=!0},this.preventDefault=function(){i.preventDefault(this.domEvent),this.defaultPrevented=!0},this.stop=function(){this.stopPropagation(),this.preventDefault()},this.getDocumentPosition=function(){return this.$pos?this.$pos:(this.$pos=this.editor.renderer.screenToTextCoordinates(this.clientX,this.clientY),this.$pos)},this.inSelection=function(){if(null!==this.$inSelection)return this.$inSelection;var e=this.editor.getSelectionRange();if(e.isEmpty())this.$inSelection=!1;else{var t=this.getDocumentPosition();this.$inSelection=e.contains(t.row,t.column)}return this.$inSelection},this.getButton=function(){return i.getButton(this.domEvent)},this.getShiftKey=function(){return this.domEvent.shiftKey},this.getAccelKey=r.isMac?function(){return this.domEvent.metaKey}:function(){return this.domEvent.ctrlKey}}).call(o.prototype)})),ace.define("ace/mouse/dragdrop_handler",["require","exports","module","ace/lib/dom","ace/lib/event","ace/lib/useragent"],(function(e,t,n){"use strict";var i=e("../lib/dom"),r=e("../lib/event"),o=e("../lib/useragent");function s(e){var t=e.editor,n=i.createElement("img");n.src="data:image/gif;base64,R0lGODlhAQABAAAAACH5BAEKAAEALAAAAAABAAEAAAICTAEAOw==",o.isOpera&&(n.style.cssText="width:1px;height:1px;position:fixed;top:0;left:0;z-index:2147483647;opacity:0;");["dragWait","dragWaitEnd","startDrag","dragReadyEnd","onMouseDrag"].forEach((function(t){e[t]=this[t]}),this),t.addEventListener("mousedown",this.onMouseDown.bind(e));var s,l,u,c,h,f,d,p,g,m,v,b=t.container,y=0;function w(){var e=f;(function(e,n){var i=Date.now(),r=!n||e.row!=n.row,o=!n||e.column!=n.column;!m||r||o?(t.$blockScrolling+=1,t.moveCursorToPosition(e),t.$blockScrolling-=1,m=i,v={x:l,y:u}):a(v.x,v.y,l,u)>5?m=null:i-m>=200&&(t.renderer.scrollCursorIntoView(),m=null)})(f=t.renderer.screenToTextCoordinates(l,u),e),function(e,n){var i=Date.now(),r=t.renderer.layerConfig.lineHeight,o=t.renderer.layerConfig.characterWidth,s=t.renderer.scroller.getBoundingClientRect(),a={x:{left:l-s.left,right:s.right-l},y:{top:u-s.top,bottom:s.bottom-u}},c=Math.min(a.x.left,a.x.right),h=Math.min(a.y.top,a.y.bottom),f={row:e.row,column:e.column};c/o<=2&&(f.column+=a.x.left<a.x.right?-3:2),h/r<=1&&(f.row+=a.y.top<a.y.bottom?-1:1);var d=e.row!=f.row,p=e.column!=f.column,m=!n||e.row!=n.row;d||p&&!m?g?i-g>=200&&t.renderer.scrollCursorIntoView(f):g=i:g=null}(f,e)}function A(){h=t.selection.toOrientedRange(),s=t.session.addMarker(h,"ace_selection",t.getSelectionStyle()),t.clearSelection(),t.isFocused()&&t.renderer.$cursorLayer.setBlinking(!1),clearInterval(c),w(),c=setInterval(w,20),y=0,r.addListener(document,"mousemove",F)}function C(){clearInterval(c),t.session.removeMarker(s),s=null,t.$blockScrolling+=1,t.selection.fromOrientedRange(h),t.$blockScrolling-=1,t.isFocused()&&!p&&t.renderer.$cursorLayer.setBlinking(!t.getReadOnly()),h=null,f=null,y=0,g=null,m=null,r.removeListener(document,"mousemove",F)}this.onDragStart=function(e){if(this.cancelDrag||!b.draggable){var i=this;return setTimeout((function(){i.startSelect(),i.captureMouse(e)}),0),e.preventDefault()}h=t.getSelectionRange();var r=e.dataTransfer;r.effectAllowed=t.getReadOnly()?"copy":"copyMove",o.isOpera&&(t.container.appendChild(n),n.scrollTop=0),r.setDragImage&&r.setDragImage(n,0,0),o.isOpera&&t.container.removeChild(n),r.clearData(),r.setData("Text",t.session.getTextRange()),p=!0,this.setState("drag")},this.onDragEnd=function(e){if(b.draggable=!1,p=!1,this.setState(null),!t.getReadOnly()){var n=e.dataTransfer.dropEffect;d||"move"!=n||t.session.remove(t.getSelectionRange()),t.renderer.$cursorLayer.setBlinking(!0)}this.editor.unsetStyle("ace_dragging"),this.editor.renderer.setCursorStyle("")},this.onDragEnter=function(e){if(!t.getReadOnly()&&x(e.dataTransfer))return l=e.clientX,u=e.clientY,s||A(),y++,e.dataTransfer.dropEffect=d=k(e),r.preventDefault(e)},this.onDragOver=function(e){if(!t.getReadOnly()&&x(e.dataTransfer))return l=e.clientX,u=e.clientY,s||(A(),y++),null!==E&&(E=null),e.dataTransfer.dropEffect=d=k(e),r.preventDefault(e)},this.onDragLeave=function(e){if(--y<=0&&s)return C(),d=null,r.preventDefault(e)},this.onDrop=function(e){if(f){var n=e.dataTransfer;if(p)switch(d){case"move":h=h.contains(f.row,f.column)?{start:f,end:f}:t.moveText(h,f);break;case"copy":h=t.moveText(h,f,!0)}else{var i=n.getData("Text");h={start:f,end:t.session.insert(f,i)},t.focus(),d=null}return C(),r.preventDefault(e)}},r.addListener(b,"dragstart",this.onDragStart.bind(e)),r.addListener(b,"dragend",this.onDragEnd.bind(e)),r.addListener(b,"dragenter",this.onDragEnter.bind(e)),r.addListener(b,"dragover",this.onDragOver.bind(e)),r.addListener(b,"dragleave",this.onDragLeave.bind(e)),r.addListener(b,"drop",this.onDrop.bind(e));var E=null;function F(){null==E&&(E=setTimeout((function(){null!=E&&s&&C()}),20))}function x(e){var t=e.types;return!t||Array.prototype.some.call(t,(function(e){return"text/plain"==e||"Text"==e}))}function k(e){var t=["copy","copymove","all","uninitialized"],n=o.isMac?e.altKey:e.ctrlKey,i="uninitialized";try{i=e.dataTransfer.effectAllowed.toLowerCase()}catch(e){}var r="none";return n&&t.indexOf(i)>=0?r="copy":["move","copymove","linkmove","all","uninitialized"].indexOf(i)>=0?r="move":t.indexOf(i)>=0&&(r="copy"),r}}function a(e,t,n,i){return Math.sqrt(Math.pow(n-e,2)+Math.pow(i-t,2))}(function(){this.dragWait=function(){Date.now()-this.mousedownEvent.time>this.editor.getDragDelay()&&this.startDrag()},this.dragWaitEnd=function(){this.editor.container.draggable=!1,this.startSelect(this.mousedownEvent.getDocumentPosition()),this.selectEnd()},this.dragReadyEnd=function(e){this.editor.renderer.$cursorLayer.setBlinking(!this.editor.getReadOnly()),this.editor.unsetStyle("ace_dragging"),this.editor.renderer.setCursorStyle(""),this.dragWaitEnd()},this.startDrag=function(){this.cancelDrag=!1;var e=this.editor;e.container.draggable=!0,e.renderer.$cursorLayer.setBlinking(!1),e.setStyle("ace_dragging");var t=o.isWin?"default":"move";e.renderer.setCursorStyle(t),this.setState("dragReady")},this.onMouseDrag=function(e){var t=this.editor.container;o.isIE&&"dragReady"==this.state&&(a(this.mousedownEvent.x,this.mousedownEvent.y,this.x,this.y)>3&&t.dragDrop());"dragWait"===this.state&&(a(this.mousedownEvent.x,this.mousedownEvent.y,this.x,this.y)>0&&(t.draggable=!1,this.startSelect(this.mousedownEvent.getDocumentPosition())))},this.onMouseDown=function(e){if(this.$dragEnabled){this.mousedownEvent=e;var t=this.editor,n=e.inSelection(),i=e.getButton();if(1===(e.domEvent.detail||1)&&0===i&&n){if(e.editor.inMultiSelectMode&&(e.getAccelKey()||e.getShiftKey()))return;this.mousedownEvent.time=Date.now();var r=e.domEvent.target||e.domEvent.srcElement;if("unselectable"in r&&(r.unselectable="on"),t.getDragDelay()){if(o.isWebKit)this.cancelDrag=!0,t.container.draggable=!0;this.setState("dragWait")}else this.startDrag();this.captureMouse(e,this.onMouseDrag.bind(this)),e.defaultPrevented=!0}}}}).call(s.prototype),t.DragdropHandler=s})),ace.define("ace/lib/net",["require","exports","module","ace/lib/dom"],(function(e,t,n){"use strict";var i=e("./dom");t.get=function(e,t){var n=new XMLHttpRequest;n.open("GET",e,!0),n.onreadystatechange=function(){4===n.readyState&&t(n.responseText)},n.send(null)},t.loadScript=function(e,t){var n=i.getDocumentHead(),r=document.createElement("script");r.src=e,n.appendChild(r),r.onload=r.onreadystatechange=function(e,n){!n&&r.readyState&&"loaded"!=r.readyState&&"complete"!=r.readyState||(r=r.onload=r.onreadystatechange=null,n||t())}},t.qualifyURL=function(e){var t=document.createElement("a");return t.href=e,t.href}})),ace.define("ace/lib/event_emitter",["require","exports","module"],(function(e,t,n){"use strict";var i={},r=function(){this.propagationStopped=!0},o=function(){this.defaultPrevented=!0};i._emit=i._dispatchEvent=function(e,t){this._eventRegistry||(this._eventRegistry={}),this._defaultHandlers||(this._defaultHandlers={});var n=this._eventRegistry[e]||[],i=this._defaultHandlers[e];if(n.length||i){"object"==typeof t&&t||(t={}),t.type||(t.type=e),t.stopPropagation||(t.stopPropagation=r),t.preventDefault||(t.preventDefault=o),n=n.slice();for(var s=0;s<n.length&&(n[s](t,this),!t.propagationStopped);s++);return i&&!t.defaultPrevented?i(t,this):void 0}},i._signal=function(e,t){var n=(this._eventRegistry||{})[e];if(n){n=n.slice();for(var i=0;i<n.length;i++)n[i](t,this)}},i.once=function(e,t){var n=this;t&&this.addEventListener(e,(function i(){n.removeEventListener(e,i),t.apply(null,arguments)}))},i.setDefaultHandler=function(e,t){var n=this._defaultHandlers;if(n||(n=this._defaultHandlers={_disabled_:{}}),n[e]){var i=n[e],r=n._disabled_[e];r||(n._disabled_[e]=r=[]),r.push(i);var o=r.indexOf(t);-1!=o&&r.splice(o,1)}n[e]=t},i.removeDefaultHandler=function(e,t){var n=this._defaultHandlers;if(n){var i=n._disabled_[e];if(n[e]==t){n[e];i&&this.setDefaultHandler(e,i.pop())}else if(i){var r=i.indexOf(t);-1!=r&&i.splice(r,1)}}},i.on=i.addEventListener=function(e,t,n){this._eventRegistry=this._eventRegistry||{};var i=this._eventRegistry[e];return i||(i=this._eventRegistry[e]=[]),-1==i.indexOf(t)&&i[n?"unshift":"push"](t),t},i.off=i.removeListener=i.removeEventListener=function(e,t){this._eventRegistry=this._eventRegistry||{};var n=this._eventRegistry[e];if(n){var i=n.indexOf(t);-1!==i&&n.splice(i,1)}},i.removeAllListeners=function(e){this._eventRegistry&&(this._eventRegistry[e]=[])},t.EventEmitter=i})),ace.define("ace/lib/app_config",["require","exports","module","ace/lib/oop","ace/lib/event_emitter"],(function(e,t,n){var i=e("./oop"),r=e("./event_emitter").EventEmitter,o={setOptions:function(e){Object.keys(e).forEach((function(t){this.setOption(t,e[t])}),this)},getOptions:function(e){var t={};return e?Array.isArray(e)||(t=e,e=Object.keys(t)):e=Object.keys(this.$options),e.forEach((function(e){t[e]=this.getOption(e)}),this),t},setOption:function(e,t){if(this["$"+e]!==t){var n=this.$options[e];if(!n)return s('misspelled option "'+e+'"');if(n.forwardTo)return this[n.forwardTo]&&this[n.forwardTo].setOption(e,t);n.handlesSet||(this["$"+e]=t),n&&n.set&&n.set.call(this,t)}},getOption:function(e){var t=this.$options[e];return t?t.forwardTo?this[t.forwardTo]&&this[t.forwardTo].getOption(e):t&&t.get?t.get.call(this):this["$"+e]:s('misspelled option "'+e+'"')}};function s(e){"undefined"!=typeof console&&console.warn&&console.warn.apply(console,arguments)}function a(e,t){var n=new Error(e);n.data=t,"object"==typeof console&&console.error&&console.error(n),setTimeout((function(){throw n}))}var l=function(){this.$defaultOptions={}};(function(){i.implement(this,r),this.defineOptions=function(e,t,n){return e.$options||(this.$defaultOptions[t]=e.$options={}),Object.keys(n).forEach((function(t){var i=n[t];"string"==typeof i&&(i={forwardTo:i}),i.name||(i.name=t),e.$options[i.name]=i,"initialValue"in i&&(e["$"+i.name]=i.initialValue)})),i.implement(e,o),this},this.resetOptions=function(e){Object.keys(e.$options).forEach((function(t){var n=e.$options[t];"value"in n&&e.setOption(t,n.value)}))},this.setDefaultValue=function(e,t,n){var i=this.$defaultOptions[e]||(this.$defaultOptions[e]={});i[t]&&(i.forwardTo?this.setDefaultValue(i.forwardTo,t,n):i[t].value=n)},this.setDefaultValues=function(e,t){Object.keys(t).forEach((function(n){this.setDefaultValue(e,n,t[n])}),this)},this.warn=s,this.reportError=a}).call(l.prototype),t.AppConfig=l})),ace.define("ace/config",["require","exports","module","ace/lib/lang","ace/lib/oop","ace/lib/net","ace/lib/app_config"],(function(e,t,i){var r=e("./lib/lang"),o=(e("./lib/oop"),e("./lib/net")),s=e("./lib/app_config").AppConfig;i.exports=t=new s;var a=function(){return this||"undefined"!=typeof window&&window}(),l={packaged:!1,workerPath:null,modePath:null,themePath:null,basePath:"",suffix:".js",$moduleUrls:{}};function u(r){if(a&&a.document){l.packaged=r||e.packaged||i.packaged||a.define&&n(38).packaged;for(var o,s={},u="",c=document.currentScript||document._currentScript,h=(c&&c.ownerDocument||document).getElementsByTagName("script"),f=0;f<h.length;f++){var d=h[f],p=d.src||d.getAttribute("src");if(p){for(var g=d.attributes,m=0,v=g.length;m<v;m++){var b=g[m];0===b.name.indexOf("data-ace-")&&(s[(o=b.name.replace(/^data-ace-/,""),o.replace(/-(.)/g,(function(e,t){return t.toUpperCase()})))]=b.value)}var y=p.match(/^(.*)\/ace(\-\w+)?\.js(\?|$)/);y&&(u=y[1])}}for(var w in u&&(s.base=s.base||u,s.packaged=!0),s.basePath=s.base,s.workerPath=s.workerPath||s.base,s.modePath=s.modePath||s.base,s.themePath=s.themePath||s.base,delete s.base,s)void 0!==s[w]&&t.set(w,s[w])}}t.get=function(e){if(!l.hasOwnProperty(e))throw new Error("Unknown config key: "+e);return l[e]},t.set=function(e,t){if(!l.hasOwnProperty(e))throw new Error("Unknown config key: "+e);l[e]=t},t.all=function(){return r.copyObject(l)},t.moduleUrl=function(e,t){if(l.$moduleUrls[e])return l.$moduleUrls[e];var n=e.split("/"),i="snippets"==(t=t||n[n.length-2]||"")?"/":"-",r=n[n.length-1];if("worker"==t&&"-"==i){var o=new RegExp("^"+t+"[\\-_]|[\\-_]"+t+"$","g");r=r.replace(o,"")}(!r||r==t)&&n.length>1&&(r=n[n.length-2]);var s=l[t+"Path"];return null==s?s=l.basePath:"/"==i&&(t=i=""),s&&"/"!=s.slice(-1)&&(s+="/"),s+t+i+r+this.get("suffix")},t.setModuleUrl=function(e,t){return l.$moduleUrls[e]=t},t.$loading={},t.loadModule=function(n,i){var r,s;Array.isArray(n)&&(s=n[0],n=n[1]);try{r=e(n)}catch(e){}if(r&&!t.$loading[n])return i&&i(r);if(t.$loading[n]||(t.$loading[n]=[]),t.$loading[n].push(i),!(t.$loading[n].length>1)){var a=function(){e([n],(function(e){t._emit("load.module",{name:n,module:e});var i=t.$loading[n];t.$loading[n]=null,i.forEach((function(t){t&&t(e)}))}))};if(!t.get("packaged"))return a();o.loadScript(t.moduleUrl(n,s),a)}},u(!0),t.init=u})),ace.define("ace/mouse/mouse_handler",["require","exports","module","ace/lib/event","ace/lib/useragent","ace/mouse/default_handlers","ace/mouse/default_gutter_handler","ace/mouse/mouse_event","ace/mouse/dragdrop_handler","ace/config"],(function(e,t,n){"use strict";var i=e("../lib/event"),r=e("../lib/useragent"),o=e("./default_handlers").DefaultHandlers,s=e("./default_gutter_handler").GutterHandler,a=e("./mouse_event").MouseEvent,l=e("./dragdrop_handler").DragdropHandler,u=e("../config"),c=function(e){var t=this;this.editor=e,new o(this),new s(this),new l(this);var n=function(t){(!document.hasFocus||!document.hasFocus()||!e.isFocused()&&document.activeElement==(e.textInput&&e.textInput.getElement()))&&window.focus(),e.focus()},a=e.renderer.getMouseEventTarget();i.addListener(a,"click",this.onMouseEvent.bind(this,"click")),i.addListener(a,"mousemove",this.onMouseMove.bind(this,"mousemove")),i.addMultiMouseDownListener([a,e.renderer.scrollBarV&&e.renderer.scrollBarV.inner,e.renderer.scrollBarH&&e.renderer.scrollBarH.inner,e.textInput&&e.textInput.getElement()].filter(Boolean),[400,300,250],this,"onMouseEvent"),i.addMouseWheelListener(e.container,this.onMouseWheel.bind(this,"mousewheel")),i.addTouchMoveListener(e.container,this.onTouchMove.bind(this,"touchmove"));var u=e.renderer.$gutter;i.addListener(u,"mousedown",this.onMouseEvent.bind(this,"guttermousedown")),i.addListener(u,"click",this.onMouseEvent.bind(this,"gutterclick")),i.addListener(u,"dblclick",this.onMouseEvent.bind(this,"gutterdblclick")),i.addListener(u,"mousemove",this.onMouseEvent.bind(this,"guttermousemove")),i.addListener(a,"mousedown",n),i.addListener(u,"mousedown",n),r.isIE&&e.renderer.scrollBarV&&(i.addListener(e.renderer.scrollBarV.element,"mousedown",n),i.addListener(e.renderer.scrollBarH.element,"mousedown",n)),e.on("mousemove",(function(n){if(!t.state&&!t.$dragDelay&&t.$dragEnabled){var i=e.renderer.screenToTextCoordinates(n.x,n.y),r=e.session.selection.getRange(),o=e.renderer;!r.isEmpty()&&r.insideStart(i.row,i.column)?o.setCursorStyle("default"):o.setCursorStyle("")}}))};(function(){this.onMouseEvent=function(e,t){this.editor._emit(e,new a(t,this.editor))},this.onMouseMove=function(e,t){var n=this.editor._eventRegistry&&this.editor._eventRegistry.mousemove;n&&n.length&&this.editor._emit(e,new a(t,this.editor))},this.onMouseWheel=function(e,t){var n=new a(t,this.editor);n.speed=2*this.$scrollSpeed,n.wheelX=t.wheelX,n.wheelY=t.wheelY,this.editor._emit(e,n)},this.onTouchMove=function(e,t){var n=new a(t,this.editor);n.speed=1,n.wheelX=t.wheelX,n.wheelY=t.wheelY,this.editor._emit(e,n)},this.setState=function(e){this.state=e},this.captureMouse=function(e,t){this.x=e.x,this.y=e.y,this.isMousePressed=!0;var n=this.editor.renderer;n.$keepTextAreaAtCursor&&(n.$keepTextAreaAtCursor=null);var o=this,s=function(e){if(e){if(r.isWebKit&&!e.which&&o.releaseMouse)return o.releaseMouse();o.x=e.clientX,o.y=e.clientY,t&&t(e),o.mouseEvent=new a(e,o.editor),o.$mouseMoved=!0}},l=function(e){clearInterval(c),u(),o[o.state+"End"]&&o[o.state+"End"](e),o.state="",null==n.$keepTextAreaAtCursor&&(n.$keepTextAreaAtCursor=!0,n.$moveTextAreaToCursor()),o.isMousePressed=!1,o.$onCaptureMouseMove=o.releaseMouse=null,e&&o.onMouseEvent("mouseup",e)},u=function(){o[o.state]&&o[o.state](),o.$mouseMoved=!1};if(r.isOldIE&&"dblclick"==e.domEvent.type)return setTimeout((function(){l(e)}));o.$onCaptureMouseMove=s,o.releaseMouse=i.capture(this.editor.container,s,l);var c=setInterval(u,20)},this.releaseMouse=null,this.cancelContextMenu=function(){var e=function(t){t&&t.domEvent&&"contextmenu"!=t.domEvent.type||(this.editor.off("nativecontextmenu",e),t&&t.domEvent&&i.stopEvent(t.domEvent))}.bind(this);setTimeout(e,10),this.editor.on("nativecontextmenu",e)}}).call(c.prototype),u.defineOptions(c.prototype,"mouseHandler",{scrollSpeed:{initialValue:2},dragDelay:{initialValue:r.isMac?150:0},dragEnabled:{initialValue:!0},focusTimout:{initialValue:0},tooltipFollowsMouse:{initialValue:!0}}),t.MouseHandler=c})),ace.define("ace/mouse/fold_handler",["require","exports","module"],(function(e,t,n){"use strict";t.FoldHandler=function(e){e.on("click",(function(t){var n=t.getDocumentPosition(),i=e.session,r=i.getFoldAt(n.row,n.column,1);r&&(t.getAccelKey()?i.removeFold(r):i.expandFold(r),t.stop())})),e.on("gutterclick",(function(t){if("foldWidgets"==e.renderer.$gutterLayer.getRegion(t)){var n=t.getDocumentPosition().row,i=e.session;i.foldWidgets&&i.foldWidgets[n]&&e.session.onFoldWidgetClick(n,t),e.isFocused()||e.focus(),t.stop()}})),e.on("gutterdblclick",(function(t){if("foldWidgets"==e.renderer.$gutterLayer.getRegion(t)){var n=t.getDocumentPosition().row,i=e.session,r=i.getParentFoldRangeData(n,!0),o=r.range||r.firstRange;if(o){n=o.start.row;var s=i.getFoldAt(n,i.getLine(n).length,1);s?i.removeFold(s):(i.addFold("...",o),e.renderer.scrollCursorIntoView({row:o.start.row,column:0}))}t.stop()}}))}})),ace.define("ace/keyboard/keybinding",["require","exports","module","ace/lib/keys","ace/lib/event"],(function(e,t,n){"use strict";var i=e("../lib/keys"),r=e("../lib/event"),o=function(e){this.$editor=e,this.$data={editor:e},this.$handlers=[],this.setDefaultHandler(e.commands)};(function(){this.setDefaultHandler=function(e){this.removeKeyboardHandler(this.$defaultHandler),this.$defaultHandler=e,this.addKeyboardHandler(e,0)},this.setKeyboardHandler=function(e){var t=this.$handlers;if(t[t.length-1]!=e){for(;t[t.length-1]&&t[t.length-1]!=this.$defaultHandler;)this.removeKeyboardHandler(t[t.length-1]);this.addKeyboardHandler(e,1)}},this.addKeyboardHandler=function(e,t){if(e){"function"!=typeof e||e.handleKeyboard||(e.handleKeyboard=e);var n=this.$handlers.indexOf(e);-1!=n&&this.$handlers.splice(n,1),null==t?this.$handlers.push(e):this.$handlers.splice(t,0,e),-1==n&&e.attach&&e.attach(this.$editor)}},this.removeKeyboardHandler=function(e){var t=this.$handlers.indexOf(e);return-1!=t&&(this.$handlers.splice(t,1),e.detach&&e.detach(this.$editor),!0)},this.getKeyboardHandler=function(){return this.$handlers[this.$handlers.length-1]},this.getStatusText=function(){var e=this.$data,t=e.editor;return this.$handlers.map((function(n){return n.getStatusText&&n.getStatusText(t,e)||""})).filter(Boolean).join(" ")},this.$callKeyboardHandlers=function(e,t,n,i){for(var o,s=!1,a=this.$editor.commands,l=this.$handlers.length;l--&&!((o=this.$handlers[l].handleKeyboard(this.$data,e,t,n,i))&&o.command&&((s="null"==o.command||a.exec(o.command,this.$editor,o.args,i))&&i&&-1!=e&&1!=o.passEvent&&1!=o.command.passEvent&&r.stopEvent(i),s)););return s||-1!=e||(o={command:"insertstring"},s=a.exec("insertstring",this.$editor,t)),s&&this.$editor._signal&&this.$editor._signal("keyboardActivity",o),s},this.onCommandKey=function(e,t,n){var r=i.keyCodeToString(n);this.$callKeyboardHandlers(t,r,n,e)},this.onTextInput=function(e){this.$callKeyboardHandlers(-1,e)}}).call(o.prototype),t.KeyBinding=o})),ace.define("ace/lib/bidiutil",["require","exports","module"],(function(e,t,n){"use strict";var i=0,r=0,o=!1,s=!1,a=!1,l=[[0,3,0,1,0,0,0],[0,3,0,1,2,2,0],[0,3,0,17,2,0,1],[0,3,5,5,4,1,0],[0,3,21,21,4,0,1],[0,3,5,5,4,2,0]],u=[[2,0,1,1,0,1,0],[2,0,1,1,0,2,0],[2,0,2,1,3,2,0],[2,0,2,33,3,1,1]],c=[18,18,18,18,18,18,18,18,18,6,5,6,8,5,18,18,18,18,18,18,18,18,18,18,18,18,18,18,5,5,5,6,8,4,4,11,11,11,4,4,4,4,4,10,9,10,9,9,2,2,2,2,2,2,2,2,2,2,9,4,4,4,4,4,4,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,4,4,4,4,4,4,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,4,4,4,4,18,18,18,18,18,18,5,18,18,18,18,18,18,18,18,18,18,18,18,18,18,18,18,18,18,18,18,18,18,18,18,18,18,9,4,11,11,11,11,4,4,4,4,0,4,4,18,4,4,11,11,2,2,4,0,4,4,4,2,0,4,4,4,4,4],h=[8,8,8,8,8,8,8,8,8,8,8,18,18,18,0,1,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,8,5,13,14,15,16,17,9,11,11,11,11,11,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,9,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,8];function f(e,t,n){if(!(r<e))if(1!=e||1!=i||s)for(var o,a,l,u,c=n.length,h=0;h<c;){if(t[h]>=e){for(o=h+1;o<c&&t[o]>=e;)o++;for(a=h,l=o-1;a<l;a++,l--)u=n[a],n[a]=n[l],n[l]=u;h=o}h++}else n.reverse()}function d(e,t,n,r){var l,u,c,h,f=t[r];switch(f){case 0:case 1:o=!1;case 4:case 3:return f;case 2:return o?3:2;case 7:return o=!0,!0,1;case 8:return 4;case 9:return r<1||r+1>=t.length||2!=(l=n[r-1])&&3!=l||2!=(u=t[r+1])&&3!=u?4:(o&&(u=3),u==l?u:4);case 10:return 2==(l=r>0?n[r-1]:5)&&r+1<t.length&&2==t[r+1]?2:4;case 11:if(r>0&&2==n[r-1])return 2;if(o)return 4;for(h=r+1,c=t.length;h<c&&11==t[h];)h++;return h<c&&2==t[h]?2:4;case 12:for(c=t.length,h=r+1;h<c&&12==t[h];)h++;if(h<c){var d=e[r],p=d>=1425&&d<=2303||64286==d;if(l=t[h],p&&(1==l||7==l))return 1}return r<1||5==(l=t[r-1])?4:n[r-1];case 5:return o=!1,s=!0,i;case 6:return a=!0,4;case 13:case 14:case 16:case 17:case 15:o=!1;case 18:return 4}}function p(e){var t=e.charCodeAt(0),n=t>>8;return 0==n?t>191?0:c[t]:5==n?/[\u0591-\u05f4]/.test(e)?1:0:6==n?/[\u0610-\u061a\u064b-\u065f\u06d6-\u06e4\u06e7-\u06ed]/.test(e)?12:/[\u0660-\u0669\u066b-\u066c]/.test(e)?3:1642==t?11:/[\u06f0-\u06f9]/.test(e)?2:7:32==n&&t<=8287?h[255&t]:254==n&&t>=65136?7:4}t.L=0,t.R=1,t.EN=2,t.ON_R=3,t.AN=4,t.R_H=5,t.B=6,t.DOT="·",t.doBidiReorder=function(e,n,c){if(e.length<2)return{};var h=e.split(""),g=new Array(h.length),m=new Array(h.length),v=[];i=c?1:0,function(e,t,n,c){var h=i?u:l,f=null,g=null,m=null,v=0,b=null,y=-1,w=null,A=null,C=[];if(!c)for(w=0,c=[];w<n;w++)c[w]=p(e[w]);for(r=i,o=!1,!1,s=!1,a=!1,A=0;A<n;A++){if(f=v,C[A]=g=d(e,c,C,A),b=240&(v=h[f][g]),v&=15,t[A]=m=h[v][5],b>0)if(16==b){for(w=y;w<A;w++)t[w]=1;y=-1}else y=-1;if(h[v][6])-1==y&&(y=A);else if(y>-1){for(w=y;w<A;w++)t[w]=m;y=-1}5==c[A]&&(t[A]=0),r|=m}if(a)for(w=0;w<n;w++)if(6==c[w]){t[w]=i;for(var E=w-1;E>=0&&8==c[E];E--)t[E]=i}}(h,v,h.length,n);for(var b=0;b<g.length;g[b]=b,b++);f(2,v,g),f(1,v,g);for(b=0;b<g.length-1;b++)3===n[b]?v[b]=t.AN:1===v[b]&&(n[b]>7&&n[b]<13||4===n[b]||18===n[b])?v[b]=t.ON_R:b>0&&"ل"===h[b-1]&&/\u0622|\u0623|\u0625|\u0627/.test(h[b])&&(v[b-1]=v[b]=t.R_H,b++);h[h.length-1]===t.DOT&&(v[h.length-1]=t.B);for(b=0;b<g.length;b++)m[b]=v[g[b]];return{logicalFromVisual:g,bidiLevels:m}},t.hasBidiCharacters=function(e,t){for(var n=!1,i=0;i<e.length;i++)t[i]=p(e.charAt(i)),n||1!=t[i]&&7!=t[i]||(n=!0);return n},t.getVisualFromLogicalIdx=function(e,t){for(var n=0;n<t.logicalFromVisual.length;n++)if(t.logicalFromVisual[n]==e)return n;return 0}})),ace.define("ace/bidihandler",["require","exports","module","ace/lib/bidiutil","ace/lib/lang","ace/lib/useragent"],(function(e,t,n){"use strict";var i=e("./lib/bidiutil"),r=e("./lib/lang"),o=e("./lib/useragent"),s=/[\u0590-\u05f4\u0600-\u06ff\u0700-\u08ac]/,a=function(e){this.session=e,this.bidiMap={},this.currentRow=null,this.bidiUtil=i,this.charWidths=[],this.EOL="¬",this.showInvisibles=!0,this.isRtlDir=!1,this.line="",this.wrapIndent=0,this.isLastRow=!1,this.EOF="¶",this.seenBidi=!1};(function(){this.isBidiRow=function(e,t,n){return!!this.seenBidi&&(e!==this.currentRow&&(this.currentRow=e,this.updateRowLine(t,n),this.updateBidiMap()),this.bidiMap.bidiLevels)},this.onChange=function(e){this.seenBidi?this.currentRow=null:"insert"==e.action&&s.test(e.lines.join("\n"))&&(this.seenBidi=!0,this.currentRow=null)},this.getDocumentRow=function(){var e=0,t=this.session.$screenRowCache;if(t.length){var n=this.session.$getRowCacheIndex(t,this.currentRow);n>=0&&(e=this.session.$docRowCache[n])}return e},this.getSplitIndex=function(){var e=0,t=this.session.$screenRowCache;if(t.length)for(var n,i=this.session.$getRowCacheIndex(t,this.currentRow);this.currentRow-e>0&&(n=this.session.$getRowCacheIndex(t,this.currentRow-e-1))===i;)i=n,e++;return e},this.updateRowLine=function(e,t){if(void 0===e&&(e=this.getDocumentRow()),this.wrapIndent=0,this.isLastRow=e===this.session.getLength()-1,this.line=this.session.getLine(e),this.session.$useWrapMode){var n=this.session.$wrapData[e];n&&(void 0===t&&(t=this.getSplitIndex()),t>0&&n.length?(this.wrapIndent=n.indent,this.line=t<n.length?this.line.substring(n[t-1],n[n.length-1]):this.line.substring(n[n.length-1])):this.line=this.line.substring(0,n[t]))}var o,s=this.session,a=0;this.line=this.line.replace(/\t|[\u1100-\u2029, \u202F-\uFFE6]/g,(function(e,t){return"\t"===e||s.isFullWidth(e.charCodeAt(0))?(o="\t"===e?s.getScreenTabSize(t+a):2,a+=o-1,r.stringRepeat(i.DOT,o)):e}))},this.updateBidiMap=function(){var e=[],t=this.isLastRow?this.EOF:this.EOL,n=this.line+(this.showInvisibles?t:i.DOT);i.hasBidiCharacters(n,e)?this.bidiMap=i.doBidiReorder(n,e,this.isRtlDir):this.bidiMap={}},this.markAsDirty=function(){this.currentRow=null},this.updateCharacterWidths=function(e){if(this.seenBidi&&this.characterWidth!==e.$characterSize.width){var t=this.characterWidth=e.$characterSize.width,n=e.$measureCharWidth("ה");this.charWidths[i.L]=this.charWidths[i.EN]=this.charWidths[i.ON_R]=t,this.charWidths[i.R]=this.charWidths[i.AN]=n,this.charWidths[i.R_H]=o.isChrome?n:.45*n,this.charWidths[i.B]=0,this.currentRow=null}},this.getShowInvisibles=function(){return this.showInvisibles},this.setShowInvisibles=function(e){this.showInvisibles=e,this.currentRow=null},this.setEolChar=function(e){this.EOL=e},this.setTextDir=function(e){this.isRtlDir=e},this.getPosLeft=function(e){e-=this.wrapIndent;var t=i.getVisualFromLogicalIdx(e>0?e-1:0,this.bidiMap),n=this.bidiMap.bidiLevels,r=0;0===e&&n[t]%2!=0&&t++;for(var o=0;o<t;o++)r+=this.charWidths[n[o]];return 0!==e&&n[t]%2==0&&(r+=this.charWidths[n[t]]),this.wrapIndent&&(r+=this.wrapIndent*this.charWidths[i.L]),r},this.getSelections=function(e,t){for(var n,r,o=this.bidiMap,s=o.bidiLevels,a=this.wrapIndent*this.charWidths[i.L],l=[],u=Math.min(e,t)-this.wrapIndent,c=Math.max(e,t)-this.wrapIndent,h=!1,f=!1,d=0,p=0;p<s.length;p++)r=o.logicalFromVisual[p],n=s[p],(h=r>=u&&r<c)&&!f?d=a:!h&&f&&l.push({left:d,width:a-d}),a+=this.charWidths[n],f=h;return h&&p===s.length&&l.push({left:d,width:a-d}),l},this.offsetToCol=function(e){var t=0,n=(e=Math.max(e,0),0),r=0,o=this.bidiMap.bidiLevels,s=this.charWidths[o[r]];for(this.wrapIndent&&(e-=this.wrapIndent*this.charWidths[i.L]);e>n+s/2;){if(n+=s,r===o.length-1){s=0;break}s=this.charWidths[o[++r]]}return r>0&&o[r-1]%2!=0&&o[r]%2==0?(e<n&&r--,t=this.bidiMap.logicalFromVisual[r]):r>0&&o[r-1]%2==0&&o[r]%2!=0?t=1+(e>n?this.bidiMap.logicalFromVisual[r]:this.bidiMap.logicalFromVisual[r-1]):this.isRtlDir&&r===o.length-1&&0===s&&o[r-1]%2==0||!this.isRtlDir&&0===r&&o[r]%2!=0?t=1+this.bidiMap.logicalFromVisual[r]:(r>0&&o[r-1]%2!=0&&0!==s&&r--,t=this.bidiMap.logicalFromVisual[r]),t+this.wrapIndent}}).call(a.prototype),t.BidiHandler=a})),ace.define("ace/range",["require","exports","module"],(function(e,t,n){"use strict";var i=function(e,t,n,i){this.start={row:e,column:t},this.end={row:n,column:i}};(function(){this.isEqual=function(e){return this.start.row===e.start.row&&this.end.row===e.end.row&&this.start.column===e.start.column&&this.end.column===e.end.column},this.toString=function(){return"Range: ["+this.start.row+"/"+this.start.column+"] -> ["+this.end.row+"/"+this.end.column+"]"},this.contains=function(e,t){return 0==this.compare(e,t)},this.compareRange=function(e){var t,n=e.end,i=e.start;return 1==(t=this.compare(n.row,n.column))?1==(t=this.compare(i.row,i.column))?2:0==t?1:0:-1==t?-2:-1==(t=this.compare(i.row,i.column))?-1:1==t?42:0},this.comparePoint=function(e){return this.compare(e.row,e.column)},this.containsRange=function(e){return 0==this.comparePoint(e.start)&&0==this.comparePoint(e.end)},this.intersects=function(e){var t=this.compareRange(e);return-1==t||0==t||1==t},this.isEnd=function(e,t){return this.end.row==e&&this.end.column==t},this.isStart=function(e,t){return this.start.row==e&&this.start.column==t},this.setStart=function(e,t){"object"==typeof e?(this.start.column=e.column,this.start.row=e.row):(this.start.row=e,this.start.column=t)},this.setEnd=function(e,t){"object"==typeof e?(this.end.column=e.column,this.end.row=e.row):(this.end.row=e,this.end.column=t)},this.inside=function(e,t){return 0==this.compare(e,t)&&(!this.isEnd(e,t)&&!this.isStart(e,t))},this.insideStart=function(e,t){return 0==this.compare(e,t)&&!this.isEnd(e,t)},this.insideEnd=function(e,t){return 0==this.compare(e,t)&&!this.isStart(e,t)},this.compare=function(e,t){return this.isMultiLine()||e!==this.start.row?e<this.start.row?-1:e>this.end.row?1:this.start.row===e?t>=this.start.column?0:-1:this.end.row===e?t<=this.end.column?0:1:0:t<this.start.column?-1:t>this.end.column?1:0},this.compareStart=function(e,t){return this.start.row==e&&this.start.column==t?-1:this.compare(e,t)},this.compareEnd=function(e,t){return this.end.row==e&&this.end.column==t?1:this.compare(e,t)},this.compareInside=function(e,t){return this.end.row==e&&this.end.column==t?1:this.start.row==e&&this.start.column==t?-1:this.compare(e,t)},this.clipRows=function(e,t){if(this.end.row>t)var n={row:t+1,column:0};else if(this.end.row<e)n={row:e,column:0};if(this.start.row>t)var r={row:t+1,column:0};else if(this.start.row<e)r={row:e,column:0};return i.fromPoints(r||this.start,n||this.end)},this.extend=function(e,t){var n=this.compare(e,t);if(0==n)return this;if(-1==n)var r={row:e,column:t};else var o={row:e,column:t};return i.fromPoints(r||this.start,o||this.end)},this.isEmpty=function(){return this.start.row===this.end.row&&this.start.column===this.end.column},this.isMultiLine=function(){return this.start.row!==this.end.row},this.clone=function(){return i.fromPoints(this.start,this.end)},this.collapseRows=function(){return 0==this.end.column?new i(this.start.row,0,Math.max(this.start.row,this.end.row-1),0):new i(this.start.row,0,this.end.row,0)},this.toScreenRange=function(e){var t=e.documentToScreenPosition(this.start),n=e.documentToScreenPosition(this.end);return new i(t.row,t.column,n.row,n.column)},this.moveBy=function(e,t){this.start.row+=e,this.start.column+=t,this.end.row+=e,this.end.column+=t}}).call(i.prototype),i.fromPoints=function(e,t){return new i(e.row,e.column,t.row,t.column)},i.comparePoints=function(e,t){return e.row-t.row||e.column-t.column},i.comparePoints=function(e,t){return e.row-t.row||e.column-t.column},t.Range=i})),ace.define("ace/selection",["require","exports","module","ace/lib/oop","ace/lib/lang","ace/lib/event_emitter","ace/range"],(function(e,t,n){"use strict";var i=e("./lib/oop"),r=e("./lib/lang"),o=e("./lib/event_emitter").EventEmitter,s=e("./range").Range,a=function(e){this.session=e,this.doc=e.getDocument(),this.clearSelection(),this.lead=this.selectionLead=this.doc.createAnchor(0,0),this.anchor=this.selectionAnchor=this.doc.createAnchor(0,0);var t=this;this.lead.on("change",(function(e){t._emit("changeCursor"),t.$isEmpty||t._emit("changeSelection"),t.$keepDesiredColumnOnChange||e.old.column==e.value.column||(t.$desiredColumn=null)})),this.selectionAnchor.on("change",(function(){t.$isEmpty||t._emit("changeSelection")}))};(function(){i.implement(this,o),this.isEmpty=function(){return this.$isEmpty||this.anchor.row==this.lead.row&&this.anchor.column==this.lead.column},this.isMultiLine=function(){return!this.isEmpty()&&this.getRange().isMultiLine()},this.getCursor=function(){return this.lead.getPosition()},this.setSelectionAnchor=function(e,t){this.anchor.setPosition(e,t),this.$isEmpty&&(this.$isEmpty=!1,this._emit("changeSelection"))},this.getSelectionAnchor=function(){return this.$isEmpty?this.getSelectionLead():this.anchor.getPosition()},this.getSelectionLead=function(){return this.lead.getPosition()},this.shiftSelection=function(e){if(this.$isEmpty)this.moveCursorTo(this.lead.row,this.lead.column+e);else{var t=this.getSelectionAnchor(),n=this.getSelectionLead(),i=this.isBackwards();i&&0===t.column||this.setSelectionAnchor(t.row,t.column+e),(i||0!==n.column)&&this.$moveSelection((function(){this.moveCursorTo(n.row,n.column+e)}))}},this.isBackwards=function(){var e=this.anchor,t=this.lead;return e.row>t.row||e.row==t.row&&e.column>t.column},this.getRange=function(){var e=this.anchor,t=this.lead;return this.isEmpty()?s.fromPoints(t,t):this.isBackwards()?s.fromPoints(t,e):s.fromPoints(e,t)},this.clearSelection=function(){this.$isEmpty||(this.$isEmpty=!0,this._emit("changeSelection"))},this.selectAll=function(){var e=this.doc.getLength()-1;this.setSelectionAnchor(0,0),this.moveCursorTo(e,this.doc.getLine(e).length)},this.setRange=this.setSelectionRange=function(e,t){t?(this.setSelectionAnchor(e.end.row,e.end.column),this.selectTo(e.start.row,e.start.column)):(this.setSelectionAnchor(e.start.row,e.start.column),this.selectTo(e.end.row,e.end.column)),this.getRange().isEmpty()&&(this.$isEmpty=!0),this.$desiredColumn=null},this.$moveSelection=function(e){var t=this.lead;this.$isEmpty&&this.setSelectionAnchor(t.row,t.column),e.call(this)},this.selectTo=function(e,t){this.$moveSelection((function(){this.moveCursorTo(e,t)}))},this.selectToPosition=function(e){this.$moveSelection((function(){this.moveCursorToPosition(e)}))},this.moveTo=function(e,t){this.clearSelection(),this.moveCursorTo(e,t)},this.moveToPosition=function(e){this.clearSelection(),this.moveCursorToPosition(e)},this.selectUp=function(){this.$moveSelection(this.moveCursorUp)},this.selectDown=function(){this.$moveSelection(this.moveCursorDown)},this.selectRight=function(){this.$moveSelection(this.moveCursorRight)},this.selectLeft=function(){this.$moveSelection(this.moveCursorLeft)},this.selectLineStart=function(){this.$moveSelection(this.moveCursorLineStart)},this.selectLineEnd=function(){this.$moveSelection(this.moveCursorLineEnd)},this.selectFileEnd=function(){this.$moveSelection(this.moveCursorFileEnd)},this.selectFileStart=function(){this.$moveSelection(this.moveCursorFileStart)},this.selectWordRight=function(){this.$moveSelection(this.moveCursorWordRight)},this.selectWordLeft=function(){this.$moveSelection(this.moveCursorWordLeft)},this.getWordRange=function(e,t){if(void 0===t){var n=e||this.lead;e=n.row,t=n.column}return this.session.getWordRange(e,t)},this.selectWord=function(){this.setSelectionRange(this.getWordRange())},this.selectAWord=function(){var e=this.getCursor(),t=this.session.getAWordRange(e.row,e.column);this.setSelectionRange(t)},this.getLineRange=function(e,t){var n,i="number"==typeof e?e:this.lead.row,r=this.session.getFoldLine(i);return r?(i=r.start.row,n=r.end.row):n=i,!0===t?new s(i,0,n,this.session.getLine(n).length):new s(i,0,n+1,0)},this.selectLine=function(){this.setSelectionRange(this.getLineRange())},this.moveCursorUp=function(){this.moveCursorBy(-1,0)},this.moveCursorDown=function(){this.moveCursorBy(1,0)},this.wouldMoveIntoSoftTab=function(e,t,n){var i=e.column,r=e.column+t;return n<0&&(i=e.column-t,r=e.column),this.session.isTabStop(e)&&this.doc.getLine(e.row).slice(i,r).split(" ").length-1==t},this.moveCursorLeft=function(){var e,t=this.lead.getPosition();if(e=this.session.getFoldAt(t.row,t.column,-1))this.moveCursorTo(e.start.row,e.start.column);else if(0===t.column)t.row>0&&this.moveCursorTo(t.row-1,this.doc.getLine(t.row-1).length);else{var n=this.session.getTabSize();this.wouldMoveIntoSoftTab(t,n,-1)&&!this.session.getNavigateWithinSoftTabs()?this.moveCursorBy(0,-n):this.moveCursorBy(0,-1)}},this.moveCursorRight=function(){var e,t=this.lead.getPosition();if(e=this.session.getFoldAt(t.row,t.column,1))this.moveCursorTo(e.end.row,e.end.column);else if(this.lead.column==this.doc.getLine(this.lead.row).length)this.lead.row<this.doc.getLength()-1&&this.moveCursorTo(this.lead.row+1,0);else{var n=this.session.getTabSize();t=this.lead;this.wouldMoveIntoSoftTab(t,n,1)&&!this.session.getNavigateWithinSoftTabs()?this.moveCursorBy(0,n):this.moveCursorBy(0,1)}},this.moveCursorLineStart=function(){var e=this.lead.row,t=this.lead.column,n=this.session.documentToScreenRow(e,t),i=this.session.screenToDocumentPosition(n,0),r=this.session.getDisplayLine(e,null,i.row,i.column).match(/^\s*/);r[0].length==t||this.session.$useEmacsStyleLineStart||(i.column+=r[0].length),this.moveCursorToPosition(i)},this.moveCursorLineEnd=function(){var e=this.lead,t=this.session.getDocumentLastRowColumnPosition(e.row,e.column);if(this.lead.column==t.column){var n=this.session.getLine(t.row);if(t.column==n.length){var i=n.search(/\s+$/);i>0&&(t.column=i)}}this.moveCursorTo(t.row,t.column)},this.moveCursorFileEnd=function(){var e=this.doc.getLength()-1,t=this.doc.getLine(e).length;this.moveCursorTo(e,t)},this.moveCursorFileStart=function(){this.moveCursorTo(0,0)},this.moveCursorLongWordRight=function(){var e=this.lead.row,t=this.lead.column,n=this.doc.getLine(e),i=n.substring(t);this.session.nonTokenRe.lastIndex=0,this.session.tokenRe.lastIndex=0;var r=this.session.getFoldAt(e,t,1);if(r)this.moveCursorTo(r.end.row,r.end.column);else{if(this.session.nonTokenRe.exec(i)&&(t+=this.session.nonTokenRe.lastIndex,this.session.nonTokenRe.lastIndex=0,i=n.substring(t)),t>=n.length)return this.moveCursorTo(e,n.length),this.moveCursorRight(),void(e<this.doc.getLength()-1&&this.moveCursorWordRight());this.session.tokenRe.exec(i)&&(t+=this.session.tokenRe.lastIndex,this.session.tokenRe.lastIndex=0),this.moveCursorTo(e,t)}},this.moveCursorLongWordLeft=function(){var e,t=this.lead.row,n=this.lead.column;if(e=this.session.getFoldAt(t,n,-1))this.moveCursorTo(e.start.row,e.start.column);else{var i=this.session.getFoldStringAt(t,n,-1);null==i&&(i=this.doc.getLine(t).substring(0,n));var o=r.stringReverse(i);if(this.session.nonTokenRe.lastIndex=0,this.session.tokenRe.lastIndex=0,this.session.nonTokenRe.exec(o)&&(n-=this.session.nonTokenRe.lastIndex,o=o.slice(this.session.nonTokenRe.lastIndex),this.session.nonTokenRe.lastIndex=0),n<=0)return this.moveCursorTo(t,0),this.moveCursorLeft(),void(t>0&&this.moveCursorWordLeft());this.session.tokenRe.exec(o)&&(n-=this.session.tokenRe.lastIndex,this.session.tokenRe.lastIndex=0),this.moveCursorTo(t,n)}},this.$shortWordEndIndex=function(e){var t,n=0,i=/\s/,r=this.session.tokenRe;if(r.lastIndex=0,this.session.tokenRe.exec(e))n=this.session.tokenRe.lastIndex;else{for(;(t=e[n])&&i.test(t);)n++;if(n<1)for(r.lastIndex=0;(t=e[n])&&!r.test(t);)if(r.lastIndex=0,n++,i.test(t)){if(n>2){n--;break}for(;(t=e[n])&&i.test(t);)n++;if(n>2)break}}return r.lastIndex=0,n},this.moveCursorShortWordRight=function(){var e=this.lead.row,t=this.lead.column,n=this.doc.getLine(e),i=n.substring(t),r=this.session.getFoldAt(e,t,1);if(r)return this.moveCursorTo(r.end.row,r.end.column);if(t==n.length){var o=this.doc.getLength();do{e++,i=this.doc.getLine(e)}while(e<o&&/^\s*$/.test(i));/^\s+/.test(i)||(i=""),t=0}var s=this.$shortWordEndIndex(i);this.moveCursorTo(e,t+s)},this.moveCursorShortWordLeft=function(){var e,t=this.lead.row,n=this.lead.column;if(e=this.session.getFoldAt(t,n,-1))return this.moveCursorTo(e.start.row,e.start.column);var i=this.session.getLine(t).substring(0,n);if(0===n){do{t--,i=this.doc.getLine(t)}while(t>0&&/^\s*$/.test(i));n=i.length,/\s+$/.test(i)||(i="")}var o=r.stringReverse(i),s=this.$shortWordEndIndex(o);return this.moveCursorTo(t,n-s)},this.moveCursorWordRight=function(){this.session.$selectLongWords?this.moveCursorLongWordRight():this.moveCursorShortWordRight()},this.moveCursorWordLeft=function(){this.session.$selectLongWords?this.moveCursorLongWordLeft():this.moveCursorShortWordLeft()},this.moveCursorBy=function(e,t){var n,i=this.session.documentToScreenPosition(this.lead.row,this.lead.column);0===t&&(0!==e&&(this.session.$bidiHandler.isBidiRow(i.row,this.lead.row)?(n=this.session.$bidiHandler.getPosLeft(i.column),i.column=Math.round(n/this.session.$bidiHandler.charWidths[0])):n=i.column*this.session.$bidiHandler.charWidths[0]),this.$desiredColumn?i.column=this.$desiredColumn:this.$desiredColumn=i.column);var r=this.session.screenToDocumentPosition(i.row+e,i.column,n);0!==e&&0===t&&r.row===this.lead.row&&r.column===this.lead.column&&this.session.lineWidgets&&this.session.lineWidgets[r.row]&&(r.row>0||e>0)&&r.row++,this.moveCursorTo(r.row,r.column+t,0===t)},this.moveCursorToPosition=function(e){this.moveCursorTo(e.row,e.column)},this.moveCursorTo=function(e,t,n){var i=this.session.getFoldAt(e,t,1);i&&(e=i.start.row,t=i.start.column),this.$keepDesiredColumnOnChange=!0;var r=this.session.getLine(e);/[\uDC00-\uDFFF]/.test(r.charAt(t))&&r.charAt(t-1)&&(this.lead.row==e&&this.lead.column==t+1?t-=1:t+=1),this.lead.setPosition(e,t),this.$keepDesiredColumnOnChange=!1,n||(this.$desiredColumn=null)},this.moveCursorToScreen=function(e,t,n){var i=this.session.screenToDocumentPosition(e,t);this.moveCursorTo(i.row,i.column,n)},this.detach=function(){this.lead.detach(),this.anchor.detach(),this.session=this.doc=null},this.fromOrientedRange=function(e){this.setSelectionRange(e,e.cursor==e.start),this.$desiredColumn=e.desiredColumn||this.$desiredColumn},this.toOrientedRange=function(e){var t=this.getRange();return e?(e.start.column=t.start.column,e.start.row=t.start.row,e.end.column=t.end.column,e.end.row=t.end.row):e=t,e.cursor=this.isBackwards()?e.start:e.end,e.desiredColumn=this.$desiredColumn,e},this.getRangeOfMovements=function(e){var t=this.getCursor();try{e(this);var n=this.getCursor();return s.fromPoints(t,n)}catch(e){return s.fromPoints(t,t)}finally{this.moveCursorToPosition(t)}},this.toJSON=function(){if(this.rangeCount)var e=this.ranges.map((function(e){var t=e.clone();return t.isBackwards=e.cursor==e.start,t}));else(e=this.getRange()).isBackwards=this.isBackwards();return e},this.fromJSON=function(e){if(null==e.start){if(this.rangeList){this.toSingleRange(e[0]);for(var t=e.length;t--;){var n=s.fromPoints(e[t].start,e[t].end);e[t].isBackwards&&(n.cursor=n.start),this.addRange(n,!0)}return}e=e[0]}this.rangeList&&this.toSingleRange(e),this.setSelectionRange(e,e.isBackwards)},this.isEqual=function(e){if((e.length||this.rangeCount)&&e.length!=this.rangeCount)return!1;if(!e.length||!this.ranges)return this.getRange().isEqual(e);for(var t=this.ranges.length;t--;)if(!this.ranges[t].isEqual(e[t]))return!1;return!0}}).call(a.prototype),t.Selection=a})),ace.define("ace/tokenizer",["require","exports","module","ace/config"],(function(e,t,n){"use strict";var i=e("./config"),r=2e3,o=function(e){for(var t in this.states=e,this.regExps={},this.matchMappings={},this.states){for(var n=this.states[t],i=[],r=0,o=this.matchMappings[t]={defaultToken:"text"},s="g",a=[],l=0;l<n.length;l++){var u=n[l];if(u.defaultToken&&(o.defaultToken=u.defaultToken),u.caseInsensitive&&(s="gi"),null!=u.regex){u.regex instanceof RegExp&&(u.regex=u.regex.toString().slice(1,-1));var c=u.regex,h=new RegExp("(?:("+c+")|(.))").exec("a").length-2;Array.isArray(u.token)?1==u.token.length||1==h?u.token=u.token[0]:h-1!=u.token.length?(this.reportError("number of classes and regexp groups doesn't match",{rule:u,groupCount:h-1}),u.token=u.token[0]):(u.tokenArray=u.token,u.token=null,u.onMatch=this.$arrayTokens):"function"!=typeof u.token||u.onMatch||(u.onMatch=h>1?this.$applyToken:u.token),h>1&&(/\\\d/.test(u.regex)?c=u.regex.replace(/\\([0-9]+)/g,(function(e,t){return"\\"+(parseInt(t,10)+r+1)})):(h=1,c=this.removeCapturingGroups(u.regex)),u.splitRegex||"string"==typeof u.token||a.push(u)),o[r]=l,r+=h,i.push(c),u.onMatch||(u.onMatch=null)}}i.length||(o[0]=0,i.push("$")),a.forEach((function(e){e.splitRegex=this.createSplitterRegexp(e.regex,s)}),this),this.regExps[t]=new RegExp("("+i.join(")|(")+")|($)",s)}};(function(){this.$setMaxTokenCount=function(e){r=0|e},this.$applyToken=function(e){var t=this.splitRegex.exec(e).slice(1),n=this.token.apply(this,t);if("string"==typeof n)return[{type:n,value:e}];for(var i=[],r=0,o=n.length;r<o;r++)t[r]&&(i[i.length]={type:n[r],value:t[r]});return i},this.$arrayTokens=function(e){if(!e)return[];var t=this.splitRegex.exec(e);if(!t)return"text";for(var n=[],i=this.tokenArray,r=0,o=i.length;r<o;r++)t[r+1]&&(n[n.length]={type:i[r],value:t[r+1]});return n},this.removeCapturingGroups=function(e){return e.replace(/\[(?:\\.|[^\]])*?\]|\\.|\(\?[:=!]|(\()/g,(function(e,t){return t?"(?:":e}))},this.createSplitterRegexp=function(e,t){if(-1!=e.indexOf("(?=")){var n=0,i=!1,r={};e.replace(/(\\.)|(\((?:\?[=!])?)|(\))|([\[\]])/g,(function(e,t,o,s,a,l){return i?i="]"!=a:a?i=!0:s?(n==r.stack&&(r.end=l+1,r.stack=-1),n--):o&&(n++,1!=o.length&&(r.stack=n,r.start=l)),e})),null!=r.end&&/^\)*$/.test(e.substr(r.end))&&(e=e.substring(0,r.start)+e.substr(r.end))}return"^"!=e.charAt(0)&&(e="^"+e),"$"!=e.charAt(e.length-1)&&(e+="$"),new RegExp(e,(t||"").replace("g",""))},this.getLineTokens=function(e,t){if(t&&"string"!=typeof t){var n=t.slice(0);"#tmp"===(t=n[0])&&(n.shift(),t=n.shift())}else n=[];var i=t||"start",o=this.states[i];o||(i="start",o=this.states[i]);var s=this.matchMappings[i],a=this.regExps[i];a.lastIndex=0;for(var l,u=[],c=0,h=0,f={type:null,value:""};l=a.exec(e);){var d=s.defaultToken,p=null,g=l[0],m=a.lastIndex;if(m-g.length>c){var v=e.substring(c,m-g.length);f.type==d?f.value+=v:(f.type&&u.push(f),f={type:d,value:v})}for(var b=0;b<l.length-2;b++)if(void 0!==l[b+1]){d=(p=o[s[b]]).onMatch?p.onMatch(g,i,n,e):p.token,p.next&&(i="string"==typeof p.next?p.next:p.next(i,n),(o=this.states[i])||(this.reportError("state doesn't exist",i),i="start",o=this.states[i]),s=this.matchMappings[i],c=m,(a=this.regExps[i]).lastIndex=m),p.consumeLineEnd&&(c=m);break}if(g)if("string"==typeof d)p&&!1===p.merge||f.type!==d?(f.type&&u.push(f),f={type:d,value:g}):f.value+=g;else if(d){f.type&&u.push(f),f={type:null,value:""};for(b=0;b<d.length;b++)u.push(d[b])}if(c==e.length)break;if(c=m,h++>r){for(h>2*e.length&&this.reportError("infinite loop with in ace tokenizer",{startState:t,line:e});c<e.length;)f.type&&u.push(f),f={value:e.substring(c,c+=2e3),type:"overflow"};i="start",n=[];break}}return f.type&&u.push(f),n.length>1&&n[0]!==i&&n.unshift("#tmp",i),{tokens:u,state:n.length?n:i}},this.reportError=i.reportError}).call(o.prototype),t.Tokenizer=o})),ace.define("ace/mode/text_highlight_rules",["require","exports","module","ace/lib/lang"],(function(e,t,n){"use strict";var i=e("../lib/lang"),r=function(){this.$rules={start:[{token:"empty_line",regex:"^$"},{defaultToken:"text"}]}};(function(){this.addRules=function(e,t){if(t)for(var n in e){for(var i=e[n],r=0;r<i.length;r++){var o=i[r];(o.next||o.onMatch)&&("string"==typeof o.next&&0!==o.next.indexOf(t)&&(o.next=t+o.next),o.nextState&&0!==o.nextState.indexOf(t)&&(o.nextState=t+o.nextState))}this.$rules[t+n]=i}else for(var n in e)this.$rules[n]=e[n]},this.getRules=function(){return this.$rules},this.embedRules=function(e,t,n,r,o){var s="function"==typeof e?(new e).getRules():e;if(r)for(var a=0;a<r.length;a++)r[a]=t+r[a];else for(var l in r=[],s)r.push(t+l);if(this.addRules(s,t),n){var u=Array.prototype[o?"push":"unshift"];for(a=0;a<r.length;a++)u.apply(this.$rules[r[a]],i.deepCopy(n))}this.$embeds||(this.$embeds=[]),this.$embeds.push(t)},this.getEmbeds=function(){return this.$embeds};var e=function(e,t){return("start"!=e||t.length)&&t.unshift(this.nextState,e),this.nextState},t=function(e,t){return t.shift(),t.shift()||"start"};this.normalizeRules=function(){var n=0,i=this.$rules;Object.keys(i).forEach((function r(o){var s=i[o];s.processed=!0;for(var a=0;a<s.length;a++){var l=s[a],u=null;Array.isArray(l)&&(u=l,l={}),!l.regex&&l.start&&(l.regex=l.start,l.next||(l.next=[]),l.next.push({defaultToken:l.token},{token:l.token+".end",regex:l.end||l.start,next:"pop"}),l.token=l.token+".start",l.push=!0);var c=l.next||l.push;if(c&&Array.isArray(c)){var h=l.stateName;h||("string"!=typeof(h=l.token)&&(h=h[0]||""),i[h]&&(h+=n++)),i[h]=c,l.next=h,r(h)}else"pop"==c&&(l.next=t);if(l.push&&(l.nextState=l.next||l.push,l.next=e,delete l.push),l.rules)for(var f in l.rules)i[f]?i[f].push&&i[f].push.apply(i[f],l.rules[f]):i[f]=l.rules[f];var d="string"==typeof l?l:l.include;if(d&&(u=Array.isArray(d)?d.map((function(e){return i[e]})):i[d]),u){var p=[a,1].concat(u);l.noEscape&&(p=p.filter((function(e){return!e.next}))),s.splice.apply(s,p),a--}l.keywordMap&&(l.token=this.createKeywordMapper(l.keywordMap,l.defaultToken||"text",l.caseInsensitive),delete l.defaultToken)}}),this)},this.createKeywordMapper=function(e,t,n,i){var r=Object.create(null);return Object.keys(e).forEach((function(t){var o=e[t];n&&(o=o.toLowerCase());for(var s=o.split(i||"|"),a=s.length;a--;)r[s[a]]=t})),Object.getPrototypeOf(r)&&(r.__proto__=null),this.$keywordList=Object.keys(r),e=null,n?function(e){return r[e.toLowerCase()]||t}:function(e){return r[e]||t}},this.getKeywords=function(){return this.$keywords}}).call(r.prototype),t.TextHighlightRules=r})),ace.define("ace/mode/behaviour",["require","exports","module"],(function(e,t,n){"use strict";var i=function(){this.$behaviours={}};(function(){this.add=function(e,t,n){switch(void 0){case this.$behaviours:this.$behaviours={};case this.$behaviours[e]:this.$behaviours[e]={}}this.$behaviours[e][t]=n},this.addBehaviours=function(e){for(var t in e)for(var n in e[t])this.add(t,n,e[t][n])},this.remove=function(e){this.$behaviours&&this.$behaviours[e]&&delete this.$behaviours[e]},this.inherit=function(e,t){if("function"==typeof e)var n=(new e).getBehaviours(t);else n=e.getBehaviours(t);this.addBehaviours(n)},this.getBehaviours=function(e){if(e){for(var t={},n=0;n<e.length;n++)this.$behaviours[e[n]]&&(t[e[n]]=this.$behaviours[e[n]]);return t}return this.$behaviours}}).call(i.prototype),t.Behaviour=i})),ace.define("ace/token_iterator",["require","exports","module","ace/range"],(function(e,t,n){"use strict";var i=e("./range").Range,r=function(e,t,n){this.$session=e,this.$row=t,this.$rowTokens=e.getTokens(t);var i=e.getTokenAt(t,n);this.$tokenIndex=i?i.index:-1};(function(){this.stepBackward=function(){for(this.$tokenIndex-=1;this.$tokenIndex<0;){if(this.$row-=1,this.$row<0)return this.$row=0,null;this.$rowTokens=this.$session.getTokens(this.$row),this.$tokenIndex=this.$rowTokens.length-1}return this.$rowTokens[this.$tokenIndex]},this.stepForward=function(){var e;for(this.$tokenIndex+=1;this.$tokenIndex>=this.$rowTokens.length;){if(this.$row+=1,e||(e=this.$session.getLength()),this.$row>=e)return this.$row=e-1,null;this.$rowTokens=this.$session.getTokens(this.$row),this.$tokenIndex=0}return this.$rowTokens[this.$tokenIndex]},this.getCurrentToken=function(){return this.$rowTokens[this.$tokenIndex]},this.getCurrentTokenRow=function(){return this.$row},this.getCurrentTokenColumn=function(){var e=this.$rowTokens,t=this.$tokenIndex,n=e[t].start;if(void 0!==n)return n;for(n=0;t>0;)n+=e[t-=1].value.length;return n},this.getCurrentTokenPosition=function(){return{row:this.$row,column:this.getCurrentTokenColumn()}},this.getCurrentTokenRange=function(){var e=this.$rowTokens[this.$tokenIndex],t=this.getCurrentTokenColumn();return new i(this.$row,t,this.$row,t+e.value.length)}}).call(r.prototype),t.TokenIterator=r})),ace.define("ace/mode/behaviour/cstyle",["require","exports","module","ace/lib/oop","ace/mode/behaviour","ace/token_iterator","ace/lib/lang"],(function(e,t,n){"use strict";var i,r=e("../../lib/oop"),o=e("../behaviour").Behaviour,s=e("../../token_iterator").TokenIterator,a=e("../../lib/lang"),l=["text","paren.rparen","punctuation.operator"],u=["text","paren.rparen","punctuation.operator","comment"],c={},h={'"':'"',"'":"'"},f=function(e){var t=-1;if(e.multiSelect&&(t=e.selection.index,c.rangeCount!=e.multiSelect.rangeCount&&(c={rangeCount:e.multiSelect.rangeCount})),c[t])return i=c[t];i=c[t]={autoInsertedBrackets:0,autoInsertedRow:-1,autoInsertedLineEnd:"",maybeInsertedBrackets:0,maybeInsertedRow:-1,maybeInsertedLineStart:"",maybeInsertedLineEnd:""}},d=function(e,t,n,i){var r=e.end.row-e.start.row;return{text:n+t+i,selection:[0,e.start.column+1,r,e.end.column+(r?0:1)]}},p=function(e){this.add("braces","insertion",(function(t,n,r,o,s){var l=r.getCursorPosition(),u=o.doc.getLine(l.row);if("{"==s){f(r);var c=r.getSelectionRange(),h=o.doc.getTextRange(c);if(""!==h&&"{"!==h&&r.getWrapBehavioursEnabled())return d(c,h,"{","}");if(p.isSaneInsertion(r,o))return/[\]\}\)]/.test(u[l.column])||r.inMultiSelectMode||e&&e.braces?(p.recordAutoInsert(r,o,"}"),{text:"{}",selection:[1,1]}):(p.recordMaybeInsert(r,o,"{"),{text:"{",selection:[1,1]})}else if("}"==s){if(f(r),"}"==u.substring(l.column,l.column+1))if(null!==o.$findOpeningBracket("}",{column:l.column+1,row:l.row})&&p.isAutoInsertedClosing(l,u,s))return p.popAutoInsertedClosing(),{text:"",selection:[1,1]}}else{if("\n"==s||"\r\n"==s){f(r);var g="";if(p.isMaybeInsertedClosing(l,u)&&(g=a.stringRepeat("}",i.maybeInsertedBrackets),p.clearMaybeInsertedClosing()),"}"===u.substring(l.column,l.column+1)){var m=o.findMatchingBracket({row:l.row,column:l.column+1},"}");if(!m)return null;var v=this.$getIndent(o.getLine(m.row))}else{if(!g)return void p.clearMaybeInsertedClosing();v=this.$getIndent(u)}var b=v+o.getTabString();return{text:"\n"+b+"\n"+v+g,selection:[1,b.length,1,b.length]}}p.clearMaybeInsertedClosing()}})),this.add("braces","deletion",(function(e,t,n,r,o){var s=r.doc.getTextRange(o);if(!o.isMultiLine()&&"{"==s){if(f(n),"}"==r.doc.getLine(o.start.row).substring(o.end.column,o.end.column+1))return o.end.column++,o;i.maybeInsertedBrackets--}})),this.add("parens","insertion",(function(e,t,n,i,r){if("("==r){f(n);var o=n.getSelectionRange(),s=i.doc.getTextRange(o);if(""!==s&&n.getWrapBehavioursEnabled())return d(o,s,"(",")");if(p.isSaneInsertion(n,i))return p.recordAutoInsert(n,i,")"),{text:"()",selection:[1,1]}}else if(")"==r){f(n);var a=n.getCursorPosition(),l=i.doc.getLine(a.row);if(")"==l.substring(a.column,a.column+1))if(null!==i.$findOpeningBracket(")",{column:a.column+1,row:a.row})&&p.isAutoInsertedClosing(a,l,r))return p.popAutoInsertedClosing(),{text:"",selection:[1,1]}}})),this.add("parens","deletion",(function(e,t,n,i,r){var o=i.doc.getTextRange(r);if(!r.isMultiLine()&&"("==o&&(f(n),")"==i.doc.getLine(r.start.row).substring(r.start.column+1,r.start.column+2)))return r.end.column++,r})),this.add("brackets","insertion",(function(e,t,n,i,r){if("["==r){f(n);var o=n.getSelectionRange(),s=i.doc.getTextRange(o);if(""!==s&&n.getWrapBehavioursEnabled())return d(o,s,"[","]");if(p.isSaneInsertion(n,i))return p.recordAutoInsert(n,i,"]"),{text:"[]",selection:[1,1]}}else if("]"==r){f(n);var a=n.getCursorPosition(),l=i.doc.getLine(a.row);if("]"==l.substring(a.column,a.column+1))if(null!==i.$findOpeningBracket("]",{column:a.column+1,row:a.row})&&p.isAutoInsertedClosing(a,l,r))return p.popAutoInsertedClosing(),{text:"",selection:[1,1]}}})),this.add("brackets","deletion",(function(e,t,n,i,r){var o=i.doc.getTextRange(r);if(!r.isMultiLine()&&"["==o&&(f(n),"]"==i.doc.getLine(r.start.row).substring(r.start.column+1,r.start.column+2)))return r.end.column++,r})),this.add("string_dquotes","insertion",(function(e,t,n,i,r){var o=i.$mode.$quotes||h;if(1==r.length&&o[r]){if(this.lineCommentStart&&-1!=this.lineCommentStart.indexOf(r))return;f(n);var s=r,a=n.getSelectionRange(),l=i.doc.getTextRange(a);if(!(""===l||1==l.length&&o[l])&&n.getWrapBehavioursEnabled())return d(a,l,s,s);if(!l){var u=n.getCursorPosition(),c=i.doc.getLine(u.row),p=c.substring(u.column-1,u.column),g=c.substring(u.column,u.column+1),m=i.getTokenAt(u.row,u.column),v=i.getTokenAt(u.row,u.column+1);if("\\"==p&&m&&/escape/.test(m.type))return null;var b,y=m&&/string|escape/.test(m.type),w=!v||/string|escape/.test(v.type);if(g==s)(b=y!==w)&&/string\.end/.test(v.type)&&(b=!1);else{if(y&&!w)return null;if(y&&w)return null;var A=i.$mode.tokenRe;A.lastIndex=0;var C=A.test(p);A.lastIndex=0;var E=A.test(p);if(C||E)return null;if(g&&!/[\s;,.})\]\\]/.test(g))return null;b=!0}return{text:b?s+s:"",selection:[1,1]}}}})),this.add("string_dquotes","deletion",(function(e,t,n,i,r){var o=i.doc.getTextRange(r);if(!r.isMultiLine()&&('"'==o||"'"==o)&&(f(n),i.doc.getLine(r.start.row).substring(r.start.column+1,r.start.column+2)==o))return r.end.column++,r}))};p.isSaneInsertion=function(e,t){var n=e.getCursorPosition(),i=new s(t,n.row,n.column);if(!this.$matchTokenType(i.getCurrentToken()||"text",l)){var r=new s(t,n.row,n.column+1);if(!this.$matchTokenType(r.getCurrentToken()||"text",l))return!1}return i.stepForward(),i.getCurrentTokenRow()!==n.row||this.$matchTokenType(i.getCurrentToken()||"text",u)},p.$matchTokenType=function(e,t){return t.indexOf(e.type||e)>-1},p.recordAutoInsert=function(e,t,n){var r=e.getCursorPosition(),o=t.doc.getLine(r.row);this.isAutoInsertedClosing(r,o,i.autoInsertedLineEnd[0])||(i.autoInsertedBrackets=0),i.autoInsertedRow=r.row,i.autoInsertedLineEnd=n+o.substr(r.column),i.autoInsertedBrackets++},p.recordMaybeInsert=function(e,t,n){var r=e.getCursorPosition(),o=t.doc.getLine(r.row);this.isMaybeInsertedClosing(r,o)||(i.maybeInsertedBrackets=0),i.maybeInsertedRow=r.row,i.maybeInsertedLineStart=o.substr(0,r.column)+n,i.maybeInsertedLineEnd=o.substr(r.column),i.maybeInsertedBrackets++},p.isAutoInsertedClosing=function(e,t,n){return i.autoInsertedBrackets>0&&e.row===i.autoInsertedRow&&n===i.autoInsertedLineEnd[0]&&t.substr(e.column)===i.autoInsertedLineEnd},p.isMaybeInsertedClosing=function(e,t){return i.maybeInsertedBrackets>0&&e.row===i.maybeInsertedRow&&t.substr(e.column)===i.maybeInsertedLineEnd&&t.substr(0,e.column)==i.maybeInsertedLineStart},p.popAutoInsertedClosing=function(){i.autoInsertedLineEnd=i.autoInsertedLineEnd.substr(1),i.autoInsertedBrackets--},p.clearMaybeInsertedClosing=function(){i&&(i.maybeInsertedBrackets=0,i.maybeInsertedRow=-1)},r.inherits(p,o),t.CstyleBehaviour=p})),ace.define("ace/unicode",["require","exports","module"],(function(e,t,n){"use strict";t.packages={},function(e){var n=/\w{4}/g;for(var i in e)t.packages[i]=e[i].replace(n,"\\u$&")}({L:"0041-005A0061-007A00AA00B500BA00C0-00D600D8-00F600F8-02C102C6-02D102E0-02E402EC02EE0370-037403760377037A-037D03860388-038A038C038E-03A103A3-03F503F7-0481048A-05250531-055605590561-058705D0-05EA05F0-05F20621-064A066E066F0671-06D306D506E506E606EE06EF06FA-06FC06FF07100712-072F074D-07A507B107CA-07EA07F407F507FA0800-0815081A082408280904-0939093D09500958-0961097109720979-097F0985-098C098F09900993-09A809AA-09B009B209B6-09B909BD09CE09DC09DD09DF-09E109F009F10A05-0A0A0A0F0A100A13-0A280A2A-0A300A320A330A350A360A380A390A59-0A5C0A5E0A72-0A740A85-0A8D0A8F-0A910A93-0AA80AAA-0AB00AB20AB30AB5-0AB90ABD0AD00AE00AE10B05-0B0C0B0F0B100B13-0B280B2A-0B300B320B330B35-0B390B3D0B5C0B5D0B5F-0B610B710B830B85-0B8A0B8E-0B900B92-0B950B990B9A0B9C0B9E0B9F0BA30BA40BA8-0BAA0BAE-0BB90BD00C05-0C0C0C0E-0C100C12-0C280C2A-0C330C35-0C390C3D0C580C590C600C610C85-0C8C0C8E-0C900C92-0CA80CAA-0CB30CB5-0CB90CBD0CDE0CE00CE10D05-0D0C0D0E-0D100D12-0D280D2A-0D390D3D0D600D610D7A-0D7F0D85-0D960D9A-0DB10DB3-0DBB0DBD0DC0-0DC60E01-0E300E320E330E40-0E460E810E820E840E870E880E8A0E8D0E94-0E970E99-0E9F0EA1-0EA30EA50EA70EAA0EAB0EAD-0EB00EB20EB30EBD0EC0-0EC40EC60EDC0EDD0F000F40-0F470F49-0F6C0F88-0F8B1000-102A103F1050-1055105A-105D106110651066106E-10701075-1081108E10A0-10C510D0-10FA10FC1100-1248124A-124D1250-12561258125A-125D1260-1288128A-128D1290-12B012B2-12B512B8-12BE12C012C2-12C512C8-12D612D8-13101312-13151318-135A1380-138F13A0-13F41401-166C166F-167F1681-169A16A0-16EA1700-170C170E-17111720-17311740-17511760-176C176E-17701780-17B317D717DC1820-18771880-18A818AA18B0-18F51900-191C1950-196D1970-19741980-19AB19C1-19C71A00-1A161A20-1A541AA71B05-1B331B45-1B4B1B83-1BA01BAE1BAF1C00-1C231C4D-1C4F1C5A-1C7D1CE9-1CEC1CEE-1CF11D00-1DBF1E00-1F151F18-1F1D1F20-1F451F48-1F4D1F50-1F571F591F5B1F5D1F5F-1F7D1F80-1FB41FB6-1FBC1FBE1FC2-1FC41FC6-1FCC1FD0-1FD31FD6-1FDB1FE0-1FEC1FF2-1FF41FF6-1FFC2071207F2090-209421022107210A-211321152119-211D212421262128212A-212D212F-2139213C-213F2145-2149214E218321842C00-2C2E2C30-2C5E2C60-2CE42CEB-2CEE2D00-2D252D30-2D652D6F2D80-2D962DA0-2DA62DA8-2DAE2DB0-2DB62DB8-2DBE2DC0-2DC62DC8-2DCE2DD0-2DD62DD8-2DDE2E2F300530063031-3035303B303C3041-3096309D-309F30A1-30FA30FC-30FF3105-312D3131-318E31A0-31B731F0-31FF3400-4DB54E00-9FCBA000-A48CA4D0-A4FDA500-A60CA610-A61FA62AA62BA640-A65FA662-A66EA67F-A697A6A0-A6E5A717-A71FA722-A788A78BA78CA7FB-A801A803-A805A807-A80AA80C-A822A840-A873A882-A8B3A8F2-A8F7A8FBA90A-A925A930-A946A960-A97CA984-A9B2A9CFAA00-AA28AA40-AA42AA44-AA4BAA60-AA76AA7AAA80-AAAFAAB1AAB5AAB6AAB9-AABDAAC0AAC2AADB-AADDABC0-ABE2AC00-D7A3D7B0-D7C6D7CB-D7FBF900-FA2DFA30-FA6DFA70-FAD9FB00-FB06FB13-FB17FB1DFB1F-FB28FB2A-FB36FB38-FB3CFB3EFB40FB41FB43FB44FB46-FBB1FBD3-FD3DFD50-FD8FFD92-FDC7FDF0-FDFBFE70-FE74FE76-FEFCFF21-FF3AFF41-FF5AFF66-FFBEFFC2-FFC7FFCA-FFCFFFD2-FFD7FFDA-FFDC",Ll:"0061-007A00AA00B500BA00DF-00F600F8-00FF01010103010501070109010B010D010F01110113011501170119011B011D011F01210123012501270129012B012D012F01310133013501370138013A013C013E014001420144014601480149014B014D014F01510153015501570159015B015D015F01610163016501670169016B016D016F0171017301750177017A017C017E-0180018301850188018C018D019201950199-019B019E01A101A301A501A801AA01AB01AD01B001B401B601B901BA01BD-01BF01C601C901CC01CE01D001D201D401D601D801DA01DC01DD01DF01E101E301E501E701E901EB01ED01EF01F001F301F501F901FB01FD01FF02010203020502070209020B020D020F02110213021502170219021B021D021F02210223022502270229022B022D022F02310233-0239023C023F0240024202470249024B024D024F-02930295-02AF037103730377037B-037D039003AC-03CE03D003D103D5-03D703D903DB03DD03DF03E103E303E503E703E903EB03ED03EF-03F303F503F803FB03FC0430-045F04610463046504670469046B046D046F04710473047504770479047B047D047F0481048B048D048F04910493049504970499049B049D049F04A104A304A504A704A904AB04AD04AF04B104B304B504B704B904BB04BD04BF04C204C404C604C804CA04CC04CE04CF04D104D304D504D704D904DB04DD04DF04E104E304E504E704E904EB04ED04EF04F104F304F504F704F904FB04FD04FF05010503050505070509050B050D050F05110513051505170519051B051D051F0521052305250561-05871D00-1D2B1D62-1D771D79-1D9A1E011E031E051E071E091E0B1E0D1E0F1E111E131E151E171E191E1B1E1D1E1F1E211E231E251E271E291E2B1E2D1E2F1E311E331E351E371E391E3B1E3D1E3F1E411E431E451E471E491E4B1E4D1E4F1E511E531E551E571E591E5B1E5D1E5F1E611E631E651E671E691E6B1E6D1E6F1E711E731E751E771E791E7B1E7D1E7F1E811E831E851E871E891E8B1E8D1E8F1E911E931E95-1E9D1E9F1EA11EA31EA51EA71EA91EAB1EAD1EAF1EB11EB31EB51EB71EB91EBB1EBD1EBF1EC11EC31EC51EC71EC91ECB1ECD1ECF1ED11ED31ED51ED71ED91EDB1EDD1EDF1EE11EE31EE51EE71EE91EEB1EED1EEF1EF11EF31EF51EF71EF91EFB1EFD1EFF-1F071F10-1F151F20-1F271F30-1F371F40-1F451F50-1F571F60-1F671F70-1F7D1F80-1F871F90-1F971FA0-1FA71FB0-1FB41FB61FB71FBE1FC2-1FC41FC61FC71FD0-1FD31FD61FD71FE0-1FE71FF2-1FF41FF61FF7210A210E210F2113212F21342139213C213D2146-2149214E21842C30-2C5E2C612C652C662C682C6A2C6C2C712C732C742C76-2C7C2C812C832C852C872C892C8B2C8D2C8F2C912C932C952C972C992C9B2C9D2C9F2CA12CA32CA52CA72CA92CAB2CAD2CAF2CB12CB32CB52CB72CB92CBB2CBD2CBF2CC12CC32CC52CC72CC92CCB2CCD2CCF2CD12CD32CD52CD72CD92CDB2CDD2CDF2CE12CE32CE42CEC2CEE2D00-2D25A641A643A645A647A649A64BA64DA64FA651A653A655A657A659A65BA65DA65FA663A665A667A669A66BA66DA681A683A685A687A689A68BA68DA68FA691A693A695A697A723A725A727A729A72BA72DA72F-A731A733A735A737A739A73BA73DA73FA741A743A745A747A749A74BA74DA74FA751A753A755A757A759A75BA75DA75FA761A763A765A767A769A76BA76DA76FA771-A778A77AA77CA77FA781A783A785A787A78CFB00-FB06FB13-FB17FF41-FF5A",Lu:"0041-005A00C0-00D600D8-00DE01000102010401060108010A010C010E01100112011401160118011A011C011E01200122012401260128012A012C012E01300132013401360139013B013D013F0141014301450147014A014C014E01500152015401560158015A015C015E01600162016401660168016A016C016E017001720174017601780179017B017D018101820184018601870189-018B018E-0191019301940196-0198019C019D019F01A001A201A401A601A701A901AC01AE01AF01B1-01B301B501B701B801BC01C401C701CA01CD01CF01D101D301D501D701D901DB01DE01E001E201E401E601E801EA01EC01EE01F101F401F6-01F801FA01FC01FE02000202020402060208020A020C020E02100212021402160218021A021C021E02200222022402260228022A022C022E02300232023A023B023D023E02410243-02460248024A024C024E03700372037603860388-038A038C038E038F0391-03A103A3-03AB03CF03D2-03D403D803DA03DC03DE03E003E203E403E603E803EA03EC03EE03F403F703F903FA03FD-042F04600462046404660468046A046C046E04700472047404760478047A047C047E0480048A048C048E04900492049404960498049A049C049E04A004A204A404A604A804AA04AC04AE04B004B204B404B604B804BA04BC04BE04C004C104C304C504C704C904CB04CD04D004D204D404D604D804DA04DC04DE04E004E204E404E604E804EA04EC04EE04F004F204F404F604F804FA04FC04FE05000502050405060508050A050C050E05100512051405160518051A051C051E0520052205240531-055610A0-10C51E001E021E041E061E081E0A1E0C1E0E1E101E121E141E161E181E1A1E1C1E1E1E201E221E241E261E281E2A1E2C1E2E1E301E321E341E361E381E3A1E3C1E3E1E401E421E441E461E481E4A1E4C1E4E1E501E521E541E561E581E5A1E5C1E5E1E601E621E641E661E681E6A1E6C1E6E1E701E721E741E761E781E7A1E7C1E7E1E801E821E841E861E881E8A1E8C1E8E1E901E921E941E9E1EA01EA21EA41EA61EA81EAA1EAC1EAE1EB01EB21EB41EB61EB81EBA1EBC1EBE1EC01EC21EC41EC61EC81ECA1ECC1ECE1ED01ED21ED41ED61ED81EDA1EDC1EDE1EE01EE21EE41EE61EE81EEA1EEC1EEE1EF01EF21EF41EF61EF81EFA1EFC1EFE1F08-1F0F1F18-1F1D1F28-1F2F1F38-1F3F1F48-1F4D1F591F5B1F5D1F5F1F68-1F6F1FB8-1FBB1FC8-1FCB1FD8-1FDB1FE8-1FEC1FF8-1FFB21022107210B-210D2110-211221152119-211D212421262128212A-212D2130-2133213E213F214521832C00-2C2E2C602C62-2C642C672C692C6B2C6D-2C702C722C752C7E-2C802C822C842C862C882C8A2C8C2C8E2C902C922C942C962C982C9A2C9C2C9E2CA02CA22CA42CA62CA82CAA2CAC2CAE2CB02CB22CB42CB62CB82CBA2CBC2CBE2CC02CC22CC42CC62CC82CCA2CCC2CCE2CD02CD22CD42CD62CD82CDA2CDC2CDE2CE02CE22CEB2CEDA640A642A644A646A648A64AA64CA64EA650A652A654A656A658A65AA65CA65EA662A664A666A668A66AA66CA680A682A684A686A688A68AA68CA68EA690A692A694A696A722A724A726A728A72AA72CA72EA732A734A736A738A73AA73CA73EA740A742A744A746A748A74AA74CA74EA750A752A754A756A758A75AA75CA75EA760A762A764A766A768A76AA76CA76EA779A77BA77DA77EA780A782A784A786A78BFF21-FF3A",Lt:"01C501C801CB01F21F88-1F8F1F98-1F9F1FA8-1FAF1FBC1FCC1FFC",Lm:"02B0-02C102C6-02D102E0-02E402EC02EE0374037A0559064006E506E607F407F507FA081A0824082809710E460EC610FC17D718431AA71C78-1C7D1D2C-1D611D781D9B-1DBF2071207F2090-20942C7D2D6F2E2F30053031-3035303B309D309E30FC-30FEA015A4F8-A4FDA60CA67FA717-A71FA770A788A9CFAA70AADDFF70FF9EFF9F",Lo:"01BB01C0-01C3029405D0-05EA05F0-05F20621-063F0641-064A066E066F0671-06D306D506EE06EF06FA-06FC06FF07100712-072F074D-07A507B107CA-07EA0800-08150904-0939093D09500958-096109720979-097F0985-098C098F09900993-09A809AA-09B009B209B6-09B909BD09CE09DC09DD09DF-09E109F009F10A05-0A0A0A0F0A100A13-0A280A2A-0A300A320A330A350A360A380A390A59-0A5C0A5E0A72-0A740A85-0A8D0A8F-0A910A93-0AA80AAA-0AB00AB20AB30AB5-0AB90ABD0AD00AE00AE10B05-0B0C0B0F0B100B13-0B280B2A-0B300B320B330B35-0B390B3D0B5C0B5D0B5F-0B610B710B830B85-0B8A0B8E-0B900B92-0B950B990B9A0B9C0B9E0B9F0BA30BA40BA8-0BAA0BAE-0BB90BD00C05-0C0C0C0E-0C100C12-0C280C2A-0C330C35-0C390C3D0C580C590C600C610C85-0C8C0C8E-0C900C92-0CA80CAA-0CB30CB5-0CB90CBD0CDE0CE00CE10D05-0D0C0D0E-0D100D12-0D280D2A-0D390D3D0D600D610D7A-0D7F0D85-0D960D9A-0DB10DB3-0DBB0DBD0DC0-0DC60E01-0E300E320E330E40-0E450E810E820E840E870E880E8A0E8D0E94-0E970E99-0E9F0EA1-0EA30EA50EA70EAA0EAB0EAD-0EB00EB20EB30EBD0EC0-0EC40EDC0EDD0F000F40-0F470F49-0F6C0F88-0F8B1000-102A103F1050-1055105A-105D106110651066106E-10701075-1081108E10D0-10FA1100-1248124A-124D1250-12561258125A-125D1260-1288128A-128D1290-12B012B2-12B512B8-12BE12C012C2-12C512C8-12D612D8-13101312-13151318-135A1380-138F13A0-13F41401-166C166F-167F1681-169A16A0-16EA1700-170C170E-17111720-17311740-17511760-176C176E-17701780-17B317DC1820-18421844-18771880-18A818AA18B0-18F51900-191C1950-196D1970-19741980-19AB19C1-19C71A00-1A161A20-1A541B05-1B331B45-1B4B1B83-1BA01BAE1BAF1C00-1C231C4D-1C4F1C5A-1C771CE9-1CEC1CEE-1CF12135-21382D30-2D652D80-2D962DA0-2DA62DA8-2DAE2DB0-2DB62DB8-2DBE2DC0-2DC62DC8-2DCE2DD0-2DD62DD8-2DDE3006303C3041-3096309F30A1-30FA30FF3105-312D3131-318E31A0-31B731F0-31FF3400-4DB54E00-9FCBA000-A014A016-A48CA4D0-A4F7A500-A60BA610-A61FA62AA62BA66EA6A0-A6E5A7FB-A801A803-A805A807-A80AA80C-A822A840-A873A882-A8B3A8F2-A8F7A8FBA90A-A925A930-A946A960-A97CA984-A9B2AA00-AA28AA40-AA42AA44-AA4BAA60-AA6FAA71-AA76AA7AAA80-AAAFAAB1AAB5AAB6AAB9-AABDAAC0AAC2AADBAADCABC0-ABE2AC00-D7A3D7B0-D7C6D7CB-D7FBF900-FA2DFA30-FA6DFA70-FAD9FB1DFB1F-FB28FB2A-FB36FB38-FB3CFB3EFB40FB41FB43FB44FB46-FBB1FBD3-FD3DFD50-FD8FFD92-FDC7FDF0-FDFBFE70-FE74FE76-FEFCFF66-FF6FFF71-FF9DFFA0-FFBEFFC2-FFC7FFCA-FFCFFFD2-FFD7FFDA-FFDC",M:"0300-036F0483-04890591-05BD05BF05C105C205C405C505C70610-061A064B-065E067006D6-06DC06DE-06E406E706E806EA-06ED07110730-074A07A6-07B007EB-07F30816-0819081B-08230825-08270829-082D0900-0903093C093E-094E0951-0955096209630981-098309BC09BE-09C409C709C809CB-09CD09D709E209E30A01-0A030A3C0A3E-0A420A470A480A4B-0A4D0A510A700A710A750A81-0A830ABC0ABE-0AC50AC7-0AC90ACB-0ACD0AE20AE30B01-0B030B3C0B3E-0B440B470B480B4B-0B4D0B560B570B620B630B820BBE-0BC20BC6-0BC80BCA-0BCD0BD70C01-0C030C3E-0C440C46-0C480C4A-0C4D0C550C560C620C630C820C830CBC0CBE-0CC40CC6-0CC80CCA-0CCD0CD50CD60CE20CE30D020D030D3E-0D440D46-0D480D4A-0D4D0D570D620D630D820D830DCA0DCF-0DD40DD60DD8-0DDF0DF20DF30E310E34-0E3A0E47-0E4E0EB10EB4-0EB90EBB0EBC0EC8-0ECD0F180F190F350F370F390F3E0F3F0F71-0F840F860F870F90-0F970F99-0FBC0FC6102B-103E1056-1059105E-10601062-10641067-106D1071-10741082-108D108F109A-109D135F1712-17141732-1734175217531772177317B6-17D317DD180B-180D18A91920-192B1930-193B19B0-19C019C819C91A17-1A1B1A55-1A5E1A60-1A7C1A7F1B00-1B041B34-1B441B6B-1B731B80-1B821BA1-1BAA1C24-1C371CD0-1CD21CD4-1CE81CED1CF21DC0-1DE61DFD-1DFF20D0-20F02CEF-2CF12DE0-2DFF302A-302F3099309AA66F-A672A67CA67DA6F0A6F1A802A806A80BA823-A827A880A881A8B4-A8C4A8E0-A8F1A926-A92DA947-A953A980-A983A9B3-A9C0AA29-AA36AA43AA4CAA4DAA7BAAB0AAB2-AAB4AAB7AAB8AABEAABFAAC1ABE3-ABEAABECABEDFB1EFE00-FE0FFE20-FE26",Mn:"0300-036F0483-04870591-05BD05BF05C105C205C405C505C70610-061A064B-065E067006D6-06DC06DF-06E406E706E806EA-06ED07110730-074A07A6-07B007EB-07F30816-0819081B-08230825-08270829-082D0900-0902093C0941-0948094D0951-095509620963098109BC09C1-09C409CD09E209E30A010A020A3C0A410A420A470A480A4B-0A4D0A510A700A710A750A810A820ABC0AC1-0AC50AC70AC80ACD0AE20AE30B010B3C0B3F0B41-0B440B4D0B560B620B630B820BC00BCD0C3E-0C400C46-0C480C4A-0C4D0C550C560C620C630CBC0CBF0CC60CCC0CCD0CE20CE30D41-0D440D4D0D620D630DCA0DD2-0DD40DD60E310E34-0E3A0E47-0E4E0EB10EB4-0EB90EBB0EBC0EC8-0ECD0F180F190F350F370F390F71-0F7E0F80-0F840F860F870F90-0F970F99-0FBC0FC6102D-10301032-10371039103A103D103E10581059105E-10601071-1074108210851086108D109D135F1712-17141732-1734175217531772177317B7-17BD17C617C9-17D317DD180B-180D18A91920-19221927192819321939-193B1A171A181A561A58-1A5E1A601A621A65-1A6C1A73-1A7C1A7F1B00-1B031B341B36-1B3A1B3C1B421B6B-1B731B801B811BA2-1BA51BA81BA91C2C-1C331C361C371CD0-1CD21CD4-1CE01CE2-1CE81CED1DC0-1DE61DFD-1DFF20D0-20DC20E120E5-20F02CEF-2CF12DE0-2DFF302A-302F3099309AA66FA67CA67DA6F0A6F1A802A806A80BA825A826A8C4A8E0-A8F1A926-A92DA947-A951A980-A982A9B3A9B6-A9B9A9BCAA29-AA2EAA31AA32AA35AA36AA43AA4CAAB0AAB2-AAB4AAB7AAB8AABEAABFAAC1ABE5ABE8ABEDFB1EFE00-FE0FFE20-FE26",Mc:"0903093E-09400949-094C094E0982098309BE-09C009C709C809CB09CC09D70A030A3E-0A400A830ABE-0AC00AC90ACB0ACC0B020B030B3E0B400B470B480B4B0B4C0B570BBE0BBF0BC10BC20BC6-0BC80BCA-0BCC0BD70C01-0C030C41-0C440C820C830CBE0CC0-0CC40CC70CC80CCA0CCB0CD50CD60D020D030D3E-0D400D46-0D480D4A-0D4C0D570D820D830DCF-0DD10DD8-0DDF0DF20DF30F3E0F3F0F7F102B102C10311038103B103C105610571062-10641067-106D108310841087-108C108F109A-109C17B617BE-17C517C717C81923-19261929-192B193019311933-193819B0-19C019C819C91A19-1A1B1A551A571A611A631A641A6D-1A721B041B351B3B1B3D-1B411B431B441B821BA11BA61BA71BAA1C24-1C2B1C341C351CE11CF2A823A824A827A880A881A8B4-A8C3A952A953A983A9B4A9B5A9BAA9BBA9BD-A9C0AA2FAA30AA33AA34AA4DAA7BABE3ABE4ABE6ABE7ABE9ABEAABEC",Me:"0488048906DE20DD-20E020E2-20E4A670-A672",N:"0030-003900B200B300B900BC-00BE0660-066906F0-06F907C0-07C90966-096F09E6-09EF09F4-09F90A66-0A6F0AE6-0AEF0B66-0B6F0BE6-0BF20C66-0C6F0C78-0C7E0CE6-0CEF0D66-0D750E50-0E590ED0-0ED90F20-0F331040-10491090-10991369-137C16EE-16F017E0-17E917F0-17F91810-18191946-194F19D0-19DA1A80-1A891A90-1A991B50-1B591BB0-1BB91C40-1C491C50-1C5920702074-20792080-20892150-21822185-21892460-249B24EA-24FF2776-27932CFD30073021-30293038-303A3192-31953220-32293251-325F3280-328932B1-32BFA620-A629A6E6-A6EFA830-A835A8D0-A8D9A900-A909A9D0-A9D9AA50-AA59ABF0-ABF9FF10-FF19",Nd:"0030-00390660-066906F0-06F907C0-07C90966-096F09E6-09EF0A66-0A6F0AE6-0AEF0B66-0B6F0BE6-0BEF0C66-0C6F0CE6-0CEF0D66-0D6F0E50-0E590ED0-0ED90F20-0F291040-10491090-109917E0-17E91810-18191946-194F19D0-19DA1A80-1A891A90-1A991B50-1B591BB0-1BB91C40-1C491C50-1C59A620-A629A8D0-A8D9A900-A909A9D0-A9D9AA50-AA59ABF0-ABF9FF10-FF19",Nl:"16EE-16F02160-21822185-218830073021-30293038-303AA6E6-A6EF",No:"00B200B300B900BC-00BE09F4-09F90BF0-0BF20C78-0C7E0D70-0D750F2A-0F331369-137C17F0-17F920702074-20792080-20892150-215F21892460-249B24EA-24FF2776-27932CFD3192-31953220-32293251-325F3280-328932B1-32BFA830-A835",P:"0021-00230025-002A002C-002F003A003B003F0040005B-005D005F007B007D00A100AB00B700BB00BF037E0387055A-055F0589058A05BE05C005C305C605F305F40609060A060C060D061B061E061F066A-066D06D40700-070D07F7-07F90830-083E0964096509700DF40E4F0E5A0E5B0F04-0F120F3A-0F3D0F850FD0-0FD4104A-104F10FB1361-13681400166D166E169B169C16EB-16ED1735173617D4-17D617D8-17DA1800-180A1944194519DE19DF1A1E1A1F1AA0-1AA61AA8-1AAD1B5A-1B601C3B-1C3F1C7E1C7F1CD32010-20272030-20432045-20512053-205E207D207E208D208E2329232A2768-277527C527C627E6-27EF2983-299829D8-29DB29FC29FD2CF9-2CFC2CFE2CFF2E00-2E2E2E302E313001-30033008-30113014-301F3030303D30A030FBA4FEA4FFA60D-A60FA673A67EA6F2-A6F7A874-A877A8CEA8CFA8F8-A8FAA92EA92FA95FA9C1-A9CDA9DEA9DFAA5C-AA5FAADEAADFABEBFD3EFD3FFE10-FE19FE30-FE52FE54-FE61FE63FE68FE6AFE6BFF01-FF03FF05-FF0AFF0C-FF0FFF1AFF1BFF1FFF20FF3B-FF3DFF3FFF5BFF5DFF5F-FF65",Pd:"002D058A05BE140018062010-20152E172E1A301C303030A0FE31FE32FE58FE63FF0D",Ps:"0028005B007B0F3A0F3C169B201A201E2045207D208D23292768276A276C276E27702772277427C527E627E827EA27EC27EE2983298529872989298B298D298F299129932995299729D829DA29FC2E222E242E262E283008300A300C300E3010301430163018301A301DFD3EFE17FE35FE37FE39FE3BFE3DFE3FFE41FE43FE47FE59FE5BFE5DFF08FF3BFF5BFF5FFF62",Pe:"0029005D007D0F3B0F3D169C2046207E208E232A2769276B276D276F27712773277527C627E727E927EB27ED27EF298429862988298A298C298E2990299229942996299829D929DB29FD2E232E252E272E293009300B300D300F3011301530173019301B301E301FFD3FFE18FE36FE38FE3AFE3CFE3EFE40FE42FE44FE48FE5AFE5CFE5EFF09FF3DFF5DFF60FF63",Pi:"00AB2018201B201C201F20392E022E042E092E0C2E1C2E20",Pf:"00BB2019201D203A2E032E052E0A2E0D2E1D2E21",Pc:"005F203F20402054FE33FE34FE4D-FE4FFF3F",Po:"0021-00230025-0027002A002C002E002F003A003B003F0040005C00A100B700BF037E0387055A-055F058905C005C305C605F305F40609060A060C060D061B061E061F066A-066D06D40700-070D07F7-07F90830-083E0964096509700DF40E4F0E5A0E5B0F04-0F120F850FD0-0FD4104A-104F10FB1361-1368166D166E16EB-16ED1735173617D4-17D617D8-17DA1800-18051807-180A1944194519DE19DF1A1E1A1F1AA0-1AA61AA8-1AAD1B5A-1B601C3B-1C3F1C7E1C7F1CD3201620172020-20272030-2038203B-203E2041-20432047-205120532055-205E2CF9-2CFC2CFE2CFF2E002E012E06-2E082E0B2E0E-2E162E182E192E1B2E1E2E1F2E2A-2E2E2E302E313001-3003303D30FBA4FEA4FFA60D-A60FA673A67EA6F2-A6F7A874-A877A8CEA8CFA8F8-A8FAA92EA92FA95FA9C1-A9CDA9DEA9DFAA5C-AA5FAADEAADFABEBFE10-FE16FE19FE30FE45FE46FE49-FE4CFE50-FE52FE54-FE57FE5F-FE61FE68FE6AFE6BFF01-FF03FF05-FF07FF0AFF0CFF0EFF0FFF1AFF1BFF1FFF20FF3CFF61FF64FF65",S:"0024002B003C-003E005E0060007C007E00A2-00A900AC00AE-00B100B400B600B800D700F702C2-02C502D2-02DF02E5-02EB02ED02EF-02FF03750384038503F604820606-0608060B060E060F06E906FD06FE07F609F209F309FA09FB0AF10B700BF3-0BFA0C7F0CF10CF20D790E3F0F01-0F030F13-0F170F1A-0F1F0F340F360F380FBE-0FC50FC7-0FCC0FCE0FCF0FD5-0FD8109E109F13601390-139917DB194019E0-19FF1B61-1B6A1B74-1B7C1FBD1FBF-1FC11FCD-1FCF1FDD-1FDF1FED-1FEF1FFD1FFE20442052207A-207C208A-208C20A0-20B8210021012103-21062108210921142116-2118211E-2123212521272129212E213A213B2140-2144214A-214D214F2190-2328232B-23E82400-24262440-244A249C-24E92500-26CD26CF-26E126E326E8-26FF2701-27042706-2709270C-27272729-274B274D274F-27522756-275E2761-276727942798-27AF27B1-27BE27C0-27C427C7-27CA27CC27D0-27E527F0-29822999-29D729DC-29FB29FE-2B4C2B50-2B592CE5-2CEA2E80-2E992E9B-2EF32F00-2FD52FF0-2FFB300430123013302030363037303E303F309B309C319031913196-319F31C0-31E33200-321E322A-32503260-327F328A-32B032C0-32FE3300-33FF4DC0-4DFFA490-A4C6A700-A716A720A721A789A78AA828-A82BA836-A839AA77-AA79FB29FDFCFDFDFE62FE64-FE66FE69FF04FF0BFF1C-FF1EFF3EFF40FF5CFF5EFFE0-FFE6FFE8-FFEEFFFCFFFD",Sm:"002B003C-003E007C007E00AC00B100D700F703F60606-060820442052207A-207C208A-208C2140-2144214B2190-2194219A219B21A021A321A621AE21CE21CF21D221D421F4-22FF2308-230B23202321237C239B-23B323DC-23E125B725C125F8-25FF266F27C0-27C427C7-27CA27CC27D0-27E527F0-27FF2900-29822999-29D729DC-29FB29FE-2AFF2B30-2B442B47-2B4CFB29FE62FE64-FE66FF0BFF1C-FF1EFF5CFF5EFFE2FFE9-FFEC",Sc:"002400A2-00A5060B09F209F309FB0AF10BF90E3F17DB20A0-20B8A838FDFCFE69FF04FFE0FFE1FFE5FFE6",Sk:"005E006000A800AF00B400B802C2-02C502D2-02DF02E5-02EB02ED02EF-02FF0375038403851FBD1FBF-1FC11FCD-1FCF1FDD-1FDF1FED-1FEF1FFD1FFE309B309CA700-A716A720A721A789A78AFF3EFF40FFE3",So:"00A600A700A900AE00B000B60482060E060F06E906FD06FE07F609FA0B700BF3-0BF80BFA0C7F0CF10CF20D790F01-0F030F13-0F170F1A-0F1F0F340F360F380FBE-0FC50FC7-0FCC0FCE0FCF0FD5-0FD8109E109F13601390-1399194019E0-19FF1B61-1B6A1B74-1B7C210021012103-21062108210921142116-2118211E-2123212521272129212E213A213B214A214C214D214F2195-2199219C-219F21A121A221A421A521A7-21AD21AF-21CD21D021D121D321D5-21F32300-2307230C-231F2322-2328232B-237B237D-239A23B4-23DB23E2-23E82400-24262440-244A249C-24E92500-25B625B8-25C025C2-25F72600-266E2670-26CD26CF-26E126E326E8-26FF2701-27042706-2709270C-27272729-274B274D274F-27522756-275E2761-276727942798-27AF27B1-27BE2800-28FF2B00-2B2F2B452B462B50-2B592CE5-2CEA2E80-2E992E9B-2EF32F00-2FD52FF0-2FFB300430123013302030363037303E303F319031913196-319F31C0-31E33200-321E322A-32503260-327F328A-32B032C0-32FE3300-33FF4DC0-4DFFA490-A4C6A828-A82BA836A837A839AA77-AA79FDFDFFE4FFE8FFEDFFEEFFFCFFFD",Z:"002000A01680180E2000-200A20282029202F205F3000",Zs:"002000A01680180E2000-200A202F205F3000",Zl:"2028",Zp:"2029",C:"0000-001F007F-009F00AD03780379037F-0383038B038D03A20526-05300557055805600588058B-059005C8-05CF05EB-05EF05F5-0605061C061D0620065F06DD070E070F074B074C07B2-07BF07FB-07FF082E082F083F-08FF093A093B094F095609570973-097809800984098D098E0991099209A909B109B3-09B509BA09BB09C509C609C909CA09CF-09D609D8-09DB09DE09E409E509FC-0A000A040A0B-0A0E0A110A120A290A310A340A370A3A0A3B0A3D0A43-0A460A490A4A0A4E-0A500A52-0A580A5D0A5F-0A650A76-0A800A840A8E0A920AA90AB10AB40ABA0ABB0AC60ACA0ACE0ACF0AD1-0ADF0AE40AE50AF00AF2-0B000B040B0D0B0E0B110B120B290B310B340B3A0B3B0B450B460B490B4A0B4E-0B550B58-0B5B0B5E0B640B650B72-0B810B840B8B-0B8D0B910B96-0B980B9B0B9D0BA0-0BA20BA5-0BA70BAB-0BAD0BBA-0BBD0BC3-0BC50BC90BCE0BCF0BD1-0BD60BD8-0BE50BFB-0C000C040C0D0C110C290C340C3A-0C3C0C450C490C4E-0C540C570C5A-0C5F0C640C650C70-0C770C800C810C840C8D0C910CA90CB40CBA0CBB0CC50CC90CCE-0CD40CD7-0CDD0CDF0CE40CE50CF00CF3-0D010D040D0D0D110D290D3A-0D3C0D450D490D4E-0D560D58-0D5F0D640D650D76-0D780D800D810D840D97-0D990DB20DBC0DBE0DBF0DC7-0DC90DCB-0DCE0DD50DD70DE0-0DF10DF5-0E000E3B-0E3E0E5C-0E800E830E850E860E890E8B0E8C0E8E-0E930E980EA00EA40EA60EA80EA90EAC0EBA0EBE0EBF0EC50EC70ECE0ECF0EDA0EDB0EDE-0EFF0F480F6D-0F700F8C-0F8F0F980FBD0FCD0FD9-0FFF10C6-10CF10FD-10FF1249124E124F12571259125E125F1289128E128F12B112B612B712BF12C112C612C712D7131113161317135B-135E137D-137F139A-139F13F5-13FF169D-169F16F1-16FF170D1715-171F1737-173F1754-175F176D17711774-177F17B417B517DE17DF17EA-17EF17FA-17FF180F181A-181F1878-187F18AB-18AF18F6-18FF191D-191F192C-192F193C-193F1941-1943196E196F1975-197F19AC-19AF19CA-19CF19DB-19DD1A1C1A1D1A5F1A7D1A7E1A8A-1A8F1A9A-1A9F1AAE-1AFF1B4C-1B4F1B7D-1B7F1BAB-1BAD1BBA-1BFF1C38-1C3A1C4A-1C4C1C80-1CCF1CF3-1CFF1DE7-1DFC1F161F171F1E1F1F1F461F471F4E1F4F1F581F5A1F5C1F5E1F7E1F7F1FB51FC51FD41FD51FDC1FF01FF11FF51FFF200B-200F202A-202E2060-206F20722073208F2095-209F20B9-20CF20F1-20FF218A-218F23E9-23FF2427-243F244B-245F26CE26E226E4-26E727002705270A270B2728274C274E2753-2755275F27602795-279727B027BF27CB27CD-27CF2B4D-2B4F2B5A-2BFF2C2F2C5F2CF2-2CF82D26-2D2F2D66-2D6E2D70-2D7F2D97-2D9F2DA72DAF2DB72DBF2DC72DCF2DD72DDF2E32-2E7F2E9A2EF4-2EFF2FD6-2FEF2FFC-2FFF3040309730983100-3104312E-3130318F31B8-31BF31E4-31EF321F32FF4DB6-4DBF9FCC-9FFFA48D-A48FA4C7-A4CFA62C-A63FA660A661A674-A67BA698-A69FA6F8-A6FFA78D-A7FAA82C-A82FA83A-A83FA878-A87FA8C5-A8CDA8DA-A8DFA8FC-A8FFA954-A95EA97D-A97FA9CEA9DA-A9DDA9E0-A9FFAA37-AA3FAA4EAA4FAA5AAA5BAA7C-AA7FAAC3-AADAAAE0-ABBFABEEABEFABFA-ABFFD7A4-D7AFD7C7-D7CAD7FC-F8FFFA2EFA2FFA6EFA6FFADA-FAFFFB07-FB12FB18-FB1CFB37FB3DFB3FFB42FB45FBB2-FBD2FD40-FD4FFD90FD91FDC8-FDEFFDFEFDFFFE1A-FE1FFE27-FE2FFE53FE67FE6C-FE6FFE75FEFD-FF00FFBF-FFC1FFC8FFC9FFD0FFD1FFD8FFD9FFDD-FFDFFFE7FFEF-FFFBFFFEFFFF",Cc:"0000-001F007F-009F",Cf:"00AD0600-060306DD070F17B417B5200B-200F202A-202E2060-2064206A-206FFEFFFFF9-FFFB",Co:"E000-F8FF",Cs:"D800-DFFF",Cn:"03780379037F-0383038B038D03A20526-05300557055805600588058B-059005C8-05CF05EB-05EF05F5-05FF06040605061C061D0620065F070E074B074C07B2-07BF07FB-07FF082E082F083F-08FF093A093B094F095609570973-097809800984098D098E0991099209A909B109B3-09B509BA09BB09C509C609C909CA09CF-09D609D8-09DB09DE09E409E509FC-0A000A040A0B-0A0E0A110A120A290A310A340A370A3A0A3B0A3D0A43-0A460A490A4A0A4E-0A500A52-0A580A5D0A5F-0A650A76-0A800A840A8E0A920AA90AB10AB40ABA0ABB0AC60ACA0ACE0ACF0AD1-0ADF0AE40AE50AF00AF2-0B000B040B0D0B0E0B110B120B290B310B340B3A0B3B0B450B460B490B4A0B4E-0B550B58-0B5B0B5E0B640B650B72-0B810B840B8B-0B8D0B910B96-0B980B9B0B9D0BA0-0BA20BA5-0BA70BAB-0BAD0BBA-0BBD0BC3-0BC50BC90BCE0BCF0BD1-0BD60BD8-0BE50BFB-0C000C040C0D0C110C290C340C3A-0C3C0C450C490C4E-0C540C570C5A-0C5F0C640C650C70-0C770C800C810C840C8D0C910CA90CB40CBA0CBB0CC50CC90CCE-0CD40CD7-0CDD0CDF0CE40CE50CF00CF3-0D010D040D0D0D110D290D3A-0D3C0D450D490D4E-0D560D58-0D5F0D640D650D76-0D780D800D810D840D97-0D990DB20DBC0DBE0DBF0DC7-0DC90DCB-0DCE0DD50DD70DE0-0DF10DF5-0E000E3B-0E3E0E5C-0E800E830E850E860E890E8B0E8C0E8E-0E930E980EA00EA40EA60EA80EA90EAC0EBA0EBE0EBF0EC50EC70ECE0ECF0EDA0EDB0EDE-0EFF0F480F6D-0F700F8C-0F8F0F980FBD0FCD0FD9-0FFF10C6-10CF10FD-10FF1249124E124F12571259125E125F1289128E128F12B112B612B712BF12C112C612C712D7131113161317135B-135E137D-137F139A-139F13F5-13FF169D-169F16F1-16FF170D1715-171F1737-173F1754-175F176D17711774-177F17DE17DF17EA-17EF17FA-17FF180F181A-181F1878-187F18AB-18AF18F6-18FF191D-191F192C-192F193C-193F1941-1943196E196F1975-197F19AC-19AF19CA-19CF19DB-19DD1A1C1A1D1A5F1A7D1A7E1A8A-1A8F1A9A-1A9F1AAE-1AFF1B4C-1B4F1B7D-1B7F1BAB-1BAD1BBA-1BFF1C38-1C3A1C4A-1C4C1C80-1CCF1CF3-1CFF1DE7-1DFC1F161F171F1E1F1F1F461F471F4E1F4F1F581F5A1F5C1F5E1F7E1F7F1FB51FC51FD41FD51FDC1FF01FF11FF51FFF2065-206920722073208F2095-209F20B9-20CF20F1-20FF218A-218F23E9-23FF2427-243F244B-245F26CE26E226E4-26E727002705270A270B2728274C274E2753-2755275F27602795-279727B027BF27CB27CD-27CF2B4D-2B4F2B5A-2BFF2C2F2C5F2CF2-2CF82D26-2D2F2D66-2D6E2D70-2D7F2D97-2D9F2DA72DAF2DB72DBF2DC72DCF2DD72DDF2E32-2E7F2E9A2EF4-2EFF2FD6-2FEF2FFC-2FFF3040309730983100-3104312E-3130318F31B8-31BF31E4-31EF321F32FF4DB6-4DBF9FCC-9FFFA48D-A48FA4C7-A4CFA62C-A63FA660A661A674-A67BA698-A69FA6F8-A6FFA78D-A7FAA82C-A82FA83A-A83FA878-A87FA8C5-A8CDA8DA-A8DFA8FC-A8FFA954-A95EA97D-A97FA9CEA9DA-A9DDA9E0-A9FFAA37-AA3FAA4EAA4FAA5AAA5BAA7C-AA7FAAC3-AADAAAE0-ABBFABEEABEFABFA-ABFFD7A4-D7AFD7C7-D7CAD7FC-D7FFFA2EFA2FFA6EFA6FFADA-FAFFFB07-FB12FB18-FB1CFB37FB3DFB3FFB42FB45FBB2-FBD2FD40-FD4FFD90FD91FDC8-FDEFFDFEFDFFFE1A-FE1FFE27-FE2FFE53FE67FE6C-FE6FFE75FEFDFEFEFF00FFBF-FFC1FFC8FFC9FFD0FFD1FFD8FFD9FFDD-FFDFFFE7FFEF-FFF8FFFEFFFF"})})),ace.define("ace/mode/text",["require","exports","module","ace/tokenizer","ace/mode/text_highlight_rules","ace/mode/behaviour/cstyle","ace/unicode","ace/lib/lang","ace/token_iterator","ace/range"],(function(e,t,n){"use strict";var i=e("../tokenizer").Tokenizer,r=e("./text_highlight_rules").TextHighlightRules,o=e("./behaviour/cstyle").CstyleBehaviour,s=e("../unicode"),a=e("../lib/lang"),l=e("../token_iterator").TokenIterator,u=e("../range").Range,c=function(){this.HighlightRules=r};(function(){this.$defaultBehaviour=new o,this.tokenRe=new RegExp("^["+s.packages.L+s.packages.Mn+s.packages.Mc+s.packages.Nd+s.packages.Pc+"\\$_]+","g"),this.nonTokenRe=new RegExp("^(?:[^"+s.packages.L+s.packages.Mn+s.packages.Mc+s.packages.Nd+s.packages.Pc+"\\$_]|\\s])+","g"),this.getTokenizer=function(){return this.$tokenizer||(this.$highlightRules=this.$highlightRules||new this.HighlightRules(this.$highlightRuleConfig),this.$tokenizer=new i(this.$highlightRules.getRules())),this.$tokenizer},this.lineCommentStart="",this.blockComment="",this.toggleCommentLines=function(e,t,n,i){var r=t.doc,o=!0,s=!0,l=1/0,u=t.getTabSize(),c=!1;if(this.lineCommentStart){if(Array.isArray(this.lineCommentStart))g=this.lineCommentStart.map(a.escapeRegExp).join("|"),d=this.lineCommentStart[0];else g=a.escapeRegExp(this.lineCommentStart),d=this.lineCommentStart;g=new RegExp("^(\\s*)(?:"+g+") ?"),c=t.getUseSoftTabs();b=function(e,t){var n=e.match(g);if(n){var i=n[1].length,o=n[0].length;f(e,i,o)||" "!=n[0][o-1]||o--,r.removeInLine(t,i,o)}};var h=d+" ",f=(v=function(e,t){o&&!/\S/.test(e)||(f(e,l,l)?r.insertInLine({row:t,column:l},h):r.insertInLine({row:t,column:l},d))},y=function(e,t){return g.test(e)},function(e,t,n){for(var i=0;t--&&" "==e.charAt(t);)i++;if(i%u!=0)return!1;for(i=0;" "==e.charAt(n++);)i++;return u>2?i%u!=u-1:i%u==0})}else{if(!this.blockComment)return!1;var d=this.blockComment.start,p=this.blockComment.end,g=new RegExp("^(\\s*)(?:"+a.escapeRegExp(d)+")"),m=new RegExp("(?:"+a.escapeRegExp(p)+")\\s*$"),v=function(e,t){y(e,t)||o&&!/\S/.test(e)||(r.insertInLine({row:t,column:e.length},p),r.insertInLine({row:t,column:l},d))},b=function(e,t){var n;(n=e.match(m))&&r.removeInLine(t,e.length-n[0].length,e.length),(n=e.match(g))&&r.removeInLine(t,n[1].length,n[0].length)},y=function(e,n){if(g.test(e))return!0;for(var i=t.getTokens(n),r=0;r<i.length;r++)if("comment"===i[r].type)return!0}}function w(e){for(var t=n;t<=i;t++)e(r.getLine(t),t)}var A=1/0;w((function(e,t){var n=e.search(/\S/);-1!==n?(n<l&&(l=n),s&&!y(e,t)&&(s=!1)):A>e.length&&(A=e.length)})),l==1/0&&(l=A,o=!1,s=!1),c&&l%u!=0&&(l=Math.floor(l/u)*u),w(s?b:v)},this.toggleBlockComment=function(e,t,n,i){var r=this.blockComment;if(r){!r.start&&r[0]&&(r=r[0]);var o,s,a=(g=new l(t,i.row,i.column)).getCurrentToken(),c=(t.selection,t.selection.toOrientedRange());if(a&&/comment/.test(a.type)){for(var h,f;a&&/comment/.test(a.type);){if(-1!=(m=a.value.indexOf(r.start))){var d=g.getCurrentTokenRow(),p=g.getCurrentTokenColumn()+m;h=new u(d,p,d,p+r.start.length);break}a=g.stepBackward()}var g;for(a=(g=new l(t,i.row,i.column)).getCurrentToken();a&&/comment/.test(a.type);){var m;if(-1!=(m=a.value.indexOf(r.end))){d=g.getCurrentTokenRow(),p=g.getCurrentTokenColumn()+m;f=new u(d,p,d,p+r.end.length);break}a=g.stepForward()}f&&t.remove(f),h&&(t.remove(h),o=h.start.row,s=-r.start.length)}else s=r.start.length,o=n.start.row,t.insert(n.end,r.end),t.insert(n.start,r.start);c.start.row==o&&(c.start.column+=s),c.end.row==o&&(c.end.column+=s),t.selection.fromOrientedRange(c)}},this.getNextLineIndent=function(e,t,n){return this.$getIndent(t)},this.checkOutdent=function(e,t,n){return!1},this.autoOutdent=function(e,t,n){},this.$getIndent=function(e){return e.match(/^\s*/)[0]},this.createWorker=function(e){return null},this.createModeDelegates=function(e){for(var t in this.$embeds=[],this.$modes={},e)e[t]&&(this.$embeds.push(t),this.$modes[t]=new e[t]);var n=["toggleBlockComment","toggleCommentLines","getNextLineIndent","checkOutdent","autoOutdent","transformAction","getCompletions"];for(t=0;t<n.length;t++)!function(e){var i=n[t],r=e[i];e[n[t]]=function(){return this.$delegator(i,arguments,r)}}(this)},this.$delegator=function(e,t,n){var i=t[0];"string"!=typeof i&&(i=i[0]);for(var r=0;r<this.$embeds.length;r++)if(this.$modes[this.$embeds[r]]){var o=i.split(this.$embeds[r]);if(!o[0]&&o[1]){t[0]=o[1];var s=this.$modes[this.$embeds[r]];return s[e].apply(s,t)}}var a=n.apply(this,t);return n?a:void 0},this.transformAction=function(e,t,n,i,r){if(this.$behaviour){var o=this.$behaviour.getBehaviours();for(var s in o)if(o[s][t]){var a=o[s][t].apply(this,arguments);if(a)return a}}},this.getKeywords=function(e){if(!this.completionKeywords){var t=this.$tokenizer.rules,n=[];for(var i in t)for(var r=t[i],o=0,s=r.length;o<s;o++)if("string"==typeof r[o].token)/keyword|support|storage/.test(r[o].token)&&n.push(r[o].regex);else if("object"==typeof r[o].token)for(var a=0,l=r[o].token.length;a<l;a++)if(/keyword|support|storage/.test(r[o].token[a])){i=r[o].regex.match(/\(.+?\)/g)[a];n.push(i.substr(1,i.length-2))}this.completionKeywords=n}return e?n.concat(this.$keywordList||[]):this.$keywordList},this.$createKeywordList=function(){return this.$highlightRules||this.getTokenizer(),this.$keywordList=this.$highlightRules.$keywordList||[]},this.getCompletions=function(e,t,n,i){return(this.$keywordList||this.$createKeywordList()).map((function(e){return{name:e,value:e,score:0,meta:"keyword"}}))},this.$id="ace/mode/text"}).call(c.prototype),t.Mode=c})),ace.define("ace/apply_delta",["require","exports","module"],(function(e,t,n){"use strict";t.applyDelta=function(e,t,n){var i=t.start.row,r=t.start.column,o=e[i]||"";switch(t.action){case"insert":if(1===t.lines.length)e[i]=o.substring(0,r)+t.lines[0]+o.substring(r);else{var s=[i,1].concat(t.lines);e.splice.apply(e,s),e[i]=o.substring(0,r)+e[i],e[i+t.lines.length-1]+=o.substring(r)}break;case"remove":var a=t.end.column,l=t.end.row;i===l?e[i]=o.substring(0,r)+o.substring(a):e.splice(i,l-i+1,o.substring(0,r)+e[l].substring(a))}}})),ace.define("ace/anchor",["require","exports","module","ace/lib/oop","ace/lib/event_emitter"],(function(e,t,n){"use strict";var i=e("./lib/oop"),r=e("./lib/event_emitter").EventEmitter,o=t.Anchor=function(e,t,n){this.$onChange=this.onChange.bind(this),this.attach(e),void 0===n?this.setPosition(t.row,t.column):this.setPosition(t,n)};(function(){function e(e,t,n){var i=n?e.column<=t.column:e.column<t.column;return e.row<t.row||e.row==t.row&&i}i.implement(this,r),this.getPosition=function(){return this.$clipPositionToDocument(this.row,this.column)},this.getDocument=function(){return this.document},this.$insertRight=!1,this.onChange=function(t){if(!(t.start.row==t.end.row&&t.start.row!=this.row||t.start.row>this.row)){var n=function(t,n,i){var r="insert"==t.action,o=(r?1:-1)*(t.end.row-t.start.row),s=(r?1:-1)*(t.end.column-t.start.column),a=t.start,l=r?a:t.end;if(e(n,a,i))return{row:n.row,column:n.column};if(e(l,n,!i))return{row:n.row+o,column:n.column+(n.row==l.row?s:0)};return{row:a.row,column:a.column}}(t,{row:this.row,column:this.column},this.$insertRight);this.setPosition(n.row,n.column,!0)}},this.setPosition=function(e,t,n){var i;if(i=n?{row:e,column:t}:this.$clipPositionToDocument(e,t),this.row!=i.row||this.column!=i.column){var r={row:this.row,column:this.column};this.row=i.row,this.column=i.column,this._signal("change",{old:r,value:i})}},this.detach=function(){this.document.removeEventListener("change",this.$onChange)},this.attach=function(e){this.document=e||this.document,this.document.on("change",this.$onChange)},this.$clipPositionToDocument=function(e,t){var n={};return e>=this.document.getLength()?(n.row=Math.max(0,this.document.getLength()-1),n.column=this.document.getLine(n.row).length):e<0?(n.row=0,n.column=0):(n.row=e,n.column=Math.min(this.document.getLine(n.row).length,Math.max(0,t))),t<0&&(n.column=0),n}}).call(o.prototype)})),ace.define("ace/document",["require","exports","module","ace/lib/oop","ace/apply_delta","ace/lib/event_emitter","ace/range","ace/anchor"],(function(e,t,n){"use strict";var i=e("./lib/oop"),r=e("./apply_delta").applyDelta,o=e("./lib/event_emitter").EventEmitter,s=e("./range").Range,a=e("./anchor").Anchor,l=function(e){this.$lines=[""],0===e.length?this.$lines=[""]:Array.isArray(e)?this.insertMergedLines({row:0,column:0},e):this.insert({row:0,column:0},e)};(function(){i.implement(this,o),this.setValue=function(e){var t=this.getLength()-1;this.remove(new s(0,0,t,this.getLine(t).length)),this.insert({row:0,column:0},e)},this.getValue=function(){return this.getAllLines().join(this.getNewLineCharacter())},this.createAnchor=function(e,t){return new a(this,e,t)},0==="aaa".split(/a/).length?this.$split=function(e){return e.replace(/\r\n|\r/g,"\n").split("\n")}:this.$split=function(e){return e.split(/\r\n|\r|\n/)},this.$detectNewLine=function(e){var t=e.match(/^.*?(\r\n|\r|\n)/m);this.$autoNewLine=t?t[1]:"\n",this._signal("changeNewLineMode")},this.getNewLineCharacter=function(){switch(this.$newLineMode){case"windows":return"\r\n";case"unix":return"\n";default:return this.$autoNewLine||"\n"}},this.$autoNewLine="",this.$newLineMode="auto",this.setNewLineMode=function(e){this.$newLineMode!==e&&(this.$newLineMode=e,this._signal("changeNewLineMode"))},this.getNewLineMode=function(){return this.$newLineMode},this.isNewLine=function(e){return"\r\n"==e||"\r"==e||"\n"==e},this.getLine=function(e){return this.$lines[e]||""},this.getLines=function(e,t){return this.$lines.slice(e,t+1)},this.getAllLines=function(){return this.getLines(0,this.getLength())},this.getLength=function(){return this.$lines.length},this.getTextRange=function(e){return this.getLinesForRange(e).join(this.getNewLineCharacter())},this.getLinesForRange=function(e){var t;if(e.start.row===e.end.row)t=[this.getLine(e.start.row).substring(e.start.column,e.end.column)];else{(t=this.getLines(e.start.row,e.end.row))[0]=(t[0]||"").substring(e.start.column);var n=t.length-1;e.end.row-e.start.row==n&&(t[n]=t[n].substring(0,e.end.column))}return t},this.insertLines=function(e,t){return console.warn("Use of document.insertLines is deprecated. Use the insertFullLines method instead."),this.insertFullLines(e,t)},this.removeLines=function(e,t){return console.warn("Use of document.removeLines is deprecated. Use the removeFullLines method instead."),this.removeFullLines(e,t)},this.insertNewLine=function(e){return console.warn("Use of document.insertNewLine is deprecated. Use insertMergedLines(position, ['', '']) instead."),this.insertMergedLines(e,["",""])},this.insert=function(e,t){return this.getLength()<=1&&this.$detectNewLine(t),this.insertMergedLines(e,this.$split(t))},this.insertInLine=function(e,t){var n=this.clippedPos(e.row,e.column),i=this.pos(e.row,e.column+t.length);return this.applyDelta({start:n,end:i,action:"insert",lines:[t]},!0),this.clonePos(i)},this.clippedPos=function(e,t){var n=this.getLength();void 0===e?e=n:e<0?e=0:e>=n&&(e=n-1,t=void 0);var i=this.getLine(e);return null==t&&(t=i.length),{row:e,column:t=Math.min(Math.max(t,0),i.length)}},this.clonePos=function(e){return{row:e.row,column:e.column}},this.pos=function(e,t){return{row:e,column:t}},this.$clipPosition=function(e){var t=this.getLength();return e.row>=t?(e.row=Math.max(0,t-1),e.column=this.getLine(t-1).length):(e.row=Math.max(0,e.row),e.column=Math.min(Math.max(e.column,0),this.getLine(e.row).length)),e},this.insertFullLines=function(e,t){var n=0;(e=Math.min(Math.max(e,0),this.getLength()))<this.getLength()?(t=t.concat([""]),n=0):(t=[""].concat(t),e--,n=this.$lines[e].length),this.insertMergedLines({row:e,column:n},t)},this.insertMergedLines=function(e,t){var n=this.clippedPos(e.row,e.column),i={row:n.row+t.length-1,column:(1==t.length?n.column:0)+t[t.length-1].length};return this.applyDelta({start:n,end:i,action:"insert",lines:t}),this.clonePos(i)},this.remove=function(e){var t=this.clippedPos(e.start.row,e.start.column),n=this.clippedPos(e.end.row,e.end.column);return this.applyDelta({start:t,end:n,action:"remove",lines:this.getLinesForRange({start:t,end:n})}),this.clonePos(t)},this.removeInLine=function(e,t,n){var i=this.clippedPos(e,t),r=this.clippedPos(e,n);return this.applyDelta({start:i,end:r,action:"remove",lines:this.getLinesForRange({start:i,end:r})},!0),this.clonePos(i)},this.removeFullLines=function(e,t){e=Math.min(Math.max(0,e),this.getLength()-1);var n=(t=Math.min(Math.max(0,t),this.getLength()-1))==this.getLength()-1&&e>0,i=t<this.getLength()-1,r=n?e-1:e,o=n?this.getLine(r).length:0,a=i?t+1:t,l=i?0:this.getLine(a).length,u=new s(r,o,a,l),c=this.$lines.slice(e,t+1);return this.applyDelta({start:u.start,end:u.end,action:"remove",lines:this.getLinesForRange(u)}),c},this.removeNewLine=function(e){e<this.getLength()-1&&e>=0&&this.applyDelta({start:this.pos(e,this.getLine(e).length),end:this.pos(e+1,0),action:"remove",lines:["",""]})},this.replace=function(e,t){return e instanceof s||(e=s.fromPoints(e.start,e.end)),0===t.length&&e.isEmpty()?e.start:t==this.getTextRange(e)?e.end:(this.remove(e),t?this.insert(e.start,t):e.start)},this.applyDeltas=function(e){for(var t=0;t<e.length;t++)this.applyDelta(e[t])},this.revertDeltas=function(e){for(var t=e.length-1;t>=0;t--)this.revertDelta(e[t])},this.applyDelta=function(e,t){var n="insert"==e.action;(n?e.lines.length<=1&&!e.lines[0]:!s.comparePoints(e.start,e.end))||(n&&e.lines.length>2e4&&this.$splitAndapplyLargeDelta(e,2e4),r(this.$lines,e,t),this._signal("change",e))},this.$splitAndapplyLargeDelta=function(e,t){for(var n=e.lines,i=n.length,r=e.start.row,o=e.start.column,s=0,a=0;;){s=a,a+=t-1;var l=n.slice(s,a);if(a>i){e.lines=l,e.start.row=r+s,e.start.column=o;break}l.push(""),this.applyDelta({start:this.pos(r+s,o),end:this.pos(r+a,o=0),action:e.action,lines:l},!0)}},this.revertDelta=function(e){this.applyDelta({start:this.clonePos(e.start),end:this.clonePos(e.end),action:"insert"==e.action?"remove":"insert",lines:e.lines.slice()})},this.indexToPosition=function(e,t){for(var n=this.$lines||this.getAllLines(),i=this.getNewLineCharacter().length,r=t||0,o=n.length;r<o;r++)if((e-=n[r].length+i)<0)return{row:r,column:e+n[r].length+i};return{row:o-1,column:n[o-1].length}},this.positionToIndex=function(e,t){for(var n=this.$lines||this.getAllLines(),i=this.getNewLineCharacter().length,r=0,o=Math.min(e.row,n.length),s=t||0;s<o;++s)r+=n[s].length+i;return r+e.column}}).call(l.prototype),t.Document=l})),ace.define("ace/background_tokenizer",["require","exports","module","ace/lib/oop","ace/lib/event_emitter"],(function(e,t,n){"use strict";var i=e("./lib/oop"),r=e("./lib/event_emitter").EventEmitter,o=function(e,t){this.running=!1,this.lines=[],this.states=[],this.currentLine=0,this.tokenizer=e;var n=this;this.$worker=function(){if(n.running){for(var e=new Date,t=n.currentLine,i=-1,r=n.doc,o=t;n.lines[t];)t++;var s=r.getLength(),a=0;for(n.running=!1;t<s;){n.$tokenizeRow(t),i=t;do{t++}while(n.lines[t]);if(++a%5==0&&new Date-e>20){n.running=setTimeout(n.$worker,20);break}}n.currentLine=t,-1==i&&(i=t),o<=i&&n.fireUpdateEvent(o,i)}}};(function(){i.implement(this,r),this.setTokenizer=function(e){this.tokenizer=e,this.lines=[],this.states=[],this.start(0)},this.setDocument=function(e){this.doc=e,this.lines=[],this.states=[],this.stop()},this.fireUpdateEvent=function(e,t){var n={first:e,last:t};this._signal("update",{data:n})},this.start=function(e){this.currentLine=Math.min(e||0,this.currentLine,this.doc.getLength()),this.lines.splice(this.currentLine,this.lines.length),this.states.splice(this.currentLine,this.states.length),this.stop(),this.running=setTimeout(this.$worker,700)},this.scheduleStart=function(){this.running||(this.running=setTimeout(this.$worker,700))},this.$updateOnChange=function(e){var t=e.start.row,n=e.end.row-t;if(0===n)this.lines[t]=null;else if("remove"==e.action)this.lines.splice(t,n+1,null),this.states.splice(t,n+1,null);else{var i=Array(n+1);i.unshift(t,1),this.lines.splice.apply(this.lines,i),this.states.splice.apply(this.states,i)}this.currentLine=Math.min(t,this.currentLine,this.doc.getLength()),this.stop()},this.stop=function(){this.running&&clearTimeout(this.running),this.running=!1},this.getTokens=function(e){return this.lines[e]||this.$tokenizeRow(e)},this.getState=function(e){return this.currentLine==e&&this.$tokenizeRow(e),this.states[e]||"start"},this.$tokenizeRow=function(e){var t=this.doc.getLine(e),n=this.states[e-1],i=this.tokenizer.getLineTokens(t,n,e);return this.states[e]+""!=i.state+""?(this.states[e]=i.state,this.lines[e+1]=null,this.currentLine>e+1&&(this.currentLine=e+1)):this.currentLine==e&&(this.currentLine=e+1),this.lines[e]=i.tokens}}).call(o.prototype),t.BackgroundTokenizer=o})),ace.define("ace/search_highlight",["require","exports","module","ace/lib/lang","ace/lib/oop","ace/range"],(function(e,t,n){"use strict";var i=e("./lib/lang"),r=(e("./lib/oop"),e("./range").Range),o=function(e,t,n){this.setRegexp(e),this.clazz=t,this.type=n||"text"};(function(){this.MAX_RANGES=500,this.setRegexp=function(e){this.regExp+""!=e+""&&(this.regExp=e,this.cache=[])},this.update=function(e,t,n,o){if(this.regExp)for(var s=o.firstRow,a=o.lastRow,l=s;l<=a;l++){var u=this.cache[l];null==u&&((u=i.getMatchOffsets(n.getLine(l),this.regExp)).length>this.MAX_RANGES&&(u=u.slice(0,this.MAX_RANGES)),u=u.map((function(e){return new r(l,e.offset,l,e.offset+e.length)})),this.cache[l]=u.length?u:"");for(var c=u.length;c--;)t.drawSingleLineMarker(e,u[c].toScreenRange(n),this.clazz,o)}}}).call(o.prototype),t.SearchHighlight=o})),ace.define("ace/edit_session/fold_line",["require","exports","module","ace/range"],(function(e,t,n){"use strict";var i=e("../range").Range;function r(e,t){this.foldData=e,Array.isArray(t)?this.folds=t:t=this.folds=[t];var n=t[t.length-1];this.range=new i(t[0].start.row,t[0].start.column,n.end.row,n.end.column),this.start=this.range.start,this.end=this.range.end,this.folds.forEach((function(e){e.setFoldLine(this)}),this)}(function(){this.shiftRow=function(e){this.start.row+=e,this.end.row+=e,this.folds.forEach((function(t){t.start.row+=e,t.end.row+=e}))},this.addFold=function(e){if(e.sameRow){if(e.start.row<this.startRow||e.endRow>this.endRow)throw new Error("Can't add a fold to this FoldLine as it has no connection");this.folds.push(e),this.folds.sort((function(e,t){return-e.range.compareEnd(t.start.row,t.start.column)})),this.range.compareEnd(e.start.row,e.start.column)>0?(this.end.row=e.end.row,this.end.column=e.end.column):this.range.compareStart(e.end.row,e.end.column)<0&&(this.start.row=e.start.row,this.start.column=e.start.column)}else if(e.start.row==this.end.row)this.folds.push(e),this.end.row=e.end.row,this.end.column=e.end.column;else{if(e.end.row!=this.start.row)throw new Error("Trying to add fold to FoldRow that doesn't have a matching row");this.folds.unshift(e),this.start.row=e.start.row,this.start.column=e.start.column}e.foldLine=this},this.containsRow=function(e){return e>=this.start.row&&e<=this.end.row},this.walk=function(e,t,n){var i,r,o=0,s=this.folds,a=!0;null==t&&(t=this.end.row,n=this.end.column);for(var l=0;l<s.length;l++){if(-1==(r=(i=s[l]).range.compareStart(t,n)))return void e(null,t,n,o,a);if(!e(null,i.start.row,i.start.column,o,a)&&e(i.placeholder,i.start.row,i.start.column,o)||0===r)return;a=!i.sameRow,o=i.end.column}e(null,t,n,o,a)},this.getNextFoldTo=function(e,t){for(var n,i,r=0;r<this.folds.length;r++){if(-1==(i=(n=this.folds[r]).range.compareEnd(e,t)))return{fold:n,kind:"after"};if(0===i)return{fold:n,kind:"inside"}}return null},this.addRemoveChars=function(e,t,n){var i,r,o=this.getNextFoldTo(e,t);if(o)if(i=o.fold,"inside"==o.kind&&i.start.column!=t&&i.start.row!=e)window.console&&window.console.log(e,t,i);else if(i.start.row==e){var s=(r=this.folds).indexOf(i);for(0===s&&(this.start.column+=n);s<r.length;s++){if((i=r[s]).start.column+=n,!i.sameRow)return;i.end.column+=n}this.end.column+=n}},this.split=function(e,t){var n=this.getNextFoldTo(e,t);if(!n||"inside"==n.kind)return null;var i=n.fold,o=this.folds,s=this.foldData,a=o.indexOf(i),l=o[a-1];this.end.row=l.end.row,this.end.column=l.end.column;var u=new r(s,o=o.splice(a,o.length-a));return s.splice(s.indexOf(this)+1,0,u),u},this.merge=function(e){for(var t=e.folds,n=0;n<t.length;n++)this.addFold(t[n]);var i=this.foldData;i.splice(i.indexOf(e),1)},this.toString=function(){var e=[this.range.toString()+": ["];return this.folds.forEach((function(t){e.push("  "+t.toString())})),e.push("]"),e.join("\n")},this.idxToPosition=function(e){for(var t=0,n=0;n<this.folds.length;n++){var i=this.folds[n];if((e-=i.start.column-t)<0)return{row:i.start.row,column:i.start.column+e};if((e-=i.placeholder.length)<0)return i.start;t=i.end.column}return{row:this.end.row,column:this.end.column+e}}}).call(r.prototype),t.FoldLine=r})),ace.define("ace/range_list",["require","exports","module","ace/range"],(function(e,t,n){"use strict";var i=e("./range").Range.comparePoints,r=function(){this.ranges=[]};(function(){this.comparePoints=i,this.pointIndex=function(e,t,n){for(var r=this.ranges,o=n||0;o<r.length;o++){var s=r[o],a=i(e,s.end);if(!(a>0)){var l=i(e,s.start);return 0===a?t&&0!==l?-o-2:o:l>0||0===l&&!t?o:-o-1}}return-o-1},this.add=function(e){var t=!e.isEmpty(),n=this.pointIndex(e.start,t);n<0&&(n=-n-1);var i=this.pointIndex(e.end,t,n);return i<0?i=-i-1:i++,this.ranges.splice(n,i-n,e)},this.addList=function(e){for(var t=[],n=e.length;n--;)t.push.apply(t,this.add(e[n]));return t},this.substractPoint=function(e){var t=this.pointIndex(e);if(t>=0)return this.ranges.splice(t,1)},this.merge=function(){for(var e,t=[],n=this.ranges,r=(n=n.sort((function(e,t){return i(e.start,t.start)})))[0],o=1;o<n.length;o++){e=r,r=n[o];var s=i(e.end,r.start);s<0||(0!=s||e.isEmpty()||r.isEmpty())&&(i(e.end,r.end)<0&&(e.end.row=r.end.row,e.end.column=r.end.column),n.splice(o,1),t.push(r),r=e,o--)}return this.ranges=n,t},this.contains=function(e,t){return this.pointIndex({row:e,column:t})>=0},this.containsPoint=function(e){return this.pointIndex(e)>=0},this.rangeAtPoint=function(e){var t=this.pointIndex(e);if(t>=0)return this.ranges[t]},this.clipRows=function(e,t){var n=this.ranges;if(n[0].start.row>t||n[n.length-1].start.row<e)return[];var i=this.pointIndex({row:e,column:0});i<0&&(i=-i-1);var r=this.pointIndex({row:t,column:0},i);r<0&&(r=-r-1);for(var o=[],s=i;s<r;s++)o.push(n[s]);return o},this.removeAll=function(){return this.ranges.splice(0,this.ranges.length)},this.attach=function(e){this.session&&this.detach(),this.session=e,this.onChange=this.$onChange.bind(this),this.session.on("change",this.onChange)},this.detach=function(){this.session&&(this.session.removeListener("change",this.onChange),this.session=null)},this.$onChange=function(e){if("insert"==e.action)var t=e.start,n=e.end;else n=e.start,t=e.end;for(var i=t.row,r=n.row-i,o=-t.column+n.column,s=this.ranges,a=0,l=s.length;a<l;a++){if(!((u=s[a]).end.row<i)){if(u.start.row>i)break;if(u.start.row==i&&u.start.column>=t.column&&(u.start.column==t.column&&this.$insertRight||(u.start.column+=o,u.start.row+=r)),u.end.row==i&&u.end.column>=t.column){if(u.end.column==t.column&&this.$insertRight)continue;u.end.column==t.column&&o>0&&a<l-1&&u.end.column>u.start.column&&u.end.column==s[a+1].start.column&&(u.end.column-=o),u.end.column+=o,u.end.row+=r}}}if(0!=r&&a<l)for(;a<l;a++){var u;(u=s[a]).start.row+=r,u.end.row+=r}}}).call(r.prototype),t.RangeList=r})),ace.define("ace/edit_session/fold",["require","exports","module","ace/range","ace/range_list","ace/lib/oop"],(function(e,t,n){"use strict";e("../range").Range;var i=e("../range_list").RangeList,r=e("../lib/oop"),o=t.Fold=function(e,t){this.foldLine=null,this.placeholder=t,this.range=e,this.start=e.start,this.end=e.end,this.sameRow=e.start.row==e.end.row,this.subFolds=this.ranges=[]};function s(e,t){e.row-=t.row,0==e.row&&(e.column-=t.column)}function a(e,t){0==e.row&&(e.column+=t.column),e.row+=t.row}r.inherits(o,i),function(){this.toString=function(){return'"'+this.placeholder+'" '+this.range.toString()},this.setFoldLine=function(e){this.foldLine=e,this.subFolds.forEach((function(t){t.setFoldLine(e)}))},this.clone=function(){var e=this.range.clone(),t=new o(e,this.placeholder);return this.subFolds.forEach((function(e){t.subFolds.push(e.clone())})),t.collapseChildren=this.collapseChildren,t},this.addSubFold=function(e){if(!this.range.isEqual(e)){if(!this.range.containsRange(e))throw new Error("A fold can't intersect already existing fold"+e.range+this.range);var t,n;t=e,n=this.start,s(t.start,n),s(t.end,n);for(var i=e.start.row,r=e.start.column,o=0,a=-1;o<this.subFolds.length&&1==(a=this.subFolds[o].range.compare(i,r));o++);var l=this.subFolds[o];if(0==a)return l.addSubFold(e);i=e.range.end.row,r=e.range.end.column;var u=o;for(a=-1;u<this.subFolds.length&&1==(a=this.subFolds[u].range.compare(i,r));u++);this.subFolds[u];if(0==a)throw new Error("A fold can't intersect already existing fold"+e.range+this.range);this.subFolds.splice(o,u-o,e);return e.setFoldLine(this.foldLine),e}},this.restoreRange=function(e){return function(e,t){a(e.start,t),a(e.end,t)}(e,this.start)}}.call(o.prototype)})),ace.define("ace/edit_session/folding",["require","exports","module","ace/range","ace/edit_session/fold_line","ace/edit_session/fold","ace/token_iterator"],(function(e,t,n){"use strict";var i=e("../range").Range,r=e("./fold_line").FoldLine,o=e("./fold").Fold,s=e("../token_iterator").TokenIterator;t.Folding=function(){this.getFoldAt=function(e,t,n){var i=this.getFoldLine(e);if(!i)return null;for(var r=i.folds,o=0;o<r.length;o++){var s=r[o];if(s.range.contains(e,t)){if(1==n&&s.range.isEnd(e,t))continue;if(-1==n&&s.range.isStart(e,t))continue;return s}}},this.getFoldsInRange=function(e){var t=e.start,n=e.end,i=this.$foldData,r=[];t.column+=1,n.column-=1;for(var o=0;o<i.length;o++){var s=i[o].range.compareRange(e);if(2!=s){if(-2==s)break;for(var a=i[o].folds,l=0;l<a.length;l++){var u=a[l];if(-2==(s=u.range.compareRange(e)))break;if(2!=s){if(42==s)break;r.push(u)}}}}return t.column-=1,n.column+=1,r},this.getFoldsInRangeList=function(e){if(Array.isArray(e)){var t=[];e.forEach((function(e){t=t.concat(this.getFoldsInRange(e))}),this)}else t=this.getFoldsInRange(e);return t},this.getAllFolds=function(){for(var e=[],t=this.$foldData,n=0;n<t.length;n++)for(var i=0;i<t[n].folds.length;i++)e.push(t[n].folds[i]);return e},this.getFoldStringAt=function(e,t,n,i){if(!(i=i||this.getFoldLine(e)))return null;for(var r,o,s={end:{column:0}},a=0;a<i.folds.length;a++){var l=(o=i.folds[a]).range.compareEnd(e,t);if(-1==l){r=this.getLine(o.start.row).substring(s.end.column,o.start.column);break}if(0===l)return null;s=o}return r||(r=this.getLine(o.start.row).substring(s.end.column)),-1==n?r.substring(0,t-s.end.column):1==n?r.substring(t-s.end.column):r},this.getFoldLine=function(e,t){var n=this.$foldData,i=0;for(t&&(i=n.indexOf(t)),-1==i&&(i=0);i<n.length;i++){var r=n[i];if(r.start.row<=e&&r.end.row>=e)return r;if(r.end.row>e)return null}return null},this.getNextFoldLine=function(e,t){var n=this.$foldData,i=0;for(t&&(i=n.indexOf(t)),-1==i&&(i=0);i<n.length;i++){var r=n[i];if(r.end.row>=e)return r}return null},this.getFoldedRowCount=function(e,t){for(var n=this.$foldData,i=t-e+1,r=0;r<n.length;r++){var o=n[r],s=o.end.row,a=o.start.row;if(s>=t){a<t&&(a>=e?i-=t-a:i=0);break}s>=e&&(i-=a>=e?s-a:s-e+1)}return i},this.$addFoldLine=function(e){return this.$foldData.push(e),this.$foldData.sort((function(e,t){return e.start.row-t.start.row})),e},this.addFold=function(e,t){var n,i=this.$foldData,s=!1;e instanceof o?n=e:(n=new o(t,e)).collapseChildren=t.collapseChildren,this.$clipRangeToDocument(n.range);var a=n.start.row,l=n.start.column,u=n.end.row,c=n.end.column;if(!(a<u||a==u&&l<=c-2))throw new Error("The range has to be at least 2 characters width");var h=this.getFoldAt(a,l,1),f=this.getFoldAt(u,c,-1);if(h&&f==h)return h.addSubFold(n);h&&!h.range.isStart(a,l)&&this.removeFold(h),f&&!f.range.isEnd(u,c)&&this.removeFold(f);var d=this.getFoldsInRange(n.range);d.length>0&&(this.removeFolds(d),d.forEach((function(e){n.addSubFold(e)})));for(var p=0;p<i.length;p++){var g=i[p];if(u==g.start.row){g.addFold(n),s=!0;break}if(a==g.end.row){if(g.addFold(n),s=!0,!n.sameRow){var m=i[p+1];if(m&&m.start.row==u){g.merge(m);break}}break}if(u<=g.start.row)break}return s||(g=this.$addFoldLine(new r(this.$foldData,n))),this.$useWrapMode?this.$updateWrapData(g.start.row,g.start.row):this.$updateRowLengthCache(g.start.row,g.start.row),this.$modified=!0,this._signal("changeFold",{data:n,action:"add"}),n},this.addFolds=function(e){e.forEach((function(e){this.addFold(e)}),this)},this.removeFold=function(e){var t=e.foldLine,n=t.start.row,i=t.end.row,r=this.$foldData,o=t.folds;if(1==o.length)r.splice(r.indexOf(t),1);else if(t.range.isEnd(e.end.row,e.end.column))o.pop(),t.end.row=o[o.length-1].end.row,t.end.column=o[o.length-1].end.column;else if(t.range.isStart(e.start.row,e.start.column))o.shift(),t.start.row=o[0].start.row,t.start.column=o[0].start.column;else if(e.sameRow)o.splice(o.indexOf(e),1);else{var s=t.split(e.start.row,e.start.column);(o=s.folds).shift(),s.start.row=o[0].start.row,s.start.column=o[0].start.column}this.$updating||(this.$useWrapMode?this.$updateWrapData(n,i):this.$updateRowLengthCache(n,i)),this.$modified=!0,this._signal("changeFold",{data:e,action:"remove"})},this.removeFolds=function(e){for(var t=[],n=0;n<e.length;n++)t.push(e[n]);t.forEach((function(e){this.removeFold(e)}),this),this.$modified=!0},this.expandFold=function(e){this.removeFold(e),e.subFolds.forEach((function(t){e.restoreRange(t),this.addFold(t)}),this),e.collapseChildren>0&&this.foldAll(e.start.row+1,e.end.row,e.collapseChildren-1),e.subFolds=[]},this.expandFolds=function(e){e.forEach((function(e){this.expandFold(e)}),this)},this.unfold=function(e,t){var n,r;if(null==e?(n=new i(0,0,this.getLength(),0),t=!0):n="number"==typeof e?new i(e,0,e,this.getLine(e).length):"row"in e?i.fromPoints(e,e):e,r=this.getFoldsInRangeList(n),t)this.removeFolds(r);else for(var o=r;o.length;)this.expandFolds(o),o=this.getFoldsInRangeList(n);if(r.length)return r},this.isRowFolded=function(e,t){return!!this.getFoldLine(e,t)},this.getRowFoldEnd=function(e,t){var n=this.getFoldLine(e,t);return n?n.end.row:e},this.getRowFoldStart=function(e,t){var n=this.getFoldLine(e,t);return n?n.start.row:e},this.getFoldDisplayLine=function(e,t,n,i,r){null==i&&(i=e.start.row),null==r&&(r=0),null==t&&(t=e.end.row),null==n&&(n=this.getLine(t).length);var o=this.doc,s="";return e.walk((function(e,t,n,a){if(!(t<i)){if(t==i){if(n<r)return;a=Math.max(r,a)}s+=null!=e?e:o.getLine(t).substring(a,n)}}),t,n),s},this.getDisplayLine=function(e,t,n,i){var r,o=this.getFoldLine(e);return o?this.getFoldDisplayLine(o,e,t,n,i):(r=this.doc.getLine(e)).substring(i||0,t||r.length)},this.$cloneFoldData=function(){var e=[];return e=this.$foldData.map((function(t){var n=t.folds.map((function(e){return e.clone()}));return new r(e,n)}))},this.toggleFold=function(e){var t,n,i=this.selection.getRange();if(i.isEmpty()){var r=i.start;if(t=this.getFoldAt(r.row,r.column))return void this.expandFold(t);(n=this.findMatchingBracket(r))?1==i.comparePoint(n)?i.end=n:(i.start=n,i.start.column++,i.end.column--):(n=this.findMatchingBracket({row:r.row,column:r.column+1}))?(1==i.comparePoint(n)?i.end=n:i.start=n,i.start.column++):i=this.getCommentFoldRange(r.row,r.column)||i}else{var o=this.getFoldsInRange(i);if(e&&o.length)return void this.expandFolds(o);1==o.length&&(t=o[0])}if(t||(t=this.getFoldAt(i.start.row,i.start.column)),t&&t.range.toString()==i.toString())this.expandFold(t);else{var s="...";if(!i.isMultiLine()){if((s=this.getTextRange(i)).length<4)return;s=s.trim().substring(0,2)+".."}this.addFold(s,i)}},this.getCommentFoldRange=function(e,t,n){var r=new s(this,e,t),o=r.getCurrentToken(),a=o.type;if(o&&/^comment|string/.test(a)){"comment"==(a=a.match(/comment|string/)[0])&&(a+="|doc-start");var l=new RegExp(a),u=new i;if(1!=n){do{o=r.stepBackward()}while(o&&l.test(o.type));r.stepForward()}if(u.start.row=r.getCurrentTokenRow(),u.start.column=r.getCurrentTokenColumn()+2,r=new s(this,e,t),-1!=n){var c=-1;do{if(o=r.stepForward(),-1==c){var h=this.getState(r.$row);l.test(h)||(c=r.$row)}else if(r.$row>c)break}while(o&&l.test(o.type));o=r.stepBackward()}else o=r.getCurrentToken();return u.end.row=r.getCurrentTokenRow(),u.end.column=r.getCurrentTokenColumn()+o.value.length-2,u}},this.foldAll=function(e,t,n){null==n&&(n=1e5);var i=this.foldWidgets;if(i){t=t||this.getLength();for(var r=e=e||0;r<t;r++)if(null==i[r]&&(i[r]=this.getFoldWidget(r)),"start"==i[r]){var o=this.getFoldWidgetRange(r);if(o&&o.isMultiLine()&&o.end.row<=t&&o.start.row>=e){r=o.end.row;try{var s=this.addFold("...",o);s&&(s.collapseChildren=n)}catch(e){}}}}},this.$foldStyles={manual:1,markbegin:1,markbeginend:1},this.$foldStyle="markbegin",this.setFoldStyle=function(e){if(!this.$foldStyles[e])throw new Error("invalid fold style: "+e+"["+Object.keys(this.$foldStyles).join(", ")+"]");if(this.$foldStyle!=e){this.$foldStyle=e,"manual"==e&&this.unfold();var t=this.$foldMode;this.$setFolding(null),this.$setFolding(t)}},this.$setFolding=function(e){this.$foldMode!=e&&(this.$foldMode=e,this.off("change",this.$updateFoldWidgets),this.off("tokenizerUpdate",this.$tokenizerUpdateFoldWidgets),this._signal("changeAnnotation"),e&&"manual"!=this.$foldStyle?(this.foldWidgets=[],this.getFoldWidget=e.getFoldWidget.bind(e,this,this.$foldStyle),this.getFoldWidgetRange=e.getFoldWidgetRange.bind(e,this,this.$foldStyle),this.$updateFoldWidgets=this.updateFoldWidgets.bind(this),this.$tokenizerUpdateFoldWidgets=this.tokenizerUpdateFoldWidgets.bind(this),this.on("change",this.$updateFoldWidgets),this.on("tokenizerUpdate",this.$tokenizerUpdateFoldWidgets)):this.foldWidgets=null)},this.getParentFoldRangeData=function(e,t){var n=this.foldWidgets;if(!n||t&&n[e])return{};for(var i,r=e-1;r>=0;){var o=n[r];if(null==o&&(o=n[r]=this.getFoldWidget(r)),"start"==o){var s=this.getFoldWidgetRange(r);if(i||(i=s),s&&s.end.row>=e)break}r--}return{range:-1!==r&&s,firstRange:i}},this.onFoldWidgetClick=function(e,t){var n={children:(t=t.domEvent).shiftKey,all:t.ctrlKey||t.metaKey,siblings:t.altKey};if(!this.$toggleFoldWidget(e,n)){var i=t.target||t.srcElement;i&&/ace_fold-widget/.test(i.className)&&(i.className+=" ace_invalid")}},this.$toggleFoldWidget=function(e,t){if(this.getFoldWidget){var n=this.getFoldWidget(e),i=this.getLine(e),r="end"===n?-1:1,o=this.getFoldAt(e,-1===r?0:i.length,r);if(o)return t.children||t.all?this.removeFold(o):this.expandFold(o),o;var s=this.getFoldWidgetRange(e,!0);if(s&&!s.isMultiLine()&&(o=this.getFoldAt(s.start.row,s.start.column,1))&&s.isEqual(o.range))return this.removeFold(o),o;if(t.siblings){var a=this.getParentFoldRangeData(e);if(a.range)var l=a.range.start.row+1,u=a.range.end.row;this.foldAll(l,u,t.all?1e4:0)}else t.children?(u=s?s.end.row:this.getLength(),this.foldAll(e+1,u,t.all?1e4:0)):s&&(t.all&&(s.collapseChildren=1e4),this.addFold("...",s));return s}},this.toggleFoldWidget=function(e){var t=this.selection.getCursor().row;t=this.getRowFoldStart(t);var n=this.$toggleFoldWidget(t,{});if(!n){var i=this.getParentFoldRangeData(t,!0);if(n=i.range||i.firstRange){t=n.start.row;var r=this.getFoldAt(t,this.getLine(t).length,1);r?this.removeFold(r):this.addFold("...",n)}}},this.updateFoldWidgets=function(e){var t=e.start.row,n=e.end.row-t;if(0===n)this.foldWidgets[t]=null;else if("remove"==e.action)this.foldWidgets.splice(t,n+1,null);else{var i=Array(n+1);i.unshift(t,1),this.foldWidgets.splice.apply(this.foldWidgets,i)}},this.tokenizerUpdateFoldWidgets=function(e){var t=e.data;t.first!=t.last&&this.foldWidgets.length>t.first&&this.foldWidgets.splice(t.first,this.foldWidgets.length)}}})),ace.define("ace/edit_session/bracket_match",["require","exports","module","ace/token_iterator","ace/range"],(function(e,t,n){"use strict";var i=e("../token_iterator").TokenIterator,r=e("../range").Range;t.BracketMatch=function(){this.findMatchingBracket=function(e,t){if(0==e.column)return null;var n=t||this.getLine(e.row).charAt(e.column-1);if(""==n)return null;var i=n.match(/([\(\[\{])|([\)\]\}])/);return i?i[1]?this.$findClosingBracket(i[1],e):this.$findOpeningBracket(i[2],e):null},this.getBracketRange=function(e){var t,n=this.getLine(e.row),i=!0,o=n.charAt(e.column-1),s=o&&o.match(/([\(\[\{])|([\)\]\}])/);if(s||(o=n.charAt(e.column),e={row:e.row,column:e.column+1},s=o&&o.match(/([\(\[\{])|([\)\]\}])/),i=!1),!s)return null;if(s[1]){if(!(a=this.$findClosingBracket(s[1],e)))return null;t=r.fromPoints(e,a),i||(t.end.column++,t.start.column--),t.cursor=t.end}else{var a;if(!(a=this.$findOpeningBracket(s[2],e)))return null;t=r.fromPoints(a,e),i||(t.start.column++,t.end.column--),t.cursor=t.start}return t},this.$brackets={")":"(","(":")","]":"[","[":"]","{":"}","}":"{"},this.$findOpeningBracket=function(e,t,n){var r=this.$brackets[e],o=1,s=new i(this,t.row,t.column),a=s.getCurrentToken();if(a||(a=s.stepForward()),a){n||(n=new RegExp("(\\.?"+a.type.replace(".","\\.").replace("rparen",".paren").replace(/\b(?:end)\b/,"(?:start|begin|end)")+")+"));for(var l=t.column-s.getCurrentTokenColumn()-2,u=a.value;;){for(;l>=0;){var c=u.charAt(l);if(c==r){if(0==(o-=1))return{row:s.getCurrentTokenRow(),column:l+s.getCurrentTokenColumn()}}else c==e&&(o+=1);l-=1}do{a=s.stepBackward()}while(a&&!n.test(a.type));if(null==a)break;l=(u=a.value).length-1}return null}},this.$findClosingBracket=function(e,t,n){var r=this.$brackets[e],o=1,s=new i(this,t.row,t.column),a=s.getCurrentToken();if(a||(a=s.stepForward()),a){n||(n=new RegExp("(\\.?"+a.type.replace(".","\\.").replace("lparen",".paren").replace(/\b(?:start|begin)\b/,"(?:start|begin|end)")+")+"));for(var l=t.column-s.getCurrentTokenColumn();;){for(var u=a.value,c=u.length;l<c;){var h=u.charAt(l);if(h==r){if(0==(o-=1))return{row:s.getCurrentTokenRow(),column:l+s.getCurrentTokenColumn()}}else h==e&&(o+=1);l+=1}do{a=s.stepForward()}while(a&&!n.test(a.type));if(null==a)break;l=0}return null}}}})),ace.define("ace/edit_session",["require","exports","module","ace/lib/oop","ace/lib/lang","ace/bidihandler","ace/config","ace/lib/event_emitter","ace/selection","ace/mode/text","ace/range","ace/document","ace/background_tokenizer","ace/search_highlight","ace/edit_session/folding","ace/edit_session/bracket_match"],(function(e,t,n){"use strict";var i=e("./lib/oop"),r=e("./lib/lang"),o=e("./bidihandler").BidiHandler,s=e("./config"),a=e("./lib/event_emitter").EventEmitter,l=e("./selection").Selection,u=e("./mode/text").Mode,c=e("./range").Range,h=e("./document").Document,f=e("./background_tokenizer").BackgroundTokenizer,d=e("./search_highlight").SearchHighlight,p=function(e,t){this.$breakpoints=[],this.$decorations=[],this.$frontMarkers={},this.$backMarkers={},this.$markerId=1,this.$undoSelect=!0,this.$foldData=[],this.id="session"+ ++p.$uid,this.$foldData.toString=function(){return this.join("\n")},this.on("changeFold",this.onChangeFold.bind(this)),this.$onChange=this.onChange.bind(this),"object"==typeof e&&e.getLine||(e=new h(e)),this.$bidiHandler=new o(this),this.setDocument(e),this.selection=new l(this),s.resetOptions(this),this.setMode(t),s._signal("session",this)};p.$uid=0,function(){i.implement(this,a),this.setDocument=function(e){this.doc&&this.doc.removeListener("change",this.$onChange),this.doc=e,e.on("change",this.$onChange),this.bgTokenizer&&this.bgTokenizer.setDocument(this.getDocument()),this.resetCaches()},this.getDocument=function(){return this.doc},this.$resetRowCache=function(e){if(!e)return this.$docRowCache=[],void(this.$screenRowCache=[]);var t=this.$docRowCache.length,n=this.$getRowCacheIndex(this.$docRowCache,e)+1;t>n&&(this.$docRowCache.splice(n,t),this.$screenRowCache.splice(n,t))},this.$getRowCacheIndex=function(e,t){for(var n=0,i=e.length-1;n<=i;){var r=n+i>>1,o=e[r];if(t>o)n=r+1;else{if(!(t<o))return r;i=r-1}}return n-1},this.resetCaches=function(){this.$modified=!0,this.$wrapData=[],this.$rowLengthCache=[],this.$resetRowCache(0),this.bgTokenizer&&this.bgTokenizer.start(0)},this.onChangeFold=function(e){var t=e.data;this.$resetRowCache(t.start.row)},this.onChange=function(e){this.$modified=!0,this.$bidiHandler.onChange(e),this.$resetRowCache(e.start.row);var t=this.$updateInternalDataOnChange(e);this.$fromUndo||!this.$undoManager||e.ignore||(this.$deltasDoc.push(e),t&&0!=t.length&&this.$deltasFold.push({action:"removeFolds",folds:t}),this.$informUndoManager.schedule()),this.bgTokenizer&&this.bgTokenizer.$updateOnChange(e),this._signal("change",e)},this.setValue=function(e){this.doc.setValue(e),this.selection.moveTo(0,0),this.$resetRowCache(0),this.$deltas=[],this.$deltasDoc=[],this.$deltasFold=[],this.setUndoManager(this.$undoManager),this.getUndoManager().reset()},this.getValue=this.toString=function(){return this.doc.getValue()},this.getSelection=function(){return this.selection},this.getState=function(e){return this.bgTokenizer.getState(e)},this.getTokens=function(e){return this.bgTokenizer.getTokens(e)},this.getTokenAt=function(e,t){var n,i=this.bgTokenizer.getTokens(e),r=0;if(null==t){var o=i.length-1;r=this.getLine(e).length}else for(o=0;o<i.length&&!((r+=i[o].value.length)>=t);o++);return(n=i[o])?(n.index=o,n.start=r-n.value.length,n):null},this.setUndoManager=function(e){if(this.$undoManager=e,this.$deltas=[],this.$deltasDoc=[],this.$deltasFold=[],this.$informUndoManager&&this.$informUndoManager.cancel(),e){var t=this;this.$syncInformUndoManager=function(){t.$informUndoManager.cancel(),t.$deltasFold.length&&(t.$deltas.push({group:"fold",deltas:t.$deltasFold}),t.$deltasFold=[]),t.$deltasDoc.length&&(t.$deltas.push({group:"doc",deltas:t.$deltasDoc}),t.$deltasDoc=[]),t.$deltas.length>0&&e.execute({action:"aceupdate",args:[t.$deltas,t],merge:t.mergeUndoDeltas}),t.mergeUndoDeltas=!1,t.$deltas=[]},this.$informUndoManager=r.delayedCall(this.$syncInformUndoManager)}},this.markUndoGroup=function(){this.$syncInformUndoManager&&this.$syncInformUndoManager()},this.$defaultUndoManager={undo:function(){},redo:function(){},reset:function(){}},this.getUndoManager=function(){return this.$undoManager||this.$defaultUndoManager},this.getTabString=function(){return this.getUseSoftTabs()?r.stringRepeat(" ",this.getTabSize()):"\t"},this.setUseSoftTabs=function(e){this.setOption("useSoftTabs",e)},this.getUseSoftTabs=function(){return this.$useSoftTabs&&!this.$mode.$indentWithTabs},this.setTabSize=function(e){this.setOption("tabSize",e)},this.getTabSize=function(){return this.$tabSize},this.isTabStop=function(e){return this.$useSoftTabs&&e.column%this.$tabSize==0},this.setNavigateWithinSoftTabs=function(e){this.setOption("navigateWithinSoftTabs",e)},this.getNavigateWithinSoftTabs=function(){return this.$navigateWithinSoftTabs},this.$overwrite=!1,this.setOverwrite=function(e){this.setOption("overwrite",e)},this.getOverwrite=function(){return this.$overwrite},this.toggleOverwrite=function(){this.setOverwrite(!this.$overwrite)},this.addGutterDecoration=function(e,t){this.$decorations[e]||(this.$decorations[e]=""),this.$decorations[e]+=" "+t,this._signal("changeBreakpoint",{})},this.removeGutterDecoration=function(e,t){this.$decorations[e]=(this.$decorations[e]||"").replace(" "+t,""),this._signal("changeBreakpoint",{})},this.getBreakpoints=function(){return this.$breakpoints},this.setBreakpoints=function(e){this.$breakpoints=[];for(var t=0;t<e.length;t++)this.$breakpoints[e[t]]="ace_breakpoint";this._signal("changeBreakpoint",{})},this.clearBreakpoints=function(){this.$breakpoints=[],this._signal("changeBreakpoint",{})},this.setBreakpoint=function(e,t){void 0===t&&(t="ace_breakpoint"),t?this.$breakpoints[e]=t:delete this.$breakpoints[e],this._signal("changeBreakpoint",{})},this.clearBreakpoint=function(e){delete this.$breakpoints[e],this._signal("changeBreakpoint",{})},this.addMarker=function(e,t,n,i){var r=this.$markerId++,o={range:e,type:n||"line",renderer:"function"==typeof n?n:null,clazz:t,inFront:!!i,id:r};return i?(this.$frontMarkers[r]=o,this._signal("changeFrontMarker")):(this.$backMarkers[r]=o,this._signal("changeBackMarker")),r},this.addDynamicMarker=function(e,t){if(e.update){var n=this.$markerId++;return e.id=n,e.inFront=!!t,t?(this.$frontMarkers[n]=e,this._signal("changeFrontMarker")):(this.$backMarkers[n]=e,this._signal("changeBackMarker")),e}},this.removeMarker=function(e){var t=this.$frontMarkers[e]||this.$backMarkers[e];if(t){var n=t.inFront?this.$frontMarkers:this.$backMarkers;t&&(delete n[e],this._signal(t.inFront?"changeFrontMarker":"changeBackMarker"))}},this.getMarkers=function(e){return e?this.$frontMarkers:this.$backMarkers},this.highlight=function(e){if(!this.$searchHighlight){var t=new d(null,"ace_selected-word","text");this.$searchHighlight=this.addDynamicMarker(t)}this.$searchHighlight.setRegexp(e)},this.highlightLines=function(e,t,n,i){"number"!=typeof t&&(n=t,t=e),n||(n="ace_step");var r=new c(e,0,t,1/0);return r.id=this.addMarker(r,n,"fullLine",i),r},this.setAnnotations=function(e){this.$annotations=e,this._signal("changeAnnotation",{})},this.getAnnotations=function(){return this.$annotations||[]},this.clearAnnotations=function(){this.setAnnotations([])},this.$detectNewLine=function(e){var t=e.match(/^.*?(\r?\n)/m);this.$autoNewLine=t?t[1]:"\n"},this.getWordRange=function(e,t){var n=this.getLine(e),i=!1;if(t>0&&(i=!!n.charAt(t-1).match(this.tokenRe)),i||(i=!!n.charAt(t).match(this.tokenRe)),i)var r=this.tokenRe;else if(/^\s+$/.test(n.slice(t-1,t+1)))r=/\s/;else r=this.nonTokenRe;var o=t;if(o>0){do{o--}while(o>=0&&n.charAt(o).match(r));o++}for(var s=t;s<n.length&&n.charAt(s).match(r);)s++;return new c(e,o,e,s)},this.getAWordRange=function(e,t){for(var n=this.getWordRange(e,t),i=this.getLine(n.end.row);i.charAt(n.end.column).match(/[ \t]/);)n.end.column+=1;return n},this.setNewLineMode=function(e){this.doc.setNewLineMode(e)},this.getNewLineMode=function(){return this.doc.getNewLineMode()},this.setUseWorker=function(e){this.setOption("useWorker",e)},this.getUseWorker=function(){return this.$useWorker},this.onReloadTokenizer=function(e){var t=e.data;this.bgTokenizer.start(t.first),this._signal("tokenizerUpdate",e)},this.$modes={},this.$mode=null,this.$modeId=null,this.setMode=function(e,t){if(e&&"object"==typeof e){if(e.getTokenizer)return this.$onChangeMode(e);var n=e,i=n.path}else i=e||"ace/mode/text";if(this.$modes["ace/mode/text"]||(this.$modes["ace/mode/text"]=new u),this.$modes[i]&&!n)return this.$onChangeMode(this.$modes[i]),void(t&&t());this.$modeId=i,s.loadModule(["mode",i],function(e){if(this.$modeId!==i)return t&&t();this.$modes[i]&&!n?this.$onChangeMode(this.$modes[i]):e&&e.Mode&&(e=new e.Mode(n),n||(this.$modes[i]=e,e.$id=i),this.$onChangeMode(e)),t&&t()}.bind(this)),this.$mode||this.$onChangeMode(this.$modes["ace/mode/text"],!0)},this.$onChangeMode=function(e,t){if(t||(this.$modeId=e.$id),this.$mode!==e){this.$mode=e,this.$stopWorker(),this.$useWorker&&this.$startWorker();var n=e.getTokenizer();if(void 0!==n.addEventListener){var i=this.onReloadTokenizer.bind(this);n.addEventListener("update",i)}if(this.bgTokenizer)this.bgTokenizer.setTokenizer(n);else{this.bgTokenizer=new f(n);var r=this;this.bgTokenizer.addEventListener("update",(function(e){r._signal("tokenizerUpdate",e)}))}this.bgTokenizer.setDocument(this.getDocument()),this.tokenRe=e.tokenRe,this.nonTokenRe=e.nonTokenRe,t||(e.attachToSession&&e.attachToSession(this),this.$options.wrapMethod.set.call(this,this.$wrapMethod),this.$setFolding(e.foldingRules),this.bgTokenizer.start(0),this._emit("changeMode"))}},this.$stopWorker=function(){this.$worker&&(this.$worker.terminate(),this.$worker=null)},this.$startWorker=function(){try{this.$worker=this.$mode.createWorker(this)}catch(e){s.warn("Could not load worker",e),this.$worker=null}},this.getMode=function(){return this.$mode},this.$scrollTop=0,this.setScrollTop=function(e){this.$scrollTop===e||isNaN(e)||(this.$scrollTop=e,this._signal("changeScrollTop",e))},this.getScrollTop=function(){return this.$scrollTop},this.$scrollLeft=0,this.setScrollLeft=function(e){this.$scrollLeft===e||isNaN(e)||(this.$scrollLeft=e,this._signal("changeScrollLeft",e))},this.getScrollLeft=function(){return this.$scrollLeft},this.getScreenWidth=function(){return this.$computeWidth(),this.lineWidgets?Math.max(this.getLineWidgetMaxWidth(),this.screenWidth):this.screenWidth},this.getLineWidgetMaxWidth=function(){if(null!=this.lineWidgetsWidth)return this.lineWidgetsWidth;var e=0;return this.lineWidgets.forEach((function(t){t&&t.screenWidth>e&&(e=t.screenWidth)})),this.lineWidgetWidth=e},this.$computeWidth=function(e){if(this.$modified||e){if(this.$modified=!1,this.$useWrapMode)return this.screenWidth=this.$wrapLimit;for(var t=this.doc.getAllLines(),n=this.$rowLengthCache,i=0,r=0,o=this.$foldData[r],s=o?o.start.row:1/0,a=t.length,l=0;l<a;l++){if(l>s){if((l=o.end.row+1)>=a)break;s=(o=this.$foldData[r++])?o.start.row:1/0}null==n[l]&&(n[l]=this.$getStringScreenWidth(t[l])[0]),n[l]>i&&(i=n[l])}this.screenWidth=i}},this.getLine=function(e){return this.doc.getLine(e)},this.getLines=function(e,t){return this.doc.getLines(e,t)},this.getLength=function(){return this.doc.getLength()},this.getTextRange=function(e){return this.doc.getTextRange(e||this.selection.getRange())},this.insert=function(e,t){return this.doc.insert(e,t)},this.remove=function(e){return this.doc.remove(e)},this.removeFullLines=function(e,t){return this.doc.removeFullLines(e,t)},this.undoChanges=function(e,t){if(e.length){this.$fromUndo=!0;for(var n=null,i=e.length-1;-1!=i;i--){var r=e[i];"doc"==r.group?(this.doc.revertDeltas(r.deltas),n=this.$getUndoSelection(r.deltas,!0,n)):r.deltas.forEach((function(e){this.addFolds(e.folds)}),this)}return this.$fromUndo=!1,n&&this.$undoSelect&&!t&&this.selection.setSelectionRange(n),n}},this.redoChanges=function(e,t){if(e.length){this.$fromUndo=!0;for(var n=null,i=0;i<e.length;i++){var r=e[i];"doc"==r.group&&(this.doc.applyDeltas(r.deltas),n=this.$getUndoSelection(r.deltas,!1,n))}return this.$fromUndo=!1,n&&this.$undoSelect&&!t&&this.selection.setSelectionRange(n),n}},this.setUndoSelect=function(e){this.$undoSelect=e},this.$getUndoSelection=function(e,t,n){function i(e){return t?"insert"!==e.action:"insert"===e.action}var r,o,s=e[0];i(s)?r=c.fromPoints(s.start,s.end):r=c.fromPoints(s.start,s.start);for(var a=1;a<e.length;a++)i(s=e[a])?(o=s.start,-1==r.compare(o.row,o.column)&&r.setStart(o),o=s.end,1==r.compare(o.row,o.column)&&r.setEnd(o),!0):(o=s.start,-1==r.compare(o.row,o.column)&&(r=c.fromPoints(s.start,s.start)),!1);if(null!=n){0===c.comparePoints(n.start,r.start)&&(n.start.column+=r.end.column-r.start.column,n.end.column+=r.end.column-r.start.column);var l=n.compareRange(r);1==l?r.setStart(n.start):-1==l&&r.setEnd(n.end)}return r},this.replace=function(e,t){return this.doc.replace(e,t)},this.moveText=function(e,t,n){var i=this.getTextRange(e),r=this.getFoldsInRange(e),o=c.fromPoints(t,t);if(!n){this.remove(e);var s=e.start.row-e.end.row;(u=s?-e.end.column:e.start.column-e.end.column)&&(o.start.row==e.end.row&&o.start.column>e.end.column&&(o.start.column+=u),o.end.row==e.end.row&&o.end.column>e.end.column&&(o.end.column+=u)),s&&o.start.row>=e.end.row&&(o.start.row+=s,o.end.row+=s)}if(o.end=this.insert(o.start,i),r.length){var a=e.start,l=o.start,u=(s=l.row-a.row,l.column-a.column);this.addFolds(r.map((function(e){return(e=e.clone()).start.row==a.row&&(e.start.column+=u),e.end.row==a.row&&(e.end.column+=u),e.start.row+=s,e.end.row+=s,e})))}return o},this.indentRows=function(e,t,n){n=n.replace(/\t/g,this.getTabString());for(var i=e;i<=t;i++)this.doc.insertInLine({row:i,column:0},n)},this.outdentRows=function(e){for(var t=e.collapseRows(),n=new c(0,0,0,0),i=this.getTabSize(),r=t.start.row;r<=t.end.row;++r){var o=this.getLine(r);n.start.row=r,n.end.row=r;for(var s=0;s<i&&" "==o.charAt(s);++s);s<i&&"\t"==o.charAt(s)?(n.start.column=s,n.end.column=s+1):(n.start.column=0,n.end.column=s),this.remove(n)}},this.$moveLines=function(e,t,n){if(e=this.getRowFoldStart(e),t=this.getRowFoldEnd(t),n<0){if((r=this.getRowFoldStart(e+n))<0)return 0;var i=r-e}else if(n>0){var r;if((r=this.getRowFoldEnd(t+n))>this.doc.getLength()-1)return 0;i=r-t}else{e=this.$clipRowToDocument(e);i=(t=this.$clipRowToDocument(t))-e+1}var o=new c(e,0,t,Number.MAX_VALUE),s=this.getFoldsInRange(o).map((function(e){return(e=e.clone()).start.row+=i,e.end.row+=i,e})),a=0==n?this.doc.getLines(e,t):this.doc.removeFullLines(e,t);return this.doc.insertFullLines(e+i,a),s.length&&this.addFolds(s),i},this.moveLinesUp=function(e,t){return this.$moveLines(e,t,-1)},this.moveLinesDown=function(e,t){return this.$moveLines(e,t,1)},this.duplicateLines=function(e,t){return this.$moveLines(e,t,0)},this.$clipRowToDocument=function(e){return Math.max(0,Math.min(e,this.doc.getLength()-1))},this.$clipColumnToRow=function(e,t){return t<0?0:Math.min(this.doc.getLine(e).length,t)},this.$clipPositionToDocument=function(e,t){if(t=Math.max(0,t),e<0)e=0,t=0;else{var n=this.doc.getLength();e>=n?(e=n-1,t=this.doc.getLine(n-1).length):t=Math.min(this.doc.getLine(e).length,t)}return{row:e,column:t}},this.$clipRangeToDocument=function(e){e.start.row<0?(e.start.row=0,e.start.column=0):e.start.column=this.$clipColumnToRow(e.start.row,e.start.column);var t=this.doc.getLength()-1;return e.end.row>t?(e.end.row=t,e.end.column=this.doc.getLine(t).length):e.end.column=this.$clipColumnToRow(e.end.row,e.end.column),e},this.$wrapLimit=80,this.$useWrapMode=!1,this.$wrapLimitRange={min:null,max:null},this.setUseWrapMode=function(e){if(e!=this.$useWrapMode){if(this.$useWrapMode=e,this.$modified=!0,this.$resetRowCache(0),e){var t=this.getLength();this.$wrapData=Array(t),this.$updateWrapData(0,t-1)}this._signal("changeWrapMode")}},this.getUseWrapMode=function(){return this.$useWrapMode},this.setWrapLimitRange=function(e,t){this.$wrapLimitRange.min===e&&this.$wrapLimitRange.max===t||(this.$wrapLimitRange={min:e,max:t},this.$modified=!0,this.$bidiHandler.markAsDirty(),this.$useWrapMode&&this._signal("changeWrapMode"))},this.adjustWrapLimit=function(e,t){var n=this.$wrapLimitRange;n.max<0&&(n={min:t,max:t});var i=this.$constrainWrapLimit(e,n.min,n.max);return i!=this.$wrapLimit&&i>1&&(this.$wrapLimit=i,this.$modified=!0,this.$useWrapMode&&(this.$updateWrapData(0,this.getLength()-1),this.$resetRowCache(0),this._signal("changeWrapLimit")),!0)},this.$constrainWrapLimit=function(e,t,n){return t&&(e=Math.max(t,e)),n&&(e=Math.min(n,e)),e},this.getWrapLimit=function(){return this.$wrapLimit},this.setWrapLimit=function(e){this.setWrapLimitRange(e,e)},this.getWrapLimitRange=function(){return{min:this.$wrapLimitRange.min,max:this.$wrapLimitRange.max}},this.$updateInternalDataOnChange=function(e){var t=this.$useWrapMode,n=e.action,i=e.start,r=e.end,o=i.row,s=r.row,a=s-o,l=null;if(this.$updating=!0,0!=a)if("remove"===n){this[t?"$wrapData":"$rowLengthCache"].splice(o,a);var u=this.$foldData;l=this.getFoldsInRange(e),this.removeFolds(l);var c=0;if(g=this.getFoldLine(r.row)){g.addRemoveChars(r.row,r.column,i.column-r.column),g.shiftRow(-a);var h=this.getFoldLine(o);h&&h!==g&&(h.merge(g),g=h),c=u.indexOf(g)+1}for(;c<u.length;c++){(g=u[c]).start.row>=r.row&&g.shiftRow(-a)}s=o}else{var f=Array(a);f.unshift(o,0);var d=t?this.$wrapData:this.$rowLengthCache;d.splice.apply(d,f);u=this.$foldData,c=0;if(g=this.getFoldLine(o)){var p=g.range.compareInside(i.row,i.column);0==p?(g=g.split(i.row,i.column))&&(g.shiftRow(a),g.addRemoveChars(s,0,r.column-i.column)):-1==p&&(g.addRemoveChars(o,0,r.column-i.column),g.shiftRow(a)),c=u.indexOf(g)+1}for(;c<u.length;c++){var g;(g=u[c]).start.row>=o&&g.shiftRow(a)}}else a=Math.abs(e.start.column-e.end.column),"remove"===n&&(l=this.getFoldsInRange(e),this.removeFolds(l),a=-a),(g=this.getFoldLine(o))&&g.addRemoveChars(o,i.column,a);return t&&this.$wrapData.length!=this.doc.getLength()&&console.error("doc.getLength() and $wrapData.length have to be the same!"),this.$updating=!1,t?this.$updateWrapData(o,s):this.$updateRowLengthCache(o,s),l},this.$updateRowLengthCache=function(e,t,n){this.$rowLengthCache[e]=null,this.$rowLengthCache[t]=null},this.$updateWrapData=function(n,i){var r,o,s=this.doc.getAllLines(),a=this.getTabSize(),l=this.$wrapData,u=this.$wrapLimit,c=n;for(i=Math.min(i,s.length-1);c<=i;)(o=this.getFoldLine(c,o))?(r=[],o.walk(function(n,i,o,a){var l;if(null!=n){(l=this.$getDisplayTokens(n,r.length))[0]=e;for(var u=1;u<l.length;u++)l[u]=t}else l=this.$getDisplayTokens(s[i].substring(a,o),r.length);r=r.concat(l)}.bind(this),o.end.row,s[o.end.row].length+1),l[o.start.row]=this.$computeWrapSplits(r,u,a),c=o.end.row+1):(r=this.$getDisplayTokens(s[c]),l[c]=this.$computeWrapSplits(r,u,a),c++)};var e=3,t=4;function n(e){return!(e<4352)&&(e>=4352&&e<=4447||e>=4515&&e<=4519||e>=4602&&e<=4607||e>=9001&&e<=9002||e>=11904&&e<=11929||e>=11931&&e<=12019||e>=12032&&e<=12245||e>=12272&&e<=12283||e>=12288&&e<=12350||e>=12353&&e<=12438||e>=12441&&e<=12543||e>=12549&&e<=12589||e>=12593&&e<=12686||e>=12688&&e<=12730||e>=12736&&e<=12771||e>=12784&&e<=12830||e>=12832&&e<=12871||e>=12880&&e<=13054||e>=13056&&e<=19903||e>=19968&&e<=42124||e>=42128&&e<=42182||e>=43360&&e<=43388||e>=44032&&e<=55203||e>=55216&&e<=55238||e>=55243&&e<=55291||e>=63744&&e<=64255||e>=65040&&e<=65049||e>=65072&&e<=65106||e>=65108&&e<=65126||e>=65128&&e<=65131||e>=65281&&e<=65376||e>=65504&&e<=65510)}this.$computeWrapSplits=function(n,i,r){if(0==n.length)return[];var o=[],s=n.length,a=0,l=0,u=this.$wrapAsCode,c=this.$indentedSoftWrap,h=i<=Math.max(2*r,8)||!1===c?0:Math.floor(i/2);function f(e){var t=n.slice(a,e),i=t.length;t.join("").replace(/12/g,(function(){i-=1})).replace(/2/g,(function(){i-=1})),o.length||(d=function(){var e=0;if(0===h)return e;if(c)for(var t=0;t<n.length;t++){var i=n[t];if(10==i)e+=1;else{if(11!=i){if(12==i)continue;break}e+=r}}return u&&!1!==c&&(e+=r),Math.min(e,h)}(),o.indent=d),l+=i,o.push(l),a=e}for(var d=0;s-a>i-d;){var p=a+i-d;if(n[p-1]>=10&&n[p]>=10)f(p);else if(n[p]!=e&&n[p]!=t){for(var g=Math.max(p-(i-(i>>2)),a-1);p>g&&n[p]<e;)p--;if(u){for(;p>g&&n[p]<e;)p--;for(;p>g&&9==n[p];)p--}else for(;p>g&&n[p]<10;)p--;p>g?f(++p):(2==n[p=a+i]&&p--,f(p-d))}else{for(;p!=a-1&&n[p]!=e;p--);if(p>a){f(p);continue}for(p=a+i;p<n.length&&n[p]==t;p++);if(p==n.length)break;f(p)}}return o},this.$getDisplayTokens=function(e,t){var i,r=[];t=t||0;for(var o=0;o<e.length;o++){var s=e.charCodeAt(o);if(9==s){i=this.getScreenTabSize(r.length+t),r.push(11);for(var a=1;a<i;a++)r.push(12)}else 32==s?r.push(10):s>39&&s<48||s>57&&s<64?r.push(9):s>=4352&&n(s)?r.push(1,2):r.push(1)}return r},this.$getStringScreenWidth=function(e,t,i){if(0==t)return[0,0];var r,o;for(null==t&&(t=1/0),i=i||0,o=0;o<e.length&&(9==(r=e.charCodeAt(o))?i+=this.getScreenTabSize(i):r>=4352&&n(r)?i+=2:i+=1,!(i>t));o++);return[i,o]},this.lineWidgets=null,this.getRowLength=function(e){if(this.lineWidgets)var t=this.lineWidgets[e]&&this.lineWidgets[e].rowCount||0;else t=0;return this.$useWrapMode&&this.$wrapData[e]?this.$wrapData[e].length+1+t:1+t},this.getRowLineCount=function(e){return this.$useWrapMode&&this.$wrapData[e]?this.$wrapData[e].length+1:1},this.getRowWrapIndent=function(e){if(this.$useWrapMode){var t=this.screenToDocumentPosition(e,Number.MAX_VALUE),n=this.$wrapData[t.row];return n.length&&n[0]<t.column?n.indent:0}return 0},this.getScreenLastRowColumn=function(e){var t=this.screenToDocumentPosition(e,Number.MAX_VALUE);return this.documentToScreenColumn(t.row,t.column)},this.getDocumentLastRowColumn=function(e,t){var n=this.documentToScreenRow(e,t);return this.getScreenLastRowColumn(n)},this.getDocumentLastRowColumnPosition=function(e,t){var n=this.documentToScreenRow(e,t);return this.screenToDocumentPosition(n,Number.MAX_VALUE/10)},this.getRowSplitData=function(e){return this.$useWrapMode?this.$wrapData[e]:void 0},this.getScreenTabSize=function(e){return this.$tabSize-e%this.$tabSize},this.screenToDocumentRow=function(e,t){return this.screenToDocumentPosition(e,t).row},this.screenToDocumentColumn=function(e,t){return this.screenToDocumentPosition(e,t).column},this.screenToDocumentPosition=function(e,t,n){if(e<0)return{row:0,column:0};var i,r,o=0,s=0,a=0,l=0,u=this.$screenRowCache,c=this.$getRowCacheIndex(u,e),h=u.length;if(h&&c>=0){a=u[c],o=this.$docRowCache[c];var f=e>u[h-1]}else f=!h;for(var d=this.getLength()-1,p=this.getNextFoldLine(o),g=p?p.start.row:1/0;a<=e&&!(a+(l=this.getRowLength(o))>e||o>=d);)a+=l,++o>g&&(o=p.end.row+1,g=(p=this.getNextFoldLine(o,p))?p.start.row:1/0),f&&(this.$docRowCache.push(o),this.$screenRowCache.push(a));if(p&&p.start.row<=o)i=this.getFoldDisplayLine(p),o=p.start.row;else{if(a+l<=e||o>d)return{row:d,column:this.getLine(d).length};i=this.getLine(o),p=null}var m=0,v=Math.floor(e-a);if(this.$useWrapMode){var b=this.$wrapData[o];b&&(r=b[v],v>0&&b.length&&(m=b.indent,s=b[v-1]||b[b.length-1],i=i.substring(s)))}return void 0!==n&&this.$bidiHandler.isBidiRow(a+v,o,v)&&(t=this.$bidiHandler.offsetToCol(n)),s+=this.$getStringScreenWidth(i,t-m)[1],this.$useWrapMode&&s>=r&&(s=r-1),p?p.idxToPosition(s):{row:o,column:s}},this.documentToScreenPosition=function(e,t){if(void 0===t)var n=this.$clipPositionToDocument(e.row,e.column);else n=this.$clipPositionToDocument(e,t);e=n.row,t=n.column;var i,r=0,o=null;(i=this.getFoldAt(e,t,1))&&(e=i.start.row,t=i.start.column);var s,a=0,l=this.$docRowCache,u=this.$getRowCacheIndex(l,e),c=l.length;if(c&&u>=0){a=l[u],r=this.$screenRowCache[u];var h=e>l[c-1]}else h=!c;for(var f=this.getNextFoldLine(a),d=f?f.start.row:1/0;a<e;){if(a>=d){if((s=f.end.row+1)>e)break;d=(f=this.getNextFoldLine(s,f))?f.start.row:1/0}else s=a+1;r+=this.getRowLength(a),a=s,h&&(this.$docRowCache.push(a),this.$screenRowCache.push(r))}var p="";f&&a>=d?(p=this.getFoldDisplayLine(f,e,t),o=f.start.row):(p=this.getLine(e).substring(0,t),o=e);var g=0;if(this.$useWrapMode){var m=this.$wrapData[o];if(m){for(var v=0;p.length>=m[v];)r++,v++;p=p.substring(m[v-1]||0,p.length),g=v>0?m.indent:0}}return{row:r,column:g+this.$getStringScreenWidth(p)[0]}},this.documentToScreenColumn=function(e,t){return this.documentToScreenPosition(e,t).column},this.documentToScreenRow=function(e,t){return this.documentToScreenPosition(e,t).row},this.getScreenLength=function(){var e=0,t=null;if(this.$useWrapMode)for(var n=this.$wrapData.length,i=0,r=(a=0,(t=this.$foldData[a++])?t.start.row:1/0);i<n;){var o=this.$wrapData[i];e+=o?o.length+1:1,++i>r&&(i=t.end.row+1,r=(t=this.$foldData[a++])?t.start.row:1/0)}else{e=this.getLength();for(var s=this.$foldData,a=0;a<s.length;a++)e-=(t=s[a]).end.row-t.start.row}return this.lineWidgets&&(e+=this.$getWidgetScreenLength()),e},this.$setFontMetrics=function(e){this.$enableVarChar&&(this.$getStringScreenWidth=function(t,n,i){if(0===n)return[0,0];var r,o;for(n||(n=1/0),i=i||0,o=0;o<t.length&&!((i+="\t"===(r=t.charAt(o))?this.getScreenTabSize(i):e.getCharacterWidth(r))>n);o++);return[i,o]})},this.destroy=function(){this.bgTokenizer&&(this.bgTokenizer.setDocument(null),this.bgTokenizer=null),this.$stopWorker()},this.isFullWidth=n}.call(p.prototype),e("./edit_session/folding").Folding.call(p.prototype),e("./edit_session/bracket_match").BracketMatch.call(p.prototype),s.defineOptions(p.prototype,"session",{wrap:{set:function(e){if(e&&"off"!=e?"free"==e?e=!0:"printMargin"==e?e=-1:"string"==typeof e&&(e=parseInt(e,10)||!1):e=!1,this.$wrap!=e)if(this.$wrap=e,e){var t="number"==typeof e?e:null;this.setWrapLimitRange(t,t),this.setUseWrapMode(!0)}else this.setUseWrapMode(!1)},get:function(){return this.getUseWrapMode()?-1==this.$wrap?"printMargin":this.getWrapLimitRange().min?this.$wrap:"free":"off"},handlesSet:!0},wrapMethod:{set:function(e){(e="auto"==e?"text"!=this.$mode.type:"text"!=e)!=this.$wrapAsCode&&(this.$wrapAsCode=e,this.$useWrapMode&&(this.$modified=!0,this.$resetRowCache(0),this.$updateWrapData(0,this.getLength()-1)))},initialValue:"auto"},indentedSoftWrap:{initialValue:!0},firstLineNumber:{set:function(){this._signal("changeBreakpoint")},initialValue:1},useWorker:{set:function(e){this.$useWorker=e,this.$stopWorker(),e&&this.$startWorker()},initialValue:!0},useSoftTabs:{initialValue:!0},tabSize:{set:function(e){isNaN(e)||this.$tabSize===e||(this.$modified=!0,this.$rowLengthCache=[],this.$tabSize=e,this._signal("changeTabSize"))},initialValue:4,handlesSet:!0},navigateWithinSoftTabs:{initialValue:!1},overwrite:{set:function(e){this._signal("changeOverwrite")},initialValue:!1},newLineMode:{set:function(e){this.doc.setNewLineMode(e)},get:function(){return this.doc.getNewLineMode()},handlesSet:!0},mode:{set:function(e){this.setMode(e)},get:function(){return this.$modeId}}}),t.EditSession=p})),ace.define("ace/search",["require","exports","module","ace/lib/lang","ace/lib/oop","ace/range"],(function(e,t,n){"use strict";var i=e("./lib/lang"),r=e("./lib/oop"),o=e("./range").Range,s=function(){this.$options={}};(function(){this.set=function(e){return r.mixin(this.$options,e),this},this.getOptions=function(){return i.copyObject(this.$options)},this.setOptions=function(e){this.$options=e},this.find=function(e){var t=this.$options,n=this.$matchIterator(e,t);if(!n)return!1;var i=null;return n.forEach((function(e,n,r,s){return i=new o(e,n,r,s),!(n==s&&t.start&&t.start.start&&0!=t.skipCurrent&&i.isEqual(t.start))||(i=null,!1)})),i},this.findAll=function(e){var t=this.$options;if(!t.needle)return[];this.$assembleRegExp(t);var n=t.range,r=n?e.getLines(n.start.row,n.end.row):e.doc.getAllLines(),s=[],a=t.re;if(t.$isMultiLine){var l,u=a.length,c=r.length-u;e:for(var h=a.offset||0;h<=c;h++){for(var f=0;f<u;f++)if(-1==r[h+f].search(a[f]))continue e;var d=r[h],p=r[h+u-1],g=d.length-d.match(a[0])[0].length,m=p.match(a[u-1])[0].length;l&&l.end.row===h&&l.end.column>g||(s.push(l=new o(h,g,h+u-1,m)),u>2&&(h=h+u-2))}}else for(var v=0;v<r.length;v++){var b=i.getMatchOffsets(r[v],a);for(f=0;f<b.length;f++){var y=b[f];s.push(new o(v,y.offset,v,y.offset+y.length))}}if(n){var w=n.start.column,A=n.start.column;for(v=0,f=s.length-1;v<f&&s[v].start.column<w&&s[v].start.row==n.start.row;)v++;for(;v<f&&s[f].end.column>A&&s[f].end.row==n.end.row;)f--;for(s=s.slice(v,f+1),v=0,f=s.length;v<f;v++)s[v].start.row+=n.start.row,s[v].end.row+=n.start.row}return s},this.replace=function(e,t){var n=this.$options,i=this.$assembleRegExp(n);if(n.$isMultiLine)return t;if(i){var r=i.exec(e);if(!r||r[0].length!=e.length)return null;if(t=e.replace(i,t),n.preserveCase){t=t.split("");for(var o=Math.min(e.length,e.length);o--;){var s=e[o];s&&s.toLowerCase()!=s?t[o]=t[o].toUpperCase():t[o]=t[o].toLowerCase()}t=t.join("")}return t}},this.$assembleRegExp=function(e,t){if(e.needle instanceof RegExp)return e.re=e.needle;var n=e.needle;if(!e.needle)return e.re=!1;e.regExp||(n=i.escapeRegExp(n)),e.wholeWord&&(n=function(e,t){function n(e){return/\w/.test(e)||t.regExp?"\\b":""}return n(e[0])+e+n(e[e.length-1])}(n,e));var r=e.caseSensitive?"gm":"gmi";if(e.$isMultiLine=!t&&/[\n\r]/.test(n),e.$isMultiLine)return e.re=this.$assembleMultilineRegExp(n,r);try{var o=new RegExp(n,r)}catch(e){o=!1}return e.re=o},this.$assembleMultilineRegExp=function(e,t){for(var n=e.replace(/\r\n|\r|\n/g,"$\n^").split("\n"),i=[],r=0;r<n.length;r++)try{i.push(new RegExp(n[r],t))}catch(e){return!1}return i},this.$matchIterator=function(e,t){var n=this.$assembleRegExp(t);if(!n)return!1;var i=1==t.backwards,r=0!=t.skipCurrent,o=t.range,s=t.start;s||(s=o?o[i?"end":"start"]:e.selection.getRange()),s.start&&(s=s[r!=i?"end":"start"]);var a=o?o.start.row:0,l=o?o.end.row:e.getLength()-1;if(i)var u=function(e){var n=s.row;if(!h(n,s.column,e)){for(n--;n>=a;n--)if(h(n,Number.MAX_VALUE,e))return;if(0!=t.wrap)for(n=l,a=s.row;n>=a;n--)if(h(n,Number.MAX_VALUE,e))return}};else u=function(e){var n=s.row;if(!h(n,s.column,e)){for(n+=1;n<=l;n++)if(h(n,0,e))return;if(0!=t.wrap)for(n=a,l=s.row;n<=l;n++)if(h(n,0,e))return}};if(t.$isMultiLine)var c=n.length,h=function(t,r,o){var s=i?t-c+1:t;if(!(s<0)){var a=e.getLine(s),l=a.search(n[0]);if(!(!i&&l<r||-1===l)){for(var u=1;u<c;u++)if(-1==(a=e.getLine(s+u)).search(n[u]))return;var h=a.match(n[c-1])[0].length;if(!(i&&h>r))return!!o(s,l,s+c-1,h)||void 0}}};else if(i)h=function(t,i,r){var o,s=e.getLine(t),a=[],l=0;for(n.lastIndex=0;o=n.exec(s);){var u=o[0].length;if(l=o.index,!u){if(l>=s.length)break;n.lastIndex=l+=1}if(o.index+u>i)break;a.push(o.index,u)}for(var c=a.length-1;c>=0;c-=2){var h=a[c-1];if(r(t,h,t,h+(u=a[c])))return!0}};else h=function(t,i,r){var o,s=e.getLine(t),a=i;for(n.lastIndex=i;o=n.exec(s);){var l=o[0].length;if(r(t,a=o.index,t,a+l))return!0;if(!l&&(n.lastIndex=a+=1,a>=s.length))return!1}};return{forEach:u}}}).call(s.prototype),t.Search=s})),ace.define("ace/keyboard/hash_handler",["require","exports","module","ace/lib/keys","ace/lib/useragent"],(function(e,t,n){"use strict";var i=e("../lib/keys"),r=e("../lib/useragent"),o=i.KEY_MODS;function s(e,t){this.platform=t||(r.isMac?"mac":"win"),this.commands={},this.commandKeyBinding={},this.addCommands(e),this.$singleCommand=!0}function a(e,t){s.call(this,e,t),this.$singleCommand=!1}a.prototype=s.prototype,function(){function e(e){return"object"==typeof e&&e.bindKey&&e.bindKey.position||(e.isDefault?-100:0)}this.addCommand=function(e){this.commands[e.name]&&this.removeCommand(e),this.commands[e.name]=e,e.bindKey&&this._buildKeyHash(e)},this.removeCommand=function(e,t){var n=e&&("string"==typeof e?e:e.name);e=this.commands[n],t||delete this.commands[n];var i=this.commandKeyBinding;for(var r in i){var o=i[r];if(o==e)delete i[r];else if(Array.isArray(o)){var s=o.indexOf(e);-1!=s&&(o.splice(s,1),1==o.length&&(i[r]=o[0]))}}},this.bindKey=function(e,t,n){if("object"==typeof e&&e&&(null==n&&(n=e.position),e=e[this.platform]),e)return"function"==typeof t?this.addCommand({exec:t,bindKey:e,name:t.name||e}):void e.split("|").forEach((function(e){var i="";if(-1!=e.indexOf(" ")){var r=e.split(/\s+/);e=r.pop(),r.forEach((function(e){var t=this.parseKeys(e),n=o[t.hashId]+t.key;i+=(i?" ":"")+n,this._addCommandToBinding(i,"chainKeys")}),this),i+=" "}var s=this.parseKeys(e),a=o[s.hashId]+s.key;this._addCommandToBinding(i+a,t,n)}),this)},this._addCommandToBinding=function(t,n,i){var r,o=this.commandKeyBinding;if(n)if(!o[t]||this.$singleCommand)o[t]=n;else{Array.isArray(o[t])?-1!=(r=o[t].indexOf(n))&&o[t].splice(r,1):o[t]=[o[t]],"number"!=typeof i&&(i=e(n));var s=o[t];for(r=0;r<s.length;r++){if(e(s[r])>i)break}s.splice(r,0,n)}else delete o[t]},this.addCommands=function(e){e&&Object.keys(e).forEach((function(t){var n=e[t];if(n){if("string"==typeof n)return this.bindKey(n,t);"function"==typeof n&&(n={exec:n}),"object"==typeof n&&(n.name||(n.name=t),this.addCommand(n))}}),this)},this.removeCommands=function(e){Object.keys(e).forEach((function(t){this.removeCommand(e[t])}),this)},this.bindKeys=function(e){Object.keys(e).forEach((function(t){this.bindKey(t,e[t])}),this)},this._buildKeyHash=function(e){this.bindKey(e.bindKey,e)},this.parseKeys=function(e){var t=e.toLowerCase().split(/[\-\+]([\-\+])?/).filter((function(e){return e})),n=t.pop(),r=i[n];if(i.FUNCTION_KEYS[r])n=i.FUNCTION_KEYS[r].toLowerCase();else{if(!t.length)return{key:n,hashId:-1};if(1==t.length&&"shift"==t[0])return{key:n.toUpperCase(),hashId:-1}}for(var o=0,s=t.length;s--;){var a=i.KEY_MODS[t[s]];if(null==a)return"undefined"!=typeof console&&console.error("invalid modifier "+t[s]+" in "+e),!1;o|=a}return{key:n,hashId:o}},this.findKeyCommand=function(e,t){var n=o[e]+t;return this.commandKeyBinding[n]},this.handleKeyboard=function(e,t,n,i){if(!(i<0)){var r=o[t]+n,s=this.commandKeyBinding[r];return e.$keyChain&&(e.$keyChain+=" "+r,s=this.commandKeyBinding[e.$keyChain]||s),!s||"chainKeys"!=s&&"chainKeys"!=s[s.length-1]?(e.$keyChain&&(t&&4!=t||1!=n.length?(-1==t||i>0)&&(e.$keyChain=""):e.$keyChain=e.$keyChain.slice(0,-r.length-1)),{command:s}):(e.$keyChain=e.$keyChain||r,{command:"null"})}},this.getStatusText=function(e,t){return t.$keyChain||""}}.call(s.prototype),t.HashHandler=s,t.MultiHashHandler=a})),ace.define("ace/commands/command_manager",["require","exports","module","ace/lib/oop","ace/keyboard/hash_handler","ace/lib/event_emitter"],(function(e,t,n){"use strict";var i=e("../lib/oop"),r=e("../keyboard/hash_handler").MultiHashHandler,o=e("../lib/event_emitter").EventEmitter,s=function(e,t){r.call(this,t,e),this.byName=this.commands,this.setDefaultHandler("exec",(function(e){return e.command.exec(e.editor,e.args||{})}))};i.inherits(s,r),function(){i.implement(this,o),this.exec=function(e,t,n){if(Array.isArray(e)){for(var i=e.length;i--;)if(this.exec(e[i],t,n))return!0;return!1}if("string"==typeof e&&(e=this.commands[e]),!e)return!1;if(t&&t.$readOnly&&!e.readOnly)return!1;if(e.isAvailable&&!e.isAvailable(t))return!1;var r={editor:t,command:e,args:n};return r.returnValue=this._emit("exec",r),this._signal("afterExec",r),!1!==r.returnValue},this.toggleRecording=function(e){if(!this.$inReplay)return e&&e._emit("changeStatus"),this.recording?(this.macro.pop(),this.removeEventListener("exec",this.$addCommandToMacro),this.macro.length||(this.macro=this.oldMacro),this.recording=!1):(this.$addCommandToMacro||(this.$addCommandToMacro=function(e){this.macro.push([e.command,e.args])}.bind(this)),this.oldMacro=this.macro,this.macro=[],this.on("exec",this.$addCommandToMacro),this.recording=!0)},this.replay=function(e){if(!this.$inReplay&&this.macro){if(this.recording)return this.toggleRecording(e);try{this.$inReplay=!0,this.macro.forEach((function(t){"string"==typeof t?this.exec(t,e):this.exec(t[0],e,t[1])}),this)}finally{this.$inReplay=!1}}},this.trimMacro=function(e){return e.map((function(e){return"string"!=typeof e[0]&&(e[0]=e[0].name),e[1]||(e=e[0]),e}))}}.call(s.prototype),t.CommandManager=s})),ace.define("ace/commands/default_commands",["require","exports","module","ace/lib/lang","ace/config","ace/range"],(function(e,t,n){"use strict";var i=e("../lib/lang"),r=e("../config"),o=e("../range").Range;function s(e,t){return{win:e,mac:t}}t.commands=[{name:"showSettingsMenu",bindKey:s("Ctrl-,","Command-,"),exec:function(e){r.loadModule("ace/ext/settings_menu",(function(t){t.init(e),e.showSettingsMenu()}))},readOnly:!0},{name:"goToNextError",bindKey:s("Alt-E","F4"),exec:function(e){r.loadModule("ace/ext/error_marker",(function(t){t.showErrorMarker(e,1)}))},scrollIntoView:"animate",readOnly:!0},{name:"goToPreviousError",bindKey:s("Alt-Shift-E","Shift-F4"),exec:function(e){r.loadModule("ace/ext/error_marker",(function(t){t.showErrorMarker(e,-1)}))},scrollIntoView:"animate",readOnly:!0},{name:"selectall",bindKey:s("Ctrl-A","Command-A"),exec:function(e){e.selectAll()},readOnly:!0},{name:"centerselection",bindKey:s(null,"Ctrl-L"),exec:function(e){e.centerSelection()},readOnly:!0},{name:"gotoline",bindKey:s("Ctrl-L","Command-L"),exec:function(e){var t=parseInt(prompt("Enter line number:"),10);isNaN(t)||e.gotoLine(t)},readOnly:!0},{name:"fold",bindKey:s("Alt-L|Ctrl-F1","Command-Alt-L|Command-F1"),exec:function(e){e.session.toggleFold(!1)},multiSelectAction:"forEach",scrollIntoView:"center",readOnly:!0},{name:"unfold",bindKey:s("Alt-Shift-L|Ctrl-Shift-F1","Command-Alt-Shift-L|Command-Shift-F1"),exec:function(e){e.session.toggleFold(!0)},multiSelectAction:"forEach",scrollIntoView:"center",readOnly:!0},{name:"toggleFoldWidget",bindKey:s("F2","F2"),exec:function(e){e.session.toggleFoldWidget()},multiSelectAction:"forEach",scrollIntoView:"center",readOnly:!0},{name:"toggleParentFoldWidget",bindKey:s("Alt-F2","Alt-F2"),exec:function(e){e.session.toggleFoldWidget(!0)},multiSelectAction:"forEach",scrollIntoView:"center",readOnly:!0},{name:"foldall",bindKey:s(null,"Ctrl-Command-Option-0"),exec:function(e){e.session.foldAll()},scrollIntoView:"center",readOnly:!0},{name:"foldOther",bindKey:s("Alt-0","Command-Option-0"),exec:function(e){e.session.foldAll(),e.session.unfold(e.selection.getAllRanges())},scrollIntoView:"center",readOnly:!0},{name:"unfoldall",bindKey:s("Alt-Shift-0","Command-Option-Shift-0"),exec:function(e){e.session.unfold()},scrollIntoView:"center",readOnly:!0},{name:"findnext",bindKey:s("Ctrl-K","Command-G"),exec:function(e){e.findNext()},multiSelectAction:"forEach",scrollIntoView:"center",readOnly:!0},{name:"findprevious",bindKey:s("Ctrl-Shift-K","Command-Shift-G"),exec:function(e){e.findPrevious()},multiSelectAction:"forEach",scrollIntoView:"center",readOnly:!0},{name:"selectOrFindNext",bindKey:s("Alt-K","Ctrl-G"),exec:function(e){e.selection.isEmpty()?e.selection.selectWord():e.findNext()},readOnly:!0},{name:"selectOrFindPrevious",bindKey:s("Alt-Shift-K","Ctrl-Shift-G"),exec:function(e){e.selection.isEmpty()?e.selection.selectWord():e.findPrevious()},readOnly:!0},{name:"find",bindKey:s("Ctrl-F","Command-F"),exec:function(e){r.loadModule("ace/ext/searchbox",(function(t){t.Search(e)}))},readOnly:!0},{name:"overwrite",bindKey:"Insert",exec:function(e){e.toggleOverwrite()},readOnly:!0},{name:"selecttostart",bindKey:s("Ctrl-Shift-Home","Command-Shift-Home|Command-Shift-Up"),exec:function(e){e.getSelection().selectFileStart()},multiSelectAction:"forEach",readOnly:!0,scrollIntoView:"animate",aceCommandGroup:"fileJump"},{name:"gotostart",bindKey:s("Ctrl-Home","Command-Home|Command-Up"),exec:function(e){e.navigateFileStart()},multiSelectAction:"forEach",readOnly:!0,scrollIntoView:"animate",aceCommandGroup:"fileJump"},{name:"selectup",bindKey:s("Shift-Up","Shift-Up|Ctrl-Shift-P"),exec:function(e){e.getSelection().selectUp()},multiSelectAction:"forEach",scrollIntoView:"cursor",readOnly:!0},{name:"golineup",bindKey:s("Up","Up|Ctrl-P"),exec:function(e,t){e.navigateUp(t.times)},multiSelectAction:"forEach",scrollIntoView:"cursor",readOnly:!0},{name:"selecttoend",bindKey:s("Ctrl-Shift-End","Command-Shift-End|Command-Shift-Down"),exec:function(e){e.getSelection().selectFileEnd()},multiSelectAction:"forEach",readOnly:!0,scrollIntoView:"animate",aceCommandGroup:"fileJump"},{name:"gotoend",bindKey:s("Ctrl-End","Command-End|Command-Down"),exec:function(e){e.navigateFileEnd()},multiSelectAction:"forEach",readOnly:!0,scrollIntoView:"animate",aceCommandGroup:"fileJump"},{name:"selectdown",bindKey:s("Shift-Down","Shift-Down|Ctrl-Shift-N"),exec:function(e){e.getSelection().selectDown()},multiSelectAction:"forEach",scrollIntoView:"cursor",readOnly:!0},{name:"golinedown",bindKey:s("Down","Down|Ctrl-N"),exec:function(e,t){e.navigateDown(t.times)},multiSelectAction:"forEach",scrollIntoView:"cursor",readOnly:!0},{name:"selectwordleft",bindKey:s("Ctrl-Shift-Left","Option-Shift-Left"),exec:function(e){e.getSelection().selectWordLeft()},multiSelectAction:"forEach",scrollIntoView:"cursor",readOnly:!0},{name:"gotowordleft",bindKey:s("Ctrl-Left","Option-Left"),exec:function(e){e.navigateWordLeft()},multiSelectAction:"forEach",scrollIntoView:"cursor",readOnly:!0},{name:"selecttolinestart",bindKey:s("Alt-Shift-Left","Command-Shift-Left|Ctrl-Shift-A"),exec:function(e){e.getSelection().selectLineStart()},multiSelectAction:"forEach",scrollIntoView:"cursor",readOnly:!0},{name:"gotolinestart",bindKey:s("Alt-Left|Home","Command-Left|Home|Ctrl-A"),exec:function(e){e.navigateLineStart()},multiSelectAction:"forEach",scrollIntoView:"cursor",readOnly:!0},{name:"selectleft",bindKey:s("Shift-Left","Shift-Left|Ctrl-Shift-B"),exec:function(e){e.getSelection().selectLeft()},multiSelectAction:"forEach",scrollIntoView:"cursor",readOnly:!0},{name:"gotoleft",bindKey:s("Left","Left|Ctrl-B"),exec:function(e,t){e.navigateLeft(t.times)},multiSelectAction:"forEach",scrollIntoView:"cursor",readOnly:!0},{name:"selectwordright",bindKey:s("Ctrl-Shift-Right","Option-Shift-Right"),exec:function(e){e.getSelection().selectWordRight()},multiSelectAction:"forEach",scrollIntoView:"cursor",readOnly:!0},{name:"gotowordright",bindKey:s("Ctrl-Right","Option-Right"),exec:function(e){e.navigateWordRight()},multiSelectAction:"forEach",scrollIntoView:"cursor",readOnly:!0},{name:"selecttolineend",bindKey:s("Alt-Shift-Right","Command-Shift-Right|Shift-End|Ctrl-Shift-E"),exec:function(e){e.getSelection().selectLineEnd()},multiSelectAction:"forEach",scrollIntoView:"cursor",readOnly:!0},{name:"gotolineend",bindKey:s("Alt-Right|End","Command-Right|End|Ctrl-E"),exec:function(e){e.navigateLineEnd()},multiSelectAction:"forEach",scrollIntoView:"cursor",readOnly:!0},{name:"selectright",bindKey:s("Shift-Right","Shift-Right"),exec:function(e){e.getSelection().selectRight()},multiSelectAction:"forEach",scrollIntoView:"cursor",readOnly:!0},{name:"gotoright",bindKey:s("Right","Right|Ctrl-F"),exec:function(e,t){e.navigateRight(t.times)},multiSelectAction:"forEach",scrollIntoView:"cursor",readOnly:!0},{name:"selectpagedown",bindKey:"Shift-PageDown",exec:function(e){e.selectPageDown()},readOnly:!0},{name:"pagedown",bindKey:s(null,"Option-PageDown"),exec:function(e){e.scrollPageDown()},readOnly:!0},{name:"gotopagedown",bindKey:s("PageDown","PageDown|Ctrl-V"),exec:function(e){e.gotoPageDown()},readOnly:!0},{name:"selectpageup",bindKey:"Shift-PageUp",exec:function(e){e.selectPageUp()},readOnly:!0},{name:"pageup",bindKey:s(null,"Option-PageUp"),exec:function(e){e.scrollPageUp()},readOnly:!0},{name:"gotopageup",bindKey:"PageUp",exec:function(e){e.gotoPageUp()},readOnly:!0},{name:"scrollup",bindKey:s("Ctrl-Up",null),exec:function(e){e.renderer.scrollBy(0,-2*e.renderer.layerConfig.lineHeight)},readOnly:!0},{name:"scrolldown",bindKey:s("Ctrl-Down",null),exec:function(e){e.renderer.scrollBy(0,2*e.renderer.layerConfig.lineHeight)},readOnly:!0},{name:"selectlinestart",bindKey:"Shift-Home",exec:function(e){e.getSelection().selectLineStart()},multiSelectAction:"forEach",scrollIntoView:"cursor",readOnly:!0},{name:"selectlineend",bindKey:"Shift-End",exec:function(e){e.getSelection().selectLineEnd()},multiSelectAction:"forEach",scrollIntoView:"cursor",readOnly:!0},{name:"togglerecording",bindKey:s("Ctrl-Alt-E","Command-Option-E"),exec:function(e){e.commands.toggleRecording(e)},readOnly:!0},{name:"replaymacro",bindKey:s("Ctrl-Shift-E","Command-Shift-E"),exec:function(e){e.commands.replay(e)},readOnly:!0},{name:"jumptomatching",bindKey:s("Ctrl-P","Ctrl-P"),exec:function(e){e.jumpToMatching()},multiSelectAction:"forEach",scrollIntoView:"animate",readOnly:!0},{name:"selecttomatching",bindKey:s("Ctrl-Shift-P","Ctrl-Shift-P"),exec:function(e){e.jumpToMatching(!0)},multiSelectAction:"forEach",scrollIntoView:"animate",readOnly:!0},{name:"expandToMatching",bindKey:s("Ctrl-Shift-M","Ctrl-Shift-M"),exec:function(e){e.jumpToMatching(!0,!0)},multiSelectAction:"forEach",scrollIntoView:"animate",readOnly:!0},{name:"passKeysToBrowser",bindKey:s(null,null),exec:function(){},passEvent:!0,readOnly:!0},{name:"copy",exec:function(e){},readOnly:!0},{name:"cut",exec:function(e){var t=e.getSelectionRange();e._emit("cut",t),e.selection.isEmpty()||(e.session.remove(t),e.clearSelection())},scrollIntoView:"cursor",multiSelectAction:"forEach"},{name:"paste",exec:function(e,t){e.$handlePaste(t)},scrollIntoView:"cursor"},{name:"removeline",bindKey:s("Ctrl-D","Command-D"),exec:function(e){e.removeLines()},scrollIntoView:"cursor",multiSelectAction:"forEachLine"},{name:"duplicateSelection",bindKey:s("Ctrl-Shift-D","Command-Shift-D"),exec:function(e){e.duplicateSelection()},scrollIntoView:"cursor",multiSelectAction:"forEach"},{name:"sortlines",bindKey:s("Ctrl-Alt-S","Command-Alt-S"),exec:function(e){e.sortLines()},scrollIntoView:"selection",multiSelectAction:"forEachLine"},{name:"togglecomment",bindKey:s("Ctrl-/","Command-/"),exec:function(e){e.toggleCommentLines()},multiSelectAction:"forEachLine",scrollIntoView:"selectionPart"},{name:"toggleBlockComment",bindKey:s("Ctrl-Shift-/","Command-Shift-/"),exec:function(e){e.toggleBlockComment()},multiSelectAction:"forEach",scrollIntoView:"selectionPart"},{name:"modifyNumberUp",bindKey:s("Ctrl-Shift-Up","Alt-Shift-Up"),exec:function(e){e.modifyNumber(1)},scrollIntoView:"cursor",multiSelectAction:"forEach"},{name:"modifyNumberDown",bindKey:s("Ctrl-Shift-Down","Alt-Shift-Down"),exec:function(e){e.modifyNumber(-1)},scrollIntoView:"cursor",multiSelectAction:"forEach"},{name:"replace",bindKey:s("Ctrl-H","Command-Option-F"),exec:function(e){r.loadModule("ace/ext/searchbox",(function(t){t.Search(e,!0)}))}},{name:"undo",bindKey:s("Ctrl-Z","Command-Z"),exec:function(e){e.undo()}},{name:"redo",bindKey:s("Ctrl-Shift-Z|Ctrl-Y","Command-Shift-Z|Command-Y"),exec:function(e){e.redo()}},{name:"copylinesup",bindKey:s("Alt-Shift-Up","Command-Option-Up"),exec:function(e){e.copyLinesUp()},scrollIntoView:"cursor"},{name:"movelinesup",bindKey:s("Alt-Up","Option-Up"),exec:function(e){e.moveLinesUp()},scrollIntoView:"cursor"},{name:"copylinesdown",bindKey:s("Alt-Shift-Down","Command-Option-Down"),exec:function(e){e.copyLinesDown()},scrollIntoView:"cursor"},{name:"movelinesdown",bindKey:s("Alt-Down","Option-Down"),exec:function(e){e.moveLinesDown()},scrollIntoView:"cursor"},{name:"del",bindKey:s("Delete","Delete|Ctrl-D|Shift-Delete"),exec:function(e){e.remove("right")},multiSelectAction:"forEach",scrollIntoView:"cursor"},{name:"backspace",bindKey:s("Shift-Backspace|Backspace","Ctrl-Backspace|Shift-Backspace|Backspace|Ctrl-H"),exec:function(e){e.remove("left")},multiSelectAction:"forEach",scrollIntoView:"cursor"},{name:"cut_or_delete",bindKey:s("Shift-Delete",null),exec:function(e){if(!e.selection.isEmpty())return!1;e.remove("left")},multiSelectAction:"forEach",scrollIntoView:"cursor"},{name:"removetolinestart",bindKey:s("Alt-Backspace","Command-Backspace"),exec:function(e){e.removeToLineStart()},multiSelectAction:"forEach",scrollIntoView:"cursor"},{name:"removetolineend",bindKey:s("Alt-Delete","Ctrl-K|Command-Delete"),exec:function(e){e.removeToLineEnd()},multiSelectAction:"forEach",scrollIntoView:"cursor"},{name:"removetolinestarthard",bindKey:s("Ctrl-Shift-Backspace",null),exec:function(e){var t=e.selection.getRange();t.start.column=0,e.session.remove(t)},multiSelectAction:"forEach",scrollIntoView:"cursor"},{name:"removetolineendhard",bindKey:s("Ctrl-Shift-Delete",null),exec:function(e){var t=e.selection.getRange();t.end.column=Number.MAX_VALUE,e.session.remove(t)},multiSelectAction:"forEach",scrollIntoView:"cursor"},{name:"removewordleft",bindKey:s("Ctrl-Backspace","Alt-Backspace|Ctrl-Alt-Backspace"),exec:function(e){e.removeWordLeft()},multiSelectAction:"forEach",scrollIntoView:"cursor"},{name:"removewordright",bindKey:s("Ctrl-Delete","Alt-Delete"),exec:function(e){e.removeWordRight()},multiSelectAction:"forEach",scrollIntoView:"cursor"},{name:"outdent",bindKey:s("Shift-Tab","Shift-Tab"),exec:function(e){e.blockOutdent()},multiSelectAction:"forEach",scrollIntoView:"selectionPart"},{name:"indent",bindKey:s("Tab","Tab"),exec:function(e){e.indent()},multiSelectAction:"forEach",scrollIntoView:"selectionPart"},{name:"blockoutdent",bindKey:s("Ctrl-[","Ctrl-["),exec:function(e){e.blockOutdent()},multiSelectAction:"forEachLine",scrollIntoView:"selectionPart"},{name:"blockindent",bindKey:s("Ctrl-]","Ctrl-]"),exec:function(e){e.blockIndent()},multiSelectAction:"forEachLine",scrollIntoView:"selectionPart"},{name:"insertstring",exec:function(e,t){e.insert(t)},multiSelectAction:"forEach",scrollIntoView:"cursor"},{name:"inserttext",exec:function(e,t){e.insert(i.stringRepeat(t.text||"",t.times||1))},multiSelectAction:"forEach",scrollIntoView:"cursor"},{name:"splitline",bindKey:s(null,"Ctrl-O"),exec:function(e){e.splitLine()},multiSelectAction:"forEach",scrollIntoView:"cursor"},{name:"transposeletters",bindKey:s("Alt-Shift-X","Ctrl-T"),exec:function(e){e.transposeLetters()},multiSelectAction:function(e){e.transposeSelections(1)},scrollIntoView:"cursor"},{name:"touppercase",bindKey:s("Ctrl-U","Ctrl-U"),exec:function(e){e.toUpperCase()},multiSelectAction:"forEach",scrollIntoView:"cursor"},{name:"tolowercase",bindKey:s("Ctrl-Shift-U","Ctrl-Shift-U"),exec:function(e){e.toLowerCase()},multiSelectAction:"forEach",scrollIntoView:"cursor"},{name:"expandtoline",bindKey:s("Ctrl-Shift-L","Command-Shift-L"),exec:function(e){var t=e.selection.getRange();t.start.column=t.end.column=0,t.end.row++,e.selection.setRange(t,!1)},multiSelectAction:"forEach",scrollIntoView:"cursor",readOnly:!0},{name:"joinlines",bindKey:s(null,null),exec:function(e){for(var t=e.selection.isBackwards(),n=t?e.selection.getSelectionLead():e.selection.getSelectionAnchor(),r=t?e.selection.getSelectionAnchor():e.selection.getSelectionLead(),s=e.session.doc.getLine(n.row).length,a=e.session.doc.getTextRange(e.selection.getRange()).replace(/\n\s*/," ").length,l=e.session.doc.getLine(n.row),u=n.row+1;u<=r.row+1;u++){var c=i.stringTrimLeft(i.stringTrimRight(e.session.doc.getLine(u)));0!==c.length&&(c=" "+c),l+=c}r.row+1<e.session.doc.getLength()-1&&(l+=e.session.doc.getNewLineCharacter()),e.clearSelection(),e.session.doc.replace(new o(n.row,0,r.row+2,0),l),a>0?(e.selection.moveCursorTo(n.row,n.column),e.selection.selectTo(n.row,n.column+a)):(s=e.session.doc.getLine(n.row).length>s?s+1:s,e.selection.moveCursorTo(n.row,s))},multiSelectAction:"forEach",readOnly:!0},{name:"invertSelection",bindKey:s(null,null),exec:function(e){var t=e.session.doc.getLength()-1,n=e.session.doc.getLine(t).length,i=e.selection.rangeList.ranges,r=[];i.length<1&&(i=[e.selection.getRange()]);for(var s=0;s<i.length;s++)s==i.length-1&&(i[s].end.row===t&&i[s].end.column===n||r.push(new o(i[s].end.row,i[s].end.column,t,n))),0===s?0===i[s].start.row&&0===i[s].start.column||r.push(new o(0,0,i[s].start.row,i[s].start.column)):r.push(new o(i[s-1].end.row,i[s-1].end.column,i[s].start.row,i[s].start.column));e.exitMultiSelectMode(),e.clearSelection();for(s=0;s<r.length;s++)e.selection.addRange(r[s],!1)},readOnly:!0,scrollIntoView:"none"}]})),ace.define("ace/editor",["require","exports","module","ace/lib/fixoldbrowsers","ace/lib/oop","ace/lib/dom","ace/lib/lang","ace/lib/useragent","ace/keyboard/textinput","ace/mouse/mouse_handler","ace/mouse/fold_handler","ace/keyboard/keybinding","ace/edit_session","ace/search","ace/range","ace/lib/event_emitter","ace/commands/command_manager","ace/commands/default_commands","ace/config","ace/token_iterator"],(function(e,t,n){"use strict";e("./lib/fixoldbrowsers");var i=e("./lib/oop"),r=e("./lib/dom"),o=e("./lib/lang"),s=e("./lib/useragent"),a=e("./keyboard/textinput").TextInput,l=e("./mouse/mouse_handler").MouseHandler,u=e("./mouse/fold_handler").FoldHandler,c=e("./keyboard/keybinding").KeyBinding,h=e("./edit_session").EditSession,f=e("./search").Search,d=e("./range").Range,p=e("./lib/event_emitter").EventEmitter,g=e("./commands/command_manager").CommandManager,m=e("./commands/default_commands").commands,v=e("./config"),b=e("./token_iterator").TokenIterator,y=function(e,t){var n=e.getContainerElement();this.container=n,this.renderer=e,this.id="editor"+ ++y.$uid,this.commands=new g(s.isMac?"mac":"win",m),"object"==typeof document&&(this.textInput=new a(e.getTextAreaContainer(),this),this.renderer.textarea=this.textInput.getElement(),this.$mouseHandler=new l(this),new u(this)),this.keyBinding=new c(this),this.$blockScrolling=0,this.$search=(new f).set({wrap:!0}),this.$historyTracker=this.$historyTracker.bind(this),this.commands.on("exec",this.$historyTracker),this.$initOperationListeners(),this._$emitInputEvent=o.delayedCall(function(){this._signal("input",{}),this.session&&this.session.bgTokenizer&&this.session.bgTokenizer.scheduleStart()}.bind(this)),this.on("change",(function(e,t){t._$emitInputEvent.schedule(31)})),this.setSession(t||new h("")),v.resetOptions(this),v._signal("editor",this)};y.$uid=0,function(){i.implement(this,p),this.$initOperationListeners=function(){this.selections=[],this.commands.on("exec",this.startOperation.bind(this),!0),this.commands.on("afterExec",this.endOperation.bind(this),!0),this.$opResetTimer=o.delayedCall(this.endOperation.bind(this)),this.on("change",function(){this.curOp||this.startOperation(),this.curOp.docChanged=!0}.bind(this),!0),this.on("changeSelection",function(){this.curOp||this.startOperation(),this.curOp.selectionChanged=!0}.bind(this),!0)},this.curOp=null,this.prevOp={},this.startOperation=function(e){if(this.curOp){if(!e||this.curOp.command)return;this.prevOp=this.curOp}e||(this.previousCommand=null,e={}),this.$opResetTimer.schedule(),this.curOp={command:e.command||{},args:e.args,scrollTop:this.renderer.scrollTop},this.curOp.command.name&&void 0!==this.curOp.command.scrollIntoView&&this.$blockScrolling++},this.endOperation=function(e){if(this.curOp){if(e&&!1===e.returnValue)return this.curOp=null;this._signal("beforeEndOperation");var t=this.curOp.command;t.name&&this.$blockScrolling>0&&this.$blockScrolling--;var n=t&&t.scrollIntoView;if(n){switch(n){case"center-animate":n="animate";case"center":this.renderer.scrollCursorIntoView(null,.5);break;case"animate":case"cursor":this.renderer.scrollCursorIntoView();break;case"selectionPart":var i=this.selection.getRange(),r=this.renderer.layerConfig;(i.start.row>=r.lastRow||i.end.row<=r.firstRow)&&this.renderer.scrollSelectionIntoView(this.selection.anchor,this.selection.lead)}"animate"==n&&this.renderer.animateScrolling(this.curOp.scrollTop)}this.prevOp=this.curOp,this.curOp=null}},this.$mergeableCommands=["backspace","del","insertstring"],this.$historyTracker=function(e){if(this.$mergeUndoDeltas){var t=this.prevOp,n=this.$mergeableCommands,i=t.command&&e.command.name==t.command.name;if("insertstring"==e.command.name){var r=e.args;void 0===this.mergeNextCommand&&(this.mergeNextCommand=!0),i=i&&this.mergeNextCommand&&(!/\s/.test(r)||/\s/.test(t.args)),this.mergeNextCommand=!0}else i=i&&-1!==n.indexOf(e.command.name);"always"!=this.$mergeUndoDeltas&&Date.now()-this.sequenceStartTime>2e3&&(i=!1),i?this.session.mergeUndoDeltas=!0:-1!==n.indexOf(e.command.name)&&(this.sequenceStartTime=Date.now())}},this.setKeyboardHandler=function(e,t){if(e&&"string"==typeof e){this.$keybindingId=e;var n=this;v.loadModule(["keybinding",e],(function(i){n.$keybindingId==e&&n.keyBinding.setKeyboardHandler(i&&i.handler),t&&t()}))}else this.$keybindingId=null,this.keyBinding.setKeyboardHandler(e),t&&t()},this.getKeyboardHandler=function(){return this.keyBinding.getKeyboardHandler()},this.setSession=function(e){if(this.session!=e){this.curOp&&this.endOperation(),this.curOp={};var t=this.session;if(t){this.session.off("change",this.$onDocumentChange),this.session.off("changeMode",this.$onChangeMode),this.session.off("tokenizerUpdate",this.$onTokenizerUpdate),this.session.off("changeTabSize",this.$onChangeTabSize),this.session.off("changeWrapLimit",this.$onChangeWrapLimit),this.session.off("changeWrapMode",this.$onChangeWrapMode),this.session.off("changeFold",this.$onChangeFold),this.session.off("changeFrontMarker",this.$onChangeFrontMarker),this.session.off("changeBackMarker",this.$onChangeBackMarker),this.session.off("changeBreakpoint",this.$onChangeBreakpoint),this.session.off("changeAnnotation",this.$onChangeAnnotation),this.session.off("changeOverwrite",this.$onCursorChange),this.session.off("changeScrollTop",this.$onScrollTopChange),this.session.off("changeScrollLeft",this.$onScrollLeftChange);var n=this.session.getSelection();n.off("changeCursor",this.$onCursorChange),n.off("changeSelection",this.$onSelectionChange)}this.session=e,e?(this.$onDocumentChange=this.onDocumentChange.bind(this),e.on("change",this.$onDocumentChange),this.renderer.setSession(e),this.$onChangeMode=this.onChangeMode.bind(this),e.on("changeMode",this.$onChangeMode),this.$onTokenizerUpdate=this.onTokenizerUpdate.bind(this),e.on("tokenizerUpdate",this.$onTokenizerUpdate),this.$onChangeTabSize=this.renderer.onChangeTabSize.bind(this.renderer),e.on("changeTabSize",this.$onChangeTabSize),this.$onChangeWrapLimit=this.onChangeWrapLimit.bind(this),e.on("changeWrapLimit",this.$onChangeWrapLimit),this.$onChangeWrapMode=this.onChangeWrapMode.bind(this),e.on("changeWrapMode",this.$onChangeWrapMode),this.$onChangeFold=this.onChangeFold.bind(this),e.on("changeFold",this.$onChangeFold),this.$onChangeFrontMarker=this.onChangeFrontMarker.bind(this),this.session.on("changeFrontMarker",this.$onChangeFrontMarker),this.$onChangeBackMarker=this.onChangeBackMarker.bind(this),this.session.on("changeBackMarker",this.$onChangeBackMarker),this.$onChangeBreakpoint=this.onChangeBreakpoint.bind(this),this.session.on("changeBreakpoint",this.$onChangeBreakpoint),this.$onChangeAnnotation=this.onChangeAnnotation.bind(this),this.session.on("changeAnnotation",this.$onChangeAnnotation),this.$onCursorChange=this.onCursorChange.bind(this),this.session.on("changeOverwrite",this.$onCursorChange),this.$onScrollTopChange=this.onScrollTopChange.bind(this),this.session.on("changeScrollTop",this.$onScrollTopChange),this.$onScrollLeftChange=this.onScrollLeftChange.bind(this),this.session.on("changeScrollLeft",this.$onScrollLeftChange),this.selection=e.getSelection(),this.selection.on("changeCursor",this.$onCursorChange),this.$onSelectionChange=this.onSelectionChange.bind(this),this.selection.on("changeSelection",this.$onSelectionChange),this.onChangeMode(),this.$blockScrolling+=1,this.onCursorChange(),this.$blockScrolling-=1,this.onScrollTopChange(),this.onScrollLeftChange(),this.onSelectionChange(),this.onChangeFrontMarker(),this.onChangeBackMarker(),this.onChangeBreakpoint(),this.onChangeAnnotation(),this.session.getUseWrapMode()&&this.renderer.adjustWrapLimit(),this.renderer.updateFull()):(this.selection=null,this.renderer.setSession(e)),this._signal("changeSession",{session:e,oldSession:t}),this.curOp=null,t&&t._signal("changeEditor",{oldEditor:this}),e&&e._signal("changeEditor",{editor:this}),e&&e.bgTokenizer&&e.bgTokenizer.scheduleStart()}},this.getSession=function(){return this.session},this.setValue=function(e,t){return this.session.doc.setValue(e),t?1==t?this.navigateFileEnd():-1==t&&this.navigateFileStart():this.selectAll(),e},this.getValue=function(){return this.session.getValue()},this.getSelection=function(){return this.selection},this.resize=function(e){this.renderer.onResize(e)},this.setTheme=function(e,t){this.renderer.setTheme(e,t)},this.getTheme=function(){return this.renderer.getTheme()},this.setStyle=function(e){this.renderer.setStyle(e)},this.unsetStyle=function(e){this.renderer.unsetStyle(e)},this.getFontSize=function(){return this.getOption("fontSize")||r.computedStyle(this.container,"fontSize")},this.setFontSize=function(e){this.setOption("fontSize",e)},this.$highlightBrackets=function(){if(this.session.$bracketHighlight&&(this.session.removeMarker(this.session.$bracketHighlight),this.session.$bracketHighlight=null),!this.$highlightPending){var e=this;this.$highlightPending=!0,setTimeout((function(){e.$highlightPending=!1;var t=e.session;if(t&&t.bgTokenizer){var n=t.findMatchingBracket(e.getCursorPosition());if(n)var i=new d(n.row,n.column,n.row,n.column+1);else if(t.$mode.getMatching)i=t.$mode.getMatching(e.session);i&&(t.$bracketHighlight=t.addMarker(i,"ace_bracket","text"))}}),50)}},this.$highlightTags=function(){if(!this.$highlightTagPending){var e=this;this.$highlightTagPending=!0,setTimeout((function(){e.$highlightTagPending=!1;var t=e.session;if(t&&t.bgTokenizer){var n=e.getCursorPosition(),i=new b(e.session,n.row,n.column),r=i.getCurrentToken();if(!r||!/\b(?:tag-open|tag-name)/.test(r.type))return t.removeMarker(t.$tagHighlight),void(t.$tagHighlight=null);if(-1==r.type.indexOf("tag-open")||(r=i.stepForward())){var o=r.value,s=0,a=i.stepBackward();if("<"==a.value)do{a=r,(r=i.stepForward())&&r.value===o&&-1!==r.type.indexOf("tag-name")&&("<"===a.value?s++:"</"===a.value&&s--)}while(r&&s>=0);else{do{r=a,a=i.stepBackward(),r&&r.value===o&&-1!==r.type.indexOf("tag-name")&&("<"===a.value?s++:"</"===a.value&&s--)}while(a&&s<=0);i.stepForward()}if(!r)return t.removeMarker(t.$tagHighlight),void(t.$tagHighlight=null);var l=i.getCurrentTokenRow(),u=i.getCurrentTokenColumn(),c=new d(l,u,l,u+r.value.length),h=t.$backMarkers[t.$tagHighlight];t.$tagHighlight&&null!=h&&0!==c.compareRange(h.range)&&(t.removeMarker(t.$tagHighlight),t.$tagHighlight=null),c&&!t.$tagHighlight&&(t.$tagHighlight=t.addMarker(c,"ace_bracket","text"))}}}),50)}},this.focus=function(){var e=this;setTimeout((function(){e.textInput.focus()})),this.textInput.focus()},this.isFocused=function(){return this.textInput.isFocused()},this.blur=function(){this.textInput.blur()},this.onFocus=function(e){this.$isFocused||(this.$isFocused=!0,this.renderer.showCursor(),this.renderer.visualizeFocus(),this._emit("focus",e))},this.onBlur=function(e){this.$isFocused&&(this.$isFocused=!1,this.renderer.hideCursor(),this.renderer.visualizeBlur(),this._emit("blur",e))},this.$cursorChange=function(){this.renderer.updateCursor()},this.onDocumentChange=function(e){var t=this.session.$useWrapMode,n=e.start.row==e.end.row?e.end.row:1/0;this.renderer.updateLines(e.start.row,n,t),this._signal("change",e),this.$cursorChange(),this.$updateHighlightActiveLine()},this.onTokenizerUpdate=function(e){var t=e.data;this.renderer.updateLines(t.first,t.last)},this.onScrollTopChange=function(){this.renderer.scrollToY(this.session.getScrollTop())},this.onScrollLeftChange=function(){this.renderer.scrollToX(this.session.getScrollLeft())},this.onCursorChange=function(){this.$cursorChange(),this.$blockScrolling||(v.warn("Automatically scrolling cursor into view after selection change","this will be disabled in the next version","set editor.$blockScrolling = Infinity to disable this message"),this.renderer.scrollCursorIntoView()),this.$highlightBrackets(),this.$highlightTags(),this.$updateHighlightActiveLine(),this._signal("changeSelection")},this.$updateHighlightActiveLine=function(){var e,t=this.getSession();if(this.$highlightActiveLine&&("line"==this.$selectionStyle&&this.selection.isMultiLine()||(e=this.getCursorPosition()),!this.renderer.$maxLines||1!==this.session.getLength()||this.renderer.$minLines>1||(e=!1)),t.$highlightLineMarker&&!e)t.removeMarker(t.$highlightLineMarker.id),t.$highlightLineMarker=null;else if(!t.$highlightLineMarker&&e){var n=new d(e.row,e.column,e.row,1/0);n.id=t.addMarker(n,"ace_active-line","screenLine"),t.$highlightLineMarker=n}else e&&(t.$highlightLineMarker.start.row=e.row,t.$highlightLineMarker.end.row=e.row,t.$highlightLineMarker.start.column=e.column,t._signal("changeBackMarker"))},this.onSelectionChange=function(e){var t=this.session;if(t.$selectionMarker&&t.removeMarker(t.$selectionMarker),t.$selectionMarker=null,this.selection.isEmpty())this.$updateHighlightActiveLine();else{var n=this.selection.getRange(),i=this.getSelectionStyle();t.$selectionMarker=t.addMarker(n,"ace_selection",i)}var r=this.$highlightSelectedWord&&this.$getSelectionHighLightRegexp();this.session.highlight(r),this._signal("changeSelection")},this.$getSelectionHighLightRegexp=function(){var e=this.session,t=this.getSelectionRange();if(!t.isEmpty()&&!t.isMultiLine()){var n=t.start.column-1,i=t.end.column+1,r=e.getLine(t.start.row),o=r.length,s=r.substring(Math.max(n,0),Math.min(i,o));if(!(n>=0&&/^[\w\d]/.test(s)||i<=o&&/[\w\d]$/.test(s)))if(s=r.substring(t.start.column,t.end.column),/^[\w\d]+$/.test(s))return this.$search.$assembleRegExp({wholeWord:!0,caseSensitive:!0,needle:s})}},this.onChangeFrontMarker=function(){this.renderer.updateFrontMarkers()},this.onChangeBackMarker=function(){this.renderer.updateBackMarkers()},this.onChangeBreakpoint=function(){this.renderer.updateBreakpoints()},this.onChangeAnnotation=function(){this.renderer.setAnnotations(this.session.getAnnotations())},this.onChangeMode=function(e){this.renderer.updateText(),this._emit("changeMode",e)},this.onChangeWrapLimit=function(){this.renderer.updateFull()},this.onChangeWrapMode=function(){this.renderer.onResize(!0)},this.onChangeFold=function(){this.$updateHighlightActiveLine(),this.renderer.updateFull()},this.getSelectedText=function(){return this.session.getTextRange(this.getSelectionRange())},this.getCopyText=function(){var e=this.getSelectedText();return this._signal("copy",e),e},this.onCopy=function(){this.commands.exec("copy",this)},this.onCut=function(){this.commands.exec("cut",this)},this.onPaste=function(e,t){var n={text:e,event:t};this.commands.exec("paste",this,n)},this.$handlePaste=function(e){"string"==typeof e&&(e={text:e}),this._signal("paste",e);var t=e.text;if(!this.inMultiSelectMode||this.inVirtualSelectionMode)this.insert(t);else{var n=t.split(/\r\n|\r|\n/),i=this.selection.rangeList.ranges;if(n.length>i.length||n.length<2||!n[1])return this.commands.exec("insertstring",this,t);for(var r=i.length;r--;){var o=i[r];o.isEmpty()||this.session.remove(o),this.session.insert(o.start,n[r])}}},this.execCommand=function(e,t){return this.commands.exec(e,this,t)},this.insert=function(e,t){var n=this.session,i=n.getMode(),r=this.getCursorPosition();if(this.getBehavioursEnabled()&&!t){var o=i.transformAction(n.getState(r.row),"insertion",this,n,e);o&&(e!==o.text&&(this.session.mergeUndoDeltas=!1,this.$mergeNextCommand=!1),e=o.text)}if("\t"==e&&(e=this.session.getTabString()),this.selection.isEmpty()){if(this.session.getOverwrite()&&-1==e.indexOf("\n")){(s=new d.fromPoints(r,r)).end.column+=e.length,this.session.remove(s)}}else{var s=this.getSelectionRange();r=this.session.remove(s),this.clearSelection()}if("\n"==e||"\r\n"==e){var a=n.getLine(r.row);if(r.column>a.search(/\S|$/)){var l=a.substr(r.column).search(/\S|$/);n.doc.removeInLine(r.row,r.column,r.column+l)}}this.clearSelection();var u=r.column,c=n.getState(r.row),h=(a=n.getLine(r.row),i.checkOutdent(c,a,e));n.insert(r,e);if(o&&o.selection&&(2==o.selection.length?this.selection.setSelectionRange(new d(r.row,u+o.selection[0],r.row,u+o.selection[1])):this.selection.setSelectionRange(new d(r.row+o.selection[0],o.selection[1],r.row+o.selection[2],o.selection[3]))),n.getDocument().isNewLine(e)){var f=i.getNextLineIndent(c,a.slice(0,r.column),n.getTabString());n.insert({row:r.row+1,column:0},f)}h&&i.autoOutdent(c,n,r.row)},this.onTextInput=function(e){this.keyBinding.onTextInput(e)},this.onCommandKey=function(e,t,n){this.keyBinding.onCommandKey(e,t,n)},this.setOverwrite=function(e){this.session.setOverwrite(e)},this.getOverwrite=function(){return this.session.getOverwrite()},this.toggleOverwrite=function(){this.session.toggleOverwrite()},this.setScrollSpeed=function(e){this.setOption("scrollSpeed",e)},this.getScrollSpeed=function(){return this.getOption("scrollSpeed")},this.setDragDelay=function(e){this.setOption("dragDelay",e)},this.getDragDelay=function(){return this.getOption("dragDelay")},this.setSelectionStyle=function(e){this.setOption("selectionStyle",e)},this.getSelectionStyle=function(){return this.getOption("selectionStyle")},this.setHighlightActiveLine=function(e){this.setOption("highlightActiveLine",e)},this.getHighlightActiveLine=function(){return this.getOption("highlightActiveLine")},this.setHighlightGutterLine=function(e){this.setOption("highlightGutterLine",e)},this.getHighlightGutterLine=function(){return this.getOption("highlightGutterLine")},this.setHighlightSelectedWord=function(e){this.setOption("highlightSelectedWord",e)},this.getHighlightSelectedWord=function(){return this.$highlightSelectedWord},this.setAnimatedScroll=function(e){this.renderer.setAnimatedScroll(e)},this.getAnimatedScroll=function(){return this.renderer.getAnimatedScroll()},this.setShowInvisibles=function(e){this.renderer.setShowInvisibles(e)},this.getShowInvisibles=function(){return this.renderer.getShowInvisibles()},this.setDisplayIndentGuides=function(e){this.renderer.setDisplayIndentGuides(e)},this.getDisplayIndentGuides=function(){return this.renderer.getDisplayIndentGuides()},this.setShowPrintMargin=function(e){this.renderer.setShowPrintMargin(e)},this.getShowPrintMargin=function(){return this.renderer.getShowPrintMargin()},this.setPrintMarginColumn=function(e){this.renderer.setPrintMarginColumn(e)},this.getPrintMarginColumn=function(){return this.renderer.getPrintMarginColumn()},this.setReadOnly=function(e){this.setOption("readOnly",e)},this.getReadOnly=function(){return this.getOption("readOnly")},this.setBehavioursEnabled=function(e){this.setOption("behavioursEnabled",e)},this.getBehavioursEnabled=function(){return this.getOption("behavioursEnabled")},this.setWrapBehavioursEnabled=function(e){this.setOption("wrapBehavioursEnabled",e)},this.getWrapBehavioursEnabled=function(){return this.getOption("wrapBehavioursEnabled")},this.setShowFoldWidgets=function(e){this.setOption("showFoldWidgets",e)},this.getShowFoldWidgets=function(){return this.getOption("showFoldWidgets")},this.setFadeFoldWidgets=function(e){this.setOption("fadeFoldWidgets",e)},this.getFadeFoldWidgets=function(){return this.getOption("fadeFoldWidgets")},this.remove=function(e){this.selection.isEmpty()&&("left"==e?this.selection.selectLeft():this.selection.selectRight());var t=this.getSelectionRange();if(this.getBehavioursEnabled()){var n=this.session,i=n.getState(t.start.row),r=n.getMode().transformAction(i,"deletion",this,n,t);if(0===t.end.column){var o=n.getTextRange(t);if("\n"==o[o.length-1]){var s=n.getLine(t.end.row);/^\s+$/.test(s)&&(t.end.column=s.length)}}r&&(t=r)}this.session.remove(t),this.clearSelection()},this.removeWordRight=function(){this.selection.isEmpty()&&this.selection.selectWordRight(),this.session.remove(this.getSelectionRange()),this.clearSelection()},this.removeWordLeft=function(){this.selection.isEmpty()&&this.selection.selectWordLeft(),this.session.remove(this.getSelectionRange()),this.clearSelection()},this.removeToLineStart=function(){this.selection.isEmpty()&&this.selection.selectLineStart(),this.session.remove(this.getSelectionRange()),this.clearSelection()},this.removeToLineEnd=function(){this.selection.isEmpty()&&this.selection.selectLineEnd();var e=this.getSelectionRange();e.start.column==e.end.column&&e.start.row==e.end.row&&(e.end.column=0,e.end.row++),this.session.remove(e),this.clearSelection()},this.splitLine=function(){this.selection.isEmpty()||(this.session.remove(this.getSelectionRange()),this.clearSelection());var e=this.getCursorPosition();this.insert("\n"),this.moveCursorToPosition(e)},this.transposeLetters=function(){if(this.selection.isEmpty()){var e=this.getCursorPosition(),t=e.column;if(0!==t){var n,i,r=this.session.getLine(e.row);t<r.length?(n=r.charAt(t)+r.charAt(t-1),i=new d(e.row,t-1,e.row,t+1)):(n=r.charAt(t-1)+r.charAt(t-2),i=new d(e.row,t-2,e.row,t)),this.session.replace(i,n),this.session.selection.moveToPosition(i.end)}}},this.toLowerCase=function(){var e=this.getSelectionRange();this.selection.isEmpty()&&this.selection.selectWord();var t=this.getSelectionRange(),n=this.session.getTextRange(t);this.session.replace(t,n.toLowerCase()),this.selection.setSelectionRange(e)},this.toUpperCase=function(){var e=this.getSelectionRange();this.selection.isEmpty()&&this.selection.selectWord();var t=this.getSelectionRange(),n=this.session.getTextRange(t);this.session.replace(t,n.toUpperCase()),this.selection.setSelectionRange(e)},this.indent=function(){var e=this.session,t=this.getSelectionRange();if(!(t.start.row<t.end.row)){if(t.start.column<t.end.column){var n=e.getTextRange(t);if(!/^\s+$/.test(n)){c=this.$getSelectedRows();return void e.indentRows(c.first,c.last,"\t")}}var i=e.getLine(t.start.row),r=t.start,s=e.getTabSize(),a=e.documentToScreenColumn(r.row,r.column);if(this.session.getUseSoftTabs())var l=s-a%s,u=o.stringRepeat(" ",l);else{for(l=a%s;" "==i[t.start.column-1]&&l;)t.start.column--,l--;this.selection.setSelectionRange(t),u="\t"}return this.insert(u)}var c=this.$getSelectedRows();e.indentRows(c.first,c.last,"\t")},this.blockIndent=function(){var e=this.$getSelectedRows();this.session.indentRows(e.first,e.last,"\t")},this.blockOutdent=function(){var e=this.session.getSelection();this.session.outdentRows(e.getRange())},this.sortLines=function(){for(var e=this.$getSelectedRows(),t=this.session,n=[],i=e.first;i<=e.last;i++)n.push(t.getLine(i));n.sort((function(e,t){return e.toLowerCase()<t.toLowerCase()?-1:e.toLowerCase()>t.toLowerCase()?1:0}));var r=new d(0,0,0,0);for(i=e.first;i<=e.last;i++){var o=t.getLine(i);r.start.row=i,r.end.row=i,r.end.column=o.length,t.replace(r,n[i-e.first])}},this.toggleCommentLines=function(){var e=this.session.getState(this.getCursorPosition().row),t=this.$getSelectedRows();this.session.getMode().toggleCommentLines(e,this.session,t.first,t.last)},this.toggleBlockComment=function(){var e=this.getCursorPosition(),t=this.session.getState(e.row),n=this.getSelectionRange();this.session.getMode().toggleBlockComment(t,this.session,n,e)},this.getNumberAt=function(e,t){var n=/[\-]?[0-9]+(?:\.[0-9]+)?/g;n.lastIndex=0;for(var i=this.session.getLine(e);n.lastIndex<t;){var r=n.exec(i);if(r.index<=t&&r.index+r[0].length>=t)return{value:r[0],start:r.index,end:r.index+r[0].length}}return null},this.modifyNumber=function(e){var t=this.selection.getCursor().row,n=this.selection.getCursor().column,i=new d(t,n-1,t,n),r=this.session.getTextRange(i);if(!isNaN(parseFloat(r))&&isFinite(r)){var o=this.getNumberAt(t,n);if(o){var s=o.value.indexOf(".")>=0?o.start+o.value.indexOf(".")+1:o.end,a=o.start+o.value.length-s,l=parseFloat(o.value);l*=Math.pow(10,a),s!==o.end&&n<s?e*=Math.pow(10,o.end-n-1):e*=Math.pow(10,o.end-n),l+=e;var u=(l/=Math.pow(10,a)).toFixed(a),c=new d(t,o.start,t,o.end);this.session.replace(c,u),this.moveCursorTo(t,Math.max(o.start+1,n+u.length-o.value.length))}}},this.removeLines=function(){var e=this.$getSelectedRows();this.session.removeFullLines(e.first,e.last),this.clearSelection()},this.duplicateSelection=function(){var e=this.selection,t=this.session,n=e.getRange(),i=e.isBackwards();if(n.isEmpty()){var r=n.start.row;t.duplicateLines(r,r)}else{var o=i?n.start:n.end,s=t.insert(o,t.getTextRange(n),!1);n.start=o,n.end=s,e.setSelectionRange(n,i)}},this.moveLinesDown=function(){this.$moveLines(1,!1)},this.moveLinesUp=function(){this.$moveLines(-1,!1)},this.moveText=function(e,t,n){return this.session.moveText(e,t,n)},this.copyLinesUp=function(){this.$moveLines(-1,!0)},this.copyLinesDown=function(){this.$moveLines(1,!0)},this.$moveLines=function(e,t){var n,i,r=this.selection;if(!r.inMultiSelectMode||this.inVirtualSelectionMode){var o=r.toOrientedRange();n=this.$getSelectedRows(o),i=this.session.$moveLines(n.first,n.last,t?0:e),t&&-1==e&&(i=0),o.moveBy(i,0),r.fromOrientedRange(o)}else{var s=r.rangeList.ranges;r.rangeList.detach(this.session),this.inVirtualSelectionMode=!0;for(var a=0,l=0,u=s.length,c=0;c<u;c++){var h=c;s[c].moveBy(a,0);for(var f=(n=this.$getSelectedRows(s[c])).first,d=n.last;++c<u;){l&&s[c].moveBy(l,0);var p=this.$getSelectedRows(s[c]);if(t&&p.first!=d)break;if(!t&&p.first>d+1)break;d=p.last}for(c--,a=this.session.$moveLines(f,d,t?0:e),t&&-1==e&&(h=c+1);h<=c;)s[h].moveBy(a,0),h++;t||(a=0),l+=a}r.fromOrientedRange(r.ranges[0]),r.rangeList.attach(this.session),this.inVirtualSelectionMode=!1}},this.$getSelectedRows=function(e){return e=(e||this.getSelectionRange()).collapseRows(),{first:this.session.getRowFoldStart(e.start.row),last:this.session.getRowFoldEnd(e.end.row)}},this.onCompositionStart=function(e){this.renderer.showComposition(this.getCursorPosition())},this.onCompositionUpdate=function(e){this.renderer.setCompositionText(e)},this.onCompositionEnd=function(){this.renderer.hideComposition()},this.getFirstVisibleRow=function(){return this.renderer.getFirstVisibleRow()},this.getLastVisibleRow=function(){return this.renderer.getLastVisibleRow()},this.isRowVisible=function(e){return e>=this.getFirstVisibleRow()&&e<=this.getLastVisibleRow()},this.isRowFullyVisible=function(e){return e>=this.renderer.getFirstFullyVisibleRow()&&e<=this.renderer.getLastFullyVisibleRow()},this.$getVisibleRowCount=function(){return this.renderer.getScrollBottomRow()-this.renderer.getScrollTopRow()+1},this.$moveByPage=function(e,t){var n=this.renderer,i=this.renderer.layerConfig,r=e*Math.floor(i.height/i.lineHeight);this.$blockScrolling++,!0===t?this.selection.$moveSelection((function(){this.moveCursorBy(r,0)})):!1===t&&(this.selection.moveCursorBy(r,0),this.selection.clearSelection()),this.$blockScrolling--;var o=n.scrollTop;n.scrollBy(0,r*i.lineHeight),null!=t&&n.scrollCursorIntoView(null,.5),n.animateScrolling(o)},this.selectPageDown=function(){this.$moveByPage(1,!0)},this.selectPageUp=function(){this.$moveByPage(-1,!0)},this.gotoPageDown=function(){this.$moveByPage(1,!1)},this.gotoPageUp=function(){this.$moveByPage(-1,!1)},this.scrollPageDown=function(){this.$moveByPage(1)},this.scrollPageUp=function(){this.$moveByPage(-1)},this.scrollToRow=function(e){this.renderer.scrollToRow(e)},this.scrollToLine=function(e,t,n,i){this.renderer.scrollToLine(e,t,n,i)},this.centerSelection=function(){var e=this.getSelectionRange(),t={row:Math.floor(e.start.row+(e.end.row-e.start.row)/2),column:Math.floor(e.start.column+(e.end.column-e.start.column)/2)};this.renderer.alignCursor(t,.5)},this.getCursorPosition=function(){return this.selection.getCursor()},this.getCursorPositionScreen=function(){return this.session.documentToScreenPosition(this.getCursorPosition())},this.getSelectionRange=function(){return this.selection.getRange()},this.selectAll=function(){this.$blockScrolling+=1,this.selection.selectAll(),this.$blockScrolling-=1},this.clearSelection=function(){this.selection.clearSelection()},this.moveCursorTo=function(e,t){this.selection.moveCursorTo(e,t)},this.moveCursorToPosition=function(e){this.selection.moveCursorToPosition(e)},this.jumpToMatching=function(e,t){var n=this.getCursorPosition(),i=new b(this.session,n.row,n.column),r=i.getCurrentToken(),o=r||i.stepForward();if(o){var s,a,l=!1,u={},c=n.column-o.start,h={")":"(","(":"(","]":"[","[":"[","{":"{","}":"{"};do{if(o.value.match(/[{}()\[\]]/g)){for(;c<o.value.length&&!l;c++)if(h[o.value[c]])switch(a=h[o.value[c]]+"."+o.type.replace("rparen","lparen"),isNaN(u[a])&&(u[a]=0),o.value[c]){case"(":case"[":case"{":u[a]++;break;case")":case"]":case"}":u[a]--,-1===u[a]&&(s="bracket",l=!0)}}else o&&-1!==o.type.indexOf("tag-name")&&(isNaN(u[o.value])&&(u[o.value]=0),"<"===r.value?u[o.value]++:"</"===r.value&&u[o.value]--,-1===u[o.value]&&(s="tag",l=!0));l||(r=o,o=i.stepForward(),c=0)}while(o&&!l);if(s){var f,p;if("bracket"===s)(f=this.session.getBracketRange(n))||(p=(f=new d(i.getCurrentTokenRow(),i.getCurrentTokenColumn()+c-1,i.getCurrentTokenRow(),i.getCurrentTokenColumn()+c-1)).start,(t||p.row===n.row&&Math.abs(p.column-n.column)<2)&&(f=this.session.getBracketRange(p)));else if("tag"===s){if(!o||-1===o.type.indexOf("tag-name"))return;var g=o.value;if(0===(f=new d(i.getCurrentTokenRow(),i.getCurrentTokenColumn()-2,i.getCurrentTokenRow(),i.getCurrentTokenColumn()-2)).compare(n.row,n.column)){l=!1;do{o=r,(r=i.stepBackward())&&(-1!==r.type.indexOf("tag-close")&&f.setEnd(i.getCurrentTokenRow(),i.getCurrentTokenColumn()+1),o.value===g&&-1!==o.type.indexOf("tag-name")&&("<"===r.value?u[g]++:"</"===r.value&&u[g]--,0===u[g]&&(l=!0)))}while(r&&!l)}o&&o.type.indexOf("tag-name")&&(p=f.start).row==n.row&&Math.abs(p.column-n.column)<2&&(p=f.end)}(p=f&&f.cursor||p)&&(e?f&&t?this.selection.setRange(f):f&&f.isEqual(this.getSelectionRange())?this.clearSelection():this.selection.selectTo(p.row,p.column):this.selection.moveTo(p.row,p.column))}}},this.gotoLine=function(e,t,n){this.selection.clearSelection(),this.session.unfold({row:e-1,column:t||0}),this.$blockScrolling+=1,this.exitMultiSelectMode&&this.exitMultiSelectMode(),this.moveCursorTo(e-1,t||0),this.$blockScrolling-=1,this.isRowFullyVisible(e-1)||this.scrollToLine(e-1,!0,n)},this.navigateTo=function(e,t){this.selection.moveTo(e,t)},this.navigateUp=function(e){if(this.selection.isMultiLine()&&!this.selection.isBackwards()){var t=this.selection.anchor.getPosition();return this.moveCursorToPosition(t)}this.selection.clearSelection(),this.selection.moveCursorBy(-e||-1,0)},this.navigateDown=function(e){if(this.selection.isMultiLine()&&this.selection.isBackwards()){var t=this.selection.anchor.getPosition();return this.moveCursorToPosition(t)}this.selection.clearSelection(),this.selection.moveCursorBy(e||1,0)},this.navigateLeft=function(e){if(this.selection.isEmpty())for(e=e||1;e--;)this.selection.moveCursorLeft();else{var t=this.getSelectionRange().start;this.moveCursorToPosition(t)}this.clearSelection()},this.navigateRight=function(e){if(this.selection.isEmpty())for(e=e||1;e--;)this.selection.moveCursorRight();else{var t=this.getSelectionRange().end;this.moveCursorToPosition(t)}this.clearSelection()},this.navigateLineStart=function(){this.selection.moveCursorLineStart(),this.clearSelection()},this.navigateLineEnd=function(){this.selection.moveCursorLineEnd(),this.clearSelection()},this.navigateFileEnd=function(){this.selection.moveCursorFileEnd(),this.clearSelection()},this.navigateFileStart=function(){this.selection.moveCursorFileStart(),this.clearSelection()},this.navigateWordRight=function(){this.selection.moveCursorWordRight(),this.clearSelection()},this.navigateWordLeft=function(){this.selection.moveCursorWordLeft(),this.clearSelection()},this.replace=function(e,t){t&&this.$search.set(t);var n=this.$search.find(this.session),i=0;return n?(this.$tryReplace(n,e)&&(i=1),null!==n&&(this.selection.setSelectionRange(n),this.renderer.scrollSelectionIntoView(n.start,n.end)),i):i},this.replaceAll=function(e,t){t&&this.$search.set(t);var n=this.$search.findAll(this.session),i=0;if(!n.length)return i;this.$blockScrolling+=1;var r=this.getSelectionRange();this.selection.moveTo(0,0);for(var o=n.length-1;o>=0;--o)this.$tryReplace(n[o],e)&&i++;return this.selection.setSelectionRange(r),this.$blockScrolling-=1,i},this.$tryReplace=function(e,t){var n=this.session.getTextRange(e);return null!==(t=this.$search.replace(n,t))?(e.end=this.session.replace(e,t),e):null},this.getLastSearchOptions=function(){return this.$search.getOptions()},this.find=function(e,t,n){t||(t={}),"string"==typeof e||e instanceof RegExp?t.needle=e:"object"==typeof e&&i.mixin(t,e);var r=this.selection.getRange();null==t.needle&&((e=this.session.getTextRange(r)||this.$search.$options.needle)||(r=this.session.getWordRange(r.start.row,r.start.column),e=this.session.getTextRange(r)),this.$search.set({needle:e})),this.$search.set(t),t.start||this.$search.set({start:r});var o=this.$search.find(this.session);return t.preventScroll?o:o?(this.revealRange(o,n),o):(t.backwards?r.start=r.end:r.end=r.start,void this.selection.setRange(r))},this.findNext=function(e,t){this.find({skipCurrent:!0,backwards:!1},e,t)},this.findPrevious=function(e,t){this.find(e,{skipCurrent:!0,backwards:!0},t)},this.revealRange=function(e,t){this.$blockScrolling+=1,this.session.unfold(e),this.selection.setSelectionRange(e),this.$blockScrolling-=1;var n=this.renderer.scrollTop;this.renderer.scrollSelectionIntoView(e.start,e.end,.5),!1!==t&&this.renderer.animateScrolling(n)},this.undo=function(){this.$blockScrolling++,this.session.getUndoManager().undo(),this.$blockScrolling--,this.renderer.scrollCursorIntoView(null,.5)},this.redo=function(){this.$blockScrolling++,this.session.getUndoManager().redo(),this.$blockScrolling--,this.renderer.scrollCursorIntoView(null,.5)},this.destroy=function(){this.renderer.destroy(),this._signal("destroy",this),this.session&&this.session.destroy()},this.setAutoScrollEditorIntoView=function(e){if(e){var t,n=this,i=!1;this.$scrollAnchor||(this.$scrollAnchor=document.createElement("div"));var r=this.$scrollAnchor;r.style.cssText="position:absolute",this.container.insertBefore(r,this.container.firstChild);var o=this.on("changeSelection",(function(){i=!0})),s=this.renderer.on("beforeRender",(function(){i&&(t=n.renderer.container.getBoundingClientRect())})),a=this.renderer.on("afterRender",(function(){if(i&&t&&(n.isFocused()||n.searchBox&&n.searchBox.isFocused())){var e=n.renderer,o=e.$cursorLayer.$pixelPos,s=e.layerConfig,a=o.top-s.offset;null!=(i=o.top>=0&&a+t.top<0||!(o.top<s.height&&o.top+t.top+s.lineHeight>window.innerHeight)&&null)&&(r.style.top=a+"px",r.style.left=o.left+"px",r.style.height=s.lineHeight+"px",r.scrollIntoView(i)),i=t=null}}));this.setAutoScrollEditorIntoView=function(e){e||(delete this.setAutoScrollEditorIntoView,this.off("changeSelection",o),this.renderer.off("afterRender",a),this.renderer.off("beforeRender",s))}}},this.$resetCursorStyle=function(){var e=this.$cursorStyle||"ace",t=this.renderer.$cursorLayer;t&&(t.setSmoothBlinking(/smooth/.test(e)),t.isBlinking=!this.$readOnly&&"wide"!=e,r.setCssClass(t.element,"ace_slim-cursors",/slim/.test(e)))}}.call(y.prototype),v.defineOptions(y.prototype,"editor",{selectionStyle:{set:function(e){this.onSelectionChange(),this._signal("changeSelectionStyle",{data:e})},initialValue:"line"},highlightActiveLine:{set:function(){this.$updateHighlightActiveLine()},initialValue:!0},highlightSelectedWord:{set:function(e){this.$onSelectionChange()},initialValue:!0},readOnly:{set:function(e){this.$resetCursorStyle()},initialValue:!1},cursorStyle:{set:function(e){this.$resetCursorStyle()},values:["ace","slim","smooth","wide"],initialValue:"ace"},mergeUndoDeltas:{values:[!1,!0,"always"],initialValue:!0},behavioursEnabled:{initialValue:!0},wrapBehavioursEnabled:{initialValue:!0},autoScrollEditorIntoView:{set:function(e){this.setAutoScrollEditorIntoView(e)}},keyboardHandler:{set:function(e){this.setKeyboardHandler(e)},get:function(){return this.keybindingId},handlesSet:!0},hScrollBarAlwaysVisible:"renderer",vScrollBarAlwaysVisible:"renderer",highlightGutterLine:"renderer",animatedScroll:"renderer",showInvisibles:"renderer",showPrintMargin:"renderer",printMarginColumn:"renderer",printMargin:"renderer",fadeFoldWidgets:"renderer",showFoldWidgets:"renderer",showLineNumbers:"renderer",showGutter:"renderer",displayIndentGuides:"renderer",fontSize:"renderer",fontFamily:"renderer",maxLines:"renderer",minLines:"renderer",scrollPastEnd:"renderer",fixedWidthGutter:"renderer",theme:"renderer",scrollSpeed:"$mouseHandler",dragDelay:"$mouseHandler",dragEnabled:"$mouseHandler",focusTimout:"$mouseHandler",tooltipFollowsMouse:"$mouseHandler",firstLineNumber:"session",overwrite:"session",newLineMode:"session",useWorker:"session",useSoftTabs:"session",tabSize:"session",wrap:"session",indentedSoftWrap:"session",foldStyle:"session",mode:"session"}),t.Editor=y})),ace.define("ace/undomanager",["require","exports","module"],(function(e,t,n){"use strict";var i=function(){this.reset()};(function(){function e(e){return{action:e.action,start:e.start,end:e.end,lines:1==e.lines.length?null:e.lines,text:1==e.lines.length?e.lines[0]:null}}function t(e){return{action:e.action,start:e.start,end:e.end,lines:e.lines||[e.text]}}function n(e,t){for(var n=new Array(e.length),i=0;i<e.length;i++){for(var r=e[i],o={group:r.group,deltas:new Array(r.length)},s=0;s<r.deltas.length;s++){var a=r.deltas[s];o.deltas[s]=t(a)}n[i]=o}return n}this.execute=function(e){var t=e.args[0];this.$doc=e.args[1],e.merge&&this.hasUndo()&&(this.dirtyCounter--,t=this.$undoStack.pop().concat(t)),this.$undoStack.push(t),this.$redoStack=[],this.dirtyCounter<0&&(this.dirtyCounter=NaN),this.dirtyCounter++},this.undo=function(e){var t=this.$undoStack.pop(),n=null;return t&&(n=this.$doc.undoChanges(t,e),this.$redoStack.push(t),this.dirtyCounter--),n},this.redo=function(e){var t=this.$redoStack.pop(),n=null;return t&&(n=this.$doc.redoChanges(this.$deserializeDeltas(t),e),this.$undoStack.push(t),this.dirtyCounter++),n},this.reset=function(){this.$undoStack=[],this.$redoStack=[],this.dirtyCounter=0},this.hasUndo=function(){return this.$undoStack.length>0},this.hasRedo=function(){return this.$redoStack.length>0},this.markClean=function(){this.dirtyCounter=0},this.isClean=function(){return 0===this.dirtyCounter},this.$serializeDeltas=function(t){return n(t,e)},this.$deserializeDeltas=function(e){return n(e,t)}}).call(i.prototype),t.UndoManager=i})),ace.define("ace/layer/gutter",["require","exports","module","ace/lib/dom","ace/lib/oop","ace/lib/lang","ace/lib/event_emitter"],(function(e,t,n){"use strict";var i=e("../lib/dom"),r=e("../lib/oop"),o=e("../lib/lang"),s=e("../lib/event_emitter").EventEmitter,a=function(e){this.element=i.createElement("div"),this.element.className="ace_layer ace_gutter-layer",e.appendChild(this.element),this.setShowFoldWidgets(this.$showFoldWidgets),this.gutterWidth=0,this.$annotations=[],this.$updateAnnotations=this.$updateAnnotations.bind(this),this.$cells=[]};(function(){r.implement(this,s),this.setSession=function(e){this.session&&this.session.removeEventListener("change",this.$updateAnnotations),this.session=e,e&&e.on("change",this.$updateAnnotations)},this.addGutterDecoration=function(e,t){window.console&&console.warn&&console.warn("deprecated use session.addGutterDecoration"),this.session.addGutterDecoration(e,t)},this.removeGutterDecoration=function(e,t){window.console&&console.warn&&console.warn("deprecated use session.removeGutterDecoration"),this.session.removeGutterDecoration(e,t)},this.setAnnotations=function(e){this.$annotations=[];for(var t=0;t<e.length;t++){var n=e[t],i=n.row,r=this.$annotations[i];r||(r=this.$annotations[i]={text:[]});var s=n.text;s=s?o.escapeHTML(s):n.html||"",-1===r.text.indexOf(s)&&r.text.push(s);var a=n.type;"error"==a?r.className=" ace_error":"warning"==a&&" ace_error"!=r.className?r.className=" ace_warning":"info"!=a||r.className||(r.className=" ace_info")}},this.$updateAnnotations=function(e){if(this.$annotations.length){var t=e.start.row,n=e.end.row-t;if(0===n);else if("remove"==e.action)this.$annotations.splice(t,n+1,null);else{var i=new Array(n+1);i.unshift(t,1),this.$annotations.splice.apply(this.$annotations,i)}}},this.update=function(e){for(var t=this.session,n=e.firstRow,r=Math.min(e.lastRow+e.gutterOffset,t.getLength()-1),o=t.getNextFoldLine(n),s=o?o.start.row:1/0,a=this.$showFoldWidgets&&t.foldWidgets,l=t.$breakpoints,u=t.$decorations,c=t.$firstLineNumber,h=0,f=t.gutterRenderer||this.$renderer,d=null,p=-1,g=n;;){if(g>s&&(g=o.end.row+1,s=(o=t.getNextFoldLine(g,o))?o.start.row:1/0),g>r){for(;this.$cells.length>p+1;)d=this.$cells.pop(),this.element.removeChild(d.element);break}(d=this.$cells[++p])||((d={element:null,textNode:null,foldWidget:null}).element=i.createElement("div"),d.textNode=document.createTextNode(""),d.element.appendChild(d.textNode),this.element.appendChild(d.element),this.$cells[p]=d);var m="ace_gutter-cell ";if(l[g]&&(m+=l[g]),u[g]&&(m+=u[g]),this.$annotations[g]&&(m+=this.$annotations[g].className),d.element.className!=m&&(d.element.className=m),(b=t.getRowLength(g)*e.lineHeight+"px")!=d.element.style.height&&(d.element.style.height=b),a){var v=a[g];null==v&&(v=a[g]=t.getFoldWidget(g))}if(v){d.foldWidget||(d.foldWidget=i.createElement("span"),d.element.appendChild(d.foldWidget));m="ace_fold-widget ace_"+v;"start"==v&&g==s&&g<o.end.row?m+=" ace_closed":m+=" ace_open",d.foldWidget.className!=m&&(d.foldWidget.className=m);var b=e.lineHeight+"px";d.foldWidget.style.height!=b&&(d.foldWidget.style.height=b)}else d.foldWidget&&(d.element.removeChild(d.foldWidget),d.foldWidget=null);var y=h=f?f.getText(t,g):g+c;y!==d.textNode.data&&(d.textNode.data=y),g++}this.element.style.height=e.minHeight+"px",(this.$fixedWidth||t.$useWrapMode)&&(h=t.getLength()+c);var w=f?f.getWidth(t,h,e):h.toString().length*e.characterWidth,A=this.$padding||this.$computePadding();(w+=A.left+A.right)===this.gutterWidth||isNaN(w)||(this.gutterWidth=w,this.element.style.width=Math.ceil(this.gutterWidth)+"px",this._emit("changeGutterWidth",w))},this.$fixedWidth=!1,this.$showLineNumbers=!0,this.$renderer="",this.setShowLineNumbers=function(e){this.$renderer=!e&&{getWidth:function(){return""},getText:function(){return""}}},this.getShowLineNumbers=function(){return this.$showLineNumbers},this.$showFoldWidgets=!0,this.setShowFoldWidgets=function(e){e?i.addCssClass(this.element,"ace_folding-enabled"):i.removeCssClass(this.element,"ace_folding-enabled"),this.$showFoldWidgets=e,this.$padding=null},this.getShowFoldWidgets=function(){return this.$showFoldWidgets},this.$computePadding=function(){if(!this.element.firstChild)return{left:0,right:0};var e=i.computedStyle(this.element.firstChild);return this.$padding={},this.$padding.left=parseInt(e.paddingLeft)+1||0,this.$padding.right=parseInt(e.paddingRight)||0,this.$padding},this.getRegion=function(e){var t=this.$padding||this.$computePadding(),n=this.element.getBoundingClientRect();return e.x<t.left+n.left?"markers":this.$showFoldWidgets&&e.x>n.right-t.right?"foldWidgets":void 0}}).call(a.prototype),t.Gutter=a})),ace.define("ace/layer/marker",["require","exports","module","ace/range","ace/lib/dom"],(function(e,t,n){"use strict";var i=e("../range").Range,r=e("../lib/dom"),o=function(e){this.element=r.createElement("div"),this.element.className="ace_layer ace_marker-layer",e.appendChild(this.element)};(function(){this.$padding=0,this.setPadding=function(e){this.$padding=e},this.setSession=function(e){this.session=e},this.setMarkers=function(e){this.markers=e},this.update=function(e){if(e){this.config=e;var t=[];for(var n in this.markers){var i=this.markers[n];if(i.range){var r=i.range.clipRows(e.firstRow,e.lastRow);if(!r.isEmpty())if(r=r.toScreenRange(this.session),i.renderer){var o=this.$getTop(r.start.row,e),s=this.$padding+(this.session.$bidiHandler.isBidiRow(r.start.row)?this.session.$bidiHandler.getPosLeft(r.start.column):r.start.column*e.characterWidth);i.renderer(t,r,s,o,e)}else"fullLine"==i.type?this.drawFullLineMarker(t,r,i.clazz,e):"screenLine"==i.type?this.drawScreenLineMarker(t,r,i.clazz,e):r.isMultiLine()?"text"==i.type?this.drawTextMarker(t,r,i.clazz,e):this.drawMultiLineMarker(t,r,i.clazz,e):this.session.$bidiHandler.isBidiRow(r.start.row)?this.drawBidiSingleLineMarker(t,r,i.clazz+" ace_start ace_br15",e):this.drawSingleLineMarker(t,r,i.clazz+" ace_start ace_br15",e)}else i.update(t,this,this.session,e)}this.element.innerHTML=t.join("")}},this.$getTop=function(e,t){return(e-t.firstRowScreen)*t.lineHeight},this.drawTextMarker=function(e,t,n,r,o){for(var s=this.session,a=t.start.row,l=t.end.row,u=a,c=0,h=0,f=s.getScreenLastRowColumn(u),d=null,p=new i(u,t.start.column,u,h);u<=l;u++)p.start.row=p.end.row=u,p.start.column=u==a?t.start.column:s.getRowWrapIndent(u),p.end.column=f,c=h,h=f,f=u+1<l?s.getScreenLastRowColumn(u+1):u==l?0:t.end.column,d=n+(u==a?" ace_start":"")+" ace_br"+((u==a||u==a+1&&t.start.column?1:0)|(c<h?2:0)|(h>f?4:0)|(u==l?8:0)),this.session.$bidiHandler.isBidiRow(u)?this.drawBidiSingleLineMarker(e,p,d,r,u==l?0:1,o):this.drawSingleLineMarker(e,p,d,r,u==l?0:1,o)},this.drawMultiLineMarker=function(e,t,n,i,r){var o,s,a,l=this.$padding;(r=r||"",this.session.$bidiHandler.isBidiRow(t.start.row))?((u=t.clone()).end.row=u.start.row,u.end.column=this.session.getLine(u.start.row).length,this.drawBidiSingleLineMarker(e,u,n+" ace_br1 ace_start",i,null,r)):(o=i.lineHeight,s=this.$getTop(t.start.row,i),a=l+t.start.column*i.characterWidth,e.push("<div class='",n," ace_br1 ace_start' style='","height:",o,"px;","right:0;","top:",s,"px;","left:",a,"px;",r,"'></div>"));if(this.session.$bidiHandler.isBidiRow(t.end.row)){var u;(u=t.clone()).start.row=u.end.row,u.start.column=0,this.drawBidiSingleLineMarker(e,u,n+" ace_br12",i,null,r)}else{var c=t.end.column*i.characterWidth;o=i.lineHeight,s=this.$getTop(t.end.row,i),e.push("<div class='",n," ace_br12' style='","height:",o,"px;","width:",c,"px;","top:",s,"px;","left:",l,"px;",r,"'></div>")}if(!((o=(t.end.row-t.start.row-1)*i.lineHeight)<=0)){s=this.$getTop(t.start.row+1,i);var h=(t.start.column?1:0)|(t.end.column?0:8);e.push("<div class='",n,h?" ace_br"+h:"","' style='","height:",o,"px;","right:0;","top:",s,"px;","left:",l,"px;",r,"'></div>")}},this.drawSingleLineMarker=function(e,t,n,i,r,o){var s=i.lineHeight,a=(t.end.column+(r||0)-t.start.column)*i.characterWidth,l=this.$getTop(t.start.row,i),u=this.$padding+t.start.column*i.characterWidth;e.push("<div class='",n,"' style='","height:",s,"px;","width:",a,"px;","top:",l,"px;","left:",u,"px;",o||"","'></div>")},this.drawBidiSingleLineMarker=function(e,t,n,i,r,o){var s=i.lineHeight,a=this.$getTop(t.start.row,i),l=this.$padding;this.session.$bidiHandler.getSelections(t.start.column,t.end.column).forEach((function(t){e.push("<div class='",n,"' style='","height:",s,"px;","width:",t.width+(r||0),"px;","top:",a,"px;","left:",l+t.left,"px;",o||"","'></div>")}))},this.drawFullLineMarker=function(e,t,n,i,r){var o=this.$getTop(t.start.row,i),s=i.lineHeight;t.start.row!=t.end.row&&(s+=this.$getTop(t.end.row,i)-o),e.push("<div class='",n,"' style='","height:",s,"px;","top:",o,"px;","left:0;right:0;",r||"","'></div>")},this.drawScreenLineMarker=function(e,t,n,i,r){var o=this.$getTop(t.start.row,i),s=i.lineHeight;e.push("<div class='",n,"' style='","height:",s,"px;","top:",o,"px;","left:0;right:0;",r||"","'></div>")}}).call(o.prototype),t.Marker=o})),ace.define("ace/layer/text",["require","exports","module","ace/lib/oop","ace/lib/dom","ace/lib/lang","ace/lib/useragent","ace/lib/event_emitter"],(function(e,t,n){"use strict";var i=e("../lib/oop"),r=e("../lib/dom"),o=e("../lib/lang"),s=(e("../lib/useragent"),e("../lib/event_emitter").EventEmitter),a=function(e){this.element=r.createElement("div"),this.element.className="ace_layer ace_text-layer",e.appendChild(this.element),this.$updateEolChar=this.$updateEolChar.bind(this)};(function(){i.implement(this,s),this.EOF_CHAR="¶",this.EOL_CHAR_LF="¬",this.EOL_CHAR_CRLF="¤",this.EOL_CHAR=this.EOL_CHAR_LF,this.TAB_CHAR="—",this.SPACE_CHAR="·",this.$padding=0,this.$updateEolChar=function(){var e="\n"==this.session.doc.getNewLineCharacter()?this.EOL_CHAR_LF:this.EOL_CHAR_CRLF;if(this.EOL_CHAR!=e)return this.EOL_CHAR=e,!0},this.setPadding=function(e){this.$padding=e,this.element.style.padding="0 "+e+"px"},this.getLineHeight=function(){return this.$fontMetrics.$characterSize.height||0},this.getCharacterWidth=function(){return this.$fontMetrics.$characterSize.width||0},this.$setFontMetrics=function(e){this.$fontMetrics=e,this.$fontMetrics.on("changeCharacterSize",function(e){this._signal("changeCharacterSize",e)}.bind(this)),this.$pollSizeChanges()},this.checkForSizeChanges=function(){this.$fontMetrics.checkForSizeChanges()},this.$pollSizeChanges=function(){return this.$pollSizeChangesTimer=this.$fontMetrics.$pollSizeChanges()},this.setSession=function(e){this.session=e,e&&this.$computeTabString()},this.showInvisibles=!1,this.setShowInvisibles=function(e){return this.showInvisibles!=e&&(this.showInvisibles=e,this.$computeTabString(),!0)},this.displayIndentGuides=!0,this.setDisplayIndentGuides=function(e){return this.displayIndentGuides!=e&&(this.displayIndentGuides=e,this.$computeTabString(),!0)},this.$tabStrings=[],this.onChangeTabSize=this.$computeTabString=function(){var e=this.session.getTabSize();this.tabSize=e;for(var t=this.$tabStrings=[0],n=1;n<e+1;n++)this.showInvisibles?t.push("<span class='ace_invisible ace_invisible_tab'>"+o.stringRepeat(this.TAB_CHAR,n)+"</span>"):t.push(o.stringRepeat(" ",n));if(this.displayIndentGuides){this.$indentGuideRe=/\s\S| \t|\t |\s$/;var i="ace_indent-guide",r="",s="";if(this.showInvisibles){i+=" ace_invisible",r=" ace_invisible_space",s=" ace_invisible_tab";var a=o.stringRepeat(this.SPACE_CHAR,this.tabSize),l=o.stringRepeat(this.TAB_CHAR,this.tabSize)}else l=a=o.stringRepeat(" ",this.tabSize);this.$tabStrings[" "]="<span class='"+i+r+"'>"+a+"</span>",this.$tabStrings["\t"]="<span class='"+i+s+"'>"+l+"</span>"}},this.updateLines=function(e,t,n){this.config.lastRow==e.lastRow&&this.config.firstRow==e.firstRow||this.scrollLines(e),this.config=e;for(var i=Math.max(t,e.firstRow),r=Math.min(n,e.lastRow),o=this.element.childNodes,s=0,a=e.firstRow;a<i;a++){if(l=this.session.getFoldLine(a)){if(l.containsRow(i)){i=l.start.row;break}a=l.end.row}s++}a=i;for(var l,u=(l=this.session.getNextFoldLine(a))?l.start.row:1/0;a>u&&(a=l.end.row+1,u=(l=this.session.getNextFoldLine(a,l))?l.start.row:1/0),!(a>r);){var c=o[s++];if(c){var h=[];this.$renderLine(h,a,!this.$useLineGroups(),a==u&&l),c.style.height=e.lineHeight*this.session.getRowLength(a)+"px",c.innerHTML=h.join("")}a++}},this.scrollLines=function(e){var t=this.config;if(this.config=e,!t||t.lastRow<e.firstRow)return this.update(e);if(e.lastRow<t.firstRow)return this.update(e);var n=this.element;if(t.firstRow<e.firstRow)for(var i=this.session.getFoldedRowCount(t.firstRow,e.firstRow-1);i>0;i--)n.removeChild(n.firstChild);if(t.lastRow>e.lastRow)for(i=this.session.getFoldedRowCount(e.lastRow+1,t.lastRow);i>0;i--)n.removeChild(n.lastChild);if(e.firstRow<t.firstRow){var r=this.$renderLinesFragment(e,e.firstRow,t.firstRow-1);n.firstChild?n.insertBefore(r,n.firstChild):n.appendChild(r)}if(e.lastRow>t.lastRow){r=this.$renderLinesFragment(e,t.lastRow+1,e.lastRow);n.appendChild(r)}},this.$renderLinesFragment=function(e,t,n){for(var i=this.element.ownerDocument.createDocumentFragment(),o=t,s=this.session.getNextFoldLine(o),a=s?s.start.row:1/0;o>a&&(o=s.end.row+1,a=(s=this.session.getNextFoldLine(o,s))?s.start.row:1/0),!(o>n);){var l=r.createElement("div"),u=[];if(this.$renderLine(u,o,!1,o==a&&s),l.innerHTML=u.join(""),this.$useLineGroups())l.className="ace_line_group",i.appendChild(l),l.style.height=e.lineHeight*this.session.getRowLength(o)+"px";else for(;l.firstChild;)i.appendChild(l.firstChild);o++}return i},this.update=function(e){this.config=e;for(var t=[],n=e.firstRow,i=e.lastRow,r=n,o=this.session.getNextFoldLine(r),s=o?o.start.row:1/0;r>s&&(r=o.end.row+1,s=(o=this.session.getNextFoldLine(r,o))?o.start.row:1/0),!(r>i);)this.$useLineGroups()&&t.push("<div class='ace_line_group' style='height:",e.lineHeight*this.session.getRowLength(r),"px'>"),this.$renderLine(t,r,!1,r==s&&o),this.$useLineGroups()&&t.push("</div>"),r++;this.element.innerHTML=t.join("")},this.$textToken={text:!0,rparen:!0,lparen:!0},this.$renderToken=function(e,t,n,i){var r=this,s=i.replace(/\t|&|<|>|( +)|([\x00-\x1f\x80-\xa0\xad\u1680\u180E\u2000-\u200f\u2028\u2029\u202F\u205F\u3000\uFEFF\uFFF9-\uFFFC])|[\u1100-\u115F\u11A3-\u11A7\u11FA-\u11FF\u2329-\u232A\u2E80-\u2E99\u2E9B-\u2EF3\u2F00-\u2FD5\u2FF0-\u2FFB\u3000-\u303E\u3041-\u3096\u3099-\u30FF\u3105-\u312D\u3131-\u318E\u3190-\u31BA\u31C0-\u31E3\u31F0-\u321E\u3220-\u3247\u3250-\u32FE\u3300-\u4DBF\u4E00-\uA48C\uA490-\uA4C6\uA960-\uA97C\uAC00-\uD7A3\uD7B0-\uD7C6\uD7CB-\uD7FB\uF900-\uFAFF\uFE10-\uFE19\uFE30-\uFE52\uFE54-\uFE66\uFE68-\uFE6B\uFF01-\uFF60\uFFE0-\uFFE6]|[\uD800-\uDBFF][\uDC00-\uDFFF]/g,(function(e,n,i,s,a){if(n)return r.showInvisibles?"<span class='ace_invisible ace_invisible_space'>"+o.stringRepeat(r.SPACE_CHAR,e.length)+"</span>":e;if("&"==e)return"&#38;";if("<"==e)return"&#60;";if(">"==e)return"&#62;";if("\t"==e){var l=r.session.getScreenTabSize(t+s);return t+=l-1,r.$tabStrings[l]}if("　"==e){var u=r.showInvisibles?"ace_cjk ace_invisible ace_invisible_space":"ace_cjk",c=r.showInvisibles?r.SPACE_CHAR:"";return t+=1,"<span class='"+u+"' style='width:"+2*r.config.characterWidth+"px'>"+c+"</span>"}return i?"<span class='ace_invisible ace_invisible_space ace_invalid'>"+r.SPACE_CHAR+"</span>":(t+=1,"<span class='ace_cjk' style='width:"+2*r.config.characterWidth+"px'>"+e+"</span>")}));if(this.$textToken[n.type])e.push(s);else{var a="ace_"+n.type.replace(/\./g," ace_"),l="";"fold"==n.type&&(l=" style='width:"+n.value.length*this.config.characterWidth+"px;' "),e.push("<span class='",a,"'",l,">",s,"</span>")}return t+i.length},this.renderIndentGuide=function(e,t,n){var i=t.search(this.$indentGuideRe);return i<=0||i>=n?t:" "==t[0]?(i-=i%this.tabSize,e.push(o.stringRepeat(this.$tabStrings[" "],i/this.tabSize)),t.substr(i)):"\t"==t[0]?(e.push(o.stringRepeat(this.$tabStrings["\t"],i)),t.substr(i)):t},this.$renderWrappedLine=function(e,t,n,i){for(var r=0,s=0,a=n[0],l=0,u=0;u<t.length;u++){var c=t[u],h=c.value;if(0==u&&this.displayIndentGuides){if(r=h.length,!(h=this.renderIndentGuide(e,h,a)))continue;r-=h.length}if(r+h.length<a)l=this.$renderToken(e,l,c,h),r+=h.length;else{for(;r+h.length>=a;)l=this.$renderToken(e,l,c,h.substring(0,a-r)),h=h.substring(a-r),r=a,i||e.push("</div>","<div class='ace_line' style='height:",this.config.lineHeight,"px'>"),e.push(o.stringRepeat(" ",n.indent)),l=0,a=n[++s]||Number.MAX_VALUE;0!=h.length&&(r+=h.length,l=this.$renderToken(e,l,c,h))}}},this.$renderSimpleLine=function(e,t){var n=0,i=t[0],r=i.value;this.displayIndentGuides&&(r=this.renderIndentGuide(e,r)),r&&(n=this.$renderToken(e,n,i,r));for(var o=1;o<t.length;o++)r=(i=t[o]).value,n=this.$renderToken(e,n,i,r)},this.$renderLine=function(e,t,n,i){if(i||0==i||(i=this.session.getFoldLine(t)),i)var r=this.$getFoldLineTokens(t,i);else r=this.session.getTokens(t);if(n||e.push("<div class='ace_line' style='height:",this.config.lineHeight*(this.$useLineGroups()?1:this.session.getRowLength(t)),"px'>"),r.length){var o=this.session.getRowSplitData(t);o&&o.length?this.$renderWrappedLine(e,r,o,n):this.$renderSimpleLine(e,r)}this.showInvisibles&&(i&&(t=i.end.row),e.push("<span class='ace_invisible ace_invisible_eol'>",t==this.session.getLength()-1?this.EOF_CHAR:this.EOL_CHAR,"</span>")),n||e.push("</div>")},this.$getFoldLineTokens=function(e,t){var n=this.session,i=[];var r=n.getTokens(e);return t.walk((function(e,t,o,s,a){null!=e?i.push({type:"fold",value:e}):(a&&(r=n.getTokens(t)),r.length&&function(e,t,n){for(var r=0,o=0;o+e[r].value.length<t;)if(o+=e[r].value.length,++r==e.length)return;for(o!=t&&((s=e[r].value.substring(t-o)).length>n-t&&(s=s.substring(0,n-t)),i.push({type:e[r].type,value:s}),o=t+s.length,r+=1);o<n&&r<e.length;){var s;(s=e[r].value).length+o>n?i.push({type:e[r].type,value:s.substring(0,n-o)}):i.push(e[r]),o+=s.length,r+=1}}(r,s,o))}),t.end.row,this.session.getLine(t.end.row).length),i},this.$useLineGroups=function(){return this.session.getUseWrapMode()},this.destroy=function(){clearInterval(this.$pollSizeChangesTimer),this.$measureNode&&this.$measureNode.parentNode.removeChild(this.$measureNode),delete this.$measureNode}}).call(a.prototype),t.Text=a})),ace.define("ace/layer/cursor",["require","exports","module","ace/lib/dom"],(function(e,t,n){"use strict";var i,r=e("../lib/dom"),o=function(e){this.element=r.createElement("div"),this.element.className="ace_layer ace_cursor-layer",e.appendChild(this.element),void 0===i&&(i=!("opacity"in this.element.style)),this.isVisible=!1,this.isBlinking=!0,this.blinkInterval=1e3,this.smoothBlinking=!1,this.cursors=[],this.cursor=this.addCursor(),r.addCssClass(this.element,"ace_hidden-cursors"),this.$updateCursors=(i?this.$updateVisibility:this.$updateOpacity).bind(this)};(function(){this.$updateVisibility=function(e){for(var t=this.cursors,n=t.length;n--;)t[n].style.visibility=e?"":"hidden"},this.$updateOpacity=function(e){for(var t=this.cursors,n=t.length;n--;)t[n].style.opacity=e?"":"0"},this.$padding=0,this.setPadding=function(e){this.$padding=e},this.setSession=function(e){this.session=e},this.setBlinking=function(e){e!=this.isBlinking&&(this.isBlinking=e,this.restartTimer())},this.setBlinkInterval=function(e){e!=this.blinkInterval&&(this.blinkInterval=e,this.restartTimer())},this.setSmoothBlinking=function(e){e==this.smoothBlinking||i||(this.smoothBlinking=e,r.setCssClass(this.element,"ace_smooth-blinking",e),this.$updateCursors(!0),this.$updateCursors=this.$updateOpacity.bind(this),this.restartTimer())},this.addCursor=function(){var e=r.createElement("div");return e.className="ace_cursor",this.element.appendChild(e),this.cursors.push(e),e},this.removeCursor=function(){if(this.cursors.length>1){var e=this.cursors.pop();return e.parentNode.removeChild(e),e}},this.hideCursor=function(){this.isVisible=!1,r.addCssClass(this.element,"ace_hidden-cursors"),this.restartTimer()},this.showCursor=function(){this.isVisible=!0,r.removeCssClass(this.element,"ace_hidden-cursors"),this.restartTimer()},this.restartTimer=function(){var e=this.$updateCursors;if(clearInterval(this.intervalId),clearTimeout(this.timeoutId),this.smoothBlinking&&r.removeCssClass(this.element,"ace_smooth-blinking"),e(!0),this.isBlinking&&this.blinkInterval&&this.isVisible){this.smoothBlinking&&setTimeout(function(){r.addCssClass(this.element,"ace_smooth-blinking")}.bind(this));var t=function(){this.timeoutId=setTimeout((function(){e(!1)}),.6*this.blinkInterval)}.bind(this);this.intervalId=setInterval((function(){e(!0),t()}),this.blinkInterval),t()}},this.getPixelPosition=function(e,t){if(!this.config||!this.session)return{left:0,top:0};e||(e=this.session.selection.getCursor());var n=this.session.documentToScreenPosition(e);return{left:this.$padding+(this.session.$bidiHandler.isBidiRow(n.row,e.row)?this.session.$bidiHandler.getPosLeft(n.column):n.column*this.config.characterWidth),top:(n.row-(t?this.config.firstRowScreen:0))*this.config.lineHeight}},this.update=function(e){this.config=e;var t=this.session.$selectionMarkers,n=0,i=0;void 0!==t&&0!==t.length||(t=[{cursor:null}]);n=0;for(var r=t.length;n<r;n++){var o=this.getPixelPosition(t[n].cursor,!0);if(!((o.top>e.height+e.offset||o.top<0)&&n>1)){var s=(this.cursors[i++]||this.addCursor()).style;this.drawCursor?this.drawCursor(s,o,e,t[n],this.session):(s.left=o.left+"px",s.top=o.top+"px",s.width=e.characterWidth+"px",s.height=e.lineHeight+"px")}}for(;this.cursors.length>i;)this.removeCursor();var a=this.session.getOverwrite();this.$setOverwrite(a),this.$pixelPos=o,this.restartTimer()},this.drawCursor=null,this.$setOverwrite=function(e){e!=this.overwrite&&(this.overwrite=e,e?r.addCssClass(this.element,"ace_overwrite-cursors"):r.removeCssClass(this.element,"ace_overwrite-cursors"))},this.destroy=function(){clearInterval(this.intervalId),clearTimeout(this.timeoutId)}}).call(o.prototype),t.Cursor=o})),ace.define("ace/scrollbar",["require","exports","module","ace/lib/oop","ace/lib/dom","ace/lib/event","ace/lib/event_emitter"],(function(e,t,n){"use strict";var i=e("./lib/oop"),r=e("./lib/dom"),o=e("./lib/event"),s=e("./lib/event_emitter").EventEmitter,a=function(e){this.element=r.createElement("div"),this.element.className="ace_scrollbar ace_scrollbar"+this.classSuffix,this.inner=r.createElement("div"),this.inner.className="ace_scrollbar-inner",this.element.appendChild(this.inner),e.appendChild(this.element),this.setVisible(!1),this.skipEvent=!1,o.addListener(this.element,"scroll",this.onScroll.bind(this)),o.addListener(this.element,"mousedown",o.preventDefault)};(function(){i.implement(this,s),this.setVisible=function(e){this.element.style.display=e?"":"none",this.isVisible=e,this.coeff=1}}).call(a.prototype);var l=function(e,t){a.call(this,e),this.scrollTop=0,this.scrollHeight=0,t.$scrollbarWidth=this.width=r.scrollbarWidth(e.ownerDocument),this.inner.style.width=this.element.style.width=(this.width||15)+5+"px",this.$minWidth=0};i.inherits(l,a),function(){this.classSuffix="-v",this.onScroll=function(){if(!this.skipEvent){if(this.scrollTop=this.element.scrollTop,1!=this.coeff){var e=this.element.clientHeight/this.scrollHeight;this.scrollTop=this.scrollTop*(1-e)/(this.coeff-e)}this._emit("scroll",{data:this.scrollTop})}this.skipEvent=!1},this.getWidth=function(){return Math.max(this.isVisible?this.width:0,this.$minWidth||0)},this.setHeight=function(e){this.element.style.height=e+"px"},this.setInnerHeight=this.setScrollHeight=function(e){this.scrollHeight=e,e>32768?(this.coeff=32768/e,e=32768):1!=this.coeff&&(this.coeff=1),this.inner.style.height=e+"px"},this.setScrollTop=function(e){this.scrollTop!=e&&(this.skipEvent=!0,this.scrollTop=e,this.element.scrollTop=e*this.coeff)}}.call(l.prototype);var u=function(e,t){a.call(this,e),this.scrollLeft=0,this.height=t.$scrollbarWidth,this.inner.style.height=this.element.style.height=(this.height||15)+5+"px"};i.inherits(u,a),function(){this.classSuffix="-h",this.onScroll=function(){this.skipEvent||(this.scrollLeft=this.element.scrollLeft,this._emit("scroll",{data:this.scrollLeft})),this.skipEvent=!1},this.getHeight=function(){return this.isVisible?this.height:0},this.setWidth=function(e){this.element.style.width=e+"px"},this.setInnerWidth=function(e){this.inner.style.width=e+"px"},this.setScrollWidth=function(e){this.inner.style.width=e+"px"},this.setScrollLeft=function(e){this.scrollLeft!=e&&(this.skipEvent=!0,this.scrollLeft=this.element.scrollLeft=e)}}.call(u.prototype),t.ScrollBar=l,t.ScrollBarV=l,t.ScrollBarH=u,t.VScrollBar=l,t.HScrollBar=u})),ace.define("ace/renderloop",["require","exports","module","ace/lib/event"],(function(e,t,n){"use strict";var i=e("./lib/event"),r=function(e,t){this.onRender=e,this.pending=!1,this.changes=0,this.window=t||window};(function(){this.schedule=function(e){if(this.changes=this.changes|e,!this.pending&&this.changes){this.pending=!0;var t=this;i.nextFrame((function(){var e;for(t.pending=!1;e=t.changes;)t.changes=0,t.onRender(e)}),this.window)}}}).call(r.prototype),t.RenderLoop=r})),ace.define("ace/layer/font_metrics",["require","exports","module","ace/lib/oop","ace/lib/dom","ace/lib/lang","ace/lib/useragent","ace/lib/event_emitter"],(function(e,t,n){var i=e("../lib/oop"),r=e("../lib/dom"),o=e("../lib/lang"),s=e("../lib/useragent"),a=e("../lib/event_emitter").EventEmitter,l=0,u=t.FontMetrics=function(e){this.el=r.createElement("div"),this.$setMeasureNodeStyles(this.el.style,!0),this.$main=r.createElement("div"),this.$setMeasureNodeStyles(this.$main.style),this.$measureNode=r.createElement("div"),this.$setMeasureNodeStyles(this.$measureNode.style),this.el.appendChild(this.$main),this.el.appendChild(this.$measureNode),e.appendChild(this.el),l||this.$testFractionalRect(),this.$measureNode.innerHTML=o.stringRepeat("X",l),this.$characterSize={width:0,height:0},this.checkForSizeChanges()};(function(){i.implement(this,a),this.$characterSize={width:0,height:0},this.$testFractionalRect=function(){var e=r.createElement("div");this.$setMeasureNodeStyles(e.style),e.style.width="0.2px",document.documentElement.appendChild(e);var t=e.getBoundingClientRect().width;l=t>0&&t<1?50:100,e.parentNode.removeChild(e)},this.$setMeasureNodeStyles=function(e,t){e.width=e.height="auto",e.left=e.top="0px",e.visibility="hidden",e.position="absolute",e.whiteSpace="pre",s.isIE<8?e["font-family"]="inherit":e.font="inherit",e.overflow=t?"hidden":"visible"},this.checkForSizeChanges=function(){var e=this.$measureSizes();if(e&&(this.$characterSize.width!==e.width||this.$characterSize.height!==e.height)){this.$measureNode.style.fontWeight="bold";var t=this.$measureSizes();this.$measureNode.style.fontWeight="",this.$characterSize=e,this.charSizes=Object.create(null),this.allowBoldFonts=t&&t.width===e.width&&t.height===e.height,this._emit("changeCharacterSize",{data:e})}},this.$pollSizeChanges=function(){if(this.$pollSizeChangesTimer)return this.$pollSizeChangesTimer;var e=this;return this.$pollSizeChangesTimer=setInterval((function(){e.checkForSizeChanges()}),500)},this.setPolling=function(e){e?this.$pollSizeChanges():this.$pollSizeChangesTimer&&(clearInterval(this.$pollSizeChangesTimer),this.$pollSizeChangesTimer=0)},this.$measureSizes=function(){if(50===l){var e=null;try{e=this.$measureNode.getBoundingClientRect()}catch(t){e={width:0,height:0}}var t={height:e.height,width:e.width/l}}else t={height:this.$measureNode.clientHeight,width:this.$measureNode.clientWidth/l};return 0===t.width||0===t.height?null:t},this.$measureCharWidth=function(e){return this.$main.innerHTML=o.stringRepeat(e,l),this.$main.getBoundingClientRect().width/l},this.getCharacterWidth=function(e){var t=this.charSizes[e];return void 0===t&&(t=this.charSizes[e]=this.$measureCharWidth(e)/this.$characterSize.width),t},this.destroy=function(){clearInterval(this.$pollSizeChangesTimer),this.el&&this.el.parentNode&&this.el.parentNode.removeChild(this.el)}}).call(u.prototype)})),ace.define("ace/virtual_renderer",["require","exports","module","ace/lib/oop","ace/lib/dom","ace/config","ace/lib/useragent","ace/layer/gutter","ace/layer/marker","ace/layer/text","ace/layer/cursor","ace/scrollbar","ace/scrollbar","ace/renderloop","ace/layer/font_metrics","ace/lib/event_emitter"],(function(e,t,n){"use strict";var i=e("./lib/oop"),r=e("./lib/dom"),o=e("./config"),s=e("./lib/useragent"),a=e("./layer/gutter").Gutter,l=e("./layer/marker").Marker,u=e("./layer/text").Text,c=e("./layer/cursor").Cursor,h=e("./scrollbar").HScrollBar,f=e("./scrollbar").VScrollBar,d=e("./renderloop").RenderLoop,p=e("./layer/font_metrics").FontMetrics,g=e("./lib/event_emitter").EventEmitter;r.importCssString('.ace_editor {position: relative;overflow: hidden;font: 12px/normal \'Monaco\', \'Menlo\', \'Ubuntu Mono\', \'Consolas\', \'source-code-pro\', monospace;direction: ltr;text-align: left;-webkit-tap-highlight-color: rgba(0, 0, 0, 0);}.ace_scroller {position: absolute;overflow: hidden;top: 0;bottom: 0;background-color: inherit;-ms-user-select: none;-moz-user-select: none;-webkit-user-select: none;user-select: none;cursor: text;}.ace_content {position: absolute;-moz-box-sizing: border-box;-webkit-box-sizing: border-box;box-sizing: border-box;min-width: 100%;}.ace_dragging .ace_scroller:before{position: absolute;top: 0;left: 0;right: 0;bottom: 0;content: \'\';background: rgba(250, 250, 250, 0.01);z-index: 1000;}.ace_dragging.ace_dark .ace_scroller:before{background: rgba(0, 0, 0, 0.01);}.ace_selecting, .ace_selecting * {cursor: text !important;}.ace_gutter {position: absolute;overflow : hidden;width: auto;top: 0;bottom: 0;left: 0;cursor: default;z-index: 4;-ms-user-select: none;-moz-user-select: none;-webkit-user-select: none;user-select: none;}.ace_gutter-active-line {position: absolute;left: 0;right: 0;}.ace_scroller.ace_scroll-left {box-shadow: 17px 0 16px -16px rgba(0, 0, 0, 0.4) inset;}.ace_gutter-cell {padding-left: 19px;padding-right: 6px;background-repeat: no-repeat;}.ace_gutter-cell.ace_error {background-image: url("data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAABAAAAAQCAMAAAAoLQ9TAAABOFBMVEX/////////QRswFAb/Ui4wFAYwFAYwFAaWGAfDRymzOSH/PxswFAb/SiUwFAYwFAbUPRvjQiDllog5HhHdRybsTi3/Tyv9Tir+Syj/UC3////XurebMBIwFAb/RSHbPx/gUzfdwL3kzMivKBAwFAbbvbnhPx66NhowFAYwFAaZJg8wFAaxKBDZurf/RB6mMxb/SCMwFAYwFAbxQB3+RB4wFAb/Qhy4Oh+4QifbNRcwFAYwFAYwFAb/QRzdNhgwFAYwFAbav7v/Uy7oaE68MBK5LxLewr/r2NXewLswFAaxJw4wFAbkPRy2PyYwFAaxKhLm1tMwFAazPiQwFAaUGAb/QBrfOx3bvrv/VC/maE4wFAbRPBq6MRO8Qynew8Dp2tjfwb0wFAbx6eju5+by6uns4uH9/f36+vr/GkHjAAAAYnRSTlMAGt+64rnWu/bo8eAA4InH3+DwoN7j4eLi4xP99Nfg4+b+/u9B/eDs1MD1mO7+4PHg2MXa347g7vDizMLN4eG+Pv7i5evs/v79yu7S3/DV7/498Yv24eH+4ufQ3Ozu/v7+y13sRqwAAADLSURBVHjaZc/XDsFgGIBhtDrshlitmk2IrbHFqL2pvXf/+78DPokj7+Fz9qpU/9UXJIlhmPaTaQ6QPaz0mm+5gwkgovcV6GZzd5JtCQwgsxoHOvJO15kleRLAnMgHFIESUEPmawB9ngmelTtipwwfASilxOLyiV5UVUyVAfbG0cCPHig+GBkzAENHS0AstVF6bacZIOzgLmxsHbt2OecNgJC83JERmePUYq8ARGkJx6XtFsdddBQgZE2nPR6CICZhawjA4Fb/chv+399kfR+MMMDGOQAAAABJRU5ErkJggg==");background-repeat: no-repeat;background-position: 2px center;}.ace_gutter-cell.ace_warning {background-image: url("data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAABAAAAAQCAMAAAAoLQ9TAAAAmVBMVEX///8AAAD///8AAAAAAABPSzb/5sAAAAB/blH/73z/ulkAAAAAAAD85pkAAAAAAAACAgP/vGz/rkDerGbGrV7/pkQICAf////e0IsAAAD/oED/qTvhrnUAAAD/yHD/njcAAADuv2r/nz//oTj/p064oGf/zHAAAAA9Nir/tFIAAAD/tlTiuWf/tkIAAACynXEAAAAAAAAtIRW7zBpBAAAAM3RSTlMAABR1m7RXO8Ln31Z36zT+neXe5OzooRDfn+TZ4p3h2hTf4t3k3ucyrN1K5+Xaks52Sfs9CXgrAAAAjklEQVR42o3PbQ+CIBQFYEwboPhSYgoYunIqqLn6/z8uYdH8Vmdnu9vz4WwXgN/xTPRD2+sgOcZjsge/whXZgUaYYvT8QnuJaUrjrHUQreGczuEafQCO/SJTufTbroWsPgsllVhq3wJEk2jUSzX3CUEDJC84707djRc5MTAQxoLgupWRwW6UB5fS++NV8AbOZgnsC7BpEAAAAABJRU5ErkJggg==");background-position: 2px center;}.ace_gutter-cell.ace_info {background-image: url("data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAABAAAAAQCAAAAAA6mKC9AAAAGXRFWHRTb2Z0d2FyZQBBZG9iZSBJbWFnZVJlYWR5ccllPAAAAAJ0Uk5TAAB2k804AAAAPklEQVQY02NgIB68QuO3tiLznjAwpKTgNyDbMegwisCHZUETUZV0ZqOquBpXj2rtnpSJT1AEnnRmL2OgGgAAIKkRQap2htgAAAAASUVORK5CYII=");background-position: 2px center;}.ace_dark .ace_gutter-cell.ace_info {background-image: url("data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAABAAAAAQBAMAAADt3eJSAAAAJFBMVEUAAAChoaGAgIAqKiq+vr6tra1ZWVmUlJSbm5s8PDxubm56enrdgzg3AAAAAXRSTlMAQObYZgAAAClJREFUeNpjYMAPdsMYHegyJZFQBlsUlMFVCWUYKkAZMxZAGdxlDMQBAG+TBP4B6RyJAAAAAElFTkSuQmCC");}.ace_scrollbar {position: absolute;right: 0;bottom: 0;z-index: 6;}.ace_scrollbar-inner {position: absolute;cursor: text;left: 0;top: 0;}.ace_scrollbar-v{overflow-x: hidden;overflow-y: scroll;top: 0;}.ace_scrollbar-h {overflow-x: scroll;overflow-y: hidden;left: 0;}.ace_print-margin {position: absolute;height: 100%;}.ace_text-input {position: absolute;z-index: 0;width: 0.5em;height: 1em;opacity: 0;background: transparent;-moz-appearance: none;appearance: none;border: none;resize: none;outline: none;overflow: hidden;font: inherit;padding: 0 1px;margin: 0 -1px;text-indent: -1em;-ms-user-select: text;-moz-user-select: text;-webkit-user-select: text;user-select: text;white-space: pre!important;}.ace_text-input.ace_composition {background: inherit;color: inherit;z-index: 1000;opacity: 1;text-indent: 0;}.ace_layer {z-index: 1;position: absolute;overflow: hidden;word-wrap: normal;white-space: pre;height: 100%;width: 100%;-moz-box-sizing: border-box;-webkit-box-sizing: border-box;box-sizing: border-box;pointer-events: none;}.ace_gutter-layer {position: relative;width: auto;text-align: right;pointer-events: auto;}.ace_text-layer {font: inherit !important;}.ace_cjk {display: inline-block;text-align: center;}.ace_cursor-layer {z-index: 4;}.ace_cursor {z-index: 4;position: absolute;-moz-box-sizing: border-box;-webkit-box-sizing: border-box;box-sizing: border-box;border-left: 2px solid;transform: translatez(0);}.ace_multiselect .ace_cursor {border-left-width: 1px;}.ace_slim-cursors .ace_cursor {border-left-width: 1px;}.ace_overwrite-cursors .ace_cursor {border-left-width: 0;border-bottom: 1px solid;}.ace_hidden-cursors .ace_cursor {opacity: 0.2;}.ace_smooth-blinking .ace_cursor {-webkit-transition: opacity 0.18s;transition: opacity 0.18s;}.ace_marker-layer .ace_step, .ace_marker-layer .ace_stack {position: absolute;z-index: 3;}.ace_marker-layer .ace_selection {position: absolute;z-index: 5;}.ace_marker-layer .ace_bracket {position: absolute;z-index: 6;}.ace_marker-layer .ace_active-line {position: absolute;z-index: 2;}.ace_marker-layer .ace_selected-word {position: absolute;z-index: 4;-moz-box-sizing: border-box;-webkit-box-sizing: border-box;box-sizing: border-box;}.ace_line .ace_fold {-moz-box-sizing: border-box;-webkit-box-sizing: border-box;box-sizing: border-box;display: inline-block;height: 11px;margin-top: -2px;vertical-align: middle;background-image:url("data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAABEAAAAJCAYAAADU6McMAAAAGXRFWHRTb2Z0d2FyZQBBZG9iZSBJbWFnZVJlYWR5ccllPAAAAJpJREFUeNpi/P//PwOlgAXGYGRklAVSokD8GmjwY1wasKljQpYACtpCFeADcHVQfQyMQAwzwAZI3wJKvCLkfKBaMSClBlR7BOQikCFGQEErIH0VqkabiGCAqwUadAzZJRxQr/0gwiXIal8zQQPnNVTgJ1TdawL0T5gBIP1MUJNhBv2HKoQHHjqNrA4WO4zY0glyNKLT2KIfIMAAQsdgGiXvgnYAAAAASUVORK5CYII="),url("data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAAEAAAA3CAYAAADNNiA5AAAAGXRFWHRTb2Z0d2FyZQBBZG9iZSBJbWFnZVJlYWR5ccllPAAAACJJREFUeNpi+P//fxgTAwPDBxDxD078RSX+YeEyDFMCIMAAI3INmXiwf2YAAAAASUVORK5CYII=");background-repeat: no-repeat, repeat-x;background-position: center center, top left;color: transparent;border: 1px solid black;border-radius: 2px;cursor: pointer;pointer-events: auto;}.ace_dark .ace_fold {}.ace_fold:hover{background-image:url("data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAABEAAAAJCAYAAADU6McMAAAAGXRFWHRTb2Z0d2FyZQBBZG9iZSBJbWFnZVJlYWR5ccllPAAAAJpJREFUeNpi/P//PwOlgAXGYGRklAVSokD8GmjwY1wasKljQpYACtpCFeADcHVQfQyMQAwzwAZI3wJKvCLkfKBaMSClBlR7BOQikCFGQEErIH0VqkabiGCAqwUadAzZJRxQr/0gwiXIal8zQQPnNVTgJ1TdawL0T5gBIP1MUJNhBv2HKoQHHjqNrA4WO4zY0glyNKLT2KIfIMAAQsdgGiXvgnYAAAAASUVORK5CYII="),url("data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAAEAAAA3CAYAAADNNiA5AAAAGXRFWHRTb2Z0d2FyZQBBZG9iZSBJbWFnZVJlYWR5ccllPAAAACBJREFUeNpi+P//fz4TAwPDZxDxD5X4i5fLMEwJgAADAEPVDbjNw87ZAAAAAElFTkSuQmCC");}.ace_tooltip {background-color: #FFF;background-image: -webkit-linear-gradient(top, transparent, rgba(0, 0, 0, 0.1));background-image: linear-gradient(to bottom, transparent, rgba(0, 0, 0, 0.1));border: 1px solid gray;border-radius: 1px;box-shadow: 0 1px 2px rgba(0, 0, 0, 0.3);color: black;max-width: 100%;padding: 3px 4px;position: fixed;z-index: 999999;-moz-box-sizing: border-box;-webkit-box-sizing: border-box;box-sizing: border-box;cursor: default;white-space: pre;word-wrap: break-word;line-height: normal;font-style: normal;font-weight: normal;letter-spacing: normal;pointer-events: none;}.ace_folding-enabled > .ace_gutter-cell {padding-right: 13px;}.ace_fold-widget {-moz-box-sizing: border-box;-webkit-box-sizing: border-box;box-sizing: border-box;margin: 0 -12px 0 1px;display: none;width: 11px;vertical-align: top;background-image: url("data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAAUAAAAFCAYAAACNbyblAAAANElEQVR42mWKsQ0AMAzC8ixLlrzQjzmBiEjp0A6WwBCSPgKAXoLkqSot7nN3yMwR7pZ32NzpKkVoDBUxKAAAAABJRU5ErkJggg==");background-repeat: no-repeat;background-position: center;border-radius: 3px;border: 1px solid transparent;cursor: pointer;}.ace_folding-enabled .ace_fold-widget {display: inline-block;   }.ace_fold-widget.ace_end {background-image: url("data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAAUAAAAFCAYAAACNbyblAAAANElEQVR42m3HwQkAMAhD0YzsRchFKI7sAikeWkrxwScEB0nh5e7KTPWimZki4tYfVbX+MNl4pyZXejUO1QAAAABJRU5ErkJggg==");}.ace_fold-widget.ace_closed {background-image: url("data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAAMAAAAGCAYAAAAG5SQMAAAAOUlEQVR42jXKwQkAMAgDwKwqKD4EwQ26sSOkVWjgIIHAzPiCgaqiqnJHZnKICBERHN194O5b9vbLuAVRL+l0YWnZAAAAAElFTkSuQmCCXA==");}.ace_fold-widget:hover {border: 1px solid rgba(0, 0, 0, 0.3);background-color: rgba(255, 255, 255, 0.2);box-shadow: 0 1px 1px rgba(255, 255, 255, 0.7);}.ace_fold-widget:active {border: 1px solid rgba(0, 0, 0, 0.4);background-color: rgba(0, 0, 0, 0.05);box-shadow: 0 1px 1px rgba(255, 255, 255, 0.8);}.ace_dark .ace_fold-widget {background-image: url("data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAAUAAAAFCAYAAACNbyblAAAAHklEQVQIW2P4//8/AzoGEQ7oGCaLLAhWiSwB146BAQCSTPYocqT0AAAAAElFTkSuQmCC");}.ace_dark .ace_fold-widget.ace_end {background-image: url("data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAAUAAAAFCAYAAACNbyblAAAAH0lEQVQIW2P4//8/AxQ7wNjIAjDMgC4AxjCVKBirIAAF0kz2rlhxpAAAAABJRU5ErkJggg==");}.ace_dark .ace_fold-widget.ace_closed {background-image: url("data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAAMAAAAFCAYAAACAcVaiAAAAHElEQVQIW2P4//+/AxAzgDADlOOAznHAKgPWAwARji8UIDTfQQAAAABJRU5ErkJggg==");}.ace_dark .ace_fold-widget:hover {box-shadow: 0 1px 1px rgba(255, 255, 255, 0.2);background-color: rgba(255, 255, 255, 0.1);}.ace_dark .ace_fold-widget:active {box-shadow: 0 1px 1px rgba(255, 255, 255, 0.2);}.ace_fold-widget.ace_invalid {background-color: #FFB4B4;border-color: #DE5555;}.ace_fade-fold-widgets .ace_fold-widget {-webkit-transition: opacity 0.4s ease 0.05s;transition: opacity 0.4s ease 0.05s;opacity: 0;}.ace_fade-fold-widgets:hover .ace_fold-widget {-webkit-transition: opacity 0.05s ease 0.05s;transition: opacity 0.05s ease 0.05s;opacity:1;}.ace_underline {text-decoration: underline;}.ace_bold {font-weight: bold;}.ace_nobold .ace_bold {font-weight: normal;}.ace_italic {font-style: italic;}.ace_error-marker {background-color: rgba(255, 0, 0,0.2);position: absolute;z-index: 9;}.ace_highlight-marker {background-color: rgba(255, 255, 0,0.2);position: absolute;z-index: 8;}.ace_br1 {border-top-left-radius    : 3px;}.ace_br2 {border-top-right-radius   : 3px;}.ace_br3 {border-top-left-radius    : 3px; border-top-right-radius:    3px;}.ace_br4 {border-bottom-right-radius: 3px;}.ace_br5 {border-top-left-radius    : 3px; border-bottom-right-radius: 3px;}.ace_br6 {border-top-right-radius   : 3px; border-bottom-right-radius: 3px;}.ace_br7 {border-top-left-radius    : 3px; border-top-right-radius:    3px; border-bottom-right-radius: 3px;}.ace_br8 {border-bottom-left-radius : 3px;}.ace_br9 {border-top-left-radius    : 3px; border-bottom-left-radius:  3px;}.ace_br10{border-top-right-radius   : 3px; border-bottom-left-radius:  3px;}.ace_br11{border-top-left-radius    : 3px; border-top-right-radius:    3px; border-bottom-left-radius:  3px;}.ace_br12{border-bottom-right-radius: 3px; border-bottom-left-radius:  3px;}.ace_br13{border-top-left-radius    : 3px; border-bottom-right-radius: 3px; border-bottom-left-radius:  3px;}.ace_br14{border-top-right-radius   : 3px; border-bottom-right-radius: 3px; border-bottom-left-radius:  3px;}.ace_br15{border-top-left-radius    : 3px; border-top-right-radius:    3px; border-bottom-right-radius: 3px; border-bottom-left-radius: 3px;}.ace_text-input-ios {position: absolute !important;top: -100000px !important;left: -100000px !important;}',"ace_editor.css");var m=function(e,t){var n=this;this.container=e||r.createElement("div"),this.$keepTextAreaAtCursor=!s.isOldIE,r.addCssClass(this.container,"ace_editor"),this.setTheme(t),this.$gutter=r.createElement("div"),this.$gutter.className="ace_gutter",this.container.appendChild(this.$gutter),this.$gutter.setAttribute("aria-hidden",!0),this.scroller=r.createElement("div"),this.scroller.className="ace_scroller",this.container.appendChild(this.scroller),this.content=r.createElement("div"),this.content.className="ace_content",this.scroller.appendChild(this.content),this.$gutterLayer=new a(this.$gutter),this.$gutterLayer.on("changeGutterWidth",this.onGutterResize.bind(this)),this.$markerBack=new l(this.content);var i=this.$textLayer=new u(this.content);this.canvas=i.element,this.$markerFront=new l(this.content),this.$cursorLayer=new c(this.content),this.$horizScroll=!1,this.$vScroll=!1,this.scrollBar=this.scrollBarV=new f(this.container,this),this.scrollBarH=new h(this.container,this),this.scrollBarV.addEventListener("scroll",(function(e){n.$scrollAnimation||n.session.setScrollTop(e.data-n.scrollMargin.top)})),this.scrollBarH.addEventListener("scroll",(function(e){n.$scrollAnimation||n.session.setScrollLeft(e.data-n.scrollMargin.left)})),this.scrollTop=0,this.scrollLeft=0,this.cursorPos={row:0,column:0},this.$fontMetrics=new p(this.container),this.$textLayer.$setFontMetrics(this.$fontMetrics),this.$textLayer.addEventListener("changeCharacterSize",(function(e){n.updateCharacterSize(),n.onResize(!0,n.gutterWidth,n.$size.width,n.$size.height),n._signal("changeCharacterSize",e)})),this.$size={width:0,height:0,scrollerHeight:0,scrollerWidth:0,$dirty:!0},this.layerConfig={width:1,padding:0,firstRow:0,firstRowScreen:0,lastRow:0,lineHeight:0,characterWidth:0,minHeight:1,maxHeight:1,offset:0,height:1,gutterOffset:1},this.scrollMargin={left:0,right:0,top:0,bottom:0,v:0,h:0},this.$loop=new d(this.$renderChanges.bind(this),this.container.ownerDocument.defaultView),this.$loop.schedule(this.CHANGE_FULL),this.updateCharacterSize(),this.setPadding(4),o.resetOptions(this),o._emit("renderer",this)};(function(){this.CHANGE_CURSOR=1,this.CHANGE_MARKER=2,this.CHANGE_GUTTER=4,this.CHANGE_SCROLL=8,this.CHANGE_LINES=16,this.CHANGE_TEXT=32,this.CHANGE_SIZE=64,this.CHANGE_MARKER_BACK=128,this.CHANGE_MARKER_FRONT=256,this.CHANGE_FULL=512,this.CHANGE_H_SCROLL=1024,i.implement(this,g),this.updateCharacterSize=function(){this.$textLayer.allowBoldFonts!=this.$allowBoldFonts&&(this.$allowBoldFonts=this.$textLayer.allowBoldFonts,this.setStyle("ace_nobold",!this.$allowBoldFonts)),this.layerConfig.characterWidth=this.characterWidth=this.$textLayer.getCharacterWidth(),this.layerConfig.lineHeight=this.lineHeight=this.$textLayer.getLineHeight(),this.$updatePrintMargin()},this.setSession=function(e){this.session&&this.session.doc.off("changeNewLineMode",this.onChangeNewLineMode),this.session=e,e&&this.scrollMargin.top&&e.getScrollTop()<=0&&e.setScrollTop(-this.scrollMargin.top),this.$cursorLayer.setSession(e),this.$markerBack.setSession(e),this.$markerFront.setSession(e),this.$gutterLayer.setSession(e),this.$textLayer.setSession(e),e&&(this.$loop.schedule(this.CHANGE_FULL),this.session.$setFontMetrics(this.$fontMetrics),this.scrollBarH.scrollLeft=this.scrollBarV.scrollTop=null,this.onChangeNewLineMode=this.onChangeNewLineMode.bind(this),this.onChangeNewLineMode(),this.session.doc.on("changeNewLineMode",this.onChangeNewLineMode))},this.updateLines=function(e,t,n){if(void 0===t&&(t=1/0),this.$changedLines?(this.$changedLines.firstRow>e&&(this.$changedLines.firstRow=e),this.$changedLines.lastRow<t&&(this.$changedLines.lastRow=t)):this.$changedLines={firstRow:e,lastRow:t},this.$changedLines.lastRow<this.layerConfig.firstRow){if(!n)return;this.$changedLines.lastRow=this.layerConfig.lastRow}this.$changedLines.firstRow>this.layerConfig.lastRow||this.$loop.schedule(this.CHANGE_LINES)},this.onChangeNewLineMode=function(){this.$loop.schedule(this.CHANGE_TEXT),this.$textLayer.$updateEolChar(),this.session.$bidiHandler.setEolChar(this.$textLayer.EOL_CHAR)},this.onChangeTabSize=function(){this.$loop.schedule(this.CHANGE_TEXT|this.CHANGE_MARKER),this.$textLayer.onChangeTabSize()},this.updateText=function(){this.$loop.schedule(this.CHANGE_TEXT)},this.updateFull=function(e){e?this.$renderChanges(this.CHANGE_FULL,!0):this.$loop.schedule(this.CHANGE_FULL)},this.updateFontSize=function(){this.$textLayer.checkForSizeChanges()},this.$changes=0,this.$updateSizeAsync=function(){this.$loop.pending?this.$size.$dirty=!0:this.onResize()},this.onResize=function(e,t,n,i){if(!(this.resizing>2)){this.resizing>0?this.resizing++:this.resizing=e?1:0;var r=this.container;i||(i=r.clientHeight||r.scrollHeight),n||(n=r.clientWidth||r.scrollWidth);var o=this.$updateCachedSize(e,t,n,i);if(!this.$size.scrollerHeight||!n&&!i)return this.resizing=0;e&&(this.$gutterLayer.$padding=null),e?this.$renderChanges(o|this.$changes,!0):this.$loop.schedule(o|this.$changes),this.resizing&&(this.resizing=0),this.scrollBarV.scrollLeft=this.scrollBarV.scrollTop=null}},this.$updateCachedSize=function(e,t,n,i){i-=this.$extraHeight||0;var r=0,o=this.$size,s={width:o.width,height:o.height,scrollerHeight:o.scrollerHeight,scrollerWidth:o.scrollerWidth};return i&&(e||o.height!=i)&&(o.height=i,r|=this.CHANGE_SIZE,o.scrollerHeight=o.height,this.$horizScroll&&(o.scrollerHeight-=this.scrollBarH.getHeight()),this.scrollBarV.element.style.bottom=this.scrollBarH.getHeight()+"px",r|=this.CHANGE_SCROLL),n&&(e||o.width!=n)&&(r|=this.CHANGE_SIZE,o.width=n,null==t&&(t=this.$showGutter?this.$gutter.offsetWidth:0),this.gutterWidth=t,this.scrollBarH.element.style.left=this.scroller.style.left=t+"px",o.scrollerWidth=Math.max(0,n-t-this.scrollBarV.getWidth()),this.scrollBarH.element.style.right=this.scroller.style.right=this.scrollBarV.getWidth()+"px",this.scroller.style.bottom=this.scrollBarH.getHeight()+"px",(this.session&&this.session.getUseWrapMode()&&this.adjustWrapLimit()||e)&&(r|=this.CHANGE_FULL)),o.$dirty=!n||!i,r&&this._signal("resize",s),r},this.onGutterResize=function(){var e=this.$showGutter?this.$gutter.offsetWidth:0;e!=this.gutterWidth&&(this.$changes|=this.$updateCachedSize(!0,e,this.$size.width,this.$size.height)),this.session.getUseWrapMode()&&this.adjustWrapLimit()?this.$loop.schedule(this.CHANGE_FULL):this.$size.$dirty?this.$loop.schedule(this.CHANGE_FULL):(this.$computeLayerConfig(),this.$loop.schedule(this.CHANGE_MARKER))},this.adjustWrapLimit=function(){var e=this.$size.scrollerWidth-2*this.$padding,t=Math.floor(e/this.characterWidth);return this.session.adjustWrapLimit(t,this.$showPrintMargin&&this.$printMarginColumn)},this.setAnimatedScroll=function(e){this.setOption("animatedScroll",e)},this.getAnimatedScroll=function(){return this.$animatedScroll},this.setShowInvisibles=function(e){this.setOption("showInvisibles",e),this.session.$bidiHandler.setShowInvisibles(e)},this.getShowInvisibles=function(){return this.getOption("showInvisibles")},this.getDisplayIndentGuides=function(){return this.getOption("displayIndentGuides")},this.setDisplayIndentGuides=function(e){this.setOption("displayIndentGuides",e)},this.setShowPrintMargin=function(e){this.setOption("showPrintMargin",e)},this.getShowPrintMargin=function(){return this.getOption("showPrintMargin")},this.setPrintMarginColumn=function(e){this.setOption("printMarginColumn",e)},this.getPrintMarginColumn=function(){return this.getOption("printMarginColumn")},this.getShowGutter=function(){return this.getOption("showGutter")},this.setShowGutter=function(e){return this.setOption("showGutter",e)},this.getFadeFoldWidgets=function(){return this.getOption("fadeFoldWidgets")},this.setFadeFoldWidgets=function(e){this.setOption("fadeFoldWidgets",e)},this.setHighlightGutterLine=function(e){this.setOption("highlightGutterLine",e)},this.getHighlightGutterLine=function(){return this.getOption("highlightGutterLine")},this.$updateGutterLineHighlight=function(){var e=this.$cursorLayer.$pixelPos,t=this.layerConfig.lineHeight;if(this.session.getUseWrapMode()){var n=this.session.selection.getCursor();n.column=0,e=this.$cursorLayer.getPixelPosition(n,!0),t*=this.session.getRowLength(n.row)}this.$gutterLineHighlight.style.top=e.top-this.layerConfig.offset+"px",this.$gutterLineHighlight.style.height=t+"px"},this.$updatePrintMargin=function(){if(this.$showPrintMargin||this.$printMarginEl){if(!this.$printMarginEl){var e=r.createElement("div");e.className="ace_layer ace_print-margin-layer",this.$printMarginEl=r.createElement("div"),this.$printMarginEl.className="ace_print-margin",e.appendChild(this.$printMarginEl),this.content.insertBefore(e,this.content.firstChild)}var t=this.$printMarginEl.style;t.left=this.characterWidth*this.$printMarginColumn+this.$padding+"px",t.visibility=this.$showPrintMargin?"visible":"hidden",this.session&&-1==this.session.$wrap&&this.adjustWrapLimit()}},this.getContainerElement=function(){return this.container},this.getMouseEventTarget=function(){return this.scroller},this.getTextAreaContainer=function(){return this.container},this.$moveTextAreaToCursor=function(){if(this.$keepTextAreaAtCursor){var e=this.layerConfig,t=this.$cursorLayer.$pixelPos.top,n=this.$cursorLayer.$pixelPos.left;t-=e.offset;var i=this.textarea.style,r=this.lineHeight;if(t<0||t>e.height-r)i.top=i.left="0";else{var o=this.characterWidth;if(this.$composition){var s=this.textarea.value.replace(/^\x01+/,"");o*=this.session.$getStringScreenWidth(s)[0]+2,r+=2}(n-=this.scrollLeft)>this.$size.scrollerWidth-o&&(n=this.$size.scrollerWidth-o),n+=this.gutterWidth,i.height=r+"px",i.width=o+"px",i.left=Math.min(n,this.$size.scrollerWidth-o)+"px",i.top=Math.min(t,this.$size.height-r)+"px"}}},this.getFirstVisibleRow=function(){return this.layerConfig.firstRow},this.getFirstFullyVisibleRow=function(){return this.layerConfig.firstRow+(0===this.layerConfig.offset?0:1)},this.getLastFullyVisibleRow=function(){var e=this.layerConfig,t=e.lastRow;return this.session.documentToScreenRow(t,0)*e.lineHeight-this.session.getScrollTop()>e.height-e.lineHeight?t-1:t},this.getLastVisibleRow=function(){return this.layerConfig.lastRow},this.$padding=null,this.setPadding=function(e){this.$padding=e,this.$textLayer.setPadding(e),this.$cursorLayer.setPadding(e),this.$markerFront.setPadding(e),this.$markerBack.setPadding(e),this.$loop.schedule(this.CHANGE_FULL),this.$updatePrintMargin()},this.setScrollMargin=function(e,t,n,i){var r=this.scrollMargin;r.top=0|e,r.bottom=0|t,r.right=0|i,r.left=0|n,r.v=r.top+r.bottom,r.h=r.left+r.right,r.top&&this.scrollTop<=0&&this.session&&this.session.setScrollTop(-r.top),this.updateFull()},this.getHScrollBarAlwaysVisible=function(){return this.$hScrollBarAlwaysVisible},this.setHScrollBarAlwaysVisible=function(e){this.setOption("hScrollBarAlwaysVisible",e)},this.getVScrollBarAlwaysVisible=function(){return this.$vScrollBarAlwaysVisible},this.setVScrollBarAlwaysVisible=function(e){this.setOption("vScrollBarAlwaysVisible",e)},this.$updateScrollBarV=function(){var e=this.layerConfig.maxHeight,t=this.$size.scrollerHeight;!this.$maxLines&&this.$scrollPastEnd&&(e-=(t-this.lineHeight)*this.$scrollPastEnd,this.scrollTop>e-t&&(e=this.scrollTop+t,this.scrollBarV.scrollTop=null)),this.scrollBarV.setScrollHeight(e+this.scrollMargin.v),this.scrollBarV.setScrollTop(this.scrollTop+this.scrollMargin.top)},this.$updateScrollBarH=function(){this.scrollBarH.setScrollWidth(this.layerConfig.width+2*this.$padding+this.scrollMargin.h),this.scrollBarH.setScrollLeft(this.scrollLeft+this.scrollMargin.left)},this.$frozen=!1,this.freeze=function(){this.$frozen=!0},this.unfreeze=function(){this.$frozen=!1},this.$renderChanges=function(e,t){if(this.$changes&&(e|=this.$changes,this.$changes=0),this.session&&this.container.offsetWidth&&!this.$frozen&&(e||t)){if(this.$size.$dirty)return this.$changes|=e,this.onResize(!0);this.lineHeight||this.$textLayer.checkForSizeChanges(),this._signal("beforeRender"),this.session&&this.session.$bidiHandler&&this.session.$bidiHandler.updateCharacterWidths(this.$fontMetrics);var n=this.layerConfig;if(e&this.CHANGE_FULL||e&this.CHANGE_SIZE||e&this.CHANGE_TEXT||e&this.CHANGE_LINES||e&this.CHANGE_SCROLL||e&this.CHANGE_H_SCROLL){if(e|=this.$computeLayerConfig(),n.firstRow!=this.layerConfig.firstRow&&n.firstRowScreen==this.layerConfig.firstRowScreen){var i=this.scrollTop+(n.firstRow-this.layerConfig.firstRow)*this.lineHeight;i>0&&(this.scrollTop=i,e|=this.CHANGE_SCROLL,e|=this.$computeLayerConfig())}n=this.layerConfig,this.$updateScrollBarV(),e&this.CHANGE_H_SCROLL&&this.$updateScrollBarH(),this.$gutterLayer.element.style.marginTop=-n.offset+"px",this.content.style.marginTop=-n.offset+"px",this.content.style.width=n.width+2*this.$padding+"px",this.content.style.height=n.minHeight+"px"}if(e&this.CHANGE_H_SCROLL&&(this.content.style.marginLeft=-this.scrollLeft+"px",this.scroller.className=this.scrollLeft<=0?"ace_scroller":"ace_scroller ace_scroll-left"),e&this.CHANGE_FULL)return this.$textLayer.update(n),this.$showGutter&&this.$gutterLayer.update(n),this.$markerBack.update(n),this.$markerFront.update(n),this.$cursorLayer.update(n),this.$moveTextAreaToCursor(),this.$highlightGutterLine&&this.$updateGutterLineHighlight(),void this._signal("afterRender");if(e&this.CHANGE_SCROLL)return e&this.CHANGE_TEXT||e&this.CHANGE_LINES?this.$textLayer.update(n):this.$textLayer.scrollLines(n),this.$showGutter&&this.$gutterLayer.update(n),this.$markerBack.update(n),this.$markerFront.update(n),this.$cursorLayer.update(n),this.$highlightGutterLine&&this.$updateGutterLineHighlight(),this.$moveTextAreaToCursor(),void this._signal("afterRender");e&this.CHANGE_TEXT?(this.$textLayer.update(n),this.$showGutter&&this.$gutterLayer.update(n)):e&this.CHANGE_LINES?(this.$updateLines()||e&this.CHANGE_GUTTER&&this.$showGutter)&&this.$gutterLayer.update(n):(e&this.CHANGE_TEXT||e&this.CHANGE_GUTTER)&&this.$showGutter&&this.$gutterLayer.update(n),e&this.CHANGE_CURSOR&&(this.$cursorLayer.update(n),this.$moveTextAreaToCursor(),this.$highlightGutterLine&&this.$updateGutterLineHighlight()),e&(this.CHANGE_MARKER|this.CHANGE_MARKER_FRONT)&&this.$markerFront.update(n),e&(this.CHANGE_MARKER|this.CHANGE_MARKER_BACK)&&this.$markerBack.update(n),this._signal("afterRender")}else this.$changes|=e},this.$autosize=function(){var e=this.session.getScreenLength()*this.lineHeight,t=this.$maxLines*this.lineHeight,n=Math.min(t,Math.max((this.$minLines||1)*this.lineHeight,e))+this.scrollMargin.v+(this.$extraHeight||0);this.$horizScroll&&(n+=this.scrollBarH.getHeight()),this.$maxPixelHeight&&n>this.$maxPixelHeight&&(n=this.$maxPixelHeight);var i=e>t;if(n!=this.desiredHeight||this.$size.height!=this.desiredHeight||i!=this.$vScroll){i!=this.$vScroll&&(this.$vScroll=i,this.scrollBarV.setVisible(i));var r=this.container.clientWidth;this.container.style.height=n+"px",this.$updateCachedSize(!0,this.$gutterWidth,r,n),this.desiredHeight=n,this._signal("autosize")}},this.$computeLayerConfig=function(){var e=this.session,t=this.$size,n=t.height<=2*this.lineHeight,i=this.session.getScreenLength()*this.lineHeight,r=this.$getLongestLine(),o=!n&&(this.$hScrollBarAlwaysVisible||t.scrollerWidth-r-2*this.$padding<0),s=this.$horizScroll!==o;s&&(this.$horizScroll=o,this.scrollBarH.setVisible(o));var a=this.$vScroll;this.$maxLines&&this.lineHeight>1&&this.$autosize();var l=this.scrollTop%this.lineHeight,u=t.scrollerHeight+this.lineHeight,c=!this.$maxLines&&this.$scrollPastEnd?(t.scrollerHeight-this.lineHeight)*this.$scrollPastEnd:0;i+=c;var h=this.scrollMargin;this.session.setScrollTop(Math.max(-h.top,Math.min(this.scrollTop,i-t.scrollerHeight+h.bottom))),this.session.setScrollLeft(Math.max(-h.left,Math.min(this.scrollLeft,r+2*this.$padding-t.scrollerWidth+h.right)));var f=!n&&(this.$vScrollBarAlwaysVisible||t.scrollerHeight-i+c<0||this.scrollTop>h.top),d=a!==f;d&&(this.$vScroll=f,this.scrollBarV.setVisible(f));var p,g,m=Math.ceil(u/this.lineHeight)-1,v=Math.max(0,Math.round((this.scrollTop-l)/this.lineHeight)),b=v+m,y=this.lineHeight;v=e.screenToDocumentRow(v,0);var w=e.getFoldLine(v);w&&(v=w.start.row),p=e.documentToScreenRow(v,0),g=e.getRowLength(v)*y,b=Math.min(e.screenToDocumentRow(b,0),e.getLength()-1),u=t.scrollerHeight+e.getRowLength(b)*y+g,l=this.scrollTop-p*y;var A=0;return this.layerConfig.width!=r&&(A=this.CHANGE_H_SCROLL),(s||d)&&(A=this.$updateCachedSize(!0,this.gutterWidth,t.width,t.height),this._signal("scrollbarVisibilityChanged"),d&&(r=this.$getLongestLine())),this.layerConfig={width:r,padding:this.$padding,firstRow:v,firstRowScreen:p,lastRow:b,lineHeight:y,characterWidth:this.characterWidth,minHeight:u,maxHeight:i,offset:l,gutterOffset:y?Math.max(0,Math.ceil((l+t.height-t.scrollerHeight)/y)):0,height:this.$size.scrollerHeight},A},this.$updateLines=function(){if(this.$changedLines){var e=this.$changedLines.firstRow,t=this.$changedLines.lastRow;this.$changedLines=null;var n=this.layerConfig;if(!(e>n.lastRow+1||t<n.firstRow))return t===1/0?(this.$showGutter&&this.$gutterLayer.update(n),void this.$textLayer.update(n)):(this.$textLayer.updateLines(n,e,t),!0)}},this.$getLongestLine=function(){var e=this.session.getScreenWidth();return this.showInvisibles&&!this.session.$useWrapMode&&(e+=1),Math.max(this.$size.scrollerWidth-2*this.$padding,Math.round(e*this.characterWidth))},this.updateFrontMarkers=function(){this.$markerFront.setMarkers(this.session.getMarkers(!0)),this.$loop.schedule(this.CHANGE_MARKER_FRONT)},this.updateBackMarkers=function(){this.$markerBack.setMarkers(this.session.getMarkers()),this.$loop.schedule(this.CHANGE_MARKER_BACK)},this.addGutterDecoration=function(e,t){this.$gutterLayer.addGutterDecoration(e,t)},this.removeGutterDecoration=function(e,t){this.$gutterLayer.removeGutterDecoration(e,t)},this.updateBreakpoints=function(e){this.$loop.schedule(this.CHANGE_GUTTER)},this.setAnnotations=function(e){this.$gutterLayer.setAnnotations(e),this.$loop.schedule(this.CHANGE_GUTTER)},this.updateCursor=function(){this.$loop.schedule(this.CHANGE_CURSOR)},this.hideCursor=function(){this.$cursorLayer.hideCursor()},this.showCursor=function(){this.$cursorLayer.showCursor()},this.scrollSelectionIntoView=function(e,t,n){this.scrollCursorIntoView(e,n),this.scrollCursorIntoView(t,n)},this.scrollCursorIntoView=function(e,t,n){if(0!==this.$size.scrollerHeight){var i=this.$cursorLayer.getPixelPosition(e),r=i.left,o=i.top,s=n&&n.top||0,a=n&&n.bottom||0,l=this.$scrollAnimation?this.session.getScrollTop():this.scrollTop;l+s>o?(t&&l+s>o+this.lineHeight&&(o-=t*this.$size.scrollerHeight),0===o&&(o=-this.scrollMargin.top),this.session.setScrollTop(o)):l+this.$size.scrollerHeight-a<o+this.lineHeight&&(t&&l+this.$size.scrollerHeight-a<o-this.lineHeight&&(o+=t*this.$size.scrollerHeight),this.session.setScrollTop(o+this.lineHeight-this.$size.scrollerHeight));var u=this.scrollLeft;u>r?(r<this.$padding+2*this.layerConfig.characterWidth&&(r=-this.scrollMargin.left),this.session.setScrollLeft(r)):u+this.$size.scrollerWidth<r+this.characterWidth?this.session.setScrollLeft(Math.round(r+this.characterWidth-this.$size.scrollerWidth)):u<=this.$padding&&r-u<this.characterWidth&&this.session.setScrollLeft(0)}},this.getScrollTop=function(){return this.session.getScrollTop()},this.getScrollLeft=function(){return this.session.getScrollLeft()},this.getScrollTopRow=function(){return this.scrollTop/this.lineHeight},this.getScrollBottomRow=function(){return Math.max(0,Math.floor((this.scrollTop+this.$size.scrollerHeight)/this.lineHeight)-1)},this.scrollToRow=function(e){this.session.setScrollTop(e*this.lineHeight)},this.alignCursor=function(e,t){"number"==typeof e&&(e={row:e,column:0});var n=this.$cursorLayer.getPixelPosition(e),i=this.$size.scrollerHeight-this.lineHeight,r=n.top-i*(t||0);return this.session.setScrollTop(r),r},this.STEPS=8,this.$calcSteps=function(e,t){var n,i,r=0,o=this.STEPS,s=[];for(r=0;r<o;++r)s.push((n=r/this.STEPS,i=e,(t-e)*(Math.pow(n-1,3)+1)+i));return s},this.scrollToLine=function(e,t,n,i){var r=this.$cursorLayer.getPixelPosition({row:e,column:0}).top;t&&(r-=this.$size.scrollerHeight/2);var o=this.scrollTop;this.session.setScrollTop(r),!1!==n&&this.animateScrolling(o,i)},this.animateScrolling=function(e,t){var n=this.scrollTop;if(this.$animatedScroll){var i=this;if(e!=n){if(this.$scrollAnimation){var r=this.$scrollAnimation.steps;if(r.length&&(e=r[0])==n)return}var o=i.$calcSteps(e,n);this.$scrollAnimation={from:e,to:n,steps:o},clearInterval(this.$timer),i.session.setScrollTop(o.shift()),i.session.$scrollTop=n,this.$timer=setInterval((function(){o.length?(i.session.setScrollTop(o.shift()),i.session.$scrollTop=n):null!=n?(i.session.$scrollTop=-1,i.session.setScrollTop(n),n=null):(i.$timer=clearInterval(i.$timer),i.$scrollAnimation=null,t&&t())}),10)}}},this.scrollToY=function(e){this.scrollTop!==e&&(this.$loop.schedule(this.CHANGE_SCROLL),this.scrollTop=e)},this.scrollToX=function(e){this.scrollLeft!==e&&(this.scrollLeft=e),this.$loop.schedule(this.CHANGE_H_SCROLL)},this.scrollTo=function(e,t){this.session.setScrollTop(t),this.session.setScrollLeft(t)},this.scrollBy=function(e,t){t&&this.session.setScrollTop(this.session.getScrollTop()+t),e&&this.session.setScrollLeft(this.session.getScrollLeft()+e)},this.isScrollableBy=function(e,t){return t<0&&this.session.getScrollTop()>=1-this.scrollMargin.top||(t>0&&this.session.getScrollTop()+this.$size.scrollerHeight-this.layerConfig.maxHeight<-1+this.scrollMargin.bottom||(e<0&&this.session.getScrollLeft()>=1-this.scrollMargin.left||(e>0&&this.session.getScrollLeft()+this.$size.scrollerWidth-this.layerConfig.width<-1+this.scrollMargin.right||void 0)))},this.pixelToScreenCoordinates=function(e,t){var n=this.scroller.getBoundingClientRect(),i=e+this.scrollLeft-n.left-this.$padding,r=i/this.characterWidth,o=Math.floor((t+this.scrollTop-n.top)/this.lineHeight),s=Math.round(r);return{row:o,column:s,side:r-s>0?1:-1,offsetX:i}},this.screenToTextCoordinates=function(e,t){var n=this.scroller.getBoundingClientRect(),i=e+this.scrollLeft-n.left-this.$padding,r=Math.round(i/this.characterWidth),o=(t+this.scrollTop-n.top)/this.lineHeight;return this.session.screenToDocumentPosition(o,Math.max(r,0),i)},this.textToScreenCoordinates=function(e,t){var n=this.scroller.getBoundingClientRect(),i=this.session.documentToScreenPosition(e,t),r=this.$padding+(this.session.$bidiHandler.isBidiRow(i.row,e)?this.session.$bidiHandler.getPosLeft(i.column):Math.round(i.column*this.characterWidth)),o=i.row*this.lineHeight;return{pageX:n.left+r-this.scrollLeft,pageY:n.top+o-this.scrollTop}},this.visualizeFocus=function(){r.addCssClass(this.container,"ace_focus")},this.visualizeBlur=function(){r.removeCssClass(this.container,"ace_focus")},this.showComposition=function(e){this.$composition||(this.$composition={keepTextAreaAtCursor:this.$keepTextAreaAtCursor,cssText:this.textarea.style.cssText}),this.$keepTextAreaAtCursor=!0,r.addCssClass(this.textarea,"ace_composition"),this.textarea.style.cssText="",this.$moveTextAreaToCursor()},this.setCompositionText=function(e){this.$moveTextAreaToCursor()},this.hideComposition=function(){this.$composition&&(r.removeCssClass(this.textarea,"ace_composition"),this.$keepTextAreaAtCursor=this.$composition.keepTextAreaAtCursor,this.textarea.style.cssText=this.$composition.cssText,this.$composition=null)},this.setTheme=function(e,t){var n=this;if(this.$themeId=e,n._dispatchEvent("themeChange",{theme:e}),e&&"string"!=typeof e)s(e);else{var i=e||this.$options.theme.initialValue;o.loadModule(["theme",i],s)}function s(i){if(n.$themeId!=e)return t&&t();if(!i||!i.cssClass)throw new Error("couldn't load module "+e+" or it didn't call define");r.importCssString(i.cssText,i.cssClass,n.container.ownerDocument),n.theme&&r.removeCssClass(n.container,n.theme.cssClass);var o="padding"in i?i.padding:"padding"in(n.theme||{})?4:n.$padding;n.$padding&&o!=n.$padding&&n.setPadding(o),n.$theme=i.cssClass,n.theme=i,r.addCssClass(n.container,i.cssClass),r.setCssClass(n.container,"ace_dark",i.isDark),n.$size&&(n.$size.width=0,n.$updateSizeAsync()),n._dispatchEvent("themeLoaded",{theme:i}),t&&t()}},this.getTheme=function(){return this.$themeId},this.setStyle=function(e,t){r.setCssClass(this.container,e,!1!==t)},this.unsetStyle=function(e){r.removeCssClass(this.container,e)},this.setCursorStyle=function(e){this.scroller.style.cursor!=e&&(this.scroller.style.cursor=e)},this.setMouseCursor=function(e){this.scroller.style.cursor=e},this.destroy=function(){this.$textLayer.destroy(),this.$cursorLayer.destroy()}}).call(m.prototype),o.defineOptions(m.prototype,"renderer",{animatedScroll:{initialValue:!1},showInvisibles:{set:function(e){this.$textLayer.setShowInvisibles(e)&&this.$loop.schedule(this.CHANGE_TEXT)},initialValue:!1},showPrintMargin:{set:function(){this.$updatePrintMargin()},initialValue:!0},printMarginColumn:{set:function(){this.$updatePrintMargin()},initialValue:80},printMargin:{set:function(e){"number"==typeof e&&(this.$printMarginColumn=e),this.$showPrintMargin=!!e,this.$updatePrintMargin()},get:function(){return this.$showPrintMargin&&this.$printMarginColumn}},showGutter:{set:function(e){this.$gutter.style.display=e?"block":"none",this.$loop.schedule(this.CHANGE_FULL),this.onGutterResize()},initialValue:!0},fadeFoldWidgets:{set:function(e){r.setCssClass(this.$gutter,"ace_fade-fold-widgets",e)},initialValue:!1},showFoldWidgets:{set:function(e){this.$gutterLayer.setShowFoldWidgets(e)},initialValue:!0},showLineNumbers:{set:function(e){this.$gutterLayer.setShowLineNumbers(e),this.$loop.schedule(this.CHANGE_GUTTER)},initialValue:!0},displayIndentGuides:{set:function(e){this.$textLayer.setDisplayIndentGuides(e)&&this.$loop.schedule(this.CHANGE_TEXT)},initialValue:!0},highlightGutterLine:{set:function(e){if(!this.$gutterLineHighlight)return this.$gutterLineHighlight=r.createElement("div"),this.$gutterLineHighlight.className="ace_gutter-active-line",void this.$gutter.appendChild(this.$gutterLineHighlight);this.$gutterLineHighlight.style.display=e?"":"none",this.$cursorLayer.$pixelPos&&this.$updateGutterLineHighlight()},initialValue:!1,value:!0},hScrollBarAlwaysVisible:{set:function(e){this.$hScrollBarAlwaysVisible&&this.$horizScroll||this.$loop.schedule(this.CHANGE_SCROLL)},initialValue:!1},vScrollBarAlwaysVisible:{set:function(e){this.$vScrollBarAlwaysVisible&&this.$vScroll||this.$loop.schedule(this.CHANGE_SCROLL)},initialValue:!1},fontSize:{set:function(e){"number"==typeof e&&(e+="px"),this.container.style.fontSize=e,this.updateFontSize()},initialValue:12},fontFamily:{set:function(e){this.container.style.fontFamily=e,this.updateFontSize()}},maxLines:{set:function(e){this.updateFull()}},minLines:{set:function(e){this.updateFull()}},maxPixelHeight:{set:function(e){this.updateFull()},initialValue:0},scrollPastEnd:{set:function(e){e=+e||0,this.$scrollPastEnd!=e&&(this.$scrollPastEnd=e,this.$loop.schedule(this.CHANGE_SCROLL))},initialValue:0,handlesSet:!0},fixedWidthGutter:{set:function(e){this.$gutterLayer.$fixedWidth=!!e,this.$loop.schedule(this.CHANGE_GUTTER)}},theme:{set:function(e){this.setTheme(e)},get:function(){return this.$themeId||this.theme},initialValue:"./theme/textmate",handlesSet:!0}}),t.VirtualRenderer=m})),ace.define("ace/worker/worker_client",["require","exports","module","ace/lib/oop","ace/lib/net","ace/lib/event_emitter","ace/config"],(function(e,t,n){"use strict";var i=e("../lib/oop"),r=e("../lib/net"),o=e("../lib/event_emitter").EventEmitter,s=e("../config");function a(e,t){var n=function(e,t){var n=t.src;r.qualifyURL(e);try{return new Blob([n],{type:"application/javascript"})}catch(e){var i=new(window.BlobBuilder||window.WebKitBlobBuilder||window.MozBlobBuilder);return i.append(n),i.getBlob("application/javascript")}}(e,t),i=(window.URL||window.webkitURL).createObjectURL(n);return new Worker(i)}var l=function(t,n,i,r,o){if(this.$sendDeltaQueue=this.$sendDeltaQueue.bind(this),this.changeListener=this.changeListener.bind(this),this.onMessage=this.onMessage.bind(this),e.nameToUrl&&!e.toUrl&&(e.toUrl=e.nameToUrl),s.get("packaged")||!e.toUrl)r=r||s.moduleUrl(n.id,"worker");else{var l=this.$normalizePath;r=r||l(e.toUrl("ace/worker/worker.js",null,"_"));var u={};t.forEach((function(t){u[t]=l(e.toUrl(t,null,"_").replace(/(\.js)?(\?.*)?$/,""))}))}this.$worker=a(r,n),o&&this.send("importScripts",o),this.$worker.postMessage({init:!0,tlns:u,module:n.id,classname:i}),this.callbackId=1,this.callbacks={},this.$worker.onmessage=this.onMessage};(function(){i.implement(this,o),this.onMessage=function(e){var t=e.data;switch(t.type){case"event":this._signal(t.name,{data:t.data});break;case"call":var n=this.callbacks[t.id];n&&(n(t.data),delete this.callbacks[t.id]);break;case"error":this.reportError(t.data);break;case"log":window.console&&console.log&&console.log.apply(console,t.data)}},this.reportError=function(e){window.console&&console.error&&console.error(e)},this.$normalizePath=function(e){return r.qualifyURL(e)},this.terminate=function(){this._signal("terminate",{}),this.deltaQueue=null,this.$worker.terminate(),this.$worker=null,this.$doc&&this.$doc.off("change",this.changeListener),this.$doc=null},this.send=function(e,t){this.$worker.postMessage({command:e,args:t})},this.call=function(e,t,n){if(n){var i=this.callbackId++;this.callbacks[i]=n,t.push(i)}this.send(e,t)},this.emit=function(e,t){try{this.$worker.postMessage({event:e,data:{data:t.data}})}catch(e){console.error(e.stack)}},this.attachToDocument=function(e){this.$doc&&this.terminate(),this.$doc=e,this.call("setValue",[e.getValue()]),e.on("change",this.changeListener)},this.changeListener=function(e){this.deltaQueue||(this.deltaQueue=[],setTimeout(this.$sendDeltaQueue,0)),"insert"==e.action?this.deltaQueue.push(e.start,e.lines):this.deltaQueue.push(e.start,e.end)},this.$sendDeltaQueue=function(){var e=this.deltaQueue;e&&(this.deltaQueue=null,e.length>50&&e.length>this.$doc.getLength()>>1?this.call("setValue",[this.$doc.getValue()]):this.emit("change",{data:e}))}}).call(l.prototype);var u=function(e,t,n){this.$sendDeltaQueue=this.$sendDeltaQueue.bind(this),this.changeListener=this.changeListener.bind(this),this.callbackId=1,this.callbacks={},this.messageBuffer=[];var i=null,r=!1,a=Object.create(o),l=this;this.$worker={},this.$worker.terminate=function(){},this.$worker.postMessage=function(e){l.messageBuffer.push(e),i&&(r?setTimeout(u):u())},this.setEmitSync=function(e){r=e};var u=function(){var e=l.messageBuffer.shift();e.command?i[e.command].apply(i,e.args):e.event&&a._signal(e.event,e.data)};a.postMessage=function(e){l.onMessage({data:e})},a.callback=function(e,t){this.postMessage({type:"call",id:t,data:e})},a.emit=function(e,t){this.postMessage({type:"event",name:e,data:t})},s.loadModule(["worker",t],(function(e){for(i=new e[n](a);l.messageBuffer.length;)u()}))};u.prototype=l.prototype,t.UIWorkerClient=u,t.WorkerClient=l,t.createWorker=a})),ace.define("ace/placeholder",["require","exports","module","ace/range","ace/lib/event_emitter","ace/lib/oop"],(function(e,t,n){"use strict";var i=e("./range").Range,r=e("./lib/event_emitter").EventEmitter,o=e("./lib/oop"),s=function(e,t,n,i,r,o){var s=this;this.length=t,this.session=e,this.doc=e.getDocument(),this.mainClass=r,this.othersClass=o,this.$onUpdate=this.onUpdate.bind(this),this.doc.on("change",this.$onUpdate),this.$others=i,this.$onCursorChange=function(){setTimeout((function(){s.onCursorChange()}))},this.$pos=n;var a=e.getUndoManager().$undoStack||e.getUndoManager().$undostack||{length:-1};this.$undoStackDepth=a.length,this.setup(),e.selection.on("changeCursor",this.$onCursorChange)};(function(){o.implement(this,r),this.setup=function(){var e=this,t=this.doc,n=this.session;this.selectionBefore=n.selection.toJSON(),n.selection.inMultiSelectMode&&n.selection.toSingleRange(),this.pos=t.createAnchor(this.$pos.row,this.$pos.column);var r=this.pos;r.$insertRight=!0,r.detach(),r.markerId=n.addMarker(new i(r.row,r.column,r.row,r.column+this.length),this.mainClass,null,!1),this.others=[],this.$others.forEach((function(n){var i=t.createAnchor(n.row,n.column);i.$insertRight=!0,i.detach(),e.others.push(i)})),n.setUndoSelect(!1)},this.showOtherMarkers=function(){if(!this.othersActive){var e=this.session,t=this;this.othersActive=!0,this.others.forEach((function(n){n.markerId=e.addMarker(new i(n.row,n.column,n.row,n.column+t.length),t.othersClass,null,!1)}))}},this.hideOtherMarkers=function(){if(this.othersActive){this.othersActive=!1;for(var e=0;e<this.others.length;e++)this.session.removeMarker(this.others[e].markerId)}},this.onUpdate=function(e){if(this.$updating)return this.updateAnchors(e);var t=e;if(t.start.row===t.end.row&&t.start.row===this.pos.row){this.$updating=!0;var n="insert"===e.action?t.end.column-t.start.column:t.start.column-t.end.column,r=t.start.column>=this.pos.column&&t.start.column<=this.pos.column+this.length+1,o=t.start.column-this.pos.column;if(this.updateAnchors(e),r&&(this.length+=n),r&&!this.session.$fromUndo)if("insert"===e.action)for(var s=this.others.length-1;s>=0;s--){var a={row:(l=this.others[s]).row,column:l.column+o};this.doc.insertMergedLines(a,e.lines)}else if("remove"===e.action)for(s=this.others.length-1;s>=0;s--){var l;a={row:(l=this.others[s]).row,column:l.column+o};this.doc.remove(new i(a.row,a.column,a.row,a.column-n))}this.$updating=!1,this.updateMarkers()}},this.updateAnchors=function(e){this.pos.onChange(e);for(var t=this.others.length;t--;)this.others[t].onChange(e);this.updateMarkers()},this.updateMarkers=function(){if(!this.$updating){var e=this,t=this.session,n=function(n,r){t.removeMarker(n.markerId),n.markerId=t.addMarker(new i(n.row,n.column,n.row,n.column+e.length),r,null,!1)};n(this.pos,this.mainClass);for(var r=this.others.length;r--;)n(this.others[r],this.othersClass)}},this.onCursorChange=function(e){if(!this.$updating&&this.session){var t=this.session.selection.getCursor();t.row===this.pos.row&&t.column>=this.pos.column&&t.column<=this.pos.column+this.length?(this.showOtherMarkers(),this._emit("cursorEnter",e)):(this.hideOtherMarkers(),this._emit("cursorLeave",e))}},this.detach=function(){this.session.removeMarker(this.pos&&this.pos.markerId),this.hideOtherMarkers(),this.doc.removeEventListener("change",this.$onUpdate),this.session.selection.removeEventListener("changeCursor",this.$onCursorChange),this.session.setUndoSelect(!0),this.session=null},this.cancel=function(){if(-1!==this.$undoStackDepth){for(var e=this.session.getUndoManager(),t=(e.$undoStack||e.$undostack).length-this.$undoStackDepth,n=0;n<t;n++)e.undo(!0);this.selectionBefore&&this.session.selection.fromJSON(this.selectionBefore)}}}).call(s.prototype),t.PlaceHolder=s})),ace.define("ace/mouse/multi_select_handler",["require","exports","module","ace/lib/event","ace/lib/useragent"],(function(e,t,n){var i=e("../lib/event"),r=e("../lib/useragent");function o(e,t){return e.row==t.row&&e.column==t.column}t.onMouseDown=function(e){var t=e.domEvent,n=t.altKey,s=t.shiftKey,a=t.ctrlKey,l=e.getAccelKey(),u=e.getButton();if(a&&r.isMac&&(u=t.button),e.editor.inMultiSelectMode&&2==u)e.editor.textInput.onContextMenu(e.domEvent);else if(a||n||l){if(0===u){var c,h=e.editor,f=h.selection,d=h.inMultiSelectMode,p=e.getDocumentPosition(),g=f.getCursor(),m=e.inSelection()||f.isEmpty()&&o(p,g),v=e.x,b=e.y,y=h.session,w=h.renderer.pixelToScreenCoordinates(v,b),A=w;if(h.$mouseHandler.$enableJumpToDef)a&&n||l&&n?c=s?"block":"add":n&&h.$blockSelectEnabled&&(c="block");else if(l&&!n){if(c="add",!d&&s)return}else n&&h.$blockSelectEnabled&&(c="block");if(c&&r.isMac&&t.ctrlKey&&h.$mouseHandler.cancelContextMenu(),"add"==c){if(!d&&m)return;if(!d){var C=f.toOrientedRange();h.addSelectionMarker(C)}var E=f.rangeList.rangeAtPoint(p);h.$blockScrolling++,h.inVirtualSelectionMode=!0,s&&(E=null,C=f.ranges[0]||C,h.removeSelectionMarker(C)),h.once("mouseup",(function(){var e=f.toOrientedRange();E&&e.isEmpty()&&o(E.cursor,e.cursor)?f.substractPoint(e.cursor):(s?f.substractPoint(C.cursor):C&&(h.removeSelectionMarker(C),f.addRange(C)),f.addRange(e)),h.$blockScrolling--,h.inVirtualSelectionMode=!1}))}else if("block"==c){var F;e.stop(),h.inVirtualSelectionMode=!0;var x=[];h.$blockScrolling++,d&&!l?f.toSingleRange():!d&&l&&(F=f.toOrientedRange(),h.addSelectionMarker(F)),s?w=y.documentToScreenPosition(f.lead):f.moveToPosition(p),h.$blockScrolling--,A={row:-1,column:-1};var k=function(){var e=h.renderer.pixelToScreenCoordinates(v,b),t=y.screenToDocumentPosition(e.row,e.column,e.offsetX);o(A,e)&&o(t,f.lead)||(A=e,h.$blockScrolling++,h.selection.moveToPosition(t),h.renderer.scrollCursorIntoView(),h.removeSelectionMarkers(x),x=f.rectangularRangeBlock(A,w),h.$mouseHandler.$clickSelection&&1==x.length&&x[0].isEmpty()&&(x[0]=h.$mouseHandler.$clickSelection.clone()),x.forEach(h.addSelectionMarker,h),h.updateSelectionMarkers(),h.$blockScrolling--)};i.capture(h.container,(function(e){v=e.clientX,b=e.clientY}),(function(e){clearInterval(S),h.removeSelectionMarkers(x),x.length||(x=[f.toOrientedRange()]),h.$blockScrolling++,F&&(h.removeSelectionMarker(F),f.toSingleRange(F));for(var t=0;t<x.length;t++)f.addRange(x[t]);h.inVirtualSelectionMode=!1,h.$mouseHandler.$clickSelection=null,h.$blockScrolling--}));var S=setInterval((function(){k()}),20);return e.preventDefault()}}}else 0===u&&e.editor.inMultiSelectMode&&e.editor.exitMultiSelectMode()}})),ace.define("ace/commands/multi_select_commands",["require","exports","module","ace/keyboard/hash_handler"],(function(e,t,n){t.defaultCommands=[{name:"addCursorAbove",exec:function(e){e.selectMoreLines(-1)},bindKey:{win:"Ctrl-Alt-Up",mac:"Ctrl-Alt-Up"},scrollIntoView:"cursor",readOnly:!0},{name:"addCursorBelow",exec:function(e){e.selectMoreLines(1)},bindKey:{win:"Ctrl-Alt-Down",mac:"Ctrl-Alt-Down"},scrollIntoView:"cursor",readOnly:!0},{name:"addCursorAboveSkipCurrent",exec:function(e){e.selectMoreLines(-1,!0)},bindKey:{win:"Ctrl-Alt-Shift-Up",mac:"Ctrl-Alt-Shift-Up"},scrollIntoView:"cursor",readOnly:!0},{name:"addCursorBelowSkipCurrent",exec:function(e){e.selectMoreLines(1,!0)},bindKey:{win:"Ctrl-Alt-Shift-Down",mac:"Ctrl-Alt-Shift-Down"},scrollIntoView:"cursor",readOnly:!0},{name:"selectMoreBefore",exec:function(e){e.selectMore(-1)},bindKey:{win:"Ctrl-Alt-Left",mac:"Ctrl-Alt-Left"},scrollIntoView:"cursor",readOnly:!0},{name:"selectMoreAfter",exec:function(e){e.selectMore(1)},bindKey:{win:"Ctrl-Alt-Right",mac:"Ctrl-Alt-Right"},scrollIntoView:"cursor",readOnly:!0},{name:"selectNextBefore",exec:function(e){e.selectMore(-1,!0)},bindKey:{win:"Ctrl-Alt-Shift-Left",mac:"Ctrl-Alt-Shift-Left"},scrollIntoView:"cursor",readOnly:!0},{name:"selectNextAfter",exec:function(e){e.selectMore(1,!0)},bindKey:{win:"Ctrl-Alt-Shift-Right",mac:"Ctrl-Alt-Shift-Right"},scrollIntoView:"cursor",readOnly:!0},{name:"splitIntoLines",exec:function(e){e.multiSelect.splitIntoLines()},bindKey:{win:"Ctrl-Alt-L",mac:"Ctrl-Alt-L"},readOnly:!0},{name:"alignCursors",exec:function(e){e.alignCursors()},bindKey:{win:"Ctrl-Alt-A",mac:"Ctrl-Alt-A"},scrollIntoView:"cursor"},{name:"findAll",exec:function(e){e.findAll()},bindKey:{win:"Ctrl-Alt-K",mac:"Ctrl-Alt-G"},scrollIntoView:"cursor",readOnly:!0}],t.multiSelectCommands=[{name:"singleSelection",bindKey:"esc",exec:function(e){e.exitMultiSelectMode()},scrollIntoView:"cursor",readOnly:!0,isAvailable:function(e){return e&&e.inMultiSelectMode}}];var i=e("../keyboard/hash_handler").HashHandler;t.keyboardHandler=new i(t.multiSelectCommands)})),ace.define("ace/multi_select",["require","exports","module","ace/range_list","ace/range","ace/selection","ace/mouse/multi_select_handler","ace/lib/event","ace/lib/lang","ace/commands/multi_select_commands","ace/search","ace/edit_session","ace/editor","ace/config"],(function(e,t,n){var i=e("./range_list").RangeList,r=e("./range").Range,o=e("./selection").Selection,s=e("./mouse/multi_select_handler").onMouseDown,a=e("./lib/event"),l=e("./lib/lang"),u=e("./commands/multi_select_commands");t.commands=u.defaultCommands.concat(u.multiSelectCommands);var c=new(0,e("./search").Search);var h=e("./edit_session").EditSession;(function(){this.getSelectionMarkers=function(){return this.$selectionMarkers}}).call(h.prototype),function(){this.ranges=null,this.rangeList=null,this.addRange=function(e,t){if(e){if(!this.inMultiSelectMode&&0===this.rangeCount){var n=this.toOrientedRange();if(this.rangeList.add(n),this.rangeList.add(e),2!=this.rangeList.ranges.length)return this.rangeList.removeAll(),t||this.fromOrientedRange(e);this.rangeList.removeAll(),this.rangeList.add(n),this.$onAddRange(n)}e.cursor||(e.cursor=e.end);var i=this.rangeList.add(e);return this.$onAddRange(e),i.length&&this.$onRemoveRange(i),this.rangeCount>1&&!this.inMultiSelectMode&&(this._signal("multiSelect"),this.inMultiSelectMode=!0,this.session.$undoSelect=!1,this.rangeList.attach(this.session)),t||this.fromOrientedRange(e)}},this.toSingleRange=function(e){e=e||this.ranges[0];var t=this.rangeList.removeAll();t.length&&this.$onRemoveRange(t),e&&this.fromOrientedRange(e)},this.substractPoint=function(e){var t=this.rangeList.substractPoint(e);if(t)return this.$onRemoveRange(t),t[0]},this.mergeOverlappingRanges=function(){var e=this.rangeList.merge();e.length?this.$onRemoveRange(e):this.ranges[0]&&this.fromOrientedRange(this.ranges[0])},this.$onAddRange=function(e){this.rangeCount=this.rangeList.ranges.length,this.ranges.unshift(e),this._signal("addRange",{range:e})},this.$onRemoveRange=function(e){if(this.rangeCount=this.rangeList.ranges.length,1==this.rangeCount&&this.inMultiSelectMode){var t=this.rangeList.ranges.pop();e.push(t),this.rangeCount=0}for(var n=e.length;n--;){var i=this.ranges.indexOf(e[n]);this.ranges.splice(i,1)}this._signal("removeRange",{ranges:e}),0===this.rangeCount&&this.inMultiSelectMode&&(this.inMultiSelectMode=!1,this._signal("singleSelect"),this.session.$undoSelect=!0,this.rangeList.detach(this.session)),(t=t||this.ranges[0])&&!t.isEqual(this.getRange())&&this.fromOrientedRange(t)},this.$initRangeList=function(){this.rangeList||(this.rangeList=new i,this.ranges=[],this.rangeCount=0)},this.getAllRanges=function(){return this.rangeCount?this.rangeList.ranges.concat():[this.getRange()]},this.splitIntoLines=function(){if(this.rangeCount>1){var e=this.rangeList.ranges,t=e[e.length-1],n=r.fromPoints(e[0].start,t.end);this.toSingleRange(),this.setSelectionRange(n,t.cursor==t.start)}else{n=this.getRange();var i=this.isBackwards(),o=n.start.row,s=n.end.row;if(o==s){if(i)var a=n.end,l=n.start;else a=n.start,l=n.end;return this.addRange(r.fromPoints(l,l)),void this.addRange(r.fromPoints(a,a))}var u=[],c=this.getLineRange(o,!0);c.start.column=n.start.column,u.push(c);for(var h=o+1;h<s;h++)u.push(this.getLineRange(h,!0));(c=this.getLineRange(s,!0)).end.column=n.end.column,u.push(c),u.forEach(this.addRange,this)}},this.toggleBlockSelection=function(){if(this.rangeCount>1){var e=this.rangeList.ranges,t=e[e.length-1],n=r.fromPoints(e[0].start,t.end);this.toSingleRange(),this.setSelectionRange(n,t.cursor==t.start)}else{var i=this.session.documentToScreenPosition(this.selectionLead),o=this.session.documentToScreenPosition(this.selectionAnchor);this.rectangularRangeBlock(i,o).forEach(this.addRange,this)}},this.rectangularRangeBlock=function(e,t,n){var i=[],o=e.column<t.column;if(o)var s=e.column,a=t.column,l=e.offsetX,u=t.offsetX;else s=t.column,a=e.column,l=t.offsetX,u=e.offsetX;var c,h,f=e.row<t.row;if(f)var d=e.row,p=t.row;else d=t.row,p=e.row;s<0&&(s=0),d<0&&(d=0),d==p&&(n=!0);for(var g=d;g<=p;g++){var m=r.fromPoints(this.session.screenToDocumentPosition(g,s,l),this.session.screenToDocumentPosition(g,a,u));if(m.isEmpty()){if(v&&(c=m.end,h=v,c.row==h.row&&c.column==h.column))break;var v=m.end}m.cursor=o?m.start:m.end,i.push(m)}if(f&&i.reverse(),!n){for(var b=i.length-1;i[b].isEmpty()&&b>0;)b--;if(b>0)for(var y=0;i[y].isEmpty();)y++;for(var w=b;w>=y;w--)i[w].isEmpty()&&i.splice(w,1)}return i}}.call(o.prototype);var f=e("./editor").Editor;function d(e){e.$multiselectOnSessionChange||(e.$onAddRange=e.$onAddRange.bind(e),e.$onRemoveRange=e.$onRemoveRange.bind(e),e.$onMultiSelect=e.$onMultiSelect.bind(e),e.$onSingleSelect=e.$onSingleSelect.bind(e),e.$multiselectOnSessionChange=t.onSessionChange.bind(e),e.$checkMultiselectChange=e.$checkMultiselectChange.bind(e),e.$multiselectOnSessionChange(e),e.on("changeSession",e.$multiselectOnSessionChange),e.on("mousedown",s),e.commands.addCommands(u.defaultCommands),function(e){var t=e.textInput.getElement(),n=!1;function i(t){n&&(e.renderer.setMouseCursor(""),n=!1)}a.addListener(t,"keydown",(function(t){var r=18==t.keyCode&&!(t.ctrlKey||t.shiftKey||t.metaKey);e.$blockSelectEnabled&&r?n||(e.renderer.setMouseCursor("crosshair"),n=!0):n&&i()})),a.addListener(t,"keyup",i),a.addListener(t,"blur",i)}(e))}(function(){this.updateSelectionMarkers=function(){this.renderer.updateCursor(),this.renderer.updateBackMarkers()},this.addSelectionMarker=function(e){e.cursor||(e.cursor=e.end);var t=this.getSelectionStyle();return e.marker=this.session.addMarker(e,"ace_selection",t),this.session.$selectionMarkers.push(e),this.session.selectionMarkerCount=this.session.$selectionMarkers.length,e},this.removeSelectionMarker=function(e){if(e.marker){this.session.removeMarker(e.marker);var t=this.session.$selectionMarkers.indexOf(e);-1!=t&&this.session.$selectionMarkers.splice(t,1),this.session.selectionMarkerCount=this.session.$selectionMarkers.length}},this.removeSelectionMarkers=function(e){for(var t=this.session.$selectionMarkers,n=e.length;n--;){var i=e[n];if(i.marker){this.session.removeMarker(i.marker);var r=t.indexOf(i);-1!=r&&t.splice(r,1)}}this.session.selectionMarkerCount=t.length},this.$onAddRange=function(e){this.addSelectionMarker(e.range),this.renderer.updateCursor(),this.renderer.updateBackMarkers()},this.$onRemoveRange=function(e){this.removeSelectionMarkers(e.ranges),this.renderer.updateCursor(),this.renderer.updateBackMarkers()},this.$onMultiSelect=function(e){this.inMultiSelectMode||(this.inMultiSelectMode=!0,this.setStyle("ace_multiselect"),this.keyBinding.addKeyboardHandler(u.keyboardHandler),this.commands.setDefaultHandler("exec",this.$onMultiSelectExec),this.renderer.updateCursor(),this.renderer.updateBackMarkers())},this.$onSingleSelect=function(e){this.session.multiSelect.inVirtualMode||(this.inMultiSelectMode=!1,this.unsetStyle("ace_multiselect"),this.keyBinding.removeKeyboardHandler(u.keyboardHandler),this.commands.removeDefaultHandler("exec",this.$onMultiSelectExec),this.renderer.updateCursor(),this.renderer.updateBackMarkers(),this._emit("changeSelection"))},this.$onMultiSelectExec=function(e){var t=e.command,n=e.editor;if(n.multiSelect){if(t.multiSelectAction)"forEach"==t.multiSelectAction?i=n.forEachSelection(t,e.args):"forEachLine"==t.multiSelectAction?i=n.forEachSelection(t,e.args,!0):"single"==t.multiSelectAction?(n.exitMultiSelectMode(),i=t.exec(n,e.args||{})):i=t.multiSelectAction(n,e.args||{});else{var i=t.exec(n,e.args||{});n.multiSelect.addRange(n.multiSelect.toOrientedRange()),n.multiSelect.mergeOverlappingRanges()}return i}},this.forEachSelection=function(e,t,n){if(!this.inVirtualSelectionMode){var i,r=n&&n.keepOrder,s=1==n||n&&n.$byLines,a=this.session,l=this.selection,u=l.rangeList,c=(r?l:u).ranges;if(!c.length)return e.exec?e.exec(this,t||{}):e(this,t||{});var h=l._eventRegistry;l._eventRegistry={};var f=new o(a);this.inVirtualSelectionMode=!0;for(var d=c.length;d--;){if(s)for(;d>0&&c[d].start.row==c[d-1].end.row;)d--;f.fromOrientedRange(c[d]),f.index=d,this.selection=a.selection=f;var p=e.exec?e.exec(this,t||{}):e(this,t||{});i||void 0===p||(i=p),f.toOrientedRange(c[d])}f.detach(),this.selection=a.selection=l,this.inVirtualSelectionMode=!1,l._eventRegistry=h,l.mergeOverlappingRanges();var g=this.renderer.$scrollAnimation;return this.onCursorChange(),this.onSelectionChange(),g&&g.from==g.to&&this.renderer.animateScrolling(g.from),i}},this.exitMultiSelectMode=function(){this.inMultiSelectMode&&!this.inVirtualSelectionMode&&this.multiSelect.toSingleRange()},this.getSelectedText=function(){var e="";if(this.inMultiSelectMode&&!this.inVirtualSelectionMode){for(var t=this.multiSelect.rangeList.ranges,n=[],i=0;i<t.length;i++)n.push(this.session.getTextRange(t[i]));var r=this.session.getDocument().getNewLineCharacter();(e=n.join(r)).length==(n.length-1)*r.length&&(e="")}else this.selection.isEmpty()||(e=this.session.getTextRange(this.getSelectionRange()));return e},this.$checkMultiselectChange=function(e,t){if(this.inMultiSelectMode&&!this.inVirtualSelectionMode){var n=this.multiSelect.ranges[0];if(this.multiSelect.isEmpty()&&t==this.multiSelect.anchor)return;var i=t==this.multiSelect.anchor?n.cursor==n.start?n.end:n.start:n.cursor;i.row==t.row&&this.session.$clipPositionToDocument(i.row,i.column).column==t.column||this.multiSelect.toSingleRange(this.multiSelect.toOrientedRange())}},this.findAll=function(e,t,n){if((t=t||{}).needle=e||t.needle,null==t.needle){var i=this.selection.isEmpty()?this.selection.getWordRange():this.selection.getRange();t.needle=this.session.getTextRange(i)}this.$search.set(t);var r=this.$search.findAll(this.session);if(!r.length)return 0;this.$blockScrolling+=1;var o=this.multiSelect;n||o.toSingleRange(r[0]);for(var s=r.length;s--;)o.addRange(r[s],!0);return i&&o.rangeList.rangeAtPoint(i.start)&&o.addRange(i,!0),this.$blockScrolling-=1,r.length},this.selectMoreLines=function(e,t){var n=this.selection.toOrientedRange(),i=n.cursor==n.end,o=this.session.documentToScreenPosition(n.cursor);this.selection.$desiredColumn&&(o.column=this.selection.$desiredColumn);var s,a=this.session.screenToDocumentPosition(o.row+e,o.column);if(n.isEmpty())u=a;else var l=this.session.documentToScreenPosition(i?n.end:n.start),u=this.session.screenToDocumentPosition(l.row+e,l.column);i?(s=r.fromPoints(a,u)).cursor=s.start:(s=r.fromPoints(u,a)).cursor=s.end;if(s.desiredColumn=o.column,this.selection.inMultiSelectMode){if(t)var c=n.cursor}else this.selection.addRange(n);this.selection.addRange(s),c&&this.selection.substractPoint(c)},this.transposeSelections=function(e){for(var t=this.session,n=t.multiSelect,i=n.ranges,r=i.length;r--;){if((a=i[r]).isEmpty()){var o=t.getWordRange(a.start.row,a.start.column);a.start.row=o.start.row,a.start.column=o.start.column,a.end.row=o.end.row,a.end.column=o.end.column}}n.mergeOverlappingRanges();var s=[];for(r=i.length;r--;){var a=i[r];s.unshift(t.getTextRange(a))}e<0?s.unshift(s.pop()):s.push(s.shift());for(r=i.length;r--;){o=(a=i[r]).clone();t.replace(a,s[r]),a.start.row=o.start.row,a.start.column=o.start.column}},this.selectMore=function(e,t,n){var i=this.session,r=i.multiSelect.toOrientedRange();if(!r.isEmpty()||((r=i.getWordRange(r.start.row,r.start.column)).cursor=-1==e?r.start:r.end,this.multiSelect.addRange(r),!n)){var o=i.getTextRange(r),s=function(e,t,n){return c.$options.wrap=!0,c.$options.needle=t,c.$options.backwards=-1==n,c.find(e)}(i,o,e);s&&(s.cursor=-1==e?s.start:s.end,this.$blockScrolling+=1,this.session.unfold(s),this.multiSelect.addRange(s),this.$blockScrolling-=1,this.renderer.scrollCursorIntoView(null,.5)),t&&this.multiSelect.substractPoint(r.cursor)}},this.alignCursors=function(){var e=this.session,t=e.multiSelect,n=t.ranges,i=-1,o=n.filter((function(e){if(e.cursor.row==i)return!0;i=e.cursor.row}));if(n.length&&o.length!=n.length-1){o.forEach((function(e){t.substractPoint(e.cursor)}));var s=0,a=1/0,u=n.map((function(t){var n=t.cursor,i=e.getLine(n.row).substr(n.column).search(/\S/g);return-1==i&&(i=0),n.column>s&&(s=n.column),i<a&&(a=i),i}));n.forEach((function(t,n){var i=t.cursor,o=s-i.column,c=u[n]-a;o>c?e.insert(i,l.stringRepeat(" ",o-c)):e.remove(new r(i.row,i.column,i.row,i.column-o+c)),t.start.column=t.end.column=s,t.start.row=t.end.row=i.row,t.cursor=t.end})),t.fromOrientedRange(n[0]),this.renderer.updateCursor(),this.renderer.updateBackMarkers()}else{var c=this.selection.getRange(),h=c.start.row,f=c.end.row,d=h==f;if(d){var p,g=this.session.getLength();do{p=this.session.getLine(f)}while(/[=:]/.test(p)&&++f<g);do{p=this.session.getLine(h)}while(/[=:]/.test(p)&&--h>0);h<0&&(h=0),f>=g&&(f=g-1)}var m=this.session.removeFullLines(h,f);m=this.$reAlignText(m,d),this.session.insert({row:h,column:0},m.join("\n")+"\n"),d||(c.start.column=0,c.end.column=m[m.length-1].length),this.selection.setRange(c)}},this.$reAlignText=function(e,t){var n,i,r,o=!0,s=!0;return e.map((function(e){var t=e.match(/(\s*)(.*?)(\s*)([=:].*)/);return t?null==n?(n=t[1].length,i=t[2].length,r=t[3].length,t):(n+i+r!=t[1].length+t[2].length+t[3].length&&(s=!1),n!=t[1].length&&(o=!1),n>t[1].length&&(n=t[1].length),i<t[2].length&&(i=t[2].length),r>t[3].length&&(r=t[3].length),t):[e]})).map(t?u:o?s?function(e){return e[2]?a(n+i-e[2].length)+e[2]+a(r)+e[4].replace(/^([=:])\s+/,"$1 "):e[0]}:u:function(e){return e[2]?a(n)+e[2]+a(r)+e[4].replace(/^([=:])\s+/,"$1 "):e[0]});function a(e){return l.stringRepeat(" ",e)}function u(e){return e[2]?a(n)+e[2]+a(i-e[2].length+r)+e[4].replace(/^([=:])\s+/,"$1 "):e[0]}}}).call(f.prototype),t.onSessionChange=function(e){var t=e.session;t&&!t.multiSelect&&(t.$selectionMarkers=[],t.selection.$initRangeList(),t.multiSelect=t.selection),this.multiSelect=t&&t.multiSelect;var n=e.oldSession;n&&(n.multiSelect.off("addRange",this.$onAddRange),n.multiSelect.off("removeRange",this.$onRemoveRange),n.multiSelect.off("multiSelect",this.$onMultiSelect),n.multiSelect.off("singleSelect",this.$onSingleSelect),n.multiSelect.lead.off("change",this.$checkMultiselectChange),n.multiSelect.anchor.off("change",this.$checkMultiselectChange)),t&&(t.multiSelect.on("addRange",this.$onAddRange),t.multiSelect.on("removeRange",this.$onRemoveRange),t.multiSelect.on("multiSelect",this.$onMultiSelect),t.multiSelect.on("singleSelect",this.$onSingleSelect),t.multiSelect.lead.on("change",this.$checkMultiselectChange),t.multiSelect.anchor.on("change",this.$checkMultiselectChange)),t&&this.inMultiSelectMode!=t.selection.inMultiSelectMode&&(t.selection.inMultiSelectMode?this.$onMultiSelect():this.$onSingleSelect())},t.MultiSelect=d,e("./config").defineOptions(f.prototype,"editor",{enableMultiselect:{set:function(e){d(this),e?(this.on("changeSession",this.$multiselectOnSessionChange),this.on("mousedown",s)):(this.off("changeSession",this.$multiselectOnSessionChange),this.off("mousedown",s))},value:!0},enableBlockSelect:{set:function(e){this.$blockSelectEnabled=e},value:!0}})})),ace.define("ace/mode/folding/fold_mode",["require","exports","module","ace/range"],(function(e,t,n){"use strict";var i=e("../../range").Range,r=t.FoldMode=function(){};(function(){this.foldingStartMarker=null,this.foldingStopMarker=null,this.getFoldWidget=function(e,t,n){var i=e.getLine(n);return this.foldingStartMarker.test(i)?"start":"markbeginend"==t&&this.foldingStopMarker&&this.foldingStopMarker.test(i)?"end":""},this.getFoldWidgetRange=function(e,t,n){return null},this.indentationBlock=function(e,t,n){var r=/\S/,o=e.getLine(t),s=o.search(r);if(-1!=s){for(var a=n||o.length,l=e.getLength(),u=t,c=t;++t<l;){var h=e.getLine(t).search(r);if(-1!=h){if(h<=s)break;c=t}}if(c>u){var f=e.getLine(c).length;return new i(u,a,c,f)}}},this.openingBracketBlock=function(e,t,n,r,o){var s={row:n,column:r+1},a=e.$findClosingBracket(t,s,o);if(a){var l=e.foldWidgets[a.row];return null==l&&(l=e.getFoldWidget(a.row)),"start"==l&&a.row>s.row&&(a.row--,a.column=e.getLine(a.row).length),i.fromPoints(s,a)}},this.closingBracketBlock=function(e,t,n,r,o){var s={row:n,column:r},a=e.$findOpeningBracket(t,s);if(a)return a.column++,s.column--,i.fromPoints(a,s)}}).call(r.prototype)})),ace.define("ace/theme/textmate",["require","exports","module","ace/lib/dom"],(function(e,t,n){"use strict";t.isDark=!1,t.cssClass="ace-tm",t.cssText='.ace-tm .ace_gutter {background: #f0f0f0;color: #333;}.ace-tm .ace_print-margin {width: 1px;background: #e8e8e8;}.ace-tm .ace_fold {background-color: #6B72E6;}.ace-tm {background-color: #FFFFFF;color: black;}.ace-tm .ace_cursor {color: black;}.ace-tm .ace_invisible {color: rgb(191, 191, 191);}.ace-tm .ace_storage,.ace-tm .ace_keyword {color: blue;}.ace-tm .ace_constant {color: rgb(197, 6, 11);}.ace-tm .ace_constant.ace_buildin {color: rgb(88, 72, 246);}.ace-tm .ace_constant.ace_language {color: rgb(88, 92, 246);}.ace-tm .ace_constant.ace_library {color: rgb(6, 150, 14);}.ace-tm .ace_invalid {background-color: rgba(255, 0, 0, 0.1);color: red;}.ace-tm .ace_support.ace_function {color: rgb(60, 76, 114);}.ace-tm .ace_support.ace_constant {color: rgb(6, 150, 14);}.ace-tm .ace_support.ace_type,.ace-tm .ace_support.ace_class {color: rgb(109, 121, 222);}.ace-tm .ace_keyword.ace_operator {color: rgb(104, 118, 135);}.ace-tm .ace_string {color: rgb(3, 106, 7);}.ace-tm .ace_comment {color: rgb(76, 136, 107);}.ace-tm .ace_comment.ace_doc {color: rgb(0, 102, 255);}.ace-tm .ace_comment.ace_doc.ace_tag {color: rgb(128, 159, 191);}.ace-tm .ace_constant.ace_numeric {color: rgb(0, 0, 205);}.ace-tm .ace_variable {color: rgb(49, 132, 149);}.ace-tm .ace_xml-pe {color: rgb(104, 104, 91);}.ace-tm .ace_entity.ace_name.ace_function {color: #0000A2;}.ace-tm .ace_heading {color: rgb(12, 7, 255);}.ace-tm .ace_list {color:rgb(185, 6, 144);}.ace-tm .ace_meta.ace_tag {color:rgb(0, 22, 142);}.ace-tm .ace_string.ace_regex {color: rgb(255, 0, 0)}.ace-tm .ace_marker-layer .ace_selection {background: rgb(181, 213, 255);}.ace-tm.ace_multiselect .ace_selection.ace_start {box-shadow: 0 0 3px 0px white;}.ace-tm .ace_marker-layer .ace_step {background: rgb(252, 255, 0);}.ace-tm .ace_marker-layer .ace_stack {background: rgb(164, 229, 101);}.ace-tm .ace_marker-layer .ace_bracket {margin: -1px 0 0 -1px;border: 1px solid rgb(192, 192, 192);}.ace-tm .ace_marker-layer .ace_active-line {background: rgba(0, 0, 0, 0.07);}.ace-tm .ace_gutter-active-line {background-color : #dcdcdc;}.ace-tm .ace_marker-layer .ace_selected-word {background: rgb(250, 250, 255);border: 1px solid rgb(200, 200, 250);}.ace-tm .ace_indent-guide {background: url("data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAAEAAAACCAYAAACZgbYnAAAAE0lEQVQImWP4////f4bLly//BwAmVgd1/w11/gAAAABJRU5ErkJggg==") right repeat-y;}',e("../lib/dom").importCssString(t.cssText,t.cssClass)})),ace.define("ace/line_widgets",["require","exports","module","ace/lib/oop","ace/lib/dom","ace/range"],(function(e,t,n){"use strict";e("./lib/oop");var i=e("./lib/dom");e("./range").Range;function r(e){this.session=e,this.session.widgetManager=this,this.session.getRowLength=this.getRowLength,this.session.$getWidgetScreenLength=this.$getWidgetScreenLength,this.updateOnChange=this.updateOnChange.bind(this),this.renderWidgets=this.renderWidgets.bind(this),this.measureWidgets=this.measureWidgets.bind(this),this.session._changedWidgets=[],this.$onChangeEditor=this.$onChangeEditor.bind(this),this.session.on("change",this.updateOnChange),this.session.on("changeFold",this.updateOnFold),this.session.on("changeEditor",this.$onChangeEditor)}(function(){this.getRowLength=function(e){var t;return t=this.lineWidgets&&this.lineWidgets[e]&&this.lineWidgets[e].rowCount||0,this.$useWrapMode&&this.$wrapData[e]?this.$wrapData[e].length+1+t:1+t},this.$getWidgetScreenLength=function(){var e=0;return this.lineWidgets.forEach((function(t){t&&t.rowCount&&!t.hidden&&(e+=t.rowCount)})),e},this.$onChangeEditor=function(e){this.attach(e.editor)},this.attach=function(e){e&&e.widgetManager&&e.widgetManager!=this&&e.widgetManager.detach(),this.editor!=e&&(this.detach(),this.editor=e,e&&(e.widgetManager=this,e.renderer.on("beforeRender",this.measureWidgets),e.renderer.on("afterRender",this.renderWidgets)))},this.detach=function(e){var t=this.editor;if(t){this.editor=null,t.widgetManager=null,t.renderer.off("beforeRender",this.measureWidgets),t.renderer.off("afterRender",this.renderWidgets);var n=this.session.lineWidgets;n&&n.forEach((function(e){e&&e.el&&e.el.parentNode&&(e._inDocument=!1,e.el.parentNode.removeChild(e.el))}))}},this.updateOnFold=function(e,t){var n=t.lineWidgets;if(n&&e.action){for(var i=e.data,r=i.start.row,o=i.end.row,s="add"==e.action,a=r+1;a<o;a++)n[a]&&(n[a].hidden=s);n[o]&&(s?n[r]?n[o].hidden=s:n[r]=n[o]:(n[r]==n[o]&&(n[r]=void 0),n[o].hidden=s))}},this.updateOnChange=function(e){var t=this.session.lineWidgets;if(t){var n=e.start.row,i=e.end.row-n;if(0===i);else if("remove"==e.action){t.splice(n+1,i).forEach((function(e){e&&this.removeLineWidget(e)}),this),this.$updateRows()}else{var r=new Array(i);r.unshift(n,0),t.splice.apply(t,r),this.$updateRows()}}},this.$updateRows=function(){var e=this.session.lineWidgets;if(e){var t=!0;e.forEach((function(e,n){if(e)for(t=!1,e.row=n;e.$oldWidget;)e.$oldWidget.row=n,e=e.$oldWidget})),t&&(this.session.lineWidgets=null)}},this.addLineWidget=function(e){this.session.lineWidgets||(this.session.lineWidgets=new Array(this.session.getLength()));var t=this.session.lineWidgets[e.row];t&&(e.$oldWidget=t,t.el&&t.el.parentNode&&(t.el.parentNode.removeChild(t.el),t._inDocument=!1)),this.session.lineWidgets[e.row]=e,e.session=this.session;var n=this.editor.renderer;e.html&&!e.el&&(e.el=i.createElement("div"),e.el.innerHTML=e.html),e.el&&(i.addCssClass(e.el,"ace_lineWidgetContainer"),e.el.style.position="absolute",e.el.style.zIndex=5,n.container.appendChild(e.el),e._inDocument=!0),e.coverGutter||(e.el.style.zIndex=3),null==e.pixelHeight&&(e.pixelHeight=e.el.offsetHeight),null==e.rowCount&&(e.rowCount=e.pixelHeight/n.layerConfig.lineHeight);var r=this.session.getFoldAt(e.row,0);if(e.$fold=r,r){var o=this.session.lineWidgets;e.row!=r.end.row||o[r.start.row]?e.hidden=!0:o[r.start.row]=e}return this.session._emit("changeFold",{data:{start:{row:e.row}}}),this.$updateRows(),this.renderWidgets(null,n),this.onWidgetChanged(e),e},this.removeLineWidget=function(e){if(e._inDocument=!1,e.session=null,e.el&&e.el.parentNode&&e.el.parentNode.removeChild(e.el),e.editor&&e.editor.destroy)try{e.editor.destroy()}catch(e){}if(this.session.lineWidgets){var t=this.session.lineWidgets[e.row];if(t==e)this.session.lineWidgets[e.row]=e.$oldWidget,e.$oldWidget&&this.onWidgetChanged(e.$oldWidget);else for(;t;){if(t.$oldWidget==e){t.$oldWidget=e.$oldWidget;break}t=t.$oldWidget}}this.session._emit("changeFold",{data:{start:{row:e.row}}}),this.$updateRows()},this.getWidgetsAtRow=function(e){for(var t=this.session.lineWidgets,n=t&&t[e],i=[];n;)i.push(n),n=n.$oldWidget;return i},this.onWidgetChanged=function(e){this.session._changedWidgets.push(e),this.editor&&this.editor.renderer.updateFull()},this.measureWidgets=function(e,t){var n=this.session._changedWidgets,i=t.layerConfig;if(n&&n.length){for(var r=1/0,o=0;o<n.length;o++){var s=n[o];if(s&&s.el&&s.session==this.session){if(!s._inDocument){if(this.session.lineWidgets[s.row]!=s)continue;s._inDocument=!0,t.container.appendChild(s.el)}s.h=s.el.offsetHeight,s.fixedWidth||(s.w=s.el.offsetWidth,s.screenWidth=Math.ceil(s.w/i.characterWidth));var a=s.h/i.lineHeight;s.coverLine&&(a-=this.session.getRowLineCount(s.row))<0&&(a=0),s.rowCount!=a&&(s.rowCount=a,s.row<r&&(r=s.row))}}r!=1/0&&(this.session._emit("changeFold",{data:{start:{row:r}}}),this.session.lineWidgetWidth=null),this.session._changedWidgets=[]}},this.renderWidgets=function(e,t){var n=t.layerConfig,i=this.session.lineWidgets;if(i){for(var r=Math.min(this.firstRow,n.firstRow),o=Math.max(this.lastRow,n.lastRow,i.length);r>0&&!i[r];)r--;this.firstRow=n.firstRow,this.lastRow=n.lastRow,t.$cursorLayer.config=n;for(var s=r;s<=o;s++){var a=i[s];if(a&&a.el)if(a.hidden)a.el.style.top=-100-(a.pixelHeight||0)+"px";else{a._inDocument||(a._inDocument=!0,t.container.appendChild(a.el));var l=t.$cursorLayer.getPixelPosition({row:s,column:0},!0).top;a.coverLine||(l+=n.lineHeight*this.session.getRowLineCount(a.row)),a.el.style.top=l-n.offset+"px";var u=a.coverGutter?0:t.gutterWidth;a.fixedWidth||(u-=t.scrollLeft),a.el.style.left=u+"px",a.fullWidth&&a.screenWidth&&(a.el.style.minWidth=n.width+2*n.padding+"px"),a.fixedWidth?a.el.style.right=t.scrollBar.getWidth()+"px":a.el.style.right=""}}}}}).call(r.prototype),t.LineWidgets=r})),ace.define("ace/ext/error_marker",["require","exports","module","ace/line_widgets","ace/lib/dom","ace/range"],(function(e,t,n){"use strict";var i=e("../line_widgets").LineWidgets,r=e("../lib/dom"),o=e("../range").Range;t.showErrorMarker=function(e,t){var n=e.session;n.widgetManager||(n.widgetManager=new i(n),n.widgetManager.attach(e));var s=e.getCursorPosition(),a=s.row,l=n.widgetManager.getWidgetsAtRow(a).filter((function(e){return"errorMarker"==e.type}))[0];l?l.destroy():a-=t;var u,c=function(e,t,n){var i=e.getAnnotations().sort(o.comparePoints);if(i.length){var r=function(e,t,n){for(var i=0,r=e.length-1;i<=r;){var o=i+r>>1,s=n(t,e[o]);if(s>0)i=o+1;else{if(!(s<0))return o;r=o-1}}return-(i+1)}(i,{row:t,column:-1},o.comparePoints);r<0&&(r=-r-1),r>=i.length?r=n>0?0:i.length-1:0===r&&n<0&&(r=i.length-1);var s=i[r];if(s&&n){if(s.row===t){do{s=i[r+=n]}while(s&&s.row===t);if(!s)return i.slice()}var a=[];t=s.row;do{a[n<0?"unshift":"push"](s),s=i[r+=n]}while(s&&s.row==t);return a.length&&a}}}(n,a,t);if(c){var h=c[0];s.column=(h.pos&&"number"!=typeof h.column?h.pos.sc:h.column)||0,s.row=h.row,u=e.renderer.$gutterLayer.$annotations[s.row]}else{if(l)return;u={text:["Looks good!"],className:"ace_ok"}}e.session.unfold(s.row),e.selection.moveToPosition(s);var f={row:s.row,fixedWidth:!0,coverGutter:!0,el:r.createElement("div"),type:"errorMarker"},d=f.el.appendChild(r.createElement("div")),p=f.el.appendChild(r.createElement("div"));p.className="error_widget_arrow "+u.className;var g=e.renderer.$cursorLayer.getPixelPosition(s).left;p.style.left=g+e.renderer.gutterWidth-5+"px",f.el.className="error_widget_wrapper",d.className="error_widget "+u.className,d.innerHTML=u.text.join("<br>"),d.appendChild(r.createElement("div"));var m=function(e,t,n){if(0===t&&("esc"===n||"return"===n))return f.destroy(),{command:"null"}};f.destroy=function(){e.$mouseHandler.isMousePressed||(e.keyBinding.removeKeyboardHandler(m),n.widgetManager.removeLineWidget(f),e.off("changeSelection",f.destroy),e.off("changeSession",f.destroy),e.off("mouseup",f.destroy),e.off("change",f.destroy))},e.keyBinding.addKeyboardHandler(m),e.on("changeSelection",f.destroy),e.on("changeSession",f.destroy),e.on("mouseup",f.destroy),e.on("change",f.destroy),e.session.widgetManager.addLineWidget(f),f.el.onmousedown=e.focus.bind(e),e.renderer.scrollCursorIntoView(null,.5,{bottom:f.el.offsetHeight})},r.importCssString("    .error_widget_wrapper {        background: inherit;        color: inherit;        border:none    }    .error_widget {        border-top: solid 2px;        border-bottom: solid 2px;        margin: 5px 0;        padding: 10px 40px;        white-space: pre-wrap;    }    .error_widget.ace_error, .error_widget_arrow.ace_error{        border-color: #ff5a5a    }    .error_widget.ace_warning, .error_widget_arrow.ace_warning{        border-color: #F1D817    }    .error_widget.ace_info, .error_widget_arrow.ace_info{        border-color: #5a5a5a    }    .error_widget.ace_ok, .error_widget_arrow.ace_ok{        border-color: #5aaa5a    }    .error_widget_arrow {        position: absolute;        border: solid 5px;        border-top-color: transparent!important;        border-right-color: transparent!important;        border-left-color: transparent!important;        top: -5px;    }","")})),ace.define("ace/ace",["require","exports","module","ace/lib/fixoldbrowsers","ace/lib/dom","ace/lib/event","ace/editor","ace/edit_session","ace/undomanager","ace/virtual_renderer","ace/worker/worker_client","ace/keyboard/hash_handler","ace/placeholder","ace/multi_select","ace/mode/folding/fold_mode","ace/theme/textmate","ace/ext/error_marker","ace/config"],(function(e,t,i){"use strict";e("./lib/fixoldbrowsers");var r=e("./lib/dom"),o=e("./lib/event"),s=e("./editor").Editor,a=e("./edit_session").EditSession,l=e("./undomanager").UndoManager,u=e("./virtual_renderer").VirtualRenderer;e("./worker/worker_client"),e("./keyboard/hash_handler"),e("./placeholder"),e("./multi_select"),e("./mode/folding/fold_mode"),e("./theme/textmate"),e("./ext/error_marker"),t.config=e("./config"),t.acequire=e,t.define=n(38),t.edit=function(e){if("string"==typeof e){var n=e;if(!(e=document.getElementById(n)))throw new Error("ace.edit can't find div #"+n)}if(e&&e.env&&e.env.editor instanceof s)return e.env.editor;var i="";if(e&&/input|textarea/i.test(e.tagName)){var a=e;i=a.value,e=r.createElement("pre"),a.parentNode.replaceChild(e,a)}else e&&(i=r.getInnerText(e),e.innerHTML="");var l=t.createEditSession(i),c=new s(new u(e));c.setSession(l);var h={document:l,editor:c,onResize:c.resize.bind(c,null)};return a&&(h.textarea=a),o.addListener(window,"resize",h.onResize),c.on("destroy",(function(){o.removeListener(window,"resize",h.onResize),h.editor.container.env=null})),c.container.env=c.env=h,c},t.createEditSession=function(e,t){var n=new a(e,t);return n.setUndoManager(new l),n},t.EditSession=a,t.UndoManager=l,t.version="1.2.9"})),ace.acequire(["ace/ace"],(function(e){for(var t in e&&(e.config.init(!0),e.define=ace.define),window.ace||(window.ace=e),e)e.hasOwnProperty(t)&&(window.ace[t]=e[t])})),e.exports=window.ace.acequire("ace/ace")},function(e,t,n){var i=n(40),r=n(87),o=n(88),s=i?i.toStringTag:void 0;e.exports=function(e){return null==e?void 0===e?"[object Undefined]":"[object Null]":s&&s in Object(e)?r(e):o(e)}},function(e,t){e.exports=function(e,t,n){return t in e?Object.defineProperty(e,t,{value:n,enumerable:!0,configurable:!0,writable:!0}):e[t]=n,e}},function(e,t,n){e.exports=n(148)()},function(e,t,n){var i=n(62),r=n(31);e.exports=function(e,t){return!t||"object"!==i(t)&&"function"!=typeof t?r(e):t}},function(e,t){function n(t){return e.exports=n=Object.setPrototypeOf?Object.getPrototypeOf:function(e){return e.__proto__||Object.getPrototypeOf(e)},n(t)}e.exports=n},function(e,t,n){var i=n(63);e.exports=function(e,t){if("function"!=typeof t&&null!==t)throw new TypeError("Super expression must either be null or a function");e.prototype=Object.create(t&&t.prototype,{constructor:{value:e,writable:!0,configurable:!0}}),t&&i(e,t)}},function(e,t,n){var i=n(76),r=n(77),o=n(78),s=n(79),a=n(80);function l(e){var t=-1,n=null==e?0:e.length;for(this.clear();++t<n;){var i=e[t];this.set(i[0],i[1])}}l.prototype.clear=i,l.prototype.delete=r,l.prototype.get=o,l.prototype.has=s,l.prototype.set=a,e.exports=l},function(e,t,n){var i=n(22);e.exports=function(e,t){for(var n=e.length;n--;)if(i(e[n][0],t))return n;return-1}},function(e,t){e.exports=function(e,t){return e===t||e!=e&&t!=t}},function(e,t,n){var i=n(27)(Object,"create");e.exports=i},function(e,t,n){var i=n(102);e.exports=function(e,t){var n=e.__data__;return i(t)?n["string"==typeof t?"string":"hash"]:n.map}},function(e,t){e.exports=function(e){return e.webpackPolyfill||(e.deprecate=function(){},e.paths=[],e.children||(e.children=[]),Object.defineProperty(e,"loaded",{enumerable:!0,get:function(){return e.l}}),Object.defineProperty(e,"id",{enumerable:!0,get:function(){return e.i}}),e.webpackPolyfill=1),e}},function(e,t){var n;n=function(){return this}();try{n=n||new Function("return this")()}catch(e){"object"==typeof window&&(n=window)}e.exports=n},function(e,t,n){var i=n(86),r=n(92);e.exports=function(e,t){var n=r(e,t);return i(n)?n:void 0}},function(e,t,n){var i=n(14),r=n(7);e.exports=function(e){if(!r(e))return!1;var t=i(e);return"[object Function]"==t||"[object GeneratorFunction]"==t||"[object AsyncFunction]"==t||"[object Proxy]"==t}},function(e,t,n){var i=n(43);e.exports=function(e,t,n){"__proto__"==t&&i?i(e,t,{configurable:!0,enumerable:!0,value:n,writable:!0}):e[t]=n}},function(e,t,n){var i=n(28),r=n(48);e.exports=function(e){return null!=e&&r(e.length)&&!i(e)}},function(e,t){e.exports=function(e){if(void 0===e)throw new ReferenceError("this hasn't been initialised - super() hasn't been called");return e}},function(e,t,n){var i=n(7),r=n(140),o=n(141),s=Math.max,a=Math.min;e.exports=function(e,t,n){var l,u,c,h,f,d,p=0,g=!1,m=!1,v=!0;if("function"!=typeof e)throw new TypeError("Expected a function");function b(t){var n=l,i=u;return l=u=void 0,p=t,h=e.apply(i,n)}function y(e){return p=e,f=setTimeout(A,t),g?b(e):h}function w(e){var n=e-d;return void 0===d||n>=t||n<0||m&&e-p>=c}function A(){var e=r();if(w(e))return C(e);f=setTimeout(A,function(e){var n=t-(e-d);return m?a(n,c-(e-p)):n}(e))}function C(e){return f=void 0,v&&l?b(e):(l=u=void 0,h)}function E(){var e=r(),n=w(e);if(l=arguments,u=this,d=e,n){if(void 0===f)return y(d);if(m)return clearTimeout(f),f=setTimeout(A,t),b(d)}return void 0===f&&(f=setTimeout(A,t)),h}return t=o(t)||0,i(n)&&(g=!!n.leading,c=(m="maxWait"in n)?s(o(n.maxWait)||0,t):c,v="trailing"in n?!!n.trailing:v),E.cancel=function(){void 0!==f&&clearTimeout(f),p=0,l=d=u=f=void 0},E.flush=function(){return void 0===f?h:C(r())},E}},function(e,t,n){"use strict";Object.defineProperty(t,"__esModule",{value:!0});var i=Object.assign||function(e){for(var t=1;t<arguments.length;t++){var n=arguments[t];for(var i in n)Object.prototype.hasOwnProperty.call(n,i)&&(e[i]=n[i])}return e},r=function(){function e(e,t){for(var n=0;n<t.length;n++){var i=t[n];i.enumerable=i.enumerable||!1,i.configurable=!0,"value"in i&&(i.writable=!0),Object.defineProperty(e,i.key,i)}}return function(t,n,i){return n&&e(t.prototype,n),i&&e(t,i),t}}(),o=n(0),s=l(o),a=l(n(16));function l(e){return e&&e.__esModule?e:{default:e}}var u={position:"absolute",top:0,left:0,visibility:"hidden",height:0,overflow:"scroll",whiteSpace:"pre"},c=["extraWidth","injectStyles","inputClassName","inputRef","inputStyle","minWidth","onAutosize","placeholderIsMinWidth"],h=function(e,t){t.style.fontSize=e.fontSize,t.style.fontFamily=e.fontFamily,t.style.fontWeight=e.fontWeight,t.style.fontStyle=e.fontStyle,t.style.letterSpacing=e.letterSpacing,t.style.textTransform=e.textTransform},f=!("undefined"==typeof window||!window.navigator)&&/MSIE |Trident\/|Edge\//.test(window.navigator.userAgent),d=function(){return f?"_"+Math.random().toString(36).substr(2,12):void 0},p=function(e){function t(e){!function(e,t){if(!(e instanceof t))throw new TypeError("Cannot call a class as a function")}(this,t);var n=function(e,t){if(!e)throw new ReferenceError("this hasn't been initialised - super() hasn't been called");return!t||"object"!=typeof t&&"function"!=typeof t?e:t}(this,(t.__proto__||Object.getPrototypeOf(t)).call(this,e));return n.inputRef=function(e){n.input=e,"function"==typeof n.props.inputRef&&n.props.inputRef(e)},n.placeHolderSizerRef=function(e){n.placeHolderSizer=e},n.sizerRef=function(e){n.sizer=e},n.state={inputWidth:e.minWidth,inputId:e.id||d()},n}return function(e,t){if("function"!=typeof t&&null!==t)throw new TypeError("Super expression must either be null or a function, not "+typeof t);e.prototype=Object.create(t&&t.prototype,{constructor:{value:e,enumerable:!1,writable:!0,configurable:!0}}),t&&(Object.setPrototypeOf?Object.setPrototypeOf(e,t):e.__proto__=t)}(t,e),r(t,[{key:"componentDidMount",value:function(){this.mounted=!0,this.copyInputStyles(),this.updateInputWidth()}},{key:"UNSAFE_componentWillReceiveProps",value:function(e){var t=e.id;t!==this.props.id&&this.setState({inputId:t||d()})}},{key:"componentDidUpdate",value:function(e,t){t.inputWidth!==this.state.inputWidth&&"function"==typeof this.props.onAutosize&&this.props.onAutosize(this.state.inputWidth),this.updateInputWidth()}},{key:"componentWillUnmount",value:function(){this.mounted=!1}},{key:"copyInputStyles",value:function(){if(this.mounted&&window.getComputedStyle){var e=this.input&&window.getComputedStyle(this.input);e&&(h(e,this.sizer),this.placeHolderSizer&&h(e,this.placeHolderSizer))}}},{key:"updateInputWidth",value:function(){if(this.mounted&&this.sizer&&void 0!==this.sizer.scrollWidth){var e=void 0;e=this.props.placeholder&&(!this.props.value||this.props.value&&this.props.placeholderIsMinWidth)?Math.max(this.sizer.scrollWidth,this.placeHolderSizer.scrollWidth)+2:this.sizer.scrollWidth+2,(e+="number"===this.props.type&&void 0===this.props.extraWidth?16:parseInt(this.props.extraWidth)||0)<this.props.minWidth&&(e=this.props.minWidth),e!==this.state.inputWidth&&this.setState({inputWidth:e})}}},{key:"getInput",value:function(){return this.input}},{key:"focus",value:function(){this.input.focus()}},{key:"blur",value:function(){this.input.blur()}},{key:"select",value:function(){this.input.select()}},{key:"renderStyles",value:function(){var e=this.props.injectStyles;return f&&e?s.default.createElement("style",{dangerouslySetInnerHTML:{__html:"input#"+this.state.inputId+"::-ms-clear {display: none;}"}}):null}},{key:"render",value:function(){var e=[this.props.defaultValue,this.props.value,""].reduce((function(e,t){return null!=e?e:t})),t=i({},this.props.style);t.display||(t.display="inline-block");var n=i({boxSizing:"content-box",width:this.state.inputWidth+"px"},this.props.inputStyle),r=function(e,t){var n={};for(var i in e)t.indexOf(i)>=0||Object.prototype.hasOwnProperty.call(e,i)&&(n[i]=e[i]);return n}(this.props,[]);return function(e){c.forEach((function(t){return delete e[t]}))}(r),r.className=this.props.inputClassName,r.id=this.state.inputId,r.style=n,s.default.createElement("div",{className:this.props.className,style:t},this.renderStyles(),s.default.createElement("input",i({},r,{ref:this.inputRef})),s.default.createElement("div",{ref:this.sizerRef,style:u},e),this.props.placeholder?s.default.createElement("div",{ref:this.placeHolderSizerRef,style:u},this.props.placeholder):null)}}]),t}(o.Component);p.propTypes={className:a.default.string,defaultValue:a.default.any,extraWidth:a.default.oneOfType([a.default.number,a.default.string]),id:a.default.string,injectStyles:a.default.bool,inputClassName:a.default.string,inputRef:a.default.func,inputStyle:a.default.object,minWidth:a.default.oneOfType([a.default.number,a.default.string]),onAutosize:a.default.func,onChange:a.default.func,placeholder:a.default.string,placeholderIsMinWidth:a.default.bool,style:a.default.object,value:a.default.any},p.defaultProps={minWidth:1,injectStyles:!0},t.default=p},function(e,t,n){var i=n(68),r=n(69),o=n(70);e.exports=function(e,t){return i(e)||r(e,t)||o()}},function(e,t,n){var i=n(74),r=n(131)((function(e,t,n){i(e,t,n)}));e.exports=r},function(e,t,n){var i=n(32),r=n(7);e.exports=function(e,t,n){var o=!0,s=!0;if("function"!=typeof e)throw new TypeError("Expected a function");return r(n)&&(o="leading"in n?!!n.leading:o,s="trailing"in n?!!n.trailing:s),i(e,t,{leading:o,maxWait:t,trailing:s})}},function(e,t,n){"use strict";
/*
object-assign
(c) Sindre Sorhus
@license MIT
*/var i=Object.getOwnPropertySymbols,r=Object.prototype.hasOwnProperty,o=Object.prototype.propertyIsEnumerable;function s(e){if(null==e)throw new TypeError("Object.assign cannot be called with null or undefined");return Object(e)}e.exports=function(){try{if(!Object.assign)return!1;var e=new String("abc");if(e[5]="de","5"===Object.getOwnPropertyNames(e)[0])return!1;for(var t={},n=0;n<10;n++)t["_"+String.fromCharCode(n)]=n;if("0123456789"!==Object.getOwnPropertyNames(t).map((function(e){return t[e]})).join(""))return!1;var i={};return"abcdefghijklmnopqrst".split("").forEach((function(e){i[e]=e})),"abcdefghijklmnopqrst"===Object.keys(Object.assign({},i)).join("")}catch(e){return!1}}()?Object.assign:function(e,t){for(var n,a,l=s(e),u=1;u<arguments.length;u++){for(var c in n=Object(arguments[u]))r.call(n,c)&&(l[c]=n[c]);if(i){a=i(n);for(var h=0;h<a.length;h++)o.call(n,a[h])&&(l[a[h]]=n[a[h]])}}return l}},function(e,t){e.exports=function(){throw new Error("define cannot be used indirect")}},function(e,t,n){var i=n(27)(n(11),"Map");e.exports=i},function(e,t,n){var i=n(11).Symbol;e.exports=i},function(e,t,n){(function(t){var n="object"==typeof t&&t&&t.Object===Object&&t;e.exports=n}).call(this,n(26))},function(e,t,n){var i=n(29),r=n(22);e.exports=function(e,t,n){(void 0===n||r(e[t],n))&&(void 0!==n||t in e)||i(e,t,n)}},function(e,t,n){var i=n(27),r=function(){try{var e=i(Object,"defineProperty");return e({},"",{}),e}catch(e){}}();e.exports=r},function(e,t,n){var i=n(116)(Object.getPrototypeOf,Object);e.exports=i},function(e,t){var n=Object.prototype;e.exports=function(e){var t=e&&e.constructor;return e===("function"==typeof t&&t.prototype||n)}},function(e,t,n){var i=n(117),r=n(12),o=Object.prototype,s=o.hasOwnProperty,a=o.propertyIsEnumerable,l=i(function(){return arguments}())?i:function(e){return r(e)&&s.call(e,"callee")&&!a.call(e,"callee")};e.exports=l},function(e,t){var n=Array.isArray;e.exports=n},function(e,t){e.exports=function(e){return"number"==typeof e&&e>-1&&e%1==0&&e<=9007199254740991}},function(e,t,n){(function(e){var i=n(11),r=n(119),o=t&&!t.nodeType&&t,s=o&&"object"==typeof e&&e&&!e.nodeType&&e,a=s&&s.exports===o?i.Buffer:void 0,l=(a?a.isBuffer:void 0)||r;e.exports=l}).call(this,n(25)(e))},function(e,t,n){var i=n(121),r=n(122),o=n(123),s=o&&o.isTypedArray,a=s?r(s):i;e.exports=a},function(e,t){e.exports=function(e,t){if(("constructor"!==t||"function"!=typeof e[t])&&"__proto__"!=t)return e[t]}},function(e,t,n){var i=n(127),r=n(129),o=n(30);e.exports=function(e){return o(e)?i(e,!0):r(e)}},function(e,t){var n=/^(?:0|[1-9]\d*)$/;e.exports=function(e,t){var i=typeof e;return!!(t=null==t?9007199254740991:t)&&("number"==i||"symbol"!=i&&n.test(e))&&e>-1&&e%1==0&&e<t}},function(e,t){e.exports=function(e){return e}},function(e,t,n){"use strict";var i,r=function(){return void 0===i&&(i=Boolean(window&&document&&document.all&&!window.atob)),i},o=function(){var e={};return function(t){if(void 0===e[t]){var n=document.querySelector(t);if(window.HTMLIFrameElement&&n instanceof window.HTMLIFrameElement)try{n=n.contentDocument.head}catch(e){n=null}e[t]=n}return e[t]}}(),s={};function a(e,t,n){for(var i=0;i<t.length;i++){var r={css:t[i][1],media:t[i][2],sourceMap:t[i][3]};s[e][i]?s[e][i](r):s[e].push(g(r,n))}}function l(e){var t=document.createElement("style"),i=e.attributes||{};if(void 0===i.nonce){var r=n.nc;r&&(i.nonce=r)}if(Object.keys(i).forEach((function(e){t.setAttribute(e,i[e])})),"function"==typeof e.insert)e.insert(t);else{var s=o(e.insert||"head");if(!s)throw new Error("Couldn't find a style target. This probably means that the value for the 'insert' parameter is invalid.");s.appendChild(t)}return t}var u,c=(u=[],function(e,t){return u[e]=t,u.filter(Boolean).join("\n")});function h(e,t,n,i){var r=n?"":i.css;if(e.styleSheet)e.styleSheet.cssText=c(t,r);else{var o=document.createTextNode(r),s=e.childNodes;s[t]&&e.removeChild(s[t]),s.length?e.insertBefore(o,s[t]):e.appendChild(o)}}function f(e,t,n){var i=n.css,r=n.media,o=n.sourceMap;if(r?e.setAttribute("media",r):e.removeAttribute("media"),o&&btoa&&(i+="\n/*# sourceMappingURL=data:application/json;base64,".concat(btoa(unescape(encodeURIComponent(JSON.stringify(o))))," */")),e.styleSheet)e.styleSheet.cssText=i;else{for(;e.firstChild;)e.removeChild(e.firstChild);e.appendChild(document.createTextNode(i))}}var d=null,p=0;function g(e,t){var n,i,r;if(t.singleton){var o=p++;n=d||(d=l(t)),i=h.bind(null,n,o,!1),r=h.bind(null,n,o,!0)}else n=l(t),i=f.bind(null,n,t),r=function(){!function(e){if(null===e.parentNode)return!1;e.parentNode.removeChild(e)}(n)};return i(e),function(t){if(t){if(t.css===e.css&&t.media===e.media&&t.sourceMap===e.sourceMap)return;i(e=t)}else r()}}e.exports=function(e,t,n){return(n=n||{}).singleton||"boolean"==typeof n.singleton||(n.singleton=r()),e=n.base?e+n.base:e,t=t||[],s[e]||(s[e]=[]),a(e,t,n),function(t){if(t=t||[],"[object Array]"===Object.prototype.toString.call(t)){s[e]||(s[e]=[]),a(e,t,n);for(var i=t.length;i<s[e].length;i++)s[e][i]();s[e].length=t.length,0===s[e].length&&delete s[e]}}}},function(e,t,n){"use strict";e.exports=function(e){var t=[];return t.toString=function(){return this.map((function(t){var n=function(e,t){var n=e[1]||"",i=e[3];if(!i)return n;if(t&&"function"==typeof btoa){var r=(s=i,a=btoa(unescape(encodeURIComponent(JSON.stringify(s)))),l="sourceMappingURL=data:application/json;charset=utf-8;base64,".concat(a),"/*# ".concat(l," */")),o=i.sources.map((function(e){return"/*# sourceURL=".concat(i.sourceRoot||"").concat(e," */")}));return[n].concat(o).concat([r]).join("\n")}var s,a,l;return[n].join("\n")}(t,e);return t[2]?"@media ".concat(t[2]," {").concat(n,"}"):n})).join("")},t.i=function(e,n){"string"==typeof e&&(e=[[null,e,""]]);for(var i=0;i<e.length;i++){var r=[].concat(e[i]);n&&(r[2]?r[2]="".concat(n," and ").concat(r[2]):r[2]=n),t.push(r)}},t}},function(e,n){e.exports=t},function(e,t){function n(){this.Diff_Timeout=1,this.Diff_EditCost=4,this.Match_Threshold=.5,this.Match_Distance=1e3,this.Patch_DeleteThreshold=.5,this.Patch_Margin=4,this.Match_MaxBits=32}n.Diff,n.prototype.diff_main=function(e,t,n,i){void 0===i&&(i=this.Diff_Timeout<=0?Number.MAX_VALUE:(new Date).getTime()+1e3*this.Diff_Timeout);var r=i;if(null==e||null==t)throw new Error("Null input. (diff_main)");if(e==t)return e?[[0,e]]:[];void 0===n&&(n=!0);var o=n,s=this.diff_commonPrefix(e,t),a=e.substring(0,s);e=e.substring(s),t=t.substring(s),s=this.diff_commonSuffix(e,t);var l=e.substring(e.length-s);e=e.substring(0,e.length-s),t=t.substring(0,t.length-s);var u=this.diff_compute_(e,t,o,r);return a&&u.unshift([0,a]),l&&u.push([0,l]),this.diff_cleanupMerge(u),u},n.prototype.diff_compute_=function(e,t,n,i){var r;if(!e)return[[1,t]];if(!t)return[[-1,e]];var o=e.length>t.length?e:t,s=e.length>t.length?t:e,a=o.indexOf(s);if(-1!=a)return r=[[1,o.substring(0,a)],[0,s],[1,o.substring(a+s.length)]],e.length>t.length&&(r[0][0]=r[2][0]=-1),r;if(1==s.length)return[[-1,e],[1,t]];var l=this.diff_halfMatch_(e,t);if(l){var u=l[0],c=l[1],h=l[2],f=l[3],d=l[4],p=this.diff_main(u,h,n,i),g=this.diff_main(c,f,n,i);return p.concat([[0,d]],g)}return n&&e.length>100&&t.length>100?this.diff_lineMode_(e,t,i):this.diff_bisect_(e,t,i)},n.prototype.diff_lineMode_=function(e,t,n){e=(c=this.diff_linesToChars_(e,t)).chars1,t=c.chars2;var i=c.lineArray,r=this.diff_main(e,t,!1,n);this.diff_charsToLines_(r,i),this.diff_cleanupSemantic(r),r.push([0,""]);for(var o=0,s=0,a=0,l="",u="";o<r.length;){switch(r[o][0]){case 1:a++,u+=r[o][1];break;case-1:s++,l+=r[o][1];break;case 0:if(s>=1&&a>=1){r.splice(o-s-a,s+a),o=o-s-a;for(var c,h=(c=this.diff_main(l,u,!1,n)).length-1;h>=0;h--)r.splice(o,0,c[h]);o+=c.length}a=0,s=0,l="",u=""}o++}return r.pop(),r},n.prototype.diff_bisect_=function(e,t,n){for(var i=e.length,r=t.length,o=Math.ceil((i+r)/2),s=o,a=2*o,l=new Array(a),u=new Array(a),c=0;c<a;c++)l[c]=-1,u[c]=-1;l[s+1]=0,u[s+1]=0;for(var h=i-r,f=h%2!=0,d=0,p=0,g=0,m=0,v=0;v<o&&!((new Date).getTime()>n);v++){for(var b=-v+d;b<=v-p;b+=2){for(var y=s+b,w=(x=b==-v||b!=v&&l[y-1]<l[y+1]?l[y+1]:l[y-1]+1)-b;x<i&&w<r&&e.charAt(x)==t.charAt(w);)x++,w++;if(l[y]=x,x>i)p+=2;else if(w>r)d+=2;else if(f){if((E=s+h-b)>=0&&E<a&&-1!=u[E])if(x>=(C=i-u[E]))return this.diff_bisectSplit_(e,t,x,w,n)}}for(var A=-v+g;A<=v-m;A+=2){for(var C,E=s+A,F=(C=A==-v||A!=v&&u[E-1]<u[E+1]?u[E+1]:u[E-1]+1)-A;C<i&&F<r&&e.charAt(i-C-1)==t.charAt(r-F-1);)C++,F++;if(u[E]=C,C>i)m+=2;else if(F>r)g+=2;else if(!f){if((y=s+h-A)>=0&&y<a&&-1!=l[y]){var x;w=s+(x=l[y])-y;if(x>=(C=i-C))return this.diff_bisectSplit_(e,t,x,w,n)}}}}return[[-1,e],[1,t]]},n.prototype.diff_bisectSplit_=function(e,t,n,i,r){var o=e.substring(0,n),s=t.substring(0,i),a=e.substring(n),l=t.substring(i),u=this.diff_main(o,s,!1,r),c=this.diff_main(a,l,!1,r);return u.concat(c)},n.prototype.diff_linesToChars_=function(e,t){var n=[],i={};function r(e){for(var t="",r=0,o=-1,s=n.length;o<e.length-1;){-1==(o=e.indexOf("\n",r))&&(o=e.length-1);var a=e.substring(r,o+1);r=o+1,(i.hasOwnProperty?i.hasOwnProperty(a):void 0!==i[a])?t+=String.fromCharCode(i[a]):(t+=String.fromCharCode(s),i[a]=s,n[s++]=a)}return t}return n[0]="",{chars1:r(e),chars2:r(t),lineArray:n}},n.prototype.diff_charsToLines_=function(e,t){for(var n=0;n<e.length;n++){for(var i=e[n][1],r=[],o=0;o<i.length;o++)r[o]=t[i.charCodeAt(o)];e[n][1]=r.join("")}},n.prototype.diff_commonPrefix=function(e,t){if(!e||!t||e.charAt(0)!=t.charAt(0))return 0;for(var n=0,i=Math.min(e.length,t.length),r=i,o=0;n<r;)e.substring(o,r)==t.substring(o,r)?o=n=r:i=r,r=Math.floor((i-n)/2+n);return r},n.prototype.diff_commonSuffix=function(e,t){if(!e||!t||e.charAt(e.length-1)!=t.charAt(t.length-1))return 0;for(var n=0,i=Math.min(e.length,t.length),r=i,o=0;n<r;)e.substring(e.length-r,e.length-o)==t.substring(t.length-r,t.length-o)?o=n=r:i=r,r=Math.floor((i-n)/2+n);return r},n.prototype.diff_commonOverlap_=function(e,t){var n=e.length,i=t.length;if(0==n||0==i)return 0;n>i?e=e.substring(n-i):n<i&&(t=t.substring(0,n));var r=Math.min(n,i);if(e==t)return r;for(var o=0,s=1;;){var a=e.substring(r-s),l=t.indexOf(a);if(-1==l)return o;s+=l,0!=l&&e.substring(r-s)!=t.substring(0,s)||(o=s,s++)}},n.prototype.diff_halfMatch_=function(e,t){if(this.Diff_Timeout<=0)return null;var n=e.length>t.length?e:t,i=e.length>t.length?t:e;if(n.length<4||2*i.length<n.length)return null;var r=this;function o(e,t,n){for(var i,o,s,a,l=e.substring(n,n+Math.floor(e.length/4)),u=-1,c="";-1!=(u=t.indexOf(l,u+1));){var h=r.diff_commonPrefix(e.substring(n),t.substring(u)),f=r.diff_commonSuffix(e.substring(0,n),t.substring(0,u));c.length<f+h&&(c=t.substring(u-f,u)+t.substring(u,u+h),i=e.substring(0,n-f),o=e.substring(n+h),s=t.substring(0,u-f),a=t.substring(u+h))}return 2*c.length>=e.length?[i,o,s,a,c]:null}var s,a,l,u,c,h=o(n,i,Math.ceil(n.length/4)),f=o(n,i,Math.ceil(n.length/2));return h||f?(s=f?h&&h[4].length>f[4].length?h:f:h,e.length>t.length?(a=s[0],l=s[1],u=s[2],c=s[3]):(u=s[0],c=s[1],a=s[2],l=s[3]),[a,l,u,c,s[4]]):null},n.prototype.diff_cleanupSemantic=function(e){for(var t=!1,n=[],i=0,r=null,o=0,s=0,a=0,l=0,u=0;o<e.length;)0==e[o][0]?(n[i++]=o,s=l,a=u,l=0,u=0,r=e[o][1]):(1==e[o][0]?l+=e[o][1].length:u+=e[o][1].length,r&&r.length<=Math.max(s,a)&&r.length<=Math.max(l,u)&&(e.splice(n[i-1],0,[-1,r]),e[n[i-1]+1][0]=1,i--,o=--i>0?n[i-1]:-1,s=0,a=0,l=0,u=0,r=null,t=!0)),o++;for(t&&this.diff_cleanupMerge(e),this.diff_cleanupSemanticLossless(e),o=1;o<e.length;){if(-1==e[o-1][0]&&1==e[o][0]){var c=e[o-1][1],h=e[o][1],f=this.diff_commonOverlap_(c,h),d=this.diff_commonOverlap_(h,c);f>=d?(f>=c.length/2||f>=h.length/2)&&(e.splice(o,0,[0,h.substring(0,f)]),e[o-1][1]=c.substring(0,c.length-f),e[o+1][1]=h.substring(f),o++):(d>=c.length/2||d>=h.length/2)&&(e.splice(o,0,[0,c.substring(0,d)]),e[o-1][0]=1,e[o-1][1]=h.substring(0,h.length-d),e[o+1][0]=-1,e[o+1][1]=c.substring(d),o++),o++}o++}},n.prototype.diff_cleanupSemanticLossless=function(e){function t(e,t){if(!e||!t)return 6;var i=e.charAt(e.length-1),r=t.charAt(0),o=i.match(n.nonAlphaNumericRegex_),s=r.match(n.nonAlphaNumericRegex_),a=o&&i.match(n.whitespaceRegex_),l=s&&r.match(n.whitespaceRegex_),u=a&&i.match(n.linebreakRegex_),c=l&&r.match(n.linebreakRegex_),h=u&&e.match(n.blanklineEndRegex_),f=c&&t.match(n.blanklineStartRegex_);return h||f?5:u||c?4:o&&!a&&l?3:a||l?2:o||s?1:0}for(var i=1;i<e.length-1;){if(0==e[i-1][0]&&0==e[i+1][0]){var r=e[i-1][1],o=e[i][1],s=e[i+1][1],a=this.diff_commonSuffix(r,o);if(a){var l=o.substring(o.length-a);r=r.substring(0,r.length-a),o=l+o.substring(0,o.length-a),s=l+s}for(var u=r,c=o,h=s,f=t(r,o)+t(o,s);o.charAt(0)===s.charAt(0);){r+=o.charAt(0),o=o.substring(1)+s.charAt(0),s=s.substring(1);var d=t(r,o)+t(o,s);d>=f&&(f=d,u=r,c=o,h=s)}e[i-1][1]!=u&&(u?e[i-1][1]=u:(e.splice(i-1,1),i--),e[i][1]=c,h?e[i+1][1]=h:(e.splice(i+1,1),i--))}i++}},n.nonAlphaNumericRegex_=/[^a-zA-Z0-9]/,n.whitespaceRegex_=/\s/,n.linebreakRegex_=/[\r\n]/,n.blanklineEndRegex_=/\n\r?\n$/,n.blanklineStartRegex_=/^\r?\n\r?\n/,n.prototype.diff_cleanupEfficiency=function(e){for(var t=!1,n=[],i=0,r=null,o=0,s=!1,a=!1,l=!1,u=!1;o<e.length;)0==e[o][0]?(e[o][1].length<this.Diff_EditCost&&(l||u)?(n[i++]=o,s=l,a=u,r=e[o][1]):(i=0,r=null),l=u=!1):(-1==e[o][0]?u=!0:l=!0,r&&(s&&a&&l&&u||r.length<this.Diff_EditCost/2&&s+a+l+u==3)&&(e.splice(n[i-1],0,[-1,r]),e[n[i-1]+1][0]=1,i--,r=null,s&&a?(l=u=!0,i=0):(o=--i>0?n[i-1]:-1,l=u=!1),t=!0)),o++;t&&this.diff_cleanupMerge(e)},n.prototype.diff_cleanupMerge=function(e){e.push([0,""]);for(var t,n=0,i=0,r=0,o="",s="";n<e.length;)switch(e[n][0]){case 1:r++,s+=e[n][1],n++;break;case-1:i++,o+=e[n][1],n++;break;case 0:i+r>1?(0!==i&&0!==r&&(0!==(t=this.diff_commonPrefix(s,o))&&(n-i-r>0&&0==e[n-i-r-1][0]?e[n-i-r-1][1]+=s.substring(0,t):(e.splice(0,0,[0,s.substring(0,t)]),n++),s=s.substring(t),o=o.substring(t)),0!==(t=this.diff_commonSuffix(s,o))&&(e[n][1]=s.substring(s.length-t)+e[n][1],s=s.substring(0,s.length-t),o=o.substring(0,o.length-t))),0===i?e.splice(n-r,i+r,[1,s]):0===r?e.splice(n-i,i+r,[-1,o]):e.splice(n-i-r,i+r,[-1,o],[1,s]),n=n-i-r+(i?1:0)+(r?1:0)+1):0!==n&&0==e[n-1][0]?(e[n-1][1]+=e[n][1],e.splice(n,1)):n++,r=0,i=0,o="",s=""}""===e[e.length-1][1]&&e.pop();var a=!1;for(n=1;n<e.length-1;)0==e[n-1][0]&&0==e[n+1][0]&&(e[n][1].substring(e[n][1].length-e[n-1][1].length)==e[n-1][1]?(e[n][1]=e[n-1][1]+e[n][1].substring(0,e[n][1].length-e[n-1][1].length),e[n+1][1]=e[n-1][1]+e[n+1][1],e.splice(n-1,1),a=!0):e[n][1].substring(0,e[n+1][1].length)==e[n+1][1]&&(e[n-1][1]+=e[n+1][1],e[n][1]=e[n][1].substring(e[n+1][1].length)+e[n+1][1],e.splice(n+1,1),a=!0)),n++;a&&this.diff_cleanupMerge(e)},n.prototype.diff_xIndex=function(e,t){var n,i=0,r=0,o=0,s=0;for(n=0;n<e.length&&(1!==e[n][0]&&(i+=e[n][1].length),-1!==e[n][0]&&(r+=e[n][1].length),!(i>t));n++)o=i,s=r;return e.length!=n&&-1===e[n][0]?s:s+(t-o)},n.prototype.diff_prettyHtml=function(e){for(var t=[],n=/&/g,i=/</g,r=/>/g,o=/\n/g,s=0;s<e.length;s++){var a=e[s][0],l=e[s][1].replace(n,"&amp;").replace(i,"&lt;").replace(r,"&gt;").replace(o,"&para;<br>");switch(a){case 1:t[s]='<ins style="background:#e6ffe6;">'+l+"</ins>";break;case-1:t[s]='<del style="background:#ffe6e6;">'+l+"</del>";break;case 0:t[s]="<span>"+l+"</span>"}}return t.join("")},n.prototype.diff_text1=function(e){for(var t=[],n=0;n<e.length;n++)1!==e[n][0]&&(t[n]=e[n][1]);return t.join("")},n.prototype.diff_text2=function(e){for(var t=[],n=0;n<e.length;n++)-1!==e[n][0]&&(t[n]=e[n][1]);return t.join("")},n.prototype.diff_levenshtein=function(e){for(var t=0,n=0,i=0,r=0;r<e.length;r++){var o=e[r][0],s=e[r][1];switch(o){case 1:n+=s.length;break;case-1:i+=s.length;break;case 0:t+=Math.max(n,i),n=0,i=0}}return t+=Math.max(n,i)},n.prototype.diff_toDelta=function(e){for(var t=[],n=0;n<e.length;n++)switch(e[n][0]){case 1:t[n]="+"+encodeURI(e[n][1]);break;case-1:t[n]="-"+e[n][1].length;break;case 0:t[n]="="+e[n][1].length}return t.join("\t").replace(/%20/g," ")},n.prototype.diff_fromDelta=function(e,t){for(var n=[],i=0,r=0,o=t.split(/\t/g),s=0;s<o.length;s++){var a=o[s].substring(1);switch(o[s].charAt(0)){case"+":try{n[i++]=[1,decodeURI(a)]}catch(e){throw new Error("Illegal escape in diff_fromDelta: "+a)}break;case"-":case"=":var l=parseInt(a,10);if(isNaN(l)||l<0)throw new Error("Invalid number in diff_fromDelta: "+a);var u=e.substring(r,r+=l);"="==o[s].charAt(0)?n[i++]=[0,u]:n[i++]=[-1,u];break;default:if(o[s])throw new Error("Invalid diff operation in diff_fromDelta: "+o[s])}}if(r!=e.length)throw new Error("Delta length ("+r+") does not equal source text length ("+e.length+").");return n},n.prototype.match_main=function(e,t,n){if(null==e||null==t||null==n)throw new Error("Null input. (match_main)");return n=Math.max(0,Math.min(n,e.length)),e==t?0:e.length?e.substring(n,n+t.length)==t?n:this.match_bitap_(e,t,n):-1},n.prototype.match_bitap_=function(e,t,n){if(t.length>this.Match_MaxBits)throw new Error("Pattern too long for this browser.");var i=this.match_alphabet_(t),r=this;function o(e,i){var o=e/t.length,s=Math.abs(n-i);return r.Match_Distance?o+s/r.Match_Distance:s?1:o}var s=this.Match_Threshold,a=e.indexOf(t,n);-1!=a&&(s=Math.min(o(0,a),s),-1!=(a=e.lastIndexOf(t,n+t.length))&&(s=Math.min(o(0,a),s)));var l,u,c=1<<t.length-1;a=-1;for(var h,f=t.length+e.length,d=0;d<t.length;d++){for(l=0,u=f;l<u;)o(d,n+u)<=s?l=u:f=u,u=Math.floor((f-l)/2+l);f=u;var p=Math.max(1,n-u+1),g=Math.min(n+u,e.length)+t.length,m=Array(g+2);m[g+1]=(1<<d)-1;for(var v=g;v>=p;v--){var b=i[e.charAt(v-1)];if(m[v]=0===d?(m[v+1]<<1|1)&b:(m[v+1]<<1|1)&b|(h[v+1]|h[v])<<1|1|h[v+1],m[v]&c){var y=o(d,v-1);if(y<=s){if(s=y,!((a=v-1)>n))break;p=Math.max(1,2*n-a)}}}if(o(d+1,n)>s)break;h=m}return a},n.prototype.match_alphabet_=function(e){for(var t={},n=0;n<e.length;n++)t[e.charAt(n)]=0;for(n=0;n<e.length;n++)t[e.charAt(n)]|=1<<e.length-n-1;return t},n.prototype.patch_addContext_=function(e,t){if(0!=t.length){for(var n=t.substring(e.start2,e.start2+e.length1),i=0;t.indexOf(n)!=t.lastIndexOf(n)&&n.length<this.Match_MaxBits-this.Patch_Margin-this.Patch_Margin;)i+=this.Patch_Margin,n=t.substring(e.start2-i,e.start2+e.length1+i);i+=this.Patch_Margin;var r=t.substring(e.start2-i,e.start2);r&&e.diffs.unshift([0,r]);var o=t.substring(e.start2+e.length1,e.start2+e.length1+i);o&&e.diffs.push([0,o]),e.start1-=r.length,e.start2-=r.length,e.length1+=r.length+o.length,e.length2+=r.length+o.length}},n.prototype.patch_make=function(e,t,i){var r,o;if("string"==typeof e&&"string"==typeof t&&void 0===i)r=e,(o=this.diff_main(r,t,!0)).length>2&&(this.diff_cleanupSemantic(o),this.diff_cleanupEfficiency(o));else if(e&&"object"==typeof e&&void 0===t&&void 0===i)o=e,r=this.diff_text1(o);else if("string"==typeof e&&t&&"object"==typeof t&&void 0===i)r=e,o=t;else{if("string"!=typeof e||"string"!=typeof t||!i||"object"!=typeof i)throw new Error("Unknown call format to patch_make.");r=e,o=i}if(0===o.length)return[];for(var s=[],a=new n.patch_obj,l=0,u=0,c=0,h=r,f=r,d=0;d<o.length;d++){var p=o[d][0],g=o[d][1];switch(l||0===p||(a.start1=u,a.start2=c),p){case 1:a.diffs[l++]=o[d],a.length2+=g.length,f=f.substring(0,c)+g+f.substring(c);break;case-1:a.length1+=g.length,a.diffs[l++]=o[d],f=f.substring(0,c)+f.substring(c+g.length);break;case 0:g.length<=2*this.Patch_Margin&&l&&o.length!=d+1?(a.diffs[l++]=o[d],a.length1+=g.length,a.length2+=g.length):g.length>=2*this.Patch_Margin&&l&&(this.patch_addContext_(a,h),s.push(a),a=new n.patch_obj,l=0,h=f,u=c)}1!==p&&(u+=g.length),-1!==p&&(c+=g.length)}return l&&(this.patch_addContext_(a,h),s.push(a)),s},n.prototype.patch_deepCopy=function(e){for(var t=[],i=0;i<e.length;i++){var r=e[i],o=new n.patch_obj;o.diffs=[];for(var s=0;s<r.diffs.length;s++)o.diffs[s]=r.diffs[s].slice();o.start1=r.start1,o.start2=r.start2,o.length1=r.length1,o.length2=r.length2,t[i]=o}return t},n.prototype.patch_apply=function(e,t){if(0==e.length)return[t,[]];e=this.patch_deepCopy(e);var n=this.patch_addPadding(e);t=n+t+n,this.patch_splitMax(e);for(var i=0,r=[],o=0;o<e.length;o++){var s,a,l=e[o].start2+i,u=this.diff_text1(e[o].diffs),c=-1;if(u.length>this.Match_MaxBits?-1!=(s=this.match_main(t,u.substring(0,this.Match_MaxBits),l))&&(-1==(c=this.match_main(t,u.substring(u.length-this.Match_MaxBits),l+u.length-this.Match_MaxBits))||s>=c)&&(s=-1):s=this.match_main(t,u,l),-1==s)r[o]=!1,i-=e[o].length2-e[o].length1;else if(r[o]=!0,i=s-l,u==(a=-1==c?t.substring(s,s+u.length):t.substring(s,c+this.Match_MaxBits)))t=t.substring(0,s)+this.diff_text2(e[o].diffs)+t.substring(s+u.length);else{var h=this.diff_main(u,a,!1);if(u.length>this.Match_MaxBits&&this.diff_levenshtein(h)/u.length>this.Patch_DeleteThreshold)r[o]=!1;else{this.diff_cleanupSemanticLossless(h);for(var f,d=0,p=0;p<e[o].diffs.length;p++){var g=e[o].diffs[p];0!==g[0]&&(f=this.diff_xIndex(h,d)),1===g[0]?t=t.substring(0,s+f)+g[1]+t.substring(s+f):-1===g[0]&&(t=t.substring(0,s+f)+t.substring(s+this.diff_xIndex(h,d+g[1].length))),-1!==g[0]&&(d+=g[1].length)}}}}return[t=t.substring(n.length,t.length-n.length),r]},n.prototype.patch_addPadding=function(e){for(var t=this.Patch_Margin,n="",i=1;i<=t;i++)n+=String.fromCharCode(i);for(i=0;i<e.length;i++)e[i].start1+=t,e[i].start2+=t;var r=e[0],o=r.diffs;if(0==o.length||0!=o[0][0])o.unshift([0,n]),r.start1-=t,r.start2-=t,r.length1+=t,r.length2+=t;else if(t>o[0][1].length){var s=t-o[0][1].length;o[0][1]=n.substring(o[0][1].length)+o[0][1],r.start1-=s,r.start2-=s,r.length1+=s,r.length2+=s}if(0==(o=(r=e[e.length-1]).diffs).length||0!=o[o.length-1][0])o.push([0,n]),r.length1+=t,r.length2+=t;else if(t>o[o.length-1][1].length){s=t-o[o.length-1][1].length;o[o.length-1][1]+=n.substring(0,s),r.length1+=s,r.length2+=s}return n},n.prototype.patch_splitMax=function(e){for(var t=this.Match_MaxBits,i=0;i<e.length;i++)if(!(e[i].length1<=t)){var r=e[i];e.splice(i--,1);for(var o=r.start1,s=r.start2,a="";0!==r.diffs.length;){var l=new n.patch_obj,u=!0;for(l.start1=o-a.length,l.start2=s-a.length,""!==a&&(l.length1=l.length2=a.length,l.diffs.push([0,a]));0!==r.diffs.length&&l.length1<t-this.Patch_Margin;){var c=r.diffs[0][0],h=r.diffs[0][1];1===c?(l.length2+=h.length,s+=h.length,l.diffs.push(r.diffs.shift()),u=!1):-1===c&&1==l.diffs.length&&0==l.diffs[0][0]&&h.length>2*t?(l.length1+=h.length,o+=h.length,u=!1,l.diffs.push([c,h]),r.diffs.shift()):(h=h.substring(0,t-l.length1-this.Patch_Margin),l.length1+=h.length,o+=h.length,0===c?(l.length2+=h.length,s+=h.length):u=!1,l.diffs.push([c,h]),h==r.diffs[0][1]?r.diffs.shift():r.diffs[0][1]=r.diffs[0][1].substring(h.length))}a=(a=this.diff_text2(l.diffs)).substring(a.length-this.Patch_Margin);var f=this.diff_text1(r.diffs).substring(0,this.Patch_Margin);""!==f&&(l.length1+=f.length,l.length2+=f.length,0!==l.diffs.length&&0===l.diffs[l.diffs.length-1][0]?l.diffs[l.diffs.length-1][1]+=f:l.diffs.push([0,f])),u||e.splice(++i,0,l)}}},n.prototype.patch_toText=function(e){for(var t=[],n=0;n<e.length;n++)t[n]=e[n];return t.join("")},n.prototype.patch_fromText=function(e){var t=[];if(!e)return t;for(var i=e.split("\n"),r=0,o=/^@@ -(\d+),?(\d*) \+(\d+),?(\d*) @@$/;r<i.length;){var s=i[r].match(o);if(!s)throw new Error("Invalid patch string: "+i[r]);var a=new n.patch_obj;for(t.push(a),a.start1=parseInt(s[1],10),""===s[2]?(a.start1--,a.length1=1):"0"==s[2]?a.length1=0:(a.start1--,a.length1=parseInt(s[2],10)),a.start2=parseInt(s[3],10),""===s[4]?(a.start2--,a.length2=1):"0"==s[4]?a.length2=0:(a.start2--,a.length2=parseInt(s[4],10)),r++;r<i.length;){var l=i[r].charAt(0);try{var u=decodeURI(i[r].substring(1))}catch(e){throw new Error("Illegal escape in patch_fromText: "+u)}if("-"==l)a.diffs.push([-1,u]);else if("+"==l)a.diffs.push([1,u]);else if(" "==l)a.diffs.push([0,u]);else{if("@"==l)break;if(""!==l)throw new Error('Invalid patch mode "'+l+'" in: '+u)}r++}}return t},n.patch_obj=function(){this.diffs=[],this.start1=null,this.start2=null,this.length1=0,this.length2=0},n.patch_obj.prototype.toString=function(){for(var e,t=["@@ -"+(0===this.length1?this.start1+",0":1==this.length1?this.start1+1:this.start1+1+","+this.length1)+" +"+(0===this.length2?this.start2+",0":1==this.length2?this.start2+1:this.start2+1+","+this.length2)+" @@\n"],n=0;n<this.diffs.length;n++){switch(this.diffs[n][0]){case 1:e="+";break;case-1:e="-";break;case 0:e=" "}t[n+1]=e+encodeURI(this.diffs[n][1])+"\n"}return t.join("").replace(/%20/g," ")},e.exports=n,e.exports.diff_match_patch=n,e.exports.DIFF_DELETE=-1,e.exports.DIFF_INSERT=1,e.exports.DIFF_EQUAL=0},function(e,t,n){(function(e,i){var r;
/**
 * @license
 * Lodash <https://lodash.com/>
 * Copyright OpenJS Foundation and other contributors <https://openjsf.org/>
 * Released under MIT license <https://lodash.com/license>
 * Based on Underscore.js 1.8.3 <http://underscorejs.org/LICENSE>
 * Copyright Jeremy Ashkenas, DocumentCloud and Investigative Reporters & Editors
 */(function(){var o="Expected a function",s="__lodash_placeholder__",a=[["ary",128],["bind",1],["bindKey",2],["curry",8],["curryRight",16],["flip",512],["partial",32],["partialRight",64],["rearg",256]],l="[object Arguments]",u="[object Array]",c="[object Boolean]",h="[object Date]",f="[object Error]",d="[object Function]",p="[object GeneratorFunction]",g="[object Map]",m="[object Number]",v="[object Object]",b="[object RegExp]",y="[object Set]",w="[object String]",A="[object Symbol]",C="[object WeakMap]",E="[object ArrayBuffer]",F="[object DataView]",x="[object Float32Array]",k="[object Float64Array]",S="[object Int8Array]",_="[object Int16Array]",D="[object Int32Array]",$="[object Uint8Array]",B="[object Uint16Array]",L="[object Uint32Array]",T=/\b__p \+= '';/g,R=/\b(__p \+=) '' \+/g,O=/(__e\(.*?\)|\b__t\)) \+\n'';/g,M=/&(?:amp|lt|gt|quot|#39);/g,P=/[&<>"']/g,I=RegExp(M.source),N=RegExp(P.source),W=/<%-([\s\S]+?)%>/g,z=/<%([\s\S]+?)%>/g,j=/<%=([\s\S]+?)%>/g,H=/\.|\[(?:[^[\]]*|(["'])(?:(?!\1)[^\\]|\\.)*?\1)\]/,V=/^\w*$/,U=/[^.[\]]+|\[(?:(-?\d+(?:\.\d+)?)|(["'])((?:(?!\2)[^\\]|\\.)*?)\2)\]|(?=(?:\.|\[\])(?:\.|\[\]|$))/g,K=/[\\^$.*+?()[\]{}|]/g,G=RegExp(K.source),q=/^\s+|\s+$/g,Y=/^\s+/,Q=/\s+$/,X=/\{(?:\n\/\* \[wrapped with .+\] \*\/)?\n?/,J=/\{\n\/\* \[wrapped with (.+)\] \*/,Z=/,? & /,ee=/[^\x00-\x2f\x3a-\x40\x5b-\x60\x7b-\x7f]+/g,te=/\\(\\)?/g,ne=/\$\{([^\\}]*(?:\\.[^\\}]*)*)\}/g,ie=/\w*$/,re=/^[-+]0x[0-9a-f]+$/i,oe=/^0b[01]+$/i,se=/^\[object .+?Constructor\]$/,ae=/^0o[0-7]+$/i,le=/^(?:0|[1-9]\d*)$/,ue=/[\xc0-\xd6\xd8-\xf6\xf8-\xff\u0100-\u017f]/g,ce=/($^)/,he=/['\n\r\u2028\u2029\\]/g,fe="\\u0300-\\u036f\\ufe20-\\ufe2f\\u20d0-\\u20ff",de="\\xac\\xb1\\xd7\\xf7\\x00-\\x2f\\x3a-\\x40\\x5b-\\x60\\x7b-\\xbf\\u2000-\\u206f \\t\\x0b\\f\\xa0\\ufeff\\n\\r\\u2028\\u2029\\u1680\\u180e\\u2000\\u2001\\u2002\\u2003\\u2004\\u2005\\u2006\\u2007\\u2008\\u2009\\u200a\\u202f\\u205f\\u3000",pe="[\\ud800-\\udfff]",ge="["+de+"]",me="["+fe+"]",ve="\\d+",be="[\\u2700-\\u27bf]",ye="[a-z\\xdf-\\xf6\\xf8-\\xff]",we="[^\\ud800-\\udfff"+de+ve+"\\u2700-\\u27bfa-z\\xdf-\\xf6\\xf8-\\xffA-Z\\xc0-\\xd6\\xd8-\\xde]",Ae="\\ud83c[\\udffb-\\udfff]",Ce="[^\\ud800-\\udfff]",Ee="(?:\\ud83c[\\udde6-\\uddff]){2}",Fe="[\\ud800-\\udbff][\\udc00-\\udfff]",xe="[A-Z\\xc0-\\xd6\\xd8-\\xde]",ke="(?:"+ye+"|"+we+")",Se="(?:"+xe+"|"+we+")",_e="(?:"+me+"|"+Ae+")"+"?",De="[\\ufe0e\\ufe0f]?"+_e+("(?:\\u200d(?:"+[Ce,Ee,Fe].join("|")+")[\\ufe0e\\ufe0f]?"+_e+")*"),$e="(?:"+[be,Ee,Fe].join("|")+")"+De,Be="(?:"+[Ce+me+"?",me,Ee,Fe,pe].join("|")+")",Le=RegExp("['’]","g"),Te=RegExp(me,"g"),Re=RegExp(Ae+"(?="+Ae+")|"+Be+De,"g"),Oe=RegExp([xe+"?"+ye+"+(?:['’](?:d|ll|m|re|s|t|ve))?(?="+[ge,xe,"$"].join("|")+")",Se+"+(?:['’](?:D|LL|M|RE|S|T|VE))?(?="+[ge,xe+ke,"$"].join("|")+")",xe+"?"+ke+"+(?:['’](?:d|ll|m|re|s|t|ve))?",xe+"+(?:['’](?:D|LL|M|RE|S|T|VE))?","\\d*(?:1ST|2ND|3RD|(?![123])\\dTH)(?=\\b|[a-z_])","\\d*(?:1st|2nd|3rd|(?![123])\\dth)(?=\\b|[A-Z_])",ve,$e].join("|"),"g"),Me=RegExp("[\\u200d\\ud800-\\udfff"+fe+"\\ufe0e\\ufe0f]"),Pe=/[a-z][A-Z]|[A-Z]{2}[a-z]|[0-9][a-zA-Z]|[a-zA-Z][0-9]|[^a-zA-Z0-9 ]/,Ie=["Array","Buffer","DataView","Date","Error","Float32Array","Float64Array","Function","Int8Array","Int16Array","Int32Array","Map","Math","Object","Promise","RegExp","Set","String","Symbol","TypeError","Uint8Array","Uint8ClampedArray","Uint16Array","Uint32Array","WeakMap","_","clearTimeout","isFinite","parseInt","setTimeout"],Ne=-1,We={};We[x]=We[k]=We[S]=We[_]=We[D]=We[$]=We["[object Uint8ClampedArray]"]=We[B]=We[L]=!0,We[l]=We[u]=We[E]=We[c]=We[F]=We[h]=We[f]=We[d]=We[g]=We[m]=We[v]=We[b]=We[y]=We[w]=We[C]=!1;var ze={};ze[l]=ze[u]=ze[E]=ze[F]=ze[c]=ze[h]=ze[x]=ze[k]=ze[S]=ze[_]=ze[D]=ze[g]=ze[m]=ze[v]=ze[b]=ze[y]=ze[w]=ze[A]=ze[$]=ze["[object Uint8ClampedArray]"]=ze[B]=ze[L]=!0,ze[f]=ze[d]=ze[C]=!1;var je={"\\":"\\","'":"'","\n":"n","\r":"r","\u2028":"u2028","\u2029":"u2029"},He=parseFloat,Ve=parseInt,Ue="object"==typeof e&&e&&e.Object===Object&&e,Ke="object"==typeof self&&self&&self.Object===Object&&self,Ge=Ue||Ke||Function("return this")(),qe=t&&!t.nodeType&&t,Ye=qe&&"object"==typeof i&&i&&!i.nodeType&&i,Qe=Ye&&Ye.exports===qe,Xe=Qe&&Ue.process,Je=function(){try{var e=Ye&&Ye.require&&Ye.require("util").types;return e||Xe&&Xe.binding&&Xe.binding("util")}catch(e){}}(),Ze=Je&&Je.isArrayBuffer,et=Je&&Je.isDate,tt=Je&&Je.isMap,nt=Je&&Je.isRegExp,it=Je&&Je.isSet,rt=Je&&Je.isTypedArray;function ot(e,t,n){switch(n.length){case 0:return e.call(t);case 1:return e.call(t,n[0]);case 2:return e.call(t,n[0],n[1]);case 3:return e.call(t,n[0],n[1],n[2])}return e.apply(t,n)}function st(e,t,n,i){for(var r=-1,o=null==e?0:e.length;++r<o;){var s=e[r];t(i,s,n(s),e)}return i}function at(e,t){for(var n=-1,i=null==e?0:e.length;++n<i&&!1!==t(e[n],n,e););return e}function lt(e,t){for(var n=null==e?0:e.length;n--&&!1!==t(e[n],n,e););return e}function ut(e,t){for(var n=-1,i=null==e?0:e.length;++n<i;)if(!t(e[n],n,e))return!1;return!0}function ct(e,t){for(var n=-1,i=null==e?0:e.length,r=0,o=[];++n<i;){var s=e[n];t(s,n,e)&&(o[r++]=s)}return o}function ht(e,t){return!!(null==e?0:e.length)&&At(e,t,0)>-1}function ft(e,t,n){for(var i=-1,r=null==e?0:e.length;++i<r;)if(n(t,e[i]))return!0;return!1}function dt(e,t){for(var n=-1,i=null==e?0:e.length,r=Array(i);++n<i;)r[n]=t(e[n],n,e);return r}function pt(e,t){for(var n=-1,i=t.length,r=e.length;++n<i;)e[r+n]=t[n];return e}function gt(e,t,n,i){var r=-1,o=null==e?0:e.length;for(i&&o&&(n=e[++r]);++r<o;)n=t(n,e[r],r,e);return n}function mt(e,t,n,i){var r=null==e?0:e.length;for(i&&r&&(n=e[--r]);r--;)n=t(n,e[r],r,e);return n}function vt(e,t){for(var n=-1,i=null==e?0:e.length;++n<i;)if(t(e[n],n,e))return!0;return!1}var bt=xt("length");function yt(e,t,n){var i;return n(e,(function(e,n,r){if(t(e,n,r))return i=n,!1})),i}function wt(e,t,n,i){for(var r=e.length,o=n+(i?1:-1);i?o--:++o<r;)if(t(e[o],o,e))return o;return-1}function At(e,t,n){return t==t?function(e,t,n){var i=n-1,r=e.length;for(;++i<r;)if(e[i]===t)return i;return-1}(e,t,n):wt(e,Et,n)}function Ct(e,t,n,i){for(var r=n-1,o=e.length;++r<o;)if(i(e[r],t))return r;return-1}function Et(e){return e!=e}function Ft(e,t){var n=null==e?0:e.length;return n?_t(e,t)/n:NaN}function xt(e){return function(t){return null==t?void 0:t[e]}}function kt(e){return function(t){return null==e?void 0:e[t]}}function St(e,t,n,i,r){return r(e,(function(e,r,o){n=i?(i=!1,e):t(n,e,r,o)})),n}function _t(e,t){for(var n,i=-1,r=e.length;++i<r;){var o=t(e[i]);void 0!==o&&(n=void 0===n?o:n+o)}return n}function Dt(e,t){for(var n=-1,i=Array(e);++n<e;)i[n]=t(n);return i}function $t(e){return function(t){return e(t)}}function Bt(e,t){return dt(t,(function(t){return e[t]}))}function Lt(e,t){return e.has(t)}function Tt(e,t){for(var n=-1,i=e.length;++n<i&&At(t,e[n],0)>-1;);return n}function Rt(e,t){for(var n=e.length;n--&&At(t,e[n],0)>-1;);return n}function Ot(e,t){for(var n=e.length,i=0;n--;)e[n]===t&&++i;return i}var Mt=kt({"À":"A","Á":"A","Â":"A","Ã":"A","Ä":"A","Å":"A","à":"a","á":"a","â":"a","ã":"a","ä":"a","å":"a","Ç":"C","ç":"c","Ð":"D","ð":"d","È":"E","É":"E","Ê":"E","Ë":"E","è":"e","é":"e","ê":"e","ë":"e","Ì":"I","Í":"I","Î":"I","Ï":"I","ì":"i","í":"i","î":"i","ï":"i","Ñ":"N","ñ":"n","Ò":"O","Ó":"O","Ô":"O","Õ":"O","Ö":"O","Ø":"O","ò":"o","ó":"o","ô":"o","õ":"o","ö":"o","ø":"o","Ù":"U","Ú":"U","Û":"U","Ü":"U","ù":"u","ú":"u","û":"u","ü":"u","Ý":"Y","ý":"y","ÿ":"y","Æ":"Ae","æ":"ae","Þ":"Th","þ":"th","ß":"ss","Ā":"A","Ă":"A","Ą":"A","ā":"a","ă":"a","ą":"a","Ć":"C","Ĉ":"C","Ċ":"C","Č":"C","ć":"c","ĉ":"c","ċ":"c","č":"c","Ď":"D","Đ":"D","ď":"d","đ":"d","Ē":"E","Ĕ":"E","Ė":"E","Ę":"E","Ě":"E","ē":"e","ĕ":"e","ė":"e","ę":"e","ě":"e","Ĝ":"G","Ğ":"G","Ġ":"G","Ģ":"G","ĝ":"g","ğ":"g","ġ":"g","ģ":"g","Ĥ":"H","Ħ":"H","ĥ":"h","ħ":"h","Ĩ":"I","Ī":"I","Ĭ":"I","Į":"I","İ":"I","ĩ":"i","ī":"i","ĭ":"i","į":"i","ı":"i","Ĵ":"J","ĵ":"j","Ķ":"K","ķ":"k","ĸ":"k","Ĺ":"L","Ļ":"L","Ľ":"L","Ŀ":"L","Ł":"L","ĺ":"l","ļ":"l","ľ":"l","ŀ":"l","ł":"l","Ń":"N","Ņ":"N","Ň":"N","Ŋ":"N","ń":"n","ņ":"n","ň":"n","ŋ":"n","Ō":"O","Ŏ":"O","Ő":"O","ō":"o","ŏ":"o","ő":"o","Ŕ":"R","Ŗ":"R","Ř":"R","ŕ":"r","ŗ":"r","ř":"r","Ś":"S","Ŝ":"S","Ş":"S","Š":"S","ś":"s","ŝ":"s","ş":"s","š":"s","Ţ":"T","Ť":"T","Ŧ":"T","ţ":"t","ť":"t","ŧ":"t","Ũ":"U","Ū":"U","Ŭ":"U","Ů":"U","Ű":"U","Ų":"U","ũ":"u","ū":"u","ŭ":"u","ů":"u","ű":"u","ų":"u","Ŵ":"W","ŵ":"w","Ŷ":"Y","ŷ":"y","Ÿ":"Y","Ź":"Z","Ż":"Z","Ž":"Z","ź":"z","ż":"z","ž":"z","Ĳ":"IJ","ĳ":"ij","Œ":"Oe","œ":"oe","ŉ":"'n","ſ":"s"}),Pt=kt({"&":"&amp;","<":"&lt;",">":"&gt;",'"':"&quot;","'":"&#39;"});function It(e){return"\\"+je[e]}function Nt(e){return Me.test(e)}function Wt(e){var t=-1,n=Array(e.size);return e.forEach((function(e,i){n[++t]=[i,e]})),n}function zt(e,t){return function(n){return e(t(n))}}function jt(e,t){for(var n=-1,i=e.length,r=0,o=[];++n<i;){var a=e[n];a!==t&&a!==s||(e[n]=s,o[r++]=n)}return o}function Ht(e){var t=-1,n=Array(e.size);return e.forEach((function(e){n[++t]=e})),n}function Vt(e){var t=-1,n=Array(e.size);return e.forEach((function(e){n[++t]=[e,e]})),n}function Ut(e){return Nt(e)?function(e){var t=Re.lastIndex=0;for(;Re.test(e);)++t;return t}(e):bt(e)}function Kt(e){return Nt(e)?function(e){return e.match(Re)||[]}(e):function(e){return e.split("")}(e)}var Gt=kt({"&amp;":"&","&lt;":"<","&gt;":">","&quot;":'"',"&#39;":"'"});var qt=function e(t){var n,i=(t=null==t?Ge:qt.defaults(Ge.Object(),t,qt.pick(Ge,Ie))).Array,r=t.Date,fe=t.Error,de=t.Function,pe=t.Math,ge=t.Object,me=t.RegExp,ve=t.String,be=t.TypeError,ye=i.prototype,we=de.prototype,Ae=ge.prototype,Ce=t["__core-js_shared__"],Ee=we.toString,Fe=Ae.hasOwnProperty,xe=0,ke=(n=/[^.]+$/.exec(Ce&&Ce.keys&&Ce.keys.IE_PROTO||""))?"Symbol(src)_1."+n:"",Se=Ae.toString,_e=Ee.call(ge),De=Ge._,$e=me("^"+Ee.call(Fe).replace(K,"\\$&").replace(/hasOwnProperty|(function).*?(?=\\\()| for .+?(?=\\\])/g,"$1.*?")+"$"),Be=Qe?t.Buffer:void 0,Re=t.Symbol,Me=t.Uint8Array,je=Be?Be.allocUnsafe:void 0,Ue=zt(ge.getPrototypeOf,ge),Ke=ge.create,qe=Ae.propertyIsEnumerable,Ye=ye.splice,Xe=Re?Re.isConcatSpreadable:void 0,Je=Re?Re.iterator:void 0,bt=Re?Re.toStringTag:void 0,kt=function(){try{var e=Zr(ge,"defineProperty");return e({},"",{}),e}catch(e){}}(),Yt=t.clearTimeout!==Ge.clearTimeout&&t.clearTimeout,Qt=r&&r.now!==Ge.Date.now&&r.now,Xt=t.setTimeout!==Ge.setTimeout&&t.setTimeout,Jt=pe.ceil,Zt=pe.floor,en=ge.getOwnPropertySymbols,tn=Be?Be.isBuffer:void 0,nn=t.isFinite,rn=ye.join,on=zt(ge.keys,ge),sn=pe.max,an=pe.min,ln=r.now,un=t.parseInt,cn=pe.random,hn=ye.reverse,fn=Zr(t,"DataView"),dn=Zr(t,"Map"),pn=Zr(t,"Promise"),gn=Zr(t,"Set"),mn=Zr(t,"WeakMap"),vn=Zr(ge,"create"),bn=mn&&new mn,yn={},wn=So(fn),An=So(dn),Cn=So(pn),En=So(gn),Fn=So(mn),xn=Re?Re.prototype:void 0,kn=xn?xn.valueOf:void 0,Sn=xn?xn.toString:void 0;function _n(e){if(Vs(e)&&!Ts(e)&&!(e instanceof Ln)){if(e instanceof Bn)return e;if(Fe.call(e,"__wrapped__"))return _o(e)}return new Bn(e)}var Dn=function(){function e(){}return function(t){if(!Hs(t))return{};if(Ke)return Ke(t);e.prototype=t;var n=new e;return e.prototype=void 0,n}}();function $n(){}function Bn(e,t){this.__wrapped__=e,this.__actions__=[],this.__chain__=!!t,this.__index__=0,this.__values__=void 0}function Ln(e){this.__wrapped__=e,this.__actions__=[],this.__dir__=1,this.__filtered__=!1,this.__iteratees__=[],this.__takeCount__=4294967295,this.__views__=[]}function Tn(e){var t=-1,n=null==e?0:e.length;for(this.clear();++t<n;){var i=e[t];this.set(i[0],i[1])}}function Rn(e){var t=-1,n=null==e?0:e.length;for(this.clear();++t<n;){var i=e[t];this.set(i[0],i[1])}}function On(e){var t=-1,n=null==e?0:e.length;for(this.clear();++t<n;){var i=e[t];this.set(i[0],i[1])}}function Mn(e){var t=-1,n=null==e?0:e.length;for(this.__data__=new On;++t<n;)this.add(e[t])}function Pn(e){var t=this.__data__=new Rn(e);this.size=t.size}function In(e,t){var n=Ts(e),i=!n&&Ls(e),r=!n&&!i&&Ps(e),o=!n&&!i&&!r&&Js(e),s=n||i||r||o,a=s?Dt(e.length,ve):[],l=a.length;for(var u in e)!t&&!Fe.call(e,u)||s&&("length"==u||r&&("offset"==u||"parent"==u)||o&&("buffer"==u||"byteLength"==u||"byteOffset"==u)||so(u,l))||a.push(u);return a}function Nn(e){var t=e.length;return t?e[Mi(0,t-1)]:void 0}function Wn(e,t){return Fo(vr(e),Yn(t,0,e.length))}function zn(e){return Fo(vr(e))}function jn(e,t,n){(void 0===n||Ds(e[t],n))&&(void 0!==n||t in e)||Gn(e,t,n)}function Hn(e,t,n){var i=e[t];Fe.call(e,t)&&Ds(i,n)&&(void 0!==n||t in e)||Gn(e,t,n)}function Vn(e,t){for(var n=e.length;n--;)if(Ds(e[n][0],t))return n;return-1}function Un(e,t,n,i){return ei(e,(function(e,r,o){t(i,e,n(e),o)})),i}function Kn(e,t){return e&&br(t,wa(t),e)}function Gn(e,t,n){"__proto__"==t&&kt?kt(e,t,{configurable:!0,enumerable:!0,value:n,writable:!0}):e[t]=n}function qn(e,t){for(var n=-1,r=t.length,o=i(r),s=null==e;++n<r;)o[n]=s?void 0:ga(e,t[n]);return o}function Yn(e,t,n){return e==e&&(void 0!==n&&(e=e<=n?e:n),void 0!==t&&(e=e>=t?e:t)),e}function Qn(e,t,n,i,r,o){var s,a=1&t,u=2&t,f=4&t;if(n&&(s=r?n(e,i,r,o):n(e)),void 0!==s)return s;if(!Hs(e))return e;var C=Ts(e);if(C){if(s=function(e){var t=e.length,n=new e.constructor(t);t&&"string"==typeof e[0]&&Fe.call(e,"index")&&(n.index=e.index,n.input=e.input);return n}(e),!a)return vr(e,s)}else{var T=no(e),R=T==d||T==p;if(Ps(e))return hr(e,a);if(T==v||T==l||R&&!r){if(s=u||R?{}:ro(e),!a)return u?function(e,t){return br(e,to(e),t)}(e,function(e,t){return e&&br(t,Aa(t),e)}(s,e)):function(e,t){return br(e,eo(e),t)}(e,Kn(s,e))}else{if(!ze[T])return r?e:{};s=function(e,t,n){var i=e.constructor;switch(t){case E:return fr(e);case c:case h:return new i(+e);case F:return function(e,t){var n=t?fr(e.buffer):e.buffer;return new e.constructor(n,e.byteOffset,e.byteLength)}(e,n);case x:case k:case S:case _:case D:case $:case"[object Uint8ClampedArray]":case B:case L:return dr(e,n);case g:return new i;case m:case w:return new i(e);case b:return function(e){var t=new e.constructor(e.source,ie.exec(e));return t.lastIndex=e.lastIndex,t}(e);case y:return new i;case A:return r=e,kn?ge(kn.call(r)):{}}var r}(e,T,a)}}o||(o=new Pn);var O=o.get(e);if(O)return O;o.set(e,s),Ys(e)?e.forEach((function(i){s.add(Qn(i,t,n,i,e,o))})):Us(e)&&e.forEach((function(i,r){s.set(r,Qn(i,t,n,r,e,o))}));var M=C?void 0:(f?u?Kr:Ur:u?Aa:wa)(e);return at(M||e,(function(i,r){M&&(i=e[r=i]),Hn(s,r,Qn(i,t,n,r,e,o))})),s}function Xn(e,t,n){var i=n.length;if(null==e)return!i;for(e=ge(e);i--;){var r=n[i],o=t[r],s=e[r];if(void 0===s&&!(r in e)||!o(s))return!1}return!0}function Jn(e,t,n){if("function"!=typeof e)throw new be(o);return wo((function(){e.apply(void 0,n)}),t)}function Zn(e,t,n,i){var r=-1,o=ht,s=!0,a=e.length,l=[],u=t.length;if(!a)return l;n&&(t=dt(t,$t(n))),i?(o=ft,s=!1):t.length>=200&&(o=Lt,s=!1,t=new Mn(t));e:for(;++r<a;){var c=e[r],h=null==n?c:n(c);if(c=i||0!==c?c:0,s&&h==h){for(var f=u;f--;)if(t[f]===h)continue e;l.push(c)}else o(t,h,i)||l.push(c)}return l}_n.templateSettings={escape:W,evaluate:z,interpolate:j,variable:"",imports:{_:_n}},_n.prototype=$n.prototype,_n.prototype.constructor=_n,Bn.prototype=Dn($n.prototype),Bn.prototype.constructor=Bn,Ln.prototype=Dn($n.prototype),Ln.prototype.constructor=Ln,Tn.prototype.clear=function(){this.__data__=vn?vn(null):{},this.size=0},Tn.prototype.delete=function(e){var t=this.has(e)&&delete this.__data__[e];return this.size-=t?1:0,t},Tn.prototype.get=function(e){var t=this.__data__;if(vn){var n=t[e];return"__lodash_hash_undefined__"===n?void 0:n}return Fe.call(t,e)?t[e]:void 0},Tn.prototype.has=function(e){var t=this.__data__;return vn?void 0!==t[e]:Fe.call(t,e)},Tn.prototype.set=function(e,t){var n=this.__data__;return this.size+=this.has(e)?0:1,n[e]=vn&&void 0===t?"__lodash_hash_undefined__":t,this},Rn.prototype.clear=function(){this.__data__=[],this.size=0},Rn.prototype.delete=function(e){var t=this.__data__,n=Vn(t,e);return!(n<0)&&(n==t.length-1?t.pop():Ye.call(t,n,1),--this.size,!0)},Rn.prototype.get=function(e){var t=this.__data__,n=Vn(t,e);return n<0?void 0:t[n][1]},Rn.prototype.has=function(e){return Vn(this.__data__,e)>-1},Rn.prototype.set=function(e,t){var n=this.__data__,i=Vn(n,e);return i<0?(++this.size,n.push([e,t])):n[i][1]=t,this},On.prototype.clear=function(){this.size=0,this.__data__={hash:new Tn,map:new(dn||Rn),string:new Tn}},On.prototype.delete=function(e){var t=Xr(this,e).delete(e);return this.size-=t?1:0,t},On.prototype.get=function(e){return Xr(this,e).get(e)},On.prototype.has=function(e){return Xr(this,e).has(e)},On.prototype.set=function(e,t){var n=Xr(this,e),i=n.size;return n.set(e,t),this.size+=n.size==i?0:1,this},Mn.prototype.add=Mn.prototype.push=function(e){return this.__data__.set(e,"__lodash_hash_undefined__"),this},Mn.prototype.has=function(e){return this.__data__.has(e)},Pn.prototype.clear=function(){this.__data__=new Rn,this.size=0},Pn.prototype.delete=function(e){var t=this.__data__,n=t.delete(e);return this.size=t.size,n},Pn.prototype.get=function(e){return this.__data__.get(e)},Pn.prototype.has=function(e){return this.__data__.has(e)},Pn.prototype.set=function(e,t){var n=this.__data__;if(n instanceof Rn){var i=n.__data__;if(!dn||i.length<199)return i.push([e,t]),this.size=++n.size,this;n=this.__data__=new On(i)}return n.set(e,t),this.size=n.size,this};var ei=Ar(li),ti=Ar(ui,!0);function ni(e,t){var n=!0;return ei(e,(function(e,i,r){return n=!!t(e,i,r)})),n}function ii(e,t,n){for(var i=-1,r=e.length;++i<r;){var o=e[i],s=t(o);if(null!=s&&(void 0===a?s==s&&!Xs(s):n(s,a)))var a=s,l=o}return l}function ri(e,t){var n=[];return ei(e,(function(e,i,r){t(e,i,r)&&n.push(e)})),n}function oi(e,t,n,i,r){var o=-1,s=e.length;for(n||(n=oo),r||(r=[]);++o<s;){var a=e[o];t>0&&n(a)?t>1?oi(a,t-1,n,i,r):pt(r,a):i||(r[r.length]=a)}return r}var si=Cr(),ai=Cr(!0);function li(e,t){return e&&si(e,t,wa)}function ui(e,t){return e&&ai(e,t,wa)}function ci(e,t){return ct(t,(function(t){return Ws(e[t])}))}function hi(e,t){for(var n=0,i=(t=ar(t,e)).length;null!=e&&n<i;)e=e[ko(t[n++])];return n&&n==i?e:void 0}function fi(e,t,n){var i=t(e);return Ts(e)?i:pt(i,n(e))}function di(e){return null==e?void 0===e?"[object Undefined]":"[object Null]":bt&&bt in ge(e)?function(e){var t=Fe.call(e,bt),n=e[bt];try{e[bt]=void 0;var i=!0}catch(e){}var r=Se.call(e);i&&(t?e[bt]=n:delete e[bt]);return r}(e):function(e){return Se.call(e)}(e)}function pi(e,t){return e>t}function gi(e,t){return null!=e&&Fe.call(e,t)}function mi(e,t){return null!=e&&t in ge(e)}function vi(e,t,n){for(var r=n?ft:ht,o=e[0].length,s=e.length,a=s,l=i(s),u=1/0,c=[];a--;){var h=e[a];a&&t&&(h=dt(h,$t(t))),u=an(h.length,u),l[a]=!n&&(t||o>=120&&h.length>=120)?new Mn(a&&h):void 0}h=e[0];var f=-1,d=l[0];e:for(;++f<o&&c.length<u;){var p=h[f],g=t?t(p):p;if(p=n||0!==p?p:0,!(d?Lt(d,g):r(c,g,n))){for(a=s;--a;){var m=l[a];if(!(m?Lt(m,g):r(e[a],g,n)))continue e}d&&d.push(g),c.push(p)}}return c}function bi(e,t,n){var i=null==(e=mo(e,t=ar(t,e)))?e:e[ko(No(t))];return null==i?void 0:ot(i,e,n)}function yi(e){return Vs(e)&&di(e)==l}function wi(e,t,n,i,r){return e===t||(null==e||null==t||!Vs(e)&&!Vs(t)?e!=e&&t!=t:function(e,t,n,i,r,o){var s=Ts(e),a=Ts(t),d=s?u:no(e),p=a?u:no(t),C=(d=d==l?v:d)==v,x=(p=p==l?v:p)==v,k=d==p;if(k&&Ps(e)){if(!Ps(t))return!1;s=!0,C=!1}if(k&&!C)return o||(o=new Pn),s||Js(e)?Hr(e,t,n,i,r,o):function(e,t,n,i,r,o,s){switch(n){case F:if(e.byteLength!=t.byteLength||e.byteOffset!=t.byteOffset)return!1;e=e.buffer,t=t.buffer;case E:return!(e.byteLength!=t.byteLength||!o(new Me(e),new Me(t)));case c:case h:case m:return Ds(+e,+t);case f:return e.name==t.name&&e.message==t.message;case b:case w:return e==t+"";case g:var a=Wt;case y:var l=1&i;if(a||(a=Ht),e.size!=t.size&&!l)return!1;var u=s.get(e);if(u)return u==t;i|=2,s.set(e,t);var d=Hr(a(e),a(t),i,r,o,s);return s.delete(e),d;case A:if(kn)return kn.call(e)==kn.call(t)}return!1}(e,t,d,n,i,r,o);if(!(1&n)){var S=C&&Fe.call(e,"__wrapped__"),_=x&&Fe.call(t,"__wrapped__");if(S||_){var D=S?e.value():e,$=_?t.value():t;return o||(o=new Pn),r(D,$,n,i,o)}}if(!k)return!1;return o||(o=new Pn),function(e,t,n,i,r,o){var s=1&n,a=Ur(e),l=a.length,u=Ur(t).length;if(l!=u&&!s)return!1;var c=l;for(;c--;){var h=a[c];if(!(s?h in t:Fe.call(t,h)))return!1}var f=o.get(e);if(f&&o.get(t))return f==t;var d=!0;o.set(e,t),o.set(t,e);var p=s;for(;++c<l;){h=a[c];var g=e[h],m=t[h];if(i)var v=s?i(m,g,h,t,e,o):i(g,m,h,e,t,o);if(!(void 0===v?g===m||r(g,m,n,i,o):v)){d=!1;break}p||(p="constructor"==h)}if(d&&!p){var b=e.constructor,y=t.constructor;b!=y&&"constructor"in e&&"constructor"in t&&!("function"==typeof b&&b instanceof b&&"function"==typeof y&&y instanceof y)&&(d=!1)}return o.delete(e),o.delete(t),d}(e,t,n,i,r,o)}(e,t,n,i,wi,r))}function Ai(e,t,n,i){var r=n.length,o=r,s=!i;if(null==e)return!o;for(e=ge(e);r--;){var a=n[r];if(s&&a[2]?a[1]!==e[a[0]]:!(a[0]in e))return!1}for(;++r<o;){var l=(a=n[r])[0],u=e[l],c=a[1];if(s&&a[2]){if(void 0===u&&!(l in e))return!1}else{var h=new Pn;if(i)var f=i(u,c,l,e,t,h);if(!(void 0===f?wi(c,u,3,i,h):f))return!1}}return!0}function Ci(e){return!(!Hs(e)||(t=e,ke&&ke in t))&&(Ws(e)?$e:se).test(So(e));var t}function Ei(e){return"function"==typeof e?e:null==e?Ka:"object"==typeof e?Ts(e)?Di(e[0],e[1]):_i(e):tl(e)}function Fi(e){if(!ho(e))return on(e);var t=[];for(var n in ge(e))Fe.call(e,n)&&"constructor"!=n&&t.push(n);return t}function xi(e){if(!Hs(e))return function(e){var t=[];if(null!=e)for(var n in ge(e))t.push(n);return t}(e);var t=ho(e),n=[];for(var i in e)("constructor"!=i||!t&&Fe.call(e,i))&&n.push(i);return n}function ki(e,t){return e<t}function Si(e,t){var n=-1,r=Os(e)?i(e.length):[];return ei(e,(function(e,i,o){r[++n]=t(e,i,o)})),r}function _i(e){var t=Jr(e);return 1==t.length&&t[0][2]?po(t[0][0],t[0][1]):function(n){return n===e||Ai(n,e,t)}}function Di(e,t){return lo(e)&&fo(t)?po(ko(e),t):function(n){var i=ga(n,e);return void 0===i&&i===t?ma(n,e):wi(t,i,3)}}function $i(e,t,n,i,r){e!==t&&si(t,(function(o,s){if(r||(r=new Pn),Hs(o))!function(e,t,n,i,r,o,s){var a=bo(e,n),l=bo(t,n),u=s.get(l);if(u)return void jn(e,n,u);var c=o?o(a,l,n+"",e,t,s):void 0,h=void 0===c;if(h){var f=Ts(l),d=!f&&Ps(l),p=!f&&!d&&Js(l);c=l,f||d||p?Ts(a)?c=a:Ms(a)?c=vr(a):d?(h=!1,c=hr(l,!0)):p?(h=!1,c=dr(l,!0)):c=[]:Gs(l)||Ls(l)?(c=a,Ls(a)?c=sa(a):Hs(a)&&!Ws(a)||(c=ro(l))):h=!1}h&&(s.set(l,c),r(c,l,i,o,s),s.delete(l));jn(e,n,c)}(e,t,s,n,$i,i,r);else{var a=i?i(bo(e,s),o,s+"",e,t,r):void 0;void 0===a&&(a=o),jn(e,s,a)}}),Aa)}function Bi(e,t){var n=e.length;if(n)return so(t+=t<0?n:0,n)?e[t]:void 0}function Li(e,t,n){var i=-1;return t=dt(t.length?t:[Ka],$t(Qr())),function(e,t){var n=e.length;for(e.sort(t);n--;)e[n]=e[n].value;return e}(Si(e,(function(e,n,r){return{criteria:dt(t,(function(t){return t(e)})),index:++i,value:e}})),(function(e,t){return function(e,t,n){var i=-1,r=e.criteria,o=t.criteria,s=r.length,a=n.length;for(;++i<s;){var l=pr(r[i],o[i]);if(l){if(i>=a)return l;var u=n[i];return l*("desc"==u?-1:1)}}return e.index-t.index}(e,t,n)}))}function Ti(e,t,n){for(var i=-1,r=t.length,o={};++i<r;){var s=t[i],a=hi(e,s);n(a,s)&&zi(o,ar(s,e),a)}return o}function Ri(e,t,n,i){var r=i?Ct:At,o=-1,s=t.length,a=e;for(e===t&&(t=vr(t)),n&&(a=dt(e,$t(n)));++o<s;)for(var l=0,u=t[o],c=n?n(u):u;(l=r(a,c,l,i))>-1;)a!==e&&Ye.call(a,l,1),Ye.call(e,l,1);return e}function Oi(e,t){for(var n=e?t.length:0,i=n-1;n--;){var r=t[n];if(n==i||r!==o){var o=r;so(r)?Ye.call(e,r,1):Zi(e,r)}}return e}function Mi(e,t){return e+Zt(cn()*(t-e+1))}function Pi(e,t){var n="";if(!e||t<1||t>9007199254740991)return n;do{t%2&&(n+=e),(t=Zt(t/2))&&(e+=e)}while(t);return n}function Ii(e,t){return Ao(go(e,t,Ka),e+"")}function Ni(e){return Nn(Da(e))}function Wi(e,t){var n=Da(e);return Fo(n,Yn(t,0,n.length))}function zi(e,t,n,i){if(!Hs(e))return e;for(var r=-1,o=(t=ar(t,e)).length,s=o-1,a=e;null!=a&&++r<o;){var l=ko(t[r]),u=n;if(r!=s){var c=a[l];void 0===(u=i?i(c,l,a):void 0)&&(u=Hs(c)?c:so(t[r+1])?[]:{})}Hn(a,l,u),a=a[l]}return e}var ji=bn?function(e,t){return bn.set(e,t),e}:Ka,Hi=kt?function(e,t){return kt(e,"toString",{configurable:!0,enumerable:!1,value:Ha(t),writable:!0})}:Ka;function Vi(e){return Fo(Da(e))}function Ui(e,t,n){var r=-1,o=e.length;t<0&&(t=-t>o?0:o+t),(n=n>o?o:n)<0&&(n+=o),o=t>n?0:n-t>>>0,t>>>=0;for(var s=i(o);++r<o;)s[r]=e[r+t];return s}function Ki(e,t){var n;return ei(e,(function(e,i,r){return!(n=t(e,i,r))})),!!n}function Gi(e,t,n){var i=0,r=null==e?i:e.length;if("number"==typeof t&&t==t&&r<=2147483647){for(;i<r;){var o=i+r>>>1,s=e[o];null!==s&&!Xs(s)&&(n?s<=t:s<t)?i=o+1:r=o}return r}return qi(e,t,Ka,n)}function qi(e,t,n,i){t=n(t);for(var r=0,o=null==e?0:e.length,s=t!=t,a=null===t,l=Xs(t),u=void 0===t;r<o;){var c=Zt((r+o)/2),h=n(e[c]),f=void 0!==h,d=null===h,p=h==h,g=Xs(h);if(s)var m=i||p;else m=u?p&&(i||f):a?p&&f&&(i||!d):l?p&&f&&!d&&(i||!g):!d&&!g&&(i?h<=t:h<t);m?r=c+1:o=c}return an(o,4294967294)}function Yi(e,t){for(var n=-1,i=e.length,r=0,o=[];++n<i;){var s=e[n],a=t?t(s):s;if(!n||!Ds(a,l)){var l=a;o[r++]=0===s?0:s}}return o}function Qi(e){return"number"==typeof e?e:Xs(e)?NaN:+e}function Xi(e){if("string"==typeof e)return e;if(Ts(e))return dt(e,Xi)+"";if(Xs(e))return Sn?Sn.call(e):"";var t=e+"";return"0"==t&&1/e==-1/0?"-0":t}function Ji(e,t,n){var i=-1,r=ht,o=e.length,s=!0,a=[],l=a;if(n)s=!1,r=ft;else if(o>=200){var u=t?null:Pr(e);if(u)return Ht(u);s=!1,r=Lt,l=new Mn}else l=t?[]:a;e:for(;++i<o;){var c=e[i],h=t?t(c):c;if(c=n||0!==c?c:0,s&&h==h){for(var f=l.length;f--;)if(l[f]===h)continue e;t&&l.push(h),a.push(c)}else r(l,h,n)||(l!==a&&l.push(h),a.push(c))}return a}function Zi(e,t){return null==(e=mo(e,t=ar(t,e)))||delete e[ko(No(t))]}function er(e,t,n,i){return zi(e,t,n(hi(e,t)),i)}function tr(e,t,n,i){for(var r=e.length,o=i?r:-1;(i?o--:++o<r)&&t(e[o],o,e););return n?Ui(e,i?0:o,i?o+1:r):Ui(e,i?o+1:0,i?r:o)}function nr(e,t){var n=e;return n instanceof Ln&&(n=n.value()),gt(t,(function(e,t){return t.func.apply(t.thisArg,pt([e],t.args))}),n)}function ir(e,t,n){var r=e.length;if(r<2)return r?Ji(e[0]):[];for(var o=-1,s=i(r);++o<r;)for(var a=e[o],l=-1;++l<r;)l!=o&&(s[o]=Zn(s[o]||a,e[l],t,n));return Ji(oi(s,1),t,n)}function rr(e,t,n){for(var i=-1,r=e.length,o=t.length,s={};++i<r;){var a=i<o?t[i]:void 0;n(s,e[i],a)}return s}function or(e){return Ms(e)?e:[]}function sr(e){return"function"==typeof e?e:Ka}function ar(e,t){return Ts(e)?e:lo(e,t)?[e]:xo(aa(e))}var lr=Ii;function ur(e,t,n){var i=e.length;return n=void 0===n?i:n,!t&&n>=i?e:Ui(e,t,n)}var cr=Yt||function(e){return Ge.clearTimeout(e)};function hr(e,t){if(t)return e.slice();var n=e.length,i=je?je(n):new e.constructor(n);return e.copy(i),i}function fr(e){var t=new e.constructor(e.byteLength);return new Me(t).set(new Me(e)),t}function dr(e,t){var n=t?fr(e.buffer):e.buffer;return new e.constructor(n,e.byteOffset,e.length)}function pr(e,t){if(e!==t){var n=void 0!==e,i=null===e,r=e==e,o=Xs(e),s=void 0!==t,a=null===t,l=t==t,u=Xs(t);if(!a&&!u&&!o&&e>t||o&&s&&l&&!a&&!u||i&&s&&l||!n&&l||!r)return 1;if(!i&&!o&&!u&&e<t||u&&n&&r&&!i&&!o||a&&n&&r||!s&&r||!l)return-1}return 0}function gr(e,t,n,r){for(var o=-1,s=e.length,a=n.length,l=-1,u=t.length,c=sn(s-a,0),h=i(u+c),f=!r;++l<u;)h[l]=t[l];for(;++o<a;)(f||o<s)&&(h[n[o]]=e[o]);for(;c--;)h[l++]=e[o++];return h}function mr(e,t,n,r){for(var o=-1,s=e.length,a=-1,l=n.length,u=-1,c=t.length,h=sn(s-l,0),f=i(h+c),d=!r;++o<h;)f[o]=e[o];for(var p=o;++u<c;)f[p+u]=t[u];for(;++a<l;)(d||o<s)&&(f[p+n[a]]=e[o++]);return f}function vr(e,t){var n=-1,r=e.length;for(t||(t=i(r));++n<r;)t[n]=e[n];return t}function br(e,t,n,i){var r=!n;n||(n={});for(var o=-1,s=t.length;++o<s;){var a=t[o],l=i?i(n[a],e[a],a,n,e):void 0;void 0===l&&(l=e[a]),r?Gn(n,a,l):Hn(n,a,l)}return n}function yr(e,t){return function(n,i){var r=Ts(n)?st:Un,o=t?t():{};return r(n,e,Qr(i,2),o)}}function wr(e){return Ii((function(t,n){var i=-1,r=n.length,o=r>1?n[r-1]:void 0,s=r>2?n[2]:void 0;for(o=e.length>3&&"function"==typeof o?(r--,o):void 0,s&&ao(n[0],n[1],s)&&(o=r<3?void 0:o,r=1),t=ge(t);++i<r;){var a=n[i];a&&e(t,a,i,o)}return t}))}function Ar(e,t){return function(n,i){if(null==n)return n;if(!Os(n))return e(n,i);for(var r=n.length,o=t?r:-1,s=ge(n);(t?o--:++o<r)&&!1!==i(s[o],o,s););return n}}function Cr(e){return function(t,n,i){for(var r=-1,o=ge(t),s=i(t),a=s.length;a--;){var l=s[e?a:++r];if(!1===n(o[l],l,o))break}return t}}function Er(e){return function(t){var n=Nt(t=aa(t))?Kt(t):void 0,i=n?n[0]:t.charAt(0),r=n?ur(n,1).join(""):t.slice(1);return i[e]()+r}}function Fr(e){return function(t){return gt(Wa(La(t).replace(Le,"")),e,"")}}function xr(e){return function(){var t=arguments;switch(t.length){case 0:return new e;case 1:return new e(t[0]);case 2:return new e(t[0],t[1]);case 3:return new e(t[0],t[1],t[2]);case 4:return new e(t[0],t[1],t[2],t[3]);case 5:return new e(t[0],t[1],t[2],t[3],t[4]);case 6:return new e(t[0],t[1],t[2],t[3],t[4],t[5]);case 7:return new e(t[0],t[1],t[2],t[3],t[4],t[5],t[6])}var n=Dn(e.prototype),i=e.apply(n,t);return Hs(i)?i:n}}function kr(e){return function(t,n,i){var r=ge(t);if(!Os(t)){var o=Qr(n,3);t=wa(t),n=function(e){return o(r[e],e,r)}}var s=e(t,n,i);return s>-1?r[o?t[s]:s]:void 0}}function Sr(e){return Vr((function(t){var n=t.length,i=n,r=Bn.prototype.thru;for(e&&t.reverse();i--;){var s=t[i];if("function"!=typeof s)throw new be(o);if(r&&!a&&"wrapper"==qr(s))var a=new Bn([],!0)}for(i=a?i:n;++i<n;){var l=qr(s=t[i]),u="wrapper"==l?Gr(s):void 0;a=u&&uo(u[0])&&424==u[1]&&!u[4].length&&1==u[9]?a[qr(u[0])].apply(a,u[3]):1==s.length&&uo(s)?a[l]():a.thru(s)}return function(){var e=arguments,i=e[0];if(a&&1==e.length&&Ts(i))return a.plant(i).value();for(var r=0,o=n?t[r].apply(this,e):i;++r<n;)o=t[r].call(this,o);return o}}))}function _r(e,t,n,r,o,s,a,l,u,c){var h=128&t,f=1&t,d=2&t,p=24&t,g=512&t,m=d?void 0:xr(e);return function v(){for(var b=arguments.length,y=i(b),w=b;w--;)y[w]=arguments[w];if(p)var A=Yr(v),C=Ot(y,A);if(r&&(y=gr(y,r,o,p)),s&&(y=mr(y,s,a,p)),b-=C,p&&b<c){var E=jt(y,A);return Or(e,t,_r,v.placeholder,n,y,E,l,u,c-b)}var F=f?n:this,x=d?F[e]:e;return b=y.length,l?y=vo(y,l):g&&b>1&&y.reverse(),h&&u<b&&(y.length=u),this&&this!==Ge&&this instanceof v&&(x=m||xr(x)),x.apply(F,y)}}function Dr(e,t){return function(n,i){return function(e,t,n,i){return li(e,(function(e,r,o){t(i,n(e),r,o)})),i}(n,e,t(i),{})}}function $r(e,t){return function(n,i){var r;if(void 0===n&&void 0===i)return t;if(void 0!==n&&(r=n),void 0!==i){if(void 0===r)return i;"string"==typeof n||"string"==typeof i?(n=Xi(n),i=Xi(i)):(n=Qi(n),i=Qi(i)),r=e(n,i)}return r}}function Br(e){return Vr((function(t){return t=dt(t,$t(Qr())),Ii((function(n){var i=this;return e(t,(function(e){return ot(e,i,n)}))}))}))}function Lr(e,t){var n=(t=void 0===t?" ":Xi(t)).length;if(n<2)return n?Pi(t,e):t;var i=Pi(t,Jt(e/Ut(t)));return Nt(t)?ur(Kt(i),0,e).join(""):i.slice(0,e)}function Tr(e){return function(t,n,r){return r&&"number"!=typeof r&&ao(t,n,r)&&(n=r=void 0),t=na(t),void 0===n?(n=t,t=0):n=na(n),function(e,t,n,r){for(var o=-1,s=sn(Jt((t-e)/(n||1)),0),a=i(s);s--;)a[r?s:++o]=e,e+=n;return a}(t,n,r=void 0===r?t<n?1:-1:na(r),e)}}function Rr(e){return function(t,n){return"string"==typeof t&&"string"==typeof n||(t=oa(t),n=oa(n)),e(t,n)}}function Or(e,t,n,i,r,o,s,a,l,u){var c=8&t;t|=c?32:64,4&(t&=~(c?64:32))||(t&=-4);var h=[e,t,r,c?o:void 0,c?s:void 0,c?void 0:o,c?void 0:s,a,l,u],f=n.apply(void 0,h);return uo(e)&&yo(f,h),f.placeholder=i,Co(f,e,t)}function Mr(e){var t=pe[e];return function(e,n){if(e=oa(e),(n=null==n?0:an(ia(n),292))&&nn(e)){var i=(aa(e)+"e").split("e");return+((i=(aa(t(i[0]+"e"+(+i[1]+n)))+"e").split("e"))[0]+"e"+(+i[1]-n))}return t(e)}}var Pr=gn&&1/Ht(new gn([,-0]))[1]==1/0?function(e){return new gn(e)}:Xa;function Ir(e){return function(t){var n=no(t);return n==g?Wt(t):n==y?Vt(t):function(e,t){return dt(t,(function(t){return[t,e[t]]}))}(t,e(t))}}function Nr(e,t,n,r,a,l,u,c){var h=2&t;if(!h&&"function"!=typeof e)throw new be(o);var f=r?r.length:0;if(f||(t&=-97,r=a=void 0),u=void 0===u?u:sn(ia(u),0),c=void 0===c?c:ia(c),f-=a?a.length:0,64&t){var d=r,p=a;r=a=void 0}var g=h?void 0:Gr(e),m=[e,t,n,r,a,d,p,l,u,c];if(g&&function(e,t){var n=e[1],i=t[1],r=n|i,o=r<131,a=128==i&&8==n||128==i&&256==n&&e[7].length<=t[8]||384==i&&t[7].length<=t[8]&&8==n;if(!o&&!a)return e;1&i&&(e[2]=t[2],r|=1&n?0:4);var l=t[3];if(l){var u=e[3];e[3]=u?gr(u,l,t[4]):l,e[4]=u?jt(e[3],s):t[4]}(l=t[5])&&(u=e[5],e[5]=u?mr(u,l,t[6]):l,e[6]=u?jt(e[5],s):t[6]);(l=t[7])&&(e[7]=l);128&i&&(e[8]=null==e[8]?t[8]:an(e[8],t[8]));null==e[9]&&(e[9]=t[9]);e[0]=t[0],e[1]=r}(m,g),e=m[0],t=m[1],n=m[2],r=m[3],a=m[4],!(c=m[9]=void 0===m[9]?h?0:e.length:sn(m[9]-f,0))&&24&t&&(t&=-25),t&&1!=t)v=8==t||16==t?function(e,t,n){var r=xr(e);return function o(){for(var s=arguments.length,a=i(s),l=s,u=Yr(o);l--;)a[l]=arguments[l];var c=s<3&&a[0]!==u&&a[s-1]!==u?[]:jt(a,u);if((s-=c.length)<n)return Or(e,t,_r,o.placeholder,void 0,a,c,void 0,void 0,n-s);var h=this&&this!==Ge&&this instanceof o?r:e;return ot(h,this,a)}}(e,t,c):32!=t&&33!=t||a.length?_r.apply(void 0,m):function(e,t,n,r){var o=1&t,s=xr(e);return function t(){for(var a=-1,l=arguments.length,u=-1,c=r.length,h=i(c+l),f=this&&this!==Ge&&this instanceof t?s:e;++u<c;)h[u]=r[u];for(;l--;)h[u++]=arguments[++a];return ot(f,o?n:this,h)}}(e,t,n,r);else var v=function(e,t,n){var i=1&t,r=xr(e);return function t(){var o=this&&this!==Ge&&this instanceof t?r:e;return o.apply(i?n:this,arguments)}}(e,t,n);return Co((g?ji:yo)(v,m),e,t)}function Wr(e,t,n,i){return void 0===e||Ds(e,Ae[n])&&!Fe.call(i,n)?t:e}function zr(e,t,n,i,r,o){return Hs(e)&&Hs(t)&&(o.set(t,e),$i(e,t,void 0,zr,o),o.delete(t)),e}function jr(e){return Gs(e)?void 0:e}function Hr(e,t,n,i,r,o){var s=1&n,a=e.length,l=t.length;if(a!=l&&!(s&&l>a))return!1;var u=o.get(e);if(u&&o.get(t))return u==t;var c=-1,h=!0,f=2&n?new Mn:void 0;for(o.set(e,t),o.set(t,e);++c<a;){var d=e[c],p=t[c];if(i)var g=s?i(p,d,c,t,e,o):i(d,p,c,e,t,o);if(void 0!==g){if(g)continue;h=!1;break}if(f){if(!vt(t,(function(e,t){if(!Lt(f,t)&&(d===e||r(d,e,n,i,o)))return f.push(t)}))){h=!1;break}}else if(d!==p&&!r(d,p,n,i,o)){h=!1;break}}return o.delete(e),o.delete(t),h}function Vr(e){return Ao(go(e,void 0,Ro),e+"")}function Ur(e){return fi(e,wa,eo)}function Kr(e){return fi(e,Aa,to)}var Gr=bn?function(e){return bn.get(e)}:Xa;function qr(e){for(var t=e.name+"",n=yn[t],i=Fe.call(yn,t)?n.length:0;i--;){var r=n[i],o=r.func;if(null==o||o==e)return r.name}return t}function Yr(e){return(Fe.call(_n,"placeholder")?_n:e).placeholder}function Qr(){var e=_n.iteratee||Ga;return e=e===Ga?Ei:e,arguments.length?e(arguments[0],arguments[1]):e}function Xr(e,t){var n,i,r=e.__data__;return("string"==(i=typeof(n=t))||"number"==i||"symbol"==i||"boolean"==i?"__proto__"!==n:null===n)?r["string"==typeof t?"string":"hash"]:r.map}function Jr(e){for(var t=wa(e),n=t.length;n--;){var i=t[n],r=e[i];t[n]=[i,r,fo(r)]}return t}function Zr(e,t){var n=function(e,t){return null==e?void 0:e[t]}(e,t);return Ci(n)?n:void 0}var eo=en?function(e){return null==e?[]:(e=ge(e),ct(en(e),(function(t){return qe.call(e,t)})))}:rl,to=en?function(e){for(var t=[];e;)pt(t,eo(e)),e=Ue(e);return t}:rl,no=di;function io(e,t,n){for(var i=-1,r=(t=ar(t,e)).length,o=!1;++i<r;){var s=ko(t[i]);if(!(o=null!=e&&n(e,s)))break;e=e[s]}return o||++i!=r?o:!!(r=null==e?0:e.length)&&js(r)&&so(s,r)&&(Ts(e)||Ls(e))}function ro(e){return"function"!=typeof e.constructor||ho(e)?{}:Dn(Ue(e))}function oo(e){return Ts(e)||Ls(e)||!!(Xe&&e&&e[Xe])}function so(e,t){var n=typeof e;return!!(t=null==t?9007199254740991:t)&&("number"==n||"symbol"!=n&&le.test(e))&&e>-1&&e%1==0&&e<t}function ao(e,t,n){if(!Hs(n))return!1;var i=typeof t;return!!("number"==i?Os(n)&&so(t,n.length):"string"==i&&t in n)&&Ds(n[t],e)}function lo(e,t){if(Ts(e))return!1;var n=typeof e;return!("number"!=n&&"symbol"!=n&&"boolean"!=n&&null!=e&&!Xs(e))||(V.test(e)||!H.test(e)||null!=t&&e in ge(t))}function uo(e){var t=qr(e),n=_n[t];if("function"!=typeof n||!(t in Ln.prototype))return!1;if(e===n)return!0;var i=Gr(n);return!!i&&e===i[0]}(fn&&no(new fn(new ArrayBuffer(1)))!=F||dn&&no(new dn)!=g||pn&&"[object Promise]"!=no(pn.resolve())||gn&&no(new gn)!=y||mn&&no(new mn)!=C)&&(no=function(e){var t=di(e),n=t==v?e.constructor:void 0,i=n?So(n):"";if(i)switch(i){case wn:return F;case An:return g;case Cn:return"[object Promise]";case En:return y;case Fn:return C}return t});var co=Ce?Ws:ol;function ho(e){var t=e&&e.constructor;return e===("function"==typeof t&&t.prototype||Ae)}function fo(e){return e==e&&!Hs(e)}function po(e,t){return function(n){return null!=n&&(n[e]===t&&(void 0!==t||e in ge(n)))}}function go(e,t,n){return t=sn(void 0===t?e.length-1:t,0),function(){for(var r=arguments,o=-1,s=sn(r.length-t,0),a=i(s);++o<s;)a[o]=r[t+o];o=-1;for(var l=i(t+1);++o<t;)l[o]=r[o];return l[t]=n(a),ot(e,this,l)}}function mo(e,t){return t.length<2?e:hi(e,Ui(t,0,-1))}function vo(e,t){for(var n=e.length,i=an(t.length,n),r=vr(e);i--;){var o=t[i];e[i]=so(o,n)?r[o]:void 0}return e}function bo(e,t){if(("constructor"!==t||"function"!=typeof e[t])&&"__proto__"!=t)return e[t]}var yo=Eo(ji),wo=Xt||function(e,t){return Ge.setTimeout(e,t)},Ao=Eo(Hi);function Co(e,t,n){var i=t+"";return Ao(e,function(e,t){var n=t.length;if(!n)return e;var i=n-1;return t[i]=(n>1?"& ":"")+t[i],t=t.join(n>2?", ":" "),e.replace(X,"{\n/* [wrapped with "+t+"] */\n")}(i,function(e,t){return at(a,(function(n){var i="_."+n[0];t&n[1]&&!ht(e,i)&&e.push(i)})),e.sort()}(function(e){var t=e.match(J);return t?t[1].split(Z):[]}(i),n)))}function Eo(e){var t=0,n=0;return function(){var i=ln(),r=16-(i-n);if(n=i,r>0){if(++t>=800)return arguments[0]}else t=0;return e.apply(void 0,arguments)}}function Fo(e,t){var n=-1,i=e.length,r=i-1;for(t=void 0===t?i:t;++n<t;){var o=Mi(n,r),s=e[o];e[o]=e[n],e[n]=s}return e.length=t,e}var xo=function(e){var t=Es(e,(function(e){return 500===n.size&&n.clear(),e})),n=t.cache;return t}((function(e){var t=[];return 46===e.charCodeAt(0)&&t.push(""),e.replace(U,(function(e,n,i,r){t.push(i?r.replace(te,"$1"):n||e)})),t}));function ko(e){if("string"==typeof e||Xs(e))return e;var t=e+"";return"0"==t&&1/e==-1/0?"-0":t}function So(e){if(null!=e){try{return Ee.call(e)}catch(e){}try{return e+""}catch(e){}}return""}function _o(e){if(e instanceof Ln)return e.clone();var t=new Bn(e.__wrapped__,e.__chain__);return t.__actions__=vr(e.__actions__),t.__index__=e.__index__,t.__values__=e.__values__,t}var Do=Ii((function(e,t){return Ms(e)?Zn(e,oi(t,1,Ms,!0)):[]})),$o=Ii((function(e,t){var n=No(t);return Ms(n)&&(n=void 0),Ms(e)?Zn(e,oi(t,1,Ms,!0),Qr(n,2)):[]})),Bo=Ii((function(e,t){var n=No(t);return Ms(n)&&(n=void 0),Ms(e)?Zn(e,oi(t,1,Ms,!0),void 0,n):[]}));function Lo(e,t,n){var i=null==e?0:e.length;if(!i)return-1;var r=null==n?0:ia(n);return r<0&&(r=sn(i+r,0)),wt(e,Qr(t,3),r)}function To(e,t,n){var i=null==e?0:e.length;if(!i)return-1;var r=i-1;return void 0!==n&&(r=ia(n),r=n<0?sn(i+r,0):an(r,i-1)),wt(e,Qr(t,3),r,!0)}function Ro(e){return(null==e?0:e.length)?oi(e,1):[]}function Oo(e){return e&&e.length?e[0]:void 0}var Mo=Ii((function(e){var t=dt(e,or);return t.length&&t[0]===e[0]?vi(t):[]})),Po=Ii((function(e){var t=No(e),n=dt(e,or);return t===No(n)?t=void 0:n.pop(),n.length&&n[0]===e[0]?vi(n,Qr(t,2)):[]})),Io=Ii((function(e){var t=No(e),n=dt(e,or);return(t="function"==typeof t?t:void 0)&&n.pop(),n.length&&n[0]===e[0]?vi(n,void 0,t):[]}));function No(e){var t=null==e?0:e.length;return t?e[t-1]:void 0}var Wo=Ii(zo);function zo(e,t){return e&&e.length&&t&&t.length?Ri(e,t):e}var jo=Vr((function(e,t){var n=null==e?0:e.length,i=qn(e,t);return Oi(e,dt(t,(function(e){return so(e,n)?+e:e})).sort(pr)),i}));function Ho(e){return null==e?e:hn.call(e)}var Vo=Ii((function(e){return Ji(oi(e,1,Ms,!0))})),Uo=Ii((function(e){var t=No(e);return Ms(t)&&(t=void 0),Ji(oi(e,1,Ms,!0),Qr(t,2))})),Ko=Ii((function(e){var t=No(e);return t="function"==typeof t?t:void 0,Ji(oi(e,1,Ms,!0),void 0,t)}));function Go(e){if(!e||!e.length)return[];var t=0;return e=ct(e,(function(e){if(Ms(e))return t=sn(e.length,t),!0})),Dt(t,(function(t){return dt(e,xt(t))}))}function qo(e,t){if(!e||!e.length)return[];var n=Go(e);return null==t?n:dt(n,(function(e){return ot(t,void 0,e)}))}var Yo=Ii((function(e,t){return Ms(e)?Zn(e,t):[]})),Qo=Ii((function(e){return ir(ct(e,Ms))})),Xo=Ii((function(e){var t=No(e);return Ms(t)&&(t=void 0),ir(ct(e,Ms),Qr(t,2))})),Jo=Ii((function(e){var t=No(e);return t="function"==typeof t?t:void 0,ir(ct(e,Ms),void 0,t)})),Zo=Ii(Go);var es=Ii((function(e){var t=e.length,n=t>1?e[t-1]:void 0;return n="function"==typeof n?(e.pop(),n):void 0,qo(e,n)}));function ts(e){var t=_n(e);return t.__chain__=!0,t}function ns(e,t){return t(e)}var is=Vr((function(e){var t=e.length,n=t?e[0]:0,i=this.__wrapped__,r=function(t){return qn(t,e)};return!(t>1||this.__actions__.length)&&i instanceof Ln&&so(n)?((i=i.slice(n,+n+(t?1:0))).__actions__.push({func:ns,args:[r],thisArg:void 0}),new Bn(i,this.__chain__).thru((function(e){return t&&!e.length&&e.push(void 0),e}))):this.thru(r)}));var rs=yr((function(e,t,n){Fe.call(e,n)?++e[n]:Gn(e,n,1)}));var os=kr(Lo),ss=kr(To);function as(e,t){return(Ts(e)?at:ei)(e,Qr(t,3))}function ls(e,t){return(Ts(e)?lt:ti)(e,Qr(t,3))}var us=yr((function(e,t,n){Fe.call(e,n)?e[n].push(t):Gn(e,n,[t])}));var cs=Ii((function(e,t,n){var r=-1,o="function"==typeof t,s=Os(e)?i(e.length):[];return ei(e,(function(e){s[++r]=o?ot(t,e,n):bi(e,t,n)})),s})),hs=yr((function(e,t,n){Gn(e,n,t)}));function fs(e,t){return(Ts(e)?dt:Si)(e,Qr(t,3))}var ds=yr((function(e,t,n){e[n?0:1].push(t)}),(function(){return[[],[]]}));var ps=Ii((function(e,t){if(null==e)return[];var n=t.length;return n>1&&ao(e,t[0],t[1])?t=[]:n>2&&ao(t[0],t[1],t[2])&&(t=[t[0]]),Li(e,oi(t,1),[])})),gs=Qt||function(){return Ge.Date.now()};function ms(e,t,n){return t=n?void 0:t,Nr(e,128,void 0,void 0,void 0,void 0,t=e&&null==t?e.length:t)}function vs(e,t){var n;if("function"!=typeof t)throw new be(o);return e=ia(e),function(){return--e>0&&(n=t.apply(this,arguments)),e<=1&&(t=void 0),n}}var bs=Ii((function(e,t,n){var i=1;if(n.length){var r=jt(n,Yr(bs));i|=32}return Nr(e,i,t,n,r)})),ys=Ii((function(e,t,n){var i=3;if(n.length){var r=jt(n,Yr(ys));i|=32}return Nr(t,i,e,n,r)}));function ws(e,t,n){var i,r,s,a,l,u,c=0,h=!1,f=!1,d=!0;if("function"!=typeof e)throw new be(o);function p(t){var n=i,o=r;return i=r=void 0,c=t,a=e.apply(o,n)}function g(e){return c=e,l=wo(v,t),h?p(e):a}function m(e){var n=e-u;return void 0===u||n>=t||n<0||f&&e-c>=s}function v(){var e=gs();if(m(e))return b(e);l=wo(v,function(e){var n=t-(e-u);return f?an(n,s-(e-c)):n}(e))}function b(e){return l=void 0,d&&i?p(e):(i=r=void 0,a)}function y(){var e=gs(),n=m(e);if(i=arguments,r=this,u=e,n){if(void 0===l)return g(u);if(f)return cr(l),l=wo(v,t),p(u)}return void 0===l&&(l=wo(v,t)),a}return t=oa(t)||0,Hs(n)&&(h=!!n.leading,s=(f="maxWait"in n)?sn(oa(n.maxWait)||0,t):s,d="trailing"in n?!!n.trailing:d),y.cancel=function(){void 0!==l&&cr(l),c=0,i=u=r=l=void 0},y.flush=function(){return void 0===l?a:b(gs())},y}var As=Ii((function(e,t){return Jn(e,1,t)})),Cs=Ii((function(e,t,n){return Jn(e,oa(t)||0,n)}));function Es(e,t){if("function"!=typeof e||null!=t&&"function"!=typeof t)throw new be(o);var n=function(){var i=arguments,r=t?t.apply(this,i):i[0],o=n.cache;if(o.has(r))return o.get(r);var s=e.apply(this,i);return n.cache=o.set(r,s)||o,s};return n.cache=new(Es.Cache||On),n}function Fs(e){if("function"!=typeof e)throw new be(o);return function(){var t=arguments;switch(t.length){case 0:return!e.call(this);case 1:return!e.call(this,t[0]);case 2:return!e.call(this,t[0],t[1]);case 3:return!e.call(this,t[0],t[1],t[2])}return!e.apply(this,t)}}Es.Cache=On;var xs=lr((function(e,t){var n=(t=1==t.length&&Ts(t[0])?dt(t[0],$t(Qr())):dt(oi(t,1),$t(Qr()))).length;return Ii((function(i){for(var r=-1,o=an(i.length,n);++r<o;)i[r]=t[r].call(this,i[r]);return ot(e,this,i)}))})),ks=Ii((function(e,t){return Nr(e,32,void 0,t,jt(t,Yr(ks)))})),Ss=Ii((function(e,t){return Nr(e,64,void 0,t,jt(t,Yr(Ss)))})),_s=Vr((function(e,t){return Nr(e,256,void 0,void 0,void 0,t)}));function Ds(e,t){return e===t||e!=e&&t!=t}var $s=Rr(pi),Bs=Rr((function(e,t){return e>=t})),Ls=yi(function(){return arguments}())?yi:function(e){return Vs(e)&&Fe.call(e,"callee")&&!qe.call(e,"callee")},Ts=i.isArray,Rs=Ze?$t(Ze):function(e){return Vs(e)&&di(e)==E};function Os(e){return null!=e&&js(e.length)&&!Ws(e)}function Ms(e){return Vs(e)&&Os(e)}var Ps=tn||ol,Is=et?$t(et):function(e){return Vs(e)&&di(e)==h};function Ns(e){if(!Vs(e))return!1;var t=di(e);return t==f||"[object DOMException]"==t||"string"==typeof e.message&&"string"==typeof e.name&&!Gs(e)}function Ws(e){if(!Hs(e))return!1;var t=di(e);return t==d||t==p||"[object AsyncFunction]"==t||"[object Proxy]"==t}function zs(e){return"number"==typeof e&&e==ia(e)}function js(e){return"number"==typeof e&&e>-1&&e%1==0&&e<=9007199254740991}function Hs(e){var t=typeof e;return null!=e&&("object"==t||"function"==t)}function Vs(e){return null!=e&&"object"==typeof e}var Us=tt?$t(tt):function(e){return Vs(e)&&no(e)==g};function Ks(e){return"number"==typeof e||Vs(e)&&di(e)==m}function Gs(e){if(!Vs(e)||di(e)!=v)return!1;var t=Ue(e);if(null===t)return!0;var n=Fe.call(t,"constructor")&&t.constructor;return"function"==typeof n&&n instanceof n&&Ee.call(n)==_e}var qs=nt?$t(nt):function(e){return Vs(e)&&di(e)==b};var Ys=it?$t(it):function(e){return Vs(e)&&no(e)==y};function Qs(e){return"string"==typeof e||!Ts(e)&&Vs(e)&&di(e)==w}function Xs(e){return"symbol"==typeof e||Vs(e)&&di(e)==A}var Js=rt?$t(rt):function(e){return Vs(e)&&js(e.length)&&!!We[di(e)]};var Zs=Rr(ki),ea=Rr((function(e,t){return e<=t}));function ta(e){if(!e)return[];if(Os(e))return Qs(e)?Kt(e):vr(e);if(Je&&e[Je])return function(e){for(var t,n=[];!(t=e.next()).done;)n.push(t.value);return n}(e[Je]());var t=no(e);return(t==g?Wt:t==y?Ht:Da)(e)}function na(e){return e?(e=oa(e))===1/0||e===-1/0?17976931348623157e292*(e<0?-1:1):e==e?e:0:0===e?e:0}function ia(e){var t=na(e),n=t%1;return t==t?n?t-n:t:0}function ra(e){return e?Yn(ia(e),0,4294967295):0}function oa(e){if("number"==typeof e)return e;if(Xs(e))return NaN;if(Hs(e)){var t="function"==typeof e.valueOf?e.valueOf():e;e=Hs(t)?t+"":t}if("string"!=typeof e)return 0===e?e:+e;e=e.replace(q,"");var n=oe.test(e);return n||ae.test(e)?Ve(e.slice(2),n?2:8):re.test(e)?NaN:+e}function sa(e){return br(e,Aa(e))}function aa(e){return null==e?"":Xi(e)}var la=wr((function(e,t){if(ho(t)||Os(t))br(t,wa(t),e);else for(var n in t)Fe.call(t,n)&&Hn(e,n,t[n])})),ua=wr((function(e,t){br(t,Aa(t),e)})),ca=wr((function(e,t,n,i){br(t,Aa(t),e,i)})),ha=wr((function(e,t,n,i){br(t,wa(t),e,i)})),fa=Vr(qn);var da=Ii((function(e,t){e=ge(e);var n=-1,i=t.length,r=i>2?t[2]:void 0;for(r&&ao(t[0],t[1],r)&&(i=1);++n<i;)for(var o=t[n],s=Aa(o),a=-1,l=s.length;++a<l;){var u=s[a],c=e[u];(void 0===c||Ds(c,Ae[u])&&!Fe.call(e,u))&&(e[u]=o[u])}return e})),pa=Ii((function(e){return e.push(void 0,zr),ot(Ea,void 0,e)}));function ga(e,t,n){var i=null==e?void 0:hi(e,t);return void 0===i?n:i}function ma(e,t){return null!=e&&io(e,t,mi)}var va=Dr((function(e,t,n){null!=t&&"function"!=typeof t.toString&&(t=Se.call(t)),e[t]=n}),Ha(Ka)),ba=Dr((function(e,t,n){null!=t&&"function"!=typeof t.toString&&(t=Se.call(t)),Fe.call(e,t)?e[t].push(n):e[t]=[n]}),Qr),ya=Ii(bi);function wa(e){return Os(e)?In(e):Fi(e)}function Aa(e){return Os(e)?In(e,!0):xi(e)}var Ca=wr((function(e,t,n){$i(e,t,n)})),Ea=wr((function(e,t,n,i){$i(e,t,n,i)})),Fa=Vr((function(e,t){var n={};if(null==e)return n;var i=!1;t=dt(t,(function(t){return t=ar(t,e),i||(i=t.length>1),t})),br(e,Kr(e),n),i&&(n=Qn(n,7,jr));for(var r=t.length;r--;)Zi(n,t[r]);return n}));var xa=Vr((function(e,t){return null==e?{}:function(e,t){return Ti(e,t,(function(t,n){return ma(e,n)}))}(e,t)}));function ka(e,t){if(null==e)return{};var n=dt(Kr(e),(function(e){return[e]}));return t=Qr(t),Ti(e,n,(function(e,n){return t(e,n[0])}))}var Sa=Ir(wa),_a=Ir(Aa);function Da(e){return null==e?[]:Bt(e,wa(e))}var $a=Fr((function(e,t,n){return t=t.toLowerCase(),e+(n?Ba(t):t)}));function Ba(e){return Na(aa(e).toLowerCase())}function La(e){return(e=aa(e))&&e.replace(ue,Mt).replace(Te,"")}var Ta=Fr((function(e,t,n){return e+(n?"-":"")+t.toLowerCase()})),Ra=Fr((function(e,t,n){return e+(n?" ":"")+t.toLowerCase()})),Oa=Er("toLowerCase");var Ma=Fr((function(e,t,n){return e+(n?"_":"")+t.toLowerCase()}));var Pa=Fr((function(e,t,n){return e+(n?" ":"")+Na(t)}));var Ia=Fr((function(e,t,n){return e+(n?" ":"")+t.toUpperCase()})),Na=Er("toUpperCase");function Wa(e,t,n){return e=aa(e),void 0===(t=n?void 0:t)?function(e){return Pe.test(e)}(e)?function(e){return e.match(Oe)||[]}(e):function(e){return e.match(ee)||[]}(e):e.match(t)||[]}var za=Ii((function(e,t){try{return ot(e,void 0,t)}catch(e){return Ns(e)?e:new fe(e)}})),ja=Vr((function(e,t){return at(t,(function(t){t=ko(t),Gn(e,t,bs(e[t],e))})),e}));function Ha(e){return function(){return e}}var Va=Sr(),Ua=Sr(!0);function Ka(e){return e}function Ga(e){return Ei("function"==typeof e?e:Qn(e,1))}var qa=Ii((function(e,t){return function(n){return bi(n,e,t)}})),Ya=Ii((function(e,t){return function(n){return bi(e,n,t)}}));function Qa(e,t,n){var i=wa(t),r=ci(t,i);null!=n||Hs(t)&&(r.length||!i.length)||(n=t,t=e,e=this,r=ci(t,wa(t)));var o=!(Hs(n)&&"chain"in n&&!n.chain),s=Ws(e);return at(r,(function(n){var i=t[n];e[n]=i,s&&(e.prototype[n]=function(){var t=this.__chain__;if(o||t){var n=e(this.__wrapped__),r=n.__actions__=vr(this.__actions__);return r.push({func:i,args:arguments,thisArg:e}),n.__chain__=t,n}return i.apply(e,pt([this.value()],arguments))})})),e}function Xa(){}var Ja=Br(dt),Za=Br(ut),el=Br(vt);function tl(e){return lo(e)?xt(ko(e)):function(e){return function(t){return hi(t,e)}}(e)}var nl=Tr(),il=Tr(!0);function rl(){return[]}function ol(){return!1}var sl=$r((function(e,t){return e+t}),0),al=Mr("ceil"),ll=$r((function(e,t){return e/t}),1),ul=Mr("floor");var cl,hl=$r((function(e,t){return e*t}),1),fl=Mr("round"),dl=$r((function(e,t){return e-t}),0);return _n.after=function(e,t){if("function"!=typeof t)throw new be(o);return e=ia(e),function(){if(--e<1)return t.apply(this,arguments)}},_n.ary=ms,_n.assign=la,_n.assignIn=ua,_n.assignInWith=ca,_n.assignWith=ha,_n.at=fa,_n.before=vs,_n.bind=bs,_n.bindAll=ja,_n.bindKey=ys,_n.castArray=function(){if(!arguments.length)return[];var e=arguments[0];return Ts(e)?e:[e]},_n.chain=ts,_n.chunk=function(e,t,n){t=(n?ao(e,t,n):void 0===t)?1:sn(ia(t),0);var r=null==e?0:e.length;if(!r||t<1)return[];for(var o=0,s=0,a=i(Jt(r/t));o<r;)a[s++]=Ui(e,o,o+=t);return a},_n.compact=function(e){for(var t=-1,n=null==e?0:e.length,i=0,r=[];++t<n;){var o=e[t];o&&(r[i++]=o)}return r},_n.concat=function(){var e=arguments.length;if(!e)return[];for(var t=i(e-1),n=arguments[0],r=e;r--;)t[r-1]=arguments[r];return pt(Ts(n)?vr(n):[n],oi(t,1))},_n.cond=function(e){var t=null==e?0:e.length,n=Qr();return e=t?dt(e,(function(e){if("function"!=typeof e[1])throw new be(o);return[n(e[0]),e[1]]})):[],Ii((function(n){for(var i=-1;++i<t;){var r=e[i];if(ot(r[0],this,n))return ot(r[1],this,n)}}))},_n.conforms=function(e){return function(e){var t=wa(e);return function(n){return Xn(n,e,t)}}(Qn(e,1))},_n.constant=Ha,_n.countBy=rs,_n.create=function(e,t){var n=Dn(e);return null==t?n:Kn(n,t)},_n.curry=function e(t,n,i){var r=Nr(t,8,void 0,void 0,void 0,void 0,void 0,n=i?void 0:n);return r.placeholder=e.placeholder,r},_n.curryRight=function e(t,n,i){var r=Nr(t,16,void 0,void 0,void 0,void 0,void 0,n=i?void 0:n);return r.placeholder=e.placeholder,r},_n.debounce=ws,_n.defaults=da,_n.defaultsDeep=pa,_n.defer=As,_n.delay=Cs,_n.difference=Do,_n.differenceBy=$o,_n.differenceWith=Bo,_n.drop=function(e,t,n){var i=null==e?0:e.length;return i?Ui(e,(t=n||void 0===t?1:ia(t))<0?0:t,i):[]},_n.dropRight=function(e,t,n){var i=null==e?0:e.length;return i?Ui(e,0,(t=i-(t=n||void 0===t?1:ia(t)))<0?0:t):[]},_n.dropRightWhile=function(e,t){return e&&e.length?tr(e,Qr(t,3),!0,!0):[]},_n.dropWhile=function(e,t){return e&&e.length?tr(e,Qr(t,3),!0):[]},_n.fill=function(e,t,n,i){var r=null==e?0:e.length;return r?(n&&"number"!=typeof n&&ao(e,t,n)&&(n=0,i=r),function(e,t,n,i){var r=e.length;for((n=ia(n))<0&&(n=-n>r?0:r+n),(i=void 0===i||i>r?r:ia(i))<0&&(i+=r),i=n>i?0:ra(i);n<i;)e[n++]=t;return e}(e,t,n,i)):[]},_n.filter=function(e,t){return(Ts(e)?ct:ri)(e,Qr(t,3))},_n.flatMap=function(e,t){return oi(fs(e,t),1)},_n.flatMapDeep=function(e,t){return oi(fs(e,t),1/0)},_n.flatMapDepth=function(e,t,n){return n=void 0===n?1:ia(n),oi(fs(e,t),n)},_n.flatten=Ro,_n.flattenDeep=function(e){return(null==e?0:e.length)?oi(e,1/0):[]},_n.flattenDepth=function(e,t){return(null==e?0:e.length)?oi(e,t=void 0===t?1:ia(t)):[]},_n.flip=function(e){return Nr(e,512)},_n.flow=Va,_n.flowRight=Ua,_n.fromPairs=function(e){for(var t=-1,n=null==e?0:e.length,i={};++t<n;){var r=e[t];i[r[0]]=r[1]}return i},_n.functions=function(e){return null==e?[]:ci(e,wa(e))},_n.functionsIn=function(e){return null==e?[]:ci(e,Aa(e))},_n.groupBy=us,_n.initial=function(e){return(null==e?0:e.length)?Ui(e,0,-1):[]},_n.intersection=Mo,_n.intersectionBy=Po,_n.intersectionWith=Io,_n.invert=va,_n.invertBy=ba,_n.invokeMap=cs,_n.iteratee=Ga,_n.keyBy=hs,_n.keys=wa,_n.keysIn=Aa,_n.map=fs,_n.mapKeys=function(e,t){var n={};return t=Qr(t,3),li(e,(function(e,i,r){Gn(n,t(e,i,r),e)})),n},_n.mapValues=function(e,t){var n={};return t=Qr(t,3),li(e,(function(e,i,r){Gn(n,i,t(e,i,r))})),n},_n.matches=function(e){return _i(Qn(e,1))},_n.matchesProperty=function(e,t){return Di(e,Qn(t,1))},_n.memoize=Es,_n.merge=Ca,_n.mergeWith=Ea,_n.method=qa,_n.methodOf=Ya,_n.mixin=Qa,_n.negate=Fs,_n.nthArg=function(e){return e=ia(e),Ii((function(t){return Bi(t,e)}))},_n.omit=Fa,_n.omitBy=function(e,t){return ka(e,Fs(Qr(t)))},_n.once=function(e){return vs(2,e)},_n.orderBy=function(e,t,n,i){return null==e?[]:(Ts(t)||(t=null==t?[]:[t]),Ts(n=i?void 0:n)||(n=null==n?[]:[n]),Li(e,t,n))},_n.over=Ja,_n.overArgs=xs,_n.overEvery=Za,_n.overSome=el,_n.partial=ks,_n.partialRight=Ss,_n.partition=ds,_n.pick=xa,_n.pickBy=ka,_n.property=tl,_n.propertyOf=function(e){return function(t){return null==e?void 0:hi(e,t)}},_n.pull=Wo,_n.pullAll=zo,_n.pullAllBy=function(e,t,n){return e&&e.length&&t&&t.length?Ri(e,t,Qr(n,2)):e},_n.pullAllWith=function(e,t,n){return e&&e.length&&t&&t.length?Ri(e,t,void 0,n):e},_n.pullAt=jo,_n.range=nl,_n.rangeRight=il,_n.rearg=_s,_n.reject=function(e,t){return(Ts(e)?ct:ri)(e,Fs(Qr(t,3)))},_n.remove=function(e,t){var n=[];if(!e||!e.length)return n;var i=-1,r=[],o=e.length;for(t=Qr(t,3);++i<o;){var s=e[i];t(s,i,e)&&(n.push(s),r.push(i))}return Oi(e,r),n},_n.rest=function(e,t){if("function"!=typeof e)throw new be(o);return Ii(e,t=void 0===t?t:ia(t))},_n.reverse=Ho,_n.sampleSize=function(e,t,n){return t=(n?ao(e,t,n):void 0===t)?1:ia(t),(Ts(e)?Wn:Wi)(e,t)},_n.set=function(e,t,n){return null==e?e:zi(e,t,n)},_n.setWith=function(e,t,n,i){return i="function"==typeof i?i:void 0,null==e?e:zi(e,t,n,i)},_n.shuffle=function(e){return(Ts(e)?zn:Vi)(e)},_n.slice=function(e,t,n){var i=null==e?0:e.length;return i?(n&&"number"!=typeof n&&ao(e,t,n)?(t=0,n=i):(t=null==t?0:ia(t),n=void 0===n?i:ia(n)),Ui(e,t,n)):[]},_n.sortBy=ps,_n.sortedUniq=function(e){return e&&e.length?Yi(e):[]},_n.sortedUniqBy=function(e,t){return e&&e.length?Yi(e,Qr(t,2)):[]},_n.split=function(e,t,n){return n&&"number"!=typeof n&&ao(e,t,n)&&(t=n=void 0),(n=void 0===n?4294967295:n>>>0)?(e=aa(e))&&("string"==typeof t||null!=t&&!qs(t))&&!(t=Xi(t))&&Nt(e)?ur(Kt(e),0,n):e.split(t,n):[]},_n.spread=function(e,t){if("function"!=typeof e)throw new be(o);return t=null==t?0:sn(ia(t),0),Ii((function(n){var i=n[t],r=ur(n,0,t);return i&&pt(r,i),ot(e,this,r)}))},_n.tail=function(e){var t=null==e?0:e.length;return t?Ui(e,1,t):[]},_n.take=function(e,t,n){return e&&e.length?Ui(e,0,(t=n||void 0===t?1:ia(t))<0?0:t):[]},_n.takeRight=function(e,t,n){var i=null==e?0:e.length;return i?Ui(e,(t=i-(t=n||void 0===t?1:ia(t)))<0?0:t,i):[]},_n.takeRightWhile=function(e,t){return e&&e.length?tr(e,Qr(t,3),!1,!0):[]},_n.takeWhile=function(e,t){return e&&e.length?tr(e,Qr(t,3)):[]},_n.tap=function(e,t){return t(e),e},_n.throttle=function(e,t,n){var i=!0,r=!0;if("function"!=typeof e)throw new be(o);return Hs(n)&&(i="leading"in n?!!n.leading:i,r="trailing"in n?!!n.trailing:r),ws(e,t,{leading:i,maxWait:t,trailing:r})},_n.thru=ns,_n.toArray=ta,_n.toPairs=Sa,_n.toPairsIn=_a,_n.toPath=function(e){return Ts(e)?dt(e,ko):Xs(e)?[e]:vr(xo(aa(e)))},_n.toPlainObject=sa,_n.transform=function(e,t,n){var i=Ts(e),r=i||Ps(e)||Js(e);if(t=Qr(t,4),null==n){var o=e&&e.constructor;n=r?i?new o:[]:Hs(e)&&Ws(o)?Dn(Ue(e)):{}}return(r?at:li)(e,(function(e,i,r){return t(n,e,i,r)})),n},_n.unary=function(e){return ms(e,1)},_n.union=Vo,_n.unionBy=Uo,_n.unionWith=Ko,_n.uniq=function(e){return e&&e.length?Ji(e):[]},_n.uniqBy=function(e,t){return e&&e.length?Ji(e,Qr(t,2)):[]},_n.uniqWith=function(e,t){return t="function"==typeof t?t:void 0,e&&e.length?Ji(e,void 0,t):[]},_n.unset=function(e,t){return null==e||Zi(e,t)},_n.unzip=Go,_n.unzipWith=qo,_n.update=function(e,t,n){return null==e?e:er(e,t,sr(n))},_n.updateWith=function(e,t,n,i){return i="function"==typeof i?i:void 0,null==e?e:er(e,t,sr(n),i)},_n.values=Da,_n.valuesIn=function(e){return null==e?[]:Bt(e,Aa(e))},_n.without=Yo,_n.words=Wa,_n.wrap=function(e,t){return ks(sr(t),e)},_n.xor=Qo,_n.xorBy=Xo,_n.xorWith=Jo,_n.zip=Zo,_n.zipObject=function(e,t){return rr(e||[],t||[],Hn)},_n.zipObjectDeep=function(e,t){return rr(e||[],t||[],zi)},_n.zipWith=es,_n.entries=Sa,_n.entriesIn=_a,_n.extend=ua,_n.extendWith=ca,Qa(_n,_n),_n.add=sl,_n.attempt=za,_n.camelCase=$a,_n.capitalize=Ba,_n.ceil=al,_n.clamp=function(e,t,n){return void 0===n&&(n=t,t=void 0),void 0!==n&&(n=(n=oa(n))==n?n:0),void 0!==t&&(t=(t=oa(t))==t?t:0),Yn(oa(e),t,n)},_n.clone=function(e){return Qn(e,4)},_n.cloneDeep=function(e){return Qn(e,5)},_n.cloneDeepWith=function(e,t){return Qn(e,5,t="function"==typeof t?t:void 0)},_n.cloneWith=function(e,t){return Qn(e,4,t="function"==typeof t?t:void 0)},_n.conformsTo=function(e,t){return null==t||Xn(e,t,wa(t))},_n.deburr=La,_n.defaultTo=function(e,t){return null==e||e!=e?t:e},_n.divide=ll,_n.endsWith=function(e,t,n){e=aa(e),t=Xi(t);var i=e.length,r=n=void 0===n?i:Yn(ia(n),0,i);return(n-=t.length)>=0&&e.slice(n,r)==t},_n.eq=Ds,_n.escape=function(e){return(e=aa(e))&&N.test(e)?e.replace(P,Pt):e},_n.escapeRegExp=function(e){return(e=aa(e))&&G.test(e)?e.replace(K,"\\$&"):e},_n.every=function(e,t,n){var i=Ts(e)?ut:ni;return n&&ao(e,t,n)&&(t=void 0),i(e,Qr(t,3))},_n.find=os,_n.findIndex=Lo,_n.findKey=function(e,t){return yt(e,Qr(t,3),li)},_n.findLast=ss,_n.findLastIndex=To,_n.findLastKey=function(e,t){return yt(e,Qr(t,3),ui)},_n.floor=ul,_n.forEach=as,_n.forEachRight=ls,_n.forIn=function(e,t){return null==e?e:si(e,Qr(t,3),Aa)},_n.forInRight=function(e,t){return null==e?e:ai(e,Qr(t,3),Aa)},_n.forOwn=function(e,t){return e&&li(e,Qr(t,3))},_n.forOwnRight=function(e,t){return e&&ui(e,Qr(t,3))},_n.get=ga,_n.gt=$s,_n.gte=Bs,_n.has=function(e,t){return null!=e&&io(e,t,gi)},_n.hasIn=ma,_n.head=Oo,_n.identity=Ka,_n.includes=function(e,t,n,i){e=Os(e)?e:Da(e),n=n&&!i?ia(n):0;var r=e.length;return n<0&&(n=sn(r+n,0)),Qs(e)?n<=r&&e.indexOf(t,n)>-1:!!r&&At(e,t,n)>-1},_n.indexOf=function(e,t,n){var i=null==e?0:e.length;if(!i)return-1;var r=null==n?0:ia(n);return r<0&&(r=sn(i+r,0)),At(e,t,r)},_n.inRange=function(e,t,n){return t=na(t),void 0===n?(n=t,t=0):n=na(n),function(e,t,n){return e>=an(t,n)&&e<sn(t,n)}(e=oa(e),t,n)},_n.invoke=ya,_n.isArguments=Ls,_n.isArray=Ts,_n.isArrayBuffer=Rs,_n.isArrayLike=Os,_n.isArrayLikeObject=Ms,_n.isBoolean=function(e){return!0===e||!1===e||Vs(e)&&di(e)==c},_n.isBuffer=Ps,_n.isDate=Is,_n.isElement=function(e){return Vs(e)&&1===e.nodeType&&!Gs(e)},_n.isEmpty=function(e){if(null==e)return!0;if(Os(e)&&(Ts(e)||"string"==typeof e||"function"==typeof e.splice||Ps(e)||Js(e)||Ls(e)))return!e.length;var t=no(e);if(t==g||t==y)return!e.size;if(ho(e))return!Fi(e).length;for(var n in e)if(Fe.call(e,n))return!1;return!0},_n.isEqual=function(e,t){return wi(e,t)},_n.isEqualWith=function(e,t,n){var i=(n="function"==typeof n?n:void 0)?n(e,t):void 0;return void 0===i?wi(e,t,void 0,n):!!i},_n.isError=Ns,_n.isFinite=function(e){return"number"==typeof e&&nn(e)},_n.isFunction=Ws,_n.isInteger=zs,_n.isLength=js,_n.isMap=Us,_n.isMatch=function(e,t){return e===t||Ai(e,t,Jr(t))},_n.isMatchWith=function(e,t,n){return n="function"==typeof n?n:void 0,Ai(e,t,Jr(t),n)},_n.isNaN=function(e){return Ks(e)&&e!=+e},_n.isNative=function(e){if(co(e))throw new fe("Unsupported core-js use. Try https://npms.io/search?q=ponyfill.");return Ci(e)},_n.isNil=function(e){return null==e},_n.isNull=function(e){return null===e},_n.isNumber=Ks,_n.isObject=Hs,_n.isObjectLike=Vs,_n.isPlainObject=Gs,_n.isRegExp=qs,_n.isSafeInteger=function(e){return zs(e)&&e>=-9007199254740991&&e<=9007199254740991},_n.isSet=Ys,_n.isString=Qs,_n.isSymbol=Xs,_n.isTypedArray=Js,_n.isUndefined=function(e){return void 0===e},_n.isWeakMap=function(e){return Vs(e)&&no(e)==C},_n.isWeakSet=function(e){return Vs(e)&&"[object WeakSet]"==di(e)},_n.join=function(e,t){return null==e?"":rn.call(e,t)},_n.kebabCase=Ta,_n.last=No,_n.lastIndexOf=function(e,t,n){var i=null==e?0:e.length;if(!i)return-1;var r=i;return void 0!==n&&(r=(r=ia(n))<0?sn(i+r,0):an(r,i-1)),t==t?function(e,t,n){for(var i=n+1;i--;)if(e[i]===t)return i;return i}(e,t,r):wt(e,Et,r,!0)},_n.lowerCase=Ra,_n.lowerFirst=Oa,_n.lt=Zs,_n.lte=ea,_n.max=function(e){return e&&e.length?ii(e,Ka,pi):void 0},_n.maxBy=function(e,t){return e&&e.length?ii(e,Qr(t,2),pi):void 0},_n.mean=function(e){return Ft(e,Ka)},_n.meanBy=function(e,t){return Ft(e,Qr(t,2))},_n.min=function(e){return e&&e.length?ii(e,Ka,ki):void 0},_n.minBy=function(e,t){return e&&e.length?ii(e,Qr(t,2),ki):void 0},_n.stubArray=rl,_n.stubFalse=ol,_n.stubObject=function(){return{}},_n.stubString=function(){return""},_n.stubTrue=function(){return!0},_n.multiply=hl,_n.nth=function(e,t){return e&&e.length?Bi(e,ia(t)):void 0},_n.noConflict=function(){return Ge._===this&&(Ge._=De),this},_n.noop=Xa,_n.now=gs,_n.pad=function(e,t,n){e=aa(e);var i=(t=ia(t))?Ut(e):0;if(!t||i>=t)return e;var r=(t-i)/2;return Lr(Zt(r),n)+e+Lr(Jt(r),n)},_n.padEnd=function(e,t,n){e=aa(e);var i=(t=ia(t))?Ut(e):0;return t&&i<t?e+Lr(t-i,n):e},_n.padStart=function(e,t,n){e=aa(e);var i=(t=ia(t))?Ut(e):0;return t&&i<t?Lr(t-i,n)+e:e},_n.parseInt=function(e,t,n){return n||null==t?t=0:t&&(t=+t),un(aa(e).replace(Y,""),t||0)},_n.random=function(e,t,n){if(n&&"boolean"!=typeof n&&ao(e,t,n)&&(t=n=void 0),void 0===n&&("boolean"==typeof t?(n=t,t=void 0):"boolean"==typeof e&&(n=e,e=void 0)),void 0===e&&void 0===t?(e=0,t=1):(e=na(e),void 0===t?(t=e,e=0):t=na(t)),e>t){var i=e;e=t,t=i}if(n||e%1||t%1){var r=cn();return an(e+r*(t-e+He("1e-"+((r+"").length-1))),t)}return Mi(e,t)},_n.reduce=function(e,t,n){var i=Ts(e)?gt:St,r=arguments.length<3;return i(e,Qr(t,4),n,r,ei)},_n.reduceRight=function(e,t,n){var i=Ts(e)?mt:St,r=arguments.length<3;return i(e,Qr(t,4),n,r,ti)},_n.repeat=function(e,t,n){return t=(n?ao(e,t,n):void 0===t)?1:ia(t),Pi(aa(e),t)},_n.replace=function(){var e=arguments,t=aa(e[0]);return e.length<3?t:t.replace(e[1],e[2])},_n.result=function(e,t,n){var i=-1,r=(t=ar(t,e)).length;for(r||(r=1,e=void 0);++i<r;){var o=null==e?void 0:e[ko(t[i])];void 0===o&&(i=r,o=n),e=Ws(o)?o.call(e):o}return e},_n.round=fl,_n.runInContext=e,_n.sample=function(e){return(Ts(e)?Nn:Ni)(e)},_n.size=function(e){if(null==e)return 0;if(Os(e))return Qs(e)?Ut(e):e.length;var t=no(e);return t==g||t==y?e.size:Fi(e).length},_n.snakeCase=Ma,_n.some=function(e,t,n){var i=Ts(e)?vt:Ki;return n&&ao(e,t,n)&&(t=void 0),i(e,Qr(t,3))},_n.sortedIndex=function(e,t){return Gi(e,t)},_n.sortedIndexBy=function(e,t,n){return qi(e,t,Qr(n,2))},_n.sortedIndexOf=function(e,t){var n=null==e?0:e.length;if(n){var i=Gi(e,t);if(i<n&&Ds(e[i],t))return i}return-1},_n.sortedLastIndex=function(e,t){return Gi(e,t,!0)},_n.sortedLastIndexBy=function(e,t,n){return qi(e,t,Qr(n,2),!0)},_n.sortedLastIndexOf=function(e,t){if(null==e?0:e.length){var n=Gi(e,t,!0)-1;if(Ds(e[n],t))return n}return-1},_n.startCase=Pa,_n.startsWith=function(e,t,n){return e=aa(e),n=null==n?0:Yn(ia(n),0,e.length),t=Xi(t),e.slice(n,n+t.length)==t},_n.subtract=dl,_n.sum=function(e){return e&&e.length?_t(e,Ka):0},_n.sumBy=function(e,t){return e&&e.length?_t(e,Qr(t,2)):0},_n.template=function(e,t,n){var i=_n.templateSettings;n&&ao(e,t,n)&&(t=void 0),e=aa(e),t=ca({},t,i,Wr);var r,o,s=ca({},t.imports,i.imports,Wr),a=wa(s),l=Bt(s,a),u=0,c=t.interpolate||ce,h="__p += '",f=me((t.escape||ce).source+"|"+c.source+"|"+(c===j?ne:ce).source+"|"+(t.evaluate||ce).source+"|$","g"),d="//# sourceURL="+(Fe.call(t,"sourceURL")?(t.sourceURL+"").replace(/[\r\n]/g," "):"lodash.templateSources["+ ++Ne+"]")+"\n";e.replace(f,(function(t,n,i,s,a,l){return i||(i=s),h+=e.slice(u,l).replace(he,It),n&&(r=!0,h+="' +\n__e("+n+") +\n'"),a&&(o=!0,h+="';\n"+a+";\n__p += '"),i&&(h+="' +\n((__t = ("+i+")) == null ? '' : __t) +\n'"),u=l+t.length,t})),h+="';\n";var p=Fe.call(t,"variable")&&t.variable;p||(h="with (obj) {\n"+h+"\n}\n"),h=(o?h.replace(T,""):h).replace(R,"$1").replace(O,"$1;"),h="function("+(p||"obj")+") {\n"+(p?"":"obj || (obj = {});\n")+"var __t, __p = ''"+(r?", __e = _.escape":"")+(o?", __j = Array.prototype.join;\nfunction print() { __p += __j.call(arguments, '') }\n":";\n")+h+"return __p\n}";var g=za((function(){return de(a,d+"return "+h).apply(void 0,l)}));if(g.source=h,Ns(g))throw g;return g},_n.times=function(e,t){if((e=ia(e))<1||e>9007199254740991)return[];var n=4294967295,i=an(e,4294967295);e-=4294967295;for(var r=Dt(i,t=Qr(t));++n<e;)t(n);return r},_n.toFinite=na,_n.toInteger=ia,_n.toLength=ra,_n.toLower=function(e){return aa(e).toLowerCase()},_n.toNumber=oa,_n.toSafeInteger=function(e){return e?Yn(ia(e),-9007199254740991,9007199254740991):0===e?e:0},_n.toString=aa,_n.toUpper=function(e){return aa(e).toUpperCase()},_n.trim=function(e,t,n){if((e=aa(e))&&(n||void 0===t))return e.replace(q,"");if(!e||!(t=Xi(t)))return e;var i=Kt(e),r=Kt(t);return ur(i,Tt(i,r),Rt(i,r)+1).join("")},_n.trimEnd=function(e,t,n){if((e=aa(e))&&(n||void 0===t))return e.replace(Q,"");if(!e||!(t=Xi(t)))return e;var i=Kt(e);return ur(i,0,Rt(i,Kt(t))+1).join("")},_n.trimStart=function(e,t,n){if((e=aa(e))&&(n||void 0===t))return e.replace(Y,"");if(!e||!(t=Xi(t)))return e;var i=Kt(e);return ur(i,Tt(i,Kt(t))).join("")},_n.truncate=function(e,t){var n=30,i="...";if(Hs(t)){var r="separator"in t?t.separator:r;n="length"in t?ia(t.length):n,i="omission"in t?Xi(t.omission):i}var o=(e=aa(e)).length;if(Nt(e)){var s=Kt(e);o=s.length}if(n>=o)return e;var a=n-Ut(i);if(a<1)return i;var l=s?ur(s,0,a).join(""):e.slice(0,a);if(void 0===r)return l+i;if(s&&(a+=l.length-a),qs(r)){if(e.slice(a).search(r)){var u,c=l;for(r.global||(r=me(r.source,aa(ie.exec(r))+"g")),r.lastIndex=0;u=r.exec(c);)var h=u.index;l=l.slice(0,void 0===h?a:h)}}else if(e.indexOf(Xi(r),a)!=a){var f=l.lastIndexOf(r);f>-1&&(l=l.slice(0,f))}return l+i},_n.unescape=function(e){return(e=aa(e))&&I.test(e)?e.replace(M,Gt):e},_n.uniqueId=function(e){var t=++xe;return aa(e)+t},_n.upperCase=Ia,_n.upperFirst=Na,_n.each=as,_n.eachRight=ls,_n.first=Oo,Qa(_n,(cl={},li(_n,(function(e,t){Fe.call(_n.prototype,t)||(cl[t]=e)})),cl),{chain:!1}),_n.VERSION="4.17.15",at(["bind","bindKey","curry","curryRight","partial","partialRight"],(function(e){_n[e].placeholder=_n})),at(["drop","take"],(function(e,t){Ln.prototype[e]=function(n){n=void 0===n?1:sn(ia(n),0);var i=this.__filtered__&&!t?new Ln(this):this.clone();return i.__filtered__?i.__takeCount__=an(n,i.__takeCount__):i.__views__.push({size:an(n,4294967295),type:e+(i.__dir__<0?"Right":"")}),i},Ln.prototype[e+"Right"]=function(t){return this.reverse()[e](t).reverse()}})),at(["filter","map","takeWhile"],(function(e,t){var n=t+1,i=1==n||3==n;Ln.prototype[e]=function(e){var t=this.clone();return t.__iteratees__.push({iteratee:Qr(e,3),type:n}),t.__filtered__=t.__filtered__||i,t}})),at(["head","last"],(function(e,t){var n="take"+(t?"Right":"");Ln.prototype[e]=function(){return this[n](1).value()[0]}})),at(["initial","tail"],(function(e,t){var n="drop"+(t?"":"Right");Ln.prototype[e]=function(){return this.__filtered__?new Ln(this):this[n](1)}})),Ln.prototype.compact=function(){return this.filter(Ka)},Ln.prototype.find=function(e){return this.filter(e).head()},Ln.prototype.findLast=function(e){return this.reverse().find(e)},Ln.prototype.invokeMap=Ii((function(e,t){return"function"==typeof e?new Ln(this):this.map((function(n){return bi(n,e,t)}))})),Ln.prototype.reject=function(e){return this.filter(Fs(Qr(e)))},Ln.prototype.slice=function(e,t){e=ia(e);var n=this;return n.__filtered__&&(e>0||t<0)?new Ln(n):(e<0?n=n.takeRight(-e):e&&(n=n.drop(e)),void 0!==t&&(n=(t=ia(t))<0?n.dropRight(-t):n.take(t-e)),n)},Ln.prototype.takeRightWhile=function(e){return this.reverse().takeWhile(e).reverse()},Ln.prototype.toArray=function(){return this.take(4294967295)},li(Ln.prototype,(function(e,t){var n=/^(?:filter|find|map|reject)|While$/.test(t),i=/^(?:head|last)$/.test(t),r=_n[i?"take"+("last"==t?"Right":""):t],o=i||/^find/.test(t);r&&(_n.prototype[t]=function(){var t=this.__wrapped__,s=i?[1]:arguments,a=t instanceof Ln,l=s[0],u=a||Ts(t),c=function(e){var t=r.apply(_n,pt([e],s));return i&&h?t[0]:t};u&&n&&"function"==typeof l&&1!=l.length&&(a=u=!1);var h=this.__chain__,f=!!this.__actions__.length,d=o&&!h,p=a&&!f;if(!o&&u){t=p?t:new Ln(this);var g=e.apply(t,s);return g.__actions__.push({func:ns,args:[c],thisArg:void 0}),new Bn(g,h)}return d&&p?e.apply(this,s):(g=this.thru(c),d?i?g.value()[0]:g.value():g)})})),at(["pop","push","shift","sort","splice","unshift"],(function(e){var t=ye[e],n=/^(?:push|sort|unshift)$/.test(e)?"tap":"thru",i=/^(?:pop|shift)$/.test(e);_n.prototype[e]=function(){var e=arguments;if(i&&!this.__chain__){var r=this.value();return t.apply(Ts(r)?r:[],e)}return this[n]((function(n){return t.apply(Ts(n)?n:[],e)}))}})),li(Ln.prototype,(function(e,t){var n=_n[t];if(n){var i=n.name+"";Fe.call(yn,i)||(yn[i]=[]),yn[i].push({name:t,func:n})}})),yn[_r(void 0,2).name]=[{name:"wrapper",func:void 0}],Ln.prototype.clone=function(){var e=new Ln(this.__wrapped__);return e.__actions__=vr(this.__actions__),e.__dir__=this.__dir__,e.__filtered__=this.__filtered__,e.__iteratees__=vr(this.__iteratees__),e.__takeCount__=this.__takeCount__,e.__views__=vr(this.__views__),e},Ln.prototype.reverse=function(){if(this.__filtered__){var e=new Ln(this);e.__dir__=-1,e.__filtered__=!0}else(e=this.clone()).__dir__*=-1;return e},Ln.prototype.value=function(){var e=this.__wrapped__.value(),t=this.__dir__,n=Ts(e),i=t<0,r=n?e.length:0,o=function(e,t,n){var i=-1,r=n.length;for(;++i<r;){var o=n[i],s=o.size;switch(o.type){case"drop":e+=s;break;case"dropRight":t-=s;break;case"take":t=an(t,e+s);break;case"takeRight":e=sn(e,t-s)}}return{start:e,end:t}}(0,r,this.__views__),s=o.start,a=o.end,l=a-s,u=i?a:s-1,c=this.__iteratees__,h=c.length,f=0,d=an(l,this.__takeCount__);if(!n||!i&&r==l&&d==l)return nr(e,this.__actions__);var p=[];e:for(;l--&&f<d;){for(var g=-1,m=e[u+=t];++g<h;){var v=c[g],b=v.iteratee,y=v.type,w=b(m);if(2==y)m=w;else if(!w){if(1==y)continue e;break e}}p[f++]=m}return p},_n.prototype.at=is,_n.prototype.chain=function(){return ts(this)},_n.prototype.commit=function(){return new Bn(this.value(),this.__chain__)},_n.prototype.next=function(){void 0===this.__values__&&(this.__values__=ta(this.value()));var e=this.__index__>=this.__values__.length;return{done:e,value:e?void 0:this.__values__[this.__index__++]}},_n.prototype.plant=function(e){for(var t,n=this;n instanceof $n;){var i=_o(n);i.__index__=0,i.__values__=void 0,t?r.__wrapped__=i:t=i;var r=i;n=n.__wrapped__}return r.__wrapped__=e,t},_n.prototype.reverse=function(){var e=this.__wrapped__;if(e instanceof Ln){var t=e;return this.__actions__.length&&(t=new Ln(this)),(t=t.reverse()).__actions__.push({func:ns,args:[Ho],thisArg:void 0}),new Bn(t,this.__chain__)}return this.thru(Ho)},_n.prototype.toJSON=_n.prototype.valueOf=_n.prototype.value=function(){return nr(this.__wrapped__,this.__actions__)},_n.prototype.first=_n.prototype.head,Je&&(_n.prototype[Je]=function(){return this}),_n}();Ge._=qt,void 0===(r=function(){return qt}.call(t,n,t,i))||(i.exports=r)}).call(this)}).call(this,n(26),n(25)(e))},function(e,t){e.exports=function(e,t){e.prototype=Object.create(t.prototype),e.prototype.constructor=e,e.__proto__=t}},function(e,t,n){var i=function(e){"use strict";var t=Object.prototype,n=t.hasOwnProperty,i="function"==typeof Symbol?Symbol:{},r=i.iterator||"@@iterator",o=i.asyncIterator||"@@asyncIterator",s=i.toStringTag||"@@toStringTag";function a(e,t,n,i){var r=t&&t.prototype instanceof c?t:c,o=Object.create(r.prototype),s=new C(i||[]);return o._invoke=function(e,t,n){var i="suspendedStart";return function(r,o){if("executing"===i)throw new Error("Generator is already running");if("completed"===i){if("throw"===r)throw o;return F()}for(n.method=r,n.arg=o;;){var s=n.delegate;if(s){var a=y(s,n);if(a){if(a===u)continue;return a}}if("next"===n.method)n.sent=n._sent=n.arg;else if("throw"===n.method){if("suspendedStart"===i)throw i="completed",n.arg;n.dispatchException(n.arg)}else"return"===n.method&&n.abrupt("return",n.arg);i="executing";var c=l(e,t,n);if("normal"===c.type){if(i=n.done?"completed":"suspendedYield",c.arg===u)continue;return{value:c.arg,done:n.done}}"throw"===c.type&&(i="completed",n.method="throw",n.arg=c.arg)}}}(e,n,s),o}function l(e,t,n){try{return{type:"normal",arg:e.call(t,n)}}catch(e){return{type:"throw",arg:e}}}e.wrap=a;var u={};function c(){}function h(){}function f(){}var d={};d[r]=function(){return this};var p=Object.getPrototypeOf,g=p&&p(p(E([])));g&&g!==t&&n.call(g,r)&&(d=g);var m=f.prototype=c.prototype=Object.create(d);function v(e){["next","throw","return"].forEach((function(t){e[t]=function(e){return this._invoke(t,e)}}))}function b(e){var t;this._invoke=function(i,r){function o(){return new Promise((function(t,o){!function t(i,r,o,s){var a=l(e[i],e,r);if("throw"!==a.type){var u=a.arg,c=u.value;return c&&"object"==typeof c&&n.call(c,"__await")?Promise.resolve(c.__await).then((function(e){t("next",e,o,s)}),(function(e){t("throw",e,o,s)})):Promise.resolve(c).then((function(e){u.value=e,o(u)}),(function(e){return t("throw",e,o,s)}))}s(a.arg)}(i,r,t,o)}))}return t=t?t.then(o,o):o()}}function y(e,t){var n=e.iterator[t.method];if(void 0===n){if(t.delegate=null,"throw"===t.method){if(e.iterator.return&&(t.method="return",t.arg=void 0,y(e,t),"throw"===t.method))return u;t.method="throw",t.arg=new TypeError("The iterator does not provide a 'throw' method")}return u}var i=l(n,e.iterator,t.arg);if("throw"===i.type)return t.method="throw",t.arg=i.arg,t.delegate=null,u;var r=i.arg;return r?r.done?(t[e.resultName]=r.value,t.next=e.nextLoc,"return"!==t.method&&(t.method="next",t.arg=void 0),t.delegate=null,u):r:(t.method="throw",t.arg=new TypeError("iterator result is not an object"),t.delegate=null,u)}function w(e){var t={tryLoc:e[0]};1 in e&&(t.catchLoc=e[1]),2 in e&&(t.finallyLoc=e[2],t.afterLoc=e[3]),this.tryEntries.push(t)}function A(e){var t=e.completion||{};t.type="normal",delete t.arg,e.completion=t}function C(e){this.tryEntries=[{tryLoc:"root"}],e.forEach(w,this),this.reset(!0)}function E(e){if(e){var t=e[r];if(t)return t.call(e);if("function"==typeof e.next)return e;if(!isNaN(e.length)){var i=-1,o=function t(){for(;++i<e.length;)if(n.call(e,i))return t.value=e[i],t.done=!1,t;return t.value=void 0,t.done=!0,t};return o.next=o}}return{next:F}}function F(){return{value:void 0,done:!0}}return h.prototype=m.constructor=f,f.constructor=h,f[s]=h.displayName="GeneratorFunction",e.isGeneratorFunction=function(e){var t="function"==typeof e&&e.constructor;return!!t&&(t===h||"GeneratorFunction"===(t.displayName||t.name))},e.mark=function(e){return Object.setPrototypeOf?Object.setPrototypeOf(e,f):(e.__proto__=f,s in e||(e[s]="GeneratorFunction")),e.prototype=Object.create(m),e},e.awrap=function(e){return{__await:e}},v(b.prototype),b.prototype[o]=function(){return this},e.AsyncIterator=b,e.async=function(t,n,i,r){var o=new b(a(t,n,i,r));return e.isGeneratorFunction(n)?o:o.next().then((function(e){return e.done?e.value:o.next()}))},v(m),m[s]="Generator",m[r]=function(){return this},m.toString=function(){return"[object Generator]"},e.keys=function(e){var t=[];for(var n in e)t.push(n);return t.reverse(),function n(){for(;t.length;){var i=t.pop();if(i in e)return n.value=i,n.done=!1,n}return n.done=!0,n}},e.values=E,C.prototype={constructor:C,reset:function(e){if(this.prev=0,this.next=0,this.sent=this._sent=void 0,this.done=!1,this.delegate=null,this.method="next",this.arg=void 0,this.tryEntries.forEach(A),!e)for(var t in this)"t"===t.charAt(0)&&n.call(this,t)&&!isNaN(+t.slice(1))&&(this[t]=void 0)},stop:function(){this.done=!0;var e=this.tryEntries[0].completion;if("throw"===e.type)throw e.arg;return this.rval},dispatchException:function(e){if(this.done)throw e;var t=this;function i(n,i){return s.type="throw",s.arg=e,t.next=n,i&&(t.method="next",t.arg=void 0),!!i}for(var r=this.tryEntries.length-1;r>=0;--r){var o=this.tryEntries[r],s=o.completion;if("root"===o.tryLoc)return i("end");if(o.tryLoc<=this.prev){var a=n.call(o,"catchLoc"),l=n.call(o,"finallyLoc");if(a&&l){if(this.prev<o.catchLoc)return i(o.catchLoc,!0);if(this.prev<o.finallyLoc)return i(o.finallyLoc)}else if(a){if(this.prev<o.catchLoc)return i(o.catchLoc,!0)}else{if(!l)throw new Error("try statement without catch or finally");if(this.prev<o.finallyLoc)return i(o.finallyLoc)}}}},abrupt:function(e,t){for(var i=this.tryEntries.length-1;i>=0;--i){var r=this.tryEntries[i];if(r.tryLoc<=this.prev&&n.call(r,"finallyLoc")&&this.prev<r.finallyLoc){var o=r;break}}o&&("break"===e||"continue"===e)&&o.tryLoc<=t&&t<=o.finallyLoc&&(o=null);var s=o?o.completion:{};return s.type=e,s.arg=t,o?(this.method="next",this.next=o.finallyLoc,u):this.complete(s)},complete:function(e,t){if("throw"===e.type)throw e.arg;return"break"===e.type||"continue"===e.type?this.next=e.arg:"return"===e.type?(this.rval=this.arg=e.arg,this.method="return",this.next="end"):"normal"===e.type&&t&&(this.next=t),u},finish:function(e){for(var t=this.tryEntries.length-1;t>=0;--t){var n=this.tryEntries[t];if(n.finallyLoc===e)return this.complete(n.completion,n.afterLoc),A(n),u}},catch:function(e){for(var t=this.tryEntries.length-1;t>=0;--t){var n=this.tryEntries[t];if(n.tryLoc===e){var i=n.completion;if("throw"===i.type){var r=i.arg;A(n)}return r}}throw new Error("illegal catch attempt")},delegateYield:function(e,t,n){return this.delegate={iterator:E(e),resultName:t,nextLoc:n},"next"===this.method&&(this.arg=void 0),u}},e}(e.exports);try{regeneratorRuntime=i}catch(e){Function("r","regeneratorRuntime = r")(i)}},function(e,t){function n(t){return"function"==typeof Symbol&&"symbol"==typeof Symbol.iterator?e.exports=n=function(e){return typeof e}:e.exports=n=function(e){return e&&"function"==typeof Symbol&&e.constructor===Symbol&&e!==Symbol.prototype?"symbol":typeof e},n(t)}e.exports=n},function(e,t){function n(t,i){return e.exports=n=Object.setPrototypeOf||function(e,t){return e.__proto__=t,e},n(t,i)}e.exports=n},function(e,t,n){"use strict";
/** @license React v16.12.0
 * react.production.min.js
 *
 * Copyright (c) Facebook, Inc. and its affiliates.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 */var i=n(37),r="function"==typeof Symbol&&Symbol.for,o=r?Symbol.for("react.element"):60103,s=r?Symbol.for("react.portal"):60106,a=r?Symbol.for("react.fragment"):60107,l=r?Symbol.for("react.strict_mode"):60108,u=r?Symbol.for("react.profiler"):60114,c=r?Symbol.for("react.provider"):60109,h=r?Symbol.for("react.context"):60110,f=r?Symbol.for("react.forward_ref"):60112,d=r?Symbol.for("react.suspense"):60113;r&&Symbol.for("react.suspense_list");var p=r?Symbol.for("react.memo"):60115,g=r?Symbol.for("react.lazy"):60116;r&&Symbol.for("react.fundamental"),r&&Symbol.for("react.responder"),r&&Symbol.for("react.scope");var m="function"==typeof Symbol&&Symbol.iterator;function v(e){for(var t="https://reactjs.org/docs/error-decoder.html?invariant="+e,n=1;n<arguments.length;n++)t+="&args[]="+encodeURIComponent(arguments[n]);return"Minified React error #"+e+"; visit "+t+" for the full message or use the non-minified dev environment for full errors and additional helpful warnings."}var b={isMounted:function(){return!1},enqueueForceUpdate:function(){},enqueueReplaceState:function(){},enqueueSetState:function(){}},y={};function w(e,t,n){this.props=e,this.context=t,this.refs=y,this.updater=n||b}function A(){}function C(e,t,n){this.props=e,this.context=t,this.refs=y,this.updater=n||b}w.prototype.isReactComponent={},w.prototype.setState=function(e,t){if("object"!=typeof e&&"function"!=typeof e&&null!=e)throw Error(v(85));this.updater.enqueueSetState(this,e,t,"setState")},w.prototype.forceUpdate=function(e){this.updater.enqueueForceUpdate(this,e,"forceUpdate")},A.prototype=w.prototype;var E=C.prototype=new A;E.constructor=C,i(E,w.prototype),E.isPureReactComponent=!0;var F={current:null},x={current:null},k=Object.prototype.hasOwnProperty,S={key:!0,ref:!0,__self:!0,__source:!0};function _(e,t,n){var i,r={},s=null,a=null;if(null!=t)for(i in void 0!==t.ref&&(a=t.ref),void 0!==t.key&&(s=""+t.key),t)k.call(t,i)&&!S.hasOwnProperty(i)&&(r[i]=t[i]);var l=arguments.length-2;if(1===l)r.children=n;else if(1<l){for(var u=Array(l),c=0;c<l;c++)u[c]=arguments[c+2];r.children=u}if(e&&e.defaultProps)for(i in l=e.defaultProps)void 0===r[i]&&(r[i]=l[i]);return{$$typeof:o,type:e,key:s,ref:a,props:r,_owner:x.current}}function D(e){return"object"==typeof e&&null!==e&&e.$$typeof===o}var $=/\/+/g,B=[];function L(e,t,n,i){if(B.length){var r=B.pop();return r.result=e,r.keyPrefix=t,r.func=n,r.context=i,r.count=0,r}return{result:e,keyPrefix:t,func:n,context:i,count:0}}function T(e){e.result=null,e.keyPrefix=null,e.func=null,e.context=null,e.count=0,10>B.length&&B.push(e)}function R(e,t,n){return null==e?0:function e(t,n,i,r){var a=typeof t;"undefined"!==a&&"boolean"!==a||(t=null);var l=!1;if(null===t)l=!0;else switch(a){case"string":case"number":l=!0;break;case"object":switch(t.$$typeof){case o:case s:l=!0}}if(l)return i(r,t,""===n?"."+O(t,0):n),1;if(l=0,n=""===n?".":n+":",Array.isArray(t))for(var u=0;u<t.length;u++){var c=n+O(a=t[u],u);l+=e(a,c,i,r)}else if(null===t||"object"!=typeof t?c=null:c="function"==typeof(c=m&&t[m]||t["@@iterator"])?c:null,"function"==typeof c)for(t=c.call(t),u=0;!(a=t.next()).done;)l+=e(a=a.value,c=n+O(a,u++),i,r);else if("object"===a)throw i=""+t,Error(v(31,"[object Object]"===i?"object with keys {"+Object.keys(t).join(", ")+"}":i,""));return l}(e,"",t,n)}function O(e,t){return"object"==typeof e&&null!==e&&null!=e.key?function(e){var t={"=":"=0",":":"=2"};return"$"+(""+e).replace(/[=:]/g,(function(e){return t[e]}))}(e.key):t.toString(36)}function M(e,t){e.func.call(e.context,t,e.count++)}function P(e,t,n){var i=e.result,r=e.keyPrefix;e=e.func.call(e.context,t,e.count++),Array.isArray(e)?I(e,i,n,(function(e){return e})):null!=e&&(D(e)&&(e=function(e,t){return{$$typeof:o,type:e.type,key:t,ref:e.ref,props:e.props,_owner:e._owner}}(e,r+(!e.key||t&&t.key===e.key?"":(""+e.key).replace($,"$&/")+"/")+n)),i.push(e))}function I(e,t,n,i,r){var o="";null!=n&&(o=(""+n).replace($,"$&/")+"/"),R(e,P,t=L(t,o,i,r)),T(t)}function N(){var e=F.current;if(null===e)throw Error(v(321));return e}var W={Children:{map:function(e,t,n){if(null==e)return e;var i=[];return I(e,i,null,t,n),i},forEach:function(e,t,n){if(null==e)return e;R(e,M,t=L(null,null,t,n)),T(t)},count:function(e){return R(e,(function(){return null}),null)},toArray:function(e){var t=[];return I(e,t,null,(function(e){return e})),t},only:function(e){if(!D(e))throw Error(v(143));return e}},createRef:function(){return{current:null}},Component:w,PureComponent:C,createContext:function(e,t){return void 0===t&&(t=null),(e={$$typeof:h,_calculateChangedBits:t,_currentValue:e,_currentValue2:e,_threadCount:0,Provider:null,Consumer:null}).Provider={$$typeof:c,_context:e},e.Consumer=e},forwardRef:function(e){return{$$typeof:f,render:e}},lazy:function(e){return{$$typeof:g,_ctor:e,_status:-1,_result:null}},memo:function(e,t){return{$$typeof:p,type:e,compare:void 0===t?null:t}},useCallback:function(e,t){return N().useCallback(e,t)},useContext:function(e,t){return N().useContext(e,t)},useEffect:function(e,t){return N().useEffect(e,t)},useImperativeHandle:function(e,t,n){return N().useImperativeHandle(e,t,n)},useDebugValue:function(){},useLayoutEffect:function(e,t){return N().useLayoutEffect(e,t)},useMemo:function(e,t){return N().useMemo(e,t)},useReducer:function(e,t,n){return N().useReducer(e,t,n)},useRef:function(e){return N().useRef(e)},useState:function(e){return N().useState(e)},Fragment:a,Profiler:u,StrictMode:l,Suspense:d,createElement:_,cloneElement:function(e,t,n){if(null==e)throw Error(v(267,e));var r=i({},e.props),s=e.key,a=e.ref,l=e._owner;if(null!=t){if(void 0!==t.ref&&(a=t.ref,l=x.current),void 0!==t.key&&(s=""+t.key),e.type&&e.type.defaultProps)var u=e.type.defaultProps;for(c in t)k.call(t,c)&&!S.hasOwnProperty(c)&&(r[c]=void 0===t[c]&&void 0!==u?u[c]:t[c])}var c=arguments.length-2;if(1===c)r.children=n;else if(1<c){u=Array(c);for(var h=0;h<c;h++)u[h]=arguments[h+2];r.children=u}return{$$typeof:o,type:e.type,key:s,ref:a,props:r,_owner:l}},createFactory:function(e){var t=_.bind(null,e);return t.type=e,t},isValidElement:D,version:"16.12.0",__SECRET_INTERNALS_DO_NOT_USE_OR_YOU_WILL_BE_FIRED:{ReactCurrentDispatcher:F,ReactCurrentBatchConfig:{suspense:null},ReactCurrentOwner:x,IsSomeRendererActing:{current:!1},assign:i}},z={default:W},j=z&&W||z;e.exports=j.default||j},function(e,t,n){"use strict";
/** @license React v16.12.0
 * react-dom.production.min.js
 *
 * Copyright (c) Facebook, Inc. and its affiliates.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 */var i=n(0),r=n(37),o=n(66);function s(e){for(var t="https://reactjs.org/docs/error-decoder.html?invariant="+e,n=1;n<arguments.length;n++)t+="&args[]="+encodeURIComponent(arguments[n]);return"Minified React error #"+e+"; visit "+t+" for the full message or use the non-minified dev environment for full errors and additional helpful warnings."}if(!i)throw Error(s(227));var a=null,l={};function u(){if(a)for(var e in l){var t=l[e],n=a.indexOf(e);if(!(-1<n))throw Error(s(96,e));if(!h[n]){if(!t.extractEvents)throw Error(s(97,e));for(var i in h[n]=t,n=t.eventTypes){var r=void 0,o=n[i],u=t,d=i;if(f.hasOwnProperty(d))throw Error(s(99,d));f[d]=o;var p=o.phasedRegistrationNames;if(p){for(r in p)p.hasOwnProperty(r)&&c(p[r],u,d);r=!0}else o.registrationName?(c(o.registrationName,u,d),r=!0):r=!1;if(!r)throw Error(s(98,i,e))}}}}function c(e,t,n){if(d[e])throw Error(s(100,e));d[e]=t,p[e]=t.eventTypes[n].dependencies}var h=[],f={},d={},p={};function g(e,t,n,i,r,o,s,a,l){var u=Array.prototype.slice.call(arguments,3);try{t.apply(n,u)}catch(e){this.onError(e)}}var m=!1,v=null,b=!1,y=null,w={onError:function(e){m=!0,v=e}};function A(e,t,n,i,r,o,s,a,l){m=!1,v=null,g.apply(w,arguments)}var C=null,E=null,F=null;function x(e,t,n){var i=e.type||"unknown-event";e.currentTarget=F(n),function(e,t,n,i,r,o,a,l,u){if(A.apply(this,arguments),m){if(!m)throw Error(s(198));var c=v;m=!1,v=null,b||(b=!0,y=c)}}(i,t,void 0,e),e.currentTarget=null}function k(e,t){if(null==t)throw Error(s(30));return null==e?t:Array.isArray(e)?Array.isArray(t)?(e.push.apply(e,t),e):(e.push(t),e):Array.isArray(t)?[e].concat(t):[e,t]}function S(e,t,n){Array.isArray(e)?e.forEach(t,n):e&&t.call(n,e)}var _=null;function D(e){if(e){var t=e._dispatchListeners,n=e._dispatchInstances;if(Array.isArray(t))for(var i=0;i<t.length&&!e.isPropagationStopped();i++)x(e,t[i],n[i]);else t&&x(e,t,n);e._dispatchListeners=null,e._dispatchInstances=null,e.isPersistent()||e.constructor.release(e)}}function $(e){if(null!==e&&(_=k(_,e)),e=_,_=null,e){if(S(e,D),_)throw Error(s(95));if(b)throw e=y,b=!1,y=null,e}}var B={injectEventPluginOrder:function(e){if(a)throw Error(s(101));a=Array.prototype.slice.call(e),u()},injectEventPluginsByName:function(e){var t,n=!1;for(t in e)if(e.hasOwnProperty(t)){var i=e[t];if(!l.hasOwnProperty(t)||l[t]!==i){if(l[t])throw Error(s(102,t));l[t]=i,n=!0}}n&&u()}};function L(e,t){var n=e.stateNode;if(!n)return null;var i=C(n);if(!i)return null;n=i[t];e:switch(t){case"onClick":case"onClickCapture":case"onDoubleClick":case"onDoubleClickCapture":case"onMouseDown":case"onMouseDownCapture":case"onMouseMove":case"onMouseMoveCapture":case"onMouseUp":case"onMouseUpCapture":(i=!i.disabled)||(i=!("button"===(e=e.type)||"input"===e||"select"===e||"textarea"===e)),e=!i;break e;default:e=!1}if(e)return null;if(n&&"function"!=typeof n)throw Error(s(231,t,typeof n));return n}var T=i.__SECRET_INTERNALS_DO_NOT_USE_OR_YOU_WILL_BE_FIRED;T.hasOwnProperty("ReactCurrentDispatcher")||(T.ReactCurrentDispatcher={current:null}),T.hasOwnProperty("ReactCurrentBatchConfig")||(T.ReactCurrentBatchConfig={suspense:null});var R=/^(.*)[\\\/]/,O="function"==typeof Symbol&&Symbol.for,M=O?Symbol.for("react.element"):60103,P=O?Symbol.for("react.portal"):60106,I=O?Symbol.for("react.fragment"):60107,N=O?Symbol.for("react.strict_mode"):60108,W=O?Symbol.for("react.profiler"):60114,z=O?Symbol.for("react.provider"):60109,j=O?Symbol.for("react.context"):60110,H=O?Symbol.for("react.concurrent_mode"):60111,V=O?Symbol.for("react.forward_ref"):60112,U=O?Symbol.for("react.suspense"):60113,K=O?Symbol.for("react.suspense_list"):60120,G=O?Symbol.for("react.memo"):60115,q=O?Symbol.for("react.lazy"):60116;O&&Symbol.for("react.fundamental"),O&&Symbol.for("react.responder"),O&&Symbol.for("react.scope");var Y="function"==typeof Symbol&&Symbol.iterator;function Q(e){return null===e||"object"!=typeof e?null:"function"==typeof(e=Y&&e[Y]||e["@@iterator"])?e:null}function X(e){if(null==e)return null;if("function"==typeof e)return e.displayName||e.name||null;if("string"==typeof e)return e;switch(e){case I:return"Fragment";case P:return"Portal";case W:return"Profiler";case N:return"StrictMode";case U:return"Suspense";case K:return"SuspenseList"}if("object"==typeof e)switch(e.$$typeof){case j:return"Context.Consumer";case z:return"Context.Provider";case V:var t=e.render;return t=t.displayName||t.name||"",e.displayName||(""!==t?"ForwardRef("+t+")":"ForwardRef");case G:return X(e.type);case q:if(e=1===e._status?e._result:null)return X(e)}return null}function J(e){var t="";do{e:switch(e.tag){case 3:case 4:case 6:case 7:case 10:case 9:var n="";break e;default:var i=e._debugOwner,r=e._debugSource,o=X(e.type);n=null,i&&(n=X(i.type)),i=o,o="",r?o=" (at "+r.fileName.replace(R,"")+":"+r.lineNumber+")":n&&(o=" (created by "+n+")"),n="\n    in "+(i||"Unknown")+o}t+=n,e=e.return}while(e);return t}var Z=!("undefined"==typeof window||void 0===window.document||void 0===window.document.createElement),ee=null,te=null,ne=null;function ie(e){if(e=E(e)){if("function"!=typeof ee)throw Error(s(280));var t=C(e.stateNode);ee(e.stateNode,e.type,t)}}function re(e){te?ne?ne.push(e):ne=[e]:te=e}function oe(){if(te){var e=te,t=ne;if(ne=te=null,ie(e),t)for(e=0;e<t.length;e++)ie(t[e])}}function se(e,t){return e(t)}function ae(e,t,n,i){return e(t,n,i)}function le(){}var ue=se,ce=!1,he=!1;function fe(){null===te&&null===ne||(le(),oe())}new Map;var de=/^[:A-Z_a-z\u00C0-\u00D6\u00D8-\u00F6\u00F8-\u02FF\u0370-\u037D\u037F-\u1FFF\u200C-\u200D\u2070-\u218F\u2C00-\u2FEF\u3001-\uD7FF\uF900-\uFDCF\uFDF0-\uFFFD][:A-Z_a-z\u00C0-\u00D6\u00D8-\u00F6\u00F8-\u02FF\u0370-\u037D\u037F-\u1FFF\u200C-\u200D\u2070-\u218F\u2C00-\u2FEF\u3001-\uD7FF\uF900-\uFDCF\uFDF0-\uFFFD\-.0-9\u00B7\u0300-\u036F\u203F-\u2040]*$/,pe=Object.prototype.hasOwnProperty,ge={},me={};function ve(e,t,n,i,r,o){this.acceptsBooleans=2===t||3===t||4===t,this.attributeName=i,this.attributeNamespace=r,this.mustUseProperty=n,this.propertyName=e,this.type=t,this.sanitizeURL=o}var be={};"children dangerouslySetInnerHTML defaultValue defaultChecked innerHTML suppressContentEditableWarning suppressHydrationWarning style".split(" ").forEach((function(e){be[e]=new ve(e,0,!1,e,null,!1)})),[["acceptCharset","accept-charset"],["className","class"],["htmlFor","for"],["httpEquiv","http-equiv"]].forEach((function(e){var t=e[0];be[t]=new ve(t,1,!1,e[1],null,!1)})),["contentEditable","draggable","spellCheck","value"].forEach((function(e){be[e]=new ve(e,2,!1,e.toLowerCase(),null,!1)})),["autoReverse","externalResourcesRequired","focusable","preserveAlpha"].forEach((function(e){be[e]=new ve(e,2,!1,e,null,!1)})),"allowFullScreen async autoFocus autoPlay controls default defer disabled disablePictureInPicture formNoValidate hidden loop noModule noValidate open playsInline readOnly required reversed scoped seamless itemScope".split(" ").forEach((function(e){be[e]=new ve(e,3,!1,e.toLowerCase(),null,!1)})),["checked","multiple","muted","selected"].forEach((function(e){be[e]=new ve(e,3,!0,e,null,!1)})),["capture","download"].forEach((function(e){be[e]=new ve(e,4,!1,e,null,!1)})),["cols","rows","size","span"].forEach((function(e){be[e]=new ve(e,6,!1,e,null,!1)})),["rowSpan","start"].forEach((function(e){be[e]=new ve(e,5,!1,e.toLowerCase(),null,!1)}));var ye=/[\-:]([a-z])/g;function we(e){return e[1].toUpperCase()}function Ae(e){switch(typeof e){case"boolean":case"number":case"object":case"string":case"undefined":return e;default:return""}}function Ce(e,t,n,i){var r=be.hasOwnProperty(t)?be[t]:null;(null!==r?0===r.type:!i&&(2<t.length&&("o"===t[0]||"O"===t[0])&&("n"===t[1]||"N"===t[1])))||(function(e,t,n,i){if(null==t||function(e,t,n,i){if(null!==n&&0===n.type)return!1;switch(typeof t){case"function":case"symbol":return!0;case"boolean":return!i&&(null!==n?!n.acceptsBooleans:"data-"!==(e=e.toLowerCase().slice(0,5))&&"aria-"!==e);default:return!1}}(e,t,n,i))return!0;if(i)return!1;if(null!==n)switch(n.type){case 3:return!t;case 4:return!1===t;case 5:return isNaN(t);case 6:return isNaN(t)||1>t}return!1}(t,n,r,i)&&(n=null),i||null===r?function(e){return!!pe.call(me,e)||!pe.call(ge,e)&&(de.test(e)?me[e]=!0:(ge[e]=!0,!1))}(t)&&(null===n?e.removeAttribute(t):e.setAttribute(t,""+n)):r.mustUseProperty?e[r.propertyName]=null===n?3!==r.type&&"":n:(t=r.attributeName,i=r.attributeNamespace,null===n?e.removeAttribute(t):(n=3===(r=r.type)||4===r&&!0===n?"":""+n,i?e.setAttributeNS(i,t,n):e.setAttribute(t,n))))}function Ee(e){var t=e.type;return(e=e.nodeName)&&"input"===e.toLowerCase()&&("checkbox"===t||"radio"===t)}function Fe(e){e._valueTracker||(e._valueTracker=function(e){var t=Ee(e)?"checked":"value",n=Object.getOwnPropertyDescriptor(e.constructor.prototype,t),i=""+e[t];if(!e.hasOwnProperty(t)&&void 0!==n&&"function"==typeof n.get&&"function"==typeof n.set){var r=n.get,o=n.set;return Object.defineProperty(e,t,{configurable:!0,get:function(){return r.call(this)},set:function(e){i=""+e,o.call(this,e)}}),Object.defineProperty(e,t,{enumerable:n.enumerable}),{getValue:function(){return i},setValue:function(e){i=""+e},stopTracking:function(){e._valueTracker=null,delete e[t]}}}}(e))}function xe(e){if(!e)return!1;var t=e._valueTracker;if(!t)return!0;var n=t.getValue(),i="";return e&&(i=Ee(e)?e.checked?"true":"false":e.value),(e=i)!==n&&(t.setValue(e),!0)}function ke(e,t){var n=t.checked;return r({},t,{defaultChecked:void 0,defaultValue:void 0,value:void 0,checked:null!=n?n:e._wrapperState.initialChecked})}function Se(e,t){var n=null==t.defaultValue?"":t.defaultValue,i=null!=t.checked?t.checked:t.defaultChecked;n=Ae(null!=t.value?t.value:n),e._wrapperState={initialChecked:i,initialValue:n,controlled:"checkbox"===t.type||"radio"===t.type?null!=t.checked:null!=t.value}}function _e(e,t){null!=(t=t.checked)&&Ce(e,"checked",t,!1)}function De(e,t){_e(e,t);var n=Ae(t.value),i=t.type;if(null!=n)"number"===i?(0===n&&""===e.value||e.value!=n)&&(e.value=""+n):e.value!==""+n&&(e.value=""+n);else if("submit"===i||"reset"===i)return void e.removeAttribute("value");t.hasOwnProperty("value")?Be(e,t.type,n):t.hasOwnProperty("defaultValue")&&Be(e,t.type,Ae(t.defaultValue)),null==t.checked&&null!=t.defaultChecked&&(e.defaultChecked=!!t.defaultChecked)}function $e(e,t,n){if(t.hasOwnProperty("value")||t.hasOwnProperty("defaultValue")){var i=t.type;if(!("submit"!==i&&"reset"!==i||void 0!==t.value&&null!==t.value))return;t=""+e._wrapperState.initialValue,n||t===e.value||(e.value=t),e.defaultValue=t}""!==(n=e.name)&&(e.name=""),e.defaultChecked=!e.defaultChecked,e.defaultChecked=!!e._wrapperState.initialChecked,""!==n&&(e.name=n)}function Be(e,t,n){"number"===t&&e.ownerDocument.activeElement===e||(null==n?e.defaultValue=""+e._wrapperState.initialValue:e.defaultValue!==""+n&&(e.defaultValue=""+n))}function Le(e,t){return e=r({children:void 0},t),(t=function(e){var t="";return i.Children.forEach(e,(function(e){null!=e&&(t+=e)})),t}(t.children))&&(e.children=t),e}function Te(e,t,n,i){if(e=e.options,t){t={};for(var r=0;r<n.length;r++)t["$"+n[r]]=!0;for(n=0;n<e.length;n++)r=t.hasOwnProperty("$"+e[n].value),e[n].selected!==r&&(e[n].selected=r),r&&i&&(e[n].defaultSelected=!0)}else{for(n=""+Ae(n),t=null,r=0;r<e.length;r++){if(e[r].value===n)return e[r].selected=!0,void(i&&(e[r].defaultSelected=!0));null!==t||e[r].disabled||(t=e[r])}null!==t&&(t.selected=!0)}}function Re(e,t){if(null!=t.dangerouslySetInnerHTML)throw Error(s(91));return r({},t,{value:void 0,defaultValue:void 0,children:""+e._wrapperState.initialValue})}function Oe(e,t){var n=t.value;if(null==n){if(n=t.defaultValue,null!=(t=t.children)){if(null!=n)throw Error(s(92));if(Array.isArray(t)){if(!(1>=t.length))throw Error(s(93));t=t[0]}n=t}null==n&&(n="")}e._wrapperState={initialValue:Ae(n)}}function Me(e,t){var n=Ae(t.value),i=Ae(t.defaultValue);null!=n&&((n=""+n)!==e.value&&(e.value=n),null==t.defaultValue&&e.defaultValue!==n&&(e.defaultValue=n)),null!=i&&(e.defaultValue=""+i)}function Pe(e){var t=e.textContent;t===e._wrapperState.initialValue&&""!==t&&null!==t&&(e.value=t)}"accent-height alignment-baseline arabic-form baseline-shift cap-height clip-path clip-rule color-interpolation color-interpolation-filters color-profile color-rendering dominant-baseline enable-background fill-opacity fill-rule flood-color flood-opacity font-family font-size font-size-adjust font-stretch font-style font-variant font-weight glyph-name glyph-orientation-horizontal glyph-orientation-vertical horiz-adv-x horiz-origin-x image-rendering letter-spacing lighting-color marker-end marker-mid marker-start overline-position overline-thickness paint-order panose-1 pointer-events rendering-intent shape-rendering stop-color stop-opacity strikethrough-position strikethrough-thickness stroke-dasharray stroke-dashoffset stroke-linecap stroke-linejoin stroke-miterlimit stroke-opacity stroke-width text-anchor text-decoration text-rendering underline-position underline-thickness unicode-bidi unicode-range units-per-em v-alphabetic v-hanging v-ideographic v-mathematical vector-effect vert-adv-y vert-origin-x vert-origin-y word-spacing writing-mode xmlns:xlink x-height".split(" ").forEach((function(e){var t=e.replace(ye,we);be[t]=new ve(t,1,!1,e,null,!1)})),"xlink:actuate xlink:arcrole xlink:role xlink:show xlink:title xlink:type".split(" ").forEach((function(e){var t=e.replace(ye,we);be[t]=new ve(t,1,!1,e,"http://www.w3.org/1999/xlink",!1)})),["xml:base","xml:lang","xml:space"].forEach((function(e){var t=e.replace(ye,we);be[t]=new ve(t,1,!1,e,"http://www.w3.org/XML/1998/namespace",!1)})),["tabIndex","crossOrigin"].forEach((function(e){be[e]=new ve(e,1,!1,e.toLowerCase(),null,!1)})),be.xlinkHref=new ve("xlinkHref",1,!1,"xlink:href","http://www.w3.org/1999/xlink",!0),["src","href","action","formAction"].forEach((function(e){be[e]=new ve(e,1,!1,e.toLowerCase(),null,!0)}));var Ie="http://www.w3.org/1999/xhtml",Ne="http://www.w3.org/2000/svg";function We(e){switch(e){case"svg":return"http://www.w3.org/2000/svg";case"math":return"http://www.w3.org/1998/Math/MathML";default:return"http://www.w3.org/1999/xhtml"}}function ze(e,t){return null==e||"http://www.w3.org/1999/xhtml"===e?We(t):"http://www.w3.org/2000/svg"===e&&"foreignObject"===t?"http://www.w3.org/1999/xhtml":e}var je,He=function(e){return"undefined"!=typeof MSApp&&MSApp.execUnsafeLocalFunction?function(t,n,i,r){MSApp.execUnsafeLocalFunction((function(){return e(t,n)}))}:e}((function(e,t){if(e.namespaceURI!==Ne||"innerHTML"in e)e.innerHTML=t;else{for((je=je||document.createElement("div")).innerHTML="<svg>"+t.valueOf().toString()+"</svg>",t=je.firstChild;e.firstChild;)e.removeChild(e.firstChild);for(;t.firstChild;)e.appendChild(t.firstChild)}}));function Ve(e,t){if(t){var n=e.firstChild;if(n&&n===e.lastChild&&3===n.nodeType)return void(n.nodeValue=t)}e.textContent=t}function Ue(e,t){var n={};return n[e.toLowerCase()]=t.toLowerCase(),n["Webkit"+e]="webkit"+t,n["Moz"+e]="moz"+t,n}var Ke={animationend:Ue("Animation","AnimationEnd"),animationiteration:Ue("Animation","AnimationIteration"),animationstart:Ue("Animation","AnimationStart"),transitionend:Ue("Transition","TransitionEnd")},Ge={},qe={};function Ye(e){if(Ge[e])return Ge[e];if(!Ke[e])return e;var t,n=Ke[e];for(t in n)if(n.hasOwnProperty(t)&&t in qe)return Ge[e]=n[t];return e}Z&&(qe=document.createElement("div").style,"AnimationEvent"in window||(delete Ke.animationend.animation,delete Ke.animationiteration.animation,delete Ke.animationstart.animation),"TransitionEvent"in window||delete Ke.transitionend.transition);var Qe=Ye("animationend"),Xe=Ye("animationiteration"),Je=Ye("animationstart"),Ze=Ye("transitionend"),et="abort canplay canplaythrough durationchange emptied encrypted ended error loadeddata loadedmetadata loadstart pause play playing progress ratechange seeked seeking stalled suspend timeupdate volumechange waiting".split(" ");function tt(e){var t=e,n=e;if(e.alternate)for(;t.return;)t=t.return;else{e=t;do{0!=(1026&(t=e).effectTag)&&(n=t.return),e=t.return}while(e)}return 3===t.tag?n:null}function nt(e){if(13===e.tag){var t=e.memoizedState;if(null===t&&(null!==(e=e.alternate)&&(t=e.memoizedState)),null!==t)return t.dehydrated}return null}function it(e){if(tt(e)!==e)throw Error(s(188))}function rt(e){if(!(e=function(e){var t=e.alternate;if(!t){if(null===(t=tt(e)))throw Error(s(188));return t!==e?null:e}for(var n=e,i=t;;){var r=n.return;if(null===r)break;var o=r.alternate;if(null===o){if(null!==(i=r.return)){n=i;continue}break}if(r.child===o.child){for(o=r.child;o;){if(o===n)return it(r),e;if(o===i)return it(r),t;o=o.sibling}throw Error(s(188))}if(n.return!==i.return)n=r,i=o;else{for(var a=!1,l=r.child;l;){if(l===n){a=!0,n=r,i=o;break}if(l===i){a=!0,i=r,n=o;break}l=l.sibling}if(!a){for(l=o.child;l;){if(l===n){a=!0,n=o,i=r;break}if(l===i){a=!0,i=o,n=r;break}l=l.sibling}if(!a)throw Error(s(189))}}if(n.alternate!==i)throw Error(s(190))}if(3!==n.tag)throw Error(s(188));return n.stateNode.current===n?e:t}(e)))return null;for(var t=e;;){if(5===t.tag||6===t.tag)return t;if(t.child)t.child.return=t,t=t.child;else{if(t===e)break;for(;!t.sibling;){if(!t.return||t.return===e)return null;t=t.return}t.sibling.return=t.return,t=t.sibling}}return null}var ot,st,at,lt=!1,ut=[],ct=null,ht=null,ft=null,dt=new Map,pt=new Map,gt=[],mt="mousedown mouseup touchcancel touchend touchstart auxclick dblclick pointercancel pointerdown pointerup dragend dragstart drop compositionend compositionstart keydown keypress keyup input textInput close cancel copy cut paste click change contextmenu reset submit".split(" "),vt="focus blur dragenter dragleave mouseover mouseout pointerover pointerout gotpointercapture lostpointercapture".split(" ");function bt(e,t,n,i){return{blockedOn:e,topLevelType:t,eventSystemFlags:32|n,nativeEvent:i}}function yt(e,t){switch(e){case"focus":case"blur":ct=null;break;case"dragenter":case"dragleave":ht=null;break;case"mouseover":case"mouseout":ft=null;break;case"pointerover":case"pointerout":dt.delete(t.pointerId);break;case"gotpointercapture":case"lostpointercapture":pt.delete(t.pointerId)}}function wt(e,t,n,i,r){return null===e||e.nativeEvent!==r?(e=bt(t,n,i,r),null!==t&&(null!==(t=ui(t))&&st(t)),e):(e.eventSystemFlags|=i,e)}function At(e){var t=li(e.target);if(null!==t){var n=tt(t);if(null!==n)if(13===(t=n.tag)){if(null!==(t=nt(n)))return e.blockedOn=t,void o.unstable_runWithPriority(e.priority,(function(){at(n)}))}else if(3===t&&n.stateNode.hydrate)return void(e.blockedOn=3===n.tag?n.stateNode.containerInfo:null)}e.blockedOn=null}function Ct(e){if(null!==e.blockedOn)return!1;var t=Bn(e.topLevelType,e.eventSystemFlags,e.nativeEvent);if(null!==t){var n=ui(t);return null!==n&&st(n),e.blockedOn=t,!1}return!0}function Et(e,t,n){Ct(e)&&n.delete(t)}function Ft(){for(lt=!1;0<ut.length;){var e=ut[0];if(null!==e.blockedOn){null!==(e=ui(e.blockedOn))&&ot(e);break}var t=Bn(e.topLevelType,e.eventSystemFlags,e.nativeEvent);null!==t?e.blockedOn=t:ut.shift()}null!==ct&&Ct(ct)&&(ct=null),null!==ht&&Ct(ht)&&(ht=null),null!==ft&&Ct(ft)&&(ft=null),dt.forEach(Et),pt.forEach(Et)}function xt(e,t){e.blockedOn===t&&(e.blockedOn=null,lt||(lt=!0,o.unstable_scheduleCallback(o.unstable_NormalPriority,Ft)))}function kt(e){function t(t){return xt(t,e)}if(0<ut.length){xt(ut[0],e);for(var n=1;n<ut.length;n++){var i=ut[n];i.blockedOn===e&&(i.blockedOn=null)}}for(null!==ct&&xt(ct,e),null!==ht&&xt(ht,e),null!==ft&&xt(ft,e),dt.forEach(t),pt.forEach(t),n=0;n<gt.length;n++)(i=gt[n]).blockedOn===e&&(i.blockedOn=null);for(;0<gt.length&&null===(n=gt[0]).blockedOn;)At(n),null===n.blockedOn&&gt.shift()}function St(e){return(e=e.target||e.srcElement||window).correspondingUseElement&&(e=e.correspondingUseElement),3===e.nodeType?e.parentNode:e}function _t(e){do{e=e.return}while(e&&5!==e.tag);return e||null}function Dt(e,t,n){(t=L(e,n.dispatchConfig.phasedRegistrationNames[t]))&&(n._dispatchListeners=k(n._dispatchListeners,t),n._dispatchInstances=k(n._dispatchInstances,e))}function $t(e){if(e&&e.dispatchConfig.phasedRegistrationNames){for(var t=e._targetInst,n=[];t;)n.push(t),t=_t(t);for(t=n.length;0<t--;)Dt(n[t],"captured",e);for(t=0;t<n.length;t++)Dt(n[t],"bubbled",e)}}function Bt(e,t,n){e&&n&&n.dispatchConfig.registrationName&&(t=L(e,n.dispatchConfig.registrationName))&&(n._dispatchListeners=k(n._dispatchListeners,t),n._dispatchInstances=k(n._dispatchInstances,e))}function Lt(e){e&&e.dispatchConfig.registrationName&&Bt(e._targetInst,null,e)}function Tt(e){S(e,$t)}function Rt(){return!0}function Ot(){return!1}function Mt(e,t,n,i){for(var r in this.dispatchConfig=e,this._targetInst=t,this.nativeEvent=n,e=this.constructor.Interface)e.hasOwnProperty(r)&&((t=e[r])?this[r]=t(n):"target"===r?this.target=i:this[r]=n[r]);return this.isDefaultPrevented=(null!=n.defaultPrevented?n.defaultPrevented:!1===n.returnValue)?Rt:Ot,this.isPropagationStopped=Ot,this}function Pt(e,t,n,i){if(this.eventPool.length){var r=this.eventPool.pop();return this.call(r,e,t,n,i),r}return new this(e,t,n,i)}function It(e){if(!(e instanceof this))throw Error(s(279));e.destructor(),10>this.eventPool.length&&this.eventPool.push(e)}function Nt(e){e.eventPool=[],e.getPooled=Pt,e.release=It}r(Mt.prototype,{preventDefault:function(){this.defaultPrevented=!0;var e=this.nativeEvent;e&&(e.preventDefault?e.preventDefault():"unknown"!=typeof e.returnValue&&(e.returnValue=!1),this.isDefaultPrevented=Rt)},stopPropagation:function(){var e=this.nativeEvent;e&&(e.stopPropagation?e.stopPropagation():"unknown"!=typeof e.cancelBubble&&(e.cancelBubble=!0),this.isPropagationStopped=Rt)},persist:function(){this.isPersistent=Rt},isPersistent:Ot,destructor:function(){var e,t=this.constructor.Interface;for(e in t)this[e]=null;this.nativeEvent=this._targetInst=this.dispatchConfig=null,this.isPropagationStopped=this.isDefaultPrevented=Ot,this._dispatchInstances=this._dispatchListeners=null}}),Mt.Interface={type:null,target:null,currentTarget:function(){return null},eventPhase:null,bubbles:null,cancelable:null,timeStamp:function(e){return e.timeStamp||Date.now()},defaultPrevented:null,isTrusted:null},Mt.extend=function(e){function t(){}function n(){return i.apply(this,arguments)}var i=this;t.prototype=i.prototype;var o=new t;return r(o,n.prototype),n.prototype=o,n.prototype.constructor=n,n.Interface=r({},i.Interface,e),n.extend=i.extend,Nt(n),n},Nt(Mt);var Wt=Mt.extend({animationName:null,elapsedTime:null,pseudoElement:null}),zt=Mt.extend({clipboardData:function(e){return"clipboardData"in e?e.clipboardData:window.clipboardData}}),jt=Mt.extend({view:null,detail:null}),Ht=jt.extend({relatedTarget:null});function Vt(e){var t=e.keyCode;return"charCode"in e?0===(e=e.charCode)&&13===t&&(e=13):e=t,10===e&&(e=13),32<=e||13===e?e:0}var Ut={Esc:"Escape",Spacebar:" ",Left:"ArrowLeft",Up:"ArrowUp",Right:"ArrowRight",Down:"ArrowDown",Del:"Delete",Win:"OS",Menu:"ContextMenu",Apps:"ContextMenu",Scroll:"ScrollLock",MozPrintableKey:"Unidentified"},Kt={8:"Backspace",9:"Tab",12:"Clear",13:"Enter",16:"Shift",17:"Control",18:"Alt",19:"Pause",20:"CapsLock",27:"Escape",32:" ",33:"PageUp",34:"PageDown",35:"End",36:"Home",37:"ArrowLeft",38:"ArrowUp",39:"ArrowRight",40:"ArrowDown",45:"Insert",46:"Delete",112:"F1",113:"F2",114:"F3",115:"F4",116:"F5",117:"F6",118:"F7",119:"F8",120:"F9",121:"F10",122:"F11",123:"F12",144:"NumLock",145:"ScrollLock",224:"Meta"},Gt={Alt:"altKey",Control:"ctrlKey",Meta:"metaKey",Shift:"shiftKey"};function qt(e){var t=this.nativeEvent;return t.getModifierState?t.getModifierState(e):!!(e=Gt[e])&&!!t[e]}function Yt(){return qt}for(var Qt=jt.extend({key:function(e){if(e.key){var t=Ut[e.key]||e.key;if("Unidentified"!==t)return t}return"keypress"===e.type?13===(e=Vt(e))?"Enter":String.fromCharCode(e):"keydown"===e.type||"keyup"===e.type?Kt[e.keyCode]||"Unidentified":""},location:null,ctrlKey:null,shiftKey:null,altKey:null,metaKey:null,repeat:null,locale:null,getModifierState:Yt,charCode:function(e){return"keypress"===e.type?Vt(e):0},keyCode:function(e){return"keydown"===e.type||"keyup"===e.type?e.keyCode:0},which:function(e){return"keypress"===e.type?Vt(e):"keydown"===e.type||"keyup"===e.type?e.keyCode:0}}),Xt=0,Jt=0,Zt=!1,en=!1,tn=jt.extend({screenX:null,screenY:null,clientX:null,clientY:null,pageX:null,pageY:null,ctrlKey:null,shiftKey:null,altKey:null,metaKey:null,getModifierState:Yt,button:null,buttons:null,relatedTarget:function(e){return e.relatedTarget||(e.fromElement===e.srcElement?e.toElement:e.fromElement)},movementX:function(e){if("movementX"in e)return e.movementX;var t=Xt;return Xt=e.screenX,Zt?"mousemove"===e.type?e.screenX-t:0:(Zt=!0,0)},movementY:function(e){if("movementY"in e)return e.movementY;var t=Jt;return Jt=e.screenY,en?"mousemove"===e.type?e.screenY-t:0:(en=!0,0)}}),nn=tn.extend({pointerId:null,width:null,height:null,pressure:null,tangentialPressure:null,tiltX:null,tiltY:null,twist:null,pointerType:null,isPrimary:null}),rn=tn.extend({dataTransfer:null}),on=jt.extend({touches:null,targetTouches:null,changedTouches:null,altKey:null,metaKey:null,ctrlKey:null,shiftKey:null,getModifierState:Yt}),sn=Mt.extend({propertyName:null,elapsedTime:null,pseudoElement:null}),an=tn.extend({deltaX:function(e){return"deltaX"in e?e.deltaX:"wheelDeltaX"in e?-e.wheelDeltaX:0},deltaY:function(e){return"deltaY"in e?e.deltaY:"wheelDeltaY"in e?-e.wheelDeltaY:"wheelDelta"in e?-e.wheelDelta:0},deltaZ:null,deltaMode:null}),ln=[["blur","blur",0],["cancel","cancel",0],["click","click",0],["close","close",0],["contextmenu","contextMenu",0],["copy","copy",0],["cut","cut",0],["auxclick","auxClick",0],["dblclick","doubleClick",0],["dragend","dragEnd",0],["dragstart","dragStart",0],["drop","drop",0],["focus","focus",0],["input","input",0],["invalid","invalid",0],["keydown","keyDown",0],["keypress","keyPress",0],["keyup","keyUp",0],["mousedown","mouseDown",0],["mouseup","mouseUp",0],["paste","paste",0],["pause","pause",0],["play","play",0],["pointercancel","pointerCancel",0],["pointerdown","pointerDown",0],["pointerup","pointerUp",0],["ratechange","rateChange",0],["reset","reset",0],["seeked","seeked",0],["submit","submit",0],["touchcancel","touchCancel",0],["touchend","touchEnd",0],["touchstart","touchStart",0],["volumechange","volumeChange",0],["drag","drag",1],["dragenter","dragEnter",1],["dragexit","dragExit",1],["dragleave","dragLeave",1],["dragover","dragOver",1],["mousemove","mouseMove",1],["mouseout","mouseOut",1],["mouseover","mouseOver",1],["pointermove","pointerMove",1],["pointerout","pointerOut",1],["pointerover","pointerOver",1],["scroll","scroll",1],["toggle","toggle",1],["touchmove","touchMove",1],["wheel","wheel",1],["abort","abort",2],[Qe,"animationEnd",2],[Xe,"animationIteration",2],[Je,"animationStart",2],["canplay","canPlay",2],["canplaythrough","canPlayThrough",2],["durationchange","durationChange",2],["emptied","emptied",2],["encrypted","encrypted",2],["ended","ended",2],["error","error",2],["gotpointercapture","gotPointerCapture",2],["load","load",2],["loadeddata","loadedData",2],["loadedmetadata","loadedMetadata",2],["loadstart","loadStart",2],["lostpointercapture","lostPointerCapture",2],["playing","playing",2],["progress","progress",2],["seeking","seeking",2],["stalled","stalled",2],["suspend","suspend",2],["timeupdate","timeUpdate",2],[Ze,"transitionEnd",2],["waiting","waiting",2]],un={},cn={},hn=0;hn<ln.length;hn++){var fn=ln[hn],dn=fn[0],pn=fn[1],gn=fn[2],mn="on"+(pn[0].toUpperCase()+pn.slice(1)),vn={phasedRegistrationNames:{bubbled:mn,captured:mn+"Capture"},dependencies:[dn],eventPriority:gn};un[pn]=vn,cn[dn]=vn}var bn={eventTypes:un,getEventPriority:function(e){return void 0!==(e=cn[e])?e.eventPriority:2},extractEvents:function(e,t,n,i){var r=cn[e];if(!r)return null;switch(e){case"keypress":if(0===Vt(n))return null;case"keydown":case"keyup":e=Qt;break;case"blur":case"focus":e=Ht;break;case"click":if(2===n.button)return null;case"auxclick":case"dblclick":case"mousedown":case"mousemove":case"mouseup":case"mouseout":case"mouseover":case"contextmenu":e=tn;break;case"drag":case"dragend":case"dragenter":case"dragexit":case"dragleave":case"dragover":case"dragstart":case"drop":e=rn;break;case"touchcancel":case"touchend":case"touchmove":case"touchstart":e=on;break;case Qe:case Xe:case Je:e=Wt;break;case Ze:e=sn;break;case"scroll":e=jt;break;case"wheel":e=an;break;case"copy":case"cut":case"paste":e=zt;break;case"gotpointercapture":case"lostpointercapture":case"pointercancel":case"pointerdown":case"pointermove":case"pointerout":case"pointerover":case"pointerup":e=nn;break;default:e=Mt}return Tt(t=e.getPooled(r,t,n,i)),t}},yn=o.unstable_UserBlockingPriority,wn=o.unstable_runWithPriority,An=bn.getEventPriority,Cn=[];function En(e){var t=e.targetInst,n=t;do{if(!n){e.ancestors.push(n);break}var i=n;if(3===i.tag)i=i.stateNode.containerInfo;else{for(;i.return;)i=i.return;i=3!==i.tag?null:i.stateNode.containerInfo}if(!i)break;5!==(t=n.tag)&&6!==t||e.ancestors.push(n),n=li(i)}while(n);for(n=0;n<e.ancestors.length;n++){t=e.ancestors[n];var r=St(e.nativeEvent);i=e.topLevelType;for(var o=e.nativeEvent,s=e.eventSystemFlags,a=null,l=0;l<h.length;l++){var u=h[l];u&&(u=u.extractEvents(i,t,o,r,s))&&(a=k(a,u))}$(a)}}var Fn=!0;function xn(e,t){kn(t,e,!1)}function kn(e,t,n){switch(An(t)){case 0:var i=Sn.bind(null,t,1);break;case 1:i=_n.bind(null,t,1);break;default:i=$n.bind(null,t,1)}n?e.addEventListener(t,i,!0):e.addEventListener(t,i,!1)}function Sn(e,t,n){ce||le();var i=$n,r=ce;ce=!0;try{ae(i,e,t,n)}finally{(ce=r)||fe()}}function _n(e,t,n){wn(yn,$n.bind(null,e,t,n))}function Dn(e,t,n,i){if(Cn.length){var r=Cn.pop();r.topLevelType=e,r.eventSystemFlags=t,r.nativeEvent=n,r.targetInst=i,e=r}else e={topLevelType:e,eventSystemFlags:t,nativeEvent:n,targetInst:i,ancestors:[]};try{if(t=En,n=e,he)t(n,void 0);else{he=!0;try{ue(t,n,void 0)}finally{he=!1,fe()}}}finally{e.topLevelType=null,e.nativeEvent=null,e.targetInst=null,e.ancestors.length=0,Cn.length<10&&Cn.push(e)}}function $n(e,t,n){if(Fn)if(0<ut.length&&-1<mt.indexOf(e))e=bt(null,e,t,n),ut.push(e);else{var i=Bn(e,t,n);null===i?yt(e,n):-1<mt.indexOf(e)?(e=bt(i,e,t,n),ut.push(e)):function(e,t,n,i){switch(t){case"focus":return ct=wt(ct,e,t,n,i),!0;case"dragenter":return ht=wt(ht,e,t,n,i),!0;case"mouseover":return ft=wt(ft,e,t,n,i),!0;case"pointerover":var r=i.pointerId;return dt.set(r,wt(dt.get(r)||null,e,t,n,i)),!0;case"gotpointercapture":return r=i.pointerId,pt.set(r,wt(pt.get(r)||null,e,t,n,i)),!0}return!1}(i,e,t,n)||(yt(e,n),Dn(e,t,n,null))}}function Bn(e,t,n){var i=St(n);if(null!==(i=li(i))){var r=tt(i);if(null===r)i=null;else{var o=r.tag;if(13===o){if(null!==(i=nt(r)))return i;i=null}else if(3===o){if(r.stateNode.hydrate)return 3===r.tag?r.stateNode.containerInfo:null;i=null}else r!==i&&(i=null)}}return Dn(e,t,n,i),null}function Ln(e){if(!Z)return!1;var t=(e="on"+e)in document;return t||((t=document.createElement("div")).setAttribute(e,"return;"),t="function"==typeof t[e]),t}var Tn=new("function"==typeof WeakMap?WeakMap:Map);function Rn(e){var t=Tn.get(e);return void 0===t&&(t=new Set,Tn.set(e,t)),t}function On(e,t,n){if(!n.has(e)){switch(e){case"scroll":kn(t,"scroll",!0);break;case"focus":case"blur":kn(t,"focus",!0),kn(t,"blur",!0),n.add("blur"),n.add("focus");break;case"cancel":case"close":Ln(e)&&kn(t,e,!0);break;case"invalid":case"submit":case"reset":break;default:-1===et.indexOf(e)&&xn(e,t)}n.add(e)}}var Mn={animationIterationCount:!0,borderImageOutset:!0,borderImageSlice:!0,borderImageWidth:!0,boxFlex:!0,boxFlexGroup:!0,boxOrdinalGroup:!0,columnCount:!0,columns:!0,flex:!0,flexGrow:!0,flexPositive:!0,flexShrink:!0,flexNegative:!0,flexOrder:!0,gridArea:!0,gridRow:!0,gridRowEnd:!0,gridRowSpan:!0,gridRowStart:!0,gridColumn:!0,gridColumnEnd:!0,gridColumnSpan:!0,gridColumnStart:!0,fontWeight:!0,lineClamp:!0,lineHeight:!0,opacity:!0,order:!0,orphans:!0,tabSize:!0,widows:!0,zIndex:!0,zoom:!0,fillOpacity:!0,floodOpacity:!0,stopOpacity:!0,strokeDasharray:!0,strokeDashoffset:!0,strokeMiterlimit:!0,strokeOpacity:!0,strokeWidth:!0},Pn=["Webkit","ms","Moz","O"];function In(e,t,n){return null==t||"boolean"==typeof t||""===t?"":n||"number"!=typeof t||0===t||Mn.hasOwnProperty(e)&&Mn[e]?(""+t).trim():t+"px"}function Nn(e,t){for(var n in e=e.style,t)if(t.hasOwnProperty(n)){var i=0===n.indexOf("--"),r=In(n,t[n],i);"float"===n&&(n="cssFloat"),i?e.setProperty(n,r):e[n]=r}}Object.keys(Mn).forEach((function(e){Pn.forEach((function(t){t=t+e.charAt(0).toUpperCase()+e.substring(1),Mn[t]=Mn[e]}))}));var Wn=r({menuitem:!0},{area:!0,base:!0,br:!0,col:!0,embed:!0,hr:!0,img:!0,input:!0,keygen:!0,link:!0,meta:!0,param:!0,source:!0,track:!0,wbr:!0});function zn(e,t){if(t){if(Wn[e]&&(null!=t.children||null!=t.dangerouslySetInnerHTML))throw Error(s(137,e,""));if(null!=t.dangerouslySetInnerHTML){if(null!=t.children)throw Error(s(60));if(!("object"==typeof t.dangerouslySetInnerHTML&&"__html"in t.dangerouslySetInnerHTML))throw Error(s(61))}if(null!=t.style&&"object"!=typeof t.style)throw Error(s(62,""))}}function jn(e,t){if(-1===e.indexOf("-"))return"string"==typeof t.is;switch(e){case"annotation-xml":case"color-profile":case"font-face":case"font-face-src":case"font-face-uri":case"font-face-format":case"font-face-name":case"missing-glyph":return!1;default:return!0}}function Hn(e,t){var n=Rn(e=9===e.nodeType||11===e.nodeType?e:e.ownerDocument);t=p[t];for(var i=0;i<t.length;i++)On(t[i],e,n)}function Vn(){}function Un(e){if(void 0===(e=e||("undefined"!=typeof document?document:void 0)))return null;try{return e.activeElement||e.body}catch(t){return e.body}}function Kn(e){for(;e&&e.firstChild;)e=e.firstChild;return e}function Gn(e,t){var n,i=Kn(e);for(e=0;i;){if(3===i.nodeType){if(n=e+i.textContent.length,e<=t&&n>=t)return{node:i,offset:t-e};e=n}e:{for(;i;){if(i.nextSibling){i=i.nextSibling;break e}i=i.parentNode}i=void 0}i=Kn(i)}}function qn(){for(var e=window,t=Un();t instanceof e.HTMLIFrameElement;){try{var n="string"==typeof t.contentWindow.location.href}catch(e){n=!1}if(!n)break;t=Un((e=t.contentWindow).document)}return t}function Yn(e){var t=e&&e.nodeName&&e.nodeName.toLowerCase();return t&&("input"===t&&("text"===e.type||"search"===e.type||"tel"===e.type||"url"===e.type||"password"===e.type)||"textarea"===t||"true"===e.contentEditable)}var Qn=null,Xn=null;function Jn(e,t){switch(e){case"button":case"input":case"select":case"textarea":return!!t.autoFocus}return!1}function Zn(e,t){return"textarea"===e||"option"===e||"noscript"===e||"string"==typeof t.children||"number"==typeof t.children||"object"==typeof t.dangerouslySetInnerHTML&&null!==t.dangerouslySetInnerHTML&&null!=t.dangerouslySetInnerHTML.__html}var ei="function"==typeof setTimeout?setTimeout:void 0,ti="function"==typeof clearTimeout?clearTimeout:void 0;function ni(e){for(;null!=e;e=e.nextSibling){var t=e.nodeType;if(1===t||3===t)break}return e}function ii(e){e=e.previousSibling;for(var t=0;e;){if(8===e.nodeType){var n=e.data;if("$"===n||"$!"===n||"$?"===n){if(0===t)return e;t--}else"/$"===n&&t++}e=e.previousSibling}return null}var ri=Math.random().toString(36).slice(2),oi="__reactInternalInstance$"+ri,si="__reactEventHandlers$"+ri,ai="__reactContainere$"+ri;function li(e){var t=e[oi];if(t)return t;for(var n=e.parentNode;n;){if(t=n[ai]||n[oi]){if(n=t.alternate,null!==t.child||null!==n&&null!==n.child)for(e=ii(e);null!==e;){if(n=e[oi])return n;e=ii(e)}return t}n=(e=n).parentNode}return null}function ui(e){return!(e=e[oi]||e[ai])||5!==e.tag&&6!==e.tag&&13!==e.tag&&3!==e.tag?null:e}function ci(e){if(5===e.tag||6===e.tag)return e.stateNode;throw Error(s(33))}function hi(e){return e[si]||null}var fi=null,di=null,pi=null;function gi(){if(pi)return pi;var e,t,n=di,i=n.length,r="value"in fi?fi.value:fi.textContent,o=r.length;for(e=0;e<i&&n[e]===r[e];e++);var s=i-e;for(t=1;t<=s&&n[i-t]===r[o-t];t++);return pi=r.slice(e,1<t?1-t:void 0)}var mi=Mt.extend({data:null}),vi=Mt.extend({data:null}),bi=[9,13,27,32],yi=Z&&"CompositionEvent"in window,wi=null;Z&&"documentMode"in document&&(wi=document.documentMode);var Ai=Z&&"TextEvent"in window&&!wi,Ci=Z&&(!yi||wi&&8<wi&&11>=wi),Ei=String.fromCharCode(32),Fi={beforeInput:{phasedRegistrationNames:{bubbled:"onBeforeInput",captured:"onBeforeInputCapture"},dependencies:["compositionend","keypress","textInput","paste"]},compositionEnd:{phasedRegistrationNames:{bubbled:"onCompositionEnd",captured:"onCompositionEndCapture"},dependencies:"blur compositionend keydown keypress keyup mousedown".split(" ")},compositionStart:{phasedRegistrationNames:{bubbled:"onCompositionStart",captured:"onCompositionStartCapture"},dependencies:"blur compositionstart keydown keypress keyup mousedown".split(" ")},compositionUpdate:{phasedRegistrationNames:{bubbled:"onCompositionUpdate",captured:"onCompositionUpdateCapture"},dependencies:"blur compositionupdate keydown keypress keyup mousedown".split(" ")}},xi=!1;function ki(e,t){switch(e){case"keyup":return-1!==bi.indexOf(t.keyCode);case"keydown":return 229!==t.keyCode;case"keypress":case"mousedown":case"blur":return!0;default:return!1}}function Si(e){return"object"==typeof(e=e.detail)&&"data"in e?e.data:null}var _i=!1;var Di={eventTypes:Fi,extractEvents:function(e,t,n,i){var r;if(yi)e:{switch(e){case"compositionstart":var o=Fi.compositionStart;break e;case"compositionend":o=Fi.compositionEnd;break e;case"compositionupdate":o=Fi.compositionUpdate;break e}o=void 0}else _i?ki(e,n)&&(o=Fi.compositionEnd):"keydown"===e&&229===n.keyCode&&(o=Fi.compositionStart);return o?(Ci&&"ko"!==n.locale&&(_i||o!==Fi.compositionStart?o===Fi.compositionEnd&&_i&&(r=gi()):(di="value"in(fi=i)?fi.value:fi.textContent,_i=!0)),o=mi.getPooled(o,t,n,i),r?o.data=r:null!==(r=Si(n))&&(o.data=r),Tt(o),r=o):r=null,(e=Ai?function(e,t){switch(e){case"compositionend":return Si(t);case"keypress":return 32!==t.which?null:(xi=!0,Ei);case"textInput":return(e=t.data)===Ei&&xi?null:e;default:return null}}(e,n):function(e,t){if(_i)return"compositionend"===e||!yi&&ki(e,t)?(e=gi(),pi=di=fi=null,_i=!1,e):null;switch(e){case"paste":return null;case"keypress":if(!(t.ctrlKey||t.altKey||t.metaKey)||t.ctrlKey&&t.altKey){if(t.char&&1<t.char.length)return t.char;if(t.which)return String.fromCharCode(t.which)}return null;case"compositionend":return Ci&&"ko"!==t.locale?null:t.data;default:return null}}(e,n))?((t=vi.getPooled(Fi.beforeInput,t,n,i)).data=e,Tt(t)):t=null,null===r?t:null===t?r:[r,t]}},$i={color:!0,date:!0,datetime:!0,"datetime-local":!0,email:!0,month:!0,number:!0,password:!0,range:!0,search:!0,tel:!0,text:!0,time:!0,url:!0,week:!0};function Bi(e){var t=e&&e.nodeName&&e.nodeName.toLowerCase();return"input"===t?!!$i[e.type]:"textarea"===t}var Li={change:{phasedRegistrationNames:{bubbled:"onChange",captured:"onChangeCapture"},dependencies:"blur change click focus input keydown keyup selectionchange".split(" ")}};function Ti(e,t,n){return(e=Mt.getPooled(Li.change,e,t,n)).type="change",re(n),Tt(e),e}var Ri=null,Oi=null;function Mi(e){$(e)}function Pi(e){if(xe(ci(e)))return e}function Ii(e,t){if("change"===e)return t}var Ni=!1;function Wi(){Ri&&(Ri.detachEvent("onpropertychange",zi),Oi=Ri=null)}function zi(e){if("value"===e.propertyName&&Pi(Oi))if(e=Ti(Oi,e,St(e)),ce)$(e);else{ce=!0;try{se(Mi,e)}finally{ce=!1,fe()}}}function ji(e,t,n){"focus"===e?(Wi(),Oi=n,(Ri=t).attachEvent("onpropertychange",zi)):"blur"===e&&Wi()}function Hi(e){if("selectionchange"===e||"keyup"===e||"keydown"===e)return Pi(Oi)}function Vi(e,t){if("click"===e)return Pi(t)}function Ui(e,t){if("input"===e||"change"===e)return Pi(t)}Z&&(Ni=Ln("input")&&(!document.documentMode||9<document.documentMode));var Ki,Gi={eventTypes:Li,_isInputEventSupported:Ni,extractEvents:function(e,t,n,i){var r=t?ci(t):window,o=r.nodeName&&r.nodeName.toLowerCase();if("select"===o||"input"===o&&"file"===r.type)var s=Ii;else if(Bi(r))if(Ni)s=Ui;else{s=Hi;var a=ji}else(o=r.nodeName)&&"input"===o.toLowerCase()&&("checkbox"===r.type||"radio"===r.type)&&(s=Vi);if(s&&(s=s(e,t)))return Ti(s,n,i);a&&a(e,r,t),"blur"===e&&(e=r._wrapperState)&&e.controlled&&"number"===r.type&&Be(r,"number",r.value)}},qi={mouseEnter:{registrationName:"onMouseEnter",dependencies:["mouseout","mouseover"]},mouseLeave:{registrationName:"onMouseLeave",dependencies:["mouseout","mouseover"]},pointerEnter:{registrationName:"onPointerEnter",dependencies:["pointerout","pointerover"]},pointerLeave:{registrationName:"onPointerLeave",dependencies:["pointerout","pointerover"]}},Yi={eventTypes:qi,extractEvents:function(e,t,n,i,r){var o="mouseover"===e||"pointerover"===e,s="mouseout"===e||"pointerout"===e;if(o&&0==(32&r)&&(n.relatedTarget||n.fromElement)||!s&&!o)return null;if(r=i.window===i?i:(r=i.ownerDocument)?r.defaultView||r.parentWindow:window,s?(s=t,null!==(t=(t=n.relatedTarget||n.toElement)?li(t):null)&&(t!==(o=tt(t))||5!==t.tag&&6!==t.tag)&&(t=null)):s=null,s===t)return null;if("mouseout"===e||"mouseover"===e)var a=tn,l=qi.mouseLeave,u=qi.mouseEnter,c="mouse";else"pointerout"!==e&&"pointerover"!==e||(a=nn,l=qi.pointerLeave,u=qi.pointerEnter,c="pointer");if(e=null==s?r:ci(s),r=null==t?r:ci(t),(l=a.getPooled(l,s,n,i)).type=c+"leave",l.target=e,l.relatedTarget=r,(i=a.getPooled(u,t,n,i)).type=c+"enter",i.target=r,i.relatedTarget=e,c=t,(a=s)&&c)e:{for(e=c,s=0,t=u=a;t;t=_t(t))s++;for(t=0,r=e;r;r=_t(r))t++;for(;0<s-t;)u=_t(u),s--;for(;0<t-s;)e=_t(e),t--;for(;s--;){if(u===e||u===e.alternate)break e;u=_t(u),e=_t(e)}u=null}else u=null;for(e=u,u=[];a&&a!==e&&(null===(s=a.alternate)||s!==e);)u.push(a),a=_t(a);for(a=[];c&&c!==e&&(null===(s=c.alternate)||s!==e);)a.push(c),c=_t(c);for(c=0;c<u.length;c++)Bt(u[c],"bubbled",l);for(c=a.length;0<c--;)Bt(a[c],"captured",i);return n===Ki?(Ki=null,[l]):(Ki=n,[l,i])}};var Qi="function"==typeof Object.is?Object.is:function(e,t){return e===t&&(0!==e||1/e==1/t)||e!=e&&t!=t},Xi=Object.prototype.hasOwnProperty;function Ji(e,t){if(Qi(e,t))return!0;if("object"!=typeof e||null===e||"object"!=typeof t||null===t)return!1;var n=Object.keys(e),i=Object.keys(t);if(n.length!==i.length)return!1;for(i=0;i<n.length;i++)if(!Xi.call(t,n[i])||!Qi(e[n[i]],t[n[i]]))return!1;return!0}var Zi=Z&&"documentMode"in document&&11>=document.documentMode,er={select:{phasedRegistrationNames:{bubbled:"onSelect",captured:"onSelectCapture"},dependencies:"blur contextmenu dragend focus keydown keyup mousedown mouseup selectionchange".split(" ")}},tr=null,nr=null,ir=null,rr=!1;function or(e,t){var n=t.window===t?t.document:9===t.nodeType?t:t.ownerDocument;return rr||null==tr||tr!==Un(n)?null:("selectionStart"in(n=tr)&&Yn(n)?n={start:n.selectionStart,end:n.selectionEnd}:n={anchorNode:(n=(n.ownerDocument&&n.ownerDocument.defaultView||window).getSelection()).anchorNode,anchorOffset:n.anchorOffset,focusNode:n.focusNode,focusOffset:n.focusOffset},ir&&Ji(ir,n)?null:(ir=n,(e=Mt.getPooled(er.select,nr,e,t)).type="select",e.target=tr,Tt(e),e))}var sr={eventTypes:er,extractEvents:function(e,t,n,i){var r,o=i.window===i?i.document:9===i.nodeType?i:i.ownerDocument;if(!(r=!o)){e:{o=Rn(o),r=p.onSelect;for(var s=0;s<r.length;s++)if(!o.has(r[s])){o=!1;break e}o=!0}r=!o}if(r)return null;switch(o=t?ci(t):window,e){case"focus":(Bi(o)||"true"===o.contentEditable)&&(tr=o,nr=t,ir=null);break;case"blur":ir=nr=tr=null;break;case"mousedown":rr=!0;break;case"contextmenu":case"mouseup":case"dragend":return rr=!1,or(n,i);case"selectionchange":if(Zi)break;case"keydown":case"keyup":return or(n,i)}return null}};B.injectEventPluginOrder("ResponderEventPlugin SimpleEventPlugin EnterLeaveEventPlugin ChangeEventPlugin SelectEventPlugin BeforeInputEventPlugin".split(" ")),C=hi,E=ui,F=ci,B.injectEventPluginsByName({SimpleEventPlugin:bn,EnterLeaveEventPlugin:Yi,ChangeEventPlugin:Gi,SelectEventPlugin:sr,BeforeInputEventPlugin:Di}),new Set;var ar=[],lr=-1;function ur(e){0>lr||(e.current=ar[lr],ar[lr]=null,lr--)}function cr(e,t){lr++,ar[lr]=e.current,e.current=t}var hr={},fr={current:hr},dr={current:!1},pr=hr;function gr(e,t){var n=e.type.contextTypes;if(!n)return hr;var i=e.stateNode;if(i&&i.__reactInternalMemoizedUnmaskedChildContext===t)return i.__reactInternalMemoizedMaskedChildContext;var r,o={};for(r in n)o[r]=t[r];return i&&((e=e.stateNode).__reactInternalMemoizedUnmaskedChildContext=t,e.__reactInternalMemoizedMaskedChildContext=o),o}function mr(e){return null!=(e=e.childContextTypes)}function vr(e){ur(dr),ur(fr)}function br(e){ur(dr),ur(fr)}function yr(e,t,n){if(fr.current!==hr)throw Error(s(168));cr(fr,t),cr(dr,n)}function wr(e,t,n){var i=e.stateNode;if(e=t.childContextTypes,"function"!=typeof i.getChildContext)return n;for(var o in i=i.getChildContext())if(!(o in e))throw Error(s(108,X(t)||"Unknown",o));return r({},n,{},i)}function Ar(e){var t=e.stateNode;return t=t&&t.__reactInternalMemoizedMergedChildContext||hr,pr=fr.current,cr(fr,t),cr(dr,dr.current),!0}function Cr(e,t,n){var i=e.stateNode;if(!i)throw Error(s(169));n?(t=wr(e,t,pr),i.__reactInternalMemoizedMergedChildContext=t,ur(dr),ur(fr),cr(fr,t)):ur(dr),cr(dr,n)}var Er=o.unstable_runWithPriority,Fr=o.unstable_scheduleCallback,xr=o.unstable_cancelCallback,kr=o.unstable_shouldYield,Sr=o.unstable_requestPaint,_r=o.unstable_now,Dr=o.unstable_getCurrentPriorityLevel,$r=o.unstable_ImmediatePriority,Br=o.unstable_UserBlockingPriority,Lr=o.unstable_NormalPriority,Tr=o.unstable_LowPriority,Rr=o.unstable_IdlePriority,Or={},Mr=void 0!==Sr?Sr:function(){},Pr=null,Ir=null,Nr=!1,Wr=_r(),zr=1e4>Wr?_r:function(){return _r()-Wr};function jr(){switch(Dr()){case $r:return 99;case Br:return 98;case Lr:return 97;case Tr:return 96;case Rr:return 95;default:throw Error(s(332))}}function Hr(e){switch(e){case 99:return $r;case 98:return Br;case 97:return Lr;case 96:return Tr;case 95:return Rr;default:throw Error(s(332))}}function Vr(e,t){return e=Hr(e),Er(e,t)}function Ur(e,t,n){return e=Hr(e),Fr(e,t,n)}function Kr(e){return null===Pr?(Pr=[e],Ir=Fr($r,qr)):Pr.push(e),Or}function Gr(){if(null!==Ir){var e=Ir;Ir=null,xr(e)}qr()}function qr(){if(!Nr&&null!==Pr){Nr=!0;var e=0;try{var t=Pr;Vr(99,(function(){for(;e<t.length;e++){var n=t[e];do{n=n(!0)}while(null!==n)}})),Pr=null}catch(t){throw null!==Pr&&(Pr=Pr.slice(e+1)),Fr($r,Gr),t}finally{Nr=!1}}}var Yr=3;function Qr(e,t,n){return 1073741821-(1+((1073741821-e+t/10)/(n/=10)|0))*n}function Xr(e,t){if(e&&e.defaultProps)for(var n in t=r({},t),e=e.defaultProps)void 0===t[n]&&(t[n]=e[n]);return t}var Jr={current:null},Zr=null,eo=null,to=null;function no(){to=eo=Zr=null}function io(e,t){var n=e.type._context;cr(Jr,n._currentValue),n._currentValue=t}function ro(e){var t=Jr.current;ur(Jr),e.type._context._currentValue=t}function oo(e,t){for(;null!==e;){var n=e.alternate;if(e.childExpirationTime<t)e.childExpirationTime=t,null!==n&&n.childExpirationTime<t&&(n.childExpirationTime=t);else{if(!(null!==n&&n.childExpirationTime<t))break;n.childExpirationTime=t}e=e.return}}function so(e,t){Zr=e,to=eo=null,null!==(e=e.dependencies)&&null!==e.firstContext&&(e.expirationTime>=t&&(zs=!0),e.firstContext=null)}function ao(e,t){if(to!==e&&!1!==t&&0!==t)if("number"==typeof t&&1073741823!==t||(to=e,t=1073741823),t={context:e,observedBits:t,next:null},null===eo){if(null===Zr)throw Error(s(308));eo=t,Zr.dependencies={expirationTime:0,firstContext:t,responders:null}}else eo=eo.next=t;return e._currentValue}var lo=!1;function uo(e){return{baseState:e,firstUpdate:null,lastUpdate:null,firstCapturedUpdate:null,lastCapturedUpdate:null,firstEffect:null,lastEffect:null,firstCapturedEffect:null,lastCapturedEffect:null}}function co(e){return{baseState:e.baseState,firstUpdate:e.firstUpdate,lastUpdate:e.lastUpdate,firstCapturedUpdate:null,lastCapturedUpdate:null,firstEffect:null,lastEffect:null,firstCapturedEffect:null,lastCapturedEffect:null}}function ho(e,t){return{expirationTime:e,suspenseConfig:t,tag:0,payload:null,callback:null,next:null,nextEffect:null}}function fo(e,t){null===e.lastUpdate?e.firstUpdate=e.lastUpdate=t:(e.lastUpdate.next=t,e.lastUpdate=t)}function po(e,t){var n=e.alternate;if(null===n){var i=e.updateQueue,r=null;null===i&&(i=e.updateQueue=uo(e.memoizedState))}else i=e.updateQueue,r=n.updateQueue,null===i?null===r?(i=e.updateQueue=uo(e.memoizedState),r=n.updateQueue=uo(n.memoizedState)):i=e.updateQueue=co(r):null===r&&(r=n.updateQueue=co(i));null===r||i===r?fo(i,t):null===i.lastUpdate||null===r.lastUpdate?(fo(i,t),fo(r,t)):(fo(i,t),r.lastUpdate=t)}function go(e,t){var n=e.updateQueue;null===(n=null===n?e.updateQueue=uo(e.memoizedState):mo(e,n)).lastCapturedUpdate?n.firstCapturedUpdate=n.lastCapturedUpdate=t:(n.lastCapturedUpdate.next=t,n.lastCapturedUpdate=t)}function mo(e,t){var n=e.alternate;return null!==n&&t===n.updateQueue&&(t=e.updateQueue=co(t)),t}function vo(e,t,n,i,o,s){switch(n.tag){case 1:return"function"==typeof(e=n.payload)?e.call(s,i,o):e;case 3:e.effectTag=-4097&e.effectTag|64;case 0:if(null==(o="function"==typeof(e=n.payload)?e.call(s,i,o):e))break;return r({},i,o);case 2:lo=!0}return i}function bo(e,t,n,i,r){lo=!1;for(var o=(t=mo(e,t)).baseState,s=null,a=0,l=t.firstUpdate,u=o;null!==l;){var c=l.expirationTime;c<r?(null===s&&(s=l,o=u),a<c&&(a=c)):(hl(c,l.suspenseConfig),u=vo(e,0,l,u,n,i),null!==l.callback&&(e.effectTag|=32,l.nextEffect=null,null===t.lastEffect?t.firstEffect=t.lastEffect=l:(t.lastEffect.nextEffect=l,t.lastEffect=l))),l=l.next}for(c=null,l=t.firstCapturedUpdate;null!==l;){var h=l.expirationTime;h<r?(null===c&&(c=l,null===s&&(o=u)),a<h&&(a=h)):(u=vo(e,0,l,u,n,i),null!==l.callback&&(e.effectTag|=32,l.nextEffect=null,null===t.lastCapturedEffect?t.firstCapturedEffect=t.lastCapturedEffect=l:(t.lastCapturedEffect.nextEffect=l,t.lastCapturedEffect=l))),l=l.next}null===s&&(t.lastUpdate=null),null===c?t.lastCapturedUpdate=null:e.effectTag|=32,null===s&&null===c&&(o=u),t.baseState=o,t.firstUpdate=s,t.firstCapturedUpdate=c,fl(a),e.expirationTime=a,e.memoizedState=u}function yo(e,t,n){null!==t.firstCapturedUpdate&&(null!==t.lastUpdate&&(t.lastUpdate.next=t.firstCapturedUpdate,t.lastUpdate=t.lastCapturedUpdate),t.firstCapturedUpdate=t.lastCapturedUpdate=null),wo(t.firstEffect,n),t.firstEffect=t.lastEffect=null,wo(t.firstCapturedEffect,n),t.firstCapturedEffect=t.lastCapturedEffect=null}function wo(e,t){for(;null!==e;){var n=e.callback;if(null!==n){e.callback=null;var i=t;if("function"!=typeof n)throw Error(s(191,n));n.call(i)}e=e.nextEffect}}var Ao=T.ReactCurrentBatchConfig,Co=(new i.Component).refs;function Eo(e,t,n,i){n=null==(n=n(i,t=e.memoizedState))?t:r({},t,n),e.memoizedState=n,null!==(i=e.updateQueue)&&0===e.expirationTime&&(i.baseState=n)}var Fo={isMounted:function(e){return!!(e=e._reactInternalFiber)&&tt(e)===e},enqueueSetState:function(e,t,n){e=e._reactInternalFiber;var i=Ja(),r=Ao.suspense;(r=ho(i=Za(i,e,r),r)).payload=t,null!=n&&(r.callback=n),po(e,r),el(e,i)},enqueueReplaceState:function(e,t,n){e=e._reactInternalFiber;var i=Ja(),r=Ao.suspense;(r=ho(i=Za(i,e,r),r)).tag=1,r.payload=t,null!=n&&(r.callback=n),po(e,r),el(e,i)},enqueueForceUpdate:function(e,t){e=e._reactInternalFiber;var n=Ja(),i=Ao.suspense;(i=ho(n=Za(n,e,i),i)).tag=2,null!=t&&(i.callback=t),po(e,i),el(e,n)}};function xo(e,t,n,i,r,o,s){return"function"==typeof(e=e.stateNode).shouldComponentUpdate?e.shouldComponentUpdate(i,o,s):!t.prototype||!t.prototype.isPureReactComponent||(!Ji(n,i)||!Ji(r,o))}function ko(e,t,n){var i=!1,r=hr,o=t.contextType;return"object"==typeof o&&null!==o?o=ao(o):(r=mr(t)?pr:fr.current,o=(i=null!=(i=t.contextTypes))?gr(e,r):hr),t=new t(n,o),e.memoizedState=null!==t.state&&void 0!==t.state?t.state:null,t.updater=Fo,e.stateNode=t,t._reactInternalFiber=e,i&&((e=e.stateNode).__reactInternalMemoizedUnmaskedChildContext=r,e.__reactInternalMemoizedMaskedChildContext=o),t}function So(e,t,n,i){e=t.state,"function"==typeof t.componentWillReceiveProps&&t.componentWillReceiveProps(n,i),"function"==typeof t.UNSAFE_componentWillReceiveProps&&t.UNSAFE_componentWillReceiveProps(n,i),t.state!==e&&Fo.enqueueReplaceState(t,t.state,null)}function _o(e,t,n,i){var r=e.stateNode;r.props=n,r.state=e.memoizedState,r.refs=Co;var o=t.contextType;"object"==typeof o&&null!==o?r.context=ao(o):(o=mr(t)?pr:fr.current,r.context=gr(e,o)),null!==(o=e.updateQueue)&&(bo(e,o,n,r,i),r.state=e.memoizedState),"function"==typeof(o=t.getDerivedStateFromProps)&&(Eo(e,t,o,n),r.state=e.memoizedState),"function"==typeof t.getDerivedStateFromProps||"function"==typeof r.getSnapshotBeforeUpdate||"function"!=typeof r.UNSAFE_componentWillMount&&"function"!=typeof r.componentWillMount||(t=r.state,"function"==typeof r.componentWillMount&&r.componentWillMount(),"function"==typeof r.UNSAFE_componentWillMount&&r.UNSAFE_componentWillMount(),t!==r.state&&Fo.enqueueReplaceState(r,r.state,null),null!==(o=e.updateQueue)&&(bo(e,o,n,r,i),r.state=e.memoizedState)),"function"==typeof r.componentDidMount&&(e.effectTag|=4)}var Do=Array.isArray;function $o(e,t,n){if(null!==(e=n.ref)&&"function"!=typeof e&&"object"!=typeof e){if(n._owner){if(n=n._owner){if(1!==n.tag)throw Error(s(309));var i=n.stateNode}if(!i)throw Error(s(147,e));var r=""+e;return null!==t&&null!==t.ref&&"function"==typeof t.ref&&t.ref._stringRef===r?t.ref:((t=function(e){var t=i.refs;t===Co&&(t=i.refs={}),null===e?delete t[r]:t[r]=e})._stringRef=r,t)}if("string"!=typeof e)throw Error(s(284));if(!n._owner)throw Error(s(290,e))}return e}function Bo(e,t){if("textarea"!==e.type)throw Error(s(31,"[object Object]"===Object.prototype.toString.call(t)?"object with keys {"+Object.keys(t).join(", ")+"}":t,""))}function Lo(e){function t(t,n){if(e){var i=t.lastEffect;null!==i?(i.nextEffect=n,t.lastEffect=n):t.firstEffect=t.lastEffect=n,n.nextEffect=null,n.effectTag=8}}function n(n,i){if(!e)return null;for(;null!==i;)t(n,i),i=i.sibling;return null}function i(e,t){for(e=new Map;null!==t;)null!==t.key?e.set(t.key,t):e.set(t.index,t),t=t.sibling;return e}function r(e,t,n){return(e=Ll(e,t)).index=0,e.sibling=null,e}function o(t,n,i){return t.index=i,e?null!==(i=t.alternate)?(i=i.index)<n?(t.effectTag=2,n):i:(t.effectTag=2,n):n}function a(t){return e&&null===t.alternate&&(t.effectTag=2),t}function l(e,t,n,i){return null===t||6!==t.tag?((t=Ol(n,e.mode,i)).return=e,t):((t=r(t,n)).return=e,t)}function u(e,t,n,i){return null!==t&&t.elementType===n.type?((i=r(t,n.props)).ref=$o(e,t,n),i.return=e,i):((i=Tl(n.type,n.key,n.props,null,e.mode,i)).ref=$o(e,t,n),i.return=e,i)}function c(e,t,n,i){return null===t||4!==t.tag||t.stateNode.containerInfo!==n.containerInfo||t.stateNode.implementation!==n.implementation?((t=Ml(n,e.mode,i)).return=e,t):((t=r(t,n.children||[])).return=e,t)}function h(e,t,n,i,o){return null===t||7!==t.tag?((t=Rl(n,e.mode,i,o)).return=e,t):((t=r(t,n)).return=e,t)}function f(e,t,n){if("string"==typeof t||"number"==typeof t)return(t=Ol(""+t,e.mode,n)).return=e,t;if("object"==typeof t&&null!==t){switch(t.$$typeof){case M:return(n=Tl(t.type,t.key,t.props,null,e.mode,n)).ref=$o(e,null,t),n.return=e,n;case P:return(t=Ml(t,e.mode,n)).return=e,t}if(Do(t)||Q(t))return(t=Rl(t,e.mode,n,null)).return=e,t;Bo(e,t)}return null}function d(e,t,n,i){var r=null!==t?t.key:null;if("string"==typeof n||"number"==typeof n)return null!==r?null:l(e,t,""+n,i);if("object"==typeof n&&null!==n){switch(n.$$typeof){case M:return n.key===r?n.type===I?h(e,t,n.props.children,i,r):u(e,t,n,i):null;case P:return n.key===r?c(e,t,n,i):null}if(Do(n)||Q(n))return null!==r?null:h(e,t,n,i,null);Bo(e,n)}return null}function p(e,t,n,i,r){if("string"==typeof i||"number"==typeof i)return l(t,e=e.get(n)||null,""+i,r);if("object"==typeof i&&null!==i){switch(i.$$typeof){case M:return e=e.get(null===i.key?n:i.key)||null,i.type===I?h(t,e,i.props.children,r,i.key):u(t,e,i,r);case P:return c(t,e=e.get(null===i.key?n:i.key)||null,i,r)}if(Do(i)||Q(i))return h(t,e=e.get(n)||null,i,r,null);Bo(t,i)}return null}function g(r,s,a,l){for(var u=null,c=null,h=s,g=s=0,m=null;null!==h&&g<a.length;g++){h.index>g?(m=h,h=null):m=h.sibling;var v=d(r,h,a[g],l);if(null===v){null===h&&(h=m);break}e&&h&&null===v.alternate&&t(r,h),s=o(v,s,g),null===c?u=v:c.sibling=v,c=v,h=m}if(g===a.length)return n(r,h),u;if(null===h){for(;g<a.length;g++)null!==(h=f(r,a[g],l))&&(s=o(h,s,g),null===c?u=h:c.sibling=h,c=h);return u}for(h=i(r,h);g<a.length;g++)null!==(m=p(h,r,g,a[g],l))&&(e&&null!==m.alternate&&h.delete(null===m.key?g:m.key),s=o(m,s,g),null===c?u=m:c.sibling=m,c=m);return e&&h.forEach((function(e){return t(r,e)})),u}function m(r,a,l,u){var c=Q(l);if("function"!=typeof c)throw Error(s(150));if(null==(l=c.call(l)))throw Error(s(151));for(var h=c=null,g=a,m=a=0,v=null,b=l.next();null!==g&&!b.done;m++,b=l.next()){g.index>m?(v=g,g=null):v=g.sibling;var y=d(r,g,b.value,u);if(null===y){null===g&&(g=v);break}e&&g&&null===y.alternate&&t(r,g),a=o(y,a,m),null===h?c=y:h.sibling=y,h=y,g=v}if(b.done)return n(r,g),c;if(null===g){for(;!b.done;m++,b=l.next())null!==(b=f(r,b.value,u))&&(a=o(b,a,m),null===h?c=b:h.sibling=b,h=b);return c}for(g=i(r,g);!b.done;m++,b=l.next())null!==(b=p(g,r,m,b.value,u))&&(e&&null!==b.alternate&&g.delete(null===b.key?m:b.key),a=o(b,a,m),null===h?c=b:h.sibling=b,h=b);return e&&g.forEach((function(e){return t(r,e)})),c}return function(e,i,o,l){var u="object"==typeof o&&null!==o&&o.type===I&&null===o.key;u&&(o=o.props.children);var c="object"==typeof o&&null!==o;if(c)switch(o.$$typeof){case M:e:{for(c=o.key,u=i;null!==u;){if(u.key===c){if(7===u.tag?o.type===I:u.elementType===o.type){n(e,u.sibling),(i=r(u,o.type===I?o.props.children:o.props)).ref=$o(e,u,o),i.return=e,e=i;break e}n(e,u);break}t(e,u),u=u.sibling}o.type===I?((i=Rl(o.props.children,e.mode,l,o.key)).return=e,e=i):((l=Tl(o.type,o.key,o.props,null,e.mode,l)).ref=$o(e,i,o),l.return=e,e=l)}return a(e);case P:e:{for(u=o.key;null!==i;){if(i.key===u){if(4===i.tag&&i.stateNode.containerInfo===o.containerInfo&&i.stateNode.implementation===o.implementation){n(e,i.sibling),(i=r(i,o.children||[])).return=e,e=i;break e}n(e,i);break}t(e,i),i=i.sibling}(i=Ml(o,e.mode,l)).return=e,e=i}return a(e)}if("string"==typeof o||"number"==typeof o)return o=""+o,null!==i&&6===i.tag?(n(e,i.sibling),(i=r(i,o)).return=e,e=i):(n(e,i),(i=Ol(o,e.mode,l)).return=e,e=i),a(e);if(Do(o))return g(e,i,o,l);if(Q(o))return m(e,i,o,l);if(c&&Bo(e,o),void 0===o&&!u)switch(e.tag){case 1:case 0:throw e=e.type,Error(s(152,e.displayName||e.name||"Component"))}return n(e,i)}}var To=Lo(!0),Ro=Lo(!1),Oo={},Mo={current:Oo},Po={current:Oo},Io={current:Oo};function No(e){if(e===Oo)throw Error(s(174));return e}function Wo(e,t){cr(Io,t),cr(Po,e),cr(Mo,Oo);var n=t.nodeType;switch(n){case 9:case 11:t=(t=t.documentElement)?t.namespaceURI:ze(null,"");break;default:t=ze(t=(n=8===n?t.parentNode:t).namespaceURI||null,n=n.tagName)}ur(Mo),cr(Mo,t)}function zo(e){ur(Mo),ur(Po),ur(Io)}function jo(e){No(Io.current);var t=No(Mo.current),n=ze(t,e.type);t!==n&&(cr(Po,e),cr(Mo,n))}function Ho(e){Po.current===e&&(ur(Mo),ur(Po))}var Vo={current:0};function Uo(e){for(var t=e;null!==t;){if(13===t.tag){var n=t.memoizedState;if(null!==n&&(null===(n=n.dehydrated)||"$?"===n.data||"$!"===n.data))return t}else if(19===t.tag&&void 0!==t.memoizedProps.revealOrder){if(0!=(64&t.effectTag))return t}else if(null!==t.child){t.child.return=t,t=t.child;continue}if(t===e)break;for(;null===t.sibling;){if(null===t.return||t.return===e)return null;t=t.return}t.sibling.return=t.return,t=t.sibling}return null}function Ko(e,t){return{responder:e,props:t}}var Go=T.ReactCurrentDispatcher,qo=T.ReactCurrentBatchConfig,Yo=0,Qo=null,Xo=null,Jo=null,Zo=null,es=null,ts=null,ns=0,is=null,rs=0,os=!1,ss=null,as=0;function ls(){throw Error(s(321))}function us(e,t){if(null===t)return!1;for(var n=0;n<t.length&&n<e.length;n++)if(!Qi(e[n],t[n]))return!1;return!0}function cs(e,t,n,i,r,o){if(Yo=o,Qo=t,Jo=null!==e?e.memoizedState:null,Go.current=null===Jo?Ds:$s,t=n(i,r),os){do{os=!1,as+=1,Jo=null!==e?e.memoizedState:null,ts=Zo,is=es=Xo=null,Go.current=$s,t=n(i,r)}while(os);ss=null,as=0}if(Go.current=_s,(e=Qo).memoizedState=Zo,e.expirationTime=ns,e.updateQueue=is,e.effectTag|=rs,e=null!==Xo&&null!==Xo.next,Yo=0,ts=es=Zo=Jo=Xo=Qo=null,ns=0,is=null,rs=0,e)throw Error(s(300));return t}function hs(){Go.current=_s,Yo=0,ts=es=Zo=Jo=Xo=Qo=null,ns=0,is=null,rs=0,os=!1,ss=null,as=0}function fs(){var e={memoizedState:null,baseState:null,queue:null,baseUpdate:null,next:null};return null===es?Zo=es=e:es=es.next=e,es}function ds(){if(null!==ts)ts=(es=ts).next,Jo=null!==(Xo=Jo)?Xo.next:null;else{if(null===Jo)throw Error(s(310));var e={memoizedState:(Xo=Jo).memoizedState,baseState:Xo.baseState,queue:Xo.queue,baseUpdate:Xo.baseUpdate,next:null};es=null===es?Zo=e:es.next=e,Jo=Xo.next}return es}function ps(e,t){return"function"==typeof t?t(e):t}function gs(e){var t=ds(),n=t.queue;if(null===n)throw Error(s(311));if(n.lastRenderedReducer=e,0<as){var i=n.dispatch;if(null!==ss){var r=ss.get(n);if(void 0!==r){ss.delete(n);var o=t.memoizedState;do{o=e(o,r.action),r=r.next}while(null!==r);return Qi(o,t.memoizedState)||(zs=!0),t.memoizedState=o,t.baseUpdate===n.last&&(t.baseState=o),n.lastRenderedState=o,[o,i]}}return[t.memoizedState,i]}i=n.last;var a=t.baseUpdate;if(o=t.baseState,null!==a?(null!==i&&(i.next=null),i=a.next):i=null!==i?i.next:null,null!==i){var l=r=null,u=i,c=!1;do{var h=u.expirationTime;h<Yo?(c||(c=!0,l=a,r=o),h>ns&&fl(ns=h)):(hl(h,u.suspenseConfig),o=u.eagerReducer===e?u.eagerState:e(o,u.action)),a=u,u=u.next}while(null!==u&&u!==i);c||(l=a,r=o),Qi(o,t.memoizedState)||(zs=!0),t.memoizedState=o,t.baseUpdate=l,t.baseState=r,n.lastRenderedState=o}return[t.memoizedState,n.dispatch]}function ms(e){var t=fs();return"function"==typeof e&&(e=e()),t.memoizedState=t.baseState=e,e=(e=t.queue={last:null,dispatch:null,lastRenderedReducer:ps,lastRenderedState:e}).dispatch=Ss.bind(null,Qo,e),[t.memoizedState,e]}function vs(e){return gs(ps)}function bs(e,t,n,i){return e={tag:e,create:t,destroy:n,deps:i,next:null},null===is?(is={lastEffect:null}).lastEffect=e.next=e:null===(t=is.lastEffect)?is.lastEffect=e.next=e:(n=t.next,t.next=e,e.next=n,is.lastEffect=e),e}function ys(e,t,n,i){var r=fs();rs|=e,r.memoizedState=bs(t,n,void 0,void 0===i?null:i)}function ws(e,t,n,i){var r=ds();i=void 0===i?null:i;var o=void 0;if(null!==Xo){var s=Xo.memoizedState;if(o=s.destroy,null!==i&&us(i,s.deps))return void bs(0,n,o,i)}rs|=e,r.memoizedState=bs(t,n,o,i)}function As(e,t){return ys(516,192,e,t)}function Cs(e,t){return ws(516,192,e,t)}function Es(e,t){return"function"==typeof t?(e=e(),t(e),function(){t(null)}):null!=t?(e=e(),t.current=e,function(){t.current=null}):void 0}function Fs(){}function xs(e,t){return fs().memoizedState=[e,void 0===t?null:t],e}function ks(e,t){var n=ds();t=void 0===t?null:t;var i=n.memoizedState;return null!==i&&null!==t&&us(t,i[1])?i[0]:(n.memoizedState=[e,t],e)}function Ss(e,t,n){if(!(25>as))throw Error(s(301));var i=e.alternate;if(e===Qo||null!==i&&i===Qo)if(os=!0,e={expirationTime:Yo,suspenseConfig:null,action:n,eagerReducer:null,eagerState:null,next:null},null===ss&&(ss=new Map),void 0===(n=ss.get(t)))ss.set(t,e);else{for(t=n;null!==t.next;)t=t.next;t.next=e}else{var r=Ja(),o=Ao.suspense;o={expirationTime:r=Za(r,e,o),suspenseConfig:o,action:n,eagerReducer:null,eagerState:null,next:null};var a=t.last;if(null===a)o.next=o;else{var l=a.next;null!==l&&(o.next=l),a.next=o}if(t.last=o,0===e.expirationTime&&(null===i||0===i.expirationTime)&&null!==(i=t.lastRenderedReducer))try{var u=t.lastRenderedState,c=i(u,n);if(o.eagerReducer=i,o.eagerState=c,Qi(c,u))return}catch(e){}el(e,r)}}var _s={readContext:ao,useCallback:ls,useContext:ls,useEffect:ls,useImperativeHandle:ls,useLayoutEffect:ls,useMemo:ls,useReducer:ls,useRef:ls,useState:ls,useDebugValue:ls,useResponder:ls,useDeferredValue:ls,useTransition:ls},Ds={readContext:ao,useCallback:xs,useContext:ao,useEffect:As,useImperativeHandle:function(e,t,n){return n=null!=n?n.concat([e]):null,ys(4,36,Es.bind(null,t,e),n)},useLayoutEffect:function(e,t){return ys(4,36,e,t)},useMemo:function(e,t){var n=fs();return t=void 0===t?null:t,e=e(),n.memoizedState=[e,t],e},useReducer:function(e,t,n){var i=fs();return t=void 0!==n?n(t):t,i.memoizedState=i.baseState=t,e=(e=i.queue={last:null,dispatch:null,lastRenderedReducer:e,lastRenderedState:t}).dispatch=Ss.bind(null,Qo,e),[i.memoizedState,e]},useRef:function(e){return e={current:e},fs().memoizedState=e},useState:ms,useDebugValue:Fs,useResponder:Ko,useDeferredValue:function(e,t){var n=ms(e),i=n[0],r=n[1];return As((function(){o.unstable_next((function(){var n=qo.suspense;qo.suspense=void 0===t?null:t;try{r(e)}finally{qo.suspense=n}}))}),[e,t]),i},useTransition:function(e){var t=ms(!1),n=t[0],i=t[1];return[xs((function(t){i(!0),o.unstable_next((function(){var n=qo.suspense;qo.suspense=void 0===e?null:e;try{i(!1),t()}finally{qo.suspense=n}}))}),[e,n]),n]}},$s={readContext:ao,useCallback:ks,useContext:ao,useEffect:Cs,useImperativeHandle:function(e,t,n){return n=null!=n?n.concat([e]):null,ws(4,36,Es.bind(null,t,e),n)},useLayoutEffect:function(e,t){return ws(4,36,e,t)},useMemo:function(e,t){var n=ds();t=void 0===t?null:t;var i=n.memoizedState;return null!==i&&null!==t&&us(t,i[1])?i[0]:(e=e(),n.memoizedState=[e,t],e)},useReducer:gs,useRef:function(){return ds().memoizedState},useState:vs,useDebugValue:Fs,useResponder:Ko,useDeferredValue:function(e,t){var n=vs(),i=n[0],r=n[1];return Cs((function(){o.unstable_next((function(){var n=qo.suspense;qo.suspense=void 0===t?null:t;try{r(e)}finally{qo.suspense=n}}))}),[e,t]),i},useTransition:function(e){var t=vs(),n=t[0],i=t[1];return[ks((function(t){i(!0),o.unstable_next((function(){var n=qo.suspense;qo.suspense=void 0===e?null:e;try{i(!1),t()}finally{qo.suspense=n}}))}),[e,n]),n]}},Bs=null,Ls=null,Ts=!1;function Rs(e,t){var n=$l(5,null,null,0);n.elementType="DELETED",n.type="DELETED",n.stateNode=t,n.return=e,n.effectTag=8,null!==e.lastEffect?(e.lastEffect.nextEffect=n,e.lastEffect=n):e.firstEffect=e.lastEffect=n}function Os(e,t){switch(e.tag){case 5:var n=e.type;return null!==(t=1!==t.nodeType||n.toLowerCase()!==t.nodeName.toLowerCase()?null:t)&&(e.stateNode=t,!0);case 6:return null!==(t=""===e.pendingProps||3!==t.nodeType?null:t)&&(e.stateNode=t,!0);case 13:default:return!1}}function Ms(e){if(Ts){var t=Ls;if(t){var n=t;if(!Os(e,t)){if(!(t=ni(n.nextSibling))||!Os(e,t))return e.effectTag=-1025&e.effectTag|2,Ts=!1,void(Bs=e);Rs(Bs,n)}Bs=e,Ls=ni(t.firstChild)}else e.effectTag=-1025&e.effectTag|2,Ts=!1,Bs=e}}function Ps(e){for(e=e.return;null!==e&&5!==e.tag&&3!==e.tag&&13!==e.tag;)e=e.return;Bs=e}function Is(e){if(e!==Bs)return!1;if(!Ts)return Ps(e),Ts=!0,!1;var t=e.type;if(5!==e.tag||"head"!==t&&"body"!==t&&!Zn(t,e.memoizedProps))for(t=Ls;t;)Rs(e,t),t=ni(t.nextSibling);if(Ps(e),13===e.tag){if(!(e=null!==(e=e.memoizedState)?e.dehydrated:null))throw Error(s(317));e:{for(e=e.nextSibling,t=0;e;){if(8===e.nodeType){var n=e.data;if("/$"===n){if(0===t){Ls=ni(e.nextSibling);break e}t--}else"$"!==n&&"$!"!==n&&"$?"!==n||t++}e=e.nextSibling}Ls=null}}else Ls=Bs?ni(e.stateNode.nextSibling):null;return!0}function Ns(){Ls=Bs=null,Ts=!1}var Ws=T.ReactCurrentOwner,zs=!1;function js(e,t,n,i){t.child=null===e?Ro(t,null,n,i):To(t,e.child,n,i)}function Hs(e,t,n,i,r){n=n.render;var o=t.ref;return so(t,r),i=cs(e,t,n,i,o,r),null===e||zs?(t.effectTag|=1,js(e,t,i,r),t.child):(t.updateQueue=e.updateQueue,t.effectTag&=-517,e.expirationTime<=r&&(e.expirationTime=0),oa(e,t,r))}function Vs(e,t,n,i,r,o){if(null===e){var s=n.type;return"function"!=typeof s||Bl(s)||void 0!==s.defaultProps||null!==n.compare||void 0!==n.defaultProps?((e=Tl(n.type,null,i,null,t.mode,o)).ref=t.ref,e.return=t,t.child=e):(t.tag=15,t.type=s,Us(e,t,s,i,r,o))}return s=e.child,r<o&&(r=s.memoizedProps,(n=null!==(n=n.compare)?n:Ji)(r,i)&&e.ref===t.ref)?oa(e,t,o):(t.effectTag|=1,(e=Ll(s,i)).ref=t.ref,e.return=t,t.child=e)}function Us(e,t,n,i,r,o){return null!==e&&Ji(e.memoizedProps,i)&&e.ref===t.ref&&(zs=!1,r<o)?oa(e,t,o):Gs(e,t,n,i,o)}function Ks(e,t){var n=t.ref;(null===e&&null!==n||null!==e&&e.ref!==n)&&(t.effectTag|=128)}function Gs(e,t,n,i,r){var o=mr(n)?pr:fr.current;return o=gr(t,o),so(t,r),n=cs(e,t,n,i,o,r),null===e||zs?(t.effectTag|=1,js(e,t,n,r),t.child):(t.updateQueue=e.updateQueue,t.effectTag&=-517,e.expirationTime<=r&&(e.expirationTime=0),oa(e,t,r))}function qs(e,t,n,i,r){if(mr(n)){var o=!0;Ar(t)}else o=!1;if(so(t,r),null===t.stateNode)null!==e&&(e.alternate=null,t.alternate=null,t.effectTag|=2),ko(t,n,i),_o(t,n,i,r),i=!0;else if(null===e){var s=t.stateNode,a=t.memoizedProps;s.props=a;var l=s.context,u=n.contextType;"object"==typeof u&&null!==u?u=ao(u):u=gr(t,u=mr(n)?pr:fr.current);var c=n.getDerivedStateFromProps,h="function"==typeof c||"function"==typeof s.getSnapshotBeforeUpdate;h||"function"!=typeof s.UNSAFE_componentWillReceiveProps&&"function"!=typeof s.componentWillReceiveProps||(a!==i||l!==u)&&So(t,s,i,u),lo=!1;var f=t.memoizedState;l=s.state=f;var d=t.updateQueue;null!==d&&(bo(t,d,i,s,r),l=t.memoizedState),a!==i||f!==l||dr.current||lo?("function"==typeof c&&(Eo(t,n,c,i),l=t.memoizedState),(a=lo||xo(t,n,a,i,f,l,u))?(h||"function"!=typeof s.UNSAFE_componentWillMount&&"function"!=typeof s.componentWillMount||("function"==typeof s.componentWillMount&&s.componentWillMount(),"function"==typeof s.UNSAFE_componentWillMount&&s.UNSAFE_componentWillMount()),"function"==typeof s.componentDidMount&&(t.effectTag|=4)):("function"==typeof s.componentDidMount&&(t.effectTag|=4),t.memoizedProps=i,t.memoizedState=l),s.props=i,s.state=l,s.context=u,i=a):("function"==typeof s.componentDidMount&&(t.effectTag|=4),i=!1)}else s=t.stateNode,a=t.memoizedProps,s.props=t.type===t.elementType?a:Xr(t.type,a),l=s.context,"object"==typeof(u=n.contextType)&&null!==u?u=ao(u):u=gr(t,u=mr(n)?pr:fr.current),(h="function"==typeof(c=n.getDerivedStateFromProps)||"function"==typeof s.getSnapshotBeforeUpdate)||"function"!=typeof s.UNSAFE_componentWillReceiveProps&&"function"!=typeof s.componentWillReceiveProps||(a!==i||l!==u)&&So(t,s,i,u),lo=!1,l=t.memoizedState,f=s.state=l,null!==(d=t.updateQueue)&&(bo(t,d,i,s,r),f=t.memoizedState),a!==i||l!==f||dr.current||lo?("function"==typeof c&&(Eo(t,n,c,i),f=t.memoizedState),(c=lo||xo(t,n,a,i,l,f,u))?(h||"function"!=typeof s.UNSAFE_componentWillUpdate&&"function"!=typeof s.componentWillUpdate||("function"==typeof s.componentWillUpdate&&s.componentWillUpdate(i,f,u),"function"==typeof s.UNSAFE_componentWillUpdate&&s.UNSAFE_componentWillUpdate(i,f,u)),"function"==typeof s.componentDidUpdate&&(t.effectTag|=4),"function"==typeof s.getSnapshotBeforeUpdate&&(t.effectTag|=256)):("function"!=typeof s.componentDidUpdate||a===e.memoizedProps&&l===e.memoizedState||(t.effectTag|=4),"function"!=typeof s.getSnapshotBeforeUpdate||a===e.memoizedProps&&l===e.memoizedState||(t.effectTag|=256),t.memoizedProps=i,t.memoizedState=f),s.props=i,s.state=f,s.context=u,i=c):("function"!=typeof s.componentDidUpdate||a===e.memoizedProps&&l===e.memoizedState||(t.effectTag|=4),"function"!=typeof s.getSnapshotBeforeUpdate||a===e.memoizedProps&&l===e.memoizedState||(t.effectTag|=256),i=!1);return Ys(e,t,n,i,o,r)}function Ys(e,t,n,i,r,o){Ks(e,t);var s=0!=(64&t.effectTag);if(!i&&!s)return r&&Cr(t,n,!1),oa(e,t,o);i=t.stateNode,Ws.current=t;var a=s&&"function"!=typeof n.getDerivedStateFromError?null:i.render();return t.effectTag|=1,null!==e&&s?(t.child=To(t,e.child,null,o),t.child=To(t,null,a,o)):js(e,t,a,o),t.memoizedState=i.state,r&&Cr(t,n,!0),t.child}function Qs(e){var t=e.stateNode;t.pendingContext?yr(0,t.pendingContext,t.pendingContext!==t.context):t.context&&yr(0,t.context,!1),Wo(e,t.containerInfo)}var Xs,Js,Zs,ea={dehydrated:null,retryTime:0};function ta(e,t,n){var i,r=t.mode,o=t.pendingProps,s=Vo.current,a=!1;if((i=0!=(64&t.effectTag))||(i=0!=(2&s)&&(null===e||null!==e.memoizedState)),i?(a=!0,t.effectTag&=-65):null!==e&&null===e.memoizedState||void 0===o.fallback||!0===o.unstable_avoidThisFallback||(s|=1),cr(Vo,1&s),null===e){if(void 0!==o.fallback&&Ms(t),a){if(a=o.fallback,(o=Rl(null,r,0,null)).return=t,0==(2&t.mode))for(e=null!==t.memoizedState?t.child.child:t.child,o.child=e;null!==e;)e.return=o,e=e.sibling;return(n=Rl(a,r,n,null)).return=t,o.sibling=n,t.memoizedState=ea,t.child=o,n}return r=o.children,t.memoizedState=null,t.child=Ro(t,null,r,n)}if(null!==e.memoizedState){if(r=(e=e.child).sibling,a){if(o=o.fallback,(n=Ll(e,e.pendingProps)).return=t,0==(2&t.mode)&&(a=null!==t.memoizedState?t.child.child:t.child)!==e.child)for(n.child=a;null!==a;)a.return=n,a=a.sibling;return(r=Ll(r,o,r.expirationTime)).return=t,n.sibling=r,n.childExpirationTime=0,t.memoizedState=ea,t.child=n,r}return n=To(t,e.child,o.children,n),t.memoizedState=null,t.child=n}if(e=e.child,a){if(a=o.fallback,(o=Rl(null,r,0,null)).return=t,o.child=e,null!==e&&(e.return=o),0==(2&t.mode))for(e=null!==t.memoizedState?t.child.child:t.child,o.child=e;null!==e;)e.return=o,e=e.sibling;return(n=Rl(a,r,n,null)).return=t,o.sibling=n,n.effectTag|=2,o.childExpirationTime=0,t.memoizedState=ea,t.child=o,n}return t.memoizedState=null,t.child=To(t,e,o.children,n)}function na(e,t){e.expirationTime<t&&(e.expirationTime=t);var n=e.alternate;null!==n&&n.expirationTime<t&&(n.expirationTime=t),oo(e.return,t)}function ia(e,t,n,i,r,o){var s=e.memoizedState;null===s?e.memoizedState={isBackwards:t,rendering:null,last:i,tail:n,tailExpiration:0,tailMode:r,lastEffect:o}:(s.isBackwards=t,s.rendering=null,s.last=i,s.tail=n,s.tailExpiration=0,s.tailMode=r,s.lastEffect=o)}function ra(e,t,n){var i=t.pendingProps,r=i.revealOrder,o=i.tail;if(js(e,t,i.children,n),0!=(2&(i=Vo.current)))i=1&i|2,t.effectTag|=64;else{if(null!==e&&0!=(64&e.effectTag))e:for(e=t.child;null!==e;){if(13===e.tag)null!==e.memoizedState&&na(e,n);else if(19===e.tag)na(e,n);else if(null!==e.child){e.child.return=e,e=e.child;continue}if(e===t)break e;for(;null===e.sibling;){if(null===e.return||e.return===t)break e;e=e.return}e.sibling.return=e.return,e=e.sibling}i&=1}if(cr(Vo,i),0==(2&t.mode))t.memoizedState=null;else switch(r){case"forwards":for(n=t.child,r=null;null!==n;)null!==(e=n.alternate)&&null===Uo(e)&&(r=n),n=n.sibling;null===(n=r)?(r=t.child,t.child=null):(r=n.sibling,n.sibling=null),ia(t,!1,r,n,o,t.lastEffect);break;case"backwards":for(n=null,r=t.child,t.child=null;null!==r;){if(null!==(e=r.alternate)&&null===Uo(e)){t.child=r;break}e=r.sibling,r.sibling=n,n=r,r=e}ia(t,!0,n,null,o,t.lastEffect);break;case"together":ia(t,!1,null,null,void 0,t.lastEffect);break;default:t.memoizedState=null}return t.child}function oa(e,t,n){null!==e&&(t.dependencies=e.dependencies);var i=t.expirationTime;if(0!==i&&fl(i),t.childExpirationTime<n)return null;if(null!==e&&t.child!==e.child)throw Error(s(153));if(null!==t.child){for(n=Ll(e=t.child,e.pendingProps,e.expirationTime),t.child=n,n.return=t;null!==e.sibling;)e=e.sibling,(n=n.sibling=Ll(e,e.pendingProps,e.expirationTime)).return=t;n.sibling=null}return t.child}function sa(e){e.effectTag|=4}function aa(e,t){switch(e.tailMode){case"hidden":t=e.tail;for(var n=null;null!==t;)null!==t.alternate&&(n=t),t=t.sibling;null===n?e.tail=null:n.sibling=null;break;case"collapsed":n=e.tail;for(var i=null;null!==n;)null!==n.alternate&&(i=n),n=n.sibling;null===i?t||null===e.tail?e.tail=null:e.tail.sibling=null:i.sibling=null}}function la(e){switch(e.tag){case 1:mr(e.type)&&vr();var t=e.effectTag;return 4096&t?(e.effectTag=-4097&t|64,e):null;case 3:if(zo(),br(),0!=(64&(t=e.effectTag)))throw Error(s(285));return e.effectTag=-4097&t|64,e;case 5:return Ho(e),null;case 13:return ur(Vo),4096&(t=e.effectTag)?(e.effectTag=-4097&t|64,e):null;case 19:return ur(Vo),null;case 4:return zo(),null;case 10:return ro(e),null;default:return null}}function ua(e,t){return{value:e,source:t,stack:J(t)}}Xs=function(e,t){for(var n=t.child;null!==n;){if(5===n.tag||6===n.tag)e.appendChild(n.stateNode);else if(4!==n.tag&&null!==n.child){n.child.return=n,n=n.child;continue}if(n===t)break;for(;null===n.sibling;){if(null===n.return||n.return===t)return;n=n.return}n.sibling.return=n.return,n=n.sibling}},Js=function(e,t,n,i,o){var s=e.memoizedProps;if(s!==i){var a,l,u=t.stateNode;switch(No(Mo.current),e=null,n){case"input":s=ke(u,s),i=ke(u,i),e=[];break;case"option":s=Le(u,s),i=Le(u,i),e=[];break;case"select":s=r({},s,{value:void 0}),i=r({},i,{value:void 0}),e=[];break;case"textarea":s=Re(u,s),i=Re(u,i),e=[];break;default:"function"!=typeof s.onClick&&"function"==typeof i.onClick&&(u.onclick=Vn)}for(a in zn(n,i),n=null,s)if(!i.hasOwnProperty(a)&&s.hasOwnProperty(a)&&null!=s[a])if("style"===a)for(l in u=s[a])u.hasOwnProperty(l)&&(n||(n={}),n[l]="");else"dangerouslySetInnerHTML"!==a&&"children"!==a&&"suppressContentEditableWarning"!==a&&"suppressHydrationWarning"!==a&&"autoFocus"!==a&&(d.hasOwnProperty(a)?e||(e=[]):(e=e||[]).push(a,null));for(a in i){var c=i[a];if(u=null!=s?s[a]:void 0,i.hasOwnProperty(a)&&c!==u&&(null!=c||null!=u))if("style"===a)if(u){for(l in u)!u.hasOwnProperty(l)||c&&c.hasOwnProperty(l)||(n||(n={}),n[l]="");for(l in c)c.hasOwnProperty(l)&&u[l]!==c[l]&&(n||(n={}),n[l]=c[l])}else n||(e||(e=[]),e.push(a,n)),n=c;else"dangerouslySetInnerHTML"===a?(c=c?c.__html:void 0,u=u?u.__html:void 0,null!=c&&u!==c&&(e=e||[]).push(a,""+c)):"children"===a?u===c||"string"!=typeof c&&"number"!=typeof c||(e=e||[]).push(a,""+c):"suppressContentEditableWarning"!==a&&"suppressHydrationWarning"!==a&&(d.hasOwnProperty(a)?(null!=c&&Hn(o,a),e||u===c||(e=[])):(e=e||[]).push(a,c))}n&&(e=e||[]).push("style",n),o=e,(t.updateQueue=o)&&sa(t)}},Zs=function(e,t,n,i){n!==i&&sa(t)};var ca="function"==typeof WeakSet?WeakSet:Set;function ha(e,t){var n=t.source,i=t.stack;null===i&&null!==n&&(i=J(n)),null!==n&&X(n.type),t=t.value,null!==e&&1===e.tag&&X(e.type);try{console.error(t)}catch(e){setTimeout((function(){throw e}))}}function fa(e){var t=e.ref;if(null!==t)if("function"==typeof t)try{t(null)}catch(t){Fl(e,t)}else t.current=null}function da(e,t){switch(t.tag){case 0:case 11:case 15:pa(2,0,t);break;case 1:if(256&t.effectTag&&null!==e){var n=e.memoizedProps,i=e.memoizedState;t=(e=t.stateNode).getSnapshotBeforeUpdate(t.elementType===t.type?n:Xr(t.type,n),i),e.__reactInternalSnapshotBeforeUpdate=t}break;case 3:case 5:case 6:case 4:case 17:break;default:throw Error(s(163))}}function pa(e,t,n){if(null!==(n=null!==(n=n.updateQueue)?n.lastEffect:null)){var i=n=n.next;do{if(0!=(i.tag&e)){var r=i.destroy;i.destroy=void 0,void 0!==r&&r()}0!=(i.tag&t)&&(r=i.create,i.destroy=r()),i=i.next}while(i!==n)}}function ga(e,t,n){switch("function"==typeof _l&&_l(t),t.tag){case 0:case 11:case 14:case 15:if(null!==(e=t.updateQueue)&&null!==(e=e.lastEffect)){var i=e.next;Vr(97<n?97:n,(function(){var e=i;do{var n=e.destroy;if(void 0!==n){var r=t;try{n()}catch(e){Fl(r,e)}}e=e.next}while(e!==i)}))}break;case 1:fa(t),"function"==typeof(n=t.stateNode).componentWillUnmount&&function(e,t){try{t.props=e.memoizedProps,t.state=e.memoizedState,t.componentWillUnmount()}catch(t){Fl(e,t)}}(t,n);break;case 5:fa(t);break;case 4:ya(e,t,n)}}function ma(e){var t=e.alternate;e.return=null,e.child=null,e.memoizedState=null,e.updateQueue=null,e.dependencies=null,e.alternate=null,e.firstEffect=null,e.lastEffect=null,e.pendingProps=null,e.memoizedProps=null,null!==t&&ma(t)}function va(e){return 5===e.tag||3===e.tag||4===e.tag}function ba(e){e:{for(var t=e.return;null!==t;){if(va(t)){var n=t;break e}t=t.return}throw Error(s(160))}switch(t=n.stateNode,n.tag){case 5:var i=!1;break;case 3:case 4:t=t.containerInfo,i=!0;break;default:throw Error(s(161))}16&n.effectTag&&(Ve(t,""),n.effectTag&=-17);e:t:for(n=e;;){for(;null===n.sibling;){if(null===n.return||va(n.return)){n=null;break e}n=n.return}for(n.sibling.return=n.return,n=n.sibling;5!==n.tag&&6!==n.tag&&18!==n.tag;){if(2&n.effectTag)continue t;if(null===n.child||4===n.tag)continue t;n.child.return=n,n=n.child}if(!(2&n.effectTag)){n=n.stateNode;break e}}for(var r=e;;){var o=5===r.tag||6===r.tag;if(o){var a=o?r.stateNode:r.stateNode.instance;if(n)if(i){var l=a;a=n,8===(o=t).nodeType?o.parentNode.insertBefore(l,a):o.insertBefore(l,a)}else t.insertBefore(a,n);else i?(8===(l=t).nodeType?(o=l.parentNode).insertBefore(a,l):(o=l).appendChild(a),null!=(l=l._reactRootContainer)||null!==o.onclick||(o.onclick=Vn)):t.appendChild(a)}else if(4!==r.tag&&null!==r.child){r.child.return=r,r=r.child;continue}if(r===e)break;for(;null===r.sibling;){if(null===r.return||r.return===e)return;r=r.return}r.sibling.return=r.return,r=r.sibling}}function ya(e,t,n){for(var i,r,o=t,a=!1;;){if(!a){a=o.return;e:for(;;){if(null===a)throw Error(s(160));switch(i=a.stateNode,a.tag){case 5:r=!1;break e;case 3:case 4:i=i.containerInfo,r=!0;break e}a=a.return}a=!0}if(5===o.tag||6===o.tag){e:for(var l=e,u=o,c=n,h=u;;)if(ga(l,h,c),null!==h.child&&4!==h.tag)h.child.return=h,h=h.child;else{if(h===u)break;for(;null===h.sibling;){if(null===h.return||h.return===u)break e;h=h.return}h.sibling.return=h.return,h=h.sibling}r?(l=i,u=o.stateNode,8===l.nodeType?l.parentNode.removeChild(u):l.removeChild(u)):i.removeChild(o.stateNode)}else if(4===o.tag){if(null!==o.child){i=o.stateNode.containerInfo,r=!0,o.child.return=o,o=o.child;continue}}else if(ga(e,o,n),null!==o.child){o.child.return=o,o=o.child;continue}if(o===t)break;for(;null===o.sibling;){if(null===o.return||o.return===t)return;4===(o=o.return).tag&&(a=!1)}o.sibling.return=o.return,o=o.sibling}}function wa(e,t){switch(t.tag){case 0:case 11:case 14:case 15:pa(4,8,t);break;case 1:break;case 5:var n=t.stateNode;if(null!=n){var i=t.memoizedProps,r=null!==e?e.memoizedProps:i;e=t.type;var o=t.updateQueue;if(t.updateQueue=null,null!==o){for(n[si]=i,"input"===e&&"radio"===i.type&&null!=i.name&&_e(n,i),jn(e,r),t=jn(e,i),r=0;r<o.length;r+=2){var a=o[r],l=o[r+1];"style"===a?Nn(n,l):"dangerouslySetInnerHTML"===a?He(n,l):"children"===a?Ve(n,l):Ce(n,a,l,t)}switch(e){case"input":De(n,i);break;case"textarea":Me(n,i);break;case"select":t=n._wrapperState.wasMultiple,n._wrapperState.wasMultiple=!!i.multiple,null!=(e=i.value)?Te(n,!!i.multiple,e,!1):t!==!!i.multiple&&(null!=i.defaultValue?Te(n,!!i.multiple,i.defaultValue,!0):Te(n,!!i.multiple,i.multiple?[]:"",!1))}}}break;case 6:if(null===t.stateNode)throw Error(s(162));t.stateNode.nodeValue=t.memoizedProps;break;case 3:(t=t.stateNode).hydrate&&(t.hydrate=!1,kt(t.containerInfo));break;case 12:break;case 13:if(n=t,null===t.memoizedState?i=!1:(i=!0,n=t.child,Wa=zr()),null!==n)e:for(e=n;;){if(5===e.tag)o=e.stateNode,i?"function"==typeof(o=o.style).setProperty?o.setProperty("display","none","important"):o.display="none":(o=e.stateNode,r=null!=(r=e.memoizedProps.style)&&r.hasOwnProperty("display")?r.display:null,o.style.display=In("display",r));else if(6===e.tag)e.stateNode.nodeValue=i?"":e.memoizedProps;else{if(13===e.tag&&null!==e.memoizedState&&null===e.memoizedState.dehydrated){(o=e.child.sibling).return=e,e=o;continue}if(null!==e.child){e.child.return=e,e=e.child;continue}}if(e===n)break e;for(;null===e.sibling;){if(null===e.return||e.return===n)break e;e=e.return}e.sibling.return=e.return,e=e.sibling}Aa(t);break;case 19:Aa(t);break;case 17:case 20:case 21:break;default:throw Error(s(163))}}function Aa(e){var t=e.updateQueue;if(null!==t){e.updateQueue=null;var n=e.stateNode;null===n&&(n=e.stateNode=new ca),t.forEach((function(t){var i=kl.bind(null,e,t);n.has(t)||(n.add(t),t.then(i,i))}))}}var Ca="function"==typeof WeakMap?WeakMap:Map;function Ea(e,t,n){(n=ho(n,null)).tag=3,n.payload={element:null};var i=t.value;return n.callback=function(){ja||(ja=!0,Ha=i),ha(e,t)},n}function Fa(e,t,n){(n=ho(n,null)).tag=3;var i=e.type.getDerivedStateFromError;if("function"==typeof i){var r=t.value;n.payload=function(){return ha(e,t),i(r)}}var o=e.stateNode;return null!==o&&"function"==typeof o.componentDidCatch&&(n.callback=function(){"function"!=typeof i&&(null===Va?Va=new Set([this]):Va.add(this),ha(e,t));var n=t.stack;this.componentDidCatch(t.value,{componentStack:null!==n?n:""})}),n}var xa,ka=Math.ceil,Sa=T.ReactCurrentDispatcher,_a=T.ReactCurrentOwner,Da=0,$a=null,Ba=null,La=0,Ta=0,Ra=null,Oa=1073741823,Ma=1073741823,Pa=null,Ia=0,Na=!1,Wa=0,za=null,ja=!1,Ha=null,Va=null,Ua=!1,Ka=null,Ga=90,qa=null,Ya=0,Qa=null,Xa=0;function Ja(){return 0!=(48&Da)?1073741821-(zr()/10|0):0!==Xa?Xa:Xa=1073741821-(zr()/10|0)}function Za(e,t,n){if(0==(2&(t=t.mode)))return 1073741823;var i=jr();if(0==(4&t))return 99===i?1073741823:1073741822;if(0!=(16&Da))return La;if(null!==n)e=Qr(e,0|n.timeoutMs||5e3,250);else switch(i){case 99:e=1073741823;break;case 98:e=Qr(e,150,100);break;case 97:case 96:e=Qr(e,5e3,250);break;case 95:e=2;break;default:throw Error(s(326))}return null!==$a&&e===La&&--e,e}function el(e,t){if(50<Ya)throw Ya=0,Qa=null,Error(s(185));if(null!==(e=tl(e,t))){var n=jr();1073741823===t?0!=(8&Da)&&0==(48&Da)?ol(e):(il(e),0===Da&&Gr()):il(e),0==(4&Da)||98!==n&&99!==n||(null===qa?qa=new Map([[e,t]]):(void 0===(n=qa.get(e))||n>t)&&qa.set(e,t))}}function tl(e,t){e.expirationTime<t&&(e.expirationTime=t);var n=e.alternate;null!==n&&n.expirationTime<t&&(n.expirationTime=t);var i=e.return,r=null;if(null===i&&3===e.tag)r=e.stateNode;else for(;null!==i;){if(n=i.alternate,i.childExpirationTime<t&&(i.childExpirationTime=t),null!==n&&n.childExpirationTime<t&&(n.childExpirationTime=t),null===i.return&&3===i.tag){r=i.stateNode;break}i=i.return}return null!==r&&($a===r&&(fl(t),4===Ta&&Nl(r,La)),Wl(r,t)),r}function nl(e){var t=e.lastExpiredTime;return 0!==t?t:Il(e,t=e.firstPendingTime)?(t=e.lastPingedTime)>(e=e.nextKnownPendingLevel)?t:e:t}function il(e){if(0!==e.lastExpiredTime)e.callbackExpirationTime=1073741823,e.callbackPriority=99,e.callbackNode=Kr(ol.bind(null,e));else{var t=nl(e),n=e.callbackNode;if(0===t)null!==n&&(e.callbackNode=null,e.callbackExpirationTime=0,e.callbackPriority=90);else{var i=Ja();if(1073741823===t?i=99:1===t||2===t?i=95:i=0>=(i=10*(1073741821-t)-10*(1073741821-i))?99:250>=i?98:5250>=i?97:95,null!==n){var r=e.callbackPriority;if(e.callbackExpirationTime===t&&r>=i)return;n!==Or&&xr(n)}e.callbackExpirationTime=t,e.callbackPriority=i,t=1073741823===t?Kr(ol.bind(null,e)):Ur(i,rl.bind(null,e),{timeout:10*(1073741821-t)-zr()}),e.callbackNode=t}}}function rl(e,t){if(Xa=0,t)return zl(e,t=Ja()),il(e),null;var n=nl(e);if(0!==n){if(t=e.callbackNode,0!=(48&Da))throw Error(s(327));if(Al(),e===$a&&n===La||ll(e,n),null!==Ba){var i=Da;Da|=16;for(var r=cl();;)try{pl();break}catch(t){ul(e,t)}if(no(),Da=i,Sa.current=r,1===Ta)throw t=Ra,ll(e,n),Nl(e,n),il(e),t;if(null===Ba)switch(r=e.finishedWork=e.current.alternate,e.finishedExpirationTime=n,i=Ta,$a=null,i){case 0:case 1:throw Error(s(345));case 2:zl(e,2<n?2:n);break;case 3:if(Nl(e,n),n===(i=e.lastSuspendedTime)&&(e.nextKnownPendingLevel=vl(r)),1073741823===Oa&&10<(r=Wa+500-zr())){if(Na){var o=e.lastPingedTime;if(0===o||o>=n){e.lastPingedTime=n,ll(e,n);break}}if(0!==(o=nl(e))&&o!==n)break;if(0!==i&&i!==n){e.lastPingedTime=i;break}e.timeoutHandle=ei(bl.bind(null,e),r);break}bl(e);break;case 4:if(Nl(e,n),n===(i=e.lastSuspendedTime)&&(e.nextKnownPendingLevel=vl(r)),Na&&(0===(r=e.lastPingedTime)||r>=n)){e.lastPingedTime=n,ll(e,n);break}if(0!==(r=nl(e))&&r!==n)break;if(0!==i&&i!==n){e.lastPingedTime=i;break}if(1073741823!==Ma?i=10*(1073741821-Ma)-zr():1073741823===Oa?i=0:(i=10*(1073741821-Oa)-5e3,0>(i=(r=zr())-i)&&(i=0),(n=10*(1073741821-n)-r)<(i=(120>i?120:480>i?480:1080>i?1080:1920>i?1920:3e3>i?3e3:4320>i?4320:1960*ka(i/1960))-i)&&(i=n)),10<i){e.timeoutHandle=ei(bl.bind(null,e),i);break}bl(e);break;case 5:if(1073741823!==Oa&&null!==Pa){o=Oa;var a=Pa;if(0>=(i=0|a.busyMinDurationMs)?i=0:(r=0|a.busyDelayMs,i=(o=zr()-(10*(1073741821-o)-(0|a.timeoutMs||5e3)))<=r?0:r+i-o),10<i){Nl(e,n),e.timeoutHandle=ei(bl.bind(null,e),i);break}}bl(e);break;default:throw Error(s(329))}if(il(e),e.callbackNode===t)return rl.bind(null,e)}}return null}function ol(e){var t=e.lastExpiredTime;if(t=0!==t?t:1073741823,e.finishedExpirationTime===t)bl(e);else{if(0!=(48&Da))throw Error(s(327));if(Al(),e===$a&&t===La||ll(e,t),null!==Ba){var n=Da;Da|=16;for(var i=cl();;)try{dl();break}catch(t){ul(e,t)}if(no(),Da=n,Sa.current=i,1===Ta)throw n=Ra,ll(e,t),Nl(e,t),il(e),n;if(null!==Ba)throw Error(s(261));e.finishedWork=e.current.alternate,e.finishedExpirationTime=t,$a=null,bl(e),il(e)}}return null}function sl(e,t){var n=Da;Da|=1;try{return e(t)}finally{0===(Da=n)&&Gr()}}function al(e,t){var n=Da;Da&=-2,Da|=8;try{return e(t)}finally{0===(Da=n)&&Gr()}}function ll(e,t){e.finishedWork=null,e.finishedExpirationTime=0;var n=e.timeoutHandle;if(-1!==n&&(e.timeoutHandle=-1,ti(n)),null!==Ba)for(n=Ba.return;null!==n;){var i=n;switch(i.tag){case 1:var r=i.type.childContextTypes;null!=r&&vr();break;case 3:zo(),br();break;case 5:Ho(i);break;case 4:zo();break;case 13:case 19:ur(Vo);break;case 10:ro(i)}n=n.return}$a=e,Ba=Ll(e.current,null),La=t,Ta=0,Ra=null,Ma=Oa=1073741823,Pa=null,Ia=0,Na=!1}function ul(e,t){for(;;){try{if(no(),hs(),null===Ba||null===Ba.return)return Ta=1,Ra=t,null;e:{var n=e,i=Ba.return,r=Ba,o=t;if(t=La,r.effectTag|=2048,r.firstEffect=r.lastEffect=null,null!==o&&"object"==typeof o&&"function"==typeof o.then){var s=o,a=0!=(1&Vo.current),l=i;do{var u;if(u=13===l.tag){var c=l.memoizedState;if(null!==c)u=null!==c.dehydrated;else{var h=l.memoizedProps;u=void 0!==h.fallback&&(!0!==h.unstable_avoidThisFallback||!a)}}if(u){var f=l.updateQueue;if(null===f){var d=new Set;d.add(s),l.updateQueue=d}else f.add(s);if(0==(2&l.mode)){if(l.effectTag|=64,r.effectTag&=-2981,1===r.tag)if(null===r.alternate)r.tag=17;else{var p=ho(1073741823,null);p.tag=2,po(r,p)}r.expirationTime=1073741823;break e}o=void 0,r=t;var g=n.pingCache;if(null===g?(g=n.pingCache=new Ca,o=new Set,g.set(s,o)):void 0===(o=g.get(s))&&(o=new Set,g.set(s,o)),!o.has(r)){o.add(r);var m=xl.bind(null,n,s,r);s.then(m,m)}l.effectTag|=4096,l.expirationTime=t;break e}l=l.return}while(null!==l);o=Error((X(r.type)||"A React component")+" suspended while rendering, but no fallback UI was specified.\n\nAdd a <Suspense fallback=...> component higher in the tree to provide a loading indicator or placeholder to display."+J(r))}5!==Ta&&(Ta=2),o=ua(o,r),l=i;do{switch(l.tag){case 3:s=o,l.effectTag|=4096,l.expirationTime=t,go(l,Ea(l,s,t));break e;case 1:s=o;var v=l.type,b=l.stateNode;if(0==(64&l.effectTag)&&("function"==typeof v.getDerivedStateFromError||null!==b&&"function"==typeof b.componentDidCatch&&(null===Va||!Va.has(b)))){l.effectTag|=4096,l.expirationTime=t,go(l,Fa(l,s,t));break e}}l=l.return}while(null!==l)}Ba=ml(Ba)}catch(e){t=e;continue}break}}function cl(){var e=Sa.current;return Sa.current=_s,null===e?_s:e}function hl(e,t){e<Oa&&2<e&&(Oa=e),null!==t&&e<Ma&&2<e&&(Ma=e,Pa=t)}function fl(e){e>Ia&&(Ia=e)}function dl(){for(;null!==Ba;)Ba=gl(Ba)}function pl(){for(;null!==Ba&&!kr();)Ba=gl(Ba)}function gl(e){var t=xa(e.alternate,e,La);return e.memoizedProps=e.pendingProps,null===t&&(t=ml(e)),_a.current=null,t}function ml(e){Ba=e;do{var t=Ba.alternate;if(e=Ba.return,0==(2048&Ba.effectTag)){e:{var n=t,i=La,o=(t=Ba).pendingProps;switch(t.tag){case 2:case 16:break;case 15:case 0:break;case 1:mr(t.type)&&vr();break;case 3:zo(),br(),(o=t.stateNode).pendingContext&&(o.context=o.pendingContext,o.pendingContext=null),(null===n||null===n.child)&&Is(t)&&sa(t);break;case 5:Ho(t),i=No(Io.current);var a=t.type;if(null!==n&&null!=t.stateNode)Js(n,t,a,o,i),n.ref!==t.ref&&(t.effectTag|=128);else if(o){var l=No(Mo.current);if(Is(t)){var u=(o=t).stateNode;n=o.type;var c=o.memoizedProps,h=i;switch(u[oi]=o,u[si]=c,a=void 0,i=u,n){case"iframe":case"object":case"embed":xn("load",i);break;case"video":case"audio":for(u=0;u<et.length;u++)xn(et[u],i);break;case"source":xn("error",i);break;case"img":case"image":case"link":xn("error",i),xn("load",i);break;case"form":xn("reset",i),xn("submit",i);break;case"details":xn("toggle",i);break;case"input":Se(i,c),xn("invalid",i),Hn(h,"onChange");break;case"select":i._wrapperState={wasMultiple:!!c.multiple},xn("invalid",i),Hn(h,"onChange");break;case"textarea":Oe(i,c),xn("invalid",i),Hn(h,"onChange")}for(a in zn(n,c),u=null,c)c.hasOwnProperty(a)&&(l=c[a],"children"===a?"string"==typeof l?i.textContent!==l&&(u=["children",l]):"number"==typeof l&&i.textContent!==""+l&&(u=["children",""+l]):d.hasOwnProperty(a)&&null!=l&&Hn(h,a));switch(n){case"input":Fe(i),$e(i,c,!0);break;case"textarea":Fe(i),Pe(i);break;case"select":case"option":break;default:"function"==typeof c.onClick&&(i.onclick=Vn)}a=u,o.updateQueue=a,(o=null!==a)&&sa(t)}else{n=t,h=a,c=o,u=9===i.nodeType?i:i.ownerDocument,l===Ie&&(l=We(h)),l===Ie?"script"===h?((c=u.createElement("div")).innerHTML="<script><\/script>",u=c.removeChild(c.firstChild)):"string"==typeof c.is?u=u.createElement(h,{is:c.is}):(u=u.createElement(h),"select"===h&&(h=u,c.multiple?h.multiple=!0:c.size&&(h.size=c.size))):u=u.createElementNS(l,h),(c=u)[oi]=n,c[si]=o,Xs(c,t),t.stateNode=c;var f=i,p=jn(h=a,n=o);switch(h){case"iframe":case"object":case"embed":xn("load",c),i=n;break;case"video":case"audio":for(i=0;i<et.length;i++)xn(et[i],c);i=n;break;case"source":xn("error",c),i=n;break;case"img":case"image":case"link":xn("error",c),xn("load",c),i=n;break;case"form":xn("reset",c),xn("submit",c),i=n;break;case"details":xn("toggle",c),i=n;break;case"input":Se(c,n),i=ke(c,n),xn("invalid",c),Hn(f,"onChange");break;case"option":i=Le(c,n);break;case"select":c._wrapperState={wasMultiple:!!n.multiple},i=r({},n,{value:void 0}),xn("invalid",c),Hn(f,"onChange");break;case"textarea":Oe(c,n),i=Re(c,n),xn("invalid",c),Hn(f,"onChange");break;default:i=n}zn(h,i),u=void 0,l=h;var g=c,m=i;for(u in m)if(m.hasOwnProperty(u)){var v=m[u];"style"===u?Nn(g,v):"dangerouslySetInnerHTML"===u?null!=(v=v?v.__html:void 0)&&He(g,v):"children"===u?"string"==typeof v?("textarea"!==l||""!==v)&&Ve(g,v):"number"==typeof v&&Ve(g,""+v):"suppressContentEditableWarning"!==u&&"suppressHydrationWarning"!==u&&"autoFocus"!==u&&(d.hasOwnProperty(u)?null!=v&&Hn(f,u):null!=v&&Ce(g,u,v,p))}switch(h){case"input":Fe(c),$e(c,n,!1);break;case"textarea":Fe(c),Pe(c);break;case"option":null!=n.value&&c.setAttribute("value",""+Ae(n.value));break;case"select":(i=c).multiple=!!n.multiple,null!=(c=n.value)?Te(i,!!n.multiple,c,!1):null!=n.defaultValue&&Te(i,!!n.multiple,n.defaultValue,!0);break;default:"function"==typeof i.onClick&&(c.onclick=Vn)}(o=Jn(a,o))&&sa(t)}null!==t.ref&&(t.effectTag|=128)}else if(null===t.stateNode)throw Error(s(166));break;case 6:if(n&&null!=t.stateNode)Zs(0,t,n.memoizedProps,o);else{if("string"!=typeof o&&null===t.stateNode)throw Error(s(166));i=No(Io.current),No(Mo.current),Is(t)?(a=(o=t).stateNode,i=o.memoizedProps,a[oi]=o,(o=a.nodeValue!==i)&&sa(t)):(a=t,(o=(9===i.nodeType?i:i.ownerDocument).createTextNode(o))[oi]=a,t.stateNode=o)}break;case 11:break;case 13:if(ur(Vo),o=t.memoizedState,0!=(64&t.effectTag)){t.expirationTime=i;break e}o=null!==o,a=!1,null===n?void 0!==t.memoizedProps.fallback&&Is(t):(a=null!==(i=n.memoizedState),o||null===i||null!==(i=n.child.sibling)&&(null!==(c=t.firstEffect)?(t.firstEffect=i,i.nextEffect=c):(t.firstEffect=t.lastEffect=i,i.nextEffect=null),i.effectTag=8)),o&&!a&&0!=(2&t.mode)&&(null===n&&!0!==t.memoizedProps.unstable_avoidThisFallback||0!=(1&Vo.current)?0===Ta&&(Ta=3):(0!==Ta&&3!==Ta||(Ta=4),0!==Ia&&null!==$a&&(Nl($a,La),Wl($a,Ia)))),(o||a)&&(t.effectTag|=4);break;case 7:case 8:case 12:break;case 4:zo();break;case 10:ro(t);break;case 9:case 14:break;case 17:mr(t.type)&&vr();break;case 19:if(ur(Vo),null===(o=t.memoizedState))break;if(a=0!=(64&t.effectTag),null===(c=o.rendering)){if(a)aa(o,!1);else if(0!==Ta||null!==n&&0!=(64&n.effectTag))for(n=t.child;null!==n;){if(null!==(c=Uo(n))){for(t.effectTag|=64,aa(o,!1),null!==(a=c.updateQueue)&&(t.updateQueue=a,t.effectTag|=4),null===o.lastEffect&&(t.firstEffect=null),t.lastEffect=o.lastEffect,o=i,a=t.child;null!==a;)n=o,(i=a).effectTag&=2,i.nextEffect=null,i.firstEffect=null,i.lastEffect=null,null===(c=i.alternate)?(i.childExpirationTime=0,i.expirationTime=n,i.child=null,i.memoizedProps=null,i.memoizedState=null,i.updateQueue=null,i.dependencies=null):(i.childExpirationTime=c.childExpirationTime,i.expirationTime=c.expirationTime,i.child=c.child,i.memoizedProps=c.memoizedProps,i.memoizedState=c.memoizedState,i.updateQueue=c.updateQueue,n=c.dependencies,i.dependencies=null===n?null:{expirationTime:n.expirationTime,firstContext:n.firstContext,responders:n.responders}),a=a.sibling;cr(Vo,1&Vo.current|2),t=t.child;break e}n=n.sibling}}else{if(!a)if(null!==(n=Uo(c))){if(t.effectTag|=64,a=!0,null!==(i=n.updateQueue)&&(t.updateQueue=i,t.effectTag|=4),aa(o,!0),null===o.tail&&"hidden"===o.tailMode&&!c.alternate){null!==(t=t.lastEffect=o.lastEffect)&&(t.nextEffect=null);break}}else zr()>o.tailExpiration&&1<i&&(t.effectTag|=64,a=!0,aa(o,!1),t.expirationTime=t.childExpirationTime=i-1);o.isBackwards?(c.sibling=t.child,t.child=c):(null!==(i=o.last)?i.sibling=c:t.child=c,o.last=c)}if(null!==o.tail){0===o.tailExpiration&&(o.tailExpiration=zr()+500),i=o.tail,o.rendering=i,o.tail=i.sibling,o.lastEffect=t.lastEffect,i.sibling=null,o=Vo.current,cr(Vo,o=a?1&o|2:1&o),t=i;break e}break;case 20:case 21:break;default:throw Error(s(156,t.tag))}t=null}if(o=Ba,1===La||1!==o.childExpirationTime){for(a=0,i=o.child;null!==i;)(n=i.expirationTime)>a&&(a=n),(c=i.childExpirationTime)>a&&(a=c),i=i.sibling;o.childExpirationTime=a}if(null!==t)return t;null!==e&&0==(2048&e.effectTag)&&(null===e.firstEffect&&(e.firstEffect=Ba.firstEffect),null!==Ba.lastEffect&&(null!==e.lastEffect&&(e.lastEffect.nextEffect=Ba.firstEffect),e.lastEffect=Ba.lastEffect),1<Ba.effectTag&&(null!==e.lastEffect?e.lastEffect.nextEffect=Ba:e.firstEffect=Ba,e.lastEffect=Ba))}else{if(null!==(t=la(Ba)))return t.effectTag&=2047,t;null!==e&&(e.firstEffect=e.lastEffect=null,e.effectTag|=2048)}if(null!==(t=Ba.sibling))return t;Ba=e}while(null!==Ba);return 0===Ta&&(Ta=5),null}function vl(e){var t=e.expirationTime;return t>(e=e.childExpirationTime)?t:e}function bl(e){var t=jr();return Vr(99,yl.bind(null,e,t)),null}function yl(e,t){do{Al()}while(null!==Ka);if(0!=(48&Da))throw Error(s(327));var n=e.finishedWork,i=e.finishedExpirationTime;if(null===n)return null;if(e.finishedWork=null,e.finishedExpirationTime=0,n===e.current)throw Error(s(177));e.callbackNode=null,e.callbackExpirationTime=0,e.callbackPriority=90,e.nextKnownPendingLevel=0;var r=vl(n);if(e.firstPendingTime=r,i<=e.lastSuspendedTime?e.firstSuspendedTime=e.lastSuspendedTime=e.nextKnownPendingLevel=0:i<=e.firstSuspendedTime&&(e.firstSuspendedTime=i-1),i<=e.lastPingedTime&&(e.lastPingedTime=0),i<=e.lastExpiredTime&&(e.lastExpiredTime=0),e===$a&&(Ba=$a=null,La=0),1<n.effectTag?null!==n.lastEffect?(n.lastEffect.nextEffect=n,r=n.firstEffect):r=n:r=n.firstEffect,null!==r){var o=Da;Da|=32,_a.current=null,Qn=Fn;var a=qn();if(Yn(a)){if("selectionStart"in a)var l={start:a.selectionStart,end:a.selectionEnd};else e:{var u=(l=(l=a.ownerDocument)&&l.defaultView||window).getSelection&&l.getSelection();if(u&&0!==u.rangeCount){l=u.anchorNode;var c=u.anchorOffset,h=u.focusNode;u=u.focusOffset;try{l.nodeType,h.nodeType}catch(e){l=null;break e}var f=0,d=-1,p=-1,g=0,m=0,v=a,b=null;t:for(;;){for(var y;v!==l||0!==c&&3!==v.nodeType||(d=f+c),v!==h||0!==u&&3!==v.nodeType||(p=f+u),3===v.nodeType&&(f+=v.nodeValue.length),null!==(y=v.firstChild);)b=v,v=y;for(;;){if(v===a)break t;if(b===l&&++g===c&&(d=f),b===h&&++m===u&&(p=f),null!==(y=v.nextSibling))break;b=(v=b).parentNode}v=y}l=-1===d||-1===p?null:{start:d,end:p}}else l=null}l=l||{start:0,end:0}}else l=null;Xn={focusedElem:a,selectionRange:l},Fn=!1,za=r;do{try{wl()}catch(e){if(null===za)throw Error(s(330));Fl(za,e),za=za.nextEffect}}while(null!==za);za=r;do{try{for(a=e,l=t;null!==za;){var w=za.effectTag;if(16&w&&Ve(za.stateNode,""),128&w){var A=za.alternate;if(null!==A){var C=A.ref;null!==C&&("function"==typeof C?C(null):C.current=null)}}switch(1038&w){case 2:ba(za),za.effectTag&=-3;break;case 6:ba(za),za.effectTag&=-3,wa(za.alternate,za);break;case 1024:za.effectTag&=-1025;break;case 1028:za.effectTag&=-1025,wa(za.alternate,za);break;case 4:wa(za.alternate,za);break;case 8:ya(a,c=za,l),ma(c)}za=za.nextEffect}}catch(e){if(null===za)throw Error(s(330));Fl(za,e),za=za.nextEffect}}while(null!==za);if(C=Xn,A=qn(),w=C.focusedElem,l=C.selectionRange,A!==w&&w&&w.ownerDocument&&function e(t,n){return!(!t||!n)&&(t===n||(!t||3!==t.nodeType)&&(n&&3===n.nodeType?e(t,n.parentNode):"contains"in t?t.contains(n):!!t.compareDocumentPosition&&!!(16&t.compareDocumentPosition(n))))}(w.ownerDocument.documentElement,w)){null!==l&&Yn(w)&&(A=l.start,void 0===(C=l.end)&&(C=A),"selectionStart"in w?(w.selectionStart=A,w.selectionEnd=Math.min(C,w.value.length)):(C=(A=w.ownerDocument||document)&&A.defaultView||window).getSelection&&(C=C.getSelection(),c=w.textContent.length,a=Math.min(l.start,c),l=void 0===l.end?a:Math.min(l.end,c),!C.extend&&a>l&&(c=l,l=a,a=c),c=Gn(w,a),h=Gn(w,l),c&&h&&(1!==C.rangeCount||C.anchorNode!==c.node||C.anchorOffset!==c.offset||C.focusNode!==h.node||C.focusOffset!==h.offset)&&((A=A.createRange()).setStart(c.node,c.offset),C.removeAllRanges(),a>l?(C.addRange(A),C.extend(h.node,h.offset)):(A.setEnd(h.node,h.offset),C.addRange(A))))),A=[];for(C=w;C=C.parentNode;)1===C.nodeType&&A.push({element:C,left:C.scrollLeft,top:C.scrollTop});for("function"==typeof w.focus&&w.focus(),w=0;w<A.length;w++)(C=A[w]).element.scrollLeft=C.left,C.element.scrollTop=C.top}Xn=null,Fn=!!Qn,Qn=null,e.current=n,za=r;do{try{for(w=i;null!==za;){var E=za.effectTag;if(36&E){var F=za.alternate;switch(C=w,(A=za).tag){case 0:case 11:case 15:pa(16,32,A);break;case 1:var x=A.stateNode;if(4&A.effectTag)if(null===F)x.componentDidMount();else{var k=A.elementType===A.type?F.memoizedProps:Xr(A.type,F.memoizedProps);x.componentDidUpdate(k,F.memoizedState,x.__reactInternalSnapshotBeforeUpdate)}var S=A.updateQueue;null!==S&&yo(0,S,x);break;case 3:var _=A.updateQueue;if(null!==_){if(a=null,null!==A.child)switch(A.child.tag){case 5:a=A.child.stateNode;break;case 1:a=A.child.stateNode}yo(0,_,a)}break;case 5:var D=A.stateNode;null===F&&4&A.effectTag&&Jn(A.type,A.memoizedProps)&&D.focus();break;case 6:case 4:case 12:break;case 13:if(null===A.memoizedState){var $=A.alternate;if(null!==$){var B=$.memoizedState;if(null!==B){var L=B.dehydrated;null!==L&&kt(L)}}}break;case 19:case 17:case 20:case 21:break;default:throw Error(s(163))}}if(128&E){A=void 0;var T=za.ref;if(null!==T){var R=za.stateNode;switch(za.tag){case 5:A=R;break;default:A=R}"function"==typeof T?T(A):T.current=A}}za=za.nextEffect}}catch(e){if(null===za)throw Error(s(330));Fl(za,e),za=za.nextEffect}}while(null!==za);za=null,Mr(),Da=o}else e.current=n;if(Ua)Ua=!1,Ka=e,Ga=t;else for(za=r;null!==za;)t=za.nextEffect,za.nextEffect=null,za=t;if(0===(t=e.firstPendingTime)&&(Va=null),1073741823===t?e===Qa?Ya++:(Ya=0,Qa=e):Ya=0,"function"==typeof Sl&&Sl(n.stateNode,i),il(e),ja)throw ja=!1,e=Ha,Ha=null,e;return 0!=(8&Da)?null:(Gr(),null)}function wl(){for(;null!==za;){var e=za.effectTag;0!=(256&e)&&da(za.alternate,za),0==(512&e)||Ua||(Ua=!0,Ur(97,(function(){return Al(),null}))),za=za.nextEffect}}function Al(){if(90!==Ga){var e=97<Ga?97:Ga;return Ga=90,Vr(e,Cl)}}function Cl(){if(null===Ka)return!1;var e=Ka;if(Ka=null,0!=(48&Da))throw Error(s(331));var t=Da;for(Da|=32,e=e.current.firstEffect;null!==e;){try{var n=e;if(0!=(512&n.effectTag))switch(n.tag){case 0:case 11:case 15:pa(128,0,n),pa(0,64,n)}}catch(t){if(null===e)throw Error(s(330));Fl(e,t)}n=e.nextEffect,e.nextEffect=null,e=n}return Da=t,Gr(),!0}function El(e,t,n){po(e,t=Ea(e,t=ua(n,t),1073741823)),null!==(e=tl(e,1073741823))&&il(e)}function Fl(e,t){if(3===e.tag)El(e,e,t);else for(var n=e.return;null!==n;){if(3===n.tag){El(n,e,t);break}if(1===n.tag){var i=n.stateNode;if("function"==typeof n.type.getDerivedStateFromError||"function"==typeof i.componentDidCatch&&(null===Va||!Va.has(i))){po(n,e=Fa(n,e=ua(t,e),1073741823)),null!==(n=tl(n,1073741823))&&il(n);break}}n=n.return}}function xl(e,t,n){var i=e.pingCache;null!==i&&i.delete(t),$a===e&&La===n?4===Ta||3===Ta&&1073741823===Oa&&zr()-Wa<500?ll(e,La):Na=!0:Il(e,n)&&(0!==(t=e.lastPingedTime)&&t<n||(e.lastPingedTime=n,e.finishedExpirationTime===n&&(e.finishedExpirationTime=0,e.finishedWork=null),il(e)))}function kl(e,t){var n=e.stateNode;null!==n&&n.delete(t),0===(t=0)&&(t=Za(t=Ja(),e,null)),null!==(e=tl(e,t))&&il(e)}xa=function(e,t,n){var i=t.expirationTime;if(null!==e){var r=t.pendingProps;if(e.memoizedProps!==r||dr.current)zs=!0;else{if(i<n){switch(zs=!1,t.tag){case 3:Qs(t),Ns();break;case 5:if(jo(t),4&t.mode&&1!==n&&r.hidden)return t.expirationTime=t.childExpirationTime=1,null;break;case 1:mr(t.type)&&Ar(t);break;case 4:Wo(t,t.stateNode.containerInfo);break;case 10:io(t,t.memoizedProps.value);break;case 13:if(null!==t.memoizedState)return 0!==(i=t.child.childExpirationTime)&&i>=n?ta(e,t,n):(cr(Vo,1&Vo.current),null!==(t=oa(e,t,n))?t.sibling:null);cr(Vo,1&Vo.current);break;case 19:if(i=t.childExpirationTime>=n,0!=(64&e.effectTag)){if(i)return ra(e,t,n);t.effectTag|=64}if(null!==(r=t.memoizedState)&&(r.rendering=null,r.tail=null),cr(Vo,Vo.current),!i)return null}return oa(e,t,n)}zs=!1}}else zs=!1;switch(t.expirationTime=0,t.tag){case 2:if(i=t.type,null!==e&&(e.alternate=null,t.alternate=null,t.effectTag|=2),e=t.pendingProps,r=gr(t,fr.current),so(t,n),r=cs(null,t,i,e,r,n),t.effectTag|=1,"object"==typeof r&&null!==r&&"function"==typeof r.render&&void 0===r.$$typeof){if(t.tag=1,hs(),mr(i)){var o=!0;Ar(t)}else o=!1;t.memoizedState=null!==r.state&&void 0!==r.state?r.state:null;var a=i.getDerivedStateFromProps;"function"==typeof a&&Eo(t,i,a,e),r.updater=Fo,t.stateNode=r,r._reactInternalFiber=t,_o(t,i,e,n),t=Ys(null,t,i,!0,o,n)}else t.tag=0,js(null,t,r,n),t=t.child;return t;case 16:if(r=t.elementType,null!==e&&(e.alternate=null,t.alternate=null,t.effectTag|=2),e=t.pendingProps,function(e){if(-1===e._status){e._status=0;var t=e._ctor;t=t(),e._result=t,t.then((function(t){0===e._status&&(t=t.default,e._status=1,e._result=t)}),(function(t){0===e._status&&(e._status=2,e._result=t)}))}}(r),1!==r._status)throw r._result;switch(r=r._result,t.type=r,o=t.tag=function(e){if("function"==typeof e)return Bl(e)?1:0;if(null!=e){if((e=e.$$typeof)===V)return 11;if(e===G)return 14}return 2}(r),e=Xr(r,e),o){case 0:t=Gs(null,t,r,e,n);break;case 1:t=qs(null,t,r,e,n);break;case 11:t=Hs(null,t,r,e,n);break;case 14:t=Vs(null,t,r,Xr(r.type,e),i,n);break;default:throw Error(s(306,r,""))}return t;case 0:return i=t.type,r=t.pendingProps,Gs(e,t,i,r=t.elementType===i?r:Xr(i,r),n);case 1:return i=t.type,r=t.pendingProps,qs(e,t,i,r=t.elementType===i?r:Xr(i,r),n);case 3:if(Qs(t),null===(i=t.updateQueue))throw Error(s(282));if(r=null!==(r=t.memoizedState)?r.element:null,bo(t,i,t.pendingProps,null,n),(i=t.memoizedState.element)===r)Ns(),t=oa(e,t,n);else{if((r=t.stateNode.hydrate)&&(Ls=ni(t.stateNode.containerInfo.firstChild),Bs=t,r=Ts=!0),r)for(n=Ro(t,null,i,n),t.child=n;n;)n.effectTag=-3&n.effectTag|1024,n=n.sibling;else js(e,t,i,n),Ns();t=t.child}return t;case 5:return jo(t),null===e&&Ms(t),i=t.type,r=t.pendingProps,o=null!==e?e.memoizedProps:null,a=r.children,Zn(i,r)?a=null:null!==o&&Zn(i,o)&&(t.effectTag|=16),Ks(e,t),4&t.mode&&1!==n&&r.hidden?(t.expirationTime=t.childExpirationTime=1,t=null):(js(e,t,a,n),t=t.child),t;case 6:return null===e&&Ms(t),null;case 13:return ta(e,t,n);case 4:return Wo(t,t.stateNode.containerInfo),i=t.pendingProps,null===e?t.child=To(t,null,i,n):js(e,t,i,n),t.child;case 11:return i=t.type,r=t.pendingProps,Hs(e,t,i,r=t.elementType===i?r:Xr(i,r),n);case 7:return js(e,t,t.pendingProps,n),t.child;case 8:case 12:return js(e,t,t.pendingProps.children,n),t.child;case 10:e:{if(i=t.type._context,r=t.pendingProps,a=t.memoizedProps,io(t,o=r.value),null!==a){var l=a.value;if(0===(o=Qi(l,o)?0:0|("function"==typeof i._calculateChangedBits?i._calculateChangedBits(l,o):1073741823))){if(a.children===r.children&&!dr.current){t=oa(e,t,n);break e}}else for(null!==(l=t.child)&&(l.return=t);null!==l;){var u=l.dependencies;if(null!==u){a=l.child;for(var c=u.firstContext;null!==c;){if(c.context===i&&0!=(c.observedBits&o)){1===l.tag&&((c=ho(n,null)).tag=2,po(l,c)),l.expirationTime<n&&(l.expirationTime=n),null!==(c=l.alternate)&&c.expirationTime<n&&(c.expirationTime=n),oo(l.return,n),u.expirationTime<n&&(u.expirationTime=n);break}c=c.next}}else a=10===l.tag&&l.type===t.type?null:l.child;if(null!==a)a.return=l;else for(a=l;null!==a;){if(a===t){a=null;break}if(null!==(l=a.sibling)){l.return=a.return,a=l;break}a=a.return}l=a}}js(e,t,r.children,n),t=t.child}return t;case 9:return r=t.type,i=(o=t.pendingProps).children,so(t,n),i=i(r=ao(r,o.unstable_observedBits)),t.effectTag|=1,js(e,t,i,n),t.child;case 14:return o=Xr(r=t.type,t.pendingProps),Vs(e,t,r,o=Xr(r.type,o),i,n);case 15:return Us(e,t,t.type,t.pendingProps,i,n);case 17:return i=t.type,r=t.pendingProps,r=t.elementType===i?r:Xr(i,r),null!==e&&(e.alternate=null,t.alternate=null,t.effectTag|=2),t.tag=1,mr(i)?(e=!0,Ar(t)):e=!1,so(t,n),ko(t,i,r),_o(t,i,r,n),Ys(null,t,i,!0,e,n);case 19:return ra(e,t,n)}throw Error(s(156,t.tag))};var Sl=null,_l=null;function Dl(e,t,n,i){this.tag=e,this.key=n,this.sibling=this.child=this.return=this.stateNode=this.type=this.elementType=null,this.index=0,this.ref=null,this.pendingProps=t,this.dependencies=this.memoizedState=this.updateQueue=this.memoizedProps=null,this.mode=i,this.effectTag=0,this.lastEffect=this.firstEffect=this.nextEffect=null,this.childExpirationTime=this.expirationTime=0,this.alternate=null}function $l(e,t,n,i){return new Dl(e,t,n,i)}function Bl(e){return!(!(e=e.prototype)||!e.isReactComponent)}function Ll(e,t){var n=e.alternate;return null===n?((n=$l(e.tag,t,e.key,e.mode)).elementType=e.elementType,n.type=e.type,n.stateNode=e.stateNode,n.alternate=e,e.alternate=n):(n.pendingProps=t,n.effectTag=0,n.nextEffect=null,n.firstEffect=null,n.lastEffect=null),n.childExpirationTime=e.childExpirationTime,n.expirationTime=e.expirationTime,n.child=e.child,n.memoizedProps=e.memoizedProps,n.memoizedState=e.memoizedState,n.updateQueue=e.updateQueue,t=e.dependencies,n.dependencies=null===t?null:{expirationTime:t.expirationTime,firstContext:t.firstContext,responders:t.responders},n.sibling=e.sibling,n.index=e.index,n.ref=e.ref,n}function Tl(e,t,n,i,r,o){var a=2;if(i=e,"function"==typeof e)Bl(e)&&(a=1);else if("string"==typeof e)a=5;else e:switch(e){case I:return Rl(n.children,r,o,t);case H:a=8,r|=7;break;case N:a=8,r|=1;break;case W:return(e=$l(12,n,t,8|r)).elementType=W,e.type=W,e.expirationTime=o,e;case U:return(e=$l(13,n,t,r)).type=U,e.elementType=U,e.expirationTime=o,e;case K:return(e=$l(19,n,t,r)).elementType=K,e.expirationTime=o,e;default:if("object"==typeof e&&null!==e)switch(e.$$typeof){case z:a=10;break e;case j:a=9;break e;case V:a=11;break e;case G:a=14;break e;case q:a=16,i=null;break e}throw Error(s(130,null==e?e:typeof e,""))}return(t=$l(a,n,t,r)).elementType=e,t.type=i,t.expirationTime=o,t}function Rl(e,t,n,i){return(e=$l(7,e,i,t)).expirationTime=n,e}function Ol(e,t,n){return(e=$l(6,e,null,t)).expirationTime=n,e}function Ml(e,t,n){return(t=$l(4,null!==e.children?e.children:[],e.key,t)).expirationTime=n,t.stateNode={containerInfo:e.containerInfo,pendingChildren:null,implementation:e.implementation},t}function Pl(e,t,n){this.tag=t,this.current=null,this.containerInfo=e,this.pingCache=this.pendingChildren=null,this.finishedExpirationTime=0,this.finishedWork=null,this.timeoutHandle=-1,this.pendingContext=this.context=null,this.hydrate=n,this.callbackNode=null,this.callbackPriority=90,this.lastExpiredTime=this.lastPingedTime=this.nextKnownPendingLevel=this.lastSuspendedTime=this.firstSuspendedTime=this.firstPendingTime=0}function Il(e,t){var n=e.firstSuspendedTime;return e=e.lastSuspendedTime,0!==n&&n>=t&&e<=t}function Nl(e,t){var n=e.firstSuspendedTime,i=e.lastSuspendedTime;n<t&&(e.firstSuspendedTime=t),(i>t||0===n)&&(e.lastSuspendedTime=t),t<=e.lastPingedTime&&(e.lastPingedTime=0),t<=e.lastExpiredTime&&(e.lastExpiredTime=0)}function Wl(e,t){t>e.firstPendingTime&&(e.firstPendingTime=t);var n=e.firstSuspendedTime;0!==n&&(t>=n?e.firstSuspendedTime=e.lastSuspendedTime=e.nextKnownPendingLevel=0:t>=e.lastSuspendedTime&&(e.lastSuspendedTime=t+1),t>e.nextKnownPendingLevel&&(e.nextKnownPendingLevel=t))}function zl(e,t){var n=e.lastExpiredTime;(0===n||n>t)&&(e.lastExpiredTime=t)}function jl(e,t,n,i){var r=t.current,o=Ja(),a=Ao.suspense;o=Za(o,r,a);e:if(n){t:{if(tt(n=n._reactInternalFiber)!==n||1!==n.tag)throw Error(s(170));var l=n;do{switch(l.tag){case 3:l=l.stateNode.context;break t;case 1:if(mr(l.type)){l=l.stateNode.__reactInternalMemoizedMergedChildContext;break t}}l=l.return}while(null!==l);throw Error(s(171))}if(1===n.tag){var u=n.type;if(mr(u)){n=wr(n,u,l);break e}}n=l}else n=hr;return null===t.context?t.context=n:t.pendingContext=n,(t=ho(o,a)).payload={element:e},null!==(i=void 0===i?null:i)&&(t.callback=i),po(r,t),el(r,o),o}function Hl(e){if(!(e=e.current).child)return null;switch(e.child.tag){case 5:default:return e.child.stateNode}}function Vl(e,t){null!==(e=e.memoizedState)&&null!==e.dehydrated&&e.retryTime<t&&(e.retryTime=t)}function Ul(e,t){Vl(e,t),(e=e.alternate)&&Vl(e,t)}function Kl(e,t,n){var i=new Pl(e,t,n=null!=n&&!0===n.hydrate),r=$l(3,null,null,2===t?7:1===t?3:0);i.current=r,r.stateNode=i,e[ai]=i.current,n&&0!==t&&function(e){var t=Rn(e);mt.forEach((function(n){On(n,e,t)})),vt.forEach((function(n){On(n,e,t)}))}(9===e.nodeType?e:e.ownerDocument),this._internalRoot=i}function Gl(e){return!(!e||1!==e.nodeType&&9!==e.nodeType&&11!==e.nodeType&&(8!==e.nodeType||" react-mount-point-unstable "!==e.nodeValue))}function ql(e,t,n,i,r){var o=n._reactRootContainer;if(o){var s=o._internalRoot;if("function"==typeof r){var a=r;r=function(){var e=Hl(s);a.call(e)}}jl(t,s,e,r)}else{if(o=n._reactRootContainer=function(e,t){if(t||(t=!(!(t=e?9===e.nodeType?e.documentElement:e.firstChild:null)||1!==t.nodeType||!t.hasAttribute("data-reactroot"))),!t)for(var n;n=e.lastChild;)e.removeChild(n);return new Kl(e,0,t?{hydrate:!0}:void 0)}(n,i),s=o._internalRoot,"function"==typeof r){var l=r;r=function(){var e=Hl(s);l.call(e)}}al((function(){jl(t,s,e,r)}))}return Hl(s)}function Yl(e,t,n){var i=3<arguments.length&&void 0!==arguments[3]?arguments[3]:null;return{$$typeof:P,key:null==i?null:""+i,children:e,containerInfo:t,implementation:n}}function Ql(e,t){var n=2<arguments.length&&void 0!==arguments[2]?arguments[2]:null;if(!Gl(t))throw Error(s(200));return Yl(e,t,null,n)}Kl.prototype.render=function(e,t){jl(e,this._internalRoot,null,void 0===t?null:t)},Kl.prototype.unmount=function(e){var t=this._internalRoot,n=void 0===e?null:e,i=t.containerInfo;jl(null,t,null,(function(){i[ai]=null,null!==n&&n()}))},ot=function(e){if(13===e.tag){var t=Qr(Ja(),150,100);el(e,t),Ul(e,t)}},st=function(e){if(13===e.tag){Ja();var t=Yr++;el(e,t),Ul(e,t)}},at=function(e){if(13===e.tag){var t=Ja();el(e,t=Za(t,e,null)),Ul(e,t)}},ee=function(e,t,n){switch(t){case"input":if(De(e,n),t=n.name,"radio"===n.type&&null!=t){for(n=e;n.parentNode;)n=n.parentNode;for(n=n.querySelectorAll("input[name="+JSON.stringify(""+t)+'][type="radio"]'),t=0;t<n.length;t++){var i=n[t];if(i!==e&&i.form===e.form){var r=hi(i);if(!r)throw Error(s(90));xe(i),De(i,r)}}}break;case"textarea":Me(e,n);break;case"select":null!=(t=n.value)&&Te(e,!!n.multiple,t,!1)}},se=sl,ae=function(e,t,n,i){var r=Da;Da|=4;try{return Vr(98,e.bind(null,t,n,i))}finally{0===(Da=r)&&Gr()}},le=function(){0==(49&Da)&&(function(){if(null!==qa){var e=qa;qa=null,e.forEach((function(e,t){zl(t,e),il(t)})),Gr()}}(),Al())},ue=function(e,t){var n=Da;Da|=2;try{return e(t)}finally{0===(Da=n)&&Gr()}};var Xl,Jl,Zl={createPortal:Ql,findDOMNode:function(e){if(null==e)return null;if(1===e.nodeType)return e;var t=e._reactInternalFiber;if(void 0===t){if("function"==typeof e.render)throw Error(s(188));throw Error(s(268,Object.keys(e)))}return e=null===(e=rt(t))?null:e.stateNode},hydrate:function(e,t,n){if(!Gl(t))throw Error(s(200));return ql(null,e,t,!0,n)},render:function(e,t,n){if(!Gl(t))throw Error(s(200));return ql(null,e,t,!1,n)},unstable_renderSubtreeIntoContainer:function(e,t,n,i){if(!Gl(n))throw Error(s(200));if(null==e||void 0===e._reactInternalFiber)throw Error(s(38));return ql(e,t,n,!1,i)},unmountComponentAtNode:function(e){if(!Gl(e))throw Error(s(40));return!!e._reactRootContainer&&(al((function(){ql(null,null,e,!1,(function(){e._reactRootContainer=null,e[ai]=null}))})),!0)},unstable_createPortal:function(){return Ql.apply(void 0,arguments)},unstable_batchedUpdates:sl,flushSync:function(e,t){if(0!=(48&Da))throw Error(s(187));var n=Da;Da|=1;try{return Vr(99,e.bind(null,t))}finally{Da=n,Gr()}},__SECRET_INTERNALS_DO_NOT_USE_OR_YOU_WILL_BE_FIRED:{Events:[ui,ci,hi,B.injectEventPluginsByName,f,Tt,function(e){S(e,Lt)},re,oe,$n,$,Al,{current:!1}]}};Jl=(Xl={findFiberByHostInstance:li,bundleType:0,version:"16.12.0",rendererPackageName:"react-dom"}).findFiberByHostInstance,function(e){if("undefined"==typeof __REACT_DEVTOOLS_GLOBAL_HOOK__)return!1;var t=__REACT_DEVTOOLS_GLOBAL_HOOK__;if(t.isDisabled||!t.supportsFiber)return!0;try{var n=t.inject(e);Sl=function(e){try{t.onCommitFiberRoot(n,e,void 0,64==(64&e.current.effectTag))}catch(e){}},_l=function(e){try{t.onCommitFiberUnmount(n,e)}catch(e){}}}catch(e){}}(r({},Xl,{overrideHookState:null,overrideProps:null,setSuspenseHandler:null,scheduleUpdate:null,currentDispatcherRef:T.ReactCurrentDispatcher,findHostInstanceByFiber:function(e){return null===(e=rt(e))?null:e.stateNode},findFiberByHostInstance:function(e){return Jl?Jl(e):null},findHostInstancesForRefresh:null,scheduleRefresh:null,scheduleRoot:null,setRefreshHandler:null,getCurrentFiber:null}));var eu={default:Zl},tu=eu&&Zl||eu;e.exports=tu.default||tu},function(e,t,n){"use strict";e.exports=n(67)},function(e,t,n){"use strict";
/** @license React v0.18.0
 * scheduler.production.min.js
 *
 * Copyright (c) Facebook, Inc. and its affiliates.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 */var i,r,o,s,a;if(Object.defineProperty(t,"__esModule",{value:!0}),"undefined"==typeof window||"function"!=typeof MessageChannel){var l=null,u=null,c=function(){if(null!==l)try{var e=t.unstable_now();l(!0,e),l=null}catch(e){throw setTimeout(c,0),e}},h=Date.now();t.unstable_now=function(){return Date.now()-h},i=function(e){null!==l?setTimeout(i,0,e):(l=e,setTimeout(c,0))},r=function(e,t){u=setTimeout(e,t)},o=function(){clearTimeout(u)},s=function(){return!1},a=t.unstable_forceFrameRate=function(){}}else{var f=window.performance,d=window.Date,p=window.setTimeout,g=window.clearTimeout;if("undefined"!=typeof console){var m=window.cancelAnimationFrame;"function"!=typeof window.requestAnimationFrame&&console.error("This browser doesn't support requestAnimationFrame. Make sure that you load a polyfill in older browsers. https://fb.me/react-polyfills"),"function"!=typeof m&&console.error("This browser doesn't support cancelAnimationFrame. Make sure that you load a polyfill in older browsers. https://fb.me/react-polyfills")}if("object"==typeof f&&"function"==typeof f.now)t.unstable_now=function(){return f.now()};else{var v=d.now();t.unstable_now=function(){return d.now()-v}}var b=!1,y=null,w=-1,A=5,C=0;s=function(){return t.unstable_now()>=C},a=function(){},t.unstable_forceFrameRate=function(e){0>e||125<e?console.error("forceFrameRate takes a positive int between 0 and 125, forcing framerates higher than 125 fps is not unsupported"):A=0<e?Math.floor(1e3/e):5};var E=new MessageChannel,F=E.port2;E.port1.onmessage=function(){if(null!==y){var e=t.unstable_now();C=e+A;try{y(!0,e)?F.postMessage(null):(b=!1,y=null)}catch(e){throw F.postMessage(null),e}}else b=!1},i=function(e){y=e,b||(b=!0,F.postMessage(null))},r=function(e,n){w=p((function(){e(t.unstable_now())}),n)},o=function(){g(w),w=-1}}function x(e,t){var n=e.length;e.push(t);e:for(;;){var i=Math.floor((n-1)/2),r=e[i];if(!(void 0!==r&&0<_(r,t)))break e;e[i]=t,e[n]=r,n=i}}function k(e){return void 0===(e=e[0])?null:e}function S(e){var t=e[0];if(void 0!==t){var n=e.pop();if(n!==t){e[0]=n;e:for(var i=0,r=e.length;i<r;){var o=2*(i+1)-1,s=e[o],a=o+1,l=e[a];if(void 0!==s&&0>_(s,n))void 0!==l&&0>_(l,s)?(e[i]=l,e[a]=n,i=a):(e[i]=s,e[o]=n,i=o);else{if(!(void 0!==l&&0>_(l,n)))break e;e[i]=l,e[a]=n,i=a}}}return t}return null}function _(e,t){var n=e.sortIndex-t.sortIndex;return 0!==n?n:e.id-t.id}var D=[],$=[],B=1,L=null,T=3,R=!1,O=!1,M=!1;function P(e){for(var t=k($);null!==t;){if(null===t.callback)S($);else{if(!(t.startTime<=e))break;S($),t.sortIndex=t.expirationTime,x(D,t)}t=k($)}}function I(e){if(M=!1,P(e),!O)if(null!==k(D))O=!0,i(N);else{var t=k($);null!==t&&r(I,t.startTime-e)}}function N(e,n){O=!1,M&&(M=!1,o()),R=!0;var i=T;try{for(P(n),L=k(D);null!==L&&(!(L.expirationTime>n)||e&&!s());){var a=L.callback;if(null!==a){L.callback=null,T=L.priorityLevel;var l=a(L.expirationTime<=n);n=t.unstable_now(),"function"==typeof l?L.callback=l:L===k(D)&&S(D),P(n)}else S(D);L=k(D)}if(null!==L)var u=!0;else{var c=k($);null!==c&&r(I,c.startTime-n),u=!1}return u}finally{L=null,T=i,R=!1}}function W(e){switch(e){case 1:return-1;case 2:return 250;case 5:return 1073741823;case 4:return 1e4;default:return 5e3}}var z=a;t.unstable_ImmediatePriority=1,t.unstable_UserBlockingPriority=2,t.unstable_NormalPriority=3,t.unstable_IdlePriority=5,t.unstable_LowPriority=4,t.unstable_runWithPriority=function(e,t){switch(e){case 1:case 2:case 3:case 4:case 5:break;default:e=3}var n=T;T=e;try{return t()}finally{T=n}},t.unstable_next=function(e){switch(T){case 1:case 2:case 3:var t=3;break;default:t=T}var n=T;T=t;try{return e()}finally{T=n}},t.unstable_scheduleCallback=function(e,n,s){var a=t.unstable_now();if("object"==typeof s&&null!==s){var l=s.delay;l="number"==typeof l&&0<l?a+l:a,s="number"==typeof s.timeout?s.timeout:W(e)}else s=W(e),l=a;return e={id:B++,callback:n,priorityLevel:e,startTime:l,expirationTime:s=l+s,sortIndex:-1},l>a?(e.sortIndex=l,x($,e),null===k(D)&&e===k($)&&(M?o():M=!0,r(I,l-a))):(e.sortIndex=s,x(D,e),O||R||(O=!0,i(N))),e},t.unstable_cancelCallback=function(e){e.callback=null},t.unstable_wrapCallback=function(e){var t=T;return function(){var n=T;T=t;try{return e.apply(this,arguments)}finally{T=n}}},t.unstable_getCurrentPriorityLevel=function(){return T},t.unstable_shouldYield=function(){var e=t.unstable_now();P(e);var n=k(D);return n!==L&&null!==L&&null!==n&&null!==n.callback&&n.startTime<=e&&n.expirationTime<L.expirationTime||s()},t.unstable_requestPaint=z,t.unstable_continueExecution=function(){O||R||(O=!0,i(N))},t.unstable_pauseExecution=function(){},t.unstable_getFirstCallbackNode=function(){return k(D)},t.unstable_Profiling=null},function(e,t){e.exports=function(e){if(Array.isArray(e))return e}},function(e,t){e.exports=function(e,t){if(Symbol.iterator in Object(e)||"[object Arguments]"===Object.prototype.toString.call(e)){var n=[],i=!0,r=!1,o=void 0;try{for(var s,a=e[Symbol.iterator]();!(i=(s=a.next()).done)&&(n.push(s.value),!t||n.length!==t);i=!0);}catch(e){r=!0,o=e}finally{try{i||null==a.return||a.return()}finally{if(r)throw o}}return n}}},function(e,t){e.exports=function(){throw new TypeError("Invalid attempt to destructure non-iterable instance")}},function(e,t){var n,i,r=e.exports={};function o(){throw new Error("setTimeout has not been defined")}function s(){throw new Error("clearTimeout has not been defined")}function a(e){if(n===setTimeout)return setTimeout(e,0);if((n===o||!n)&&setTimeout)return n=setTimeout,setTimeout(e,0);try{return n(e,0)}catch(t){try{return n.call(null,e,0)}catch(t){return n.call(this,e,0)}}}!function(){try{n="function"==typeof setTimeout?setTimeout:o}catch(e){n=o}try{i="function"==typeof clearTimeout?clearTimeout:s}catch(e){i=s}}();var l,u=[],c=!1,h=-1;function f(){c&&l&&(c=!1,l.length?u=l.concat(u):h=-1,u.length&&d())}function d(){if(!c){var e=a(f);c=!0;for(var t=u.length;t;){for(l=u,u=[];++h<t;)l&&l[h].run();h=-1,t=u.length}l=null,c=!1,function(e){if(i===clearTimeout)return clearTimeout(e);if((i===s||!i)&&clearTimeout)return i=clearTimeout,clearTimeout(e);try{i(e)}catch(t){try{return i.call(null,e)}catch(t){return i.call(this,e)}}}(e)}}function p(e,t){this.fun=e,this.array=t}function g(){}r.nextTick=function(e){var t=new Array(arguments.length-1);if(arguments.length>1)for(var n=1;n<arguments.length;n++)t[n-1]=arguments[n];u.push(new p(e,t)),1!==u.length||c||a(d)},p.prototype.run=function(){this.fun.apply(null,this.array)},r.title="browser",r.browser=!0,r.env={},r.argv=[],r.version="",r.versions={},r.on=g,r.addListener=g,r.once=g,r.off=g,r.removeListener=g,r.removeAllListeners=g,r.emit=g,r.prependListener=g,r.prependOnceListener=g,r.listeners=function(e){return[]},r.binding=function(e){throw new Error("process.binding is not supported")},r.cwd=function(){return"/"},r.chdir=function(e){throw new Error("process.chdir is not supported")},r.umask=function(){return 0}},function(e,t){ace.define("ace/mode/python_highlight_rules",["require","exports","module","ace/lib/oop","ace/mode/text_highlight_rules"],(function(e,t,n){"use strict";var i=e("../lib/oop"),r=e("./text_highlight_rules").TextHighlightRules,o=function(){var e=this.createKeywordMapper({"invalid.deprecated":"debugger","support.function":"abs|divmod|input|open|staticmethod|all|enumerate|int|ord|str|any|eval|isinstance|pow|sum|basestring|execfile|issubclass|print|super|binfile|iter|property|tuple|bool|filter|len|range|type|bytearray|float|list|raw_input|unichr|callable|format|locals|reduce|unicode|chr|frozenset|long|reload|vars|classmethod|getattr|map|repr|xrange|cmp|globals|max|reversed|zip|compile|hasattr|memoryview|round|__import__|complex|hash|min|set|apply|delattr|help|next|setattr|buffer|dict|hex|object|slice|coerce|dir|id|oct|sorted|intern","constant.language":"True|False|None|NotImplemented|Ellipsis|__debug__",keyword:"and|as|assert|break|class|continue|def|del|elif|else|except|exec|finally|for|from|global|if|import|in|is|lambda|not|or|pass|print|raise|return|try|while|with|yield|async|await"},"identifier"),t="(?:r|u|ur|R|U|UR|Ur|uR)?",n="(?:(?:(?:[1-9]\\d*)|(?:0))|(?:0[oO]?[0-7]+)|(?:0[xX][\\dA-Fa-f]+)|(?:0[bB][01]+))",i="(?:(?:(?:\\d+)?(?:\\.\\d+))|(?:(?:\\d+)\\.))",r="(?:(?:(?:(?:(?:(?:\\d+)?(?:\\.\\d+))|(?:(?:\\d+)\\.))|(?:\\d+))(?:[eE][+-]?\\d+))|"+i+")",o="\\\\(x[0-9A-Fa-f]{2}|[0-7]{3}|[\\\\abfnrtv'\"]|U[0-9A-Fa-f]{8}|u[0-9A-Fa-f]{4})";this.$rules={start:[{token:"comment",regex:"#.*$"},{token:"string",regex:t+'"{3}',next:"qqstring3"},{token:"string",regex:t+'"(?=.)',next:"qqstring"},{token:"string",regex:t+"'{3}",next:"qstring3"},{token:"string",regex:t+"'(?=.)",next:"qstring"},{token:"constant.numeric",regex:"(?:"+r+"|\\d+)[jJ]\\b"},{token:"constant.numeric",regex:r},{token:"constant.numeric",regex:n+"[lL]\\b"},{token:"constant.numeric",regex:n+"\\b"},{token:e,regex:"[a-zA-Z_$][a-zA-Z0-9_$]*\\b"},{token:"keyword.operator",regex:"\\+|\\-|\\*|\\*\\*|\\/|\\/\\/|%|<<|>>|&|\\||\\^|~|<|>|<=|=>|==|!=|<>|="},{token:"paren.lparen",regex:"[\\[\\(\\{]"},{token:"paren.rparen",regex:"[\\]\\)\\}]"},{token:"text",regex:"\\s+"}],qqstring3:[{token:"constant.language.escape",regex:o},{token:"string",regex:'"{3}',next:"start"},{defaultToken:"string"}],qstring3:[{token:"constant.language.escape",regex:o},{token:"string",regex:"'{3}",next:"start"},{defaultToken:"string"}],qqstring:[{token:"constant.language.escape",regex:o},{token:"string",regex:"\\\\$",next:"qqstring"},{token:"string",regex:'"|$',next:"start"},{defaultToken:"string"}],qstring:[{token:"constant.language.escape",regex:o},{token:"string",regex:"\\\\$",next:"qstring"},{token:"string",regex:"'|$",next:"start"},{defaultToken:"string"}]}};i.inherits(o,r),t.PythonHighlightRules=o})),ace.define("ace/mode/folding/pythonic",["require","exports","module","ace/lib/oop","ace/mode/folding/fold_mode"],(function(e,t,n){"use strict";var i=e("../../lib/oop"),r=e("./fold_mode").FoldMode,o=t.FoldMode=function(e){this.foldingStartMarker=new RegExp("([\\[{])(?:\\s*)$|("+e+")(?:\\s*)(?:#.*)?$")};i.inherits(o,r),function(){this.getFoldWidgetRange=function(e,t,n){var i=e.getLine(n).match(this.foldingStartMarker);if(i)return i[1]?this.openingBracketBlock(e,i[1],n,i.index):i[2]?this.indentationBlock(e,n,i.index+i[2].length):this.indentationBlock(e,n)}}.call(o.prototype)})),ace.define("ace/mode/python",["require","exports","module","ace/lib/oop","ace/mode/text","ace/mode/python_highlight_rules","ace/mode/folding/pythonic","ace/range"],(function(e,t,n){"use strict";var i=e("../lib/oop"),r=e("./text").Mode,o=e("./python_highlight_rules").PythonHighlightRules,s=e("./folding/pythonic").FoldMode,a=e("../range").Range,l=function(){this.HighlightRules=o,this.foldingRules=new s("\\:"),this.$behaviour=this.$defaultBehaviour};i.inherits(l,r),function(){this.lineCommentStart="#",this.getNextLineIndent=function(e,t,n){var i=this.$getIndent(t),r=this.getTokenizer().getLineTokens(t,e).tokens;if(r.length&&"comment"==r[r.length-1].type)return i;"start"==e&&(t.match(/^.*[\{\(\[:]\s*$/)&&(i+=n));return i};var e={pass:1,return:1,raise:1,break:1,continue:1};this.checkOutdent=function(t,n,i){if("\r\n"!==i&&"\r"!==i&&"\n"!==i)return!1;var r=this.getTokenizer().getLineTokens(n.trim(),t).tokens;if(!r)return!1;do{var o=r.pop()}while(o&&("comment"==o.type||"text"==o.type&&o.value.match(/^\s+$/)));return!!o&&("keyword"==o.type&&e[o.value])},this.autoOutdent=function(e,t,n){n+=1;var i=this.$getIndent(t.getLine(n)),r=t.getTabString();i.slice(-r.length)==r&&t.remove(new a(n,i.length-r.length,n,i.length))},this.$id="ace/mode/python"}.call(l.prototype),t.Mode=l}))},function(e,t){ace.define("ace/theme/monokai",["require","exports","module","ace/lib/dom"],(function(e,t,n){t.isDark=!0,t.cssClass="ace-monokai",t.cssText=".ace-monokai .ace_gutter {background: #2F3129;color: #8F908A}.ace-monokai .ace_print-margin {width: 1px;background: #555651}.ace-monokai {background-color: #272822;color: #F8F8F2}.ace-monokai .ace_cursor {color: #F8F8F0}.ace-monokai .ace_marker-layer .ace_selection {background: #49483E}.ace-monokai.ace_multiselect .ace_selection.ace_start {box-shadow: 0 0 3px 0px #272822;}.ace-monokai .ace_marker-layer .ace_step {background: rgb(102, 82, 0)}.ace-monokai .ace_marker-layer .ace_bracket {margin: -1px 0 0 -1px;border: 1px solid #49483E}.ace-monokai .ace_marker-layer .ace_active-line {background: #202020}.ace-monokai .ace_gutter-active-line {background-color: #272727}.ace-monokai .ace_marker-layer .ace_selected-word {border: 1px solid #49483E}.ace-monokai .ace_invisible {color: #52524d}.ace-monokai .ace_entity.ace_name.ace_tag,.ace-monokai .ace_keyword,.ace-monokai .ace_meta.ace_tag,.ace-monokai .ace_storage {color: #F92672}.ace-monokai .ace_punctuation,.ace-monokai .ace_punctuation.ace_tag {color: #fff}.ace-monokai .ace_constant.ace_character,.ace-monokai .ace_constant.ace_language,.ace-monokai .ace_constant.ace_numeric,.ace-monokai .ace_constant.ace_other {color: #AE81FF}.ace-monokai .ace_invalid {color: #F8F8F0;background-color: #F92672}.ace-monokai .ace_invalid.ace_deprecated {color: #F8F8F0;background-color: #AE81FF}.ace-monokai .ace_support.ace_constant,.ace-monokai .ace_support.ace_function {color: #66D9EF}.ace-monokai .ace_fold {background-color: #A6E22E;border-color: #F8F8F2}.ace-monokai .ace_storage.ace_type,.ace-monokai .ace_support.ace_class,.ace-monokai .ace_support.ace_type {font-style: italic;color: #66D9EF}.ace-monokai .ace_entity.ace_name.ace_function,.ace-monokai .ace_entity.ace_other,.ace-monokai .ace_entity.ace_other.ace_attribute-name,.ace-monokai .ace_variable {color: #A6E22E}.ace-monokai .ace_variable.ace_parameter {font-style: italic;color: #FD971F}.ace-monokai .ace_string {color: #E6DB74}.ace-monokai .ace_comment {color: #75715E}.ace-monokai .ace_indent-guide {background: url(data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAAEAAAACCAYAAACZgbYnAAAAEklEQVQImWPQ0FD0ZXBzd/wPAAjVAoxeSgNeAAAAAElFTkSuQmCC) right repeat-y}",e("../lib/dom").importCssString(t.cssText,t.cssClass)}))},function(e,t,n){var i=n(75),r=n(42),o=n(106),s=n(108),a=n(7),l=n(52),u=n(51);e.exports=function e(t,n,c,h,f){t!==n&&o(n,(function(o,l){if(f||(f=new i),a(o))s(t,n,l,c,e,h,f);else{var d=h?h(u(t,l),o,l+"",t,n,f):void 0;void 0===d&&(d=o),r(t,l,d)}}),l)}},function(e,t,n){var i=n(20),r=n(81),o=n(82),s=n(83),a=n(84),l=n(85);function u(e){var t=this.__data__=new i(e);this.size=t.size}u.prototype.clear=r,u.prototype.delete=o,u.prototype.get=s,u.prototype.has=a,u.prototype.set=l,e.exports=u},function(e,t){e.exports=function(){this.__data__=[],this.size=0}},function(e,t,n){var i=n(21),r=Array.prototype.splice;e.exports=function(e){var t=this.__data__,n=i(t,e);return!(n<0)&&(n==t.length-1?t.pop():r.call(t,n,1),--this.size,!0)}},function(e,t,n){var i=n(21);e.exports=function(e){var t=this.__data__,n=i(t,e);return n<0?void 0:t[n][1]}},function(e,t,n){var i=n(21);e.exports=function(e){return i(this.__data__,e)>-1}},function(e,t,n){var i=n(21);e.exports=function(e,t){var n=this.__data__,r=i(n,e);return r<0?(++this.size,n.push([e,t])):n[r][1]=t,this}},function(e,t,n){var i=n(20);e.exports=function(){this.__data__=new i,this.size=0}},function(e,t){e.exports=function(e){var t=this.__data__,n=t.delete(e);return this.size=t.size,n}},function(e,t){e.exports=function(e){return this.__data__.get(e)}},function(e,t){e.exports=function(e){return this.__data__.has(e)}},function(e,t,n){var i=n(20),r=n(39),o=n(93);e.exports=function(e,t){var n=this.__data__;if(n instanceof i){var s=n.__data__;if(!r||s.length<199)return s.push([e,t]),this.size=++n.size,this;n=this.__data__=new o(s)}return n.set(e,t),this.size=n.size,this}},function(e,t,n){var i=n(28),r=n(89),o=n(7),s=n(91),a=/^\[object .+?Constructor\]$/,l=Function.prototype,u=Object.prototype,c=l.toString,h=u.hasOwnProperty,f=RegExp("^"+c.call(h).replace(/[\\^$.*+?()[\]{}|]/g,"\\$&").replace(/hasOwnProperty|(function).*?(?=\\\()| for .+?(?=\\\])/g,"$1.*?")+"$");e.exports=function(e){return!(!o(e)||r(e))&&(i(e)?f:a).test(s(e))}},function(e,t,n){var i=n(40),r=Object.prototype,o=r.hasOwnProperty,s=r.toString,a=i?i.toStringTag:void 0;e.exports=function(e){var t=o.call(e,a),n=e[a];try{e[a]=void 0;var i=!0}catch(e){}var r=s.call(e);return i&&(t?e[a]=n:delete e[a]),r}},function(e,t){var n=Object.prototype.toString;e.exports=function(e){return n.call(e)}},function(e,t,n){var i,r=n(90),o=(i=/[^.]+$/.exec(r&&r.keys&&r.keys.IE_PROTO||""))?"Symbol(src)_1."+i:"";e.exports=function(e){return!!o&&o in e}},function(e,t,n){var i=n(11)["__core-js_shared__"];e.exports=i},function(e,t){var n=Function.prototype.toString;e.exports=function(e){if(null!=e){try{return n.call(e)}catch(e){}try{return e+""}catch(e){}}return""}},function(e,t){e.exports=function(e,t){return null==e?void 0:e[t]}},function(e,t,n){var i=n(94),r=n(101),o=n(103),s=n(104),a=n(105);function l(e){var t=-1,n=null==e?0:e.length;for(this.clear();++t<n;){var i=e[t];this.set(i[0],i[1])}}l.prototype.clear=i,l.prototype.delete=r,l.prototype.get=o,l.prototype.has=s,l.prototype.set=a,e.exports=l},function(e,t,n){var i=n(95),r=n(20),o=n(39);e.exports=function(){this.size=0,this.__data__={hash:new i,map:new(o||r),string:new i}}},function(e,t,n){var i=n(96),r=n(97),o=n(98),s=n(99),a=n(100);function l(e){var t=-1,n=null==e?0:e.length;for(this.clear();++t<n;){var i=e[t];this.set(i[0],i[1])}}l.prototype.clear=i,l.prototype.delete=r,l.prototype.get=o,l.prototype.has=s,l.prototype.set=a,e.exports=l},function(e,t,n){var i=n(23);e.exports=function(){this.__data__=i?i(null):{},this.size=0}},function(e,t){e.exports=function(e){var t=this.has(e)&&delete this.__data__[e];return this.size-=t?1:0,t}},function(e,t,n){var i=n(23),r=Object.prototype.hasOwnProperty;e.exports=function(e){var t=this.__data__;if(i){var n=t[e];return"__lodash_hash_undefined__"===n?void 0:n}return r.call(t,e)?t[e]:void 0}},function(e,t,n){var i=n(23),r=Object.prototype.hasOwnProperty;e.exports=function(e){var t=this.__data__;return i?void 0!==t[e]:r.call(t,e)}},function(e,t,n){var i=n(23);e.exports=function(e,t){var n=this.__data__;return this.size+=this.has(e)?0:1,n[e]=i&&void 0===t?"__lodash_hash_undefined__":t,this}},function(e,t,n){var i=n(24);e.exports=function(e){var t=i(this,e).delete(e);return this.size-=t?1:0,t}},function(e,t){e.exports=function(e){var t=typeof e;return"string"==t||"number"==t||"symbol"==t||"boolean"==t?"__proto__"!==e:null===e}},function(e,t,n){var i=n(24);e.exports=function(e){return i(this,e).get(e)}},function(e,t,n){var i=n(24);e.exports=function(e){return i(this,e).has(e)}},function(e,t,n){var i=n(24);e.exports=function(e,t){var n=i(this,e),r=n.size;return n.set(e,t),this.size+=n.size==r?0:1,this}},function(e,t,n){var i=n(107)();e.exports=i},function(e,t){e.exports=function(e){return function(t,n,i){for(var r=-1,o=Object(t),s=i(t),a=s.length;a--;){var l=s[e?a:++r];if(!1===n(o[l],l,o))break}return t}}},function(e,t,n){var i=n(42),r=n(109),o=n(110),s=n(113),a=n(114),l=n(46),u=n(47),c=n(118),h=n(49),f=n(28),d=n(7),p=n(120),g=n(50),m=n(51),v=n(124);e.exports=function(e,t,n,b,y,w,A){var C=m(e,n),E=m(t,n),F=A.get(E);if(F)i(e,n,F);else{var x=w?w(C,E,n+"",e,t,A):void 0,k=void 0===x;if(k){var S=u(E),_=!S&&h(E),D=!S&&!_&&g(E);x=E,S||_||D?u(C)?x=C:c(C)?x=s(C):_?(k=!1,x=r(E,!0)):D?(k=!1,x=o(E,!0)):x=[]:p(E)||l(E)?(x=C,l(C)?x=v(C):d(C)&&!f(C)||(x=a(E))):k=!1}k&&(A.set(E,x),y(x,E,b,w,A),A.delete(E)),i(e,n,x)}}},function(e,t,n){(function(e){var i=n(11),r=t&&!t.nodeType&&t,o=r&&"object"==typeof e&&e&&!e.nodeType&&e,s=o&&o.exports===r?i.Buffer:void 0,a=s?s.allocUnsafe:void 0;e.exports=function(e,t){if(t)return e.slice();var n=e.length,i=a?a(n):new e.constructor(n);return e.copy(i),i}}).call(this,n(25)(e))},function(e,t,n){var i=n(111);e.exports=function(e,t){var n=t?i(e.buffer):e.buffer;return new e.constructor(n,e.byteOffset,e.length)}},function(e,t,n){var i=n(112);e.exports=function(e){var t=new e.constructor(e.byteLength);return new i(t).set(new i(e)),t}},function(e,t,n){var i=n(11).Uint8Array;e.exports=i},function(e,t){e.exports=function(e,t){var n=-1,i=e.length;for(t||(t=Array(i));++n<i;)t[n]=e[n];return t}},function(e,t,n){var i=n(115),r=n(44),o=n(45);e.exports=function(e){return"function"!=typeof e.constructor||o(e)?{}:i(r(e))}},function(e,t,n){var i=n(7),r=Object.create,o=function(){function e(){}return function(t){if(!i(t))return{};if(r)return r(t);e.prototype=t;var n=new e;return e.prototype=void 0,n}}();e.exports=o},function(e,t){e.exports=function(e,t){return function(n){return e(t(n))}}},function(e,t,n){var i=n(14),r=n(12);e.exports=function(e){return r(e)&&"[object Arguments]"==i(e)}},function(e,t,n){var i=n(30),r=n(12);e.exports=function(e){return r(e)&&i(e)}},function(e,t){e.exports=function(){return!1}},function(e,t,n){var i=n(14),r=n(44),o=n(12),s=Function.prototype,a=Object.prototype,l=s.toString,u=a.hasOwnProperty,c=l.call(Object);e.exports=function(e){if(!o(e)||"[object Object]"!=i(e))return!1;var t=r(e);if(null===t)return!0;var n=u.call(t,"constructor")&&t.constructor;return"function"==typeof n&&n instanceof n&&l.call(n)==c}},function(e,t,n){var i=n(14),r=n(48),o=n(12),s={};s["[object Float32Array]"]=s["[object Float64Array]"]=s["[object Int8Array]"]=s["[object Int16Array]"]=s["[object Int32Array]"]=s["[object Uint8Array]"]=s["[object Uint8ClampedArray]"]=s["[object Uint16Array]"]=s["[object Uint32Array]"]=!0,s["[object Arguments]"]=s["[object Array]"]=s["[object ArrayBuffer]"]=s["[object Boolean]"]=s["[object DataView]"]=s["[object Date]"]=s["[object Error]"]=s["[object Function]"]=s["[object Map]"]=s["[object Number]"]=s["[object Object]"]=s["[object RegExp]"]=s["[object Set]"]=s["[object String]"]=s["[object WeakMap]"]=!1,e.exports=function(e){return o(e)&&r(e.length)&&!!s[i(e)]}},function(e,t){e.exports=function(e){return function(t){return e(t)}}},function(e,t,n){(function(e){var i=n(41),r=t&&!t.nodeType&&t,o=r&&"object"==typeof e&&e&&!e.nodeType&&e,s=o&&o.exports===r&&i.process,a=function(){try{var e=o&&o.require&&o.require("util").types;return e||s&&s.binding&&s.binding("util")}catch(e){}}();e.exports=a}).call(this,n(25)(e))},function(e,t,n){var i=n(125),r=n(52);e.exports=function(e){return i(e,r(e))}},function(e,t,n){var i=n(126),r=n(29);e.exports=function(e,t,n,o){var s=!n;n||(n={});for(var a=-1,l=t.length;++a<l;){var u=t[a],c=o?o(n[u],e[u],u,n,e):void 0;void 0===c&&(c=e[u]),s?r(n,u,c):i(n,u,c)}return n}},function(e,t,n){var i=n(29),r=n(22),o=Object.prototype.hasOwnProperty;e.exports=function(e,t,n){var s=e[t];o.call(e,t)&&r(s,n)&&(void 0!==n||t in e)||i(e,t,n)}},function(e,t,n){var i=n(128),r=n(46),o=n(47),s=n(49),a=n(53),l=n(50),u=Object.prototype.hasOwnProperty;e.exports=function(e,t){var n=o(e),c=!n&&r(e),h=!n&&!c&&s(e),f=!n&&!c&&!h&&l(e),d=n||c||h||f,p=d?i(e.length,String):[],g=p.length;for(var m in e)!t&&!u.call(e,m)||d&&("length"==m||h&&("offset"==m||"parent"==m)||f&&("buffer"==m||"byteLength"==m||"byteOffset"==m)||a(m,g))||p.push(m);return p}},function(e,t){e.exports=function(e,t){for(var n=-1,i=Array(e);++n<e;)i[n]=t(n);return i}},function(e,t,n){var i=n(7),r=n(45),o=n(130),s=Object.prototype.hasOwnProperty;e.exports=function(e){if(!i(e))return o(e);var t=r(e),n=[];for(var a in e)("constructor"!=a||!t&&s.call(e,a))&&n.push(a);return n}},function(e,t){e.exports=function(e){var t=[];if(null!=e)for(var n in Object(e))t.push(n);return t}},function(e,t,n){var i=n(132),r=n(139);e.exports=function(e){return i((function(t,n){var i=-1,o=n.length,s=o>1?n[o-1]:void 0,a=o>2?n[2]:void 0;for(s=e.length>3&&"function"==typeof s?(o--,s):void 0,a&&r(n[0],n[1],a)&&(s=o<3?void 0:s,o=1),t=Object(t);++i<o;){var l=n[i];l&&e(t,l,i,s)}return t}))}},function(e,t,n){var i=n(54),r=n(133),o=n(135);e.exports=function(e,t){return o(r(e,t,i),e+"")}},function(e,t,n){var i=n(134),r=Math.max;e.exports=function(e,t,n){return t=r(void 0===t?e.length-1:t,0),function(){for(var o=arguments,s=-1,a=r(o.length-t,0),l=Array(a);++s<a;)l[s]=o[t+s];s=-1;for(var u=Array(t+1);++s<t;)u[s]=o[s];return u[t]=n(l),i(e,this,u)}}},function(e,t){e.exports=function(e,t,n){switch(n.length){case 0:return e.call(t);case 1:return e.call(t,n[0]);case 2:return e.call(t,n[0],n[1]);case 3:return e.call(t,n[0],n[1],n[2])}return e.apply(t,n)}},function(e,t,n){var i=n(136),r=n(138)(i);e.exports=r},function(e,t,n){var i=n(137),r=n(43),o=n(54),s=r?function(e,t){return r(e,"toString",{configurable:!0,enumerable:!1,value:i(t),writable:!0})}:o;e.exports=s},function(e,t){e.exports=function(e){return function(){return e}}},function(e,t){var n=Date.now;e.exports=function(e){var t=0,i=0;return function(){var r=n(),o=16-(r-i);if(i=r,o>0){if(++t>=800)return arguments[0]}else t=0;return e.apply(void 0,arguments)}}},function(e,t,n){var i=n(22),r=n(30),o=n(53),s=n(7);e.exports=function(e,t,n){if(!s(n))return!1;var a=typeof t;return!!("number"==a?r(n)&&o(t,n.length):"string"==a&&t in n)&&i(n[t],e)}},function(e,t,n){var i=n(11);e.exports=function(){return i.Date.now()}},function(e,t,n){var i=n(7),r=n(142),o=/^\s+|\s+$/g,s=/^[-+]0x[0-9a-f]+$/i,a=/^0b[01]+$/i,l=/^0o[0-7]+$/i,u=parseInt;e.exports=function(e){if("number"==typeof e)return e;if(r(e))return NaN;if(i(e)){var t="function"==typeof e.valueOf?e.valueOf():e;e=i(t)?t+"":t}if("string"!=typeof e)return 0===e?e:+e;e=e.replace(o,"");var n=a.test(e);return n||l.test(e)?u(e.slice(2),n?2:8):s.test(e)?NaN:+e}},function(e,t,n){var i=n(14),r=n(12);e.exports=function(e){return"symbol"==typeof e||r(e)&&"[object Symbol]"==i(e)}},function(e,t){e.exports=function(e,t){throw new Error("Decorating class property failed. Please ensure that proposal-class-properties is enabled and runs after the decorators transform.")}},function(e,t,n){var i=n(55),r=n(145);"string"==typeof(r=r.__esModule?r.default:r)&&(r=[[e.i,r,""]]);var o={insert:"head",singleton:!1},s=(i(e.i,r,o),r.locals?r.locals:{});e.exports=s},function(e,t,n){(t=n(56)(!1)).push([e.i,".inliner{fontsize:18px}.inliner.inliner-notebook{width:200px;position:absolute;right:20px;top:130px;z-index:100;background:white;padding:10px;border:1px solid #777}.inliner.inliner-lab{padding:10px;background:white}.inliner ul{list-style-position:inside;padding-left:0}.inliner .inline-targets{font-size:12px;margin-bottom:5px}.inliner .inline-targets .inline-targets-missing{font-size:14px;color:red;font-weight:bold}.inliner h3{font-size:14px}.inliner .inline-module-suggestion{font-size:12px}.inliner .inline-spinner{position:absolute;right:20px;top:30px;border:4px solid #ccc;border-top:4px solid #3498db;border-radius:50%;width:30px;height:30px;animation:spin 1s linear infinite}@keyframes spin{0%{transform:rotate(0deg)}100%{transform:rotate(360deg)}}.inliner .inline-diff-panel{position:fixed;bottom:0;width:1140px;left:calc(50% - 1140px/2);height:600px;background:white;border:1px solid #777}\n",""]),e.exports=t},function(e,t,n){var i=n(55),r=n(147);"string"==typeof(r=r.__esModule?r.default:r)&&(r=[[e.i,r,""]]);var o={insert:"head",singleton:!1},s=(i(e.i,r,o),r.locals?r.locals:{});e.exports=s},function(e,t,n){(t=n(56)(!1)).push([e.i,"/*! Ace-diff | github.com/ace-diff/ace-diff */.acediff__wrap{display:-webkit-box;display:-ms-flexbox;display:flex;-webkit-box-orient:horizontal;-webkit-box-direction:normal;-ms-flex-direction:row;flex-direction:row;position:absolute;bottom:0;top:0;left:0;height:100%;width:100%;overflow:auto}.acediff__gutter{-webkit-box-flex:0;-ms-flex:0 0 60px;flex:0 0 60px;border-left:1px solid #bcbcbc;border-right:1px solid #bcbcbc;overflow:hidden}.acediff__gutter,.acediff__gutter svg{background-color:#efefef}.acediff__left,.acediff__right{height:100%;-webkit-box-flex:1;-ms-flex:1;flex:1}.acediff__diffLine{background-color:#d8f2ff;border-top:1px solid #a2d7f2;border-bottom:1px solid #a2d7f2;position:absolute;z-index:4}.acediff__diffLine.targetOnly{height:0 !important;border-top:1px solid #a2d7f2;border-bottom:0;position:absolute}.acediff__connector{fill:#d8f2ff;stroke:#a2d7f2}.acediff__copy--left,.acediff__copy--right{position:relative}.acediff__copy--left div,.acediff__copy--right div{color:#000;text-shadow:1px 1px rgba(255,255,255,0.7);position:absolute;margin:2px 3px;cursor:pointer}.acediff__copy--right div:hover{color:#004ea0}.acediff__copy--left{float:right}.acediff__copy--left div{right:0}.acediff__copy--left div:hover{color:#c98100}\n",""]),e.exports=t},function(e,t,n){"use strict";var i=n(149);function r(){}function o(){}o.resetWarningCache=r,e.exports=function(){function e(e,t,n,r,o,s){if(s!==i){var a=new Error("Calling PropTypes validators directly is not supported by the `prop-types` package. Use PropTypes.checkPropTypes() to call them. Read more at http://fb.me/use-check-prop-types");throw a.name="Invariant Violation",a}}function t(){return e}e.isRequired=e;var n={array:e,bool:e,func:e,number:e,object:e,string:e,symbol:e,any:e,arrayOf:t,element:e,elementType:e,instanceOf:t,node:e,objectOf:t,oneOf:t,oneOfType:t,shape:t,exact:t,checkPropTypes:o,resetWarningCache:r};return n.PropTypes=n,n}},function(e,t,n){"use strict";e.exports="SECRET_DO_NOT_PASS_THIS_OR_YOU_WILL_BE_FIRED"},function(e,t,n){"use strict";n.r(t);var i=n(1),r=n.n(i),o=n(3),s=n.n(o),a=n(31),l=n.n(a),u=n(15),c=n.n(u),h=n(5),f=n.n(h),d=n(6),p=n.n(d),g=n(17),m=n.n(g),v=n(18),b=n.n(v),y=n(19),w=n.n(y),A=n(9),C=n.n(A),E=n(57),F=n.n(E),x=n(0),k=n.n(x),S=n(8),_=n.n(S),D=n(34),B=n.n(D),L=n(2);if(!x.useState)throw new Error("mobx-react-lite requires React with Hooks support");if(!L.n)throw new Error("mobx-react-lite requires mobx at least version 4 to be available");var T=!1;function R(){return T}
/*! *****************************************************************************
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License"); you may not use
this file except in compliance with the License. You may obtain a copy of the
License at http://www.apache.org/licenses/LICENSE-2.0

THIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED
WARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,
MERCHANTABLITY OR NON-INFRINGEMENT.

See the Apache Version 2.0 License for specific language governing permissions
and limitations under the License.
***************************************************************************** */var O=function(){return(O=Object.assign||function(e){for(var t,n=1,i=arguments.length;n<i;n++)for(var r in t=arguments[n])Object.prototype.hasOwnProperty.call(t,r)&&(e[r]=t[r]);return e}).apply(this,arguments)};function M(e,t){var n="function"==typeof Symbol&&e[Symbol.iterator];if(!n)return e;var i,r,o=n.call(e),s=[];try{for(;(void 0===t||t-- >0)&&!(i=o.next()).done;)s.push(i.value)}catch(e){r={error:e}}finally{try{i&&!i.done&&(n=o.return)&&n.call(o)}finally{if(r)throw r.error}}return s}function P(e){return e.current?Object(L.h)(e.current):"<unknown>"}var I=[];function N(){var e=M(Object(x.useState)(0),2)[1];return Object(x.useCallback)((function(){e((function(e){return e+1}))}),[])}var W={};function z(e,t,n){if(void 0===t&&(t="observed"),void 0===n&&(n=W),R())return e();var i=(n.useForceUpdate||N)(),r=Object(x.useRef)(null);r.current||(r.current=new L.b("observer("+t+")",(function(){i()})));var o,s,a=function(){r.current&&!r.current.isDisposed&&(r.current.dispose(),r.current=null)};if(Object(x.useDebugValue)(r,P),function(e){Object(x.useEffect)((function(){return e}),I)}((function(){a()})),r.current.track((function(){try{o=e()}catch(e){s=e}})),s)throw a(),s;return o}function j(e,t){if(R())return e;var n,i,r,o=O({forwardRef:!1},t),s=e.displayName||e.name,a=function(t,n){return z((function(){return e(t,n)}),s)};return a.displayName=s,n=o.forwardRef?Object(x.memo)(Object(x.forwardRef)(a)):Object(x.memo)(a),i=e,r=n,Object.keys(i).forEach((function(e){i.hasOwnProperty(e)&&!H[e]&&Object.defineProperty(r,e,Object.getOwnPropertyDescriptor(i,e))})),n.displayName=s,n}var H={$$typeof:!0,render:!0,compare:!0,type:!0};function V(e){var t=e.children,n=e.render,i=t||n;return"function"!=typeof i?null:z(i)}function U(e,t,n,i,r){var o="children"===t?"render":"children",s="function"==typeof e[t],a="function"==typeof e[o];return s&&a?new Error("MobX Observer: Do not use children and render in the same time in`"+n):s||a?null:new Error("Invalid prop `"+r+"` of type `"+typeof e[t]+"` supplied to `"+n+"`, expected `function`.")}V.propTypes={children:U,render:U},V.displayName="Observer";var K=0,G={};function q(e){return G[e]||(G[e]=function(e){if("function"==typeof Symbol)return Symbol(e);var t="__$mobx-react "+e+" ("+K+")";return K++,t}(e)),G[e]}function Y(e,t){if(Q(e,t))return!0;if("object"!=typeof e||null===e||"object"!=typeof t||null===t)return!1;var n=Object.keys(e),i=Object.keys(t);if(n.length!==i.length)return!1;for(var r=0;r<n.length;r++)if(!hasOwnProperty.call(t,n[r])||!Q(e[n[r]],t[n[r]]))return!1;return!0}function Q(e,t){return e===t?0!==e||1/e==1/t:e!=e&&t!=t}function X(e,t,n){Object.hasOwnProperty.call(e,t)?e[t]=n:Object.defineProperty(e,t,{enumerable:!1,configurable:!0,writable:!0,value:n})}var J=q("patchMixins"),Z=q("patchedDefinition");function ee(e,t){for(var n=this,i=[],r=arguments.length-2;r-- >0;)i[r]=arguments[r+2];t.locks++;try{var o;return null!=e&&(o=e.apply(this,i)),o}finally{t.locks--,0===t.locks&&t.methods.forEach((function(e){e.apply(n,i)}))}}function te(e,t){return function(){for(var n=[],i=arguments.length;i--;)n[i]=arguments[i];ee.call.apply(ee,[this,e,t].concat(n))}}function ne(e,t,n){var i=function(e,t){var n=e[J]=e[J]||{},i=n[t]=n[t]||{};return i.locks=i.locks||0,i.methods=i.methods||[],i}(e,t);i.methods.indexOf(n)<0&&i.methods.push(n);var r=Object.getOwnPropertyDescriptor(e,t);if(!r||!r[Z]){var o=function e(t,n,i,r,o){var s,a=te(o,r);return(s={})[Z]=!0,s.get=function(){return a},s.set=function(o){if(this===t)a=te(o,r);else{var s=e(this,n,i,r,o);Object.defineProperty(this,n,s)}},s.configurable=!0,s.enumerable=i,s}(e,t,r?r.enumerable:void 0,i,e[t]);Object.defineProperty(e,t,o)}}var ie=L.a||"$mobx",re=q("isUnmounted"),oe=q("skipRender"),se=q("isForcingUpdate");function ae(e,t){return R()&&console.warn("[mobx-react] It seems that a re-rendering of a React component is triggered while in static (server-side) mode. Please make sure components are rendered only once server-side."),this.state!==t||!Y(this.props,e)}function le(e,t){var n=q("reactProp_"+t+"_valueHolder"),i=q("reactProp_"+t+"_atomHolder");function r(){return this[i]||X(this,i,Object(L.g)("reactive "+t)),this[i]}Object.defineProperty(e,t,{configurable:!0,enumerable:!0,get:function(){return r.call(this).reportObserved(),this[n]},set:function(e){this[se]||Y(this[n],e)?X(this,n,e):(X(this,n,e),X(this,oe,!0),r.call(this).reportChanged(),X(this,oe,!1))}})}var ue="function"==typeof Symbol&&Symbol.for,ce=ue?Symbol.for("react.forward_ref"):"function"==typeof x.forwardRef&&Object(x.forwardRef)((function(){})).$$typeof,he=ue?Symbol.for("react.memo"):"function"==typeof x.memo&&Object(x.memo)((function(){})).$$typeof;function fe(e){if(!0===e.isMobxInjector&&console.warn("Mobx observer: You are trying to use 'observer' on a component that already has 'inject'. Please apply 'observer' before applying 'inject'"),he&&e.$$typeof===he)throw new Error("Mobx observer: You are trying to use 'observer' on function component wrapped to either another observer or 'React.memo'. The observer already applies 'React.memo' for you.");if(ce&&e.$$typeof===ce){var t=e.render;if("function"!=typeof t)throw new Error("render property of ForwardRef was not a function");return Object(x.forwardRef)((function(){var e=arguments;return k.a.createElement(V,null,(function(){return t.apply(void 0,e)}))}))}return"function"!=typeof e||e.prototype&&e.prototype.render||e.isReactClass||Object.prototype.isPrototypeOf.call(x.Component,e)?function(e){var t=e.prototype;if(t.componentWillReact)throw new Error("The componentWillReact life-cycle event is no longer supported");if(e.__proto__!==x.PureComponent)if(t.shouldComponentUpdate){if(t.shouldComponentUpdate!==ae)throw new Error("It is not allowed to use shouldComponentUpdate in observer based components.")}else t.shouldComponentUpdate=ae;le(t,"props"),le(t,"state");var n=t.render;return t.render=function(){return function(e){var t=this;if(!0===R())return e.call(this);X(this,oe,!1),X(this,se,!1);var n=this.displayName||this.name||this.constructor&&(this.constructor.displayName||this.constructor.name)||"<component>",i=e.bind(this),r=!1,o=new L.b(n+".render()",(function(){if(!r&&(r=!0,!0!==t[re])){var e=!0;try{X(t,se,!0),t[oe]||x.Component.prototype.forceUpdate.call(t),e=!1}finally{X(t,se,!1),e&&o.dispose()}}}));function s(){r=!1;var e=void 0,t=void 0;if(o.track((function(){try{t=Object(L.c)(!1,i)}catch(t){e=t}})),e)throw e;return t}return o.reactComponent=this,s[ie]=o,this.render=s,s.call(this)}.call(this,n)},ne(t,"componentWillUnmount",(function(){!0!==R()&&(this.render[ie]&&this.render[ie].dispose(),this[re]=!0)})),e}(e):j(e)}k.a.createContext({});q("disposeOnUnmountProto"),q("disposeOnUnmountInst");function de(e){function t(t,n,i,r,o,s){for(var a=[],l=arguments.length-6;l-- >0;)a[l]=arguments[l+6];return Object(L.p)((function(){return r=r||"<<anonymous>>",s=s||i,null==n[i]?t?new Error("The "+o+" `"+s+"` is marked as required in `"+r+"`, but its value is `"+(null===n[i]?"null":"undefined")+"`."):null:e.apply(void 0,[n,i,r,o,s].concat(a))}))}var n=t.bind(null,!1);return n.isRequired=t.bind(null,!0),n}function pe(e){var t=typeof e;return Array.isArray(e)?"array":e instanceof RegExp?"object":function(e,t){return"symbol"===e||"Symbol"===t["@@toStringTag"]||"function"==typeof Symbol&&t instanceof Symbol}(t,e)?"symbol":t}function ge(e,t){return de((function(n,i,r,o,s){return Object(L.p)((function(){if(e&&pe(n[i])===t.toLowerCase())return null;var o;switch(t){case"Array":o=L.i;break;case"Object":o=L.k;break;case"Map":o=L.j;break;default:throw new Error("Unexpected mobxType: "+t)}var a=n[i];if(!o(a)){var l=function(e){var t=pe(e);if("object"===t){if(e instanceof Date)return"date";if(e instanceof RegExp)return"regexp"}return t}(a),u=e?" or javascript `"+t.toLowerCase()+"`":"";return new Error("Invalid prop `"+s+"` of type `"+l+"` supplied to `"+r+"`, expected `mobx.Observable"+t+"`"+u+".")}return null}))}))}function me(e,t){return de((function(n,i,r,o,s){for(var a=[],l=arguments.length-5;l-- >0;)a[l]=arguments[l+5];return Object(L.p)((function(){if("function"!=typeof t)return new Error("Property `"+s+"` of component `"+r+"` has invalid PropType notation.");var l=ge(e,"Array")(n,i,r);if(l instanceof Error)return l;for(var u=n[i],c=0;c<u.length;c++)if((l=t.apply(void 0,[u,c,r,o,s+"["+c+"]"].concat(a)))instanceof Error)return l;return null}))}))}ge(!1,"Array"),me.bind(null,!1),ge(!1,"Map"),ge(!1,"Object"),ge(!0,"Array"),me.bind(null,!0),ge(!0,"Object");if(!x.Component)throw new Error("mobx-react requires React to be available");if(!L.l)throw new Error("mobx-react requires mobx to be available");"function"==typeof S.unstable_batchedUpdates&&Object(L.f)({reactionScheduler:S.unstable_batchedUpdates});var ve=n(13),be=n.n(ve),ye=(n(72),n(73),n(35)),we=n.n(ye),Ae=n(32),Ce=n.n(Ae),Ee=n(36),Fe=n.n(Ee),xe=n(58),ke=n.n(xe);function Se(e=""){return e.replace(/\r\n/g,"\n")}function _e(e,t,n,i){const r=e+(n-e)/2;return`M ${e} ${t} C ${r},${t} ${r},${i} ${n},${i}`}function De(e,t){const n=`js-${t}-${Math.random().toString(36).substr(2,5)}`,i=e.querySelector(`.${t}`);if(i)return i.id=i.id||n,i.id;const r=document.createElement("div");return e.appendChild(r),r.className=t,r.id=n,r.id}var $e={on:function(e,t,n,i){const r="document"===e?document:document.querySelector(e);r.addEventListener(t,e=>{const t=r.querySelectorAll(n),{target:o}=e;for(let n=0,s=t.length;n<s;n+=1){let s=o;const a=t[n];for(;s&&s!==r;)s===a&&i.call(a,e),s=s.parentNode}})}};const Be=be.a.acequire||be.a.require,{Range:Le}=Be("ace/range"),Te=0,Re=-1,Oe=1,Me="right",Pe="left",Ie="rtl",Ne="ltr",We="http://www.w3.org/2000/svg",ze="specific",je="broad";function He(e){var t;(this.options=we()({mode:null,theme:null,element:null,diffGranularity:je,lockScrolling:!1,showDiffs:!0,showConnectors:!0,maxDiffs:5e3,left:{id:null,content:null,mode:null,theme:null,editable:!0,copyLinkEnabled:!0},right:{id:null,content:null,mode:null,theme:null,editable:!0,copyLinkEnabled:!0},classes:{gutterID:"acediff__gutter",diff:"acediff__diffLine",connector:"acediff__connector",newCodeConnectorLink:"acediff__newCodeConnector",newCodeConnectorLinkContent:"&#8594;",deletedCodeConnectorLink:"acediff__deletedCodeConnector",deletedCodeConnectorLinkContent:"&#8592;",copyRightContainer:"acediff__copy--right",copyLeftContainer:"acediff__copy--left"},connectorYOffset:0},e),null!==this.options.element)?(this.options.element instanceof HTMLElement?this.el=this.options.element:this.el=document.body.querySelector(this.options.element),this.el?(this.options.left.id=De(this.el,"acediff__left"),this.options.classes.gutterID=De(this.el,"acediff__gutter"),this.options.right.id=De(this.el,"acediff__right"),this.el.innerHTML=`<div class="acediff__wrap">${this.el.innerHTML}</div>`,this.editors={left:{ace:be.a.edit(this.options.left.id),markers:[],lineLengths:[]},right:{ace:be.a.edit(this.options.right.id),markers:[],lineLengths:[]},editorHeight:null},this.editors.left.ace.getSession().setMode(Ve(this,Pe)),this.editors.right.ace.getSession().setMode(Ve(this,Me)),this.editors.left.ace.setReadOnly(!this.options.left.editable),this.editors.right.ace.setReadOnly(!this.options.right.editable),this.editors.left.ace.setTheme(Ue(this,Pe)),this.editors.right.ace.setTheme(Ue(this,Me)),this.editors.left.ace.setValue(Se(this.options.left.content),-1),this.editors.right.ace.setValue(Se(this.options.right.content),-1),this.editors.editorHeight=(t=this,document.getElementById(t.options.left.id).offsetHeight),setTimeout(()=>{this.lineHeight=this.editors.left.ace.renderer.lineHeight,function(e){e.editors.left.ace.getSession().on("changeScrollTop",Fe()(t=>{Qe(e,"left",t)},16)),e.editors.right.ace.getSession().on("changeScrollTop",Fe()(t=>{Qe(e,"right",t)},16));const t=e.diff.bind(e);e.editors.left.ace.on("change",t),e.editors.right.ace.on("change",t),e.options.left.copyLinkEnabled&&$e.on(`#${e.options.classes.gutterID}`,"click",`.${e.options.classes.newCodeConnectorLink}`,t=>{Ge(e,t,Ne)});e.options.right.copyLinkEnabled&&$e.on(`#${e.options.classes.gutterID}`,"click",`.${e.options.classes.deletedCodeConnectorLink}`,t=>{Ge(e,t,Ie)});const n=Ce()(()=>{e.editors.availableHeight=document.getElementById(e.options.left.id).offsetHeight,e.diff()},250);window.addEventListener("resize",n),Ke=()=>{window.removeEventListener("resize",n)}}(this),function(e){e.copyRightContainer=document.createElement("div"),e.copyRightContainer.setAttribute("class",e.options.classes.copyRightContainer),e.copyLeftContainer=document.createElement("div"),e.copyLeftContainer.setAttribute("class",e.options.classes.copyLeftContainer),document.getElementById(e.options.classes.gutterID).appendChild(e.copyRightContainer),document.getElementById(e.options.classes.gutterID).appendChild(e.copyLeftContainer)}(this),ot(this),this.diff()},1)):console.error(`Can't find the specified element ${this.options.element}`)):console.error("You need to specify an element for Ace-diff")}function Ve(e,t){let n=e.options.mode;return t===Pe&&null!==e.options.left.mode&&(n=e.options.left.mode),t===Me&&null!==e.options.right.mode&&(n=e.options.right.mode),n}function Ue(e,t){let n=e.options.theme;return t===Pe&&null!==e.options.left.theme&&(n=e.options.left.theme),t===Me&&null!==e.options.right.theme&&(n=e.options.right.theme),n}He.prototype={setOptions(e){we()(this.options,e),this.diff()},getNumDiffs(){return this.diffs.length},getEditors(){return{left:this.editors.left.ace,right:this.editors.right.ace}},diff(){const e=new ke.a,t=this.editors.left.ace.getSession().getValue(),n=this.editors.right.ace.getSession().getValue(),i=e.diff_main(n,t);e.diff_cleanupSemantic(i),this.editors.left.lineLengths=qe(this.editors.left),this.editors.right.lineLengths=qe(this.editors.right);const r=[],o={left:0,right:0};i.forEach((function(e,t,n){const s=e[0];let a=e[1];n[t+1]&&a.endsWith("\n")&&n[t+1][1].startsWith("\n")&&(a+="\n",i[t][1]=a,i[t+1][1]=i[t+1][1].replace(/^\n/,"")),0!==a.length&&(s===Te?(o.left+=a.length,o.right+=a.length):s===Re?(r.push(Je(this,Re,o.left,o.right,a)),o.right+=a.length):s===Oe&&(r.push(Je(this,Oe,o.left,o.right,a)),o.left+=a.length))}),this),this.diffs=function(e,t){const n=[];function i(t){return e.options.diffGranularity===ze?t<1:t<=1}t.forEach((e,t)=>{if(0===t)return void n.push(e);let r=!1;for(let t=0;t<n.length;t++)if(i(Math.abs(e.leftStartLine-n[t].leftEndLine))&&i(Math.abs(e.rightStartLine-n[t].rightEndLine))){n[t].leftStartLine=Math.min(e.leftStartLine,n[t].leftStartLine),n[t].rightStartLine=Math.min(e.rightStartLine,n[t].rightStartLine),n[t].leftEndLine=Math.max(e.leftEndLine,n[t].leftEndLine),n[t].rightEndLine=Math.max(e.rightEndLine,n[t].rightEndLine),r=!0;break}r||n.push(e)});const r=[];return n.forEach(e=>{e.leftStartLine===e.leftEndLine&&e.rightStartLine===e.rightEndLine||r.push(e)}),r}(this,r),this.diffs.length>this.options.maxDiffs||(Xe(this),at(this))},destroy(){const e=this.editors.left.ace.getValue();this.editors.left.ace.destroy();let t=this.editors.left.ace.container,n=t.cloneNode(!1);n.textContent=e,t.parentNode.replaceChild(n,t);const i=this.editors.right.ace.getValue();this.editors.right.ace.destroy(),t=this.editors.right.ace.container,n=t.cloneNode(!1),n.textContent=i,t.parentNode.replaceChild(n,t),document.getElementById(this.options.classes.gutterID).innerHTML="",Ke()}};let Ke=()=>{};function Ge(e,t,n){const i=parseInt(t.target.getAttribute("data-diff-index"),10),r=e.diffs[i];let o,s,a,l,u,c;n===Ne?(o=e.editors.left,s=e.editors.right,a=r.leftStartLine,l=r.leftEndLine,u=r.rightStartLine,c=r.rightEndLine):(o=e.editors.right,s=e.editors.left,a=r.rightStartLine,l=r.rightEndLine,u=r.leftStartLine,c=r.leftEndLine);let h="";for(let e=a;e<l;e+=1)h+=`${tt(o,e)}\n`;const f=s.ace.getSession().getScrollTop();s.ace.getSession().replace(new Le(u,0,c,0),h),s.ace.getSession().setScrollTop(parseInt(f,10)),e.diff()}function qe(e){const t=e.ace.getSession().doc.getAllLines(),n=[];return t.forEach(e=>{n.push(e.length+1)}),n}function Ye(e,t,n,i,r){i<n&&(i=n);const o=`${r} ${i>n?"lines":"targetOnly"}`;i--,(t=e.editors[t]).markers.push(t.ace.session.addMarker(new Le(n,0,i,1),o,"fullLine"))}function Qe(e,t,n){Xe(e),at(e),function(e){const t=e.editors.left.ace.getSession().getScrollTop(),n=e.editors.right.ace.getSession().getScrollTop();e.copyRightContainer.style.cssText=`top: ${-t}px`,e.copyLeftContainer.style.cssText=`top: ${-n}px`}(e)}function Xe(e){e.editors.left.markers.forEach((function(e){this.editors.left.ace.getSession().removeMarker(e)}),e),e.editors.right.markers.forEach((function(e){this.editors.right.ace.getSession().removeMarker(e)}),e)}function Je(e,t,n,i,r){let o={},s=/^\n/.test(r);if(t===Oe){var a=Ze(e.editors.left,n,r),l=nt(e.editors.right,i),u=et(e.editors.right,l);const t=et(e.editors.left,a.startLine);let f=l;0===et(e.editors.left,a.startLine)&&s&&(s=!1),0===a.startChar&&it(e.editors.right,i,s)&&(f=l+1);var c=a.startLine===a.endLine,h=0;(a.startChar>0||c&&r.length<t)&&u>0&&a.startChar<t&&h++,o={leftStartLine:a.startLine,leftEndLine:a.endLine+1,rightStartLine:f,rightEndLine:f+h}}else{a=Ze(e.editors.right,i,r),l=nt(e.editors.left,n),u=et(e.editors.left,l);const t=et(e.editors.right,a.startLine);let f=l;0===et(e.editors.right,a.startLine)&&s&&(s=!1),0===a.startChar&&it(e.editors.left,n,s)&&(f=l+1);c=a.startLine===a.endLine,h=0;(a.startChar>0||c&&r.length<t)&&u>0&&a.startChar<t&&h++,o={leftStartLine:f,leftEndLine:f+h,rightStartLine:a.startLine,rightEndLine:a.endLine+1}}return o}function Ze(e,t,n){const i={startLine:0,startChar:0,endLine:0,endChar:0},r=t+n.length;let o=0,s=!1,a=!1;e.lineLengths.forEach((e,n)=>{o+=e,!s&&t<o&&(i.startLine=n,i.startChar=t-o+e,s=!0),!a&&r<=o&&(i.endLine=n,i.endChar=r-o+e,a=!0)}),i.startChar>0&&et(e,i.startLine)===i.startChar&&(i.startLine++,i.startChar=0),0===i.endChar&&i.endLine--;const l=/\n$/.test(n);return i.startChar>0&&l&&i.endLine++,i}function et(e,t){return tt(e,t).length}function tt(e,t){return e.ace.getSession().doc.getLine(t)}function nt(e,t){let n=e.ace.getSession().doc.getAllLines(),i=0,r=0;for(let e=0;e<n.length;e++)if(r+=n[e].length+1,t<=r){i=e;break}return i}function it(e,t,n){let i=e.ace.getSession().doc.getAllLines(),r=0,o=!1;for(let e=0;e<i.length;e++){r+=i[e].length+1;let s=r;if(n&&s--,t===s){o=!0;break}}return o}function rt(e){const t=document.createElement("div"),n={class:e.className,style:`top:${e.topOffset}px`,title:e.tooltip,"data-diff-index":e.diffIndex};for(const e in n)t.setAttribute(e,n[e]);return t.innerHTML=e.arrowContent,t}function ot(e){e.gutterHeight=document.getElementById(e.options.classes.gutterID).clientHeight,e.gutterWidth=document.getElementById(e.options.classes.gutterID).clientWidth;const t=st(e,Pe),n=st(e,Me),i=Math.max(t,n,e.gutterHeight);e.gutterSVG=document.createElementNS(We,"svg"),e.gutterSVG.setAttribute("width",e.gutterWidth),e.gutterSVG.setAttribute("height",i),document.getElementById(e.options.classes.gutterID).appendChild(e.gutterSVG)}function st(e,t){return(t===Pe?e.editors.left:e.editors.right).ace.getSession().getLength()*e.lineHeight}function at(e){!function(e){document.getElementById(e.options.classes.gutterID).removeChild(e.gutterSVG),ot(e)}(e),function(e){e.copyLeftContainer.innerHTML="",e.copyRightContainer.innerHTML=""}(e),e.diffs.forEach((function(e,t){this.options.showDiffs&&(Ye(this,Pe,e.leftStartLine,e.leftEndLine,this.options.classes.diff),Ye(this,Me,e.rightStartLine,e.rightEndLine,this.options.classes.diff),this.options.showConnectors&&function(e,t,n,i,r){const o=e.editors.left.ace.getSession().getScrollTop(),s=e.editors.right.ace.getSession().getScrollTop();e.connectorYOffset=1;const a=t*e.lineHeight-o+.5,l=e.gutterWidth+1,u=i*e.lineHeight-s+.5,c=n*e.lineHeight-o+e.connectorYOffset+.5,h=e.gutterWidth+1,f=r*e.lineHeight-s+e.connectorYOffset+.5,d=`${_e(-1,a,l,u)} ${`L${l},${u} ${h},${f}`} ${_e(h,f,-1,c)} ${`L-1,${c} -1,${a}`}`,p=document.createElementNS(We,"path");p.setAttribute("d",d),p.setAttribute("class",e.options.classes.connector),e.gutterSVG.appendChild(p)}(this,e.leftStartLine,e.leftEndLine,e.rightStartLine,e.rightEndLine),function(e,t,n){if(t.leftEndLine>t.leftStartLine&&e.options.left.copyLinkEnabled){var i=rt({className:e.options.classes.newCodeConnectorLink,topOffset:t.leftStartLine*e.lineHeight,tooltip:"Copy to right",diffIndex:n,arrowContent:e.options.classes.newCodeConnectorLinkContent});e.copyRightContainer.appendChild(i)}if(t.rightEndLine>t.rightStartLine&&e.options.right.copyLinkEnabled){i=rt({className:e.options.classes.deletedCodeConnectorLink,topOffset:t.rightStartLine*e.lineHeight,tooltip:"Copy to left",diffIndex:n,arrowContent:e.options.classes.deletedCodeConnectorLinkContent});e.copyLeftContainer.appendChild(i)}}(this,e,t))}),e)}var lt=He;function ut(e,t){if(null==e)return{};var n,i,r=function(e,t){if(null==e)return{};var n,i,r={},o=Object.keys(e);for(i=0;i<o.length;i++)n=o[i],t.indexOf(n)>=0||(r[n]=e[n]);return r}(e,t);if(Object.getOwnPropertySymbols){var o=Object.getOwnPropertySymbols(e);for(i=0;i<o.length;i++)n=o[i],t.indexOf(n)>=0||Object.prototype.propertyIsEnumerable.call(e,n)&&(r[n]=e[n])}return r}function ct(){return(ct=Object.assign||function(e){for(var t=1;t<arguments.length;t++){var n=arguments[t];for(var i in n)Object.prototype.hasOwnProperty.call(n,i)&&(e[i]=n[i])}return e}).apply(this,arguments)}function ht(e){return function(e){if(Array.isArray(e)){for(var t=0,n=new Array(e.length);t<e.length;t++)n[t]=e[t];return n}}(e)||function(e){if(Symbol.iterator in Object(e)||"[object Arguments]"===Object.prototype.toString.call(e))return Array.from(e)}(e)||function(){throw new TypeError("Invalid attempt to spread non-iterable instance")}()}function ft(e,t,n){return t in e?Object.defineProperty(e,t,{value:n,enumerable:!0,configurable:!0,writable:!0}):e[t]=n,e}function dt(e){for(var t=1;t<arguments.length;t++){var n=null!=arguments[t]?arguments[t]:{},i=Object.keys(Object(n));"function"==typeof Object.getOwnPropertySymbols&&(i=i.concat(Object.getOwnPropertySymbols(n).filter((function(e){return Object.getOwnPropertyDescriptor(n,e).enumerable})))),i.forEach((function(t){ft(e,t,n[t])}))}return e}function pt(e,t){if(!(e instanceof t))throw new TypeError("Cannot call a class as a function")}function gt(e,t){for(var n=0;n<t.length;n++){var i=t[n];i.enumerable=i.enumerable||!1,i.configurable=!0,"value"in i&&(i.writable=!0),Object.defineProperty(e,i.key,i)}}function mt(e,t,n){return t&&gt(e.prototype,t),n&&gt(e,n),e}function vt(e){return(vt="function"==typeof Symbol&&"symbol"==typeof Symbol.iterator?function(e){return typeof e}:function(e){return e&&"function"==typeof Symbol&&e.constructor===Symbol&&e!==Symbol.prototype?"symbol":typeof e})(e)}function bt(e){if(void 0===e)throw new ReferenceError("this hasn't been initialised - super() hasn't been called");return e}function yt(e,t){return!t||"object"!==vt(t)&&"function"!=typeof t?bt(e):t}function wt(e){return(wt=Object.setPrototypeOf?Object.getPrototypeOf:function(e){return e.__proto__||Object.getPrototypeOf(e)})(e)}function At(e,t){return(At=Object.setPrototypeOf||function(e,t){return e.__proto__=t,e})(e,t)}function Ct(e,t){if("function"!=typeof t&&null!==t)throw new TypeError("Super expression must either be null or a function");e.prototype=Object.create(t&&t.prototype,{constructor:{value:e,writable:!0,configurable:!0}}),t&&At(e,t)}function Et(e,t){if(e.length!==t.length)return!1;for(var n=0;n<e.length;n++)if(e[n]!==t[n])return!1;return!0}var Ft=function(e,t){var n;void 0===t&&(t=Et);var i,r=[],o=!1;return function(){for(var s=[],a=0;a<arguments.length;a++)s[a]=arguments[a];return o&&n===this&&t(s,r)?i:(i=e.apply(this,s),o=!0,n=this,r=s,i)}},xt=n(60),kt=n.n(xt);var St=function(){function e(e){this.isSpeedy=void 0===e.speedy||e.speedy,this.tags=[],this.ctr=0,this.nonce=e.nonce,this.key=e.key,this.container=e.container,this.before=null}var t=e.prototype;return t.insert=function(e){if(this.ctr%(this.isSpeedy?65e3:1)==0){var t,n=function(e){var t=document.createElement("style");return t.setAttribute("data-emotion",e.key),void 0!==e.nonce&&t.setAttribute("nonce",e.nonce),t.appendChild(document.createTextNode("")),t}(this);t=0===this.tags.length?this.before:this.tags[this.tags.length-1].nextSibling,this.container.insertBefore(n,t),this.tags.push(n)}var i=this.tags[this.tags.length-1];if(this.isSpeedy){var r=function(e){if(e.sheet)return e.sheet;for(var t=0;t<document.styleSheets.length;t++)if(document.styleSheets[t].ownerNode===e)return document.styleSheets[t]}(i);try{var o=105===e.charCodeAt(1)&&64===e.charCodeAt(0);r.insertRule(e,o?0:r.cssRules.length)}catch(e){0}}else i.appendChild(document.createTextNode(e));this.ctr++},t.flush=function(){this.tags.forEach((function(e){return e.parentNode.removeChild(e)})),this.tags=[],this.ctr=0},e}();var _t=function(e){function t(e,t,i){var r=t.trim().split(p);t=r;var o=r.length,s=e.length;switch(s){case 0:case 1:var a=0;for(e=0===s?"":e[0]+" ";a<o;++a)t[a]=n(e,t[a],i).trim();break;default:var l=a=0;for(t=[];a<o;++a)for(var u=0;u<s;++u)t[l++]=n(e[u]+" ",r[a],i).trim()}return t}function n(e,t,n){var i=t.charCodeAt(0);switch(33>i&&(i=(t=t.trim()).charCodeAt(0)),i){case 38:return t.replace(g,"$1"+e.trim());case 58:return e.trim()+t.replace(g,"$1"+e.trim());default:if(0<1*n&&0<t.indexOf("\f"))return t.replace(g,(58===e.charCodeAt(0)?"":"$1")+e.trim())}return e+t}function i(e,t,n,o){var s=e+";",a=2*t+3*n+4*o;if(944===a){e=s.indexOf(":",9)+1;var l=s.substring(e,s.length-1).trim();return l=s.substring(0,e).trim()+l+";",1===D||2===D&&r(l,1)?"-webkit-"+l+l:l}if(0===D||2===D&&!r(s,1))return s;switch(a){case 1015:return 97===s.charCodeAt(10)?"-webkit-"+s+s:s;case 951:return 116===s.charCodeAt(3)?"-webkit-"+s+s:s;case 963:return 110===s.charCodeAt(5)?"-webkit-"+s+s:s;case 1009:if(100!==s.charCodeAt(4))break;case 969:case 942:return"-webkit-"+s+s;case 978:return"-webkit-"+s+"-moz-"+s+s;case 1019:case 983:return"-webkit-"+s+"-moz-"+s+"-ms-"+s+s;case 883:if(45===s.charCodeAt(8))return"-webkit-"+s+s;if(0<s.indexOf("image-set(",11))return s.replace(x,"$1-webkit-$2")+s;break;case 932:if(45===s.charCodeAt(4))switch(s.charCodeAt(5)){case 103:return"-webkit-box-"+s.replace("-grow","")+"-webkit-"+s+"-ms-"+s.replace("grow","positive")+s;case 115:return"-webkit-"+s+"-ms-"+s.replace("shrink","negative")+s;case 98:return"-webkit-"+s+"-ms-"+s.replace("basis","preferred-size")+s}return"-webkit-"+s+"-ms-"+s+s;case 964:return"-webkit-"+s+"-ms-flex-"+s+s;case 1023:if(99!==s.charCodeAt(8))break;return"-webkit-box-pack"+(l=s.substring(s.indexOf(":",15)).replace("flex-","").replace("space-between","justify"))+"-webkit-"+s+"-ms-flex-pack"+l+s;case 1005:return f.test(s)?s.replace(h,":-webkit-")+s.replace(h,":-moz-")+s:s;case 1e3:switch(t=(l=s.substring(13).trim()).indexOf("-")+1,l.charCodeAt(0)+l.charCodeAt(t)){case 226:l=s.replace(y,"tb");break;case 232:l=s.replace(y,"tb-rl");break;case 220:l=s.replace(y,"lr");break;default:return s}return"-webkit-"+s+"-ms-"+l+s;case 1017:if(-1===s.indexOf("sticky",9))break;case 975:switch(t=(s=e).length-10,a=(l=(33===s.charCodeAt(t)?s.substring(0,t):s).substring(e.indexOf(":",7)+1).trim()).charCodeAt(0)+(0|l.charCodeAt(7))){case 203:if(111>l.charCodeAt(8))break;case 115:s=s.replace(l,"-webkit-"+l)+";"+s;break;case 207:case 102:s=s.replace(l,"-webkit-"+(102<a?"inline-":"")+"box")+";"+s.replace(l,"-webkit-"+l)+";"+s.replace(l,"-ms-"+l+"box")+";"+s}return s+";";case 938:if(45===s.charCodeAt(5))switch(s.charCodeAt(6)){case 105:return l=s.replace("-items",""),"-webkit-"+s+"-webkit-box-"+l+"-ms-flex-"+l+s;case 115:return"-webkit-"+s+"-ms-flex-item-"+s.replace(C,"")+s;default:return"-webkit-"+s+"-ms-flex-line-pack"+s.replace("align-content","").replace(C,"")+s}break;case 973:case 989:if(45!==s.charCodeAt(3)||122===s.charCodeAt(4))break;case 931:case 953:if(!0===F.test(e))return 115===(l=e.substring(e.indexOf(":")+1)).charCodeAt(0)?i(e.replace("stretch","fill-available"),t,n,o).replace(":fill-available",":stretch"):s.replace(l,"-webkit-"+l)+s.replace(l,"-moz-"+l.replace("fill-",""))+s;break;case 962:if(s="-webkit-"+s+(102===s.charCodeAt(5)?"-ms-"+s:"")+s,211===n+o&&105===s.charCodeAt(13)&&0<s.indexOf("transform",10))return s.substring(0,s.indexOf(";",27)+1).replace(d,"$1-webkit-$2")+s}return s}function r(e,t){var n=e.indexOf(1===t?":":"{"),i=e.substring(0,3!==t?n:10);return n=e.substring(n+1,e.length-1),T(2!==t?i:i.replace(E,"$1"),n,t)}function o(e,t){var n=i(t,t.charCodeAt(0),t.charCodeAt(1),t.charCodeAt(2));return n!==t+";"?n.replace(A," or ($1)").substring(4):"("+t+")"}function s(e,t,n,i,r,o,s,a,u,c){for(var h,f=0,d=t;f<L;++f)switch(h=B[f].call(l,e,d,n,i,r,o,s,a,u,c)){case void 0:case!1:case!0:case null:break;default:d=h}if(d!==t)return d}function a(e){return void 0!==(e=e.prefix)&&(T=null,e?"function"!=typeof e?D=1:(D=2,T=e):D=0),a}function l(e,n){var a=e;if(33>a.charCodeAt(0)&&(a=a.trim()),a=[a],0<L){var l=s(-1,n,a,a,S,k,0,0,0,0);void 0!==l&&"string"==typeof l&&(n=l)}var h=function e(n,a,l,h,f){for(var d,p,g,y,A,C=0,E=0,F=0,x=0,B=0,T=0,O=g=d=0,M=0,P=0,I=0,N=0,W=l.length,z=W-1,j="",H="",V="",U="";M<W;){if(p=l.charCodeAt(M),M===z&&0!==E+x+F+C&&(0!==E&&(p=47===E?10:47),x=F=C=0,W++,z++),0===E+x+F+C){if(M===z&&(0<P&&(j=j.replace(c,"")),0<j.trim().length)){switch(p){case 32:case 9:case 59:case 13:case 10:break;default:j+=l.charAt(M)}p=59}switch(p){case 123:for(d=(j=j.trim()).charCodeAt(0),g=1,N=++M;M<W;){switch(p=l.charCodeAt(M)){case 123:g++;break;case 125:g--;break;case 47:switch(p=l.charCodeAt(M+1)){case 42:case 47:e:{for(O=M+1;O<z;++O)switch(l.charCodeAt(O)){case 47:if(42===p&&42===l.charCodeAt(O-1)&&M+2!==O){M=O+1;break e}break;case 10:if(47===p){M=O+1;break e}}M=O}}break;case 91:p++;case 40:p++;case 34:case 39:for(;M++<z&&l.charCodeAt(M)!==p;);}if(0===g)break;M++}switch(g=l.substring(N,M),0===d&&(d=(j=j.replace(u,"").trim()).charCodeAt(0)),d){case 64:switch(0<P&&(j=j.replace(c,"")),p=j.charCodeAt(1)){case 100:case 109:case 115:case 45:P=a;break;default:P=$}if(N=(g=e(a,P,g,p,f+1)).length,0<L&&(A=s(3,g,P=t($,j,I),a,S,k,N,p,f,h),j=P.join(""),void 0!==A&&0===(N=(g=A.trim()).length)&&(p=0,g="")),0<N)switch(p){case 115:j=j.replace(w,o);case 100:case 109:case 45:g=j+"{"+g+"}";break;case 107:g=(j=j.replace(m,"$1 $2"))+"{"+g+"}",g=1===D||2===D&&r("@"+g,3)?"@-webkit-"+g+"@"+g:"@"+g;break;default:g=j+g,112===h&&(H+=g,g="")}else g="";break;default:g=e(a,t(a,j,I),g,h,f+1)}V+=g,g=I=P=O=d=0,j="",p=l.charCodeAt(++M);break;case 125:case 59:if(1<(N=(j=(0<P?j.replace(c,""):j).trim()).length))switch(0===O&&(d=j.charCodeAt(0),45===d||96<d&&123>d)&&(N=(j=j.replace(" ",":")).length),0<L&&void 0!==(A=s(1,j,a,n,S,k,H.length,h,f,h))&&0===(N=(j=A.trim()).length)&&(j="\0\0"),d=j.charCodeAt(0),p=j.charCodeAt(1),d){case 0:break;case 64:if(105===p||99===p){U+=j+l.charAt(M);break}default:58!==j.charCodeAt(N-1)&&(H+=i(j,d,p,j.charCodeAt(2)))}I=P=O=d=0,j="",p=l.charCodeAt(++M)}}switch(p){case 13:case 10:47===E?E=0:0===1+d&&107!==h&&0<j.length&&(P=1,j+="\0"),0<L*R&&s(0,j,a,n,S,k,H.length,h,f,h),k=1,S++;break;case 59:case 125:if(0===E+x+F+C){k++;break}default:switch(k++,y=l.charAt(M),p){case 9:case 32:if(0===x+C+E)switch(B){case 44:case 58:case 9:case 32:y="";break;default:32!==p&&(y=" ")}break;case 0:y="\\0";break;case 12:y="\\f";break;case 11:y="\\v";break;case 38:0===x+E+C&&(P=I=1,y="\f"+y);break;case 108:if(0===x+E+C+_&&0<O)switch(M-O){case 2:112===B&&58===l.charCodeAt(M-3)&&(_=B);case 8:111===T&&(_=T)}break;case 58:0===x+E+C&&(O=M);break;case 44:0===E+F+x+C&&(P=1,y+="\r");break;case 34:case 39:0===E&&(x=x===p?0:0===x?p:x);break;case 91:0===x+E+F&&C++;break;case 93:0===x+E+F&&C--;break;case 41:0===x+E+C&&F--;break;case 40:if(0===x+E+C){if(0===d)switch(2*B+3*T){case 533:break;default:d=1}F++}break;case 64:0===E+F+x+C+O+g&&(g=1);break;case 42:case 47:if(!(0<x+C+F))switch(E){case 0:switch(2*p+3*l.charCodeAt(M+1)){case 235:E=47;break;case 220:N=M,E=42}break;case 42:47===p&&42===B&&N+2!==M&&(33===l.charCodeAt(N+2)&&(H+=l.substring(N,M+1)),y="",E=0)}}0===E&&(j+=y)}T=B,B=p,M++}if(0<(N=H.length)){if(P=a,0<L&&(void 0!==(A=s(2,H,P,n,S,k,N,h,f,h))&&0===(H=A).length))return U+H+V;if(H=P.join(",")+"{"+H+"}",0!=D*_){switch(2!==D||r(H,2)||(_=0),_){case 111:H=H.replace(b,":-moz-$1")+H;break;case 112:H=H.replace(v,"::-webkit-input-$1")+H.replace(v,"::-moz-$1")+H.replace(v,":-ms-input-$1")+H}_=0}}return U+H+V}($,a,n,0,0);return 0<L&&(void 0!==(l=s(-2,h,a,a,S,k,h.length,0,0,0))&&(h=l)),"",_=0,k=S=1,h}var u=/^\0+/g,c=/[\0\r\f]/g,h=/: */g,f=/zoo|gra/,d=/([,: ])(transform)/g,p=/,\r+?/g,g=/([\t\r\n ])*\f?&/g,m=/@(k\w+)\s*(\S*)\s*/,v=/::(place)/g,b=/:(read-only)/g,y=/[svh]\w+-[tblr]{2}/,w=/\(\s*(.*)\s*\)/g,A=/([\s\S]*?);/g,C=/-self|flex-/g,E=/[^]*?(:[rp][el]a[\w-]+)[^]*/,F=/stretch|:\s*\w+\-(?:conte|avail)/,x=/([^-])(image-set\()/,k=1,S=1,_=0,D=1,$=[],B=[],L=0,T=null,R=0;return l.use=function e(t){switch(t){case void 0:case null:L=B.length=0;break;default:if("function"==typeof t)B[L++]=t;else if("object"==typeof t)for(var n=0,i=t.length;n<i;++n)e(t[n]);else R=0|!!t}return e},l.set=a,void 0!==e&&a(e),l};function Dt(e){e&&$t.current.insert(e+"}")}var $t={current:null},Bt=function(e,t,n,i,r,o,s,a,l,u){switch(e){case 1:switch(t.charCodeAt(0)){case 64:return $t.current.insert(t+";"),"";case 108:if(98===t.charCodeAt(2))return""}break;case 2:if(0===a)return t+"/*|*/";break;case 3:switch(a){case 102:case 112:return $t.current.insert(n[0]+t),"";default:return t+(0===u?"/*|*/":"")}case-2:t.split("/*|*/}").forEach(Dt)}},Lt=function(e){void 0===e&&(e={});var t,n=e.key||"css";void 0!==e.prefix&&(t={prefix:e.prefix});var i=new _t(t);var r,o={};r=e.container||document.head;var s,a=document.querySelectorAll("style[data-emotion-"+n+"]");Array.prototype.forEach.call(a,(function(e){e.getAttribute("data-emotion-"+n).split(" ").forEach((function(e){o[e]=!0})),e.parentNode!==r&&r.appendChild(e)})),i.use(e.stylisPlugins)(Bt),s=function(e,t,n,r){var o=t.name;$t.current=n,i(e,t.styles),r&&(l.inserted[o]=!0)};var l={key:n,sheet:new St({key:n,container:r,nonce:e.nonce,speedy:e.speedy}),nonce:e.nonce,inserted:o,registered:{},insert:s};return l};function Tt(e,t,n){var i="";return n.split(" ").forEach((function(n){void 0!==e[n]?t.push(e[n]):i+=n+" "})),i}var Rt=function(e,t,n){var i=e.key+"-"+t.name;if(!1===n&&void 0===e.registered[i]&&(e.registered[i]=t.styles),void 0===e.inserted[t.name]){var r=t;do{e.insert("."+i,r,e.sheet,!0);r=r.next}while(void 0!==r)}};var Ot=function(e){for(var t,n=e.length,i=n^n,r=0;n>=4;)t=1540483477*(65535&(t=255&e.charCodeAt(r)|(255&e.charCodeAt(++r))<<8|(255&e.charCodeAt(++r))<<16|(255&e.charCodeAt(++r))<<24))+((1540483477*(t>>>16)&65535)<<16),i=1540483477*(65535&i)+((1540483477*(i>>>16)&65535)<<16)^(t=1540483477*(65535&(t^=t>>>24))+((1540483477*(t>>>16)&65535)<<16)),n-=4,++r;switch(n){case 3:i^=(255&e.charCodeAt(r+2))<<16;case 2:i^=(255&e.charCodeAt(r+1))<<8;case 1:i=1540483477*(65535&(i^=255&e.charCodeAt(r)))+((1540483477*(i>>>16)&65535)<<16)}return i=1540483477*(65535&(i^=i>>>13))+((1540483477*(i>>>16)&65535)<<16),((i^=i>>>15)>>>0).toString(36)},Mt={animationIterationCount:1,borderImageOutset:1,borderImageSlice:1,borderImageWidth:1,boxFlex:1,boxFlexGroup:1,boxOrdinalGroup:1,columnCount:1,columns:1,flex:1,flexGrow:1,flexPositive:1,flexShrink:1,flexNegative:1,flexOrder:1,gridRow:1,gridRowEnd:1,gridRowSpan:1,gridRowStart:1,gridColumn:1,gridColumnEnd:1,gridColumnSpan:1,gridColumnStart:1,msGridRow:1,msGridRowSpan:1,msGridColumn:1,msGridColumnSpan:1,fontWeight:1,lineHeight:1,opacity:1,order:1,orphans:1,tabSize:1,widows:1,zIndex:1,zoom:1,WebkitLineClamp:1,fillOpacity:1,floodOpacity:1,stopOpacity:1,strokeDasharray:1,strokeDashoffset:1,strokeMiterlimit:1,strokeOpacity:1,strokeWidth:1};var Pt=/[A-Z]|^ms/g,It=/_EMO_([^_]+?)_([^]*?)_EMO_/g,Nt=function(e){return 45===e.charCodeAt(1)},Wt=function(e){return null!=e&&"boolean"!=typeof e},zt=function(e){var t={};return function(n){return void 0===t[n]&&(t[n]=e(n)),t[n]}}((function(e){return Nt(e)?e:e.replace(Pt,"-$&").toLowerCase()})),jt=function(e,t){switch(e){case"animation":case"animationName":if("string"==typeof t)return t.replace(It,(function(e,t,n){return Vt={name:t,styles:n,next:Vt},t}))}return 1===Mt[e]||Nt(e)||"number"!=typeof t||0===t?t:t+"px"};function Ht(e,t,n,i){if(null==n)return"";if(void 0!==n.__emotion_styles)return n;switch(typeof n){case"boolean":return"";case"object":if(1===n.anim)return Vt={name:n.name,styles:n.styles,next:Vt},n.name;if(void 0!==n.styles){var r=n.next;if(void 0!==r)for(;void 0!==r;)Vt={name:r.name,styles:r.styles,next:Vt},r=r.next;return n.styles+";"}return function(e,t,n){var i="";if(Array.isArray(n))for(var r=0;r<n.length;r++)i+=Ht(e,t,n[r],!1);else for(var o in n){var s=n[o];if("object"!=typeof s)null!=t&&void 0!==t[s]?i+=o+"{"+t[s]+"}":Wt(s)&&(i+=zt(o)+":"+jt(o,s)+";");else if(!Array.isArray(s)||"string"!=typeof s[0]||null!=t&&void 0!==t[s[0]]){var a=Ht(e,t,s,!1);switch(o){case"animation":case"animationName":i+=zt(o)+":"+a+";";break;default:i+=o+"{"+a+"}"}}else for(var l=0;l<s.length;l++)Wt(s[l])&&(i+=zt(o)+":"+jt(o,s[l])+";")}return i}(e,t,n);case"function":if(void 0!==e){var o=Vt,s=n(e);return Vt=o,Ht(e,t,s,i)}break;case"string":}if(null==t)return n;var a=t[n];return void 0===a||i?n:a}var Vt,Ut=/label:\s*([^\s;\n{]+)\s*;/g;var Kt=function(e,t,n){if(1===e.length&&"object"==typeof e[0]&&null!==e[0]&&void 0!==e[0].styles)return e[0];var i=!0,r="";Vt=void 0;var o=e[0];null==o||void 0===o.raw?(i=!1,r+=Ht(n,t,o,!1)):r+=o[0];for(var s=1;s<e.length;s++)r+=Ht(n,t,e[s],46===r.charCodeAt(r.length-1)),i&&(r+=o[s]);Ut.lastIndex=0;for(var a,l="";null!==(a=Ut.exec(r));)l+="-"+a[1];return{name:Ot(r)+l,styles:r,next:Vt}};var Gt=function(){for(var e=arguments.length,t=new Array(e),n=0;n<e;n++)t[n]=arguments[n];return Kt(t)},qt=Object(x.createContext)("undefined"!=typeof HTMLElement?Lt():null),Yt=Object(x.createContext)({}),Qt=qt.Provider,Xt=function(e){return Object(x.forwardRef)((function(t,n){return Object(x.createElement)(qt.Consumer,null,(function(i){return e(t,i,n)}))}))},Jt="__EMOTION_TYPE_PLEASE_DO_NOT_USE__",Zt=Object.prototype.hasOwnProperty,en=function(e,t,n,i){var r=null===n?t.css:t.css(n);"string"==typeof r&&void 0!==e.registered[r]&&(r=e.registered[r]);var o=t[Jt],s=[r],a="";"string"==typeof t.className?a=Tt(e.registered,s,t.className):null!=t.className&&(a=t.className+" ");var l=Kt(s);Rt(e,l,"string"==typeof o);a+=e.key+"-"+l.name;var u={};for(var c in t)Zt.call(t,c)&&"css"!==c&&c!==Jt&&(u[c]=t[c]);return u.ref=i,u.className=a,Object(x.createElement)(o,u)},tn=Xt((function(e,t,n){return"function"==typeof e.css?Object(x.createElement)(Yt.Consumer,null,(function(i){return en(t,e,i,n)})):en(t,e,null,n)}));var nn=function(e,t){var n=arguments;if(null==t||!Zt.call(t,"css"))return x.createElement.apply(void 0,n);var i=n.length,r=new Array(i);r[0]=tn;var o={};for(var s in t)Zt.call(t,s)&&(o[s]=t[s]);o[Jt]=e,r[1]=o;for(var a=2;a<i;a++)r[a]=n[a];return x.createElement.apply(null,r)},rn=(x.Component,function e(t){for(var n=t.length,i=0,r="";i<n;i++){var o=t[i];if(null!=o){var s=void 0;switch(typeof o){case"boolean":break;case"object":if(Array.isArray(o))s=e(o);else for(var a in s="",o)o[a]&&a&&(s&&(s+=" "),s+=a);break;default:s=o}s&&(r&&(r+=" "),r+=s)}}return r});function on(e,t,n){var i=[],r=Tt(e,i,n);return i.length<2?n:r+t(i)}var sn=Xt((function(e,t){return Object(x.createElement)(Yt.Consumer,null,(function(n){var i=function(){for(var e=arguments.length,n=new Array(e),i=0;i<e;i++)n[i]=arguments[i];var r=Kt(n,t.registered);return Rt(t,r,!1),t.key+"-"+r.name},r={css:i,cx:function(){for(var e=arguments.length,n=new Array(e),r=0;r<e;r++)n[r]=arguments[r];return on(t.registered,i,rn(n))},theme:n},o=e.children(r);return!0,o}))})),an=n(16),ln=n.n(an),un=function(){};function cn(e,t){return t?"-"===t[0]?e+t:e+"__"+t:e}function hn(e,t,n){var i=[n];if(t&&e)for(var r in t)t.hasOwnProperty(r)&&t[r]&&i.push("".concat(cn(e,r)));return i.filter((function(e){return e})).map((function(e){return String(e).trim()})).join(" ")}var fn=function(e){return Array.isArray(e)?e.filter(Boolean):"object"===vt(e)&&null!==e?[e]:[]};function dn(e){return[document.documentElement,document.body,window].indexOf(e)>-1}function pn(e){return dn(e)?window.pageYOffset:e.scrollTop}function gn(e,t){dn(e)?window.scrollTo(0,t):e.scrollTop=t}function mn(e,t,n,i){return n*((e=e/i-1)*e*e+1)+t}function vn(e,t){var n=arguments.length>2&&void 0!==arguments[2]?arguments[2]:200,i=arguments.length>3&&void 0!==arguments[3]?arguments[3]:un,r=pn(e),o=t-r,s=10,a=0;function l(){var t=mn(a+=s,r,o,n);gn(e,t),a<n?window.requestAnimationFrame(l):i(e)}l()}function bn(){try{return document.createEvent("TouchEvent"),!0}catch(e){return!1}}var yn=n(33),wn=n.n(yn);function An(e){var t=e.maxHeight,n=e.menuEl,i=e.minHeight,r=e.placement,o=e.shouldScroll,s=e.isFixedPosition,a=e.theme.spacing,l=function(e){var t=getComputedStyle(e),n="absolute"===t.position,i=/(auto|scroll)/,r=document.documentElement;if("fixed"===t.position)return r;for(var o=e;o=o.parentElement;)if(t=getComputedStyle(o),(!n||"static"!==t.position)&&i.test(t.overflow+t.overflowY+t.overflowX))return o;return r}(n),u={placement:"bottom",maxHeight:t};if(!n||!n.offsetParent)return u;var c=l.getBoundingClientRect().height,h=n.getBoundingClientRect(),f=h.bottom,d=h.height,p=h.top,g=n.offsetParent.getBoundingClientRect().top,m=window.innerHeight,v=pn(l),b=parseInt(getComputedStyle(n).marginBottom,10),y=parseInt(getComputedStyle(n).marginTop,10),w=g-y,A=m-p,C=w+v,E=c-v-p,F=f-m+v+b,x=v+p-y;switch(r){case"auto":case"bottom":if(A>=d)return{placement:"bottom",maxHeight:t};if(E>=d&&!s)return o&&vn(l,F,160),{placement:"bottom",maxHeight:t};if(!s&&E>=i||s&&A>=i)return o&&vn(l,F,160),{placement:"bottom",maxHeight:s?A-b:E-b};if("auto"===r||s){var k=t,S=s?w:C;return S>=i&&(k=Math.min(S-b-a.controlHeight,t)),{placement:"top",maxHeight:k}}if("bottom"===r)return gn(l,F),{placement:"bottom",maxHeight:t};break;case"top":if(w>=d)return{placement:"top",maxHeight:t};if(C>=d&&!s)return o&&vn(l,x,160),{placement:"top",maxHeight:t};if(!s&&C>=i||s&&w>=i){var _=t;return(!s&&C>=i||s&&w>=i)&&(_=s?w-y:C-y),o&&vn(l,x,160),{placement:"top",maxHeight:_}}return{placement:"bottom",maxHeight:t};default:throw new Error('Invalid placement provided "'.concat(r,'".'))}return u}var Cn=function(e){return"auto"===e?"bottom":e},En=function(e){function t(){var e,n;pt(this,t);for(var i=arguments.length,r=new Array(i),o=0;o<i;o++)r[o]=arguments[o];return ft(bt(bt(n=yt(this,(e=wt(t)).call.apply(e,[this].concat(r))))),"state",{maxHeight:n.props.maxMenuHeight,placement:null}),ft(bt(bt(n)),"getPlacement",(function(e){var t=n.props,i=t.minMenuHeight,r=t.maxMenuHeight,o=t.menuPlacement,s=t.menuPosition,a=t.menuShouldScrollIntoView,l=t.theme,u=n.context.getPortalPlacement;if(e){var c="fixed"===s,h=An({maxHeight:r,menuEl:e,minHeight:i,placement:o,shouldScroll:a&&!c,isFixedPosition:c,theme:l});u&&u(h),n.setState(h)}})),ft(bt(bt(n)),"getUpdatedProps",(function(){var e=n.props.menuPlacement,t=n.state.placement||Cn(e);return dt({},n.props,{placement:t,maxHeight:n.state.maxHeight})})),n}return Ct(t,e),mt(t,[{key:"render",value:function(){return(0,this.props.children)({ref:this.getPlacement,placerProps:this.getUpdatedProps()})}}]),t}(x.Component);ft(En,"contextTypes",{getPortalPlacement:ln.a.func});var Fn=function(e){var t=e.theme,n=t.spacing.baseUnit;return{color:t.colors.neutral40,padding:"".concat(2*n,"px ").concat(3*n,"px"),textAlign:"center"}},xn=Fn,kn=Fn,Sn=function(e){var t=e.children,n=e.className,i=e.cx,r=e.getStyles,o=e.innerProps;return nn("div",ct({css:r("noOptionsMessage",e),className:i({"menu-notice":!0,"menu-notice--no-options":!0},n)},o),t)};Sn.defaultProps={children:"No options"};var _n=function(e){var t=e.children,n=e.className,i=e.cx,r=e.getStyles,o=e.innerProps;return nn("div",ct({css:r("loadingMessage",e),className:i({"menu-notice":!0,"menu-notice--loading":!0},n)},o),t)};_n.defaultProps={children:"Loading..."};var Dn=function(e){function t(){var e,n;pt(this,t);for(var i=arguments.length,r=new Array(i),o=0;o<i;o++)r[o]=arguments[o];return ft(bt(bt(n=yt(this,(e=wt(t)).call.apply(e,[this].concat(r))))),"state",{placement:null}),ft(bt(bt(n)),"getPortalPlacement",(function(e){var t=e.placement;t!==Cn(n.props.menuPlacement)&&n.setState({placement:t})})),n}return Ct(t,e),mt(t,[{key:"getChildContext",value:function(){return{getPortalPlacement:this.getPortalPlacement}}},{key:"render",value:function(){var e=this.props,t=e.appendTo,n=e.children,i=e.controlElement,r=e.menuPlacement,o=e.menuPosition,s=e.getStyles,a="fixed"===o;if(!t&&!a||!i)return null;var l=this.state.placement||Cn(r),u=function(e){var t=e.getBoundingClientRect();return{bottom:t.bottom,height:t.height,left:t.left,right:t.right,top:t.top,width:t.width}}(i),c=a?0:window.pageYOffset,h=u[l]+c,f=nn("div",{css:s("menuPortal",{offset:h,position:o,rect:u})},n);return t?Object(S.createPortal)(f,t):f}}]),t}(x.Component);ft(Dn,"childContextTypes",{getPortalPlacement:ln.a.func});var $n=Array.isArray,Bn=Object.keys,Ln=Object.prototype.hasOwnProperty;function Tn(e,t){try{return function e(t,n){if(t===n)return!0;if(t&&n&&"object"==vt(t)&&"object"==vt(n)){var i,r,o,s=$n(t),a=$n(n);if(s&&a){if((r=t.length)!=n.length)return!1;for(i=r;0!=i--;)if(!e(t[i],n[i]))return!1;return!0}if(s!=a)return!1;var l=t instanceof Date,u=n instanceof Date;if(l!=u)return!1;if(l&&u)return t.getTime()==n.getTime();var c=t instanceof RegExp,h=n instanceof RegExp;if(c!=h)return!1;if(c&&h)return t.toString()==n.toString();var f=Bn(t);if((r=f.length)!==Bn(n).length)return!1;for(i=r;0!=i--;)if(!Ln.call(n,f[i]))return!1;for(i=r;0!=i--;)if(!("_owner"===(o=f[i])&&t.$$typeof||e(t[o],n[o])))return!1;return!0}return t!=t&&n!=n}(e,t)}catch(e){if(e.message&&e.message.match(/stack|recursion/i))return console.warn("Warning: react-fast-compare does not handle circular references.",e.name,e.message),!1;throw e}}var Rn=function(e){function t(){return pt(this,t),yt(this,wt(t).apply(this,arguments))}return Ct(t,e),mt(t,[{key:"render",value:function(){var e=this.props,t=e.children,n=e.className,i=e.cx,r=e.isMulti,o=e.getStyles,s=e.hasValue;return nn("div",{css:o("valueContainer",this.props),className:i({"value-container":!0,"value-container--is-multi":r,"value-container--has-value":s},n)},t)}}]),t}(x.Component);function On(){var e,t,n=(e=["\n  0%, 80%, 100% { opacity: 0; }\n  40% { opacity: 1; }\n"],t||(t=e.slice(0)),Object.freeze(Object.defineProperties(e,{raw:{value:Object.freeze(t)}})));return On=function(){return n},n}var Mn={name:"19bqh2r",styles:"display:inline-block;fill:currentColor;line-height:1;stroke:currentColor;stroke-width:0;"},Pn=function(e){var t=e.size,n=ut(e,["size"]);return nn("svg",ct({height:t,width:t,viewBox:"0 0 20 20","aria-hidden":"true",focusable:"false",css:Mn},n))},In=function(e){return nn(Pn,ct({size:20},e),nn("path",{d:"M14.348 14.849c-0.469 0.469-1.229 0.469-1.697 0l-2.651-3.030-2.651 3.029c-0.469 0.469-1.229 0.469-1.697 0-0.469-0.469-0.469-1.229 0-1.697l2.758-3.15-2.759-3.152c-0.469-0.469-0.469-1.228 0-1.697s1.228-0.469 1.697 0l2.652 3.031 2.651-3.031c0.469-0.469 1.228-0.469 1.697 0s0.469 1.229 0 1.697l-2.758 3.152 2.758 3.15c0.469 0.469 0.469 1.229 0 1.698z"}))},Nn=function(e){return nn(Pn,ct({size:20},e),nn("path",{d:"M4.516 7.548c0.436-0.446 1.043-0.481 1.576 0l3.908 3.747 3.908-3.747c0.533-0.481 1.141-0.446 1.574 0 0.436 0.445 0.408 1.197 0 1.615-0.406 0.418-4.695 4.502-4.695 4.502-0.217 0.223-0.502 0.335-0.787 0.335s-0.57-0.112-0.789-0.335c0 0-4.287-4.084-4.695-4.502s-0.436-1.17 0-1.615z"}))},Wn=function(e){var t=e.isFocused,n=e.theme,i=n.spacing.baseUnit,r=n.colors;return{label:"indicatorContainer",color:t?r.neutral60:r.neutral20,display:"flex",padding:2*i,transition:"color 150ms",":hover":{color:t?r.neutral80:r.neutral40}}},zn=Wn,jn=Wn,Hn=function(){var e=Gt.apply(void 0,arguments),t="animation-"+e.name;return{name:t,styles:"@keyframes "+t+"{"+e.styles+"}",anim:1,toString:function(){return"_EMO_"+this.name+"_"+this.styles+"_EMO_"}}}(On()),Vn=function(e){var t=e.delay,n=e.offset;return nn("span",{css:Gt({animation:"".concat(Hn," 1s ease-in-out ").concat(t,"ms infinite;"),backgroundColor:"currentColor",borderRadius:"1em",display:"inline-block",marginLeft:n?"1em":null,height:"1em",verticalAlign:"top",width:"1em"},"")})},Un=function(e){var t=e.className,n=e.cx,i=e.getStyles,r=e.innerProps,o=e.isRtl;return nn("div",ct({},r,{css:i("loadingIndicator",e),className:n({indicator:!0,"loading-indicator":!0},t)}),nn(Vn,{delay:0,offset:o}),nn(Vn,{delay:160,offset:!0}),nn(Vn,{delay:320,offset:!o}))};Un.defaultProps={size:4};var Kn=function(e){return{label:"input",background:0,border:0,fontSize:"inherit",opacity:e?0:1,outline:0,padding:0,color:"inherit"}},Gn=function(e){var t=e.children,n=e.innerProps;return nn("div",n,t)},qn=Gn,Yn=Gn,Qn=function(e){function t(){return pt(this,t),yt(this,wt(t).apply(this,arguments))}return Ct(t,e),mt(t,[{key:"render",value:function(){var e=this.props,t=e.children,n=e.innerProps;return nn("div",n,t||nn(In,{size:14}))}}]),t}(x.Component),Xn=function(e){function t(){return pt(this,t),yt(this,wt(t).apply(this,arguments))}return Ct(t,e),mt(t,[{key:"render",value:function(){var e=this,t=this.props,n=t.children,i=t.className,r=t.components,o=t.cx,s=t.data,a=t.getStyles,l=t.innerProps,u=t.isDisabled,c=t.removeProps,h=t.selectProps,f=r.Container,d=r.Label,p=r.Remove;return nn(sn,null,(function(t){var r=t.css,g=t.cx;return nn(f,{data:s,innerProps:dt({},l,{className:g(r(a("multiValue",e.props)),o({"multi-value":!0,"multi-value--is-disabled":u},i))}),selectProps:h},nn(d,{data:s,innerProps:{className:g(r(a("multiValueLabel",e.props)),o({"multi-value__label":!0},i))},selectProps:h},n),nn(p,{data:s,innerProps:dt({className:g(r(a("multiValueRemove",e.props)),o({"multi-value__remove":!0},i))},c),selectProps:h}))}))}}]),t}(x.Component);ft(Xn,"defaultProps",{cropWithEllipsis:!0});var Jn={ClearIndicator:function(e){var t=e.children,n=e.className,i=e.cx,r=e.getStyles,o=e.innerProps;return nn("div",ct({},o,{css:r("clearIndicator",e),className:i({indicator:!0,"clear-indicator":!0},n)}),t||nn(In,null))},Control:function(e){var t=e.children,n=e.cx,i=e.getStyles,r=e.className,o=e.isDisabled,s=e.isFocused,a=e.innerRef,l=e.innerProps,u=e.menuIsOpen;return nn("div",ct({ref:a,css:i("control",e),className:n({control:!0,"control--is-disabled":o,"control--is-focused":s,"control--menu-is-open":u},r)},l),t)},DropdownIndicator:function(e){var t=e.children,n=e.className,i=e.cx,r=e.getStyles,o=e.innerProps;return nn("div",ct({},o,{css:r("dropdownIndicator",e),className:i({indicator:!0,"dropdown-indicator":!0},n)}),t||nn(Nn,null))},DownChevron:Nn,CrossIcon:In,Group:function(e){var t=e.children,n=e.className,i=e.cx,r=e.getStyles,o=e.Heading,s=e.headingProps,a=e.label,l=e.theme,u=e.selectProps;return nn("div",{css:r("group",e),className:i({group:!0},n)},nn(o,ct({},s,{selectProps:u,theme:l,getStyles:r,cx:i}),a),nn("div",null,t))},GroupHeading:function(e){var t=e.className,n=e.cx,i=e.getStyles,r=e.theme,o=(e.selectProps,ut(e,["className","cx","getStyles","theme","selectProps"]));return nn("div",ct({css:i("groupHeading",dt({theme:r},o)),className:n({"group-heading":!0},t)},o))},IndicatorsContainer:function(e){var t=e.children,n=e.className,i=e.cx,r=e.getStyles;return nn("div",{css:r("indicatorsContainer",e),className:i({indicators:!0},n)},t)},IndicatorSeparator:function(e){var t=e.className,n=e.cx,i=e.getStyles,r=e.innerProps;return nn("span",ct({},r,{css:i("indicatorSeparator",e),className:n({"indicator-separator":!0},t)}))},Input:function(e){var t=e.className,n=e.cx,i=e.getStyles,r=e.innerRef,o=e.isHidden,s=e.isDisabled,a=e.theme,l=(e.selectProps,ut(e,["className","cx","getStyles","innerRef","isHidden","isDisabled","theme","selectProps"]));return nn("div",{css:i("input",dt({theme:a},l))},nn(wn.a,ct({className:n({input:!0},t),inputRef:r,inputStyle:Kn(o),disabled:s},l)))},LoadingIndicator:Un,Menu:function(e){var t=e.children,n=e.className,i=e.cx,r=e.getStyles,o=e.innerRef,s=e.innerProps;return nn("div",ct({css:r("menu",e),className:i({menu:!0},n)},s,{ref:o}),t)},MenuList:function(e){var t=e.children,n=e.className,i=e.cx,r=e.getStyles,o=e.isMulti,s=e.innerRef;return nn("div",{css:r("menuList",e),className:i({"menu-list":!0,"menu-list--is-multi":o},n),ref:s},t)},MenuPortal:Dn,LoadingMessage:_n,NoOptionsMessage:Sn,MultiValue:Xn,MultiValueContainer:qn,MultiValueLabel:Yn,MultiValueRemove:Qn,Option:function(e){var t=e.children,n=e.className,i=e.cx,r=e.getStyles,o=e.isDisabled,s=e.isFocused,a=e.isSelected,l=e.innerRef,u=e.innerProps;return nn("div",ct({css:r("option",e),className:i({option:!0,"option--is-disabled":o,"option--is-focused":s,"option--is-selected":a},n),ref:l},u),t)},Placeholder:function(e){var t=e.children,n=e.className,i=e.cx,r=e.getStyles,o=e.innerProps;return nn("div",ct({css:r("placeholder",e),className:i({placeholder:!0},n)},o),t)},SelectContainer:function(e){var t=e.children,n=e.className,i=e.cx,r=e.getStyles,o=e.innerProps,s=e.isDisabled,a=e.isRtl;return nn("div",ct({css:r("container",e),className:i({"--is-disabled":s,"--is-rtl":a},n)},o),t)},SingleValue:function(e){var t=e.children,n=e.className,i=e.cx,r=e.getStyles,o=e.isDisabled,s=e.innerProps;return nn("div",ct({css:r("singleValue",e),className:i({"single-value":!0,"single-value--is-disabled":o},n)},s),t)},ValueContainer:Rn},Zn=[{base:"A",letters:/[\u0041\u24B6\uFF21\u00C0\u00C1\u00C2\u1EA6\u1EA4\u1EAA\u1EA8\u00C3\u0100\u0102\u1EB0\u1EAE\u1EB4\u1EB2\u0226\u01E0\u00C4\u01DE\u1EA2\u00C5\u01FA\u01CD\u0200\u0202\u1EA0\u1EAC\u1EB6\u1E00\u0104\u023A\u2C6F]/g},{base:"AA",letters:/[\uA732]/g},{base:"AE",letters:/[\u00C6\u01FC\u01E2]/g},{base:"AO",letters:/[\uA734]/g},{base:"AU",letters:/[\uA736]/g},{base:"AV",letters:/[\uA738\uA73A]/g},{base:"AY",letters:/[\uA73C]/g},{base:"B",letters:/[\u0042\u24B7\uFF22\u1E02\u1E04\u1E06\u0243\u0182\u0181]/g},{base:"C",letters:/[\u0043\u24B8\uFF23\u0106\u0108\u010A\u010C\u00C7\u1E08\u0187\u023B\uA73E]/g},{base:"D",letters:/[\u0044\u24B9\uFF24\u1E0A\u010E\u1E0C\u1E10\u1E12\u1E0E\u0110\u018B\u018A\u0189\uA779]/g},{base:"DZ",letters:/[\u01F1\u01C4]/g},{base:"Dz",letters:/[\u01F2\u01C5]/g},{base:"E",letters:/[\u0045\u24BA\uFF25\u00C8\u00C9\u00CA\u1EC0\u1EBE\u1EC4\u1EC2\u1EBC\u0112\u1E14\u1E16\u0114\u0116\u00CB\u1EBA\u011A\u0204\u0206\u1EB8\u1EC6\u0228\u1E1C\u0118\u1E18\u1E1A\u0190\u018E]/g},{base:"F",letters:/[\u0046\u24BB\uFF26\u1E1E\u0191\uA77B]/g},{base:"G",letters:/[\u0047\u24BC\uFF27\u01F4\u011C\u1E20\u011E\u0120\u01E6\u0122\u01E4\u0193\uA7A0\uA77D\uA77E]/g},{base:"H",letters:/[\u0048\u24BD\uFF28\u0124\u1E22\u1E26\u021E\u1E24\u1E28\u1E2A\u0126\u2C67\u2C75\uA78D]/g},{base:"I",letters:/[\u0049\u24BE\uFF29\u00CC\u00CD\u00CE\u0128\u012A\u012C\u0130\u00CF\u1E2E\u1EC8\u01CF\u0208\u020A\u1ECA\u012E\u1E2C\u0197]/g},{base:"J",letters:/[\u004A\u24BF\uFF2A\u0134\u0248]/g},{base:"K",letters:/[\u004B\u24C0\uFF2B\u1E30\u01E8\u1E32\u0136\u1E34\u0198\u2C69\uA740\uA742\uA744\uA7A2]/g},{base:"L",letters:/[\u004C\u24C1\uFF2C\u013F\u0139\u013D\u1E36\u1E38\u013B\u1E3C\u1E3A\u0141\u023D\u2C62\u2C60\uA748\uA746\uA780]/g},{base:"LJ",letters:/[\u01C7]/g},{base:"Lj",letters:/[\u01C8]/g},{base:"M",letters:/[\u004D\u24C2\uFF2D\u1E3E\u1E40\u1E42\u2C6E\u019C]/g},{base:"N",letters:/[\u004E\u24C3\uFF2E\u01F8\u0143\u00D1\u1E44\u0147\u1E46\u0145\u1E4A\u1E48\u0220\u019D\uA790\uA7A4]/g},{base:"NJ",letters:/[\u01CA]/g},{base:"Nj",letters:/[\u01CB]/g},{base:"O",letters:/[\u004F\u24C4\uFF2F\u00D2\u00D3\u00D4\u1ED2\u1ED0\u1ED6\u1ED4\u00D5\u1E4C\u022C\u1E4E\u014C\u1E50\u1E52\u014E\u022E\u0230\u00D6\u022A\u1ECE\u0150\u01D1\u020C\u020E\u01A0\u1EDC\u1EDA\u1EE0\u1EDE\u1EE2\u1ECC\u1ED8\u01EA\u01EC\u00D8\u01FE\u0186\u019F\uA74A\uA74C]/g},{base:"OI",letters:/[\u01A2]/g},{base:"OO",letters:/[\uA74E]/g},{base:"OU",letters:/[\u0222]/g},{base:"P",letters:/[\u0050\u24C5\uFF30\u1E54\u1E56\u01A4\u2C63\uA750\uA752\uA754]/g},{base:"Q",letters:/[\u0051\u24C6\uFF31\uA756\uA758\u024A]/g},{base:"R",letters:/[\u0052\u24C7\uFF32\u0154\u1E58\u0158\u0210\u0212\u1E5A\u1E5C\u0156\u1E5E\u024C\u2C64\uA75A\uA7A6\uA782]/g},{base:"S",letters:/[\u0053\u24C8\uFF33\u1E9E\u015A\u1E64\u015C\u1E60\u0160\u1E66\u1E62\u1E68\u0218\u015E\u2C7E\uA7A8\uA784]/g},{base:"T",letters:/[\u0054\u24C9\uFF34\u1E6A\u0164\u1E6C\u021A\u0162\u1E70\u1E6E\u0166\u01AC\u01AE\u023E\uA786]/g},{base:"TZ",letters:/[\uA728]/g},{base:"U",letters:/[\u0055\u24CA\uFF35\u00D9\u00DA\u00DB\u0168\u1E78\u016A\u1E7A\u016C\u00DC\u01DB\u01D7\u01D5\u01D9\u1EE6\u016E\u0170\u01D3\u0214\u0216\u01AF\u1EEA\u1EE8\u1EEE\u1EEC\u1EF0\u1EE4\u1E72\u0172\u1E76\u1E74\u0244]/g},{base:"V",letters:/[\u0056\u24CB\uFF36\u1E7C\u1E7E\u01B2\uA75E\u0245]/g},{base:"VY",letters:/[\uA760]/g},{base:"W",letters:/[\u0057\u24CC\uFF37\u1E80\u1E82\u0174\u1E86\u1E84\u1E88\u2C72]/g},{base:"X",letters:/[\u0058\u24CD\uFF38\u1E8A\u1E8C]/g},{base:"Y",letters:/[\u0059\u24CE\uFF39\u1EF2\u00DD\u0176\u1EF8\u0232\u1E8E\u0178\u1EF6\u1EF4\u01B3\u024E\u1EFE]/g},{base:"Z",letters:/[\u005A\u24CF\uFF3A\u0179\u1E90\u017B\u017D\u1E92\u1E94\u01B5\u0224\u2C7F\u2C6B\uA762]/g},{base:"a",letters:/[\u0061\u24D0\uFF41\u1E9A\u00E0\u00E1\u00E2\u1EA7\u1EA5\u1EAB\u1EA9\u00E3\u0101\u0103\u1EB1\u1EAF\u1EB5\u1EB3\u0227\u01E1\u00E4\u01DF\u1EA3\u00E5\u01FB\u01CE\u0201\u0203\u1EA1\u1EAD\u1EB7\u1E01\u0105\u2C65\u0250]/g},{base:"aa",letters:/[\uA733]/g},{base:"ae",letters:/[\u00E6\u01FD\u01E3]/g},{base:"ao",letters:/[\uA735]/g},{base:"au",letters:/[\uA737]/g},{base:"av",letters:/[\uA739\uA73B]/g},{base:"ay",letters:/[\uA73D]/g},{base:"b",letters:/[\u0062\u24D1\uFF42\u1E03\u1E05\u1E07\u0180\u0183\u0253]/g},{base:"c",letters:/[\u0063\u24D2\uFF43\u0107\u0109\u010B\u010D\u00E7\u1E09\u0188\u023C\uA73F\u2184]/g},{base:"d",letters:/[\u0064\u24D3\uFF44\u1E0B\u010F\u1E0D\u1E11\u1E13\u1E0F\u0111\u018C\u0256\u0257\uA77A]/g},{base:"dz",letters:/[\u01F3\u01C6]/g},{base:"e",letters:/[\u0065\u24D4\uFF45\u00E8\u00E9\u00EA\u1EC1\u1EBF\u1EC5\u1EC3\u1EBD\u0113\u1E15\u1E17\u0115\u0117\u00EB\u1EBB\u011B\u0205\u0207\u1EB9\u1EC7\u0229\u1E1D\u0119\u1E19\u1E1B\u0247\u025B\u01DD]/g},{base:"f",letters:/[\u0066\u24D5\uFF46\u1E1F\u0192\uA77C]/g},{base:"g",letters:/[\u0067\u24D6\uFF47\u01F5\u011D\u1E21\u011F\u0121\u01E7\u0123\u01E5\u0260\uA7A1\u1D79\uA77F]/g},{base:"h",letters:/[\u0068\u24D7\uFF48\u0125\u1E23\u1E27\u021F\u1E25\u1E29\u1E2B\u1E96\u0127\u2C68\u2C76\u0265]/g},{base:"hv",letters:/[\u0195]/g},{base:"i",letters:/[\u0069\u24D8\uFF49\u00EC\u00ED\u00EE\u0129\u012B\u012D\u00EF\u1E2F\u1EC9\u01D0\u0209\u020B\u1ECB\u012F\u1E2D\u0268\u0131]/g},{base:"j",letters:/[\u006A\u24D9\uFF4A\u0135\u01F0\u0249]/g},{base:"k",letters:/[\u006B\u24DA\uFF4B\u1E31\u01E9\u1E33\u0137\u1E35\u0199\u2C6A\uA741\uA743\uA745\uA7A3]/g},{base:"l",letters:/[\u006C\u24DB\uFF4C\u0140\u013A\u013E\u1E37\u1E39\u013C\u1E3D\u1E3B\u017F\u0142\u019A\u026B\u2C61\uA749\uA781\uA747]/g},{base:"lj",letters:/[\u01C9]/g},{base:"m",letters:/[\u006D\u24DC\uFF4D\u1E3F\u1E41\u1E43\u0271\u026F]/g},{base:"n",letters:/[\u006E\u24DD\uFF4E\u01F9\u0144\u00F1\u1E45\u0148\u1E47\u0146\u1E4B\u1E49\u019E\u0272\u0149\uA791\uA7A5]/g},{base:"nj",letters:/[\u01CC]/g},{base:"o",letters:/[\u006F\u24DE\uFF4F\u00F2\u00F3\u00F4\u1ED3\u1ED1\u1ED7\u1ED5\u00F5\u1E4D\u022D\u1E4F\u014D\u1E51\u1E53\u014F\u022F\u0231\u00F6\u022B\u1ECF\u0151\u01D2\u020D\u020F\u01A1\u1EDD\u1EDB\u1EE1\u1EDF\u1EE3\u1ECD\u1ED9\u01EB\u01ED\u00F8\u01FF\u0254\uA74B\uA74D\u0275]/g},{base:"oi",letters:/[\u01A3]/g},{base:"ou",letters:/[\u0223]/g},{base:"oo",letters:/[\uA74F]/g},{base:"p",letters:/[\u0070\u24DF\uFF50\u1E55\u1E57\u01A5\u1D7D\uA751\uA753\uA755]/g},{base:"q",letters:/[\u0071\u24E0\uFF51\u024B\uA757\uA759]/g},{base:"r",letters:/[\u0072\u24E1\uFF52\u0155\u1E59\u0159\u0211\u0213\u1E5B\u1E5D\u0157\u1E5F\u024D\u027D\uA75B\uA7A7\uA783]/g},{base:"s",letters:/[\u0073\u24E2\uFF53\u00DF\u015B\u1E65\u015D\u1E61\u0161\u1E67\u1E63\u1E69\u0219\u015F\u023F\uA7A9\uA785\u1E9B]/g},{base:"t",letters:/[\u0074\u24E3\uFF54\u1E6B\u1E97\u0165\u1E6D\u021B\u0163\u1E71\u1E6F\u0167\u01AD\u0288\u2C66\uA787]/g},{base:"tz",letters:/[\uA729]/g},{base:"u",letters:/[\u0075\u24E4\uFF55\u00F9\u00FA\u00FB\u0169\u1E79\u016B\u1E7B\u016D\u00FC\u01DC\u01D8\u01D6\u01DA\u1EE7\u016F\u0171\u01D4\u0215\u0217\u01B0\u1EEB\u1EE9\u1EEF\u1EED\u1EF1\u1EE5\u1E73\u0173\u1E77\u1E75\u0289]/g},{base:"v",letters:/[\u0076\u24E5\uFF56\u1E7D\u1E7F\u028B\uA75F\u028C]/g},{base:"vy",letters:/[\uA761]/g},{base:"w",letters:/[\u0077\u24E6\uFF57\u1E81\u1E83\u0175\u1E87\u1E85\u1E98\u1E89\u2C73]/g},{base:"x",letters:/[\u0078\u24E7\uFF58\u1E8B\u1E8D]/g},{base:"y",letters:/[\u0079\u24E8\uFF59\u1EF3\u00FD\u0177\u1EF9\u0233\u1E8F\u00FF\u1EF7\u1E99\u1EF5\u01B4\u024F\u1EFF]/g},{base:"z",letters:/[\u007A\u24E9\uFF5A\u017A\u1E91\u017C\u017E\u1E93\u1E95\u01B6\u0225\u0240\u2C6C\uA763]/g}],ei=function(e){for(var t=0;t<Zn.length;t++)e=e.replace(Zn[t].letters,Zn[t].base);return e},ti=function(e){return e.replace(/^\s+|\s+$/g,"")},ni=function(e){return"".concat(e.label," ").concat(e.value)},ii={name:"1laao21-a11yText",styles:"label:a11yText;z-index:9999;border:0;clip:rect(1px, 1px, 1px, 1px);height:1px;width:1px;position:absolute;overflow:hidden;padding:0;white-space:nowrap;"},ri=function(e){return nn("span",ct({css:ii},e))},oi=function(e){function t(){return pt(this,t),yt(this,wt(t).apply(this,arguments))}return Ct(t,e),mt(t,[{key:"render",value:function(){var e=this.props,t=(e.in,e.out,e.onExited,e.appear,e.enter,e.exit,e.innerRef),n=(e.emotion,ut(e,["in","out","onExited","appear","enter","exit","innerRef","emotion"]));return nn("input",ct({ref:t},n,{css:Gt({label:"dummyInput",background:0,border:0,fontSize:"inherit",outline:0,padding:0,width:1,color:"transparent",left:-100,opacity:0,position:"relative",transform:"scale(0)"},"")}))}}]),t}(x.Component),si=function(e){function t(){return pt(this,t),yt(this,wt(t).apply(this,arguments))}return Ct(t,e),mt(t,[{key:"componentDidMount",value:function(){this.props.innerRef(Object(S.findDOMNode)(this))}},{key:"componentWillUnmount",value:function(){this.props.innerRef(null)}},{key:"render",value:function(){return this.props.children}}]),t}(x.Component),ai=["boxSizing","height","overflow","paddingRight","position"],li={boxSizing:"border-box",overflow:"hidden",position:"relative",height:"100%"};function ui(e){e.preventDefault()}function ci(e){e.stopPropagation()}function hi(){var e=this.scrollTop,t=this.scrollHeight,n=e+this.offsetHeight;0===e?this.scrollTop=1:n===t&&(this.scrollTop=e-1)}function fi(){return"ontouchstart"in window||navigator.maxTouchPoints}var di=!(!window.document||!window.document.createElement),pi=0,gi=function(e){function t(){var e,n;pt(this,t);for(var i=arguments.length,r=new Array(i),o=0;o<i;o++)r[o]=arguments[o];return ft(bt(bt(n=yt(this,(e=wt(t)).call.apply(e,[this].concat(r))))),"originalStyles",{}),ft(bt(bt(n)),"listenerOptions",{capture:!1,passive:!1}),n}return Ct(t,e),mt(t,[{key:"componentDidMount",value:function(){var e=this;if(di){var t=this.props,n=t.accountForScrollbars,i=t.touchScrollTarget,r=document.body,o=r&&r.style;if(n&&ai.forEach((function(t){var n=o&&o[t];e.originalStyles[t]=n})),n&&pi<1){var s=parseInt(this.originalStyles.paddingRight,10)||0,a=document.body?document.body.clientWidth:0,l=window.innerWidth-a+s||0;Object.keys(li).forEach((function(e){var t=li[e];o&&(o[e]=t)})),o&&(o.paddingRight="".concat(l,"px"))}r&&fi()&&(r.addEventListener("touchmove",ui,this.listenerOptions),i&&(i.addEventListener("touchstart",hi,this.listenerOptions),i.addEventListener("touchmove",ci,this.listenerOptions))),pi+=1}}},{key:"componentWillUnmount",value:function(){var e=this;if(di){var t=this.props,n=t.accountForScrollbars,i=t.touchScrollTarget,r=document.body,o=r&&r.style;pi=Math.max(pi-1,0),n&&pi<1&&ai.forEach((function(t){var n=e.originalStyles[t];o&&(o[t]=n)})),r&&fi()&&(r.removeEventListener("touchmove",ui,this.listenerOptions),i&&(i.removeEventListener("touchstart",hi,this.listenerOptions),i.removeEventListener("touchmove",ci,this.listenerOptions)))}}},{key:"render",value:function(){return null}}]),t}(x.Component);ft(gi,"defaultProps",{accountForScrollbars:!0});var mi={name:"1dsbpcp",styles:"position:fixed;left:0;bottom:0;right:0;top:0;"},vi=function(e){function t(){var e,n;pt(this,t);for(var i=arguments.length,r=new Array(i),o=0;o<i;o++)r[o]=arguments[o];return ft(bt(bt(n=yt(this,(e=wt(t)).call.apply(e,[this].concat(r))))),"state",{touchScrollTarget:null}),ft(bt(bt(n)),"getScrollTarget",(function(e){e!==n.state.touchScrollTarget&&n.setState({touchScrollTarget:e})})),ft(bt(bt(n)),"blurSelectInput",(function(){document.activeElement&&document.activeElement.blur()})),n}return Ct(t,e),mt(t,[{key:"render",value:function(){var e=this.props,t=e.children,n=e.isEnabled,i=this.state.touchScrollTarget;return n?nn("div",null,nn("div",{onClick:this.blurSelectInput,css:mi}),nn(si,{innerRef:this.getScrollTarget},t),i?nn(gi,{touchScrollTarget:i}):null):t}}]),t}(x.PureComponent),bi=function(e){function t(){var e,n;pt(this,t);for(var i=arguments.length,r=new Array(i),o=0;o<i;o++)r[o]=arguments[o];return ft(bt(bt(n=yt(this,(e=wt(t)).call.apply(e,[this].concat(r))))),"isBottom",!1),ft(bt(bt(n)),"isTop",!1),ft(bt(bt(n)),"scrollTarget",void 0),ft(bt(bt(n)),"touchStart",void 0),ft(bt(bt(n)),"cancelScroll",(function(e){e.preventDefault(),e.stopPropagation()})),ft(bt(bt(n)),"handleEventDelta",(function(e,t){var i=n.props,r=i.onBottomArrive,o=i.onBottomLeave,s=i.onTopArrive,a=i.onTopLeave,l=n.scrollTarget,u=l.scrollTop,c=l.scrollHeight,h=l.clientHeight,f=n.scrollTarget,d=t>0,p=c-h-u,g=!1;p>t&&n.isBottom&&(o&&o(e),n.isBottom=!1),d&&n.isTop&&(a&&a(e),n.isTop=!1),d&&t>p?(r&&!n.isBottom&&r(e),f.scrollTop=c,g=!0,n.isBottom=!0):!d&&-t>u&&(s&&!n.isTop&&s(e),f.scrollTop=0,g=!0,n.isTop=!0),g&&n.cancelScroll(e)})),ft(bt(bt(n)),"onWheel",(function(e){n.handleEventDelta(e,e.deltaY)})),ft(bt(bt(n)),"onTouchStart",(function(e){n.touchStart=e.changedTouches[0].clientY})),ft(bt(bt(n)),"onTouchMove",(function(e){var t=n.touchStart-e.changedTouches[0].clientY;n.handleEventDelta(e,t)})),ft(bt(bt(n)),"getScrollTarget",(function(e){n.scrollTarget=e})),n}return Ct(t,e),mt(t,[{key:"componentDidMount",value:function(){this.startListening(this.scrollTarget)}},{key:"componentWillUnmount",value:function(){this.stopListening(this.scrollTarget)}},{key:"startListening",value:function(e){e&&("function"==typeof e.addEventListener&&e.addEventListener("wheel",this.onWheel,!1),"function"==typeof e.addEventListener&&e.addEventListener("touchstart",this.onTouchStart,!1),"function"==typeof e.addEventListener&&e.addEventListener("touchmove",this.onTouchMove,!1))}},{key:"stopListening",value:function(e){"function"==typeof e.removeEventListener&&e.removeEventListener("wheel",this.onWheel,!1),"function"==typeof e.removeEventListener&&e.removeEventListener("touchstart",this.onTouchStart,!1),"function"==typeof e.removeEventListener&&e.removeEventListener("touchmove",this.onTouchMove,!1)}},{key:"render",value:function(){return k.a.createElement(si,{innerRef:this.getScrollTarget},this.props.children)}}]),t}(x.Component),yi=function(e){function t(){return pt(this,t),yt(this,wt(t).apply(this,arguments))}return Ct(t,e),mt(t,[{key:"render",value:function(){var e=this.props,t=e.isEnabled,n=ut(e,["isEnabled"]);return t?k.a.createElement(bi,n):this.props.children}}]),t}(x.Component);ft(yi,"defaultProps",{isEnabled:!0});var wi=function(e){var t=arguments.length>1&&void 0!==arguments[1]?arguments[1]:{},n=t.isSearchable,i=t.isMulti,r=t.label,o=t.isDisabled;switch(e){case"menu":return"Use Up and Down to choose options".concat(o?"":", press Enter to select the currently focused option",", press Escape to exit the menu, press Tab to select the option and exit the menu.");case"input":return"".concat(r||"Select"," is focused ").concat(n?",type to refine list":"",", press Down to open the menu, ").concat(i?" press left to focus selected values":"");case"value":return"Use left and right to toggle between focused values, press Backspace to remove the currently focused value"}},Ai=function(e,t){var n=t.value,i=t.isDisabled;if(n)switch(e){case"deselect-option":case"pop-value":case"remove-value":return"option ".concat(n,", deselected.");case"select-option":return"option ".concat(n,i?" is disabled. Select another option.":", selected.")}},Ci=function(e){return!!e.isDisabled},Ei={clearIndicator:jn,container:function(e){var t=e.isDisabled;return{label:"container",direction:e.isRtl?"rtl":null,pointerEvents:t?"none":null,position:"relative"}},control:function(e){var t=e.isDisabled,n=e.isFocused,i=e.theme,r=i.colors,o=i.borderRadius,s=i.spacing;return{label:"control",alignItems:"center",backgroundColor:t?r.neutral5:r.neutral0,borderColor:t?r.neutral10:n?r.primary:r.neutral20,borderRadius:o,borderStyle:"solid",borderWidth:1,boxShadow:n?"0 0 0 1px ".concat(r.primary):null,cursor:"default",display:"flex",flexWrap:"wrap",justifyContent:"space-between",minHeight:s.controlHeight,outline:"0 !important",position:"relative",transition:"all 100ms","&:hover":{borderColor:n?r.primary:r.neutral30}}},dropdownIndicator:zn,group:function(e){var t=e.theme.spacing;return{paddingBottom:2*t.baseUnit,paddingTop:2*t.baseUnit}},groupHeading:function(e){var t=e.theme.spacing;return{label:"group",color:"#999",cursor:"default",display:"block",fontSize:"75%",fontWeight:"500",marginBottom:"0.25em",paddingLeft:3*t.baseUnit,paddingRight:3*t.baseUnit,textTransform:"uppercase"}},indicatorsContainer:function(){return{alignItems:"center",alignSelf:"stretch",display:"flex",flexShrink:0}},indicatorSeparator:function(e){var t=e.isDisabled,n=e.theme,i=n.spacing.baseUnit,r=n.colors;return{label:"indicatorSeparator",alignSelf:"stretch",backgroundColor:t?r.neutral10:r.neutral20,marginBottom:2*i,marginTop:2*i,width:1}},input:function(e){var t=e.isDisabled,n=e.theme,i=n.spacing,r=n.colors;return{margin:i.baseUnit/2,paddingBottom:i.baseUnit/2,paddingTop:i.baseUnit/2,visibility:t?"hidden":"visible",color:r.neutral80}},loadingIndicator:function(e){var t=e.isFocused,n=e.size,i=e.theme,r=i.colors,o=i.spacing.baseUnit;return{label:"loadingIndicator",color:t?r.neutral60:r.neutral20,display:"flex",padding:2*o,transition:"color 150ms",alignSelf:"center",fontSize:n,lineHeight:1,marginRight:n,textAlign:"center",verticalAlign:"middle"}},loadingMessage:kn,menu:function(e){var t,n=e.placement,i=e.theme,r=i.borderRadius,o=i.spacing,s=i.colors;return ft(t={label:"menu"},function(e){return e?{bottom:"top",top:"bottom"}[e]:"bottom"}(n),"100%"),ft(t,"backgroundColor",s.neutral0),ft(t,"borderRadius",r),ft(t,"boxShadow","0 0 0 1px hsla(0, 0%, 0%, 0.1), 0 4px 11px hsla(0, 0%, 0%, 0.1)"),ft(t,"marginBottom",o.menuGutter),ft(t,"marginTop",o.menuGutter),ft(t,"position","absolute"),ft(t,"width","100%"),ft(t,"zIndex",1),t},menuList:function(e){var t=e.maxHeight,n=e.theme.spacing.baseUnit;return{maxHeight:t,overflowY:"auto",paddingBottom:n,paddingTop:n,position:"relative",WebkitOverflowScrolling:"touch"}},menuPortal:function(e){var t=e.rect,n=e.offset,i=e.position;return{left:t.left,position:i,top:n,width:t.width,zIndex:1}},multiValue:function(e){var t=e.theme,n=t.spacing,i=t.borderRadius;return{label:"multiValue",backgroundColor:t.colors.neutral10,borderRadius:i/2,display:"flex",margin:n.baseUnit/2,minWidth:0}},multiValueLabel:function(e){var t=e.theme,n=t.borderRadius,i=t.colors,r=e.cropWithEllipsis;return{borderRadius:n/2,color:i.neutral80,fontSize:"85%",overflow:"hidden",padding:3,paddingLeft:6,textOverflow:r?"ellipsis":null,whiteSpace:"nowrap"}},multiValueRemove:function(e){var t=e.theme,n=t.spacing,i=t.borderRadius,r=t.colors;return{alignItems:"center",borderRadius:i/2,backgroundColor:e.isFocused&&r.dangerLight,display:"flex",paddingLeft:n.baseUnit,paddingRight:n.baseUnit,":hover":{backgroundColor:r.dangerLight,color:r.danger}}},noOptionsMessage:xn,option:function(e){var t=e.isDisabled,n=e.isFocused,i=e.isSelected,r=e.theme,o=r.spacing,s=r.colors;return{label:"option",backgroundColor:i?s.primary:n?s.primary25:"transparent",color:t?s.neutral20:i?s.neutral0:"inherit",cursor:"default",display:"block",fontSize:"inherit",padding:"".concat(2*o.baseUnit,"px ").concat(3*o.baseUnit,"px"),width:"100%",userSelect:"none",WebkitTapHighlightColor:"rgba(0, 0, 0, 0)",":active":{backgroundColor:!t&&(i?s.primary:s.primary50)}}},placeholder:function(e){var t=e.theme,n=t.spacing;return{label:"placeholder",color:t.colors.neutral50,marginLeft:n.baseUnit/2,marginRight:n.baseUnit/2,position:"absolute",top:"50%",transform:"translateY(-50%)"}},singleValue:function(e){var t=e.isDisabled,n=e.theme,i=n.spacing,r=n.colors;return{label:"singleValue",color:t?r.neutral40:r.neutral80,marginLeft:i.baseUnit/2,marginRight:i.baseUnit/2,maxWidth:"calc(100% - ".concat(2*i.baseUnit,"px)"),overflow:"hidden",position:"absolute",textOverflow:"ellipsis",whiteSpace:"nowrap",top:"50%",transform:"translateY(-50%)"}},valueContainer:function(e){var t=e.theme.spacing;return{alignItems:"center",display:"flex",flex:1,flexWrap:"wrap",padding:"".concat(t.baseUnit/2,"px ").concat(2*t.baseUnit,"px"),WebkitOverflowScrolling:"touch",position:"relative",overflow:"hidden"}}};var Fi,xi={borderRadius:4,colors:{primary:"#2684FF",primary75:"#4C9AFF",primary50:"#B2D4FF",primary25:"#DEEBFF",danger:"#DE350B",dangerLight:"#FFBDAD",neutral0:"hsl(0, 0%, 100%)",neutral5:"hsl(0, 0%, 95%)",neutral10:"hsl(0, 0%, 90%)",neutral20:"hsl(0, 0%, 80%)",neutral30:"hsl(0, 0%, 70%)",neutral40:"hsl(0, 0%, 60%)",neutral50:"hsl(0, 0%, 50%)",neutral60:"hsl(0, 0%, 40%)",neutral70:"hsl(0, 0%, 30%)",neutral80:"hsl(0, 0%, 20%)",neutral90:"hsl(0, 0%, 10%)"},spacing:{baseUnit:4,controlHeight:38,menuGutter:8}},ki={backspaceRemovesValue:!0,blurInputOnSelect:bn(),captureMenuScroll:!bn(),closeMenuOnSelect:!0,closeMenuOnScroll:!1,components:{},controlShouldRenderValue:!0,escapeClearsValue:!1,filterOption:function(e,t){var n=dt({ignoreCase:!0,ignoreAccents:!0,stringify:ni,trim:!0,matchFrom:"any"},Fi),i=n.ignoreCase,r=n.ignoreAccents,o=n.stringify,s=n.trim,a=n.matchFrom,l=s?ti(t):t,u=s?ti(o(e)):o(e);return i&&(l=l.toLowerCase(),u=u.toLowerCase()),r&&(l=ei(l),u=ei(u)),"start"===a?u.substr(0,l.length)===l:u.indexOf(l)>-1},formatGroupLabel:function(e){return e.label},getOptionLabel:function(e){return e.label},getOptionValue:function(e){return e.value},isDisabled:!1,isLoading:!1,isMulti:!1,isRtl:!1,isSearchable:!0,isOptionDisabled:Ci,loadingMessage:function(){return"Loading..."},maxMenuHeight:300,minMenuHeight:140,menuIsOpen:!1,menuPlacement:"bottom",menuPosition:"absolute",menuShouldBlockScroll:!1,menuShouldScrollIntoView:!function(){try{return/Android|webOS|iPhone|iPad|iPod|BlackBerry|IEMobile|Opera Mini/i.test(navigator.userAgent)}catch(e){return!1}}(),noOptionsMessage:function(){return"No options"},openMenuOnFocus:!1,openMenuOnClick:!0,options:[],pageSize:5,placeholder:"Select...",screenReaderStatus:function(e){var t=e.count;return"".concat(t," result").concat(1!==t?"s":""," available")},styles:{},tabIndex:"0",tabSelectsValue:!0},Si=1,_i=function(e){function t(e){var n;pt(this,t),ft(bt(bt(n=yt(this,wt(t).call(this,e)))),"state",{ariaLiveSelection:"",ariaLiveContext:"",focusedOption:null,focusedValue:null,inputIsHidden:!1,isFocused:!1,menuOptions:{render:[],focusable:[]},selectValue:[]}),ft(bt(bt(n)),"blockOptionHover",!1),ft(bt(bt(n)),"isComposing",!1),ft(bt(bt(n)),"clearFocusValueOnUpdate",!1),ft(bt(bt(n)),"commonProps",void 0),ft(bt(bt(n)),"components",void 0),ft(bt(bt(n)),"hasGroups",!1),ft(bt(bt(n)),"initialTouchX",0),ft(bt(bt(n)),"initialTouchY",0),ft(bt(bt(n)),"inputIsHiddenAfterUpdate",void 0),ft(bt(bt(n)),"instancePrefix",""),ft(bt(bt(n)),"openAfterFocus",!1),ft(bt(bt(n)),"scrollToFocusedOptionOnUpdate",!1),ft(bt(bt(n)),"userIsDragging",void 0),ft(bt(bt(n)),"controlRef",null),ft(bt(bt(n)),"getControlRef",(function(e){n.controlRef=e})),ft(bt(bt(n)),"focusedOptionRef",null),ft(bt(bt(n)),"getFocusedOptionRef",(function(e){n.focusedOptionRef=e})),ft(bt(bt(n)),"menuListRef",null),ft(bt(bt(n)),"getMenuListRef",(function(e){n.menuListRef=e})),ft(bt(bt(n)),"inputRef",null),ft(bt(bt(n)),"getInputRef",(function(e){n.inputRef=e})),ft(bt(bt(n)),"cacheComponents",(function(e){n.components=dt({},Jn,{components:e}.components)})),ft(bt(bt(n)),"focus",n.focusInput),ft(bt(bt(n)),"blur",n.blurInput),ft(bt(bt(n)),"onChange",(function(e,t){var i=n.props;(0,i.onChange)(e,dt({},t,{name:i.name}))})),ft(bt(bt(n)),"setValue",(function(e){var t=arguments.length>1&&void 0!==arguments[1]?arguments[1]:"set-value",i=arguments.length>2?arguments[2]:void 0,r=n.props,o=r.closeMenuOnSelect,s=r.isMulti;n.onInputChange("",{action:"set-value"}),o&&(n.inputIsHiddenAfterUpdate=!s,n.onMenuClose()),n.clearFocusValueOnUpdate=!0,n.onChange(e,{action:t,option:i})})),ft(bt(bt(n)),"selectOption",(function(e){var t=n.props,i=t.blurInputOnSelect,r=t.isMulti,o=n.state.selectValue;if(r)if(n.isOptionSelected(e,o)){var s=n.getOptionValue(e);n.setValue(o.filter((function(e){return n.getOptionValue(e)!==s})),"deselect-option",e),n.announceAriaLiveSelection({event:"deselect-option",context:{value:n.getOptionLabel(e)}})}else n.isOptionDisabled(e,o)?n.announceAriaLiveSelection({event:"select-option",context:{value:n.getOptionLabel(e),isDisabled:!0}}):(n.setValue([].concat(ht(o),[e]),"select-option",e),n.announceAriaLiveSelection({event:"select-option",context:{value:n.getOptionLabel(e)}}));else n.isOptionDisabled(e,o)?n.announceAriaLiveSelection({event:"select-option",context:{value:n.getOptionLabel(e),isDisabled:!0}}):(n.setValue(e,"select-option"),n.announceAriaLiveSelection({event:"select-option",context:{value:n.getOptionLabel(e)}}));i&&n.blurInput()})),ft(bt(bt(n)),"removeValue",(function(e){var t=n.state.selectValue,i=n.getOptionValue(e),r=t.filter((function(e){return n.getOptionValue(e)!==i}));n.onChange(r.length?r:null,{action:"remove-value",removedValue:e}),n.announceAriaLiveSelection({event:"remove-value",context:{value:e?n.getOptionLabel(e):""}}),n.focusInput()})),ft(bt(bt(n)),"clearValue",(function(){var e=n.props.isMulti;n.onChange(e?[]:null,{action:"clear"})})),ft(bt(bt(n)),"popValue",(function(){var e=n.state.selectValue,t=e[e.length-1],i=e.slice(0,e.length-1);n.announceAriaLiveSelection({event:"pop-value",context:{value:t?n.getOptionLabel(t):""}}),n.onChange(i.length?i:null,{action:"pop-value",removedValue:t})})),ft(bt(bt(n)),"getOptionLabel",(function(e){return n.props.getOptionLabel(e)})),ft(bt(bt(n)),"getOptionValue",(function(e){return n.props.getOptionValue(e)})),ft(bt(bt(n)),"getStyles",(function(e,t){var i=Ei[e](t);i.boxSizing="border-box";var r=n.props.styles[e];return r?r(i,t):i})),ft(bt(bt(n)),"getElementId",(function(e){return"".concat(n.instancePrefix,"-").concat(e)})),ft(bt(bt(n)),"getActiveDescendentId",(function(){var e=n.props.menuIsOpen,t=n.state,i=t.menuOptions,r=t.focusedOption;if(r&&e){var o=i.focusable.indexOf(r),s=i.render[o];return s&&s.key}})),ft(bt(bt(n)),"announceAriaLiveSelection",(function(e){var t=e.event,i=e.context;n.setState({ariaLiveSelection:Ai(t,i)})})),ft(bt(bt(n)),"announceAriaLiveContext",(function(e){var t=e.event,i=e.context;n.setState({ariaLiveContext:wi(t,dt({},i,{label:n.props["aria-label"]}))})})),ft(bt(bt(n)),"onMenuMouseDown",(function(e){0===e.button&&(e.stopPropagation(),e.preventDefault(),n.focusInput())})),ft(bt(bt(n)),"onMenuMouseMove",(function(e){n.blockOptionHover=!1})),ft(bt(bt(n)),"onControlMouseDown",(function(e){var t=n.props.openMenuOnClick;n.state.isFocused?n.props.menuIsOpen?"INPUT"!==e.target.tagName&&n.onMenuClose():t&&n.openMenu("first"):(t&&(n.openAfterFocus=!0),n.focusInput()),"INPUT"!==e.target.tagName&&e.preventDefault()})),ft(bt(bt(n)),"onDropdownIndicatorMouseDown",(function(e){if(!(e&&"mousedown"===e.type&&0!==e.button||n.props.isDisabled)){var t=n.props,i=t.isMulti,r=t.menuIsOpen;n.focusInput(),r?(n.inputIsHiddenAfterUpdate=!i,n.onMenuClose()):n.openMenu("first"),e.preventDefault(),e.stopPropagation()}})),ft(bt(bt(n)),"onClearIndicatorMouseDown",(function(e){e&&"mousedown"===e.type&&0!==e.button||(n.clearValue(),e.stopPropagation(),n.openAfterFocus=!1,"touchend"===e.type?n.focusInput():setTimeout((function(){return n.focusInput()})))})),ft(bt(bt(n)),"onScroll",(function(e){"boolean"==typeof n.props.closeMenuOnScroll?e.target instanceof HTMLElement&&dn(e.target)&&n.props.onMenuClose():"function"==typeof n.props.closeMenuOnScroll&&n.props.closeMenuOnScroll(e)&&n.props.onMenuClose()})),ft(bt(bt(n)),"onCompositionStart",(function(){n.isComposing=!0})),ft(bt(bt(n)),"onCompositionEnd",(function(){n.isComposing=!1})),ft(bt(bt(n)),"onTouchStart",(function(e){var t=e.touches.item(0);t&&(n.initialTouchX=t.clientX,n.initialTouchY=t.clientY,n.userIsDragging=!1)})),ft(bt(bt(n)),"onTouchMove",(function(e){var t=e.touches.item(0);if(t){var i=Math.abs(t.clientX-n.initialTouchX),r=Math.abs(t.clientY-n.initialTouchY);n.userIsDragging=i>5||r>5}})),ft(bt(bt(n)),"onTouchEnd",(function(e){n.userIsDragging||(n.controlRef&&!n.controlRef.contains(e.target)&&n.menuListRef&&!n.menuListRef.contains(e.target)&&n.blurInput(),n.initialTouchX=0,n.initialTouchY=0)})),ft(bt(bt(n)),"onControlTouchEnd",(function(e){n.userIsDragging||n.onControlMouseDown(e)})),ft(bt(bt(n)),"onClearIndicatorTouchEnd",(function(e){n.userIsDragging||n.onClearIndicatorMouseDown(e)})),ft(bt(bt(n)),"onDropdownIndicatorTouchEnd",(function(e){n.userIsDragging||n.onDropdownIndicatorMouseDown(e)})),ft(bt(bt(n)),"handleInputChange",(function(e){var t=e.currentTarget.value;n.inputIsHiddenAfterUpdate=!1,n.onInputChange(t,{action:"input-change"}),n.onMenuOpen()})),ft(bt(bt(n)),"onInputFocus",(function(e){var t=n.props,i=t.isSearchable,r=t.isMulti;n.props.onFocus&&n.props.onFocus(e),n.inputIsHiddenAfterUpdate=!1,n.announceAriaLiveContext({event:"input",context:{isSearchable:i,isMulti:r}}),n.setState({isFocused:!0}),(n.openAfterFocus||n.props.openMenuOnFocus)&&n.openMenu("first"),n.openAfterFocus=!1})),ft(bt(bt(n)),"onInputBlur",(function(e){n.menuListRef&&n.menuListRef.contains(document.activeElement)?n.inputRef.focus():(n.props.onBlur&&n.props.onBlur(e),n.onInputChange("",{action:"input-blur"}),n.onMenuClose(),n.setState({focusedValue:null,isFocused:!1}))})),ft(bt(bt(n)),"onOptionHover",(function(e){n.blockOptionHover||n.state.focusedOption===e||n.setState({focusedOption:e})})),ft(bt(bt(n)),"shouldHideSelectedOptions",(function(){var e=n.props,t=e.hideSelectedOptions,i=e.isMulti;return void 0===t?i:t})),ft(bt(bt(n)),"onKeyDown",(function(e){var t=n.props,i=t.isMulti,r=t.backspaceRemovesValue,o=t.escapeClearsValue,s=t.inputValue,a=t.isClearable,l=t.isDisabled,u=t.menuIsOpen,c=t.onKeyDown,h=t.tabSelectsValue,f=t.openMenuOnFocus,d=n.state,p=d.focusedOption,g=d.focusedValue,m=d.selectValue;if(!(l||"function"==typeof c&&(c(e),e.defaultPrevented))){switch(n.blockOptionHover=!0,e.key){case"ArrowLeft":if(!i||s)return;n.focusValue("previous");break;case"ArrowRight":if(!i||s)return;n.focusValue("next");break;case"Delete":case"Backspace":if(s)return;if(g)n.removeValue(g);else{if(!r)return;i?n.popValue():a&&n.clearValue()}break;case"Tab":if(n.isComposing)return;if(e.shiftKey||!u||!h||!p||f&&n.isOptionSelected(p,m))return;n.selectOption(p);break;case"Enter":if(229===e.keyCode)break;if(u){if(!p)return;if(n.isComposing)return;n.selectOption(p);break}return;case"Escape":u?(n.inputIsHiddenAfterUpdate=!1,n.onInputChange("",{action:"menu-close"}),n.onMenuClose()):a&&o&&n.clearValue();break;case" ":if(s)return;if(!u){n.openMenu("first");break}if(!p)return;n.selectOption(p);break;case"ArrowUp":u?n.focusOption("up"):n.openMenu("last");break;case"ArrowDown":u?n.focusOption("down"):n.openMenu("first");break;case"PageUp":if(!u)return;n.focusOption("pageup");break;case"PageDown":if(!u)return;n.focusOption("pagedown");break;case"Home":if(!u)return;n.focusOption("first");break;case"End":if(!u)return;n.focusOption("last");break;default:return}e.preventDefault()}}));var i=e.value;n.cacheComponents=Ft(n.cacheComponents,Tn).bind(bt(bt(n))),n.cacheComponents(e.components),n.instancePrefix="react-select-"+(n.props.instanceId||++Si);var r=fn(i),o=e.menuIsOpen?n.buildMenuOptions(e,r):{render:[],focusable:[]};return n.state.menuOptions=o,n.state.selectValue=r,n}return Ct(t,e),mt(t,[{key:"componentDidMount",value:function(){this.startListeningComposition(),this.startListeningToTouch(),this.props.closeMenuOnScroll&&document&&document.addEventListener&&document.addEventListener("scroll",this.onScroll,!0),this.props.autoFocus&&this.focusInput()}},{key:"UNSAFE_componentWillReceiveProps",value:function(e){var t=this.props,n=t.options,i=t.value,r=t.menuIsOpen,o=t.inputValue;if(this.cacheComponents(e.components),e.value!==i||e.options!==n||e.menuIsOpen!==r||e.inputValue!==o){var s=fn(e.value),a=e.menuIsOpen?this.buildMenuOptions(e,s):{render:[],focusable:[]},l=this.getNextFocusedValue(s),u=this.getNextFocusedOption(a.focusable);this.setState({menuOptions:a,selectValue:s,focusedOption:u,focusedValue:l})}null!=this.inputIsHiddenAfterUpdate&&(this.setState({inputIsHidden:this.inputIsHiddenAfterUpdate}),delete this.inputIsHiddenAfterUpdate)}},{key:"componentDidUpdate",value:function(e){var t,n,i,r,o,s=this.props,a=s.isDisabled,l=s.menuIsOpen,u=this.state.isFocused;(u&&!a&&e.isDisabled||u&&l&&!e.menuIsOpen)&&this.focusInput(),this.menuListRef&&this.focusedOptionRef&&this.scrollToFocusedOptionOnUpdate&&(t=this.menuListRef,n=this.focusedOptionRef,i=t.getBoundingClientRect(),r=n.getBoundingClientRect(),o=n.offsetHeight/3,r.bottom+o>i.bottom?gn(t,Math.min(n.offsetTop+n.clientHeight-t.offsetHeight+o,t.scrollHeight)):r.top-o<i.top&&gn(t,Math.max(n.offsetTop-o,0))),this.scrollToFocusedOptionOnUpdate=!1}},{key:"componentWillUnmount",value:function(){this.stopListeningComposition(),this.stopListeningToTouch(),document.removeEventListener("scroll",this.onScroll,!0)}},{key:"onMenuOpen",value:function(){this.props.onMenuOpen()}},{key:"onMenuClose",value:function(){var e=this.props,t=e.isSearchable,n=e.isMulti;this.announceAriaLiveContext({event:"input",context:{isSearchable:t,isMulti:n}}),this.onInputChange("",{action:"menu-close"}),this.props.onMenuClose()}},{key:"onInputChange",value:function(e,t){this.props.onInputChange(e,t)}},{key:"focusInput",value:function(){this.inputRef&&this.inputRef.focus()}},{key:"blurInput",value:function(){this.inputRef&&this.inputRef.blur()}},{key:"openMenu",value:function(e){var t=this.state,n=t.menuOptions,i=t.selectValue,r=t.isFocused,o=this.props.isMulti,s="first"===e?0:n.focusable.length-1;if(!o){var a=n.focusable.indexOf(i[0]);a>-1&&(s=a)}this.scrollToFocusedOptionOnUpdate=!(r&&this.menuListRef),this.inputIsHiddenAfterUpdate=!1,this.onMenuOpen(),this.setState({focusedValue:null,focusedOption:n.focusable[s]}),this.announceAriaLiveContext({event:"menu"})}},{key:"focusValue",value:function(e){var t=this.props,n=t.isMulti,i=t.isSearchable,r=this.state,o=r.selectValue,s=r.focusedValue;if(n){this.setState({focusedOption:null});var a=o.indexOf(s);s||(a=-1,this.announceAriaLiveContext({event:"value"}));var l=o.length-1,u=-1;if(o.length){switch(e){case"previous":u=0===a?0:-1===a?l:a-1;break;case"next":a>-1&&a<l&&(u=a+1)}-1===u&&this.announceAriaLiveContext({event:"input",context:{isSearchable:i,isMulti:n}}),this.setState({inputIsHidden:-1!==u,focusedValue:o[u]})}}}},{key:"focusOption",value:function(){var e=arguments.length>0&&void 0!==arguments[0]?arguments[0]:"first",t=this.props.pageSize,n=this.state,i=n.focusedOption,r=n.menuOptions,o=r.focusable;if(o.length){var s=0,a=o.indexOf(i);i||(a=-1,this.announceAriaLiveContext({event:"menu"})),"up"===e?s=a>0?a-1:o.length-1:"down"===e?s=(a+1)%o.length:"pageup"===e?(s=a-t)<0&&(s=0):"pagedown"===e?(s=a+t)>o.length-1&&(s=o.length-1):"last"===e&&(s=o.length-1),this.scrollToFocusedOptionOnUpdate=!0,this.setState({focusedOption:o[s],focusedValue:null}),this.announceAriaLiveContext({event:"menu",context:{isDisabled:Ci(o[s])}})}}},{key:"getTheme",value:function(){return this.props.theme?"function"==typeof this.props.theme?this.props.theme(xi):dt({},xi,this.props.theme):xi}},{key:"getCommonProps",value:function(){var e=this.clearValue,t=this.getStyles,n=this.setValue,i=this.selectOption,r=this.props,o=r.classNamePrefix,s=r.isMulti,a=r.isRtl,l=r.options,u=this.state.selectValue,c=this.hasValue();return{cx:hn.bind(null,o),clearValue:e,getStyles:t,getValue:function(){return u},hasValue:c,isMulti:s,isRtl:a,options:l,selectOption:i,setValue:n,selectProps:r,theme:this.getTheme()}}},{key:"getNextFocusedValue",value:function(e){if(this.clearFocusValueOnUpdate)return this.clearFocusValueOnUpdate=!1,null;var t=this.state,n=t.focusedValue,i=t.selectValue.indexOf(n);if(i>-1){if(e.indexOf(n)>-1)return n;if(i<e.length)return e[i]}return null}},{key:"getNextFocusedOption",value:function(e){var t=this.state.focusedOption;return t&&e.indexOf(t)>-1?t:e[0]}},{key:"hasValue",value:function(){return this.state.selectValue.length>0}},{key:"hasOptions",value:function(){return!!this.state.menuOptions.render.length}},{key:"countOptions",value:function(){return this.state.menuOptions.focusable.length}},{key:"isClearable",value:function(){var e=this.props,t=e.isClearable,n=e.isMulti;return void 0===t?n:t}},{key:"isOptionDisabled",value:function(e,t){return"function"==typeof this.props.isOptionDisabled&&this.props.isOptionDisabled(e,t)}},{key:"isOptionSelected",value:function(e,t){var n=this;if(t.indexOf(e)>-1)return!0;if("function"==typeof this.props.isOptionSelected)return this.props.isOptionSelected(e,t);var i=this.getOptionValue(e);return t.some((function(e){return n.getOptionValue(e)===i}))}},{key:"filterOption",value:function(e,t){return!this.props.filterOption||this.props.filterOption(e,t)}},{key:"formatOptionLabel",value:function(e,t){if("function"==typeof this.props.formatOptionLabel){var n=this.props.inputValue,i=this.state.selectValue;return this.props.formatOptionLabel(e,{context:t,inputValue:n,selectValue:i})}return this.getOptionLabel(e)}},{key:"formatGroupLabel",value:function(e){return this.props.formatGroupLabel(e)}},{key:"startListeningComposition",value:function(){document&&document.addEventListener&&(document.addEventListener("compositionstart",this.onCompositionStart,!1),document.addEventListener("compositionend",this.onCompositionEnd,!1))}},{key:"stopListeningComposition",value:function(){document&&document.removeEventListener&&(document.removeEventListener("compositionstart",this.onCompositionStart),document.removeEventListener("compositionend",this.onCompositionEnd))}},{key:"startListeningToTouch",value:function(){document&&document.addEventListener&&(document.addEventListener("touchstart",this.onTouchStart,!1),document.addEventListener("touchmove",this.onTouchMove,!1),document.addEventListener("touchend",this.onTouchEnd,!1))}},{key:"stopListeningToTouch",value:function(){document&&document.removeEventListener&&(document.removeEventListener("touchstart",this.onTouchStart),document.removeEventListener("touchmove",this.onTouchMove),document.removeEventListener("touchend",this.onTouchEnd))}},{key:"buildMenuOptions",value:function(e,t){var n=this,i=e.inputValue,r=void 0===i?"":i,o=e.options,s=function(e,i){var o=n.isOptionDisabled(e,t),s=n.isOptionSelected(e,t),a=n.getOptionLabel(e),l=n.getOptionValue(e);if(!(n.shouldHideSelectedOptions()&&s||!n.filterOption({label:a,value:l,data:e},r))){var u=o?void 0:function(){return n.onOptionHover(e)},c=o?void 0:function(){return n.selectOption(e)},h="".concat(n.getElementId("option"),"-").concat(i);return{innerProps:{id:h,onClick:c,onMouseMove:u,onMouseOver:u,tabIndex:-1},data:e,isDisabled:o,isSelected:s,key:h,label:a,type:"option",value:l}}};return o.reduce((function(e,t,i){if(t.options){n.hasGroups||(n.hasGroups=!0);var r=t.options.map((function(t,n){var r=s(t,"".concat(i,"-").concat(n));return r&&e.focusable.push(t),r})).filter(Boolean);if(r.length){var o="".concat(n.getElementId("group"),"-").concat(i);e.render.push({type:"group",key:o,data:t,options:r})}}else{var a=s(t,"".concat(i));a&&(e.render.push(a),e.focusable.push(t))}return e}),{render:[],focusable:[]})}},{key:"constructAriaLiveMessage",value:function(){var e=this.state,t=e.ariaLiveContext,n=e.selectValue,i=e.focusedValue,r=e.focusedOption,o=this.props,s=o.options,a=o.menuIsOpen,l=o.inputValue,u=o.screenReaderStatus,c=i?function(e){var t=e.focusedValue,n=e.getOptionLabel,i=e.selectValue;return"value ".concat(n(t)," focused, ").concat(i.indexOf(t)+1," of ").concat(i.length,".")}({focusedValue:i,getOptionLabel:this.getOptionLabel,selectValue:n}):"",h=r&&a?function(e){var t=e.focusedOption,n=e.getOptionLabel,i=e.options;return"option ".concat(n(t)," focused").concat(t.isDisabled?" disabled":"",", ").concat(i.indexOf(t)+1," of ").concat(i.length,".")}({focusedOption:r,getOptionLabel:this.getOptionLabel,options:s}):"",f=function(e){var t=e.inputValue,n=e.screenReaderMessage;return"".concat(n).concat(t?" for search term "+t:"",".")}({inputValue:l,screenReaderMessage:u({count:this.countOptions()})});return"".concat(c," ").concat(h," ").concat(f," ").concat(t)}},{key:"renderInput",value:function(){var e=this.props,t=e.isDisabled,n=e.isSearchable,i=e.inputId,r=e.inputValue,o=e.tabIndex,s=this.components.Input,a=this.state.inputIsHidden,l=i||this.getElementId("input");if(!n)return k.a.createElement(oi,{id:l,innerRef:this.getInputRef,onBlur:this.onInputBlur,onChange:un,onFocus:this.onInputFocus,readOnly:!0,disabled:t,tabIndex:o,value:""});var u={"aria-autocomplete":"list","aria-label":this.props["aria-label"],"aria-labelledby":this.props["aria-labelledby"]},c=this.commonProps,h=c.cx,f=c.theme,d=c.selectProps;return k.a.createElement(s,ct({autoCapitalize:"none",autoComplete:"off",autoCorrect:"off",cx:h,getStyles:this.getStyles,id:l,innerRef:this.getInputRef,isDisabled:t,isHidden:a,onBlur:this.onInputBlur,onChange:this.handleInputChange,onFocus:this.onInputFocus,selectProps:d,spellCheck:"false",tabIndex:o,theme:f,type:"text",value:r},u))}},{key:"renderPlaceholderOrValue",value:function(){var e=this,t=this.components,n=t.MultiValue,i=t.MultiValueContainer,r=t.MultiValueLabel,o=t.MultiValueRemove,s=t.SingleValue,a=t.Placeholder,l=this.commonProps,u=this.props,c=u.controlShouldRenderValue,h=u.isDisabled,f=u.isMulti,d=u.inputValue,p=u.placeholder,g=this.state,m=g.selectValue,v=g.focusedValue,b=g.isFocused;if(!this.hasValue()||!c)return d?null:k.a.createElement(a,ct({},l,{key:"placeholder",isDisabled:h,isFocused:b}),p);if(f)return m.map((function(t,s){var a=t===v;return k.a.createElement(n,ct({},l,{components:{Container:i,Label:r,Remove:o},isFocused:a,isDisabled:h,key:e.getOptionValue(t),index:s,removeProps:{onClick:function(){return e.removeValue(t)},onTouchEnd:function(){return e.removeValue(t)},onMouseDown:function(e){e.preventDefault(),e.stopPropagation()}},data:t}),e.formatOptionLabel(t,"value"))}));if(d)return null;var y=m[0];return k.a.createElement(s,ct({},l,{data:y,isDisabled:h}),this.formatOptionLabel(y,"value"))}},{key:"renderClearIndicator",value:function(){var e=this.components.ClearIndicator,t=this.commonProps,n=this.props,i=n.isDisabled,r=n.isLoading,o=this.state.isFocused;if(!this.isClearable()||!e||i||!this.hasValue()||r)return null;var s={onMouseDown:this.onClearIndicatorMouseDown,onTouchEnd:this.onClearIndicatorTouchEnd,"aria-hidden":"true"};return k.a.createElement(e,ct({},t,{innerProps:s,isFocused:o}))}},{key:"renderLoadingIndicator",value:function(){var e=this.components.LoadingIndicator,t=this.commonProps,n=this.props,i=n.isDisabled,r=n.isLoading,o=this.state.isFocused;if(!e||!r)return null;return k.a.createElement(e,ct({},t,{innerProps:{"aria-hidden":"true"},isDisabled:i,isFocused:o}))}},{key:"renderIndicatorSeparator",value:function(){var e=this.components,t=e.DropdownIndicator,n=e.IndicatorSeparator;if(!t||!n)return null;var i=this.commonProps,r=this.props.isDisabled,o=this.state.isFocused;return k.a.createElement(n,ct({},i,{isDisabled:r,isFocused:o}))}},{key:"renderDropdownIndicator",value:function(){var e=this.components.DropdownIndicator;if(!e)return null;var t=this.commonProps,n=this.props.isDisabled,i=this.state.isFocused,r={onMouseDown:this.onDropdownIndicatorMouseDown,onTouchEnd:this.onDropdownIndicatorTouchEnd,"aria-hidden":"true"};return k.a.createElement(e,ct({},t,{innerProps:r,isDisabled:n,isFocused:i}))}},{key:"renderMenu",value:function(){var e=this,t=this.components,n=t.Group,i=t.GroupHeading,r=t.Menu,o=t.MenuList,s=t.MenuPortal,a=t.LoadingMessage,l=t.NoOptionsMessage,u=t.Option,c=this.commonProps,h=this.state,f=h.focusedOption,d=h.menuOptions,p=this.props,g=p.captureMenuScroll,m=p.inputValue,v=p.isLoading,b=p.loadingMessage,y=p.minMenuHeight,w=p.maxMenuHeight,A=p.menuIsOpen,C=p.menuPlacement,E=p.menuPosition,F=p.menuPortalTarget,x=p.menuShouldBlockScroll,S=p.menuShouldScrollIntoView,_=p.noOptionsMessage,D=p.onMenuScrollToTop,$=p.onMenuScrollToBottom;if(!A)return null;var B,L=function(t){var n=f===t.data;return t.innerRef=n?e.getFocusedOptionRef:void 0,k.a.createElement(u,ct({},c,t,{isFocused:n}),e.formatOptionLabel(t.data,"menu"))};if(this.hasOptions())B=d.render.map((function(t){if("group"===t.type){t.type;var r=ut(t,["type"]),o="".concat(t.key,"-heading");return k.a.createElement(n,ct({},c,r,{Heading:i,headingProps:{id:o},label:e.formatGroupLabel(t.data)}),t.options.map((function(e){return L(e)})))}if("option"===t.type)return L(t)}));else if(v){var T=b({inputValue:m});if(null===T)return null;B=k.a.createElement(a,c,T)}else{var R=_({inputValue:m});if(null===R)return null;B=k.a.createElement(l,c,R)}var O={minMenuHeight:y,maxMenuHeight:w,menuPlacement:C,menuPosition:E,menuShouldScrollIntoView:S},M=k.a.createElement(En,ct({},c,O),(function(t){var n=t.ref,i=t.placerProps,s=i.placement,a=i.maxHeight;return k.a.createElement(r,ct({},c,O,{innerRef:n,innerProps:{onMouseDown:e.onMenuMouseDown,onMouseMove:e.onMenuMouseMove},isLoading:v,placement:s}),k.a.createElement(yi,{isEnabled:g,onTopArrive:D,onBottomArrive:$},k.a.createElement(vi,{isEnabled:x},k.a.createElement(o,ct({},c,{innerRef:e.getMenuListRef,isLoading:v,maxHeight:a}),B))))}));return F||"fixed"===E?k.a.createElement(s,ct({},c,{appendTo:F,controlElement:this.controlRef,menuPlacement:C,menuPosition:E}),M):M}},{key:"renderFormField",value:function(){var e=this,t=this.props,n=t.delimiter,i=t.isDisabled,r=t.isMulti,o=t.name,s=this.state.selectValue;if(o&&!i){if(r){if(n){var a=s.map((function(t){return e.getOptionValue(t)})).join(n);return k.a.createElement("input",{name:o,type:"hidden",value:a})}var l=s.length>0?s.map((function(t,n){return k.a.createElement("input",{key:"i-".concat(n),name:o,type:"hidden",value:e.getOptionValue(t)})})):k.a.createElement("input",{name:o,type:"hidden"});return k.a.createElement("div",null,l)}var u=s[0]?this.getOptionValue(s[0]):"";return k.a.createElement("input",{name:o,type:"hidden",value:u})}}},{key:"renderLiveRegion",value:function(){return this.state.isFocused?k.a.createElement(ri,{"aria-live":"polite"},k.a.createElement("p",{id:"aria-selection-event"}," ",this.state.ariaLiveSelection),k.a.createElement("p",{id:"aria-context"}," ",this.constructAriaLiveMessage())):null}},{key:"render",value:function(){var e=this.components,t=e.Control,n=e.IndicatorsContainer,i=e.SelectContainer,r=e.ValueContainer,o=this.props,s=o.className,a=o.id,l=o.isDisabled,u=o.menuIsOpen,c=this.state.isFocused,h=this.commonProps=this.getCommonProps();return k.a.createElement(i,ct({},h,{className:s,innerProps:{id:a,onKeyDown:this.onKeyDown},isDisabled:l,isFocused:c}),this.renderLiveRegion(),k.a.createElement(t,ct({},h,{innerRef:this.getControlRef,innerProps:{onMouseDown:this.onControlMouseDown,onTouchEnd:this.onControlTouchEnd},isDisabled:l,isFocused:c,menuIsOpen:u}),k.a.createElement(r,ct({},h,{isDisabled:l}),this.renderPlaceholderOrValue(),this.renderInput()),k.a.createElement(n,ct({},h,{isDisabled:l}),this.renderClearIndicator(),this.renderLoadingIndicator(),this.renderIndicatorSeparator(),this.renderDropdownIndicator())),this.renderMenu(),this.renderFormField())}}]),t}(x.Component);ft(_i,"defaultProps",ki);var Di,$i,Bi,Li,Ti,Ri,Oi,Mi,Pi,Ii,Ni,Wi,zi,ji,Hi={defaultInputValue:"",defaultMenuIsOpen:!1,defaultValue:null},Vi=(x.Component,Di=_i,Bi=$i=function(e){function t(){var e,n;pt(this,t);for(var i=arguments.length,r=new Array(i),o=0;o<i;o++)r[o]=arguments[o];return ft(bt(bt(n=yt(this,(e=wt(t)).call.apply(e,[this].concat(r))))),"select",void 0),ft(bt(bt(n)),"state",{inputValue:void 0!==n.props.inputValue?n.props.inputValue:n.props.defaultInputValue,menuIsOpen:void 0!==n.props.menuIsOpen?n.props.menuIsOpen:n.props.defaultMenuIsOpen,value:void 0!==n.props.value?n.props.value:n.props.defaultValue}),ft(bt(bt(n)),"onChange",(function(e,t){n.callProp("onChange",e,t),n.setState({value:e})})),ft(bt(bt(n)),"onInputChange",(function(e,t){var i=n.callProp("onInputChange",e,t);n.setState({inputValue:void 0!==i?i:e})})),ft(bt(bt(n)),"onMenuOpen",(function(){n.callProp("onMenuOpen"),n.setState({menuIsOpen:!0})})),ft(bt(bt(n)),"onMenuClose",(function(){n.callProp("onMenuClose"),n.setState({menuIsOpen:!1})})),n}return Ct(t,e),mt(t,[{key:"focus",value:function(){this.select.focus()}},{key:"blur",value:function(){this.select.blur()}},{key:"getProp",value:function(e){return void 0!==this.props[e]?this.props[e]:this.state[e]}},{key:"callProp",value:function(e){if("function"==typeof this.props[e]){for(var t,n=arguments.length,i=new Array(n>1?n-1:0),r=1;r<n;r++)i[r-1]=arguments[r];return(t=this.props)[e].apply(t,i)}}},{key:"render",value:function(){var e=this,t=this.props,n=(t.defaultInputValue,t.defaultMenuIsOpen,t.defaultValue,ut(t,["defaultInputValue","defaultMenuIsOpen","defaultValue"]));return k.a.createElement(Di,ct({},n,{ref:function(t){e.select=t},inputValue:this.getProp("inputValue"),menuIsOpen:this.getProp("menuIsOpen"),onChange:this.onChange,onInputChange:this.onInputChange,onMenuClose:this.onMenuClose,onMenuOpen:this.onMenuOpen,value:this.getProp("value")}))}}]),t}(x.Component),ft($i,"defaultProps",Hi),Bi),Ui=n(10),Ki=n.n(Ui),Gi=n(4),qi=n.n(Gi),Yi=(n(143),n(59)),Qi=n.n(Yi),Xi=function(){function e(){f()(this,e)}return p()(e,[{key:"format_trace",value:function(e){return e.join("\n").replace(/[\u001b\u009b][[()#;?]*(?:[0-9]{1,4}(?:;[0-9]{0,4})*)?[0-9A-ORZcf-nqry=><]/g,"")}}]),e}(),Ji=function(){return Li},Zi=function(e){return Ji().execute(e,!1)},er=function(e){return Ji().execute(e,!0)},tr=function(){function e(t){f()(this,e),this.name=t}var t,n,i,o,a,l,u;return p()(e,[{key:"setup",value:(u=s()(r.a.mark((function e(t){var n;return r.a.wrap((function(e){for(;;)switch(e.prev=e.next){case 0:return n="\nfrom inliner import Inliner\nimport json\n".concat(this.name," = Inliner(").concat(JSON.stringify(t),", [], globls=globals())"),e.abrupt("return",Zi(n));case 2:case"end":return e.stop()}}),e,this)}))),function(e){return u.apply(this,arguments)})},{key:"make_program",value:(l=s()(r.a.mark((function e(){var t,n;return r.a.wrap((function(e){for(;;)switch(e.prev=e.next){case 0:return t="print(".concat(this.name,".make_program(comments=True))"),e.next=3,er(t);case 3:return n=e.sent,e.abrupt("return",n.trimEnd());case 5:case"end":return e.stop()}}),e,this)}))),function(){return l.apply(this,arguments)})},{key:"target_suggestions",value:(a=s()(r.a.mark((function e(){var t,n;return r.a.wrap((function(e){for(;;)switch(e.prev=e.next){case 0:return t="\nimport json\nprint(json.dumps(".concat(this.name,".modules()))"),e.next=3,er(t);case 3:return n=e.sent,e.abrupt("return",JSON.parse(n));case 5:case"end":return e.stop()}}),e,this)}))),function(){return a.apply(this,arguments)})},{key:"undo",value:(o=s()(r.a.mark((function e(){return r.a.wrap((function(e){for(;;)switch(e.prev=e.next){case 0:return e.abrupt("return",Zi("".concat(this.name,".undo()")));case 1:case"end":return e.stop()}}),e,this)}))),function(){return o.apply(this,arguments)})},{key:"run_pass",value:(i=s()(r.a.mark((function e(t){var n,i,o,s=arguments;return r.a.wrap((function(e){for(;;)switch(e.prev=e.next){case 0:return n=s.length>1&&void 0!==s[1]&&s[1],i=n?"fixpoint(inliner.".concat(t,")"):"".concat(t,"()"),e.next=4,er("print(".concat(this.name,".").concat(i,")"));case 4:if("True"==(o=(o=e.sent).trim())||"False"==o){e.next=8;break}throw"".concat(t,": ").concat(o);case 8:return e.abrupt("return","True"==o);case 9:case"end":return e.stop()}}),e,this)}))),function(e){return i.apply(this,arguments)})},{key:"sync_targets",value:(n=s()(r.a.mark((function e(t){var n;return r.a.wrap((function(e){for(;;)switch(e.prev=e.next){case 0:return n="\nimport json\nfor target in json.loads('".concat(JSON.stringify(t),"'):\n    ").concat(this.name,".add_target(target)"),e.abrupt("return",Zi(n));case 2:case"end":return e.stop()}}),e,this)}))),function(e){return n.apply(this,arguments)})},{key:"last_pass",value:(t=s()(r.a.mark((function e(){return r.a.wrap((function(e){for(;;)switch(e.prev=e.next){case 0:return e.abrupt("return",er("print(".concat(this.name,".history[-1][1])")));case 1:case"end":return e.stop()}}),e,this)}))),function(){return t.apply(this,arguments)})}]),e}(),nr=function(e,t,n){var i=n.value;n.value=s()(r.a.mark((function e(){var t,n,o,s,a,l=arguments;return r.a.wrap((function(e){for(;;)switch(e.prev=e.next){case 0:for((t=!this.notebook_state.show_spinner)&&(this.notebook_state.show_spinner=!0),e.prev=2,n=l.length,o=new Array(n),s=0;s<n;s++)o[s]=l[s];return e.next=6,i.apply(this,o);case 6:return a=e.sent,e.abrupt("return",a);case 8:return e.prev=8,t&&(this.notebook_state.show_spinner=!1),e.finish(8);case 11:case"end":return e.stop()}}),e,this,[[2,,8,11]])})))},ir=(Ti=function(){function e(t,n){f()(this,e),Ki()(this,"cell_id",Ri,this),Ki()(this,"targets",Oi,this),Ki()(this,"target_suggestions",Mi,this),Ki()(this,"program_history",Pi,this),this.cell_id=n.current_cell,this.set_cell_text=t,this.bridge=new tr("inliner"),this.notebook_state=n}var t,n,i,o,a,l,u;return p()(e,[{key:"setup",value:(u=s()(r.a.mark((function e(t){var n=this;return r.a.wrap((function(e){for(;;)switch(e.prev=e.next){case 0:return e.next=2,this.bridge.setup(t);case 2:return e.next=4,this.update_cell();case 4:Object(L.d)((function(){n.bridge.sync_targets(n.targets)}));case 5:case"end":return e.stop()}}),e,this)}))),function(e){return u.apply(this,arguments)})},{key:"update_cell",value:(l=s()(r.a.mark((function e(){var t;return r.a.wrap((function(e){for(;;)switch(e.prev=e.next){case 0:return e.next=2,this.bridge.make_program();case 2:t=e.sent,this.set_cell_text(t),this.program_history.push(t);case 5:case"end":return e.stop()}}),e,this)}))),function(){return l.apply(this,arguments)})},{key:"refresh_target_suggestions",value:(a=s()(r.a.mark((function e(){var t,n=this;return r.a.wrap((function(e){for(;;)switch(e.prev=e.next){case 0:return e.next=2,this.bridge.target_suggestions();case 2:t=e.sent,t=Qi.a.pickBy(t,(function(e,t){return-1==n.targets.indexOf(t)})),this.target_suggestions.clear(),this.target_suggestions.merge(t);case 6:case"end":return e.stop()}}),e,this)}))),function(){return a.apply(this,arguments)})},{key:"undo",value:(o=s()(r.a.mark((function e(){return r.a.wrap((function(e){for(;;)switch(e.prev=e.next){case 0:return e.next=2,this.bridge.undo();case 2:return e.next=4,this.update_cell();case 4:this.program_history.pop(),this.program_history.pop();case 6:case"end":return e.stop()}}),e,this)}))),function(){return o.apply(this,arguments)})},{key:"run_pass",value:(i=s()(r.a.mark((function e(t){var n,i,o=arguments;return r.a.wrap((function(e){for(;;)switch(e.prev=e.next){case 0:if(n=o.length>1&&void 0!==o[1]&&o[1],"inline"!=t||0!=this.targets.length){e.next=3;break}throw"Must have at least one inline target to run inline pass";case 3:return e.next=5,this.bridge.run_pass(t,n);case 5:return i=e.sent,e.next=8,this.update_cell();case 8:return e.abrupt("return",i);case 9:case"end":return e.stop()}}),e,this)}))),function(e){return i.apply(this,arguments)})},{key:"last_pass",value:(n=s()(r.a.mark((function e(){return r.a.wrap((function(e){for(;;)switch(e.prev=e.next){case 0:return e.abrupt("return",this.bridge.last_pass());case 1:case"end":return e.stop()}}),e,this)}))),function(){return n.apply(this,arguments)})},{key:"autoschedule_noinline",value:function(){return this.autoschedule(!1)}},{key:"autoschedule",value:(t=s()(r.a.mark((function e(){var t,n,i,o=this,a=arguments;return r.a.wrap((function(e){for(;;)switch(e.prev=e.next){case 0:return t=!(a.length>0&&void 0!==a[0])||a[0],n=function(){var e=s()(r.a.mark((function e(t){var n,i,s,a,l,u,c,h;return r.a.wrap((function(e){for(;;)switch(e.prev=e.next){case 0:n=!1,i=!0,s=!1,a=void 0,e.prev=5,l=t[Symbol.iterator]();case 7:if(i=(u=l.next()).done){e.next=18;break}return c=u.value,console.log("Running ".concat(c)),e.next=12,o.run_pass(c,"deadcode"==c);case 12:h=e.sent,n=n||h,console.log("Finished ".concat(c,", change: ").concat(h));case 15:i=!0,e.next=7;break;case 18:e.next=24;break;case 20:e.prev=20,e.t0=e.catch(5),s=!0,a=e.t0;case 24:e.prev=24,e.prev=25,i||null==l.return||l.return();case 27:if(e.prev=27,!s){e.next=30;break}throw a;case 30:return e.finish(27);case 31:return e.finish(24);case 32:if(n){e.next=34;break}return e.abrupt("return");case 34:e.next=0;break;case 36:case"end":return e.stop()}}),e,null,[[5,20,24,32],[25,,27,31]])})));return function(t){return e.apply(this,arguments)}}(),i=["deadcode","copy_propagation","lifetimes","simplify_varargs","partial_eval","expand_tuples","clean_imports","array_index"],t&&i.unshift("inline"),e.next=6,n(i);case 6:return e.next=8,this.run_pass("expand_self");case 8:return e.next=10,n(i);case 10:return e.next=12,this.run_pass("remove_suffixes");case 12:case"end":return e.stop()}}),e,this)}))),function(){return t.apply(this,arguments)})}]),e}(),Ri=qi()(Ti.prototype,"cell_id",[L.l],{configurable:!0,enumerable:!0,writable:!0,initializer:null}),Oi=qi()(Ti.prototype,"targets",[L.l],{configurable:!0,enumerable:!0,writable:!0,initializer:function(){return[]}}),Mi=qi()(Ti.prototype,"target_suggestions",[L.l],{configurable:!0,enumerable:!0,writable:!0,initializer:function(){return new Map}}),Pi=qi()(Ti.prototype,"program_history",[L.l],{configurable:!0,enumerable:!0,writable:!0,initializer:function(){return[]}}),qi()(Ti.prototype,"setup",[nr],Object.getOwnPropertyDescriptor(Ti.prototype,"setup"),Ti.prototype),qi()(Ti.prototype,"update_cell",[nr],Object.getOwnPropertyDescriptor(Ti.prototype,"update_cell"),Ti.prototype),qi()(Ti.prototype,"refresh_target_suggestions",[nr],Object.getOwnPropertyDescriptor(Ti.prototype,"refresh_target_suggestions"),Ti.prototype),qi()(Ti.prototype,"undo",[nr],Object.getOwnPropertyDescriptor(Ti.prototype,"undo"),Ti.prototype),qi()(Ti.prototype,"run_pass",[nr],Object.getOwnPropertyDescriptor(Ti.prototype,"run_pass"),Ti.prototype),qi()(Ti.prototype,"autoschedule",[nr],Object.getOwnPropertyDescriptor(Ti.prototype,"autoschedule"),Ti.prototype),Ti),rr=(Ii=function(){function e(){f()(this,e),Ki()(this,"states",Ni,this),Ki()(this,"current_cell",Wi,this),Ki()(this,"show_spinner",zi,this)}return p()(e,[{key:"add_state",value:function(e){this.states[e.cell_id]=e}},{key:"current_state",get:function(){return this.states[this.current_cell]}}]),e}(),Ni=qi()(Ii.prototype,"states",[L.l],{configurable:!0,enumerable:!0,writable:!0,initializer:function(){return{}}}),Wi=qi()(Ii.prototype,"current_cell",[L.l],{configurable:!0,enumerable:!0,writable:!0,initializer:function(){return null}}),zi=qi()(Ii.prototype,"show_spinner",[L.l],{configurable:!0,enumerable:!0,writable:!0,initializer:function(){return!1}}),qi()(Ii.prototype,"current_state",[L.e],Object.getOwnPropertyDescriptor(Ii.prototype,"current_state"),Ii.prototype),Ii);n(144),n(146);function or(e,t){var n=Object.keys(e);if(Object.getOwnPropertySymbols){var i=Object.getOwnPropertySymbols(e);t&&(i=i.filter((function(t){return Object.getOwnPropertyDescriptor(e,t).enumerable}))),n.push.apply(n,i)}return n}var sr=k.a.createContext(null),ar=function(){var e=s()(r.a.mark((function e(t,n,i){var o,s;return r.a.wrap((function(e){for(;;)switch(e.prev=e.next){case 0:return e.prev=0,e.next=3,i();case 3:return o=e.sent,e.abrupt("return",o);case 7:return e.prev=7,e.t0=e.catch(0),e.next=11,n.last_pass();case 11:s=e.sent,e.t0="Inliner failed during operation: <code>".concat(t,"</code><br />\nThe most recent pass was: <code>").concat(s,"</code><br />\nThe Python error was:<pre>").concat(e.t0,"</pre>"),Ji().show_error(e.t0);case 14:case"end":return e.stop()}}),e,null,[[0,7]])})));return function(t,n,i){return e.apply(this,arguments)}}(),lr=function(){var e=k.a.useContext(sr),t=function(){var t=s()(r.a.mark((function t(){var n,i,o,s;return r.a.wrap((function(t){for(;;)switch(t.prev=t.next){case 0:return n=Ji().get_and_insert(),i=n.text,n.cell_id,o=n.set_cell_text,s=new ir(o,e),t.next=4,ar("create_new",s,(function(){return s.setup(i)}));case 4:e.add_state(s),ar("refresh_target_suggestions",s,(function(){return s.refresh_target_suggestions()}));case 6:case"end":return t.stop()}}),t)})));return function(){return t.apply(this,arguments)}}();return k.a.createElement("button",{className:"inline-btn inline-create",onClick:t},k.a.createElement("i",{className:"fa fa-plus"}))},ur=fe((function(){var e=k.a.useContext(sr).current_state;return k.a.createElement("button",{className:"inline-btn inline-undo",onClick:function(){return ar("undo",e,(function(){return e.undo()}))}},k.a.createElement("i",{className:"fa fa-undo"}))})),cr=fe((function(){var e=k.a.useContext(sr).current_state,t=Array.from(e.target_suggestions.entries()).map((function(e){var t=B()(e,2),n=t[0],i=t[1];return{label:"".concat(n," (").concat(i,")"),value:n}}));return k.a.createElement("div",null,k.a.createElement("div",{className:"inline-targets"},e.targets.length>0?e.targets.map((function(e){return k.a.createElement("div",{key:e},"- ",e)})):k.a.createElement("span",{className:"inline-targets-missing"},"No inline targets added")),k.a.createElement("h3",null,"Suggestions",k.a.createElement("button",{className:"inline-refresh-targets",onClick:function(){return ar("refresh_target_suggestions",e,(function(){return e.refresh_target_suggestions()}))}},k.a.createElement("i",{className:"fa fa-refresh"}))),k.a.createElement("div",{className:"inline-target-suggestions"},k.a.createElement(Vi,{options:t,value:"",styles:{menu:function(e){return function(e){for(var t=1;t<arguments.length;t++){var n=null!=arguments[t]?arguments[t]:{};t%2?or(Object(n),!0).forEach((function(t){c()(e,t,n[t])})):Object.getOwnPropertyDescriptors?Object.defineProperties(e,Object.getOwnPropertyDescriptors(n)):or(Object(n)).forEach((function(t){Object.defineProperty(e,t,Object.getOwnPropertyDescriptor(n,t))}))}return e}({},e,{fontFamily:'"Source Sans Pro", monospace'})}},placeholder:"Suggestions...",onChange:function(t){var n=t.value;e.targets.push(n),e.target_suggestions.delete(n)}})))})),hr=fe((function(){var e=k.a.useContext(sr).current_state;return k.a.createElement("div",{className:"inline-passes"},k.a.createElement("div",null,k.a.createElement("button",{className:"inline-btn",onClick:function(){return ar("autoschedule",e,(function(){return e.autoschedule()}))}},"Autoschedule")),k.a.createElement("div",null,k.a.createElement("button",{className:"inline-btn",onClick:function(){return ar("autoschedule_noinline",e,(function(){return e.autoschedule_noinline()}))}},"Autoschedule (no inline)")),["inline","deadcode","copy_propagation","clean_imports","expand_self","lifetimes","simplify_varargs","remove_suffixes","expand_tuples","partial_eval","array_index"].map((function(t){var n=t.replace("_"," ");return n=n.charAt(0).toUpperCase()+n.slice(1),k.a.createElement("div",{key:t},k.a.createElement("button",{className:"inline-btn",onClick:function(){return ar(t,e,(function(){return e.run_pass(t)}))}},n))})))})),fr=fe((function(){return k.a.useContext(sr).show_spinner?k.a.createElement("div",{className:"inline-spinner"}):null})),dr=fe(ji=function(e){function t(){return f()(this,t),m()(this,b()(t).apply(this,arguments))}return w()(t,e),p()(t,[{key:"componentDidMount",value:function(){this._set_diff()}},{key:"componentDidUpdate",value:function(){this._set_diff()}},{key:"_set_diff",value:function(){var e=this.props.state;if(e){var t=e.program_history.length;if(t>=2){var n=e.program_history[t-2],i=e.program_history[t-1];this.diff&&this.diff.destroy(),this.diff=new lt({element:this._el,mode:"ace/mode/python",left:{content:n,copyLinkEnabled:!1,editable:!1},right:{content:i,copyLinkEnabled:!1,editable:!1}})}}}},{key:"componentWillUnmount",value:function(){this.diff&&this.diff.destroy()}},{key:"render",value:function(){var e=this,t=this.props.state;return t&&t.program_history.length,k.a.createElement("div",{className:"inline-diff-panel",ref:function(t){e._el=t}})}}]),t}(k.a.Component))||ji,pr=fe((function(){var e=k.a.useState(!1),t=B()(e,2),n=t[0],i=t[1],r=k.a.useContext(sr).current_state;return k.a.createElement("span",null,k.a.createElement("button",{onClick:function(){return i(!n)}},k.a.createElement("i",{className:"fa fa-info"})),n?k.a.createElement(dr,{state:r}):null)})),gr=fe((function(e){var t=k.a.useContext(sr).current_state;return k.a.createElement("div",{className:"inliner ".concat(e.className)},k.a.createElement("h1",null,"Inliner"),k.a.createElement(lr,null),k.a.createElement(ur,null),k.a.createElement(pr,null),t?k.a.createElement("div",null,k.a.createElement("hr",null),k.a.createElement("div",null,k.a.createElement("h2",null,"Targets"),k.a.createElement(cr,null)),k.a.createElement("hr",null),k.a.createElement("div",null,k.a.createElement("h2",null,"Passes"),k.a.createElement(hr,null))):null,k.a.createElement(fr,null))}));n.d(t,"load_ipython_extension",(function(){return br}));var mr=function(e){function t(e){var n;f()(this,t),(n=m()(this,b()(t).call(this)))._state=e;return["select.Cell","set_dirty.Notebook"].forEach((function(e){C.a.notebook.events.on(e,(function(){return e=C.a.notebook.get_selected_cell(),void(n._state.current_cell=e.cell_id);var e}))})),n}return w()(t,e),p()(t,[{key:"show_error",value:function(e){F.a.modal({title:"Inliner error",body:$(e),buttons:{Dismiss:{}}})}},{key:"get_and_insert",value:function(){var e=C.a.notebook.get_selected_cell(),t=C.a.notebook.insert_cell_below();return C.a.notebook.select_next(),this._state.current_cell=t.cell_id,{text:e.get_text(),cell_id:t.cell_id,set_cell_text:function(e){return t.set_text(e)}}}},{key:"execute",value:function(e,t){return e=e.trim(),new Promise((function(n,i){C.a.notebook.kernel.execute(e,{shell:{reply:function(e){t||("error"==e.content.status?(console.error(e),i(format_trace(e))):n(e))}},iopub:{output:function(e){t&&("error"==e.msg_type?(console.error(e),i(format_trace(e))):setTimeout((function(){n(e.content.text)}),100))}}})}))}}]),t}(Xi),vr=function(e){function t(e){var n,i;return f()(this,t),n=m()(this,b()(t).call(this,e)),c()(l()(n),"state",{show:!0}),n._state=new rr,i=new mr(n._state),Li=i,n}return w()(t,e),p()(t,[{key:"componentDidMount",value:function(){var e=this;C.a.toolbar.add_buttons_group([C.a.keyboard_manager.actions.register({help:"Inliner",icon:"fa-bolt",handler:function(){e.setState({show:!e.state.show})}},"toggle-inliner","inliner")])}},{key:"render",value:function(){var e={display:this.state.show?"block":"none"};return k.a.createElement("div",{style:e},k.a.createElement(sr.Provider,{value:this._state},k.a.createElement(gr,{className:"inliner-notebook"})))}}]),t}(k.a.Component);function br(){return yr.apply(this,arguments)}function yr(){return(yr=s()(r.a.mark((function e(){var t;return r.a.wrap((function(e){for(;;)switch(e.prev=e.next){case 0:return e.next=2,C.a.notebook.config.loaded;case 2:t=document.createElement("div"),document.body.appendChild(t),_.a.render(k.a.createElement(vr,null),t),console.log("[Inliner] Extension initialized");case 6:case"end":return e.stop()}}),e)})))).apply(this,arguments)}}])}));
//# sourceMappingURL=notebook.js.map